/*
 * Copyright (c) 2003-2008 by Cosylab d. d.
 *
 * This file is part of CosyBeans.
 *
 * CosyBeans is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * CosyBeans is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with CosyBeans.  If not, see <http://www.gnu.org/licenses/>.
 */

package com.cosylab.application;

import java.io.BufferedInputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.InputStream;
import java.net.URL;
import java.util.Properties;
import java.util.logging.ConsoleHandler;
import java.util.logging.Level;
import java.util.logging.Logger;



/**
 * This class searches for configuration file in several location in well defined order.
 * 
 * @author ikriznar
 *
 */
public final class ConfigurationLoader {
	
	private ConfigurationLoader() {
		super();
	}
	
	/**
	 * Tries to return class loader from provided class. If not possible, 
	 * tries to return default class loader. If no usable class loader is found, then <code>null</code>
	 * is returned.
	 * 
	 * @param c class for witch class loader is searched
	 * @return class loader instance or <code>null</code>
	 */
	public static ClassLoader findClassLoader(Class<?> c) {
		Logger l= Logger.getLogger(ConfigurationLoader.class.getName());
		ClassLoader cl=null;
		
		if (c == null) {
			c= ConfigurationLoader.class;
		}
			
		try {
			cl= c.getClassLoader();
		} catch (Exception e) {
			l.log(Level.FINEST,"Failed to load class loader for '"+c.getName()+"'",e);
		}
		
		if (cl!=null) {
			return cl;
		}
		
		try {
			cl= ClassLoader.getSystemClassLoader();
		} catch (Exception e) {
			l.log(Level.FINEST,"Failed to load system class loader for '"+c.getName()+"'",e);
		}
		
		return cl;
	}
	
	/**
	 * Loads the configuration from the file <code>name</code>.properties
	 * @param c the class loader supplier
	 * @param name the name of the file (without extension)
	 * @return properties as read from the file
	 */
	public static Properties loadConfiguration(Class<?> c, String name) {
		return loadConfigurationFile(c,name+".properties");
	}
	
	/**
	 * Loads configuration from the file <code>name</code>.properties using
	 * default class loader.
	 * 
	 * @param name the name of the file (without extension)
	 * @return properties as read from the file
	 */
	public static Properties loadConfiguration(String name) {
		return loadConfigurationFile(null,name+".properties");
	}
	
	/**
	 * Loads the configuration from the file with the given name using
	 * the supplied class's classloader.
	 * 
	 * @param c class loader source
	 * @param propFileName the name of the properties file (with extension)
	 * @return properties as read from the file
	 */
	public static Properties loadConfigurationFile(Class<?> c, String propFileName) {
		Logger l= Logger.getLogger(ConfigurationLoader.class.getName());
		
		Properties p= new Properties();
		
		/*
		 * Tries to load from classpath (eg. configuration in jar)
		 */
		
		ClassLoader cl = findClassLoader(c);
		if (cl!=null) {
			
			URL url= cl.getResource(".");
			InputStream is=null;
			File f= null;
			try {
				f= new File(url.toURI());
				f= new File(f,propFileName);
				if (f.exists()) {
					is= new BufferedInputStream(new FileInputStream(f));
					p.load(is);
					is.close();
					l.log(Level.CONFIG,"Configuration '"+propFileName+"' loaded from file '"+f.getAbsolutePath()+"'.");
				} else {
					l.log(Level.FINEST,"Configuration '"+propFileName+"' does not exist as file '"+f.getAbsolutePath()+"'.");
				}
			} catch (Exception e) {
				l.log(Level.FINEST,"Failed to load '"+propFileName+"' from file '"+f+"'.",e);
			} finally {
				try {
					if (is!=null) is.close();
				} catch (Exception e) {
					// ignore
				}
			}
			
			is= cl.getResourceAsStream(propFileName);
			l.log(Level.FINE,"Configuration '"+propFileName+"' not found on classpath.");
			if (is!=null) {
				try {
					p.load(is);
					l.log(Level.CONFIG,"Configuration '"+propFileName+"' loaded from classpath.");
					is.close();
				} catch (Exception e) {
					l.log(Level.FINEST,"Failed to load '"+propFileName+"' from classpath.",e);
				} finally {
					try {
						if (is!=null) is.close();
					} catch (Exception e) {
						// ignore
					}
				}
			}
		}
		
		/*
		 * Tries to load from current directory.
		 */
		File f= new File(System.getProperty("user.dir"));
		InputStream is=null;
		try {
			if (f.exists()) {
				f= new File(f,propFileName);
				if (f.exists()) {
					is= new BufferedInputStream(new FileInputStream(f));
					p.load(is);
					is.close();
					l.log(Level.CONFIG,"Configuration '"+propFileName+"' loaded from file '"+f.getAbsolutePath()+"'.");
				}
			} else {
				l.log(Level.FINEST,"Configuration '"+propFileName+"' does not exist as file '"+f.getAbsolutePath()+"'.");
			}
		} catch (Exception e) {
			l.log(Level.FINEST,"Failed to load '"+propFileName+"' from file '"+f.getAbsolutePath()+"'.",e);
		} finally {
			try {
				if (is!=null) is.close();
			} catch (Exception e) {
				// ignore
			}
		}
		
		
		return p;
	}
	
	public static void main(String[] args) {
		Logger l= Logger.getLogger(ConfigurationLoader.class.getName());
		l.setLevel(Level.ALL);
		ConsoleHandler c= new ConsoleHandler();
		c.setLevel(Level.ALL);
		l.addHandler(c);
		
		Properties p= ConfigurationLoader.loadConfiguration("project");
		c.flush();
		p.list(System.out);
		
	}

}
