/*
 * Copyright (c) 2003-2008 by Cosylab d. d.
 *
 * This file is part of CosyBeans.
 *
 * CosyBeans is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * CosyBeans is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with CosyBeans.  If not, see <http://www.gnu.org/licenses/>.
 */

package com.cosylab.application;

/**
 * Interface defining an object capable of managing simple plugins. The methods
 * of this interface allow the user to install a plug in, register plugin
 * listener (for receiving containment and state changes regarding the
 * plugins) and browse the list of installed plugins. Note that in principle
 * the plugin manager can be an object completely separate from the
 * <code>CosyPanel</code> instance, which is affected by installed plugins.
 * Moreover, there could be one manager per a large  number of
 * <code>CosyPanels</code> (e.g. one manager per application), which could
 * reduce the memory footprint of the application.
 * 
 * <p>
 * All methods except for the <code>installPlugIn</code> have straightforward
 * behavior. The implementation must be thread safe, especially the
 * combination of install and browse methods.
 * </p>
 * 
 * <p>
 * A plugin is added to the manager after it has been installed successfully. A
 * plugin is removed from the manager after a destroy has been attempted.
 * Adding and destroying the plugin are the responsibilities of the manager.
 * </p>
 * 
 * <p>
 * Listeners wanting to receive the updates about the plugin state should
 * listen primarily to <code>PlugInEvent</code>s and not
 * <code>CosyPanelEvent</code>s;  in case of failed plugin installation, the
 * plugin (in the role of cosy component) will first be added and then removed
 * from the cosy panel (two event notifications).  However,
 * <code>PlugInEvent</code> will be raised by the plugin manager <b>only
 * if</b> the installation of the plugin completes successfully.
 * </p>
 * 
 * <p>
 * Calling <code>destroy()</code> on manager also destroys all contained
 * plugins.  Plugins GUI components are first removed from
 * <code>CosyApplicationPanel</code>, then <code>destroyed</code> is called on
 * plugin and finally they are removed from cosy containment tree- parent is
 * set to null.
 * </p>
 */
public interface PlugInManager
{
	/**
	 * Installs a new plugin into this manager. Installing a plugin involves,
	 * in the order prescribed below:
	 * 
	 * <ul>
	 * <li>
	 * Checking if the plugin is already installed (check by class type). If
	 * so, return.
	 * </li>
	 * <li>
	 * Instantiate the plugin with the default no-arg constructor.
	 * </li>
	 * <li>
	 * Add the plugin to the <code>CosyPanel</code> instance, blocking on
	 * <code>getHierarchyLock</code>. Within the same guarded section invoke
	 * <code>setCosyPanelParent</code> on the plugin, passing the cosy panel
	 * itself as a parameter.
	 * </li>
	 * <li>
	 * Invoke <code>installPlugIn(this)</code> on the plugin, allowing the
	 * plugin to initialize. NOTE: if this method fails by throwing an
	 * exception, remove the plugin as cosy component from the cosy panel.
	 * </li>
	 * <li>
	 * Inform the plugin listeners that a new plugin was installed, if no
	 * exception was thrown in the preceding call.
	 * </li>
	 * </ul>
	 * 
	 * <p>
	 * The method must actively guard against the cyclic references among
	 * plugins.  Therefore, if, during installation of plug A, plug B is being
	 * installed, which  requests plug A installation (cycle), the plugin
	 * manager must raise a <code>PlugInException</code>.
	 * </p>
	 *
	 * @param plugType the Class type marking the plug to be installed. The
	 *        plug must implement the <code>PlugIn</code> interface, otherwise
	 *        a <code>PlugInException</code> must be raised.
	 */
	public void installPlugIn(Class<? extends PlugIn> plugType) throws PlugInException;

	/**
	 * Removes plugin with provided type, if installed. First destroy() is
	 * called, plugin must remove itself from CosyApplicationPanel.
	 *
	 * @param plugType to be removed
	 *
	 * @throws PlugInException
	 */
	public void removePlugIn(Class<? extends PlugIn> plugType);

	/**
	 * Returns a list of all plugins installed by this plugin manager.
	 *
	 * @return PlugIn[] plugins
	 */
	public PlugIn[] getPlugIns();

	/**
	 * Returns a plugin installed with <code>plugType</code> type or
	 * <code>null</code> if such plugin has not been installed.
	 *
	 * @return PlugIn the plugin of type <code>plugType</code> or
	 *         <code>null</code>
	 */
	public <T extends PlugIn> T getPlugIn(Class<T> plugType);
	/**
	 * Returns a plugin with <code>plugType</code> type. If manager does not have recquired plugub, trys to install it first.
	 * 
	 *
	 * @return PlugIn the plugin of type <code>plugType</code> or
	 *         <code>null</code>
	 */
	public <T extends PlugIn> T acquirePlugIn(Class<T> plugType) throws PlugInException;

	/**
	 * Adds a plugin listener.
	 *
	 * @param l listener object.
	 */
	public void addPlugInListener(PlugInListener l);

	/**
	 * Remove a plugin listener.
	 *
	 * @param l listener object
	 */
	public void removePlugInListener(PlugInListener l);

	/**
	 * Returns all plugin listeners as per Java 1.4 beans specifications.
	 *
	 * @return PlugInListener[] array of currently registered listeners
	 */
	public PlugInListener[] getPlugInListeners();
	
	/**
	 * Sets the panel which is the owner of the plugin.
	 * 
	 * @param owner
	 */
	public void setOwnerPanel(ApplicationPanel owner);
	
	/**
	 * Returns the owner panel.
	 * 
	 * @return
	 */
	public ApplicationPanel getOwnerPanel();
}

/* __oOo__ */
