/*
 * Copyright (c) 2003-2008 by Cosylab d. d.
 *
 * This file is part of Java-Common.
 *
 * Java-Common is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Java-Common is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Java-Common.  If not, see <http://www.gnu.org/licenses/>.
 */

package com.cosylab.application.state.impl;

import com.cosylab.application.state.*;

import com.cosylab.util.StringUtilities;

import java.awt.Color;
import java.awt.Dimension;
import java.awt.Font;
import java.awt.Rectangle;

import java.io.IOException;
import java.io.Serializable;
import java.io.Writer;

import java.util.Iterator;
import java.util.LinkedHashMap;
import java.util.Set;


/**
 * This is the default implementation of the <code>State</code> interface. It
 * is basically extended properties. It uses the <code>LinkedHashMap</code>
 * for the intermidiate storage. This implementation should be sufficient for
 * the majority application and it is not likely that a component need to
 * implement its own state object to exchange with the
 * <code>StateKeeper</code>
 *
 * @author dvitas
 */
public class DefaultState implements State, Serializable
{
	/**
     * 
     */
    private static final long serialVersionUID = 1L;
    private static final String keyID = "ID";
	private static boolean typed = false;
	private static final String typeMark = ".type";
	protected LinkedHashMap<String, Object> map = new LinkedHashMap<>();

	/**
	 * Turn on/off storing type information about stored objects. Default is
	 * off - objects are stored without information about its type. If it is
	 * turned on, all objects are written with its type information so the
	 * function <code>getType()</code> can be used.
	 *
	 * @param typed on/off
	 */
	public static void setTyped(boolean typed)
	{
		DefaultState.typed = typed;
	}

	/**
					 *
					 */
	public DefaultState()
	{
		super();
	}
	
	private void checkKeyName(String name) {
		if (name==null) {
			throw new IllegalArgumentException("Key name can not be 'null'.");
		}
		int i= name.indexOf(' ');
		if (i>=0) {
			throw new IllegalArgumentException("Key name '"+name+"' can not contain whitespace character(s).");
		}
		i= name.indexOf('*');
		if (i>=0) {
			throw new IllegalArgumentException("Key name '"+name+"' can not contain '*' character(s).");
		}
	}

	/* (non-Javadoc)
	 * @see com.cosylab.application.state.State#putState(java.lang.String, com.cosylab.application.state.State)
	 */
	public void putState(String key, State state)
	{
		if (state != null) {
			checkKeyName(key);
			map.put(key, state);
		}
	}

	/* (non-Javadoc)
	 * @see com.cosylab.application.state.State#createChild(java.lang.String)
	 */
	public State createState(String key)
	{
		checkKeyName(key);
		State s = new DefaultState();
		map.put(key, s);

		return s;
	}

	/* (non-Javadoc)
	 * @see com.cosylab.application.state.State#getBoolean(java.lang.String)
	 */
	public boolean getBoolean(String key, boolean defaultValue)
	{
		String val = getString(key, null);

		if (val == null) {
			return defaultValue;
		}

		return Boolean.valueOf(val).booleanValue();
	}

	/* (non-Javadoc)
	 * @see com.cosylab.application.state.State#getChild(java.lang.String)
	 */
	public State getState(String key)
	{
		Object obj = map.get(key);

		if (obj == null) {
			return null;
		}

		if (!(obj instanceof State)) {
			return null;
		}

		return (State)obj;
	}

	/* (non-Javadoc)
	 * @see com.cosylab.application.state.State#getClas(java.lang.String)
	 */
	public Class<?> getClass(String key)
	{
		String value = getString(key, null);

		if (value == null) {
			return null;
		}

		try {
			return Class.forName(value);
		} catch (ClassNotFoundException e) {
			return null;
		}
	}

	/* (non-Javadoc)
	 * @see com.cosylab.application.state.State#getColor(java.lang.String)
	 */
	public Color getColor(String key)
	{
		String value = getString(key, null);

		if (value == null) {
			return null;
		}

		try {
			int rgb = (int)Long.parseLong(value, 16);
			return new Color(rgb, true);
		} catch (NumberFormatException e) {
			return null;
		}
	}

	/* (non-Javadoc)
	 * @see com.cosylab.application.state.State#getDimension(java.lang.String)
	 */
	public Dimension getDimension(String key)
	{
		// test if exists
		Object obj = map.get(key + ".width");

		if (obj == null) {
			return null;
		}

		Dimension dim = new Dimension();
		dim.width = getInt(key + ".width", 0);
		dim.height = getInt(key + ".height", 0);

		return dim;
	}

	/* (non-Javadoc)
	 * @see com.cosylab.application.state.State#getDouble(java.lang.String)
	 */
	public double getDouble(String key, double defaultValue)
	{
		Object obj = map.get(key);

		if (obj == null) {
			return defaultValue;
		}

		if (obj instanceof Number) {
			return ((Number)obj).doubleValue();
		}

		return Double.parseDouble(obj.toString());
	}

	/* (non-Javadoc)
	 * @see com.cosylab.application.state.State#getDoubleSeq(java.lang.String)
	 */
	public double[] getDoubleSeq(String key)
	{
		int length = getInt(key + ".length", -1);

		if (length == -1) {
			return null;
		}

		double[] retVal = new double[length];

		for (int i = 0; i < retVal.length; i++) {
			retVal[i] = getDouble(key + '_' + i + '_', Double.NaN);
		}

		return retVal;
	}
	
	public Color[] getColorSeq(String key)
	{
		int length = getInt(key + ".length", -1);

		if (length < 0) {
			return null;
		}

		Color[] retVal = new Color[length];

		for (int i = 0; i < retVal.length; i++) {
			retVal[i] = getColor(key + '_' + i + '_');
		}

		return retVal;
	}

	/* (non-Javadoc)
	 * @see com.cosylab.application.state.State#getFont(java.lang.String)
	 */
	public Font getFont(String key)
	{
		String value = getString(key, null);

		if (value == null) {
			return null;
		}

		Font f;

		try {
			f = Font.decode(value);
		} catch (NumberFormatException e) {
			return null;
		}

		return f;
	}

	public Font[] getFontSeq(String key)
	{
		int length = getInt(key + ".length", -1);

		if (length < 0) {
			return null;
		}

		Font[] retVal = new Font[length];

		for (int i = 0; i < retVal.length; i++) {
			retVal[i] = getFont(key + '_' + i + '_');
		}

		return retVal;
	}
	
	/* (non-Javadoc)
	 * @see com.cosylab.application.state.State#getID()
	 */
	public String getID()
	{
		return getString(keyID, null);
	}

	/* (non-Javadoc)
	 * @see com.cosylab.application.state.State#getInt(java.lang.String)
	 */
	public int getInt(String key, int defaultValue)
	{
		Object obj = map.get(key);

		if (obj == null) {
			return defaultValue;
		}

		if (obj instanceof Number) {
			return ((Number)obj).intValue();
		}

		return Integer.parseInt(obj.toString());
	}

	/* (non-Javadoc)
	 * @see com.cosylab.application.state.State#getIntSeq(java.lang.String)
	 */
	public int[] getIntSeq(String key)
	{
		int length = getInt(key + ".length", -1);

		if (length == -1) {
			return null;
		}

		int[] retVal = new int[length];

		for (int i = 0; i < retVal.length; i++) {
			retVal[i] = getInt(key + '_' + i + '_', Integer.MAX_VALUE);
		}

		return retVal;
	}

	/* (non-Javadoc)
	 * @see com.cosylab.application.state.State#getObject()
	 */
	public Object getObject(String key)
	{
		return map.get(key);
	}

	/* (non-Javadoc)
	 * @see com.cosylab.application.state.State#getType(java.lang.String)
	 */
	public Class<?> getType(String key)
	{
		String className = getString(key + typeMark, null);

		if (className == null) {
			return null;
		}

		try {
			return StringUtilities.classFromString(className);
		} catch (Exception e) {
			// no op
		}

		return null;
	}

	/* (non-Javadoc)
	 * @see com.cosylab.application.state.State#getRectangle(java.lang.String)
	 */
	public Rectangle getRectangle(String key)
	{
		// test if exists
		Object obj = map.get(key + ".x");

		if (obj == null) {
			return null;
		}

		Rectangle r = new Rectangle();
		r.x = getInt(key + ".x", 0);
		r.y = getInt(key + ".y", 0);
		r.width = getInt(key + ".width", 0);
		r.height = getInt(key + ".height", 0);

		return r;
	}

	/* (non-Javadoc)
	 * @see com.cosylab.application.state.State#getString(java.lang.String)
	 */
	public String getString(String key, String defaultValue)
	{
		Object obj = map.get(key);

		if (obj == null) {
			return defaultValue;
		}

		return obj.toString();
	}

	/* (non-Javadoc)
	 * @see com.cosylab.application.state.State#getStringSeq(java.lang.String)
	 */
	public String[] getStringSeq(String key)
	{
		int length = getInt(key + ".length", -1);

		if (length == -1) {
			return null;
		}

		String[] retVal = new String[length];

		for (int i = 0; i < retVal.length; i++) {
			retVal[i] = getString(key + '_' + i + '_', null);
		}

		return retVal;
	}

	/* (non-Javadoc)
	 * @see com.cosylab.application.state.State#keySet()
	 */
	public Set<String> keySet()
	{
		return map.keySet();
	}

	/* (non-Javadoc)
	 * @see com.cosylab.application.state.State#putBoolean(java.lang.String, boolean)
	 */
	public void putBoolean(String key, boolean value)
	{
		putString(key, Boolean.toString(value));

		if (DefaultState.typed) {
			map.put(key + typeMark, Boolean.TYPE);
		}
	}

	/* (non-Javadoc)
	 * @see com.cosylab.application.state.State#putClas(java.lang.String, java.lang.Class)
	 */
	public void putClass(String key, Class<?> value)
	{
		if (value == null) {
			return;
		}

		putString(key, value.getName());

		if (DefaultState.typed) {
			map.put(key + typeMark, value.getClass().getName());
		}
	}

	/* (non-Javadoc)
	 * @see com.cosylab.application.state.State#putColor(java.lang.String, java.awt.Color)
	 */
	public void putColor(String key, Color value)
	{
		if (value == null) {
			return;
		}

		putString(key, Integer.toHexString(value.getRGB()));

		if (DefaultState.typed) {
			map.put(key + typeMark, value.getClass().getName());
		}
	}

	/* (non-Javadoc)
	 * @see com.cosylab.application.state.State#putDimension(java.lang.String, java.awt.Dimension)
	 */
	public void putDimension(String key, Dimension value)
	{
		if (value == null) {
			return;
		}

		putInt(key + ".width", value.width);
		putInt(key + ".height", value.height);

		if (DefaultState.typed) {
			map.put(key + typeMark, value.getClass().getName());
		}
	}

	/* (non-Javadoc)
	 * @see com.cosylab.application.state.State#putDouble(java.lang.String, double)
	 */
	public void putDouble(String key, double value)
	{
		checkKeyName(key);
		map.put(key, new Double(value));

		if (DefaultState.typed) {
			map.put(key + typeMark, Double.TYPE.getName());
		}
	}

	/* (non-Javadoc)
	 * @see com.cosylab.application.state.State#putDoubleSeq(java.lang.String, double[])
	 */
	public void putDoubleSeq(String key, double[] value)
	{
		if (value == null) {
			return;
		}

		putInt(key + ".length", value.length);

		for (int i = 0; i < value.length; i++) {
			putDouble(key + '_' + i + '_', value[i]);
		}

		if (DefaultState.typed) {
			map.put(key + typeMark, value.getClass().getName());
		}
	}
	
	/* (non-Javadoc)
	 * @see com.cosylab.application.state.State#putDoubleSeq(java.lang.String, double[])
	 */
	public void putColorSeq(String key, Color[] value)
	{
		if (value == null) {
			return;
		}

		putInt(key + ".length", value.length);

		for (int i = 0; i < value.length; i++) {
			putColor(key + '_' + i + '_', value[i]);
		}

		if (DefaultState.typed) {
			map.put(key + typeMark, value.getClass().getName());
		}
	}

	/* (non-Javadoc)
	 * @see com.cosylab.application.state.State#putFont(java.lang.String, java.awt.Font)
	 */
	public void putFont(String key, Font value)
	{
		if (value == null) {
			return;
		}

		String strStyle;

		if (value.isBold()) {
			strStyle = value.isItalic() ? "bolditalic" : "bold";
		} else {
			strStyle = value.isItalic() ? "italic" : "plain";
		}

		putString(key, value.getName() + "-" + strStyle + "-" + value.getSize());

		if (DefaultState.typed) {
			map.put(key + typeMark, value.getClass().getName());
		}
	}
	
	public void putFontSeq(String key, Font[] value)
	{
		if (value == null) {
			return;
		}

		putInt(key + ".length", value.length);

		for (int i = 0; i < value.length; i++) {
			putFont(key + '_' + i + '_', value[i]);
		}

		if (DefaultState.typed) {
			map.put(key + typeMark, value.getClass().getName());
		}
	}

	/* (non-Javadoc)
	 * @see com.cosylab.application.state.State#putInt(java.lang.String, int)
	 */
	public void putInt(String key, int value)
	{
		checkKeyName(key);
		map.put(key, new Integer(value));

		if (DefaultState.typed) {
			map.put(key + typeMark, Integer.TYPE.getName());
		}
	}

	/* (non-Javadoc)
	 * @see com.cosylab.application.state.State#putIntSeq(java.lang.String, int[])
	 */
	public void putIntSeq(String key, int[] value)
	{
		if (value == null) {
			return;
		}

		putInt(key + ".length", value.length);

		for (int i = 0; i < value.length; i++) {
			putInt(key + '_' + i + '_', value[i]);
		}

		if (DefaultState.typed) {
			map.put(key + typeMark, value.getClass().getName());
		}
	}

	/* (non-Javadoc)
	 * @see com.cosylab.application.state.State#putRectangle(java.lang.String, java.awt.Rectangle)
	 */
	public void putRectangle(String key, Rectangle value)
	{
		if (value == null) {
			return;
		}

		putInt(key + ".x", value.x);
		putInt(key + ".y", value.y);
		putInt(key + ".width", value.width);
		putInt(key + ".height", value.height);

		if (DefaultState.typed) {
			map.put(key + typeMark, value.getClass().getName());
		}
	}

	/* (non-Javadoc)
	 * @see com.cosylab.application.state.State#putString(java.lang.String, java.lang.String)
	 */
	public void putString(String key, String value)
	{
		if (value == null) {
			return;
		}

		checkKeyName(key);
		map.put(key, value);

		if (DefaultState.typed) {
			map.put(key + typeMark, value.getClass().getName());
		}
	}

	/* (non-Javadoc)
	 * @see com.cosylab.application.state.State#putStringSeq(java.lang.String, java.lang.String[])
	 */
	public void putStringSeq(String key, String[] value)
	{
		if (value == null) {
			return;
		}

		putInt(key + ".length", value.length);

		for (int i = 0; i < value.length; i++) {
			putString(key + '_' + i + '_', value[i]);
		}

		if (DefaultState.typed) {
			map.put(key + typeMark, value.getClass().getName());
		}
	}

	/* (non-Javadoc)
	 * @see com.cosylab.application.state.State#setID()
	 */
	public void setID(String ID)
	{
		putString(keyID, ID);
	}

    private String processValueString(String inString) {
        if (inString==null) return "null";
        int len=inString.length();
        StringBuffer retBuf=new StringBuffer();
        for (int i = 0; i < len; i++) {
            char c=inString.charAt(i);
            if (c>127 || c=='&' || c=='"' || c=='<' || c=='>') {
                retBuf.append("&#"+((int)c)+";");
            } else {
                retBuf.append(c);
            }
        }
        return retBuf.toString();
    }
    
	private void writeState(Writer writer, String name, State state, String linePrefix)
		throws IOException
	{
		boolean hasChilds = false;
		writer.write(linePrefix+"<" + name);

		Iterator<String> iter = state.keySet().iterator();

		while (iter.hasNext()) {
			String key = iter.next();
			Object value = state.getObject(key);

			if (value instanceof State) {
				hasChilds = true;

				continue;
			}

			String aa = processValueString(value.toString());
			writer.write(" " + key + "=\"" + aa + "\" ");
		}

		if (hasChilds) {
			writer.write(">\n");
		} else {
			writer.write("/>\n");
		}

		iter = state.keySet().iterator();

		while (iter.hasNext()) {
			String key = iter.next();
			Object value = state.getObject(key);

			if (value instanceof State) {
				writeState(writer, key, (State)value, linePrefix+"  ");

				continue;
			}
		}

		if (hasChilds) {
			writer.write(linePrefix+"</" + name + ">\n");
		}
	}

	/* (non-Javadoc)
	 * @see com.cosylab.application.state.State#write(java.io.Writer)
	 */
	public void writeXML(Writer writer) throws IOException
	{
		writer.write("<?xml version=\"1.0\" standalone=\"yes\"?>\n<states>\n");
		writeState(writer, "state", this, "");
		writer.write("</states>\n");
		writer.close();
	}

	/* (non-Javadoc)
	 * @see java.lang.Object#toString()
	 */
	@Override
    public String toString()
	{
		StringBuffer sb = new StringBuffer();
		Iterator<String> iter = map.keySet().iterator();
		sb.append("ID=" + getID());
		sb.append('\n');

		while (iter.hasNext()) {
			Object key = iter.next();
			Object value = map.get(key);
			sb.append(key.toString());
			sb.append('=');
			sb.append(value.toString());
			sb.append('\n');
		}

		sb.append('\n');

		return new String(sb);
	}

	/* (non-Javadoc)
	 * @see java.lang.Object#hashCode()
	 */
	@Override
	public int hashCode() {
		final int prime = 31;
		int result = 1;
		result = prime * result + ((map == null) ? 0 : map.hashCode());
		return result;
	}

	/* (non-Javadoc)
	 * @see java.lang.Object#equals(java.lang.Object)
	 */
	@Override
	public boolean equals(Object obj) {
		if (this == obj) return true;
		if (obj == null) return false;
		if (getClass() != obj.getClass()) return false;
		DefaultState other = (DefaultState)obj;
		if (map == null) {
			if (other.map != null) return false;
		} else if (!map.equals(other.map)) return false;
		return true;
	}
	
	
}

/* __oOo__ */
