/*
 * Copyright (c) 2003-2008 by Cosylab d. d.
 *
 * This file is part of CosyBeans.
 *
 * CosyBeans is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * CosyBeans is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with CosyBeans.  If not, see <http://www.gnu.org/licenses/>.
 */

package com.cosylab.gui;

import java.io.File;
import java.util.Objects;

import javax.swing.Icon;

import com.cosylab.gui.components.util.IconHelper;

/**
 * <code>ValueIconPair</code> combines a long value with a specifi icon.
 * 
 * @author Jaka Bobnar, Cosylab
 *
 */
public class ValueIconPair implements Comparable<Object> {
	
	private Icon icon;
	private Long value;
	private String iconName;
	
	/**
	 * Constructs a new ValueIconPair with the given value and icon.
	 * 
	 * @param value
	 * @param icon
	 */
	public ValueIconPair(long value, Icon icon) {
		this(new Long(value),icon);
	}
	
	/**
	 * Constructs a new ValueIconPair with the given value and icon.
	 * The icon name is in this case the icon.toString().
	 * 
	 * @param value
	 * @param icon
	 */
	public ValueIconPair(Long value, Icon icon) {
		this.icon = icon;
		this.value = value;
		this.iconName = icon.toString();
	}
	
	/**
	 * Constructs a new ValueIconPair with the given value and path to the icon in 
	 * the classpath. 
	 * 
	 * @param value
	 * @param iconName
	 */
	public ValueIconPair(Long value, String iconName) {
		this.value = value;
		this.icon = IconHelper.createIcon(iconName);
		this.iconName = iconName;
	}
	
	/**
	 * Constructs a new ValueIconPair with the given value and the File describing
	 * the image icon. The file can point to any location on the disk as long as it
	 * is accessible. The file does not need to be in the classpath. The name of the
	 * file is the absolute path of the icon file.
	 * 
	 * @param value
	 * @param iconFile
	 */
	public ValueIconPair(Long value, File iconFile) {
		this.value = value;
		this.icon = IconHelper.createIcon(iconFile);
		this.iconName = iconFile.getAbsolutePath();
	}

	/**
	 * Returns the icon.
	 * 
	 * @return
	 */
	public Icon getIcon() {
		return icon;
	}

	/**
	 * Returns the value.
	 * 
	 * @return
	 */
	public Long getValue() {
		return value;
	}
	
	/**
	 * Returns the icon name.
	 * 
	 * @return
	 */
	public String getIconName() {
		return iconName;
	}
	
	/*
	 * (non-Javadoc)
	 * @see java.lang.Object#equals(java.lang.Object)
	 */
	@Override
    public boolean equals(Object o) {
		if (!(o instanceof ValueIconPair)) return false;
		ValueIconPair d = (ValueIconPair)o;
		if (value != null)
			return (value.equals(d.value) && d.icon.equals(icon));
		else 
			return (d.value == null && d.icon.equals(icon));
		
	}
	
	@Override
	public int hashCode() {
	    return Objects.hash(value,icon);
	}
	
	/*
	 * (non-Javadoc)
	 * @see java.lang.Object#toString()
	 */
	@Override
	public String toString() {
		return value + "= " + iconName ;
	}
	
	/* (non-Javadoc)
	 * @see java.lang.Comparable#compareTo(T)
	 */
	public int compareTo(Object o) {
		if (o instanceof ValueIconPair) {
			ValueIconPair v= (ValueIconPair)o;
			return (int)(value-v.value);
		}
		return 0;
	}
}
