/*
 * Copyright (c) 2003-2008 by Cosylab d. d.
 *
 * This file is part of CosyBeans.
 *
 * CosyBeans is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * CosyBeans is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with CosyBeans.  If not, see <http://www.gnu.org/licenses/>.
 */

package com.cosylab.gui.adapters;

import java.awt.Component;
import java.awt.Dimension;
import java.awt.GridBagConstraints;
import java.awt.GridBagLayout;
import java.awt.Insets;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;
import java.beans.Customizer;
import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;
import java.beans.PropertyVetoException;
import java.util.List;

import javax.swing.DefaultComboBoxModel;
import javax.swing.DefaultListCellRenderer;
import javax.swing.DefaultListModel;
import javax.swing.JButton;
import javax.swing.JComboBox;
import javax.swing.JFrame;
import javax.swing.JLabel;
import javax.swing.JList;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.ScrollPaneConstants;

import com.cosylab.gui.components.util.IconHelper;
import com.cosylab.gui.displayers.ConvertibleDisplayer;
import com.cosylab.gui.util.GUIUtilities;

/**
 * <code>ConverterCustomizer</code> is an implementation of the 
 * <code>java.beans.Customizer</code> which enables selection and 
 * setting of a ConverterChain. Customizer can be used together with
 * <code>{@link ConvertibleDisplayer}</code>.
 * 
 * @author <a href="mailto:jaka.bobnar@cosylab.com">Jaka Bobnar</a>
 * @version $Id: ConverterCustomizer.java,v 1.21 2008-04-22 12:31:02 jbobnar Exp $
 *
 */
public class ConverterCustomizer extends JPanel implements Customizer {
	private static final long serialVersionUID = 1L;

	private class AvailableCellRenderer extends DefaultListCellRenderer {

		private static final long serialVersionUID = 1L;

		public AvailableCellRenderer() {
			setOpaque(true);
		}

		@Override
        public Component getListCellRendererComponent(JList<?> list, Object value,
				int index, boolean isSelected, boolean cellHasFocus) {

			String name = ((Class<?>) value).getSimpleName();
			if (name != null && name.endsWith("Converter")) name = name.substring(0, name.lastIndexOf("Converter"));
			
			return super.getListCellRendererComponent(list,name,index,isSelected,cellHasFocus);
		}

	}

	private JComboBox<Class<Converter>> functionsList = null;
	
	private JList<Converter> selectedFunctionsList = null;

	private DefaultComboBoxModel<Class<Converter>> functionModel;
	
	private DefaultListModel<Converter> selectedModel;

	private JPanel buttonPanel = null;

	private JLabel functionsLabel = null;

	private JButton addButton = null, removeButton = null;

	private JButton editButton = null;

	private ConvertibleDisplayer convertibleDisplayer = null;

	protected Converter converter = null;
	
	private JScrollPane scrollPane = null;

	private Class<Converter>[] converters;

	/**
	 * Constructs a new ConverterCustomizer which offers Identity, 
	 * Multiplier, Linear, Potential, Exponential, and logarithmic converter.
	 *
	 */
	@SuppressWarnings({ "unchecked" })
	public ConverterCustomizer() {
		this(new Class[] { IdentityConverter.class, MultiplierConverter.class, LinearConverter.class,
				PotentialConverter.class, ExponentialConverter.class, LogarithmicConverter.class
				});
	}

	/**
	 * Constructs a new ConverterCustomizer which offers only the specified converters.
	 * 
	 * @param converters
	 */
	@SuppressWarnings("unchecked")
	public ConverterCustomizer(Class<? extends Converter>[] converters) {
		this.converters = (Class<Converter>[]) converters;
		initialize();
	}

	private void initialize() {
		setLayout(new GridBagLayout());
		add(getFunctionsList(), new GridBagConstraints(0, 0, 2, 1, 1, 0, GridBagConstraints.NORTHEAST, GridBagConstraints.HORIZONTAL, new Insets(7, 7, 2, 4), 1, 1));
		add(getAddButton(), new GridBagConstraints(2, 0, 1, 1, 0, 0, GridBagConstraints.NORTHWEST, GridBagConstraints.NONE, new Insets(7, 2, 2, 7), 0, 0));
		add(getScrollPane(), new GridBagConstraints(0, 1, 3, 1, 1, 1, GridBagConstraints.WEST, GridBagConstraints.BOTH, new Insets(4, 7, 4, 7), 1, 1));
		add(getButtonPanel(), new GridBagConstraints(0, 2, 3, 1, 1, 0, GridBagConstraints.NORTHWEST, GridBagConstraints.HORIZONTAL, new Insets(2, 7, 4, 7), 0, 0));
		add(getFunctionsLabel(), new GridBagConstraints(0, 3, 3, 1, 1, 0, GridBagConstraints.SOUTHWEST, GridBagConstraints.HORIZONTAL, new Insets(4, 7, 7, 7), 0, 0));
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see java.beans.Customizer#setObject(java.lang.Object)
	 */
	public void setObject(Object bean) {
		if (!(bean instanceof ConvertibleDisplayer)) {
			throw new IllegalArgumentException("Bean " + bean.getClass()
					+ " must implement ConvertibleDisplayer!");
		}
		convertibleDisplayer = (ConvertibleDisplayer) bean;

		if (convertibleDisplayer.getConverter() != null) {
			setConverter(convertibleDisplayer.getConverter());
		}

		convertibleDisplayer.addPropertyChangeListener(
				ConvertibleDisplayer.CONVERTER_PROPERTY, new PropertyChangeListener() {
					public void propertyChange(PropertyChangeEvent evt) {
						if (!ConvertibleDisplayer.CONVERTER_PROPERTY.equals(evt
								.getPropertyName()))
							return;
						setConverter(convertibleDisplayer.getConverter());
					}
				});
	}

	private void applyUsedConverters(Converter... converters) {
		selectedModel.clear();
		for (Converter c : converters) {
			if (c != null)
				selectedModel.addElement(c);
		}

		applySelectedFunctions();
	}

	private JPanel getButtonPanel() {
		if (buttonPanel == null) {
			buttonPanel = new JPanel();
			buttonPanel.setLayout(new GridBagLayout());
//			buttonPanel.setSize(new Dimension(81, 72));
			buttonPanel.add(getEditButton(), new GridBagConstraints(0, 0, 1, 1,
					1, 1, GridBagConstraints.WEST, GridBagConstraints.NONE,
					new Insets(0, 0, 0, 5), 0, 0));
			buttonPanel.add(getRemoveButton(), new GridBagConstraints(1, 0, 1,
					1, 1, 1, GridBagConstraints.EAST,
					GridBagConstraints.NONE, new Insets(0, 5, 0, 0), 0, 0));
			buttonPanel.setMaximumSize(new Dimension(100, 100));
			
		}
		return buttonPanel;
	}
	
	private JButton getEditButton() {
		if (editButton == null) {
			editButton = new JButton("Edit");
			editButton.setEnabled(false);
			editButton.setToolTipText("Edit function");
//			GUIUtilities.setFixedSize(editButton,new Dimension(100, 22));
			editButton.addActionListener(new ActionListener() {
				public void actionPerformed(ActionEvent e) {
					int i = selectedFunctionsList.getSelectedIndex();
					if (i >= 0)
						edit(selectedModel.get(i));
					refreshSelectedFunctionsList();
					applyFunctionsText();
				}
			});
		}

		return editButton;
	}

	private JButton getAddButton() {
		if (addButton == null) {
			addButton = new JButton("Add", IconHelper.createIcon("icons/navigation/Plus16.gif"));
			addButton.setToolTipText("Add function");
			GUIUtilities.setFixedSize(addButton,new Dimension(100, 22));
			addButton.addActionListener(new ActionListener() {
				public void actionPerformed(ActionEvent e) {
					addToSelectedList();
				}
			});
		}

		return addButton;
	}

	/**
	 * Returns the converter.
	 * 
	 * @return
	 */
	public Converter getConverter() {
		return converter;
	}

	/**
	 * Sets converter to this customizer.
	 * 
	 * @param c
	 */
	public void setConverter(Converter c) {
		if (converter != null && converter.equals(c) || (converter == null && c == null))
			return;
		Converter old = converter;
		converter = c;
		if (c instanceof ConverterChain) {
			applyUsedConverters(((ConverterChain) c).getConverters());
		} else {
			applyUsedConverters(c);
		}

		firePropertyChange(ConvertibleDisplayer.CONVERTER_PROPERTY, old, converter);
	}

	private JLabel getFunctionsLabel() {
		if (functionsLabel == null) {
			functionsLabel = new JLabel();
			Dimension d = new Dimension(200, 21);
			functionsLabel.setMaximumSize(d);
			functionsLabel.setPreferredSize(d);
			functionsLabel.setMinimumSize(new Dimension(100, 10));
		}

		return functionsLabel;
	}

	private JComboBox<Class<Converter>> getFunctionsList() {
		if (functionsList == null) {
			functionsList = new JComboBox<>();
			functionModel = new DefaultComboBoxModel<>();
			for (Class<Converter> c : converters) {
				functionModel.addElement(c);
			}
			functionsList.setModel(functionModel);
			functionsList.setRenderer(new AvailableCellRenderer());
			GUIUtilities.setFixedSize(functionsList,new Dimension(100, 21));
			functionsList.setSelectedIndex(0);
			functionsList.addMouseListener(new MouseAdapter() {
				@Override
                public void mouseClicked(MouseEvent e) {
					if (e.getClickCount() == 2 && functionsList.isEnabled()) {
						addToSelectedList();
					}
				}
			});
		}

		return functionsList;
	}

	private JButton getRemoveButton() {
		if (removeButton == null) {
			removeButton = new JButton("Remove", IconHelper.createIcon("icons/navigation/Minus16.gif"));
			removeButton.setToolTipText("Remove function");
//			GUIUtilities.setFixedSize(removeButton,new Dimension(100, 22));
			removeButton.addActionListener(new ActionListener() {
				public void actionPerformed(ActionEvent e) {
					removeFromSelectedList();
				}
			});
		}

		return removeButton;
	}

	private void removeFromSelectedList() {
		List<Converter> selected = getSelectedFunctionsList().getSelectedValuesList();
		if (selected != null) {
			for (Object o : selected) {
				selectedModel.removeElement(o);
			}
		}

		applySelectedFunctions();
	}

	@SuppressWarnings("unchecked")
    private void addToSelectedList() {
		Class<Converter> conv = (Class<Converter>)getFunctionsList().getSelectedItem();
		if (conv == null)
			return;
		Converter c = getConverterFromClass(conv);
		
		if (c != null) {
			if (!checkEditable(c) || edit(c)) {
				selectedModel.addElement(c);
			}
		}
		applySelectedFunctions();

	}

	private void applySelectedFunctions() {
		int k = selectedModel.size();
		if (k == 0) {
			converter = null;
		} else if (k == 1) {
			converter = selectedModel.get(0);
		} else {
			Converter[] c = new Converter[k];
			for (int i = 0; i < k; i++) {
				c[i] = selectedModel.get(i);
			}
			try {
				converter = new ConverterChain(c);
			} catch (PropertyVetoException e) {
				e.printStackTrace();
			}
		}

		applyFunctionsText();
		try {
			if (this.convertibleDisplayer != null)
				convertibleDisplayer.setConverter(converter);
		} catch (PropertyVetoException e) {
			e.printStackTrace();
		}

		firePropertyChange(ConvertibleDisplayer.CONVERTER_PROPERTY, null, converter);
	}

	private void applyFunctionsText() {
		String s = "INPUT";
		int k = selectedModel.size();
		for (int i = 0; i < k; i++) {
			s = composeFunction(selectedModel.get(i), s);
		}
		
		s = "OUTPUT = " + s;

		getFunctionsLabel().setText(s);
		getFunctionsLabel().setToolTipText(s);
	}

	private Converter getConverterFromClass(Class<Converter> conv) {
		try {
			return conv.getConstructor().newInstance();
		} catch (Exception e) {
			e.printStackTrace();
		}
		return null;
	}

	private JList<Converter> getSelectedFunctionsList() {
		if (selectedFunctionsList == null) {
			selectedFunctionsList = new JList<>();
			selectedModel = new DefaultListModel<>();
			selectedFunctionsList.setModel(selectedModel);
			selectedFunctionsList.addMouseListener(new MouseAdapter() {
				@Override
                public void mouseClicked(MouseEvent e) {
					if (e.getClickCount() == 2
							&& selectedFunctionsList.isEnabled()) {
						edit(selectedModel.get(getSelectedFunctionsList().getSelectedIndex()));
					} else if (getSelectedFunctionsList().getSelectedIndex() != -1){
						if (checkEditable(selectedModel.get(getSelectedFunctionsList().getSelectedIndex())))
							getEditButton().setEnabled(true);
						else
							getEditButton().setEnabled(false);
					}
				}
			});
		}
		
		return selectedFunctionsList;
	}
	
	private JScrollPane getScrollPane(){
		if (scrollPane == null) {
			scrollPane = new JScrollPane(getSelectedFunctionsList(), ScrollPaneConstants.VERTICAL_SCROLLBAR_AS_NEEDED, JScrollPane.HORIZONTAL_SCROLLBAR_NEVER);
			scrollPane.setPreferredSize(new Dimension(200,200));
		}

		return scrollPane;
	}
	
	
	/**
	 * Refreshes list of selected functions.
	 */
	public void refreshSelectedFunctionsList(){
		getSelectedFunctionsList().repaint();
		applyFunctionsText();
	}

	/**
	 * Composes the function string displayed in the GUI.
	 * 
	 * @param c
	 * @param input
	 * @return
	 */
	protected String composeFunction(Converter c, String input) {
		return ConverterUtilities.composeFunctionString(c, input);
	}

	/**
	 * Check if converter is editable.
	 * 
	 * @param converter
	 */
	protected boolean checkEditable(Converter converter) {
		return ConverterUtilities.isEditable(converter);
	}
	
	/**
	 * Displays editor dialog to edit converter. 
	 * 
	 * @param converter
	 * @return returns true if editing was confirmed, false if it was cancelled 
	 */
	protected boolean edit(Converter converter) {
		return ConverterUtilities.showEditorDialog(this, converter);
	}

	public static void main(String[] args) {
		JFrame f = new JFrame("ConverterCustomizer");
		f.add(new ConverterCustomizer());
		f.pack();
		f.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
		f.setVisible(true);
	}
}
