/*
 * Copyright (c) 2003-2008 by Cosylab d. d.
 *
 * This file is part of CosyBeans.
 *
 * CosyBeans is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * CosyBeans is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with CosyBeans.  If not, see <http://www.gnu.org/licenses/>.
 */

package com.cosylab.gui.adapters;

import java.util.Map;

import com.cosylab.gui.displayers.DataConsumer;
import com.cosylab.gui.displayers.DataState;
import com.cosylab.gui.displayers.DoubleConsumer;
import com.cosylab.util.CommonException;


/**
 * <code>DataInterceptor</code> filters data that belongs to DoubleConsumer.
 * <code>DataInterceptor</code> saves a certain amount of data (defined with
 * timeInterval in constructor). After that time elapses average value and
 * standard deviation are calculated. Interceptor than  updates consumer with
 * four values: average value at the begining and end of the interval that all
 * data belongs too, min and max value on this interval.
 * 
 * <p>
 * <code>DataInterceptor</code> can be used with chart consumers when great
 * ammount of data needs to be displayed. 
 * </p> 
 *
 * @author <a href="mailto:jaka.bobnar@cosylab.com">Jaka Bobnar</a>
 * @version $Id: DataInterceptor.java,v 1.6 2008-04-22 12:31:02 jbobnar Exp $
 *
 * @since VERSION
 */
public class DataInterceptor implements DoubleConsumer
{
	/**
	 * <code>Value</code> represents a single point for the consumer. It holds
	 * timestamp and value of the point.
	 *
	 * @author <a href="mailto:jaka.bobnar@cosylab.com">Jaka Bobnar</a>
	 * @version $Id: DataInterceptor.java,v 1.6 2008-04-22 12:31:02 jbobnar Exp $
	 *
	 * @since VERSION
	 */
	class Value
	{
		double value;
		long timestamp;

		/**
		 * Creates a new Value object.
		 *
		 * @param timestamp timestamp of the point
		 * @param value value of the point
		 */
		public Value(long timestamp, double value)
		{
			this.value = value;
			this.timestamp = timestamp;
		}
	}

	/** Consumer that communicates with displayer. */
	private DoubleConsumer consumer;

	/** Time interval that average and deviation are calculated in. */
	private long timeInterval;
	private double deviation;
	private double average;
	private int numberOfValues;
	private double sum;
	private double sumSquare;
	private long startTime;
	private long lastTime;
	private Value maxValue;
	private Value minValue;

	/**
	 * Default constructor.
	 */
	public DataInterceptor()
	{
		this.timeInterval = 0;
		reset();
	}

	/**
	 * Constructor.
	 *
	 * @param consumer consumers that sends data to displayer
	 * @param timeInterval time interval in millis
	 */
	public DataInterceptor(DoubleConsumer consumer, long timeInterval)
	{
		this.consumer = consumer;
		this.timeInterval = timeInterval;
		reset();
	}

	/**
	 * Sets the time interval for calculating average and deviation.
	 *
	 * @param timeInterval
	 */
	public void setTimeInterval(long timeInterval)
	{
		this.timeInterval = timeInterval;
		reset();
	}

	/**
	 * Returns the time interval set for this interceptor.
	 *
	 * @return time interval
	 */
	public long getTimeInterval()
	{
		return timeInterval;
	}

	/**
	 * Returns the value of standard deviation for the last dispatched data.
	 *
	 * @return standard deviation
	 */
	public double getDeviation()
	{
		return deviation;
	}

	/**
	 * Returns the average value on the interval of the last dispatched data.
	 *
	 * @return average
	 */
	public double getAverage()
	{
		return average;
	}

	/* (non-Javadoc)
	 * @see com.cosylab.gui.displayers.DoubleConsumer#updateValue(long, double)
	 */
	public void updateValue(long timestamp, double value)
		throws CommonException
	{
		if (numberOfValues == 0) {
			startTime = timestamp;
			maxValue = new Value(timestamp, value);
			minValue = new Value(timestamp, value);
		}

		sumSquare += value*value;
		sum += value;
		numberOfValues++;

		if (timestamp > startTime + timeInterval) {
			lastTime = timestamp;
			dispatch();
		}
	}

	private void dispatch() throws CommonException
	{
		average = sum / numberOfValues;
		// calculate deviation on the interval
		deviation = Math.sqrt(sumSquare/numberOfValues - average*average);

		// update consumer
		consumer.updateValue(startTime, average);
		if (maxValue.timestamp <= minValue.timestamp) {
		    consumer.updateValue(maxValue.timestamp, maxValue.value);
		    consumer.updateValue(minValue.timestamp, minValue.value);
		} else {
		    consumer.updateValue(minValue.timestamp, minValue.value);
		    consumer.updateValue(maxValue.timestamp, maxValue.value);
		}
		consumer.updateValue(lastTime, average);
		
		reset();
	}

	private void reset()
	{
		sum = 0;
		numberOfValues = 0;
		sumSquare = 0;
		maxValue = null;
		minValue = null;
	}

	/**
	 * Sets consumer of which this <code>DataInterceptor</code> intercepts
	 * data.
	 *
	 * @param newConsumer
	 */
	public void setConsumer(DoubleConsumer newConsumer)
	{
		this.consumer = newConsumer;
		reset();
	}

	/* (non-Javadoc)
	 * @see com.cosylab.gui.displayers.DataConsumer#getDataConsumer(java.lang.Class)
	 */
	public DataConsumer getDataConsumer(Class type)
	{
		if (type.isAssignableFrom(this.getClass())) {
			return this;
		}

		return null;
	}

	/* (non-Javadoc)
	 * @see com.cosylab.gui.displayers.DataConsumer#getDefaultDataConsumer()
	 */
	public DataConsumer getDefaultDataConsumer()
	{
		return this;
	}

	/* (non-Javadoc)
	 * @see com.cosylab.gui.displayers.DataConsumer#updateDataState(com.cosylab.gui.displayers.DataState)
	 */
	public void updateDataState(DataState datastate)
	{

	}

	/* (non-Javadoc)
	 * @see com.cosylab.gui.displayers.DataConsumer#setCharacteristics(java.util.Map)
	 */
	public void setCharacteristics(Map characteristics)
	{
		consumer.setCharacteristics(characteristics);
	}

	/* (non-Javadoc)
	 * @see com.cosylab.gui.displayers.DataConsumer#getName()
	 */
	public String getName()
	{
		return consumer.getName();
	}

	/* (non-Javadoc)
	 * @see com.cosylab.gui.displayers.DataConsumer#getSupportedCharacteristics()
	 */
	public String[] getSupportedCharacteristics()
	{
		return consumer.getSupportedCharacteristics();
	}

	/* (non-Javadoc)
	 * @see com.cosylab.gui.displayers.DataConsumer#getSupportedConsumerTypes()
	 */
	public Class[] getSupportedConsumerTypes()
	{
		return new Class[]{ DoubleConsumer.class };
	}
}

/* __oOo__ */
