/*
 * Copyright (c) 2003-2008 by Cosylab d. d.
 *
 * This file is part of CosyBeans.
 *
 * CosyBeans is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * CosyBeans is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with CosyBeans.  If not, see <http://www.gnu.org/licenses/>.
 */

package com.cosylab.gui.adapters;

import java.beans.PropertyVetoException;
import java.util.Map;

import com.cosylab.gui.displayers.DataSource;

/**
 * <code>DataSourceJoint</code> combines several DataSources and dispatch all
 * their data in a prescribed manner. Each DataSource should be added to this
 * joint together with an identifer that specifies what this joint should do
 * with the data received from that particular source. The first DataSource in
 * the joint should always have an identifier FIRST_IN_A_ROW, following by data
 * sources with any other operation identifier. There can be only one source
 * with identifier FIRST_IN_A_ROW.
 * <p>
 * When a new <code>DataSource</code> is added it should be the last in the
 * chain of all data sources if not specified otherwise. Value from each
 * <code>DataSource</code> should meet an arythmetic operation with the value
 * calculated from sources that are higher up in the chain (were added to this
 * joint earlier).
 * </p>
 * <p>
 * addDataSource(source1, FIRST_IN_A_ROW);
 * addDataSource(source2, ADD);
 * addDataSource(source3, MULTIPLY); 
 * should give a result of (source1 + source2)* source3.
 * </p>
 * 
 * @author Jaka Bobnar, Cosylab
 * 
 * @param <T> type of DataSources added to this joint
 */
public interface DataSourceJoint<T extends DataSource> extends DataSource {

	/** Identifier for the first DataSource */
	public static final String FIRST_IN_A_ROW = "first";

	/** Identifier for multiplication */
	public static final String MULTIPLY = "multiply";

	/** Identifier for division */
	public static final String DIVIDE = "divide";

	/** Identifier for addition */
	public static final String ADD = "add";

	/** Identifier for substraction */
	public static final String SUBSTRACT = "substract";

	/**
	 * Adds a <code>DataSource</code> to the chain of source as the last
	 * dataSource. The operation between added DataSource and the existing chain
	 * is specified by operation parameter, which can be one of the public
	 * identifiers.
	 * 
	 * @param dataSource
	 * @param operation
	 * @throws IllegalArgumentException
	 *             when a given datasource have operation FIRST_IN_A_ROW and a
	 *             DataSource with this identifier already exists
	 */
	public void addDataSource(T dataSource, String operation)
			throws IllegalArgumentException, PropertyVetoException;

	/**
	 * Adds a <code>DataSource</code> to the chain of source at the position
	 * specified by index. The operation between added DataSource and the
	 * existing chain is specified by operation parameter, which can be one of
	 * the public identifiers. The FIRST_IN_A_ROW is specified by index 0. If a
	 * DataSource is added as first in a row it would replace the existing first
	 * DataSource and its identifier should be automatically changed to
	 * FIRST_IN_A_ROW if a different one is given.
	 * 
	 * @param dataSource
	 * @param operation
	 * @param index
	 * @throws IllegalArgumentException
	 *             when a given index is invalid (larger than the number of data 
	 *             sources in this joint)
	 */
	public void addDataSource(T dataSource, String operation, int index)
			throws IllegalArgumentException, PropertyVetoException;

	/**
	 * Removes the dataSource from this joint. When a single
	 * <code>DataSource</code> is removed, it should not have any effects on
	 * the complete chain of sources. The sources that are further down in the
	 * chain should be moves upward. Method should return identifier of the
	 * operation associated with the removed data source.
	 * 
	 * @param dataSource DataSource to be removed
	 * @return operation associated with the data source
	 */
	public String removeDataSource(T dataSource);

	/**
	 * Returns a Map of all dataSources and their identifiers. The map is
	 * expected to use data sources as keys and operation identifiers as values.
	 * The order of data sources should be preserved.
	 * 
	 * @return a map of all data sources
	 */
	public Map<T, String> getDataSources();
}
