/*
 * Copyright (c) 2003-2008 by Cosylab d. d.
 *
 * This file is part of CosyBeans.
 *
 * CosyBeans is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * CosyBeans is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with CosyBeans.  If not, see <http://www.gnu.org/licenses/>.
 */

package com.cosylab.gui.adapters;

import java.util.HashMap;
import java.util.Map;
import java.util.Objects;

import com.cosylab.gui.displayers.CommonDisplayer;

/**
 * <code>ExponentialConverter</code> uses the supplied values
 * as the exponential factor to the specified base. If the received value is
 * x, the returned value is <code>base^x</code>.
 * 
 * @author <a href="mailto:jaka.bobnar@cosylab.com">Jaka Bobnar</a>
 * @version $Id: ExponentialConverter.java,v 1.7 2008-04-22 12:31:02 jbobnar Exp $
 *
 */
public class ExponentialConverter extends SimpleConverterSupport {

	private static final long serialVersionUID = -1789030566494308338L;

	public static final String SHORT_NAME = "exponential";
	
	private double base = Math.E, logBase = 1;
	
	/**
	 * Constructs a new ExponentialConverter with base e = 2.71828...
	 * 
	 * @see Math#E
	 */
	public ExponentialConverter() {
		setBase(Math.E);
	}
	
	/**
	 * Constructs a new ExponentialConverter with the provided base.
	 * 
	 * @param base
	 */
	public ExponentialConverter(double base){
		setBase(base);
	}
	
	/*
	 * (non-Javadoc)
	 * @see com.cosylab.gui.adapters.SimpleConverterSupport#inverseTransform(double)
	 */
	@Override
	protected double inverseTransform(double value) {
		if (value <= 0) return 0;
		return Math.log(value)/getLogBase();
	}

	/*
	 * (non-Javadoc)
	 * @see com.cosylab.gui.adapters.SimpleConverterSupport#transform(double)
	 */
	@Override
	protected double transform(double value) {
		return Math.pow(getBase(), value);
	}
	
	/*
	 * (non-Javadoc)
	 * @see com.cosylab.gui.adapters.DataConsumerDispatcher#getName()
	 */
	@Override
    public String getName(){
		return SHORT_NAME;
	}
	
	/*
	 * (non-Javadoc)
	 * @see java.lang.Object#toString()
	 */
	@Override
    public String toString(){
		return SHORT_NAME + ": " + getBase();
	}
	
	/**
	 * Returns the base used in this converter.
	 * 
	 * @return
	 */
	public double getBase() {
		return base;
	}

	/**
	 * Sets the base factor used for converting values.
	 * 
	 * @param base
	 */
	public void setBase(double base) {
		if (base <= 0) return;
		this.base = base;
		setLogBase(Math.log(base));
	}

	protected double getLogBase() {
		return logBase;
	}

	protected void setLogBase(double logBase) {
		this.logBase = logBase;
	}
	
	@Override
    public void setCharacteristics(Map characteristics)
	{
		// We store unmodified characteristic for possible later reinitialization of 
		// consumers, eg. if multiplication factor changes
		cacheLastCharacteristics(characteristics);

		HashMap<String,Object> map = new HashMap<>(characteristics);

		double min;
		double max;
		
		// Modifying the graph_min and graph_max characteristics.
		if (map.containsKey(CommonDisplayer.C_GRAPH_MIN)
		    && map.containsKey(CommonDisplayer.C_GRAPH_MAX)) {
			min = ((Number)map.get(CommonDisplayer.C_GRAPH_MIN)).doubleValue();
			max = ((Number)map.get(CommonDisplayer.C_GRAPH_MAX)).doubleValue();
				
			map.put(CommonDisplayer.C_GRAPH_MIN,
			    new Double(transform(min)));
			map.put(CommonDisplayer.C_GRAPH_MAX,
			    new Double(transform(max)));
		}
		
		// Modifying the minimum and maximum characteristics.
		if (map.containsKey(CommonDisplayer.C_MINIMUM)
		    && map.containsKey(CommonDisplayer.C_MAXIMUM)) {
			min = ((Number)map.get(CommonDisplayer.C_MINIMUM)).doubleValue();
			max = ((Number)map.get(CommonDisplayer.C_MAXIMUM)).doubleValue();
						
			map.put(CommonDisplayer.C_MINIMUM,
			    new Double(transform(min)));
			map.put(CommonDisplayer.C_MAXIMUM,
			    new Double(transform(max)));
		}

		// Modifying the units characteristics
		map.put(CommonDisplayer.C_UNITS,
		    "log_"+getBase()+"("+ map.get(CommonDisplayer.C_UNITS)+")");

		super.setCharacteristics(map);
	}
	
	/*
	 * (non-Javadoc)
	 * @see java.lang.Object#equals(java.lang.Object)
	 */
	@Override
	public boolean equals(Object obj) {
		if (!(obj instanceof ExponentialConverter)) return false;
		ExponentialConverter c = (ExponentialConverter)obj;
		return c.getBase() == getBase();
	}
	
	@Override
	public int hashCode() {
	    return Objects.hash(SHORT_NAME, base);
	}
}
