/*
 * Copyright (c) 2003-2008 by Cosylab d. d.
 *
 * This file is part of CosyBeans.
 *
 * CosyBeans is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * CosyBeans is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with CosyBeans.  If not, see <http://www.gnu.org/licenses/>.
 */

package com.cosylab.gui.adapters;

import java.util.HashMap;
import java.util.Map;
import java.util.Objects;

import com.cosylab.gui.displayers.CommonDisplayer;
import com.cosylab.gui.displayers.DoubleConsumer;

/**
 * <code>PotentialConverter</code> transforms the incomming value 
 * in the way that it exponents it to the given exponent factor.
 * If the incomming value is x, the transformed value is x^exponent.
 * 
 * @author <a href="mailto:jaka.bobnar@cosylab.com">Jaka Bobnar</a>
 * @version $Id: PotentialConverter.java,v 1.8 2008-04-22 12:31:02 jbobnar Exp $
 *
 */
public class PotentialConverter extends SimpleConverterSupport implements
		DoubleConsumer {

	private static final long serialVersionUID = 3449182777513744784L;

	public static final String SHORT_NAME = "Potential";
	
	private double exponent = 1;
	
	/**
	 * Constructs a new PotentialConverter with exponent factor 1.
	 * 
	 */
	public PotentialConverter() {
		this(1);
	}
	
	/**
	 * Constructs a new PotentialConverter with the supplied exponent.
	 * 
	 * @param exponent
	 */
	public PotentialConverter(double exponent){
		setExponent(exponent);
	}
	
	/*
	 * (non-Javadoc)
	 * @see com.cosylab.gui.adapters.SimpleConverterSupport#inverseTransform(double)
	 */
	@Override
	protected double inverseTransform(double value) {
		return Math.pow(value, 1.0 / getExponent());
	}
	
	/**
	 * Returns the exponent factor.
	 * 
	 * @return
	 */
	public double getExponent() {
		return exponent;
	}

	/*
	 * (non-Javadoc)
	 * @see com.cosylab.gui.adapters.SimpleConverterSupport#transform(double)
	 */
	@Override
	protected double transform(double value) {
		return Math.pow(value, getExponent());
	}
	
	/**
	 * Sets the exponent factor of this converter.
	 * 
	 * @param exponent
	 */
	public void  setExponent(double exponent){
		if (exponent <= 0) this.exponent = 1;
		else this.exponent = exponent;
	}
	
	/*
	 * (non-Javadoc)
	 * @see com.cosylab.gui.adapters.DataConsumerDispatcher#getName()
	 */
	@Override
	public String getName(){
		return SHORT_NAME;
	}
	
	/*
	 * (non-Javadoc)
	 * @see java.lang.Object#toString()
	 */
	@Override
	public String toString(){
		return SHORT_NAME + ": " + getExponent();
	}
	
	/*
	 * (non-Javadoc)
	 * @see com.cosylab.gui.adapters.AbstractConverter#setCharacteristics(java.util.Map)
	 */
	@Override
	public void setCharacteristics(Map characteristics)
	{
		// We store unmodified characteristic for possible later reinitialization of 
		// consumers, eg. if multiplication factor changes
		cacheLastCharacteristics(characteristics);

		HashMap map = new HashMap(characteristics);

		double min;
		double max;
		
		// Modifying the graph_min and graph_max characteristics.
		if (map.containsKey(CommonDisplayer.C_GRAPH_MIN)
		    && map.containsKey(CommonDisplayer.C_GRAPH_MAX)) {
			min = ((Number)map.get(CommonDisplayer.C_GRAPH_MIN)).doubleValue();
			max = ((Number)map.get(CommonDisplayer.C_GRAPH_MAX)).doubleValue();
				
			map.put(CommonDisplayer.C_GRAPH_MIN,
			    new Double(transform(min)));
			map.put(CommonDisplayer.C_GRAPH_MAX,
			    new Double(transform(max)));
		}
		
		// Modifying the minimum and maximum characteristics.
		if (map.containsKey(CommonDisplayer.C_MINIMUM)
		    && map.containsKey(CommonDisplayer.C_MAXIMUM)) {
			min = ((Number)map.get(CommonDisplayer.C_MINIMUM)).doubleValue();
			max = ((Number)map.get(CommonDisplayer.C_MAXIMUM)).doubleValue();
						
			map.put(CommonDisplayer.C_MINIMUM,
			    new Double(transform(min)));
			map.put(CommonDisplayer.C_MAXIMUM,
			    new Double(transform(max)));
		}

//		// Modifying the units characteristics
		map.put(CommonDisplayer.C_UNITS,
		    map.get(CommonDisplayer.C_UNITS) +"^1/"+getExponent());

		super.setCharacteristics(map);
	}
	
	/*
	 * (non-Javadoc)
	 * @see java.lang.Object#equals(java.lang.Object)
	 */
	@Override
	public boolean equals(Object obj) {
		if (!(obj instanceof PotentialConverter)) return false;
		PotentialConverter c = (PotentialConverter)obj;
		return c.getExponent() == exponent;
	}
	
	@Override
	public int hashCode() {
	    return Objects.hash(SHORT_NAME, exponent);
	}
}
