/*
 * Copyright (c) 2003-2008 by Cosylab d. d.
 *
 * This file is part of CosyBeans-Common.
 *
 * CosyBeans-Common is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * CosyBeans-Common is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with CosyBeans-Common.  If not, see <http://www.gnu.org/licenses/>.
 */

package com.cosylab.gui.components;

import com.cosylab.application.state.State;
import com.cosylab.gui.components.util.CosyTransferHandler;
import com.cosylab.util.Debug;
import com.cosylab.util.PrintfFormat;

import java.awt.Color;
import java.awt.GridBagConstraints;
import java.awt.Insets;
import java.beans.Beans;

import javax.swing.JComponent;
import javax.swing.JLabel;
import javax.swing.TransferHandler;


/**
 * An extension of AbstractDisplayerPanel used for numerical displayers.
 *
 * @author <a href="mailto:jernej.kamenik@cosylab.com">Jernej Kamenik</a>
 * @version $id$
 */
public abstract class AbstractNumericDisplayerPanel
	extends AbstractDisplayerPanel
{
	private static final long serialVersionUID = 1L;
	private Class<? extends Number> numberType = Double.class;
	private Number max;
	private Number min;
	private PrintfFormat formatter;
	private ResizableTextLabel maxLabel;
	private ResizableTextLabel minLabel;
	private ResizableTextLabel unitsLabel;
	private String format;
	private String units;
	private boolean boundsVisible;
	private boolean unitsVisible;
	private boolean unitsShownWithTitle;

	/**
	 * Default constructor.
	 */
	public AbstractNumericDisplayerPanel()
	{
		this(null, false, Layout.DYNAMIC, true, true, true, 1,
		    Integer.MAX_VALUE, null, true, false);
	}

	/**
	 * Creates a new AbstractNumericDisplayerPanel object with initial
	 * parameters.
	 *
	 * @param title initial value
	 * @param titleVisible initial value
	 * @param layoutOrientation initial value
	 * @param resizable initial value
	 * @param enabled initial value
	 * @param enchanced initial value
	 * @param minTitleFontSize initial value
	 * @param maxTitleFontSize initial value
	 * @param units initial value
	 * @param unitsVisible initial value
	 * @param unitsShownWithTitle initial value
	 */
	public AbstractNumericDisplayerPanel(String title, boolean titleVisible,
	    Layout layoutOrientation, boolean resizable, boolean enabled,
	    boolean enchanced, int minTitleFontSize, int maxTitleFontSize,
	    String units, boolean unitsVisible, boolean unitsShownWithTitle)
	{
		super(title, titleVisible, layoutOrientation, resizable, enabled,
		    enchanced, minTitleFontSize, maxTitleFontSize);

		formatter = new PrintfFormat("%3.2f");

		max = new Double(Double.POSITIVE_INFINITY);
		min = new Double(Double.NEGATIVE_INFINITY);

		internalSetFormat();
		internalSetMaximum();
		internalSetMinimum();

		setUnits(units);
		setUnitsVisible(unitsVisible);
		setUnitsShownWithTitle(unitsShownWithTitle);
	}

	/**
	 * Sets the visibility of the numerical bounds.
	 *
	 * @param b wether the bounds should be displayed in the panel.
	 */
	public void setBoundsVisible(boolean b)
	{
		if (b == boundsVisible) {
			return;
		}

		boundsVisible = b;
		layoutDisplayer();
		firePropertyChange("boundsVisible", !b, b);
	}

	/**
	 * Determines wether the component should display the numerical bounds of
	 * the value component
	 *
	 * @return <code>true</code> if the bounds are displayed,
	 *         <code>false</code> otherwise.
	 */
	public boolean isBoundsVisible()
	{
		return boundsVisible;
	}

	/**
	 * Sets the C-style format String used for formatting numerical values
	 *
	 * @param value format String
	 */
	public void setFormat(String value)
	{
		if (value == null && format == null
		    || value != null && value.equals(format)) {
			return;
		}

		String oldVal = format;
		format = value;
		internalSetFormat();
		layoutDisplayer();
		firePropertyChange("format", oldVal, getFormat());
	}

	/**
	 * Gets the C-style format string used for formatting numerical display
	 *
	 * @return the C-style format String
	 */
	public String getFormat()
	{
		return format;
	}

	/*
	 * If bounds are visible only vertical layout is supported.
	 *
	 * @see com.cosylab.gui.components.DisplayerPanel#setLayoutOrientation(int)
	 */
	@Override
	public void setLayoutType(Layout i)
	{
		if (isBoundsVisible() && (i != Layout.VERTICAL)) {
			return;
		}

		super.setLayoutType(i);
	}

	/*
	 * If bounds are visible only vertical layout is supported.
	 *
	 * @see com.cosylab.gui.components.DisplayerPanel#getLayoutOrientation()
	 */
	@Override
	public Layout getLayoutType()
	{
		if (boundsVisible) {
			return Layout.VERTICAL;
		}

		return super.getLayoutType();
	}

	/*
	 * (non-Javadoc)
	 * @see com.cosylab.gui.components.AbstractDisplayerPanel#setState(com.cosylab.application.state.State)
	 */
	@Override
	public void setState(State state)
	{
		super.setState(state);

		setBoundsVisible(state.getBoolean("boundsVisible", isBoundsVisible()));
		setUnitsVisible(state.getBoolean("unitsVisible", isUnitsVisible()));
		setUnits(state.getString("units", getUnits()));
		setFormat(state.getString("format", getFormat()));
	}

	/*
	 * (non-Javadoc)
	 * @see com.cosylab.gui.components.AbstractDisplayerPanel#getState()
	 */
	@Override
	public State getState()
	{
		State state = super.getState();

		state.putBoolean("boundsVisible", isBoundsVisible());
		state.putBoolean("unitsVisible", isUnitsVisible());
		state.putString("units", getUnits());
		state.putString("format", getFormat());

		return state;
	}

	/**
	 * Sets the units String to denote the physical units of numerical values
	 * displayed
	 *
	 * @param value units String
	 */
	public void setUnits(String value)
	{
		if (value == null && units == null
		    || value != null && value.equals(units)) {
			return;
		}

		units = value;

		internalSetUnits();
		layoutDisplayer();

		firePropertyChange("units", null, value);
	}

	/**
	 * Gets the physical units of the value displayed by the numeric component
	 *
	 * @return units String of numeric value displayed
	 */
	public String getUnits()
	{
		return units;
	}

	/**
	 * Set to <code>true</code> to display units.
	 *
	 * @param b <code>true</code> if units are to be displayed.
	 */
	public void setUnitsVisible(boolean b)
	{
		if (b == unitsVisible) {
			return;
		}
		
		unitsVisible = b;
		layoutDisplayer();
		firePropertyChange("unitsVisible", !b, b);
		repaint();
	}

	/**
	 * Returns <code>true</code> if units are displayed.
	 *
	 * @return boolean <code>true</code> if units are displayed.
	 */
	public boolean isUnitsVisible()
	{
		return unitsVisible;
	}

	/**
	 * Gets the component displaying the maximum numerical value. This method
	 * is used by the NumericDisplayerPanelLayout.
	 *
	 * @return the component displaying the nuemrical maximum
	 */
	protected JComponent getMaxComponent()
	{
		if (maxLabel == null) {
			maxLabel = new ResizableTextLabel();
			maxLabel.setHorizontalAlignment(JLabel.CENTER);
			maxLabel.setResizable(isResizable());
			maxLabel.setEnabled(isEnabled());
			maxLabel.setEnhanced(isEnhanced());
//			maxLabel.setVisible(isBoundsVisible());
			maxLabel.setMaximumFontSize(getTitleMaximumFontSize());
			maxLabel.setMinimumFontSize(getTitleMinimumFontSize());
			maxLabel.setBackground(getBackground());
			maxLabel.setForeground(getForeground());
			internalSetMaximum();
		}

		return maxLabel;
	}

	/*
	 * Auxilary method to be used for changing the dispalyed numerical bounds.
	 */
	protected void setMaximumValue(Number value)
	{
		if (value == null) {
			if (numberType == Double.class) {
				value = new Double(Double.POSITIVE_INFINITY);
			}

			if (numberType == Long.class) {
				value = new Long(Long.MAX_VALUE);
			}
		}

		max = value;
		internalSetMaximum();
	}

	/*
	 * Auxilary method to be used for changing the dispalyed numerical bounds.
	 */
	protected Number getMaximumValue()
	{
		return max;
	}

	/**
	 * Returns the component displaying the minimum numerical value. This method
	 * is used by the NumericDisplayerPanelLayout.
	 *
	 * @return the component displaying the nuemrical minimum
	 */
	protected JComponent getMinComponent()
	{
		if (minLabel == null) {
			minLabel = new ResizableTextLabel();
			minLabel.setHorizontalAlignment(JLabel.CENTER);
			minLabel.setResizable(isResizable());
			minLabel.setEnabled(isEnabled());
			minLabel.setEnhanced(isEnhanced());
//			minLabel.setVisible(isBoundsVisible());
			minLabel.setMaximumFontSize(getTitleMaximumFontSize());
			minLabel.setMinimumFontSize(getTitleMinimumFontSize());
			minLabel.setBackground(getBackground());
			minLabel.setForeground(getForeground());
			internalSetMinimum();
		}

		return minLabel;
	}

	/*
	 * Auxilary method to be used for changing the dispalyed numerical bounds.
	 */
	protected void setMinimumValue(Number value)
	{
		if (value == null) {
			if (numberType == Double.class) {
				value = new Double(Double.NEGATIVE_INFINITY);
			}

			if (numberType == Long.class) {
				value = new Long(-Long.MAX_VALUE);
			}
		}

		min = value;
		internalSetMinimum();
	}

	/*
	 * Auxilary method to be used for changing the dispalyed numerical bounds.
	 */
	protected Number getMinimumValue()
	{
		return min;
	}

	/*
	 * Sets a new Number type to the panel.
	 *
	 * @param class1 can be any of <code>double.class</code>,
	 *        <code>long.class</code><code>Double.class</code> or
	 *        <code>Long.class</code> otherwise an
	 *        <code>IllegalArgumentException</code> is thrown. <i>Primitive
	 *        type classes are automatically converted to wrapper type
	 *        classes.</i>
	 *
	 * @throws IllegalArgumentException if the type entered is not Long.class,
	 * long.class, Double.class or double.class
	 */
	protected void setNumberType(Class<? extends Number> class1)
	{
		if (class1 == null) {
			throw new IllegalArgumentException("Number type must not be null.");
		}

		if (class1 != Double.class && class1 != double.class
		    && class1 != Long.class && class1 != long.class) {
			throw new IllegalArgumentException(
			    "Class must represent a  valid number type.");
		}

		if (class1 == numberType) {
			return;
		}

		if ((class1 == Double.class) || (class1 == double.class)) {
			numberType = Double.class;
			min = new Double(min.doubleValue());
			max = new Double(max.doubleValue());
		} else if ((class1 == Long.class) || (class1 == long.class)) {
			numberType = Long.class;
			min = new Long(min.longValue());
			max = new Long(max.longValue());
		} else {
			throw new IllegalArgumentException("Number type.");
		}

		if (getFormat() != null) {
			setFormat(null);
		} 
		internalSetFormat();
	}

	/**
	 * Determines the number type currently set to this NumericDispalyerPanel.
	 * Only Numbers of selected type can be dispalyed by the component.
	 *
	 * @return <code>Double.class</code> or <code>Long.class</code>
	 */
	protected Class<? extends Number> getNumberType()
	{
		return numberType;
	}

	/**
	 * Returns the GUI component diplaying units.
	 *
	 * @return the GUI component diplaying units
	 */
	protected JComponent getUnitsComponent()
	{
		if (unitsLabel == null) {
			unitsLabel = new ResizableTextLabel();
			unitsLabel.setHorizontalAlignment(JLabel.LEFT);
			unitsLabel.setResizable(isResizable());
			unitsLabel.setEnhanced(isEnhanced());
			unitsLabel.setEnabled(isEnabled());
			unitsLabel.setMaximumFontSize(getTitleMaximumFontSize());
			unitsLabel.setMinimumFontSize(getTitleMinimumFontSize());
			unitsLabel.setBackground(getBackground());
			unitsLabel.setForeground(getForeground());
			internalSetUnits();
		}

		return unitsLabel;
	}
		
	@Override
	public void setPopupEnabled(boolean enabled) {
		super.setPopupEnabled(enabled);
		if (enabled) {
			getUnitsComponent().addMouseListener(getPopupManager().getMouseHook());
			getMinComponent().addMouseListener(getPopupManager().getMouseHook());
			getMaxComponent().addMouseListener(getPopupManager().getMouseHook());
		} else {
			getUnitsComponent().removeMouseListener(getPopupManager().getMouseHook());
			getMinComponent().removeMouseListener(getPopupManager().getMouseHook());
			getMaxComponent().removeMouseListener(getPopupManager().getMouseHook());
		}		
	}

	/**
	 * Overriden to enable/disable min/max labels.
	 */
	@Override
	protected void internalSetEnabled()
	{
		super.internalSetEnabled();

		if (maxLabel != null) {
			maxLabel.setEnabled(isEnabled());
		}

		if (minLabel != null) {
			minLabel.setEnabled(isEnabled());
		}

		if (unitsLabel != null) {
			unitsLabel.setEnabled(isEnabled());
		}
	}

	/**
	 * Overriden to set enahnced mode to min/max labels.
	 */
	@Override
	protected void internalSetEnhanced()
	{
		super.internalSetEnhanced();

		if (maxLabel != null) {
			maxLabel.setEnhanced(isEnhanced());
		}

		if (minLabel != null) {
			minLabel.setEnhanced(isEnhanced());
		}

		if (unitsLabel != null) {
			unitsLabel.setEnhanced(isEnhanced());
		}
	}

	protected void internalSetFormat()
	{
		String tFormat = format;

		if (format == null) {
			tFormat = NumberField.createDefaultFormat(getNumberType());
			format = tFormat;
		}
		
		formatter = new PrintfFormat(tFormat);

		internalSetMaximum();
		internalSetMinimum();
	}

	protected void internalSetMaximum()
	{
		if (maxLabel != null && formatter != null) {
			try {
				maxLabel.setText(formatter.sprintf(getMaximumValue()));
			} catch (IllegalArgumentException iae) {
				Debug.out("Bad format '" + format + "': " + iae);
			}
		}
	}

	protected void internalSetMinimum()
	{
		if (minLabel != null && formatter != null) {
			try {
				minLabel.setText(formatter.sprintf(getMinimumValue()));
			} catch (IllegalArgumentException iae) {
				Debug.out("Bad format '" + format + "': " + iae);
			}
		}
	}

	/**
	 * Overriden to implement additional functionality.
	 *
	 * @see com.cosylab.gui.components.AbstractDisplayerPanel#setResizable(boolean)
	 */
	@Override
	protected void internalSetResizable()
	{
		super.internalSetResizable();

		if (minLabel != null) {
			minLabel.setResizable(isResizable());
		}

		if (maxLabel != null) {
			maxLabel.setResizable(isResizable());
		}

		if (unitsLabel != null) {
			unitsLabel.setResizable(isResizable());
		}
	}

	protected void internalSetUnits()
	{
		if (unitsLabel != null) {
			String tUnits = units;

			if (units == null && Beans.isDesignTime()) {
				tUnits = "<u>";
			}

			//if (tUnits == null) {
			//	tUnits = "";
			//}
			unitsLabel.setText(tUnits);
		}
	}
	
	/*
	 * (non-Javadoc)
	 * @see com.cosylab.gui.components.AbstractDisplayerPanel#internalSetTitle()
	 */
	@Override
	protected void internalSetTitle()
	{
		super.internalSetTitle();

		String u = units;

		if (unitsShownWithTitle) {
			if (units == null && Beans.isDesignTime()) {
				u = "<u>";
			}

			getTitleComponent().setText(getTitleComponent().getText() + " ["
			    + u + "]");
		}
	}

	/* (non-Javadoc)
	 * @see com.cosylab.gui.components.DisplayerPanel#layoutDisplayer()
	 */
	@Override
	protected void layoutDisplayer()
	{
		if (isUnitsVisible() && isTitleVisible() && isBoundsVisible()
		    && getTitle() != null && getUnits() != null
		    && getTitle().length() > 0 && getUnits().length() > 0) {
			layoutValueAndTitleAndUnitsAndBounds();
		} else if (isUnitsVisible() && isBoundsVisible() && getUnits() != null
		    && getUnits().length() > 0) {
			layoutValueAndUnitsAndBounds();
		} else if (isUnitsVisible() && isTitleVisible() && getTitle() != null
		    && getUnits() != null && getTitle().length() > 0
		    && getUnits().length() > 0) {
			layoutValueAndTitleAndUnits();
		} else if (isUnitsVisible() && getUnits() != null
		    && getUnits().length() > 0) {
			layoutValueAndUnits();
		} else if (isBoundsVisible() && isTitleVisible() && getTitle() != null
		    && getTitle().length() > 0) {
			layoutValueAndTitleAndBounds();
		} else if (isBoundsVisible()) {
			layoutValueAndBounds();
		} else {
			super.layoutDisplayer();
			//do not revalidate. it was done in super method
			return;
		}
		validate();
	}

	/*
	 * Lays out both value and bounds components.
	 */
	protected void layoutValueAndBounds()
	{
		removeAll();

		Insets insets = (!isResizable()) ? new Insets(2, 2, 2, 2)
			: new Insets((getSize().height * 3) / 100,
			    (getSize().width * 3) / 100, (getSize().height * 3) / 100,
			    (getSize().width * 3) / 100);
		int fill = (isResizable() ? GridBagConstraints.BOTH
			: GridBagConstraints.NONE);
		double w = isResizable() ? 1.0 : 0.0;

		double wValue = getFixedDisplayerSize() != null ? 0 : w;
		double hValue = getFixedDisplayerSize() != null ? 0 : w*1.2;
		int fillValue = getFixedDisplayerSize() != null ? GridBagConstraints.NONE : fill;
		add(getValueComponent(),
		    new GridBagConstraints(1, 1, 2, 1, wValue, hValue,
		        GridBagConstraints.CENTER, fillValue, insets, 0, 0));
		add(getMinComponent(),
		    new GridBagConstraints(1, 2, 1, 1, w, w * 0.5,
		        GridBagConstraints.CENTER, fill, insets, 0, 0));
		add(getMaxComponent(),
		    new GridBagConstraints(2, 2, 1, 1, w, w * 0.5,
		        GridBagConstraints.CENTER, fill, insets, 0, 0));
	}

	/*
	 * Lays out both value, title and bounds components.
	 */
	protected void layoutValueAndTitleAndBounds()
	{
		removeAll();

		Insets insets = (!isResizable()) ? new Insets(2, 2, 2, 2)
			: new Insets((getSize().height * 3) / 100,
			    (getSize().width * 3) / 100, (getSize().height * 3) / 100,
			    (getSize().width * 3) / 100);
		int fill = (isResizable() ? GridBagConstraints.BOTH
			: GridBagConstraints.NONE);
		double w = isResizable() ? 1.0 : 0.0;

		double wValue = getFixedDisplayerSize() != null ? 0 : w;
		double hValue = getFixedDisplayerSize() != null ? 0 : w*1.2;
		int fillValue = getFixedDisplayerSize() != null ? GridBagConstraints.NONE : fill;
		add(getTitleComponent(),
		    new GridBagConstraints(1, 1, 2, 1, w, w, GridBagConstraints.CENTER,
		        fill, insets, 0, 0));
		add(getValueComponent(),
		    new GridBagConstraints(1, 2, 2, 1, wValue, hValue,
		        GridBagConstraints.EAST, fillValue, insets, 0, 0));
		add(getMinComponent(),
		    new GridBagConstraints(1, 3, 1, 1, w, w * 0.5,
		        GridBagConstraints.CENTER, fill, insets, 0, 0));
		add(getMaxComponent(),
		    new GridBagConstraints(2, 3, 1, 1, w, w * 0.5,
		        GridBagConstraints.CENTER, fill, insets, 0, 0));
	}

	protected void layoutValueAndTitleAndUnits()
	{
		removeAll();

		Insets insets = !isResizable() ? new Insets(2, 2, 2, 2)
			: new Insets(getSize().height * 3 / 100, getSize().width * 3 / 100,
			    getSize().height * 3 / 100, getSize().width * 3 / 100);
		int fill = (isResizable() ? GridBagConstraints.BOTH
			: GridBagConstraints.NONE);
		double w = isResizable() ? 1.0 : 0.0;

		double wValue = getFixedDisplayerSize() != null ? 0 : w;
		double hValue = getFixedDisplayerSize() != null ? 0 : w*1.2;
		int fillValue = getFixedDisplayerSize() != null ? GridBagConstraints.NONE : fill;
		if (getLayoutType() == Layout.HORIZONTAL
		    || (1. * getSize().height) / (getTitleComponent().getPreferredSize().height
		    + getValueComponent().getPreferredSize().height) < (1. * getSize().width
		    - insets.left - insets.right) / (getTitleComponent()
		    .getPreferredSize().width
		    + getValueComponent().getPreferredSize().width
		    + getUnitsComponent().getPreferredSize().width)
		    && getLayoutType() == Layout.DYNAMIC) {
			add(getTitleComponent(),
			    new GridBagConstraints(1, 1, 1, 1, w, w,
			        GridBagConstraints.CENTER, fill, insets, 0, 0));
			add(getValueComponent(),
			    new GridBagConstraints(2, 1, 1, 1, hValue, wValue,
			        GridBagConstraints.EAST, fillValue, insets, 0, 0));
			add(getUnitsComponent(),
			    new GridBagConstraints(3, 1, 1, 1, w * 0.33, w * 0.33,
			        GridBagConstraints.WEST, fill, insets, 0, 0));
		} else {
			add(getTitleComponent(),
			    new GridBagConstraints(1, 1, 2, 1, w, w,
			        GridBagConstraints.CENTER, fill, insets, 0, 0));
			add(getValueComponent(),
			    new GridBagConstraints(1, 2, 1, 1, hValue, hValue,
			        GridBagConstraints.EAST, fillValue, insets, 0, 0));
			add(getUnitsComponent(),
			    new GridBagConstraints(2, 2, 1, 1, w * 0.33, w * 0.33,
			        GridBagConstraints.WEST, fill, insets, 0, 0));
		}
	}

	protected void layoutValueAndTitleAndUnitsAndBounds()
	{
		removeAll();

		Insets insets = !isResizable() ? new Insets(2, 2, 2, 2)
			: new Insets(getSize().height * 3 / 100, getSize().width * 3 / 100,
			    getSize().height * 3 / 100, getSize().width * 3 / 100);
		int fill = (isResizable() ? GridBagConstraints.BOTH
			: GridBagConstraints.NONE);
		double w = isResizable() ? 1.0 : 0.0;

		double hValue = getFixedDisplayerSize() != null ? 0 : w*1.2;
		int fillValue = getFixedDisplayerSize() != null ? GridBagConstraints.NONE : fill;
		add(getTitleComponent(),
		    new GridBagConstraints(1, 1, 2, 1, w, w, GridBagConstraints.CENTER,
		        fill, insets, 0, 0));
		add(getValueComponent(),
		    new GridBagConstraints(1, 2, 2, 1, hValue, hValue,
		        GridBagConstraints.EAST, fillValue, insets, 0, 0));
		add(getUnitsComponent(),
		    new GridBagConstraints(3, 2, 1, 1, w, w, GridBagConstraints.WEST,
		        fill, insets, 0, 0));
		add(getMinComponent(),
		    new GridBagConstraints(1, 3, 1, 1, w, w * 0.5,
		        GridBagConstraints.CENTER, fill, insets, 0, 0));
		add(getMaxComponent(),
		    new GridBagConstraints(2, 3, 1, 1, w, w * 0.5,
		        GridBagConstraints.CENTER, fill, insets, 0, 0));
	}

	protected void layoutValueAndUnits()
	{
		removeAll();

		Insets insets = !isResizable() ? new Insets(2, 2, 2, 2)
			: new Insets(getSize().height * 3 / 100, getSize().width * 3 / 100,
			    getSize().height * 3 / 100, getSize().width * 3 / 100);
		int fill = (isResizable() ? GridBagConstraints.BOTH
			: GridBagConstraints.NONE);
		double w = isResizable() ? 1.0 : 0.0;

		double wValue = getFixedDisplayerSize() != null ? 0 : w;
		double hValue = getFixedDisplayerSize() != null ? 0 : w*1.2;
		int fillValue = getFixedDisplayerSize() != null ? GridBagConstraints.NONE : fill;
		add(getValueComponent(),
		    new GridBagConstraints(1, 1, 1, 1, hValue, wValue,
		        GridBagConstraints.EAST, fillValue, insets, 0, 0));
		add(getUnitsComponent(),
		    new GridBagConstraints(2, 1, 1, 1, w * 0.33, w,
		        GridBagConstraints.WEST, fill, insets, 0, 0));
	}

	protected void layoutValueAndUnitsAndBounds()
	{
		removeAll();

		Insets insets = !isResizable() ? new Insets(2, 2, 2, 2)
			: new Insets(getSize().height * 3 / 100, getSize().width * 3 / 100,
			    getSize().height * 3 / 100, getSize().width * 3 / 100);
		int fill = (isResizable() ? GridBagConstraints.BOTH
			: GridBagConstraints.NONE);
		double w = isResizable() ? 1.0 : 0.0;

		double wValue = getFixedDisplayerSize() != null ? 0 : w;
		double hValue = getFixedDisplayerSize() != null ? 0 : w*1.2;
		int fillValue = getFixedDisplayerSize() != null ? GridBagConstraints.NONE : fill;
		add(getValueComponent(),
		    new GridBagConstraints(1, 1, 2, 1, hValue, wValue,
		        GridBagConstraints.EAST, fillValue, insets, 0, 0));
		add(getUnitsComponent(),
		    new GridBagConstraints(2, 1, 1, 1, w, w * 1.2,
		        GridBagConstraints.WEST, fill, insets, 0, 0));
		add(getMinComponent(),
		    new GridBagConstraints(1, 2, 1, 1, w, w * 0.5,
		        GridBagConstraints.CENTER, fill, insets, 0, 0));
		add(getMaxComponent(),
		    new GridBagConstraints(2, 2, 1, 1, w, w * 0.5,
		        GridBagConstraints.CENTER, fill, insets, 0, 0));
	}

	/**
	 * If <code>true</code> units will be shown in title label.
	 *
	 * @return if <code>true</code> units will be shown in title label
	 */
	public boolean isUnitsShownWithTitle()
	{
		return unitsShownWithTitle;
	}

	/**
	 * If set to <code>true</code> units will be shown in title label
	 *
	 * @param unitsShownWithTitle if <code>true</code> units will be shown in
	 *        title label
	 */
	public void setUnitsShownWithTitle(boolean unitsShownWithTitle)
	{
		if (unitsShownWithTitle == this.unitsShownWithTitle) {
			return;
		}

		this.unitsShownWithTitle = unitsShownWithTitle;
		internalSetTitle();
		layoutDisplayer();
		firePropertyChange("unitsShownWithTitle", !unitsShownWithTitle,
		    unitsShownWithTitle);
	}
	
	/* (non-Javadoc)
     * @see com.cosylab.gui.components.AbstractDisplayerPanel#setTitleMaximumFontSize(int)
     */
	@Override
    public void setTitleMaximumFontSize(int max) {
		if (maxLabel!=null) maxLabel.setMaximumFontSize(max);
		if (minLabel!=null) minLabel.setMaximumFontSize(max);
		if (unitsLabel!=null) unitsLabel.setMaximumFontSize(max);
        super.setTitleMaximumFontSize(max);
    }
    
    /* (non-Javadoc)
     * @see com.cosylab.gui.components.AbstractDisplayerPanel#setTitleMinimumFontSize(int)
     */
    @Override
    public void setTitleMinimumFontSize(int min) {
		if (maxLabel!=null) maxLabel.setMinimumFontSize(min);
		if (minLabel!=null) minLabel.setMinimumFontSize(min);
		if (unitsLabel!=null) unitsLabel.setMinimumFontSize(min);
        super.setTitleMinimumFontSize(min);
    }
    
    /* (non-Javadoc)
     * @see javax.swing.JComponent#setTransferHandler(javax.swing.TransferHandler)
     */
    @Override
    public void setTransferHandler(TransferHandler newHandler) {
    	super.setTransferHandler(newHandler);
    	CosyTransferHandler.registerTransferHandler(this,newHandler,new JComponent[]{getMaxComponent(),getMinComponent(),getUnitsComponent()});
    }
    
    /*
     * (non-Javadoc)
     * @see com.cosylab.gui.components.AbstractDisplayerPanel#setBackground(java.awt.Color)
     */
    @Override
    public void setBackground(Color bg) {
    	super.setBackground(bg);
    	getUnitsComponent().setBackground(bg);
    	getMinComponent().setBackground(bg);
    	getMaxComponent().setBackground(bg);
    }
    
    /*
     * (non-Javadoc)
     * @see com.cosylab.gui.components.AbstractDisplayerPanel#setForeground(java.awt.Color)
     */
    @Override
    public void setForeground(Color fg) {
    	super.setForeground(fg);
    	getUnitsComponent().setForeground(fg);
    	getMinComponent().setForeground(fg);
    	getMaxComponent().setForeground(fg);
    }
}

/* __oOo__ */
