/*
 * Copyright (c) 2003-2008 by Cosylab d. d.
 *
 * This file is part of CosyBeans-Common.
 *
 * CosyBeans-Common is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * CosyBeans-Common is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with CosyBeans-Common.  If not, see <http://www.gnu.org/licenses/>.
 */

package com.cosylab.gui.components;

import java.awt.GridBagConstraints;
import java.awt.GridBagLayout;
import java.awt.Insets;
import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;

import java.util.Calendar;
import java.util.Date;
import java.util.GregorianCalendar;

import javax.swing.JDialog;


/**
 * This is a container for SimpleDateSelector and SimpleTimeSelector components,
 * combining the functionality of both. The components consists of six
 * user-editable fields for year, month, day, hour, minute and seconds. All
 * fields are validating allowing user to select only valid date and time
 * values. These are the obtained using getDate and setDate methods. 
 * <p>
 * When a new date is set the component notifies PropertyChangeListeners through
 * property <code>date</code>.
 * </p>
 *
 * @author <a href="mailto:jernej.kamenik@cosylab.com">Jernej Kamenik</a>
 * @version $id$
 */
public class DateTimeSelector extends javax.swing.JPanel implements PropertyChangeListener 
{
	
	private static final long serialVersionUID = 1L;
	private SimpleDateSelector dateChooser;
	private SimpleTimeSelector timeChooser;
	private Calendar calendar = new GregorianCalendar();

	/**
	 * Constructs a new DateTimeSelector.
	 */
	public DateTimeSelector()
	{
		super(new GridBagLayout());
		createComponents();
	}

	/**
	 * Internal helper routine creates all the components. 
	 */
	protected void createComponents()
	{
		dateChooser = new SimpleDateSelector();
		dateChooser.setBorder(null);
		dateChooser.addPropertyChangeListener("date",this);
		add(dateChooser, new GridBagConstraints(0,0,1,1,1,0,GridBagConstraints.CENTER,GridBagConstraints.HORIZONTAL,new Insets(0,0,4,0),0,0));

		timeChooser = new SimpleTimeSelector();
		timeChooser.setBorder(null);
		timeChooser.addPropertyChangeListener("time",this);
		add(timeChooser,new GridBagConstraints(0,1,1,1,1,0,GridBagConstraints.CENTER,GridBagConstraints.HORIZONTAL,new Insets(0,0,0,0),0,0));
	}

	/**
	 * Returns the date displayed by this component. The result is
	 * java.util.Date object. Only the following fields are set: SECOND,
	 * MINUTE, HOUR_OF_DAY, DAY_OF_MONTH, MONTH, YEAR. For more information on
	 * these fields see the java.util.Calendar. 
	 *
	 * @return java.util.Date
	 */
	public Date getDate()
	{
		calendar.clear();
		calendar.set(Calendar.SECOND, timeChooser.getSeconds());
		calendar.set(Calendar.MINUTE, timeChooser.getMinute());
		calendar.set(Calendar.HOUR_OF_DAY, timeChooser.getHour());
		calendar.set(Calendar.YEAR, dateChooser.getYear());
		calendar.set(Calendar.MONTH, dateChooser.getMonth());
		calendar.set(Calendar.DAY_OF_MONTH, dateChooser.getDay());

		return calendar.getTime();
	}

	/**
	 * Sets the date to display in this selector. Value is set through the
	 * java.util.Date object. Only the following fields are considerer:
	 * SECOND, MINUTE, HOUR_OF_DAY, DAY_OF_MONTH, MONTH, YEAR. For more
	 * information on these fields see the java.util.Calendar class. 
	 *
	 * @param date java.util.Date
	 */
	public void setDate(Date date)
	{
		timeChooser.setTime(date);
		dateChooser.setDate(date);
	}

	/**
	 * Sets the enabled state of this component and the components contained in
	 * it. 
	 *
	 * @param how boolean
	 */
	@Override
	public void setEnabled(boolean how)
	{
		super.setEnabled(how);
		dateChooser.setEnabled(how);
		timeChooser.setEnabled(how);
	}
	
	@Override
	public void propertyChange(PropertyChangeEvent evt) {
		firePropertyChange("date",null,getDate());		
	}
	
	/**
	 * Set the range of the year chooser.
	 * 
	 * @param start starting year
	 * @param range number of years displayed
	 */
	public void setYearRange(int start, int range) {
		dateChooser.setYearRange(start,range);
	}
	
	/**
	 * Run test applet.
	 *
	 * @param args command line arguments
	 */
	public static void main(String[] args)
	{
		JDialog d = new JDialog();
		DateTimeSelector dtc = new DateTimeSelector();

		GridBagConstraints c = new GridBagConstraints();

		c.fill = GridBagConstraints.BOTH;

		d.getContentPane().setLayout(new GridBagLayout());
		d.getContentPane().add(dtc, c);

		d.setModal(true);
		d.setVisible(true);
		System.out.println(dtc.getDate().toString());
	}
}

/* __oOo__ */
