/*
 * Copyright (c) 2003-2008 by Cosylab d. d.
 *
 * This file is part of CosyBeans-Common.
 *
 * CosyBeans-Common is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * CosyBeans-Common is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with CosyBeans-Common.  If not, see <http://www.gnu.org/licenses/>.
 */

package com.cosylab.gui.components;

import java.awt.Color;
import java.awt.Cursor;
import java.awt.Desktop;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;
import java.io.IOException;
import java.net.URI;
import java.net.URISyntaxException;

import javax.swing.JLabel;

/**
 * 
 * <code>HyperLinkLabel</code> is an extension of the JLabel which provides a 
 * hyperlink and opens a desktop browser with this link when clicked.
 * The hyper link is aslo visible in the tooltip.
 * <p>
 * This class will work properly only in Java 1.6, since URI launch is not 
 * available in earlier versions.
 * </p>
 *
 * @author <a href="mailto:jaka.bobnar@cosylab.com">Jaka Bobnar</a>
 *
 */
public class HyperLinkLabel extends JLabel {

	private static final long serialVersionUID = 1L;
	private String hyperLink;
	private URI uri;
	
	/**
	 * Constructs a new HyperLinkLabel.
	 *
	 */
	public HyperLinkLabel() {
		super();
		initialize();
	}
	
	/**
	 * Constructs a new HyperLinkLabel with the supplied text visible on the label.
	 * 
	 * @see JLabel
	 * @param text the text to be printed on the label
	 */
	public HyperLinkLabel(String text) {
		super(text);
		initialize();
	}
	
	/**
	 * Constructs a new HyperLinkLabel with the supplied text visible on the
	 * label and the hyperlink attached to it.
	 * 
	 * @param text the text to be printed on the label
	 * @param hyperLink the hyperlink of this label
	 * @throws URISyntaxException if the provided hyperlink is not a valid URI
	 */
	public HyperLinkLabel(String text, String hyperLink) throws URISyntaxException {
		super(text);
		initialize();
		setHyperLink(hyperLink);
		setForeground(Color.BLUE);
	}
	
	/**
	 * Initializes the layout.
	 *
	 */
	private void initialize() {
		setForeground(Color.BLUE);
		addMouseListener(new MouseAdapter(){
			@Override
			public void mouseClicked(MouseEvent e) {
				if (uri == null) return;
				try {
					Desktop.getDesktop().browse(uri);
				} catch (IOException e1) {
					e1.printStackTrace();
				}
			}
			@Override
            public void mouseEntered(MouseEvent e){
				e.getComponent().setCursor(Cursor.getPredefinedCursor(Cursor.HAND_CURSOR));
		    }
		});
	}
	
	/**
	 * Sets the hyperlink for this label. The hyperlink is visible in the 
	 * tooltip of the label and is opened if the label is clicked.
	 * 
	 * @param url the hyperlink
	 * @throws URISyntaxException if the provided hyperlink cannot be transformed
	 * 			to a valid URI
	 */
	public void setHyperLink(String url) throws URISyntaxException {
		if (this.hyperLink != null && this.hyperLink.equals(url)) return;
		String oldValue = this.hyperLink;
		this.hyperLink = url;
		uri = new URI(hyperLink);
		setToolTipText(hyperLink);
		firePropertyChange("hyperLink", oldValue, this.hyperLink);
	}
	
	/**
	 * Returns the hyperlink of this label.
	 * 
	 * @return the hyperlink
	 */
	public String getHyperLink() {
		return hyperLink;
	}
}
