/*
 * Copyright (c) 2003-2008 by Cosylab d. d.
 *
 * This file is part of CosyBeans-Common.
 *
 * CosyBeans-Common is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * CosyBeans-Common is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with CosyBeans-Common.  If not, see <http://www.gnu.org/licenses/>.
 */

package com.cosylab.gui.components;

import java.awt.Dimension;
import java.awt.GridBagConstraints;
import java.awt.GridBagLayout;
import java.awt.Insets;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.beans.Customizer;
import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;

import javax.swing.ButtonGroup;
import javax.swing.JCheckBox;
import javax.swing.JPanel;
import javax.swing.JRadioButton;

import com.cosylab.gui.components.range2.RangedValuePolicy;
import com.cosylab.gui.components.range2.RescalingValuePolicy;
import com.cosylab.gui.components.range2.ShiftValuePolicy;
import com.cosylab.gui.components.range2.TrimValuePolicy;

/**
 * <code>LabelledGaugerCustomizer</code> is a customizer for
 * <code>LabelledGauger</code> which allows customization of its most
 * common properties.
 * 
 * @author <a href="mailto:jaka.bobnar@cosylab.com">Jaka Bobnar</a>
 * @version $Id: LabelledGaugerCustomizer.java,v 1.13 2008-04-22 12:28:40 jbobnar Exp $
 * @see LabelledGauger
 */
public class LabelledGaugerCustomizer extends AbstractDisplayerPanelCustomizer {
	private static final long serialVersionUID = 1L;
	
	/** Value display tab name. */
	public static final String VALUE_DISPLAY = "Value Display";
	/** Tag for minimum property. */
	public static final String GRAPH_MIN = "minimum";
	/** Tag for maximum property. */
	public static final String GRAPH_MAX = "maximum";
	/** Tag for unitsVisibl property. */
	public static final String UNITS_VISIBLE = "unitsVisible";
	/** Tag for format property. */
	public static final String FORMAT = "format";
	/** Tag for units property. */
	public static final String UNITS = "units";
	/** Scale tab name */
	public static final String VISUAL_SPECIFIC = "Value Display/Scale";
	/** Names of the basic visual properties. */
	public static final String[] VISUAL_BASIC_PROPERTIES = { TITLE, TITLE_VISIBLE, RESIZABLE };
	/** Names of the value display properties. */
	public static String[] VALUE_DISPLAY_PROPERTIES = {
		GRAPH_MIN, GRAPH_MAX, UNITS, UNITS_VISIBLE, FORMAT
	};
	
	private class WPanel extends JPanel implements Customizer
	{
		private static final long serialVersionUID = 1L;
		private ButtonGroup scaleGroup;
		private LabelledGauger disp;

		// GUI elements
		private JCheckBox logarithmic;
		private JRadioButton fixedButton;
		private JRadioButton slideButton;
		private JRadioButton stretchButton;
		private boolean initialized = false;
		private ActionListener al = null;
		public WPanel() {
			al = new ActionListener(){

				public void actionPerformed(ActionEvent e) {
					applySettings();
				}
				
			};
			initialize();
		}

		public void applySettings()
		{
			applyDefaultProperties();
			update();
		}

		private JRadioButton getFixedButton()
		{
			if (fixedButton == null) {
				fixedButton = new JRadioButton("Fixed");
				getScaleGroup().add(fixedButton);
				fixedButton.addActionListener(al);
			}

			return fixedButton;
		}

		private JCheckBox getLogarithmic()
		{
			if (logarithmic == null) {
				logarithmic = new JCheckBox("Logarithmic");
				logarithmic.setPreferredSize(new Dimension(150, 21));
				logarithmic.setMinimumSize(new Dimension(75, 21));
				logarithmic.setToolTipText("Use logarithmic scale.");
				logarithmic.addActionListener(al);
			}

			return logarithmic;
		}

		private ButtonGroup getScaleGroup()
		{
			if (scaleGroup == null) {
				scaleGroup = new ButtonGroup();
			}

			return scaleGroup;
		}

		private JRadioButton getSlideButton()
		{
			if (slideButton == null) {
				slideButton = new JRadioButton("Slide");
				getScaleGroup().add(slideButton);
				slideButton.addActionListener(al);
			}

			return slideButton;
		}

		private JRadioButton getStretchButton()
		{
			if (stretchButton == null) {
				stretchButton = new JRadioButton("Stretch");
				getScaleGroup().add(stretchButton);
				stretchButton.addActionListener(al);
			}

			return stretchButton;
		}

		/*
		 * Sets editor values to displayer.
		 */
		private void applyDefaultProperties()
		{
			if (disp == null) {
				return;
			}
			boolean log = getLogarithmic().isSelected();
			disp.setLogarithmicScale(log);

			RangedValuePolicy policy = new TrimValuePolicy();

			if (getFixedButton().isSelected()) {
				policy = new TrimValuePolicy();
			}

			if (getSlideButton().isSelected()) {
				policy = new ShiftValuePolicy();
			}

			if (getStretchButton().isSelected()) {
				policy = new RescalingValuePolicy();
			}

			disp.setValuePolicy(policy);
			LabelledGaugerCustomizer.this.firePropertyChange("logarithmicScale", !log, log);
			LabelledGaugerCustomizer.this.firePropertyChange("scaleMode", null, policy);
		}

		private void initialize()
		{
			if (initialized) {
				return;
			}

			initialized = true;

			setLayout(new GridBagLayout());
			setSize(300, 450);
			setBorder(new GroupBoxBorder("Scale preferences"));

			Insets insets = new Insets(2, 2, 2, 2);

			add(getLogarithmic(),
			    new GridBagConstraints(0, 1, 3, 1, 1.0, 1.0,
			        GridBagConstraints.WEST, GridBagConstraints.HORIZONTAL,
			        insets, 0, 0));

			add(getFixedButton(),
			    new GridBagConstraints(0, 0, 1, 1, 1.0, 1.0,
			        GridBagConstraints.WEST, GridBagConstraints.NONE, insets,
			        0, 0));

			add(getStretchButton(),
			    new GridBagConstraints(1, 0, 1, 1, 1.0, 1.0,
			        GridBagConstraints.WEST, GridBagConstraints.NONE, insets,
			        0, 0));

			add(getSlideButton(),
			    new GridBagConstraints(2, 0, 1, 1, 1.0, 1.0,
			        GridBagConstraints.WEST, GridBagConstraints.NONE, insets,
			        0, 0));
		}

		/*
		 * Updates editor with values from diplayer.
		 */
		private void update()
		{
			if (disp != null) {
				getLogarithmic().setSelected(disp.isLogarithmicScale());

				RangedValuePolicy policy = disp.getValuePolicy();
				if (policy instanceof TrimValuePolicy) {
					getFixedButton().setSelected(true);
				} else if (policy instanceof ShiftValuePolicy) {
					getSlideButton().setSelected(true);
				} else if (policy instanceof RescalingValuePolicy) {
					getStretchButton().setSelected(true);
				}
			}
		}

		public void setObject(Object bean) {
			disp = (LabelledGauger) bean;
			disp.addPropertyChangeListener("scaleMode", new PropertyChangeListener(){
				public void propertyChange(PropertyChangeEvent evt) {
					update();
				}
			});
			disp.addPropertyChangeListener("logarithmicScale", new PropertyChangeListener(){
				public void propertyChange(PropertyChangeEvent evt) {
					update();
				}
			});
			update();
			
		}
	}
	
	
	public LabelledGaugerCustomizer() {
		addCustomizer(VISUAL_SPECIFIC, new WPanel());
		addCustomizerTable(VALUE_DISPLAY, VALUE_DISPLAY_PROPERTIES);
		addCustomizerTable(VISUAL_BASIC, VISUAL_BASIC_PROPERTIES);
		setSize(357, 352);
	}
}
