/*
 * Copyright (c) 2003-2008 by Cosylab d. d.
 *
 * This file is part of CosyBeans-Common.
 *
 * CosyBeans-Common is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * CosyBeans-Common is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with CosyBeans-Common.  If not, see <http://www.gnu.org/licenses/>.
 */

package com.cosylab.gui.components;

import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.event.ActionListener;
import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;

import javax.swing.JComponent;
import javax.swing.JFrame;
import javax.swing.JTextField;
import javax.swing.SwingConstants;

import com.cosylab.application.state.State;
import com.cosylab.events.SetListener;
import com.cosylab.gui.components.customizer.AbstractCustomizerPanel;
import com.cosylab.gui.components.numberfield.NumberDescriptor;
import com.cosylab.gui.components.util.PopupManageable;

/**
 * <code>LabelledNumberField</code> is extended <code>NumberField</code> with
 * title label and units label.
 *
 * @author <a href="mailto:igor.kriznar@cosylab.com">Igor Kriznar</a>
 * @author <a href="mailto:jernej.kamenik@cosylab.com">Jernej Kamenik</a>
 * @version $id$
 */
public class LabelledNumberField extends AbstractNumericDisplayerPanel implements PopupManageable
{ 
	private static final long serialVersionUID = 1L;
	private NumberField number;
	private AbstractCustomizerPanel customizer;

	/**
	 * Creates a new LabelledNumberField object.
	 */
	public LabelledNumberField()
	{
		this(null, "3.2f", null, null, false);
	}

	/**
	 * Creates a new LabelledNumberField object.
	 *
	 * @param format initial format
	 */
	public LabelledNumberField(String format)
	{
		this(null, format, null, null, false);
	}

	/**
	 * Creates a new LabelledNumberField object.
	 *
	 * @param value initial value
	 */
	public LabelledNumberField(Number value)
	{
		this(value, "3.2f", null, null, false);
	}

	/**
	 * Creates a new LabelledNumberField object.
	 *
	 * @param value initial value
	 * @param format initial format
	 */
	public LabelledNumberField(Number value, String format)
	{
		this(value, format, null, null);
	}

	/**
	 * Creates a new LabelledNumberField object.
	 *
	 * @param value initial value
	 * @param format initial format
	 * @param title initial title
	 * @param units initial units
	 */
	public LabelledNumberField(Number value, String format, String title,
	    String units)
	{
		this(value, format, title, units, false);
	}

	/**
	 * Creates a new LabelledNumberField object.
	 *
	 * @param value initial value
	 * @param format initial format
	 * @param title initial title
	 * @param units initial units
	 * @param unitsWithTitle if <code>true</code> units shown in title label
	 *        and not in separate label
	 */
	public LabelledNumberField(Number value, String format, String title,
	    String units, boolean unitsWithTitle)
	{
		this(value, format, title, units, true, true, true, unitsWithTitle);
	}

	/**
	 * Creates a new LabelledNumberField object.
	 *
	 * @param value initial value
	 * @param format initial format
	 * @param title initial title
	 * @param units initial units
	 * @param resizable inital resizable
	 * @param enchanced inital enchanced
	 * @param editable initial editable
	 * @param unitsWithTitle if <code>true</code> units shown in title label
	 *        and not in separate label
	 */
	public LabelledNumberField(Number value, String format, String title,
	    String units, boolean resizable, boolean enchanced, boolean editable,
	    boolean unitsWithTitle)
	{
		super(title, true, Layout.HORIZONTAL, resizable, true, enchanced, 1,
		    Integer.MAX_VALUE, units, !unitsWithTitle, unitsWithTitle);

		setEditable(editable);
		setFormat(format);
		setMaximumValue(new Double(Double.MAX_VALUE));
		setMinimumValue(new Double(-Double.MAX_VALUE));

		if (value != null) {
			setNumberType(value.getClass());
			setNumberValue(value);
		}

		
	}

	/*
	 * (non-Javadoc)
	 * @see com.cosylab.gui.components.AbstractDisplayerPanel#getCustomizer()
	 */
	@Override
    public AbstractCustomizerPanel getCustomizer()
	{
		if (customizer == null) {
			customizer = AbstractCustomizerPanel.findCustomizer(this);
		}

		return customizer;
	}
	
	/**
	 * Sets number of columns on the <code>NumberField</code>.
	 *
	 * @see NumberField#setColumns(int)
	 */
	public void setColumns(int columns)
	{
		int oldValue = getColumns();
		getNumber().setColumns(columns);
		firePropertyChange("columns", oldValue, columns);
	}

	/**
	 * Returns number of columns on the <code>NumberField</code>.
	 *
	 * @see JTextField#getColumns()
	 */
	public int getColumns()
	{
		return getNumber().getColumns();
	}

	/**
	 * Sets the value of type <code>double</code> to be displayed.
	 *
	 * @param newValue
	 */
	public void setDoubleValue(double newValue)
	{
		double oldValue = getDoubleValue();
		getNumber().setDoubleValue(newValue);
		firePropertyChange("doubleValue", oldValue, newValue);
	}

	/**
	 * Returns the value of type <code>double</code>
	 *
	 * @return double
	 */
	public double getDoubleValue()
	{
		return getNumber().getDoubleValue();
	}

	/**
	 * Enables/disables user editing of number.
	 *
	 * @param editable if <code>true</code> user is allowed to set value.
	 */
	public void setEditable(boolean editable)
	{
		boolean oldValue = isEditable();
		getNumber().setEditable(editable);
		firePropertyChange("editable", oldValue, editable);
	}

	/**
	 * Returns <code>true</code> if user is allowed to set value.
	 *
	 * @return boolean <code>true</code> if user is allowed to input number
	 */
	public boolean isEditable()
	{
		return getNumber().isEditable();
	}

	/**
	 * Sets the value of type <code>long</code> to be displayed.
	 *
	 * @param newValue
	 */
	public void setLongValue(long newValue)
	{
		long oldValue = getLongValue();
		getNumber().setLongValue(newValue);
		firePropertyChange("longValue", oldValue, newValue);
	}

	/**
	 * Returns the value of type <code>long</code>
	 *
	 * @return long
	 */
	public long getLongValue()
	{
		return getNumber().getLongValue();
	}

	/**
	 * Sets the maximum allowed value that can be entered in this NumberField.
	 * Setting maximum to null means no maximum is set. This method does not
	 * check for validity of the limit. If maximum is set less than minimum
	 * then no value can be set.
	 *
	 * @param newMaximum Number
	 */
	@Override
    public void setMaximumValue(Number newMaximum)
	{
		Number oldValue = getMaximumValue();
		getNumber().setMaximum(newMaximum);
		super.setMaximumValue(number.getMaximum());
		firePropertyChange("maximumValue", oldValue, newMaximum);
	}

	/**
	 * Returns the maximum value allowed by this NumberField. Can be null if no
	 * maximum is set.
	 *
	 * @return Number maximum allowed value.
	 */
	@Override
    public Number getMaximumValue()
	{
		return getNumber().getMaximum();
	}

	/**
	 * Sets the minimum allowed value that can be entered in this NumberField.
	 * Setting minimum to null means no minimum is set. This method does not
	 * check for validity of the limit. If minimum is set more than maximum
	 * then no value can be set.
	 *
	 * @param newMinimum Number
	 */
	@Override
    public void setMinimumValue(Number newMinimum)
	{
		Number oldValue = getMinimumValue();
		getNumber().setMinimum(newMinimum);
		super.setMinimumValue(number.getMinimum());
		firePropertyChange("minimumValue", oldValue, newMinimum);
	}

	/**
	 * Returns the minimum value allowed by this NumberField. Can be null if no
	 * miminum is set.
	 *
	 * @return Number minimum allowed value
	 */
	@Override
    public Number getMinimumValue()
	{
		return getNumber().getMinimum();
	}

	/**
	 * Sets the number format.
	 *
	 * @param newNumberFormat
	 */
	@Override
    public void setNumberType(Class<? extends Number> newNumberFormat)
	{
		Class<? extends Number> oldValue = getNumberType();
		getNumber().setNumberType(newNumberFormat);
		super.setNumberType(newNumberFormat);
		firePropertyChange("numberType", oldValue, newNumberFormat);
	}

	/**
	 * Returns the type of displayed number.
	 *
	 * @return type of displayed number
	 */
	@Override
    public Class<? extends Number> getNumberType()
	{
		return getNumber().getNumberType();
	}

	/**
	 * Sets the numeric value for this component. New value is subject to
	 * min/max test if any of them is set. If new value does not fall within
	 * these bounds, setting value will do nothing.
	 *
	 * @param newValue Number
	 */
	public void setNumberValue(Number newValue)
	{
		Number oldValue = getNumberValue();
		getNumber().setValue(newValue);
		firePropertyChange("numberValue", oldValue, newValue);
	}

	/**
	 * Returns generic representation of the value using the
	 * <code>Number</code> class.
	 *
	 * @return java.lang.Number
	 */
	public Number getNumberValue()
	{
		return getNumber().getValue();
	}

	/*
	 * (non-Javadoc)
	 * @see com.cosylab.gui.components.AbstractNumericDisplayerPanel#setState(com.cosylab.application.state.State)
	 */
	@Override
    public void setState(State state)
	{
		super.setState(state);

		setEditable(state.getBoolean("editable", isEditable()));
		setColumns(state.getInt("columns", getColumns()));
		setTiltingEnabled(state.getBoolean("tiltingEnabled", isTiltingEnabled()));
	}

	/**
	 * Enables/disables tilting when value is out of bounds.
	 * 
	 * @param b true if tilting should be enabled
	 * @see NumberField#setTiltingEnabled(boolean)
	 */
	public void setTiltingEnabled(boolean b)
	{
		if (isTiltingEnabled() == b) return;
		getNumber().setTiltingEnabled(b);
		firePropertyChange("tiltingEnabled", !b, b);
	}

	/*
	 * (non-Javadoc)
	 * @see com.cosylab.gui.components.AbstractNumericDisplayerPanel#getState()
	 */
	@Override
    public State getState()
	{
		State state = super.getState();

		state.putBoolean("editable", isEditable());
		state.putBoolean("tiltingEnabled", isTiltingEnabled());
		state.putInt("columns", getColumns());

		return state;
	}

	/**
	 * Returns true if tilting is enabled.
	 * 
	 * @return true if tilting enabled
	 * @see #setTiltingEnabled(boolean)
	 * @see NumberField#isTiltingEnabled()
	 */
	public boolean isTiltingEnabled()
	{
		return getNumber().isTiltingEnabled();
	}

	/**
	 * Adds action listener on <code>NumberField</code>.
	 *
	 * @param l
	 */
	public void addActionListener(ActionListener l)
	{
		number.addActionListener(l);
	}

	/**
	 * Adds set listener to the <code>NumberField</code>.
	 *
	 * @see NumberField#addSetListener(SetListener)
	 */
	public void addSetListener(SetListener l)
	{
		getNumber().addSetListener(l);
	}

	/**
	 * Removes action listener from <code>NumberField</code>.
	 *
	 * @param l
	 */
	public void removeActionListener(ActionListener l)
	{
		number.removeActionListener(l);
	}

	/**
	 * Removes set listener from the <code>NumberField</code>.
	 *
	 * @see NumberField#removeSetListener(SetListener)
	 */
	public void removeSetListener(SetListener l)
	{
		getNumber().removeSetListener(l);
	}

	/**
	 * The <code>NumberField</code> component, that does number editing and
	 * rendering.
	 *
	 * @return the <code>NumberField</code> component
	 */
	protected NumberField getNumber()
	{
		if (number == null) {
			number = new NumberField();
			number.setResizable(true);
			number.setSettingPolicy(NumberField.SET_ON_APPLY);
			number.setHorizontalAlignment(SwingConstants.RIGHT);
		}

		return number;
	}
	
	/* (non-Javadoc)
	 * @see com.cosylab.gui.components.DisplayerPanel#getValueComponent()
	 */
	@Override
    protected JComponent getValueComponent()
	{
		return getNumber();
	}
	
	/*
	 * (non-Javadoc)
	 * @see com.cosylab.gui.components.AbstractNumericDisplayerPanel#internalSetEnabled()
	 */
	@Override
    protected void internalSetEnabled()
	{
		super.internalSetEnabled();
		getNumber().setEnabled(isEnabled());
	}

	/*
	 * (non-Javadoc)
	 * @see com.cosylab.gui.components.AbstractNumericDisplayerPanel#internalSetEnhanced()
	 */
	@Override
    protected void internalSetEnhanced()
	{
		super.internalSetEnhanced();
		getNumber().setEnhanced(isEnhanced());
	}

	/*
	 * (non-Javadoc)
	 * @see com.cosylab.gui.components.AbstractNumericDisplayerPanel#internalSetFormat()
	 */
	@Override
    protected void internalSetFormat()
	{
		super.internalSetFormat();

		String tFormat = getFormat();

		if (tFormat == null) {
			tFormat = NumberField.createDefaultFormat(getNumberType());
		}

		getNumber().setFormat(tFormat);
	}

	/*
	 * (non-Javadoc)
	 * @see com.cosylab.gui.components.AbstractNumericDisplayerPanel#internalSetResizable()
	 */
	@Override
    protected void internalSetResizable()
	{
		super.internalSetResizable();
		getNumber().setResizable(isResizable());
		getNumber().revalidate();
	}

	/**
	 * Returns descriptor object, to which visualization of number as strring
	 * is dlegated.
	 *
	 * @return number visualization
	 */
	public NumberDescriptor getNumberDescriptor()
	{
		return getNumber().getNumberDescriptor();
	}

	/**
	 * When this is set with non-null value, this NumberField uses this
	 * descriptor to convert Number to string and back. By this user can
	 * define his own visualization of Number.
	 *
	 * @param numberDescriptor new Number visualization, can be
	 *        <code>null</code>
	 */
	public void setNumberDescriptor(NumberDescriptor numberDescriptor)
	{
		NumberDescriptor oldValue = getNumberDescriptor();
		getNumber().setNumberDescriptor(numberDescriptor);
		firePropertyChange("numberDescriptor", oldValue, numberDescriptor);
	}
	/**
	 * Sets the setting policy of thi Numberdisplayer. SET_AS_TYPED updates the
	 * Numberfield value when typed. SET_ON_APPLY updates the Numberfield only
	 * on actionPerformed events (when the user presses ENTER) SET_ON_EXIT
	 * updates the numberfield when it loses focus
	 *
	 * @param newPolicy new policy
	 * @see NumberField#setSettingPolicy(int)
	 * @throws IllegalArgumentException
	 */
	public void setSettingPolicy(int newPolicy)
	{
		int oldValue = getSettingPolicy();
		getNumber().setSettingPolicy(newPolicy);
		firePropertyChange("settingPolicy", oldValue, newPolicy);
	}
	/**
	 * Returns the currently used seting policy by this Numberfield
	 *
	 * @return int currently used setting policy.
	 * @see NumberField#getSettingPolicy()
	 */
	public int getSettingPolicy()
	{
		return getNumber().getSettingPolicy();
	}
	
	/**
	 * If warnOutOfBounds is set to true, user will be notified when a
	 * value is set ot of the preset bounds.
	 * 
	 * @param b true if warnings are turned on
	 */
	public void setWarnOutOfBounds(Boolean b) {
		Boolean oldValue = isWarnOutOfBounds();
		getNumber().setWarnOutOfBounds(b);
		firePropertyChange("warnOutOfBounds", oldValue, b);
	}
	
	/**
	 * Returns true if the displayer shows a warning when out of bounds.
	 * 
	 * @return true if warning is shown
	 */
	public boolean isWarnOutOfBounds() {
		return getNumber().isWarnOutOfBounds();
	}
	
	/**
	 * Sets the color which indicates the out of bounds state.
	 * 
	 * @param c new out of bounds state color
	 */
	public void setOutOfBoundsColor(Color c) {
		Color oldValue = getOutOfBoundsColor();
		getNumber().setOutOfBoundsColor(c);
		firePropertyChange("outOfBoundColor", oldValue, c);
	}
	
	/**
	 * Returns th color which indicates the out of bounds state.
	 * @return
	 */
	public Color getOutOfBoundsColor() {
		return getNumber().getWarningColor();
	}
	
	/*
	 * (non-Javadoc)
	 * @see com.cosylab.gui.components.AbstractNumericDisplayerPanel#setBackground(java.awt.Color)
	 */
	@Override
	public void setBackground(Color bg) {
		Color fieldBg = getNumber().getBackground();
		super.setBackground(bg);
		getNumber().setBackground(fieldBg);
	}
		
	/**
	 * Run test applet.
	 *
	 * @param args command line parameters
	 */
	public static void main(String[] args)
	{
		JFrame frame = new JFrame("NumberField Testing Applet");
		frame.getContentPane().setLayout(new BorderLayout(10, 10));

		//frame.getContentPane().add(new LabelledNumberField(new Long(10), "%3d",
		//        "Teddy", "Bears"), BorderLayout.CENTER);
		final LabelledNumberField field= new LabelledNumberField(new Double(10),
        "%1.0f", "Current", "A",true);
		field.setTitleMaximumFontSize(12);
		field.setMaximumValue(1000);
		field.setMinimumValue(0);
		field.setDoubleValue(0);
		field.setUnitsVisible(true);
		field.setTiltingEnabled(true);
		frame.getContentPane().add(field, BorderLayout.CENTER);
		frame.setSize(300, 100);
		frame.addWindowListener(new WindowAdapter() {
				@Override
                public void windowClosing(WindowEvent e)
				{
					System.exit(0);
				}
			});
		frame.setVisible(true);
		
		new Thread(new Runnable(){
			public void run() {
				double i = 0;
				while(true) {
					
					
					try {
						Thread.sleep(10000);
					} catch (InterruptedException e) {
						// TODO Auto-generated catch block
						e.printStackTrace();
					}
					field.setDoubleValue(i+=1500);
				}
				
			}
		}).start();
	}

}

/* __oOo__ */
