/*
 * Copyright (c) 2003-2008 by Cosylab d. d.
 *
 * This file is part of CosyBeans-Common.
 *
 * CosyBeans-Common is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * CosyBeans-Common is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with CosyBeans-Common.  If not, see <http://www.gnu.org/licenses/>.
 */

package com.cosylab.gui.components;

import com.cosylab.gui.components.util.IconHelper;

import java.util.Vector;

import javax.swing.Icon;


/**
 * This class extends <code>javax.swing.JLabel</code> so that it allows display
 * of multiple icons. It maintains internal list of icons that associates
 * integer values with <code>Icon</code> objects. This index is guaranteed to
 * be valid even if the desired icon could not be found, is not valid or is
 * null. In such cases the corresponding icon will not be displayed.
 * 
 * <p>
 * Icons can be provided by either the calling class via the
 * <code>javax.swing.Icon</code> object or they can be loaded from the
 * projects resource automatically. Two methods are provided to add the icons:
 * <code>addIcon, addIcons</code> The later is preffered when adding multiple
 * icons at the same time. <br>
 * Use the <code>setIconByIndex</code> method to set which icon should be displayed.<br>
 * The disabled state icon of the JLabel is not overriden and can be still
 * specified.
 * </p>
 * 
 * <p></p>
 *
 * @author <a href="mailto:ales.pucelj@cosylab.com">Ales Pucelj</a>
 * @version $id$
 */
public class MultiIconLabel extends javax.swing.JLabel
{

	private static final long serialVersionUID = 1L;
	/** List storing all icons */
	private final Vector<Icon> iconList = new Vector<>();

	/**
	 * Default constructor for MultiIconLabel.
	 */
	public MultiIconLabel()
	{
		super();
	}

	/**
	 * Constructs a MultiIconLabel and loads the icons from the specified
	 * resource names.
	 * 
	 * <p></p>
	 *
	 * @param resourceNames String[]
	 */
	public MultiIconLabel(String[] resourceNames)
	{
		this();
		addIcons(resourceNames);
	}

	/**
	 * Constructs a MultiIconLabel and loads the icons from the specified
	 * resource names. Also sets the horizontal alignement of the text.
	 * 
	 * <p></p>
	 *
	 * @param resourceNames String[]
	 * @param horizontalAlignement int.
	 */
	public MultiIconLabel(String[] resourceNames, int horizontalAlignement)
	{
		this();
		addIcons(resourceNames);
		setHorizontalAlignment(horizontalAlignement);
	}

	/**
	 * Constructs a MultiIconLabel and adds the icons to the icon list.
	 *
	 * @param icons Icon[]
	 */
	public MultiIconLabel(Icon[] icons)
	{
		this();
		addIcons(icons);
	}

	/**
	 * Constructs a MultiIconLabel and adds the icons to the icon list. Also
	 * sets the horizontal alignement of the text.
	 * 
	 * <p></p>
	 *
	 * @param icons javax.swing.Icon[]
	 * @param horizontalAlignment int
	 */
	public MultiIconLabel(javax.swing.Icon[] icons, int horizontalAlignment)
	{
		this(icons);
		setHorizontalAlignment(horizontalAlignment);
	}

	/**
	 * Appends the icon specified by resourceName to the end of internal list
	 * of icons. This method calls addIcon(Icon) and loadFromResource(String).
	 * If resourceName does not exist in the projects resources, method adds
	 * empty icon.
	 * 
	 * <p></p>
	 *
	 * @param resourceName java.lang.String
	 */
	public void addIcon(String resourceName)
	{
		addIcon(loadFromResource(resourceName));
	}

	/**
	 * Appends the icon to the list of icons.
	 * 
	 * <p></p>
	 *
	 * @param icon javax.swing.Icon
	 */
	public void addIcon(Icon icon)
	{
		iconList.add(icon);
	}

	/**
	 * Appends all icons specified by resourceNames to the end of internal list
	 * of icons. For adding multiple icons, this method outperforms individual
	 * calls to <code>addIcon</code> method.
	 * 
	 * <p>
	 * After calling this method, the list of icons will grow by the size of
	 * the array, regardless of its contentrs. If any of the icons is null or
	 * of invalid type, the associated icon will not be displayed.
	 * </p>
	 * 
	 * <p></p>
	 *
	 * @param resourceNames java.lang.String[]
	 */
	public void addIcons(String[] resourceNames)
	{
		if (resourceNames != null) {
			for (int i = 0; i < resourceNames.length; i++) {
				addIcon(resourceNames[i]);
			}
		}
	}

	/**
	 * Appends all icons specified by icons parameter to the internal list of
	 * icons. For adding multiple icons, this method outperforms individual
	 * calls to <code>addIcon</code> method.
	 * 
	 * <p>
	 * After calling this method, the list of icons will grow by the size of
	 * the array, regardless of its contentrs. If any of the icons is null or
	 * of invalid type, the associated icon will not be displayed. <br>
	 * </p>
	 *
	 * @param icons javax.swing.Icon[]
	 */
	public void addIcons(Icon[] icons)
	{
		if (icons != null) {
			for (int i = 0; i < icons.length; i++) {
				addIcon(icons[i]);
			}
		}
	}

	/**
	 * Returns the Icon resource assigned to the specified index. If the icon
	 * has not been defined, returns null.
	 * 
	 * <p></p>
	 *
	 * @param index int
	 *
	 * @return javax.swing.Icon
	 */
	public Icon getIconByIndex(int index)
	{
		if ((index >= 0) && (index < iconList.size())) {
			return iconList.get(index);
		}

		return null;
	}

	/**
	 * Loads the Icon from the projects resource. Returns null if the resource
	 * does not exist.
	 * 
	 * <p></p>
	 *
	 * @param resourceName java.lang.String
	 *
	 * @return javax.swing.Icon
	 */
	protected Icon loadFromResource(String resourceName)
	{
		return IconHelper.createIcon(resourceName);
	}

	/**
	 * Displays the icon specified by index. If the icon with this index has
	 * not been defined, no icon will be displayed.
	 * 
	 * <p></p>
	 *
	 * @param index int
	 */
	public void setIconByIndex(int index)
	{
		setIcon(getIconByIndex(index));
	}

	/**
	 * Returns number of available icons.
	 *
	 * @return Number of available icons.
	 */
	public int getIconCount()
	{
		return iconList.size();
	}
}

/* __oOo__ */
