/*
 * Copyright (c) 2003-2008 by Cosylab d. d.
 *
 * This file is part of CosyBeans-Common.
 *
 * CosyBeans-Common is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * CosyBeans-Common is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with CosyBeans-Common.  If not, see <http://www.gnu.org/licenses/>.
 */

package com.cosylab.gui.components;

import com.cosylab.gui.components.util.CosyUIElements;
import com.cosylab.gui.components.util.FontHelper;
import com.cosylab.gui.components.util.IconHelper;

import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Component;
import java.awt.Container;
import java.awt.Font;

import javax.swing.AbstractButton;
import javax.swing.Action;
import javax.swing.Icon;
import javax.swing.JComponent;
import javax.swing.JPanel;


/**
 * TitledPanel is an extension of the Swing's JPanel, which contains a
 * <code>PanelTitleBar</code> at the top. The rest of the panel's space is
 * occupied by the <code>content</code> read-write property, which  is an
 * arbitrary <code>Component</code>. By default, the content property is set
 * to <code>null</code>. <br>
 * The title bar shows an <code>Icon</code> which is passed as a parameter to
 * the constructor, and  a <code>String</code> read-write property named
 * title. Toolbar buttons can be added to the panel using the
 * <code>addTitleBarButton()</code> method, passing a <code>JButton</code> as
 * a parameter. The buttons added to the panel may be associated with an
 * arbitrary action, which will not be affected by the TitledPanel.Note that
 * buttons will not retain their original color,  border or size after being
 * added to the TitledPanel. <br>
 * Width of the title bar (and thus also the size of the toolbar buttons) and
 * the space between the panel's border and its content  can both be
 * determined when constructing the panel.
 *
 * @author <a href="mailto:miha.kadunc@cosylab.com">Miha Kadunc</a>
 * @version $id$
 *
 * @see com.cosylab.gui.components.PanelTitleBar
 * @see javax.swing.JPanel
 */
public class TitledPanel extends JPanel
{
	private static final long serialVersionUID = 1L;
	protected PanelTitleBar bar = null;
	private Container contentPane = null;

	//private int toolBarHeight;
	//private int borderWidth = 0;
	private boolean initialized = false;

	/**
	 * Constructor for TitledPanel. The name parameter is not visible in the
	 * panel, so it can be set to null. Title and Icon will be used for
	 * constructing the <code>PanelTitleBar</code>. The title bar will be
	 * created with the default width of 24 pixels and there will be no space
	 * between the border of the panel and its contents (titlebar and
	 * <code>content</code> component)
	 *
	 * @param name The name of the panel.
	 * @param title The title which will be displayed in the title bar.
	 * @param icon The icon to be rendered in the upper-left corner of the
	 *        panel.
	 */
	public TitledPanel(String name, String title, Icon icon)
	{
		this(name, title, icon, 20);
	}

	private TitledPanel(String name, PanelTitleBar bar)
	{
		super(new BorderLayout());
		this.bar = bar;
		setName(name);
	}

	/**
	 * Creates a new TitledPanel object.
	 */
	public TitledPanel()
	{
		this("TitledPanel", "Title",
		    IconHelper.createIcon("icons/general/ComposeMail16.gif"));
	}

	/**
	 * Constructor for TitledPanel. The name parameter is not visible in the
	 * panel, so it can be set to null. Title and Icon will be used for
	 * constructing the <code>PanelTitleBar</code>.
	 *
	 * @param name The name of the panel.
	 * @param title The title which will be displayed in the title bar.
	 * @param icon The icon to be rendered in the upper-left corner of the
	 *        panel.
	 * @param titleBarHeight Height of the title bar rendered in the top area
	 *        of the panel.
	 */
	public TitledPanel(String name, String title, Icon icon, int titleBarHeight)
	{
		this(name, new PanelTitleBar(title, icon, titleBarHeight));
	}

	/**
	 * Adds a <code>JButton</code> to this panel's title bar. Note that all the
	 * user has to provide is the button's icon and an associated action. All
	 * the behaviour needed for the button to fit into the title bar will be
	 * added to the button by the PanelTitleBar.  The button will be resized
	 * to a square with the same size as the  height of the bar. Its
	 * background color will be changed  to that of the bar and the border
	 * will be removed.  Some mouse and keyboard sensitivity will be added to
	 * the button.
	 *
	 * @param button The button to be added to the title bar.
	 */
	public void addTitleBarButton(AbstractButton button)
	{
		bar.addButton(button);
	}

	/**
	 * Adds a button to this panel's title bar at specified index.
	 *
	 * @param index location of added button
	 * @param button
	 */
	public void addTitleBarButton(int index, AbstractButton button)
	{
		bar.addButton(index, button);
	}

	/**
	 * Removes a button from this panel's title bar.
	 *
	 * @param button
	 */
	public void removeTitleBarButton(AbstractButton button)
	{
		bar.removeButton(button);
	}

	/**
	 * Adds a button activating the specified action to this panel's title bar.
	 *
	 * @param action
	 */
	public void addTitleBarAction(Action action)
	{
		bar.addActionButton(action);
	}

	/**
	 * Adds a button activating the specified action to this panel's title bar
	 * at specified index.
	 *
	 * @param index location of added action
	 * @param action
	 */
	public void addTitleBarAction(int index, Action action)
	{
		bar.addActionButton(index, action);
	}

	/**
	 * @see PanelTitleBar#containsAction(Action)
	 */
	public boolean containsTitleBarAction(Action action)
	{
		return bar.containsAction(action);
	}

	/**
	 * @see PanelTitleBar#containsButton(AbstractButton)
	 */
	public boolean containsTitleBarAction(AbstractButton button)
	{
		return bar.containsButton(button);
	}

	/**
	 * Removes the button associated with the specified action from this
	 * panel's title bar.
	 *
	 * @param action
	 */
	public void removeTitleBarAction(Action action)
	{
		bar.removeActionButton(action);
	}

	private void initialize()
	{
		initialized = true;
		super.setFocusable(false);
		bar.getTitleLabel().setFont(FontHelper.getDefaultFont());
		bar.setVisible(true);

		super.add(bar, BorderLayout.NORTH);
	}

	/**
	 * Sets this panel's title, which is diaplayed in the title bar. If the
	 * title is too long for the space which is available to the panel, it is
	 * trimmed and the missing parts are replaced by ...
	 *
	 * @param title The text that should be displayed as the title.
	 */
	public void setTitle(String title)
	{
		bar.setTitle(title);
	}

	/**
	 * Gets the title.
	 *
	 * @return Returns the <code>String</code> displayed as the title.
	 *
	 * @see #getTitle
	 */
	public String getTitle()
	{
		return bar.getTitle();
	}

	/**
	 * Returns the left background color of this panel's title bar.
	 *
	 * @return Color
	 */
	public Color getTitleBackgroundLeft()
	{
		return bar.getBackgroundStart();
	}

	/**
	 * Gets the right background color of this panel's title bar.
	 *
	 * @return Color
	 */
	public Color getTitleBackgroundRight()
	{
		return bar.getBackground();
	}

	/**
	 * Sets the background colors of this panel's title bar.
	 *
	 * @param left
	 * @param right
	 */
	public void setTitleBackground(Color left, Color right)
	{
		if (left.equals(right)) {
			bar.setBackground(right);
			bar.getTitleLabel().setGradientEnabled(false);
		} else {
			bar.getTitleLabel().setGradientEnabled(true);
			bar.setBackground(left, right);
		}
	}

	/**
	 * Sets the background colors of this panel's title bar.
	 *
	 * @param left
	 */
	public void setTitleBackgroundLeft(Color left)
	{
		bar.getTitleLabel().setGradientEnabled(true);
		bar.setBackgroundStart(left);
	}

	/**
	 * Sets the background colors of this panel's title bar.
	 *
	 * @param right
	 */
	public void setTitleBackgroundRight(Color right)
	{
		bar.getTitleLabel().setGradientEnabled(true);
		bar.setBackground(right);
	}

	/**
	 * Return this panel's title bar icon.
	 *
	 * @return Icon
	 */
	public Icon getIcon()
	{
		return bar.getIcon();
	}

	/**
	 * Sets the icon of this panel's title bar.
	 *
	 * @param icon
	 */
	public void setIcon(Icon icon)
	{
		bar.setIcon(icon);
	}

	/**
	 * Gets the <code>content</code> Container.
	 *
	 * @return Returns the Component contained in this panel as the content
	 *         pane.
	 */
	public Container getContentPane()
	{
		if (contentPane == null) {
			contentPane = new JPanel();
			contentPane.setName(getName() + "ContentPane");
		}

		return contentPane;
	}

	/**
	 * Sets the content. The content is added to the panel below the title bar,
	 * covering all the available area.
	 *
	 * @param content The content to set
	 */
	public void setContentPane(Container content)
	{
		if (!initialized) {
			initialize();
		}

		if (this.contentPane != content) {
			if (this.contentPane != null) {
				remove(this.contentPane);
			}

			if ((content instanceof JComponent)
			    && (((JComponent)content).getBorder() == null)) {
				((JComponent)content).setBorder(CosyUIElements.getPlainBorder(
				        true));
			}

			this.contentPane = content;
			super.add(content, BorderLayout.CENTER);
		}

		invalidate();
		repaint();
	}

	protected GradientLabel getTitleBarLabel()
	{
		return bar.getTitleLabel();
	}

	/**
	 * Sets the visibility of the title bar.
	 *
	 * @param visible true if title bar is visible
	 */
	public void setTitleBarVisible(boolean visible)
	{
		boolean old = bar.isVisible();

		if (old != visible) {
			bar.setVisible(visible);
			firePropertyChange("titleBarVisible", old, visible);
			revalidate();
			repaint();
		}
	}

	/**
	 * Returns true if title bar is visible or false otherwise.
	 *
	 * @return true if title bar is visible
	 */
	public boolean isTitleBarVisible()
	{
		return bar.isVisible();
	}

	/*
	 * (non-Javadoc)
	 * @see java.awt.Container#add(java.awt.Component, int)
	 */
	@Override
    public Component add(Component comp, int index)
	{
		if (!initialized) {
			initialize();
		}

		if (comp instanceof AbstractButton) {
			addTitleBarButton((AbstractButton)comp);

			return comp;
		}

		if (comp != bar && comp != contentPane) {
			setContentPane((JComponent)comp);
		}

		return comp;
	}

	/*
	 * (non-Javadoc)
	 * @see java.awt.Container#add(java.awt.Component, java.lang.Object, int)
	 */
	@Override
    public void add(Component comp, Object constraints, int index)
	{
		if (!initialized) {
			initialize();
		}

		if (comp instanceof AbstractButton) {
			addTitleBarButton((AbstractButton)comp);

			return;
		}

		if (comp != bar && comp != contentPane) {
			setContentPane((JComponent)comp);
		}
	}

	/*
	 * (non-Javadoc)
	 * @see java.awt.Container#add(java.awt.Component, java.lang.Object)
	 */
	@Override
    public void add(Component comp, Object constraints)
	{
		if (!initialized) {
			initialize();
		}

		if (comp instanceof AbstractButton) {
			addTitleBarButton((AbstractButton)comp);

			return;
		}

		if (comp != bar && comp != contentPane) {
			setContentPane((JComponent)comp);
		}
	}

	/*
	 * (non-Javadoc)
	 * @see java.awt.Container#add(java.awt.Component)
	 */
	@Override
    public Component add(Component comp)
	{
		if (!initialized) {
			initialize();
		}

		if (comp instanceof AbstractButton) {
			addTitleBarButton((AbstractButton)comp);
		} else if (comp != bar && comp != contentPane) {
			setContentPane((JComponent)comp);
		}

		return comp;
	}

	/*
	 * (non-Javadoc)
	 * @see java.awt.Container#add(java.lang.String, java.awt.Component)
	 */
	@Override
    public Component add(String name, Component comp)
	{
		if (!initialized) {
			initialize();
		}

		if (comp instanceof AbstractButton) {
			addTitleBarButton((AbstractButton)comp);

			return comp;
		}

		if (comp != bar && comp != contentPane) {
			setContentPane((JComponent)comp);
		}

		return comp;
	}

	/*
	 * (non-Javadoc)
	 * @see javax.swing.JComponent#setFont(java.awt.Font)
	 */
	@Override
    public void setFont(Font font)
	{
		if (bar != null) {
			bar.setFont(font);
		}

		super.setFont(font);
	}

	/*
	 * (non-Javadoc)
	 * @see javax.swing.JComponent#addNotify()
	 */
	@Override
    public void addNotify()
	{
		if (!initialized) {
			initialize();
		}

		super.addNotify();
	}

	/*
	 * (non-Javadoc)
	 * @see java.awt.Container#remove(java.awt.Component)
	 */
	@Override
    public void remove(Component comp)
	{
		if (comp instanceof AbstractButton) {
			removeTitleBarButton((AbstractButton)comp);

			return;
		}

		super.remove(comp);
	}
}

/* __oOo__ */
