/*
 * Copyright (c) 2003-2008 by Cosylab d. d.
 *
 * This file is part of CosyBeans-Common.
 *
 * CosyBeans-Common is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * CosyBeans-Common is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with CosyBeans-Common.  If not, see <http://www.gnu.org/licenses/>.
 */

package com.cosylab.gui.components.customizer;

import java.awt.BorderLayout;
import java.awt.Component;
import java.awt.FlowLayout;
import java.awt.Font;
import java.awt.font.FontRenderContext;
import java.beans.BeanInfo;
import java.beans.Customizer;
import java.beans.IntrospectionException;
import java.beans.Introspector;
import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;
import java.beans.PropertyDescriptor;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.LinkedHashSet;
import java.util.List;

import javax.swing.AbstractAction;
import javax.swing.AbstractListModel;
import javax.swing.JButton;
import javax.swing.JDialog;
import javax.swing.JList;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.JSplitPane;
import javax.swing.event.ListSelectionEvent;
import javax.swing.event.ListSelectionListener;

import com.cosylab.application.AutoScalableDialog;

/**
 * <code>AbstractCustomizerPanel</code> is the base class which implements 
 * the <code>java.beans.Customizer</code>. This class itself is a container
 * which does not provide any customization functionality. It serves as
 * a container, which can hold other implementations of Customizer and present
 * all of them in a tabbed display.
 * 
 * @author Alen Vrecko
 * 
 */
public abstract class AbstractCustomizerPanel extends JSplitPane implements Customizer {

	private static final long serialVersionUID = -1910612715311768339L;
	private Object bean;
	private JDialog dialog;
	private LinkedHashSet<Customizer> customizers;
	private HashMap<String, Component> aspectToComponent;
	private AspectModel model;
	private JList<String> list;
	private JScrollPane rightPanel;

	class AspectModel extends AbstractListModel<String> {
		private static final long serialVersionUID = 1L;
		private List<String> aspects= new ArrayList<>();
		
		/**
		 * Creates new instance of AbstractCustomizerPanel.AspectModel.
		 */
		public AspectModel() {
		}

		/* (non-Javadoc)
		 * @see javax.swing.ListModel#getSize()
		 */
		public int getSize() {
			return aspects.size();
		}

		/* (non-Javadoc)
		 * @see javax.swing.ListModel#getElementAt(int)
		 */
		public String getElementAt(int index) {
			return aspects.get(index);
		}
		
		public void add(String name) {
			if (aspects.contains(name)) {
				return;
			}
			aspects.add(name);
			Collections.sort(aspects);
			fireContentsChanged(this,0,getSize()-1);
		}
		
	}
	
	/**
	 * The default constructor.
	 * 
	 */
	public AbstractCustomizerPanel() {
		initialize();
	}

	/**
	 * Adds a Customizer to this container. NOTE: The Customizer should be a
	 * simple panel.
	 * 
	 * @param aspect
	 * @param customizer
	 */
	public void addCustomizer(String aspect, Customizer customizer) {

		model.add(aspect);
		customizers.add(customizer);
		customizer.addPropertyChangeListener(new PropertyChangeListener() {

			public void propertyChange(PropertyChangeEvent evt) {
				firePropertyChange(evt.getPropertyName(), evt.getOldValue(), evt
						.getNewValue());

			}
		});

		aspectToComponent.put(aspect, (Component) customizer);
		
		// Sets divider location to fit current aspect, if it doesn't already
		Font f = (new JScrollPane()).getFont();
		int l = (int) f.getStringBounds(aspect, new FontRenderContext(f.getTransform(), false, false)).getWidth();
		l += 25;
		if (getDividerLocation() < l) setDividerLocation(l);
	}

	/**
	 * Adds a placeholder for aspect.
	 * 
	 * @param aspect
	 * @param panel
	 */
	public void addCustomizerPlaceholder(String aspect, JPanel panel) {
		model.add(aspect);
		aspectToComponent.put(aspect, panel);
	}

	/**
	 * Creates a customizer for the table.
	 * 
	 * @param aspect
	 * @param properties
	 */
	public void addCustomizerTable(String aspect, String... properties) {
		addCustomizer(aspect, CustomizerFactory.createCustomizer(properties));
	}

	/**
	 * Creates a customizer from the properties.
	 * 
	 * @param aspect
	 * @param properties
	 */
	public void addCustomizerTable(String aspect, PropertyDescriptor... properties) {
		addCustomizer(aspect, CustomizerFactory.createCustomizer(properties));
	}

	/**
	 * Use this to put inherit from other customizers.
	 * 
	 * @param abstractCustomizer
	 */
	public void addCustomizer(AbstractCustomizerPanel abstractCustomizer) {
		// puts all aspects and aspect <-> component mappings to this Customizer
		for (int i=0; i< abstractCustomizer.model.getSize(); i++) {
			model.add(abstractCustomizer.model.getElementAt(i));
		}
		this.aspectToComponent.putAll(abstractCustomizer.aspectToComponent);

		// puts all registered customizers from the other one tho this one

		for (Customizer customizer : abstractCustomizer.customizers) {
			this.customizers.add(customizer);

		}

		// register the event delegation

		abstractCustomizer.addPropertyChangeListener(new PropertyChangeListener() {

			public void propertyChange(PropertyChangeEvent evt) {
				firePropertyChange(evt.getPropertyName(), evt.getOldValue(), evt
						.getNewValue());

			}
		});

	}

	/**
	 * Tries to determine customizer for provided visual component in following
	 * order:
	 * 
	 * <ol>
	 * <li> Checks BeanInfo for the component.</li>
	 * <li> Tries to guess from the component class name. If component is name
	 * is MyComponent, then tries to load customizer class
	 * MyComponentCustomizer.</li>
	 * </ol>
	 * 
	 * @param disp
	 *            the component fo which ccustomizer is returned
	 * @return customizer foor the component
	 */
	public static AbstractCustomizerPanel findCustomizer(Object bean) {

		// step 1 obtain from the BeanInfo
		Class<?> beanType = bean.getClass();
		Class<?> customizerClass;
		AbstractCustomizerPanel c=null;

		try {
			BeanInfo beanInfo = Introspector.getBeanInfo(beanType);

			customizerClass = beanInfo.getBeanDescriptor().getCustomizerClass();

			if (customizerClass != null) {
				// determine if we have an AbstractCustomizerPanel

				if (AbstractCustomizerPanel.class.isAssignableFrom(customizerClass)) {
					try {
						c= (AbstractCustomizerPanel) customizerClass.newInstance();
					} catch (InstantiationException e) {
						e.printStackTrace();
					} catch (IllegalAccessException e) {
						e.printStackTrace();
					}
				}
			}

		} catch (IntrospectionException e) {
			e.printStackTrace();
		}
		
		if (c == null) {
			Class<?> beanClass = beanType;
			while (true) {
    			// step 2 : Obtain from file name
    			try {
    	
    				customizerClass = Class.forName(beanClass.getName() + "Customizer");
    				if (customizerClass != null) {
    					// determine if we have an AbstractCustomizerPanel
    	
    					if (AbstractCustomizerPanel.class.isAssignableFrom(customizerClass)) {
    						try {
    							c= (AbstractCustomizerPanel) customizerClass.newInstance();
    						} catch (InstantiationException e) {
    							e.printStackTrace();
    						} catch (IllegalAccessException e) {
    							e.printStackTrace();
    						}
    					}
    					break;
    				}
    			} catch (ClassNotFoundException e) {
    				//ignore
    				beanClass = beanClass.getSuperclass();
    				if (beanClass.equals(Object.class)) break;
    			}
			}
		}
		if (c!=null) {
			c.setObject(bean);
		}
		return c;
	}

	private void initialize() {
		customizers = new LinkedHashSet<>();
		model= new AspectModel();
		aspectToComponent = new HashMap<>();
		list = new JList<>();
		list.setModel(model);
		list.addListSelectionListener(new ListSelectionListener() {

			public void valueChanged(ListSelectionEvent e) {

				Component component = aspectToComponent.get(list.getSelectedValue());

				if (component != null) {
					rightPanel.setViewportView(component);
				}

			}
		});

		setLeftComponent(new JScrollPane(list));
		
		rightPanel= new JScrollPane();
		setRightComponent(rightPanel);
		setSize(500,300);
		setDividerLocation(30);
	}

	/*
	 * (non-Javadoc)
	 * @see java.beans.Customizer#setObject(java.lang.Object)
	 */
	public void setObject(Object bean) {
		this.bean = bean;

		for (Customizer customizer : customizers) {
			customizer.setObject(bean);
		}
		
		if (getName()==null) {
			String s= bean.getClass().getName();
			int i= s.lastIndexOf('.');
			if (i>-1 && i+1<s.length()) {
				setName(s.substring(i+1));
			} else {
				setName(s);
			}
		}
	}

	/**
	 * <p>
	 * Shows Customizer dialog centered around customized displayer.
	 * </p>
	 * 
	 * <p>
	 * Dialog title is composed from Customizer name. Initial size is set to the
	 * size of Customizer.
	 * </p>
	 * 
	 * @return a <code>JDialog</code> instance with customizer
	 * 
	 */
	public JDialog showDialog() {
		return showDialog(bean instanceof Component ? (Component) bean : null);
	}

	/**
	 * <p>
	 * Shows customizer dialog centered around specified component.
	 * </p>
	 * 
	 * <p>
	 * Dialog title is composed from customiser name. Initial size is set to the
	 * size of customiser.
	 * </p>
	 * 
	 * @param c
	 *            a component on which customizer dialog to be centered
	 * 
	 * @return a <code>javax.swing.JDialog</code> instance with customizer
	 * 
	 */
	public JDialog showDialog(Component c) {
		if (dialog == null) {
			dialog = new AutoScalableDialog();
			dialog.setTitle(getName() + " Customizer");
			dialog.setModal(false);
			dialog.setSize(getSize().width + dialog.getInsets().left
					+ dialog.getInsets().right, getSize().height + dialog.getInsets().top
					+ dialog.getInsets().bottom);
			dialog.getContentPane().setLayout(new BorderLayout());
			dialog.getContentPane().add(this, BorderLayout.CENTER);
			JPanel p = new JPanel();
			p.setLayout(new FlowLayout(FlowLayout.RIGHT));
			p.add(new JButton(new AbstractAction("Close") {
				private static final long serialVersionUID = 1L;
				public void actionPerformed(java.awt.event.ActionEvent e) {
					dialog.dispose();
				}
			}));
			dialog.getContentPane().add(p, BorderLayout.SOUTH);
			dialog.setLocationRelativeTo(c);
		}

		dialog.setVisible(false);
		dialog.setVisible(true);
		dialog.toFront();
		return dialog;
	}
	
	/**
	 * Hides the dialog if it is visible.
	 *
	 */
	public void closeDialog() {
		if (dialog != null) {
			dialog.setVisible(false);
		}
	}
	
	/**
	 * Selects an aspect in the aspect list.
	 * 
	 * @param aspect the aspect to select
	 */
	public void selectAspect(String aspect) {
		list.setSelectedValue(aspect, true);
	}

}
