/*
 * Copyright (c) 2003-2008 by Cosylab d. d.
 *
 * This file is part of CosyBeans-Common.
 *
 * CosyBeans-Common is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * CosyBeans-Common is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with CosyBeans-Common.  If not, see <http://www.gnu.org/licenses/>.
 */

package com.cosylab.gui.components.gauger;

import java.awt.Point;
import java.awt.geom.Point2D;


/**
 * Class representing polar point defined by radius and angle.
 *
 * @author <a href="mailto:ales.pucelj@cosylab.com">Ales Pucelj</a>
 * @version $id$
 */
public class PolarPoint
{
	private static final double DEG_TO_RAD = Math.PI / 180.0;
	private static final double RAD_TO_DEG = 180.0 / Math.PI;

	private double r;

	private double phi;

	/**
	 * Creates a new PolarPoint object.
	 */
	public PolarPoint()
	{
	}

	/**
	 * Creates a new PolarPoint object.
	 *
	 * @param r Radius parameter.
	 * @param phi Angle parameter in radians.
	 */
	public PolarPoint(double r, double phi)
	{
		this.r = r;
		this.phi = phi;
	}

	/**
	 * Converts degrees to radians.
	 *
	 * @param angle angle in degrees.
	 *
	 * @return Angle in radians.
	 */
	public static final double degToRad(double angle)
	{
		return angle * DEG_TO_RAD;
	}

	/**
	 * Converts radians to degrees.
	 *
	 * @param angle in radians.
	 *
	 * @return angle in degrees.
	 */
	public static final double radToDeg(double angle)
	{
		return angle * RAD_TO_DEG;
	}

	/**
	 * Converts polar point to cartesian coordinates.
	 *
	 * @param r Radius parameter.
	 * @param phi Angle parameter.
	 *
	 * @return Point in cartesian space.
	 */
	public static final Point2D polarToCartesian(double r, double phi)
	{
		double phiInRadians = degToRad(phi);
		double x = (r * Math.cos(phiInRadians));
		double y = -(r * Math.sin(phiInRadians));

		return new Point2D.Double(x, y);
	}

	/**
	 * Converts polar point to cartesian coordinates. Result will be placed in
	 * point specified. This method will not create new instance of point to
	 * hold the result, but the provided instance will be returned.
	 *
	 * @param p Point which will hold the result.
	 * @param r Radius parameter.
	 * @param phi Angle parameter.
	 *
	 * @return p parameter.
	 */
	public static final Point2D polarToCartesian(Point2D p, double r, double phi)
	{
		double rPhi = degToRad(phi);
		p.setLocation((r * Math.cos(rPhi)), -(r * Math.sin(rPhi)));

		return p;
	}

	/**
	 * Converts polar point to cartesian coordinates.
	 *
	 * @param pp Point in polar coordinates.
	 * @param p Point in cartesian coordinates.
	 */
	public static final void polarToCartesian(PolarPoint pp, Point2D p)
	{
		double rPhi = degToRad(pp.phi);

		p.setLocation(pp.r * Math.cos(rPhi), pp.r * Math.sin(rPhi));
	}

	/**
	 * Converts this point to cartesian coordinate representation.
	 *
	 * @return Point in cartesian coordinates.
	 */
	public Point2D toPoint()
	{
		return polarToCartesian(r, phi);
	}

	/**
	 * Converts this point to cartesian coordinates and puts result in provided
	 * instance of point.
	 *
	 * @param p Object that will hold the result.
	 *
	 * @return Resulting point, which is the same object as one passed as 
	 * 		   parameter.
	 */
	public Point2D toPoint(Point p)
	{
		return polarToCartesian(p, r, phi);
	}
}

/* __oOo__ */
