/*
 * Copyright (c) 2003-2008 by Cosylab d. d.
 *
 * This file is part of CosyBeans-Common.
 *
 * CosyBeans-Common is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * CosyBeans-Common is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with CosyBeans-Common.  If not, see <http://www.gnu.org/licenses/>.
 */

package com.cosylab.gui.components.gauger;

import java.awt.BasicStroke;
import java.awt.Color;
import java.awt.Graphics2D;
import java.awt.Point;
import java.awt.Rectangle;
import java.awt.geom.Line2D;

import com.cosylab.gui.components.Gauger;


/**
 * Component defining and rendering the needle.
 *
 * @author <a href="mailto:ales.pucelj@cosylab.com">Ales Pucelj</a>
 * @version $id$
 */
public class ThinNeedle implements ActiveElement
{
	/* Scale appearance */

	/** Relative length of needle as fraction of scale space height */
	private final static double RELATIVE_NEEDLE_SIZE = 0.6;
	private Gauger gauger;
	private boolean changeSignificant;
	private Point p;
	private double angle;
	private Rectangle bounds = new Rectangle(0, 0, 0, 0);
	private double currentPosition = 0.5;
	protected ScaleTransform transform = null;
	//private Line poly = new Polygon(new int[3], new int[3], 3);
	
	private Point p1 = new Point(0, 0);
	private Point p2 = new Point(0, 0);
	
	private Line2D line = new Line2D.Double();

	/**
	 * Needle constructor comment.
	 *
	 * @param gauger on which the needle will be rendered.
	 */
	public ThinNeedle(Gauger gauger)
	{
		p = new Point();
		angle = 0;
		this.gauger = gauger;
		changeSignificant = true;
	}

	/**
	 * Updates the polygon representation of the needle.
	 */
	protected synchronized void calculatePolygon()
	{
		double sin = Math.sin(angle-Math.PI/2.0);
		double cos = Math.cos(angle-Math.PI/2.0);

		double x1 = (int)(p1.x * sin - p1.y * cos + p.x);
		double y1 = (int)(p1.x * cos + p1.y * sin + p.y);

		double x2 = (int)(p2.x * sin - p2.y * cos + p.x);
		double y2 = (int)(p2.x * cos + p2.y * sin + p.y);
		
		line.setLine(x1, y1, x2, y2);
		
		bounds = line.getBounds();
		bounds.grow(3, 3);
	}

	private synchronized final void invalidate()
	{
		double size = transform.scaleHeight(0.5) * RELATIVE_NEEDLE_SIZE;
		p2.x = 0;
		p2.y = (int)size;
	}

	/**
	 * Returns bounding rectangle of needle.
	 *
	 * @return Bounding rectangle of needle.
	 */
	public Rectangle getBounds()
	{
		return bounds;
	}

	/**
	 * Returns true if needle has moved enough to require immediate redraw.
	 *
	 * @return True if needle should be redrawn immediatelly.
	 */
	public boolean isChangeSignificant()
	{
		return changeSignificant;
	}

	protected void updatePosition(double x)
	{
		if((transform != null)) {
			transform.mapUVtoXY(x, 0, p);
			angle = transform.getAngle(x);
			calculatePolygon();
		}
	}

	/**
	 * Sets position of needle in scale space coordinates. Value is relative
	 * position between 0.0 and 1.0 inclusively.
	 *
	 * @param proportionalPosition New position of needle.
	 */
	public void setPosition(double proportionalPosition)
	{
		updatePosition(proportionalPosition);

		if(changeSignificant) {
			currentPosition = proportionalPosition;
		}
	}

	/**
	 * Renders needle.
	 *
	 * @see ActiveElement#render(Graphics2D)
	 */
	public synchronized void render(Graphics2D g)
	{
		g.setColor(gauger.getNeedleColor());

		g.setColor(Color.BLACK);
		BasicStroke bs = new BasicStroke(2.0f);
		g.setStroke(bs);
		g.draw(line);

	}

	/**
	 * Sets transform defining the gauger. This transform is used to render the
	 * needle at appropriate position and angle.
	 *
	 * @see ActiveElement#setTransform(ScaleTransform)
	 */
	public void setTransform(ScaleTransform transform)
	{
		this.transform = transform;
		invalidate();
		updatePosition(currentPosition);
	}
}

/* __oOo__ */
