/*
 * Copyright (c) 2003-2008 by Cosylab d. d.
 *
 * This file is part of CosyBeans-Common.
 *
 * CosyBeans-Common is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * CosyBeans-Common is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with CosyBeans-Common.  If not, see <http://www.gnu.org/licenses/>.
 */

package com.cosylab.gui.components.gauger;

import java.awt.Font;
import java.awt.FontMetrics;
import java.awt.Graphics2D;
import java.awt.Rectangle;

import com.cosylab.gui.components.Gauger;
import com.cosylab.util.PrintfFormat;


/**
/**
 * Value label displays value of gauger in textual form.
 *
 * @author <a href="mailto:ales.pucelj@cosylab.com">Ales Pucelj</a>
 * @version $id$
 */
public class ValueLabel
{
	private boolean unitsVisible = true;

	private Gauger gauger = null;

	//	private Font font = new Font("dialog", Font.PLAIN, 12);
	private String DEFAULT_FORMAT = "%0.2f";
	private String DEFAULT_TEXT = "val";
	private String text = DEFAULT_TEXT;
//	private String oldText = DEFAULT_TEXT;
	private String formatString = DEFAULT_FORMAT;
	private String units = "";
	private PrintfFormat formatter;
	private double lastValue = 0.0;
	private Rectangle bounds;
	private boolean changeSignificant = true;
	private int x;
	private int y;

	/**
	 * Constructor for ValueLabel.
	 *
	 * @param gauger Gauger on which this value will display.
	 */
	public ValueLabel(Gauger gauger)
	{
		super();
		this.gauger = gauger;
		
		/**
		 * This is initial height for bounding box, since first time, text has
		 * not been drawn yet and height is not properly set. It is initialized
		 * to height of font as later in draw().
		 */
		bounds = new Rectangle(0, 0, 1, 16);
	}

	/**
	 * Returns bounds of this label.
	 *
	 * @return Bounds of text.
	 */
	public Rectangle getBounds()
	{
		return bounds;
	}

	/**
	 * Returns format used by this label.
	 *
	 * @return C-Style format.
	 */
	public String getFormat()
	{
		return formatString;
	}

	/**
	 * Returns units symbol to append to text.
	 *
	 * @return Units.
	 */
	public String getUnits()
	{
		return units;
	}

	private PrintfFormat getFormatter()
	{
		if(formatter == null) {
			formatter = new PrintfFormat(formatString + " " + (unitsVisible ? units : ""));
		}

		return formatter;
	}

	/**
	 * Sets new value to display. Value will be formatted according to the
	 * format specified and will have units symbol appended to it.
	 *
	 * @param value New value to display.
	 */
	public void setValue(double value)
	{
		String newText;
		try {
			newText = getFormatter().sprintf(value);
		} catch (Exception e) {
			newText = Double.toString(value);
		}
		lastValue = value;

		if(!newText.equals(text)) {
			text = newText;

			changeSignificant = true;
		} else {
			changeSignificant = false;
		}
	}

	/**
	 * Returns whether the change is signifficant.
	 *
	 * @return True if text has changed.
	 */
	public boolean isChangeSignificant()
	{
		return changeSignificant;
	}

	/**
	 * Renders this component.
	 *
	 * @param g Context to render to.
	 */
	public void draw(Graphics2D g)
	{
		Font f = gauger.getValueLabelFont();
		g.setFont(f);
		
		
		FontMetrics metrics = g.getFontMetrics();
		
		
		//		metrics.getStringBounds(outText, g);
		int width = metrics.stringWidth(text);
		int height = metrics.getHeight(); //+metrics.getDescent();
		bounds.setBounds(x - width / 2,
		    y - (metrics.getAscent() + metrics.getDescent()) / 2, width,
		    height);
		g.drawString(text, x - width/2, y + height/2 );
	}

	/**
	 * Sets center position of this component. 
	 *
	 * @param x coordinate of label center.
	 * @param y coordinate of label center.
	 */
	public void setPosition(int x, int y)
	{
		this.x = x;
		this.y = y;
	}

	/**
	 * Sets format for this label.
	 *
	 * @param newFormat C-Style format.
	 * @see setValue()
	 */
	public void setFormat(String newFormat)
	{
		if(newFormat != formatString) {
			formatString = newFormat;

			formatter = null;
			setValue(lastValue);
		}
	}

	/**
	 * Sets units to append to text.
	 *
	 * @param newUnits Units symbol.
	 * @see setValue()
	 */
	public void setUnits(String newUnits)
	{
		units = newUnits;
		formatter = null;
		setValue(lastValue);
	}

	/**
	 * Sets the units visibility.
	 * 
	 * @param b
	 */
	public void setUnitsVisible(boolean b) {
		this.unitsVisible = b;
		formatter = null;
		setValue(lastValue);
	}

	/**
	 * Returns true if units are visible.
	 * 
	 * @return
	 */
	public boolean isUnitsVisible() {
		return unitsVisible;
	}
}

/* __oOo__ */
