/*
 * Copyright (c) 2003-2008 by Cosylab d. d.
 *
 * This file is part of CosyBeans-Common.
 *
 * CosyBeans-Common is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * CosyBeans-Common is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with CosyBeans-Common.  If not, see <http://www.gnu.org/licenses/>.
 */

package com.cosylab.gui.components.ledder;
import java.awt.Color;

/**
 * This is a specialized class for rendering LED icons. It is implemented as a
 * minimal ray-tracer with specular lighting overlay rendering.<p>
 * NOTE: Due to performance reasons this class is not thread-safe and should 
 * not be shared in multiple instances.
 * 
 * @author <a href="mailto:ales.pucelj@cosylab.com">Ales Pucelj</a>
 * @version $id$
 */
public class LedRenderer {

	private int centerX;
	private int centerY;

	private double radiusOffset;

	private int radius;

	private int red;
	private int green;
	private int blue;

	private double highlightFactor = 0.85;
	private double highlightFocus = 3;

	private double shadowIntensity = 1.5;

	private int ro1;
	private int ro2;

	private final float[] colorHSB = new float[3];

	private double[] l = getLight(-0.2, -0.5);

	/**
	 * @param x int
	 * @param y int
	 * @param r int
	 * @param lpp LedPaintParameters
	 */
	public LedRenderer(int x, int y, int r, LedPaintParameters lpp) {
		super();

		centerX = x;
		centerY = y;
		radius = r;

		l = getLight(lpp.lightX, lpp.lightY);

		highlightFactor = lpp.highlightFactor;
		highlightFocus = lpp.highlightFocus;
		shadowIntensity = lpp.shadowIntensity;

		radiusOffset = (1 - lpp.borderSize) * radius;
		ro1 = (int) (radiusOffset * radiusOffset);
		ro2 = radius * radius;

		red = lpp.color.getRed();
		green = lpp.color.getGreen();
		blue = lpp.color.getBlue();

	}

	/**
	 * @param x double
	 * @param p double
	 * @return float
	 */
	private final float getHighlight(double x, final double p) {

		x = (x - highlightFactor) / (1 - highlightFactor);

		return (float) ((x < 0) ? p : p * (1 - Math.pow(x, highlightFocus)));
	}

	/**
	 * @param x double
	 * @param p double
	 * @return float
	 */
	private final float getIntensity(final double x, final double p) {
		return (float) (p * Math.pow(x, shadowIntensity));
		//		return (float) (p * x * x * shadowIntensity);
	}

	/**
	 * @param angle double
	 * @return int
	 */
	private final int getColor(final double angle) {
		Color.RGBtoHSB(red, green, blue, colorHSB);

		return Color.HSBtoRGB(
			colorHSB[0],
			getHighlight(angle, colorHSB[1]),
			getIntensity(angle, colorHSB[2]));
	}

	/**
	 * @param x double
	 * @param y double
	 * @param v double[]
	 * @return double[]
	 */
	public double[] getLight(double x, double y) {
		double[] v = new double[3];
		v[0] = x;
		v[1] = y;
		v[2] = Math.sqrt(1.0 - x * x - y * y);

		return v;
	}

	/**
	 * @param x int
	 * @param y int
	 * @param w int
	 * @param h int
	 * @return int[]
	 */
	public final int[] renderSection(
		final int x,
		final int y,
		final int w,
		final int h) {

		final int[] data = new int[w * h * 4];

		int col;
		int base = 0;
		double dot;
		double angle;

		for (int py = 0; py < h; py++) {

			base = 4 * (py * w);
			int dy = y + py - centerY;
			int dx = x - centerX;

			for (int px = 0; px < w; px++) {

				int d = dx * dx + dy * dy;

				if (d < ro1) {

					dot = dx * l[0]	+ dy * l[1]	+ Math.sqrt(ro2 - d) * l[2];

					angle = (1 + dot / radius) / 2;

					col = getColor(angle);

				} else {
					col = (d <= ro2) ? 0xFF000000 : 0x0;
				}

				data[base + 0] = (col >> 0x10) & 0xFF;
				data[base + 1] = (col >> 0x08) & 0xFF;
				data[base + 2] = (col >> 0x00) & 0xFF;
				data[base + 3] = (col >> 0x18) & 0xFF;

				base += 4;
				dx++;
			}
		}
		return data;
	}
}
