/*
 * Copyright (c) 2003-2008 by Cosylab d. d.
 *
 * This file is part of CosyBeans-Common.
 *
 * CosyBeans-Common is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * CosyBeans-Common is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with CosyBeans-Common.  If not, see <http://www.gnu.org/licenses/>.
 */

package com.cosylab.gui.components.ledder;
import java.awt.Color;

/**
 * This is a specialized class for rendering LED icons. It is implemented as a
 * minimal ray-tracer with specular lighting overlay rendering.<p>
 * NOTE: Due to performance reasons this class is not thread-safe and should 
 * not be shared in multiple instances.
 * 
 * @author <a href="mailto:ales.pucelj@cosylab.com">Ales Pucelj</a>
 * @version $id$
 */
public class LedRendererOld {

	private int centerX;
	private int centerY;

	private double radiusOffset;

	private int radius;

	private int red;
	private int green;
	private int blue;

	private double highlightFactor = 0.85;
	private double highlightFocus = 3;

	private double shadowIntensity = 1.5;

	private final float[] colorHSB = new float[3];

	private double[] light = getLight(-0.2, -0.5, new double[3]);

	/**
	 * @param x int
	 * @param y int
	 * @param r int
	 * @param lpp LedPaintParameters
	 */
	public LedRendererOld(int x, int y, int r, LedPaintParameters lpp) {
		super();

		centerX = x;
		centerY = y;
		radius = r;

		light = getLight(lpp.lightX, lpp.lightY, new double[3]);

		highlightFactor = lpp.highlightFactor;
		highlightFocus = lpp.highlightFocus;

		shadowIntensity = lpp.shadowIntensity;

		radiusOffset = (1 - lpp.borderSize) * radius;

		red = lpp.color.getRed();
		green = lpp.color.getGreen();
		blue = lpp.color.getBlue();
	}

	/**
	 * @param v double[]
	 * @return double
	 */
	private static double vLen(final double[] v) {
		return Math.sqrt(v[0] * v[0] + v[1] * v[1] + v[2] * v[2]);
	}

	/**
	 * @param x double
	 * @param y double
	 * @param r double
	 * @return double
	 */
	private static final double getZ(
		final double x,
		final double y,
		final double r) {

		return Math.sqrt(r * r - x * x - y * y);
	}

	/**
	 * @param v double[]
	 * @return double[]
	 */
	private static final double[] norm(final double[] v) {
		double l = vLen(v);
		v[0] /= l;
		v[1] /= l;
		v[2] /= l;

		return v;
	}

	/**
	 * @param x double
	 * @param p double
	 * @return float
	 */
	private final float getHighlight(double x, final double p) {

		x = (x - highlightFactor) / (1 - highlightFactor);

		return (float) ((x < 0) ? p : p * (1 - Math.pow(x, highlightFocus)));
	}

	/**
	 * @param x double
	 * @param p double
	 * @return float
	 */
	private final float getIntensity(final double x, final double p) {
		return (float) (p * Math.pow(x, shadowIntensity));
//		return (float) (p * x * x * shadowIntensity);
	}

	/**
	 * @param angle double
	 * @return int
	 */
	private final int getColor(final double angle) {
		Color.RGBtoHSB(red, green, blue, colorHSB);

		return Color.HSBtoRGB(
			colorHSB[0],
			getHighlight(angle, colorHSB[1]),
			getIntensity(angle, colorHSB[2]));
	}

	/**
	 * @param x double
	 * @param y double
	 * @param v double[]
	 * @return double[]
	 */
	public double[] getLight(double x, double y, double[] v) {
		v[0] = x;
		v[1] = y;
		v[2] = getZ(x, y, 1);

		return norm(v);
	}

 /*
	public int[] renderSectionOld(int x, int y, int w, int h) {

		int[] data = new int[w * h * 4];

		int col = 0;
		int base = 0;

		for (int py = 0; py < h; py++) {

			base = 4 * (py * w);

			for (int px = 0; px < w; px++) {
				n[0] = x + px - centerX;
				n[1] = y + py - centerY;
				n[2] = 0.0;

				double l = vLen(n);

				if (l < radiusOffset) {
					n[2] = getZ(n[0], n[1], radius);

					col = getColor((1 + dot(norm(n), light)) / 2);
				} else {
					col = (l <= radius) ? 0xFF000000 : 0x0;
				}

				data[base + 0] = (col >> 0x10) & 0xFF;
				data[base + 1] = (col >> 0x08) & 0xFF;
				data[base + 2] = (col >> 0x00) & 0xFF;
				data[base + 3] = (col >> 0x18) & 0xFF;

				base += 4;
			}
		}
		return data;
	}
	 */

	/**
	 * @param x int
	 * @param y int
	 * @param w int
	 * @param h int
	 * @return int[]
	 */
	public final int[] renderSection(
		final int x,
		final int y,
		final int w,
		final int h) {

		int[] data = new int[w * h * 4];

		int col = 0;
		int base = 0;

		int ro1 = (int) radiusOffset * (int) radiusOffset;
		int ro2 = radius * radius;

		for (int py = 0; py < h; py++) {

			base = 4 * (py * w);
			int dy = y + py - centerY;

			for (int px = 0; px < w; px++) {
				int dx = x + px - centerX;

				// double l = vLen(n);
				int l = dx * dx + dy * dy;

				if (l < ro1) {
					// n[2] = getZ(n[0], n[1], radius);
					int dz = (int) Math.sqrt(ro2 - l);

					double dot = dx * light[0] + dy * light[1] + dz * light[2];

					//					double norm = Math.sqrt(dx * dx + dy * dy + dz * dz);

					// col = getColor((1 + dot(norm(n), light)) / 2);
					//					col = getColor((1 + dot / norm) / 2);
					double angle = (1 + dot / radius) / 2;

					col = getColor(angle);

				} else {
					col = (l <= ro2) ? 0xFF000000 : 0x0;
				}

				data[base + 0] = (col >> 0x10) & 0xFF;
				data[base + 1] = (col >> 0x08) & 0xFF;
				data[base + 2] = (col >> 0x00) & 0xFF;
				data[base + 3] = (col >> 0x18) & 0xFF;

				base += 4;
			}
		}
		return data;
	}
}

/*
					double saturation;
					double hue = 0;

					int cmin = (red < green) ? red : green;
					cmin = (cmin < blue) ? cmin : blue;

					int cmax = (red > green) ? red : green;
					cmax = (cmax > blue) ? red : blue;

					int delta = cmax - cmin;

					saturation = (cmax == 0) ? 0 : delta / cmax;

					if (saturation == 0) {
						hue = 0;
					} else
					if (red == cmax) {
						hue = 60 * (green - blue) / delta;
					} else {
						if (green == cmax) {
							hue = 120 + 60 * (blue - red) / delta;
						} else {
							if (blue == cmax) {
								hue = 240 + 60 * (red - green) / delta;
							}
						}
					}
					if (hue < 0) {
						hue += 360;
					}
*/
