/*
 * Copyright (c) 2003-2008 by Cosylab d. d.
 *
 * This file is part of CosyBeans-Common.
 *
 * CosyBeans-Common is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * CosyBeans-Common is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with CosyBeans-Common.  If not, see <http://www.gnu.org/licenses/>.
 */

package com.cosylab.gui.components.numberfield;

import com.cosylab.logging.DebugLogger;

import com.cosylab.util.PrintfFormat;
import com.cosylab.util.StringUtilities;

import java.util.logging.Level;
import java.util.logging.Logger;


/**
 * AnglesNumberDescriptor displays double as angle string.
 *
 * @author <a href="mailto:igor.kriznar@cosylab.com">Igor Kriznar</a>
 */
public class AngleNumberDescriptor implements NumberDescriptor
{
	/**
	 * @see NumberDescriptor
	 */
	public final static String NAME = "Angle Visualization";

	/**
	 * @see NumberDescriptor
	 */
	public final static String DESCRIPTION = "Displays double value in degree format: XX\u00B0XX'XX\".";
	private final Logger logger = DebugLogger.getLogger("AND", Level.OFF);
	private char secondSeparator = '"';
	private char minuteSeparator = '\'';
	private char degreeSeparator = '\u00B0';
	private Class<? extends Number> numberType;
	private String format;
	private PrintfFormat formatter;
	private boolean useFullFormat = false;
	private int decimals=0;

	/**
	 * Creates a new AngleNumberDescriptor object.
	 */
	public AngleNumberDescriptor()
	{
		super();
	}

	/**
	 * Creates a new AngleNumberDescriptor object.
	 *
	 * @param full if <code>true</code> number is printed in full format
	 */
	public AngleNumberDescriptor(boolean full)
	{
		this();
		useFullFormat = full;
	}

	/**
	 * Creates a new AngleNumberDescriptor with specified separator characters.
	 *
	 * @param degreeSeparator separator after degree part of value
	 * @param minuteSeparator separator after minutes part of value
	 * @param secondSeparator separator after seconds part of value, if 0 it is
	 *        not diplayed.
	 */
	public AngleNumberDescriptor(char degreeSeparator, char minuteSeparator,
	    char secondSeparator)
	{
		this();
		setDegreeSeparator(degreeSeparator);
		setMinuteSeparator(minuteSeparator);
		setSecondSeparator(secondSeparator);
	}

	/**
	 * Creates a new AngleNumberDescriptor with specified separator characters.
	 *
	 * @param degreeSeparator separator after degree part of value
	 * @param minuteSeparator separator after minutes part of value
	 * @param secondSeparator separator after seconds part of value, if 0 it is
	 *        not diplayed.
	 * @param full if <code>true</code> number is printed in full format
	 */
	public AngleNumberDescriptor(char degreeSeparator, char minuteSeparator,
	    char secondSeparator, boolean full)
	{
		this(degreeSeparator, minuteSeparator, secondSeparator);
		useFullFormat = full;
	}

	/**
	 * Test applet
	 *
	 * @param args cmd
	 */
	public static void main(String[] args)
	{
		AngleNumberDescriptor a = new AngleNumberDescriptor();

		Double d = new Double(3.4434);
		String s = a.printString(d);
		System.out.println(d + " -> " + s + " -> " + a.parseNumber(s));
	}

	/*
	 * (non-Javadoc)
	 *
	 * @see com.cosylab.gui.components.numberfield.NumberDescriptor#parseNumber(java.lang.String)
	 */
	public Number parseNumber(String number)
	{
		logger.info("parsing '" + number + "'");

		try {
			double deg = 0.0;
			double min = 0.0;
			double sec = 0.0;
			int degI = -1;
			int minI = -1;
			int secI = -1;
			double sign = 1.0;

			number = number.trim();

			if (number.charAt(0) == '-') {
				sign = -1.;
				number = number.substring(1);
			}

			degI = number.indexOf(degreeSeparator);

			if (degI == 0) {
				number = number.substring(1);
			} else if (degI > 0) {
				deg = Double.parseDouble(number.substring(0, degI));
				number = number.substring(degI + 1);
			}

			minI = number.indexOf(minuteSeparator);

			if (minI == 0) {
				number = number.substring(1);
			} else if (minI > 0) {
				min = Double.parseDouble(number.substring(0, minI)) / 60.0;
				number = number.substring(minI + 1);
			}

			if (secondSeparator > 0) {
				secI = number.indexOf(secondSeparator);

				if (secI == 0) {
					number = number.substring(1);
				} else if (secI > 0) {
					sec = Double.parseDouble(number.substring(0, secI)) / 3600.0;
					number = number.substring(secI + 1);
				}
			}

			/*if (degI > -1 || minI > -1) {
			    sec = StringUtilities.parseDouble(number) / 3600.0;
			    number = "";
			}*/
			if (degI < 0 && minI < 0 && secI < 0) {
				deg = StringUtilities.parseDouble(number);
			} else if (number.length() > 0) {
				if (degI > -1 && minI < 0 && secI < 0) {
					min = StringUtilities.parseDouble(number) / 60.0;
				} else if (minI > -1 && secI < 0) {
					sec = StringUtilities.parseDouble(number) / 3600.0;
				} else {
					logger.info("returning 'null'");

					return null;
				}
			}

			double d = sign * (deg + min + sec);

			if (Double.isNaN(d)) {
				logger.info("returning 'null' for NaN");

				return null;
			}

			logger.info("returning '" + d + "'");

			return new Double(d);
		} catch (Exception e) {
			logger.info("returning 'null' for " + e);

			return null;
		}
	}

	/*
	 * (non-Javadoc)
	 *
	 * @see com.cosylab.gui.components.numberfield.NumberDescriptor#toString(java.lang.Number)
	 */
	public String printString(Number number)
	{
		logger.info("value '" + number + "'");

		double value = number.doubleValue();
		StringBuffer sb = new StringBuffer();

		if (value < 0) {
			sb.append('-');
			value = Math.abs(value);
		}

		double tmp1 = Math.floor(value);
		value = (value - tmp1) * 60.0;
		double tmp2 = Math.floor(value);
		value = (value - tmp2) * 60.0;

		// solve seconds rounding problem
		if (formatter != null) {
			if (decimals==0) {
				value= Math.round(value);
			} else {
				value= Math.floor(value*Math.pow(10.0,decimals)+0.5)/Math.pow(10.0,decimals);
			}
			if (value>=60.0) {
				value-=60.0;
				tmp2+=1.0;
				if (tmp2>=60.0) {
					tmp2-=60;
					tmp1+=1.0;
				}
			}
		}
		
		if (((int)tmp1) < 10) {
			sb.append('0');
		}

		sb.append((int)tmp1);
		sb.append(degreeSeparator);


		if (tmp2 > 0.0000001 || useFullFormat) {
			if (((int)tmp2) < 10) {
				sb.append('0');
			}

			sb.append((int)tmp2);
			sb.append(minuteSeparator);


			//tmp= Math.floor(value);
			if (value > 0.0000001 || useFullFormat) {
				if (((int)value) < 10) {
					sb.append('0');
				}

				if (formatter != null) {
					String s = formatter.sprintf(value).trim();

					if (s.charAt(0) == '+') {
						s = s.substring(1);
					}

					sb.append(s);
				} else {
					sb.append(value);
				}

				if (secondSeparator != 0) {
					sb.append(secondSeparator);
				}
			}
		}

		logger.info("string '" + sb.toString() + "'");

		return sb.toString();
	}

	/* (non-Javadoc)
	 * @see com.cosylab.gui.components.numberfield.NumberDescriptor#toEditString(java.lang.Number)
	 */
	public String printEditString(Number number)
	{
		return printString(number).trim();
	}

	/*
	 * (non-Javadoc)
	 *
	 * @see com.cosylab.gui.components.numberfield.NumberDescriptor#getFormat()
	 */
	public String getFormat()
	{
		return format;
	}

	/*
	 * (non-Javadoc)
	 *
	 * @see com.cosylab.gui.components.numberfield.NumberDescriptor#setFormt(java.lang.String)
	 */
	public void setFormat(String newFormat)
	{
		format = newFormat;
		if (format==null) {
			formatter=null;
		} else {
			formatter = new PrintfFormat(format);
			decimals=0;
			int dot= format.indexOf('.');
			if (dot>0) {
				int f= format.indexOf('f',dot);
				if (f>0) {
					decimals= Integer.parseInt(format.substring(dot+1,f));
				}
			}
		}
	}

	/*
	 * (non-Javadoc)
	 *
	 * @see com.cosylab.gui.components.numberfield.NumberDescriptor#setNumberType(java.lang.Class)
	 */
	public void setNumberType(Class<? extends Number> newNumberFormat)
	{
		numberType = newNumberFormat;
	}

	/*
	 * (non-Javadoc)
	 *
	 * @see com.cosylab.gui.components.numberfield.NumberDescriptor#getNumberType()
	 */
	public Class<? extends Number> getNumberType()
	{
		return numberType;
	}

	/**
	 * Separator printed after integral part of number, thus degree sign \u00B0
	 * by default.
	 *
	 * @return separator after degree part
	 */
	public char getDegreeSeparator()
	{
		return degreeSeparator;
	}

	/**
	 * Separator printed after integral part of minutes, thus minutes sign ' in
	 * degree notation by default.
	 *
	 * @return separator after minutes part
	 */
	public char getMinuteSeparator()
	{
		return minuteSeparator;
	}

	/**
	 * Sets separator printed after integral part of minutes (60 minutes is one
	 * degree), thus minutes sign ' in degree notation by default.
	 *
	 * @param minuteSeparator new separator after minutes part
	 */
	public void setMinuteSeparator(char minuteSeparator)
	{
		this.minuteSeparator = minuteSeparator;
	}

	/**
	 * Separator printed after seconds (60 seconds is one minute), thus seconds
	 * sign " in degree notation by default. If 0 then is not printed.
	 *
	 * @return separator after seconds part
	 */
	public char getSecondSeparator()
	{
		return secondSeparator;
	}

	/**
	 * Sets separator printed after seconds (60 seconds is one minute), thus
	 * seconds sign " in degree notation by default. If 0 then is not printed.
	 *
	 * @param secondSeparator new separator after seconds part
	 */
	public void setSecondSeparator(char secondSeparator)
	{
		this.secondSeparator = secondSeparator;
	}

	/**
	 * Separator printed after integral part of number, thus degree sign \u00B0
	 * by default.
	 *
	 * @param degreeSeparator new separator after degree part
	 */
	public void setDegreeSeparator(char degreeSeparator)
	{
		this.degreeSeparator = degreeSeparator;
	}

	/**
	 * If <code>true</code> then value is printed in full format with all
	 * separators, thus 0 is printed as 00\u00B000'00"
	 *
	 * @return if <code>true</code> number is printed in full format
	 */
	public boolean isUseFullFormat()
	{
		return useFullFormat;
	}

	/**
	 * If set to <code>true</code> then value is printed in full format with
	 * all separators, thus 0 is printed as 00\u00B000'00"
	 *
	 * @param useFullFormat new full print flag
	 */
	public void setUseFullFormat(boolean useFullFormat)
	{
		this.useFullFormat = useFullFormat;
	}
}

/* __oOo__ */
