/*
 * Copyright (c) 2003-2008 by Cosylab d. d.
 *
 * This file is part of CosyBeans-Common.
 *
 * CosyBeans-Common is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * CosyBeans-Common is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with CosyBeans-Common.  If not, see <http://www.gnu.org/licenses/>.
 */

package com.cosylab.gui.components.range2;

/**
 * This is the implementation of <code>RangedValuePolicy</code> that allows 
 * values of <code>RangedValue</code> to be only within the specified range.
 * This range is defined by allowedMinimum and allowedMaximum. These values
 * are independent from minimum and maximum defined in <code>RangedValue</code>
 * in that no value can be outside these bounds, including minimum and maximum,
 * but will be cropped to these values.
 * 
 * @author <a href="mailto:ales.pucelj@cosylab.com">Ales Pucelj</a>
 * @version $id$
 */
public class LimitedRangePolicy extends RangedValuePolicy {

	// Absolute allowed minimum for all three parameter in RangedValue.
	private double allowedMin = -Double.MAX_VALUE;

	// Absolute allowed maximum for all three parameter in RangedValue.
	private double allowedMax = Double.MAX_VALUE;

	/**
	 * Constructor for LimitedRangedValuePolicy.
	 * @param rangedValue
	 */
	public LimitedRangePolicy(double allowedMinimum, double allowedMaximum) {
		if (allowedMinimum > allowedMaximum) {
			double temp = allowedMinimum;
			allowedMinimum = allowedMaximum;
			allowedMaximum = temp;
		}
		allowedMin = allowedMinimum;
		allowedMax = allowedMaximum;
	}

	/**
	 * Convenience method to check for bounds acceptability.
	 * 
	 * @param value to check against bounds.
	 * @return adjusted value.
	 */
	private final double checkBounds(double value) {
		if (value < allowedMin) {
			value = allowedMin;
		}
		if (value > allowedMax) {
			value = allowedMax;
		}
		return value;
	}

	/**
	 * @see com.cosylab.gui.components.range.RangedValuePolicy#validate(RangedValueHolder)
	 */
	@Override
	public RangedValue validate(RangedValue params) {
		params = new RangedValue(checkBounds(params.getMinimum()),checkBounds(params.getMaximum()),
				checkBounds(params.getValue()));		
		return validatePeerPolicy(params);
	}
}
