/*
 * Copyright (c) 2003-2008 by Cosylab d. d.
 *
 * This file is part of CosyBeans-Common.
 *
 * CosyBeans-Common is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * CosyBeans-Common is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with CosyBeans-Common.  If not, see <http://www.gnu.org/licenses/>.
 */

package com.cosylab.gui.components.range2;

/**
 * Implementation of linear range.
 *
 * @author <a href="mailto:ales.pucelj@cosylab.com">Ales Pucelj</a>
 * @version $id$
 */
public class LinearRange implements Range
{
	
	private TickCalculator defaultTickCalculator;

	/**
	 * Returns the parameters.
	 *
	 * @see com.cosylab.gui.components.range.Range#checkValue(double)
	 */
	public double validate(double value)
	{
		return value;
	}

	/**
	 * Converts relative value in this range to absolute. If span of range is
	 * infinite, return value is 0. Conversion is implemented for all values
	 * even outside the 0.0 - 1.0 range.
	 *
	 * @param relative value in range 0.0 to 1.0 inclusively.
	 *
	 * @return absolute value.
	 *
	 * @see com.cosylab.gui.components.range.Range#toAbsolute(double)
	 */
	public double toAbsolute(double relative, RangedValue ranged)
	{
		final double min = ranged.getMinimum();
		final double max = ranged.getMaximum();

		if (Double.isInfinite(max) || (Double.isInfinite(min))) {
			return 0.0;
		}

		return min + relative * (max - min);
	}

	/**
	 * Converts absolute value into relative (defined in 0.0 to 1.0 range). If
	 * span of the range is infinite, return value is 0.5. Conversion is
	 * defined for all values of even outside the range.
	 *
	 * @param absolute value.
	 *
	 * @return relative value.
	 *
	 * @see com.cosylab.gui.components.range.Range#toRelative(double)
	 */
	public double toRelative(double absolute, RangedValue ranged)
	{
		double range = ranged.getMaximum() - ranged.getMinimum();

		if (range == 0.0) {
			return 0.5;
		}

		if (Double.isInfinite(range)) {
			return 0.5;
		}

		return (absolute - ranged.getMinimum()) / range;
	}
	
	/*
	 * (non-Javadoc)
	 * @see com.cosylab.gui.components.range2.Range#getDefaultTickCalculator()
	 */
	public TickCalculator getDefaultTickCalculator() {
		if (defaultTickCalculator == null) {
			defaultTickCalculator = new LinearTickCalculator(3, false);
		}
		return defaultTickCalculator;
	}
}

/* __oOo__ */
