/*
 * Copyright (c) 2003-2008 by Cosylab d. d.
 *
 * This file is part of CosyBeans-Common.
 *
 * CosyBeans-Common is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * CosyBeans-Common is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with CosyBeans-Common.  If not, see <http://www.gnu.org/licenses/>.
 */

package com.cosylab.gui.components.range2;

/**
 * This interface is a definition of range, a pair of two values denoting 
 * minimum and maximum. Range can be interpreted as a function of two
 * parameters defined between minimum and maximum.<p>
 * Two operations are defined on this range, conversion from absolute value
 * to relative and vice-versa. Absolute values are of the same scale as
 * minimum and maximum, relative values are defined as an interval between
 * 0 and 1, where <code>toRelative(getMinimum())</code> returns 0 and 
 * <code>toRelative(getMaximum())</code> returns 1.<p>
 * 
 * Additionally, each range provides <code>TickCollector</code>, which generates
 * information used when rendering scale ticks.<p>
 * 
 * @author <a href="mailto:ales.pucelj@cosylab.com">Ales Pucelj</a>
 * @version $id$
 */
public interface Range {

	/**
	 * Converts relative value (in 0 to 1 range) to absolute value.<p>
	 * These conditions must be true:
	 * <ul>
	 * <li><code>toAbsolute(0.0) == getMinimum()</code></li>
	 * <li><code>toAbsolute(1.0) == getMaximum()</code></li>
	 * <li><code>toAbsolute(toRelative(x)) == x</code></li>
	 * <li>toAbsolute(x) must always return value between getMinimum() and 
	 *     getMaximum() for x between 0.0 and 1.0.</li>
	 * </ul>
	 * This method does not need to return correct result for values outside the
	 * 0 to 1 range.
	 * 
	 * @param relative value to convert to absolute.
	 * @return Absolute value.
	 */
	double toAbsolute(double relative, RangedValue ranged);
	
	/**
	 * Converts absolute value (in <code>getMinimum()</code> to 
	 * <code>getMaximum()</code> range) to relative value (0 to 1 range).<p>
	 * These conditions must be true:
	 * <ul>
	 * <li><code>toRelative(getMinimum()) == 0.0</code></li>
	 * <li><code>toRelative(getMaximum()) == 1.0</code></li>
	 * <li><code>toRelative(toAbsolute(x)) == x</code></li>
	 * <li>toRelative must return value between 0 and 1 for all values between
	 * <code>getMinimum()</code> and <code>getMaximum()</code>.
	 * </ul>
	 * 
	 * @param absolute value to convert to relative.
	 * @return relative value.
	 */
	double toRelative(double absolute, RangedValue ranged);
	
	/**
	 * Checks the value for acceptability. This is acceptance test for each and
	 * every value that is used with this range. Result will be adjusted value
	 * of the parameter.<p>
	 * For some scales i.e. logarithmic, no value can be set below certain 
	 * value (0.0 in case of logarithmic). This method will ensure, that 
	 * whenever incorrect value is passed, it returns a value that is 
	 * acceptable. In case of logarithmic range, return value will never be
	 * less than 0.0.
	 * 
	 * @param value to test.
	 * @return acceptable value.
	 */
	double validate(double value);
	
	/**
	 * Returns the default TickCaluclator employed by this Range.
	 * 
	 * @return
	 */
	public TickCalculator getDefaultTickCalculator();
}
