/*
 * Copyright (c) 2003-2008 by Cosylab d. d.
 *
 * This file is part of CosyBeans-Common.
 *
 * CosyBeans-Common is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * CosyBeans-Common is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with CosyBeans-Common.  If not, see <http://www.gnu.org/licenses/>.
 */

package com.cosylab.gui.components.range2;

import java.io.Serializable;

/**
 * This class stores values representing RangedValue. No checks are made for
 * validity of the values. Only purpose of this class is to provide data
 * transfer between <code>RangedValue</code> and <code>RangedValuePolicy</code>.
 * As such, this class will normaly be used only internaly by 
 * <code>RangedValue</code>.
 * 
 * @author <a href="mailto:ales.pucelj@cosylab.com">Ales Pucelj</a>
 * @version $id$
 */
public final class RangedValue implements Serializable {

	private static final long serialVersionUID = 1L;

	private double min;
	
	private double max;
	
	private double val;
	
	public RangedValue() {
		this(0, 100, 1);
	}

	/**
	 * @param minimum double
	 * @param maximum double
	 * @param value double
	 */
	public RangedValue(double minimum, double maximum, double value) {
		min = minimum;
		max = maximum;
		val = value;
	}
	
	/**
	 * Returns the minimum.
	 * 
	 * @return minimum
	 */
	public final double getMinimum() {
		return min;
	}
	
	/**
	 * Returns the maximum.
	 * 
	 * @return maximum
	 */
	public final double getMaximum() {
		return max;
	}
	
	/**
	 * Returns the value.
	 * 
	 * @return value
	 */
	public final double getValue() {
		return val;
	}
		
	/**
	 * Check if the value is valid within bounds.
	 * 
	 * @return
	 */
	public final boolean isValid() {
		return ((min <= val) && (val <= max));
	}
	
	/*
	 * (non-Javadoc)
	 * @see java.lang.Object#toString()
	 */
	@Override
	public String toString() {
		return getClass().getName()+" { min = "+min+", max = "+max+", val = "+val+" }";
	}
	
	/**
	 * Returns a new RangedValue with min/max the same as this object, but with the
	 * given value.
	 * 
	 * @param value new value for the RangedValue
	 * @return new RangedValue
	 */
	public RangedValue deriveWithValue(double value) {
		return new RangedValue(min, max, value);
	}
	
	/**
	 * Returns a new RangedValue with min and value the same as this object, but with the
	 * given maximum.
	 * 
	 * @param maximum new maximum for the RangedValue
	 * @return new RangedValue
	 */
	public RangedValue deriveWithMax(double maximum) {
		return new RangedValue(min, maximum, val);
	}
	
	/**
	 * Returns a new RangedValue with max and value the same as this object, but with the
	 * given minimum.
	 * 
	 * @param value new minimum for the RangedValue
	 * @return new RangedValue
	 */
	public RangedValue deriveWithMin(double minimum) {
		return new RangedValue(minimum, max, val);
	}
	
}
