/*
 * Copyright (c) 2003-2008 by Cosylab d. d.
 *
 * This file is part of CosyBeans-Common.
 *
 * CosyBeans-Common is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * CosyBeans-Common is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with CosyBeans-Common.  If not, see <http://www.gnu.org/licenses/>.
 */

package com.cosylab.gui.components.range2;

import java.util.EventObject;

/**
 * This is the event object holding the information about the change in
 * <code>RangedValue</code>. This implementation notifies about changes in
 * minimum, maximum, value or policy changes. More than one change can be
 * carried by this event.
 * 
 * @author <a href="mailto:ales.pucelj@cosylab.com">Ales Pucelj</a>
 * @version $id$
 */
public class RangedValueEvent extends EventObject {

	private static final long serialVersionUID = 1L;

	// Bit indicating the minimum has changed
	public static final int MINIMUM_CHANGED = 1;

	// Bit indicating the maximum has changed
	public static final int MAXIMUM_CHANGED = 2;

	// Bit indicating the value has changed
	public static final int VALUE_CHANGED = 4;
	
	// Bit indicating the value has changed
	public static final int RAW_VALUE_CHANGED = 8;

	// Bits indicating policies have changed
	public static final int POLICY_CHANGED = 16;

	private int change;

	/**
	 * Constructor for RangedValueEvent. This constructor initializes the 
	 * fields to notify of change in minimum, maximum and/or value.
	 * 
	 * @param source
	 * @param min boolean Has minimum changed.
	 * @param max boolean Has maximum changed.
	 * @param val boolean Has value changed.
	 */
	public RangedValueEvent(
		Object source,
		boolean min,
		boolean max,
		boolean val,
		boolean raw) {
			
		super(source);
		change = 0;
		
		setBit(MINIMUM_CHANGED, min);
		setBit(MAXIMUM_CHANGED, max);
		setBit(VALUE_CHANGED, val);
		setBit(RAW_VALUE_CHANGED, raw);
	}

	/**
	 * Constructor for RangedValueEvent. This constructor initializes the fields
	 * to notify of change in policies. 
	 * 
	 * @param source Object
	 * @param policyAdded boolean True if policy added, false if removed.
	 */
	public RangedValueEvent(
		Object source,
		boolean policyChanged) {
		super(source);
		setBit(POLICY_CHANGED, policyChanged);
	}

	/**
	 * @param bit int
	 * @param state boolean
	 */
	protected final void setBit(final int bit, final boolean state) {
		if (state) {
			change |= bit;
		} else {
			change &= ~bit;
		}
	}

	/**
	 * @param bit int
	 * @return boolean
	 */
	private final boolean bitSet(int bit) {
		return (change & bit) > 0;
	}

	/**
	 * Reuturns whether minimum has changed.
	 * 
	 * @return boolean
	 */
	public final boolean isMinimumChanged() {
		return bitSet(MINIMUM_CHANGED);
	}

	/**
	 * Returns whether value has changed.
	 * 
	 * @return boolean
	 */
	public final boolean isMaximumChanged() {
		return bitSet(MAXIMUM_CHANGED);
	}

	/**
	 * Returns whether value has changed.
	 * 
	 * @return boolean
	 */
	public final boolean isValueChanged() {
		return bitSet(VALUE_CHANGED);
	}
	
	/**
	 * Returns whether value has changed.
	 * 
	 * @return boolean
	 */
	public final boolean isRawValueChanged() {
		return bitSet(RAW_VALUE_CHANGED);
	}

	/**
	 * Returns whether minimum and maximum have changed.
	 * 
	 * @return boolean
	 */
	public final boolean isMinMaxChanged() {
		return isMaximumChanged() && isMinimumChanged();
	}

	/**
	 * Returns whether minimum or maximum have changed.
	 * 
	 * @return boolean
	 */
	public final boolean isMinOrMaxChanged() {
		return isMinimumChanged() || isMaximumChanged();
	}
	
	/**
	 * Returns whether the policies have changed.
	 * 
	 * @return boolean
	 */
	public final boolean isPolicyChanged() {
		return bitSet(POLICY_CHANGED);
	}

	/**
	 * Returns the change mask. This value can be then checked agains the
	 * constants defined by this event to determine which changes have occured.
	 * 
	 * @return int
	 */
	public final int getChange() {
		return change;
	}

}
