/*
 * Copyright (c) 2003-2008 by Cosylab d. d.
 *
 * This file is part of CosyBeans-Common.
 *
 * CosyBeans-Common is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * CosyBeans-Common is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with CosyBeans-Common.  If not, see <http://www.gnu.org/licenses/>.
 */

package com.cosylab.gui.components.range2;

/**
 * Defines value policy that maintains constant span. Whenever value moves
 * outside the bounds of the range, min and max will be adjusted to follow the
 * value or will be set to limit of the range.
 *
 * @author <a href="mailto:ales.pucelj@cosylab.com">Ales Pucelj</a>
 * @version $id$
 */
public class ShiftValuePolicy extends RangedValuePolicy
{
	private double shiftFactor;

	/**
	 * Default constructor for ShiftValuePolicy.
	 */
	public ShiftValuePolicy()
	{
		this(0.25);
	}

	/**
	 * Default constructor for ShiftValuePolicy.
	 *
	 * @param factor double
	 */
	public ShiftValuePolicy(double factor)
	{
		super();
		setFactor(factor);
	}

	/**
	 * Calculates new limits for the range. If possible, the scale will only be
	 * shifted by ammount specified by factor. In case that is insufficient,
	 * the limit will be adjusted so the new limit is offset from value by
	 * factor ammount.
	 *
	 * @param params RangedValueHolder
	 *
	 * @see com.cosylab.gui.components.range.RangedValuePolicy#validate(RangedValueHolder)
	 */
	@Override
	public RangedValue validate(RangedValue params)
	{
		double minimum = params.getMinimum();
		double maximum = params.getMaximum();
		double value = params.getValue();
		double span = maximum - minimum;
		double delta = span * getFactor();

		double newMinimum = minimum;
		double newMaximum = maximum;

		if (value < minimum) {
			// move scale down
			newMinimum = minimum - delta;

			if (newMinimum > value) {
				newMinimum = value - delta;
			}

			//newMaximum = maximum + newMinimum - minimum;
			newMaximum = newMinimum + span;
		}

		if (value > maximum) {
			// move scale up
			newMaximum = maximum + delta;

			if (newMaximum < value) {
				newMaximum = value + delta;
			}

			//newMinimum = minimum + newMaximum - maximum;
			newMinimum = newMaximum - span;
		}

		params = new RangedValue(newMinimum, newMaximum, params.getValue());
		
		return validatePeerPolicy(params);
	}

	/**
	 * Returns factor ammount.
	 *
	 * @return double current factor ammount.
	 */
	public double getFactor()
	{
		return shiftFactor;
	}

	/**
	 * Sets new shift factor. This ammount will be between 0 and 1 and will be
	 * adjusted to fit into this range.
	 *
	 * @param factor double
	 */
	public void setFactor(double factor)
	{
		factor = Math.min(1.0, factor);
		factor = Math.max(0.0, factor);

		shiftFactor = factor;
	}
}

/* __oOo__ */
