/*
 * Copyright (c) 2003-2008 by Cosylab d. d. This file is part of CosyBeans-Common.
 * CosyBeans-Common is free software: you can redistribute it and/or modify it under the
 * terms of the GNU General Public License as published by the Free Software Foundation,
 * either version 3 of the License, or (at your option) any later version.
 * CosyBeans-Common is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
 * PARTICULAR PURPOSE. See the GNU General Public License for more details. You should
 * have received a copy of the GNU General Public License along with CosyBeans-Common. If
 * not, see <http://www.gnu.org/licenses/>.
 */

package com.cosylab.gui.components.slider;

import java.awt.Color;
import java.awt.Component;
import java.awt.Dimension;
import java.awt.Font;
import java.awt.GridBagConstraints;
import java.awt.GridBagLayout;
import java.awt.Insets;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;
import java.beans.Beans;

import javax.swing.Icon;
import javax.swing.JButton;
import javax.swing.JPanel;
import javax.swing.JTextField;
import javax.swing.Timer;

import com.cosylab.events.SetListener;
import com.cosylab.gui.components.NumberField;
import com.cosylab.gui.components.ResizableTextLabel;
import com.cosylab.gui.components.Slider;
import com.cosylab.gui.components.Slider.Button;
import com.cosylab.gui.components.util.ColorHelper;
import com.cosylab.gui.components.util.IconHelper;
import com.cosylab.util.PrintfFormat;

/**
 * Button bar for exact control over slider behaviour. Consists of six buttons (3 for
 * decrementing, 3 for incrementing the value) and editable text field for entering exact
 * values. Text field is updated to reflect current value.
 * 
 * @author <a href="mailto:ales.pucelj@cosylab.com">Ales Pucelj</a>
 * @version $id$
 */
public class NavigationBar extends JPanel {
	private static final long serialVersionUID = 1L;
	
	private class HoldMouseListener extends MouseAdapter {
		private Button button;
		public HoldMouseListener(Button button) {
			this.button = button;
		}
		@Override
		public void mousePressed(MouseEvent me) {
			if (slider.isContinuousModeEnabled()) {
				setAutoUpdateButton(this.button);
				getAutoUpdateTimer().start();
			}
		}
		@Override
		public void mouseReleased(MouseEvent me) {
			getAutoUpdateTimer().stop();
		}
	}

	private boolean unitsVisible = true;
	private String units;
	private String format = "%1.4f";

	private NumberField valueEditor;
	private ResizableTextLabel unitsLabel;
	private Slider slider;

	private PrintfFormat formater;

	private JButton fastIncrement;
	private JButton fastDecrement;
	private JButton bitIncrement;
	private JButton bitDecrement;
	private JButton slowIncrement;
	private JButton slowDecrement;
	
	private Button autoUpdateButton = null;

	private Timer autoUpdateTimer;
	// Default inital delay before auto incerement/decrement is applied
	private int continuousModeDelayTime = 1000;
	// Defautl time between individual increment/decremtn repetitions
	private int continuousModeRepeatTime = 100;

	/**
	 * Constructor for NavigationBar.
	 * 
	 * @param owner Slider
	 * @throws IllegalArgumentException if owner is null.
	 */
	public NavigationBar(Slider owner) {
		super();

		if (owner == null) {
			throw new IllegalArgumentException("Cannot create new NavigationBar with null parameter.");
		}

		slider = owner;
		initialize();
	}
	private JButton createButton(String description, String iconName, int x, int min) {
		JButton button = new JButton();
		button.setBackground(ColorHelper.getCosyControl());
		button.setToolTipText(description);
		button.setPreferredSize(new Dimension(25,20));
		button.setMinimumSize(new Dimension(min,18));

		try {
			button.setIcon(IconHelper.createIcon("icons/navigation/" + iconName));
		} catch (Exception e) {
			System.out.println("Icon resource not found: " + iconName);
		}
		add(button,new GridBagConstraints(x,0,1,1,1,0,GridBagConstraints.CENTER,GridBagConstraints.HORIZONTAL,new Insets(1,1,1,1),0,0));
		return button;
	}

	/*
	 * (non-Javadoc)
	 * @see javax.swing.JComponent#setEnabled(boolean)
	 */
	@Override
	public void setEnabled(boolean enabled) {
		super.setEnabled(enabled);
		/*
		 * If Slider is disabled, valueEditor NumberField should still follow current
		 * value setting. On the other hand user should not be able to edit it.
		 */
		valueEditor.setEditable(enabled);
		/*
		 * Setting valueEditor to not focusable prevents updating slider value of a
		 * disabled slider when focus is lost.
		 */
		valueEditor.setFocusable(enabled);
		unitsLabel.setEnabled(enabled);
		fastIncrement.setEnabled(enabled);
		fastDecrement.setEnabled(enabled);
		bitIncrement.setEnabled(enabled);
		bitDecrement.setEnabled(enabled);
		slowIncrement.setEnabled(enabled);
		slowDecrement.setEnabled(enabled);
	}

	/**
	 * Sets the value to the text field.
	 * 
	 * @param value new value
	 */
	public void setValue(double value) {
		valueEditor.setDoubleValue(value);
		valueEditor.setToolTipText(String.valueOf(value));
	}

	/**
	 * Returns text field value.
	 * 
	 * @return current value of text field.
	 */
	public double getValue() {
		return valueEditor.getDoubleValue();
	}

	/**
	 * Sets the continuous mode delay time for the buttons hold update.
	 * 
	 * @param continuousModeDelayTime the initial delay time in millis
	 */
	public void setContinuousModeDelayTime(int continuousModeDelayTime) {
		this.continuousModeDelayTime = continuousModeDelayTime;
		this.getAutoUpdateTimer().setInitialDelay(continuousModeDelayTime);
	}
	
	/**
	 * Sets the continuous mode delay time for the buttons hold update.
	 * 
	 * @return the initial delay time in millis
	 */
	public int getContinuousModeDelayTime() {
		return continuousModeDelayTime;
	}
	
	/**
	 * Sets the continuous mode repeat time for the buttons hold update.
	 * 
	 * @param continuousModeRepeatTime the update period in millis
	 */
	public void setContinuousModeRepeatTime(int continuousModeRepeatTime) {
		this.continuousModeRepeatTime = continuousModeRepeatTime;
		this.getAutoUpdateTimer().setDelay(continuousModeRepeatTime);
	}
	
	/**
	 * Returns the continuous mode repeat time for the buttons hold update.
	 * 
	 * @return the update period in millis
	 */
	public int getContinuousModeRepeatTime() {
		return continuousModeRepeatTime;
	}
	
	/**
	 * Returns current format of text field.
	 * 
	 * @return current format of text field.
	 */
	public String getFormat() {
		return format;
	}

	/**
	 * Sets new format for text field. This format is used only when displaying the value.
	 * 
	 * @param format to set for textfield.
	 */
	public void setFormat(String format) {
		this.format = format;

		if (format == null) {
			format = "%1.4f";
		}

		valueEditor.setFormat(format);
		formater = new PrintfFormat(format);
	}

	/**
	 * Sets units to be displayed next to text field.
	 * 
	 * @param units string to set.
	 */
	public void setUnits(String units) {
		this.units = units;

		if (units == null && Beans.isDesignTime()) {
			units = "<u>";
		}

		unitsLabel.setText((unitsVisible ? units : ""));
	}

	/**
	 * Returns units currently displayed next to edit field.
	 * 
	 * @return current units string.
	 */
	public String getUnits() {
		return units;
	}

	private void initialize() {
		
		setLayout(new GridBagLayout());
		setOpaque(false);

		formater = new PrintfFormat("%1.4f");

		fastDecrement = createButton("Fast decrement","ArrowLeftLeft16.gif",0,15);
		fastDecrement.addMouseListener(new HoldMouseListener(Button.FAST_DECREMENT));
		fastDecrement.addActionListener(new ActionListener(){
			public void actionPerformed(ActionEvent e) {
				slider.decrementLarge();
			}
		});

		slowDecrement = createButton("Slow decrement","ArrowLeft16.gif",1,12);
		slowDecrement.addMouseListener(new HoldMouseListener(Button.SLOW_DECREMENT));
		slowDecrement.addActionListener(new ActionListener(){
			public void actionPerformed(ActionEvent e) {
				slider.decrementSmall();				
			}
		});

		bitDecrement = createButton("Bit decrement","Minus16.gif",2,12);
		bitDecrement.addMouseListener(new HoldMouseListener(Button.BIT_DECREMENT));
		bitDecrement.addActionListener(new ActionListener(){
			public void actionPerformed(ActionEvent e) {
				slider.decrementBit();
			}
		});

		bitIncrement = createButton("Bit increment","Plus16.gif",5,12);
		bitIncrement.addMouseListener(new HoldMouseListener(Button.BIT_INCREMENT));
		bitIncrement.addActionListener(new ActionListener(){
			public void actionPerformed(ActionEvent e) {
				slider.incrementBit();
			}
		});

		slowIncrement = createButton("Slow increment","ArrowRight16.gif",6,12);
		slowIncrement.addMouseListener(new HoldMouseListener(Button.SLOW_INCREMENT));
		slowIncrement.addActionListener(new ActionListener(){
			public void actionPerformed(ActionEvent e) {
				slider.incrementSmall();
			}
		});

		fastIncrement = createButton("Fast increment","ArrowRightRight16.gif",7,15);
		fastIncrement.addMouseListener(new HoldMouseListener(Button.FAST_INCREMENT));
		fastIncrement.addActionListener(new ActionListener(){
			public void actionPerformed(ActionEvent e) {
				slider.incrementLarge();
			}
		});

		valueEditor = new NumberField();
		valueEditor.setNumberType(Double.class);
		valueEditor.setFormat("%1.4f");
		valueEditor.setResizable(false);
		valueEditor.setHorizontalAlignment(JTextField.CENTER);

		valueEditor.setPreferredSize(new java.awt.Dimension(50,20));
		
		add(valueEditor,new GridBagConstraints(3,0,1,1,8,0,GridBagConstraints.CENTER,GridBagConstraints.HORIZONTAL,new Insets(0,3,0,0),0,0));
		
		unitsLabel = new ResizableTextLabel();
		unitsLabel.setOpaque(false);
		unitsLabel.setResizable(false);

		if (Beans.isDesignTime()) {
			unitsLabel.setText("<u>");
		}

		add(unitsLabel,new GridBagConstraints(4,0,1,1,0,0,GridBagConstraints.CENTER,GridBagConstraints.HORIZONTAL,new Insets(0,1,0,3),0,0));
	}

	/**
	 * Automatically updates the value on auto update timer tick. This method is called
	 * when one of the auto update buttons is pressed and held down.
	 */
	private void autoUpdate() {
		if (slider.isContinuousModeEnabled()) {
			synchronized (this) {
				// Value updated only after button released
				// supressPropChEvent = supressSetEvent = true;
				if (getAutoUpdateButton() == Button.BIT_DECREMENT) {
					slider.decrementBit();
				} else if (getAutoUpdateButton() == Button.BIT_INCREMENT) {
					slider.incrementBit();
				} else if (getAutoUpdateButton() == Button.SLOW_DECREMENT) {
					slider.decrementSmall();
				} else if (getAutoUpdateButton() == Button.SLOW_INCREMENT) {
					slider.incrementSmall();
				} else if (getAutoUpdateButton() == Button.FAST_DECREMENT) {
					slider.decrementLarge();
				} else if (getAutoUpdateButton() == Button.FAST_INCREMENT) {
					slider.incrementLarge();
				}
			}
		} else {
			if (this.getAutoUpdateTimer().isRunning()) this.getAutoUpdateTimer().stop();
		}
	}

	private Timer getAutoUpdateTimer() {
		if (autoUpdateTimer == null) {
			autoUpdateTimer = new Timer(continuousModeRepeatTime,new ActionListener() {
				public void actionPerformed(ActionEvent e) {
					if (slider.isContinuousModeEnabled()) autoUpdate();
				}
			});
			autoUpdateTimer.setInitialDelay(continuousModeDelayTime);
		}
		return autoUpdateTimer;
	}

	/**
	 * Convinience format with Slider format
	 * 
	 * @param value the value
	 * @return formated value as String
	 */
	public String format(double value) {
		return formater.sprintf(value);
	}

	/**
	 * Sets the minimum value.
	 * 
	 * @param min minimum
	 */
	public void setMinimum(double min) {
		valueEditor.setMinimum(new Double(min));
	}

	/**
	 * Sets the maximum value.
	 * 
	 * @param min maximum
	 */
	public void setMaximum(double min) {
		valueEditor.setMaximum(new Double(min));
	}

	/**
	 * Sets the units visibility.
	 * 
	 * @param b true if units are visible
	 */
	public void setUnitsVisible(boolean b) {
		this.unitsVisible = b;
		unitsLabel.setText((unitsVisible ? units : ""));
	}

	/**
	 * Returns true if units are visible.
	 * 
	 * @return true if visible
	 */
	public boolean isUnitsVisible() {
		return unitsVisible;
	}

	/*
	 * (non-Javadoc)
	 * @see javax.swing.JComponent#setBackground(java.awt.Color)
	 */
	@Override
	public void setBackground(Color bg) {
		super.setBackground(bg);
		Component[] comp = getComponents();
		for (Component c : comp) {
			if (c instanceof JButton) {
				c.setBackground(bg);
			}
		}
	}

	/*
	 * (non-Javadoc)
	 * @see javax.swing.JComponent#setForeground(java.awt.Color)
	 */
	@Override
	public void setForeground(Color fg) {
		super.setForeground(fg);
		Component[] comp = getComponents();
		for (Component c : comp) {
			c.setForeground(fg);
		}
	}

	/**
	 * Adds a set listener which receives notification when a new value is set in the
	 * field.
	 * 
	 * @param sl
	 */
	public void addSetListener(SetListener sl) {
		valueEditor.addSetListener(sl);
	}

	/**
	 * Removes a set listener.
	 * 
	 * @param sl
	 */
	public void removeSetListener(SetListener sl) {
		valueEditor.removeSetListener(sl);
	}

	/**
	 * Sets text for buttons.
	 * 
	 * @see
	 * @param button
	 * @param text
	 */
	public void setButtonText(Button button, String text) {
		if (button.equals(Button.FAST_INCREMENT)) {
			fastIncrement.setText(text);
		} else if (button.equals(Button.FAST_DECREMENT)) {
			fastDecrement.setText(text);
		} else if (button.equals(Button.SLOW_INCREMENT)) {
			slowIncrement.setText(text);
		} else if (button.equals(Button.SLOW_DECREMENT)) {
			slowDecrement.setText(text);
		} else if (button.equals(Button.BIT_INCREMENT)) {
			bitIncrement.setText(text);
		} else if (button.equals(Button.BIT_DECREMENT)) {
			bitDecrement.setText(text);
		}
	}

	/**
	 * Returns the button text.
	 * 
	 * @param button
	 * @return
	 */
	public String getButtonText(Button button) {
		if (button.equals(Button.FAST_INCREMENT)) {
			return fastIncrement.getText();
		} else if (button.equals(Button.FAST_DECREMENT)) {
			return fastDecrement.getText();
		} else if (button.equals(Button.SLOW_INCREMENT)) {
			return slowIncrement.getText();
		} else if (button.equals(Button.SLOW_DECREMENT)) {
			return slowDecrement.getText();
		} else if (button.equals(Button.BIT_INCREMENT)) {
			return bitIncrement.getText();
		} else if (button.equals(Button.BIT_DECREMENT)) {
			return bitDecrement.getText();
		}
		return null;
	}

	/**
	 * Sets tooltips for buttons.
	 * 
	 * @see
	 * @param button
	 * @param tooltip
	 */
	public void setButtonTooltip(Button button, String tooltip) {
		if (button.equals(Button.FAST_INCREMENT)) {
			fastIncrement.setToolTipText(tooltip);
		} else if (button.equals(Button.FAST_DECREMENT)) {
			fastDecrement.setToolTipText(tooltip);
		} else if (button.equals(Button.SLOW_INCREMENT)) {
			slowIncrement.setToolTipText(tooltip);
		} else if (button.equals(Button.SLOW_DECREMENT)) {
			slowDecrement.setToolTipText(tooltip);
		} else if (button.equals(Button.BIT_INCREMENT)) {
			bitIncrement.setToolTipText(tooltip);
		} else if (button.equals(Button.BIT_DECREMENT)) {
			bitDecrement.setToolTipText(tooltip);
		}
	}

	/**
	 * Returns the button tooltip.
	 * 
	 * @param button
	 * @return
	 */
	public String getButtonTooltip(Button button) {
		if (button.equals(Button.FAST_INCREMENT)) {
			return fastIncrement.getToolTipText();
		} else if (button.equals(Button.FAST_DECREMENT)) {
			return fastDecrement.getToolTipText();
		} else if (button.equals(Button.SLOW_INCREMENT)) {
			return slowIncrement.getToolTipText();
		} else if (button.equals(Button.SLOW_DECREMENT)) {
			return slowDecrement.getToolTipText();
		} else if (button.equals(Button.BIT_INCREMENT)) {
			return bitIncrement.getToolTipText();
		} else if (button.equals(Button.BIT_DECREMENT)) {
			return bitDecrement.getToolTipText();
		}
		return null;
	}

	/**
	 * Sets the visibility of the button.
	 * 
	 * @param button the button
	 * @param visible true if button should be visible
	 */
	public void setButtonVisible(Button button, boolean visible) {
		if (button.equals(Button.FAST_INCREMENT)) {
			fastIncrement.setVisible(visible);
		} else if (button.equals(Button.FAST_DECREMENT)) {
			fastDecrement.setVisible(visible);
		} else if (button.equals(Button.SLOW_INCREMENT)) {
			slowIncrement.setVisible(visible);
		} else if (button.equals(Button.SLOW_DECREMENT)) {
			slowDecrement.setVisible(visible);
		} else if (button.equals(Button.BIT_INCREMENT)) {
			bitIncrement.setVisible(visible);
		} else if (button.equals(Button.BIT_DECREMENT)) {
			bitDecrement.setVisible(visible);
		}
	}

	/**
	 * Returns the visibility of the button.
	 * 
	 * @param button
	 * @return visibility of the button
	 */
	public boolean isButtonVisible(Button button) {
		if (button.equals(Button.FAST_INCREMENT)) {
			return fastIncrement.isVisible();
		} else if (button.equals(Button.FAST_DECREMENT)) {
			return fastDecrement.isVisible();
		} else if (button.equals(Button.SLOW_INCREMENT)) {
			return slowIncrement.isVisible();
		} else if (button.equals(Button.SLOW_DECREMENT)) {
			return slowDecrement.isVisible();
		} else if (button.equals(Button.BIT_INCREMENT)) {
			return bitIncrement.isVisible();
		} else if (button.equals(Button.BIT_DECREMENT)) {
			return bitDecrement.isVisible();
		}
		return false;
	}

	/**
	 * Sets the font used to display units.
	 * 
	 * @param font new font
	 */
	public void setUnitsFont(Font font) {
		unitsLabel.setFont(font);
	}

	/**
	 * Returns the font used to display units.
	 * 
	 * @return
	 */
	public Font getUnitsFont() {
		return unitsLabel.getFont();
	}

	public Button getAutoUpdateButton() {
		return autoUpdateButton;
	}

	public void setAutoUpdateButton(Button autoUpdateButton) {
		this.autoUpdateButton = autoUpdateButton;
	}

	public void setButtonIcon(Button button, Icon icon) {
		if (button.equals(Button.FAST_INCREMENT)) {
			fastIncrement.setIcon(icon);
		} else if (button.equals(Button.FAST_DECREMENT)) {
			fastDecrement.setIcon(icon);
		} else if (button.equals(Button.SLOW_INCREMENT)) {
			slowIncrement.setIcon(icon);
		} else if (button.equals(Button.SLOW_DECREMENT)) {
			slowDecrement.setIcon(icon);
		} else if (button.equals(Button.BIT_INCREMENT)) {
			bitIncrement.setIcon(icon);
		} else if (button.equals(Button.BIT_DECREMENT)) {
			bitDecrement.setIcon(icon);
		}
	}

	public Icon getButtonIcon(Button button) {
		if (button.equals(Button.FAST_INCREMENT)) {
			return fastIncrement.getIcon();
		} else if (button.equals(Button.FAST_DECREMENT)) {
			return fastDecrement.getIcon();
		} else if (button.equals(Button.SLOW_INCREMENT)) {
			return slowIncrement.getIcon();
		} else if (button.equals(Button.SLOW_DECREMENT)) {
			return slowDecrement.getIcon();
		} else if (button.equals(Button.BIT_INCREMENT)) {
			return bitIncrement.getIcon();
		} else if (button.equals(Button.BIT_DECREMENT)) {
			return bitDecrement.getIcon();
		}
		return null;
	}
}

/* __oOo__ */
