/*
 * Copyright (c) 2003-2008 by Cosylab d. d.
 *
 * This file is part of CosyBeans-Common.
 *
 * CosyBeans-Common is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * CosyBeans-Common is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with CosyBeans-Common.  If not, see <http://www.gnu.org/licenses/>.
 */

package com.cosylab.gui.components.table;

import java.util.Vector;

import javax.swing.JTable;
import javax.swing.ListSelectionModel;
import javax.swing.table.TableColumnModel;
import javax.swing.table.TableModel;


/**
 * <code>QueueTable</code> uses <code>QueueTableModelInterceptor</code> update
 * table  in ferformance and thread safe way. User can use this table exactely
 * the same way as <code>JTable</code>, all the event intercetion is done
 * automagically inside this implementation.
 *
 * @author <a href="mailto:igor.kriznar@cosylab.com">Igor Kriznar</a>
 * @version $Id: QueueTable.java,v 1.6 2008-04-22 12:28:40 jbobnar Exp $
 *
 * @since May 3, 2004.
 */
public class QueueTable extends JTable
{
	private static final long serialVersionUID = 1L;
	private TableModel model;
	private QueueTableModelInterceptor interceptor;

	/**
	 * Constructs a new QueueTable object.
	 * @see JTable#JTable()
	 */
	public QueueTable()
	{
		super();
	}

	/**
	 * Constructs a new QueueTable object with the given number of rows and columns.
	 * 
	 * @param numRows number of rows
	 * @param numColumns number of columns
	 * @see JTable#JTable(int, int)
	 */
	public QueueTable(int numRows, int numColumns)
	{
		super(numRows, numColumns);
	}

	/**
	 * Constructs a new QueueTable with the given TableModel.
	 * 
	 * @param dm table model
	 * @see JTable#JTable(TableModel)
	 */
	public QueueTable(TableModel dm)
	{
		super(dm);
	}

	/**
	 * Constructs a new QueueTable with initial values and column names.
	 * 
	 * @param rowData initial data
	 * @param columnNames column names
	 * @see JTable#JTable(Object[][], Object[])
	 */
	public QueueTable(Object[][] rowData, Object[] columnNames)
	{
		super(rowData, columnNames);
	}

	/**
	 * Constructs a new QueueTable with initial values and column names.
	 * 
	 * @param rowData initial data
	 * @param columnNames column names
	 * @see JTable#JTable(Vector, Vector)
	 */
	public QueueTable(Vector rowData, Vector columnNames)
	{
		super(rowData, columnNames);
	}

	/**
	 * Constructs a new QueueTable using the given models.
	 * 
	 * @param dm table model
	 * @param cm column model
	 * @see JTable#JTable(TableModel, TableColumnModel)
	 */
	public QueueTable(TableModel dm, TableColumnModel cm)
	{
		super(dm, cm);
	}

	/**
	 * Constructs a new QueueTable using the given models.
	 * 
	 * @param dm table model
	 * @param cm column model
	 * @param sm selection model
	 * @see JTable#JTable(TableModel, TableColumnModel, ListSelectionModel)
	 */
	public QueueTable(TableModel dm, TableColumnModel cm, ListSelectionModel sm)
	{
		super(dm, cm, sm);
	}

	/* (non-Javadoc)
	 * @see javax.swing.JTable#setModel(javax.swing.table.TableModel)
	 */
	@Override
    public void setModel(TableModel dataModel)
	{
		model = dataModel;

		if (model == null) {
			super.setModel(null);
		} else {
			//super.setModel(model);
			QueueTableModelInterceptor i=new QueueTableModelInterceptor(model);
			if (interceptor!=null) {
				i.setRelaxationTime(interceptor.getRelaxationTime());
				interceptor.destroy();
			}
			super.setModel(interceptor=i);
		}
	}
	
	/* (non-Javadoc)
	 * @see javax.swing.JTable#getModel()
	 */
	@Override
    public TableModel getModel() {
		return model;
	}

	/**
	 * Returns minimum time betwean two table updates interceptor will wait.
	 * @return minimum time betwean two table updates 
	 */
	public int getRelaxationTime() {
		return interceptor.getRelaxationTime();
	}

	/**
	 * Sets minimum time betwean two table updates interceptor will wait.
	 * @param relaxationTime
	 */
	public void setRelaxationTime(int relaxationTime) {
		interceptor.setRelaxationTime(relaxationTime);
	}
}

/* __oOo__ */
