/*
 * Copyright (c) 2003-2008 by Cosylab d. d.
 *
 * This file is part of CosyBeans-Common.
 *
 * CosyBeans-Common is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * CosyBeans-Common is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with CosyBeans-Common.  If not, see <http://www.gnu.org/licenses/>.
 */

package com.cosylab.gui.components.table.cells;

/**
 * LongEnumCell holds enumeration valies. Each long value from min to max
 * represents one enumeration state, which can have own name and value.
 *
 * @author <a href="mailto:igor.kriznar@cosylab.com">Igor Kriznar</a>
 */
public class LongEnumCell extends LongCell
{
	/** State description property tag. */
	public static final String STATE_DESCRIPTION = "stateDescription";

	/** State value property tag */
	public static final String STATE_VALUE = "stateValue";

	/**
	 * Constructs a new LongEnumCell.
	 *
	 */
	public LongEnumCell()
	{
		super();

	}

	/**
	 * Constructs a new LongEnumCell.
	 *
	 * @param identifier
	 * @param dataSource
	 * @param value
	 * @param commands
	 */
	public LongEnumCell(String identifier, Object dataSource, long value,
	    Command[] commands)
	{
		super(identifier, dataSource, value, commands);
	}

	/**
	 * Creates a new LongEnumCell object.
	 *
	 * @param identifier
	 * @param dataSource 
	 * @param value 
	 * @param commands 
	 * @param descriptions 
	 * @param values 
	 */
	public LongEnumCell(String identifier, Object dataSource, long value,
	    Command[] commands, String[] descriptions, Object[] values)
	{
		super(identifier, dataSource, value, commands);

		for (int i = 0; i < values.length; i++) {
			putCharacteristic(STATE_VALUE + i, values[i]);
		}

		for (int i = 0; i < descriptions.length; i++) {
			putCharacteristic(STATE_DESCRIPTION + i, descriptions[i]);
		}
	}

	/**
	 * Returns the state description under the index <code>i</code>.
	 *
	 * @param i index of the requested description
	 *
	 * @return the description
	 */
	public String getStateDescription(long i)
	{
		return (String)getCharacteristic(STATE_DESCRIPTION + i);
	}

	/**
	 * Returns the value under the index <code>i</code>.
	 *
	 * @param i index of the requested value
	 *
	 * @return the value
	 */
	public Object getStateValue(long i)
	{
		return getCharacteristic(STATE_DESCRIPTION + i);
	}

	/**
	 * Returns the label name of the state under the index <code>i</code>.
	 *
	 * @param l index of the requested label
	 *
	 * @return the label
	 */
	public String getStateLabel(long l)
	{
		String s = getStateDescription(l);

		if (s == null) {
			StringBuffer sb = new StringBuffer(8);
			sb.append('<');
			sb.append(l);
			sb.append('>');
			s = sb.toString();
		}

		return s;
	}

	/**
	 * Trys to lookup state table and find enum index with this label. If value
	 * is not found, -1 is returned.
	 *
	 * @param label enum state description label
	 *
	 * @return state index or -1
	 */
	public long toLongValue(String label)
	{
		label = label.trim();

		if (label == null || label.length() < 3) {
			return -1;
		}

		if (label.charAt(0) == '<' && label.charAt(label.length() - 1) == '>') {
			// this is generated label
			return Long.parseLong(label.substring(1, label.length() - 1));
		}

		// this is label from descriptions
		int min = getMinimum(new Long(0)).intValue();
		int max = getMaximum(new Long(15)).intValue();

		for (int i = min; i <= max; i++) {
			if (label.equals(getCharacteristic(STATE_DESCRIPTION + i))) {
				return i;
			}
		}

		return -1;
	}
}

/* __oOo__ */
