/*
 * Copyright (c) 2003-2008 by Cosylab d. d.
 *
 * This file is part of CosyBeans-Common.
 *
 * CosyBeans-Common is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * CosyBeans-Common is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with CosyBeans-Common.  If not, see <http://www.gnu.org/licenses/>.
 */

package com.cosylab.gui.components.util;

import java.awt.event.ActionEvent;

import java.util.ArrayList;
import java.util.Iterator;

import javax.swing.AbstractAction;
import javax.swing.Action;
import javax.swing.Icon;


/**
 * <p>
 * An action which holds a list of other <code>Action</code> objects.
 * Convenient for containment of functionally similar actions.
 * </p>
 * 
 * <p>
 * With asotiation with <code>Action</code> can be used for creation of menus.
 * Set <code>putValue(MENU_CONTEXT,MENU_CONTEXT_RADIO)</code> to produce group
 * of radio menu items.
 * </p>
 * 
 * <p>
 * <code>Action</code> will use separator menu item, where null
 * actionitemisplaced or Action has <code>SEPARATOR</code> boolean attribute.
 * </p>
 *
 * @author <a href="mailto:miha.kadunc@cosylab.com">Miha Kadunc</a>
 * @version $id$
 */
public class ActionList extends AbstractAction
{
	private static final long serialVersionUID = 1L;

	/**
	 * Key name for a Boolean value indicating that the added action should be
	 * represented in the GUI as a separator.
	 */
	public static final String SEPARATOR = "_group_separator_";

	/**
	 * Key name for a String value indicating if this list is to be presented
	 * in menu with buttons or radio buttons.
	 *
	 * @see javax.swing.Action#getValue(java.lang.String)
	 */
	public static final String MENU_CONTEXT = "menuContext";

	/** DOCUMENT ME! */
	public static final String MENU_CONTEXT_BUTTON = "button";

	/** DOCUMENT ME! */
	public static final String MENU_CONTEXT_RADIO = "radio";

	/** DOCUMENT ME! */
	public static String ACTIONS = "actions";
	private ArrayList<Action> actions = null;

	/**
	 * Creates a new ActionList object.
	 */
	public ActionList()
	{
		this(null, null);
	}

	/**
	 * Creates a new ActionList object.
	 *
	 * @param name 
	 */
	public ActionList(String name)
	{
		this(name, null);
	}

	/**
	 * Creates a new ActionList object.
	 *
	 * @param name 
	 * @param icon 
	 */
	public ActionList(String name, Icon icon)
	{
		this(name, icon, null);
	}

	/**
	 * Creates a new ActionList object.
	 *
	 * @param name 
	 * @param icon 
	 * @param actions 
	 */
	public ActionList(String name, Icon icon, Action[] actions)
	{
		super(name, icon);
		this.actions = new ArrayList<>();

		if (actions != null) {
			for (int i = 0; i < actions.length; i++) {
				this.actions.add(actions[i]);
			}

			putValue(ACTIONS, this.actions.toArray(new Action[size()]));
		}

		putValue(MENU_CONTEXT, MENU_CONTEXT_BUTTON);
	}

	/**
	 * Adds Action to end of the list.
	 *
	 * @param action Action to be added
	 */
	public void addAction(Action action)
	{
		addAction(size(), action);
	}

	/**
	 * Checks if this list contains the given action.
	 *
	 * @param action 
	 *
	 * @return true if action exists in the list
	 */
	public boolean contains(Action action)
	{
		return actions.contains(action);
	}

	/**
	 * Checks if the list containes action with the given name.
	 *
	 * @param name name of the action
	 *
	 * @return true if action is present in the list
	 */
	public boolean containsActionWithName(String name)
	{
		return (firstIndexOfActionWithName(name) < 0);
	}

	/**
	 * Returns the index of the first action with the given name. 
	 *
	 * @param name name of the action
	 *
	 * @return index of the action in the list
	 */
	public int firstIndexOfActionWithName(String name)
	{
		for (Iterator iter = actions.iterator(); iter.hasNext();) {
			Action element = (Action)iter.next();

			if ((element != null) && element.getValue(NAME).equals(name)) {
				return actions.indexOf(element);
			}
		}

		return -1;
	}

	/**
	 * Adds all actions from ActionList to end of the this list in specified
	 * order.
	 *
	 * @param action actions to be added
	 */
	public void addAll(ActionList action)
	{
		Action[] a = action.toArray();

		for (int i = 0; i < a.length; i++) {
			addAction(a[i]);
		}
	}

	/* (non-Javadoc)
	 * @see com.cosylab.gui.components.util.ActionList#addAction(javax.swing.Action, int)
	 */
	public synchronized void addAction(int index, Action action)
	{
		actions.add(index, action);
		putValue(ACTIONS, actions.toArray(new Action[size()]));
	}

	/**
	 * Returns Action under provided index.
	 *
	 * @param index the index of requested action
	 *
	 * @return the action
	 */
	public synchronized Action getAction(int index)
	{
		return actions.get(index);
	}

	/**
	 * Returns Action which command ID corresponds to the provided parameter. 
	 *
	 * @param ID the command key string of action
	 *
	 * @return Action which command ID corresponds to the provided parameter
	 * 
	 * @see Action#ACTION_COMMAND_KEY
	 */
	public synchronized Action getAction(String ID)
	{
		Iterator it = actions.iterator();

		while (it.hasNext()) {
			Action a = (Action)it.next();

			if (ID.equals(a.getValue(ACTION_COMMAND_KEY))) {
				return a;
			}
		}

		return null;
	}

	/* (non-Javadoc)
	 * @see com.cosylab.gui.components.util.ActionList#indexOf(javax.swing.Action)
	 */
	public int indexOf(Action action)
	{
		return actions.indexOf(action);
	}

	/* (non-Javadoc)
	 * @see com.cosylab.gui.components.util.ActionList#removeAction(javax.swing.Action)
	 */
	public void removeAction(Action action)
	{
		actions.remove(action);
		putValue(ACTIONS, actions.toArray(new Action[size()]));
	}

	/* (non-Javadoc)
	 * @see com.cosylab.gui.components.util.ActionList#removeAction(int)
	 */
	public Action removeAction(int index)
	{
		Action a = actions.remove(index);
		putValue(ACTIONS, actions.toArray(new Action[size()]));

		return a;
	}

	/**
	 * Returns the size of this list (number of actions in the list).
	 *
	 * @return the size
	 */
	public int size()
	{
		return actions.size();
	}
	
	/*
	 * (non-Javadoc)
	 * @see java.awt.event.ActionListener#actionPerformed(java.awt.event.ActionEvent)
	 */
	public void actionPerformed(ActionEvent e)
	{
	}

	/**
	 * Returns all contained actions in an array.
	 *
	 * @return returns all contained actions in an array
	 */
	public Action[] toArray()
	{
		return actions.toArray(new Action[actions.size()]);
	}

	/**
	 * Returns iterator over contained actions.
	 *
	 * @return iterator over contained actions
	 */
	public Iterator iterator()
	{
		return actions.iterator();
	}

	/**
	 * Removes all actions from this list.
	 * 
	 */
	public void clear()
	{
		actions.clear();
	}
}

/* __oOo__ */
