/*
 * Copyright (c) 2003-2008 by Cosylab d. d.
 *
 * This file is part of CosyBeans-Common.
 *
 * CosyBeans-Common is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * CosyBeans-Common is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with CosyBeans-Common.  If not, see <http://www.gnu.org/licenses/>.
 */

package com.cosylab.gui.components.util;

import java.awt.Color;
import java.util.ArrayList;

/**
 * ColorManager provides a sequence of colors which can be used for example in charts.
 * 
 * @author <a href="mailto:jaka.bobnar@cosylab.com">Jaka Bobnar</a>
 * @version $Id: ColorManager.java,v 1.8 2008-08-14 11:01:56 jbobnar Exp $
 *
 */
public class ColorManager {
	
	public static final Color[] DEFAULT_COLORS= new Color[]{Color.BLACK,Color.RED,Color.BLUE,Color.GREEN.darker(),Color.ORANGE,Color.DARK_GRAY,Color.RED.darker(),Color.BLUE.darker(),Color.GREEN.darker().darker(),Color.ORANGE.darker()};
	private Color[] availableColors = DEFAULT_COLORS;
	private ArrayList<Color> unusedColors = new ArrayList<>(3);
	private int count = 0;
	
	/**
	 * Loads Default colors as the colors used by this manager.
	 * 
	 */
	public void loadDefaultColors() {
		loadColors(DEFAULT_COLORS);
	}
	
	/**
	 * Loads the colors among which manager will pick a color when requested.
	 * 
	 * @param colors color bank (null or zero-length value not permitted)
	 */
	public void loadColors(Color[] colors) {
		if (colors == null) throw new IllegalArgumentException("Colors cannot be null.");
		else if (colors.length == 0) throw new IllegalArgumentException("There should be at least one Color available.");
		synchronized (unusedColors) {
			availableColors = new Color[colors.length];
			System.arraycopy(colors, 0, availableColors, 0, colors.length);
			if (count >= availableColors.length) {
				count %= availableColors.length;
			}
		}
	}
	
	/**
	 * Picks the next color in the sequence.
	 * 
	 * @return the next color
	 */
	public Color pickColor() {
		synchronized (unusedColors) {
			if (unusedColors.size() > 0) {
				Color c = unusedColors.remove(0);
				if (c != null) {
					count++;
					return c;
				}
			}
			return availableColors[count++%availableColors.length];
		}
	}
		
	/**
	 * Drops the color into a container with unused colors. When a new color will
	 * be picked, the manager will first choose among the dropped color and when
	 * there will be no more, it will continue the normal sequence as stated in the
	 * loaded colors.
	 * 
	 * @param color color to be dropped into the bank
	 */
	public void dropColor(Color color) {
		if (color == null) throw new IllegalArgumentException("Dropped color cannot be null.");
		synchronized (unusedColors) {
			unusedColors.add(color);
			count--;
			if (count <=0 ) count = 0;
		}
	}
	
	/**
	 * Returns the current index at which the color is picked from the array.
	 * @return color index
	 */
	public int getColorIndex() {
		return count;
	}
	
	/**
	 * Sets the index at which the next color will be picked from array.
	 * @param index new color index
	 */
	public void setColorIndex(int index) {
		this.count = index % availableColors.length;
	}
	
	/**
	 * Sets the color which will be the next to be picked out from the array.
	 * It the provided color does not exist nothing will change.
	 * 
	 * @param color the next color
	 */
	public void setIndexByColor(Color color) {
		for (int i = 0; i < availableColors.length; i++) {
			if (availableColors[i].equals(color)) {
				setColorIndex(i);
				return;
			}
		}
	}
	
	/**
	 * Empties the container with all dropped colors.
	 *
	 */
	public void reset() {
		synchronized (unusedColors) {
			unusedColors.clear();
			count = 0;
		}
	}
	
	/**
	 * Previews the next color that will be picked.
	 * 
	 * @return the next color
	 */
	public Color previewNextColor() {
		synchronized (unusedColors) {
			if (unusedColors.size() > 0) {
				return unusedColors.get(0);
			}
			return availableColors[count%availableColors.length];
		}
	}
	
}
