/*
 * Copyright (c) 2003-2008 by Cosylab d. d.
 *
 * This file is part of CosyBeans-Common.
 *
 * CosyBeans-Common is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * CosyBeans-Common is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with CosyBeans-Common.  If not, see <http://www.gnu.org/licenses/>.
 */

package com.cosylab.gui.components.util;

import java.awt.Image;
import java.awt.image.ImageProducer;
import java.io.File;
import java.net.URL;
import java.util.HashMap;

import javax.swing.Icon;
import javax.swing.ImageIcon;


/**
 * IconHelper acts as an interface to the all icons currently in classpath.
 * IconHelper is a singleton class that maintains internal cache to all
 * accessed icons. This class may be used to minimize the memory consumption
 * where application uses static icons. If icons are changed independently,
 * this class should not be used.
 *
 * @author <a href="mailto:ales.pucelj@cosylab.com">Ales Pucelj</a>
 * @version $id$
 */
public final class IconHelper {
	
	public static String ICONS_NAVIGATION_DOWN16= "icons/navigation/Down16.gif";
	public static String ICONS_NAVIGATION_DOWN24= "icons/navigation/Down24.gif";
	public static String ICONS_NAVIGATION_MINUS16= "icons/navigation/Minus16.gif";
	public static String ICONS_NAVIGATION_PLUS16= "icons/navigation/Plus16.gif";
	public static String ICONS_NAVIGATION_UP16= "icons/navigation/Up16.gif";
	public static String ICONS_NAVIGATION_UP24= "icons/navigation/Up24.gif";
	public static String ICONS_GENERAL_DELETE16= "icons/general/Delete16.gif";
	public static String ICONS_GENERAL_DELETE24= "icons/general/Delete24.gif";
	public static String ICONS_GENERAL_REMOVE16= "icons/general/Remove16.gif";
	public static String ICONS_GENERAL_REMOVE24= "icons/general/Remove24.gif";
	
	
    /** cache for icons. */
    private static HashMap<String, Icon> cache;

    /**
     * Protected constructor for creation of singleton class.
     */
    private IconHelper() {
        super();
    }

    /**
     * Returns icon from the cache. If icon cannot be found in cache, return
     * value will be <code>null</code>.
     *
     * @param name String
     * @return Icon instance of an icon or null if not found
     */
    private static synchronized Icon getCachedIcon(String name) {
    	if (cache==null) {
    		return null;
    	}
        if (cache.containsKey(name)) {
            return cache.get(name);
        }

        return null;
    }

    /**
     * Puts icon in the cache. If an icon with the specified name has already been stored,
     * current icon will be overwritten.
     *
     * @param name
     * @param icon
     */
    private static synchronized void putCachedIcon(String name, Icon icon) {
    	if (cache==null) {
    		cache = new HashMap<>();
    	}
        cache.put(name, icon);
    }

    /**
     * Attempts to load the icon from the resources in classpath.
     * If the icon cannot be found or other error occurs, the return value is null.
     *
     * @param resource String
     * @param cl ClassLoader
     * @return Icon icon instance or null
     */
    private static final Icon loadResource(String resource, ClassLoader cl) {
        try {
        	URL url = cl.getResource(resource);
        	if (!(url.getContent() instanceof ImageProducer)) return null;
            return new ImageIcon(url);
        } catch (Exception e) {
            return null;
        }
    }

    /**
     * Returns instance of the icon by loading it using the <code>
     * ClassLoader</code> instance provided. Icons are cached. If an icon with
     * the same name is requested several times, <code>createIcon</code> will
     * return the same instance each time. If icons will be manipulated
     * individually, do not use this method to load icons. If icon is not found
     * in the classpath, return value will be null.
     *
     * @param name String
     * @param cl ClassLoader Instance of class loader to use
     * @return Icon Instance of Icon or null if icon could not be found.
     */
    public static Icon createIcon(String name, ClassLoader cl) {
        Icon icon = getCachedIcon(name);

        if (icon == null) {
            icon = loadResource(name, cl);

            if (icon != null) {
                putCachedIcon(name, icon);
            } else {
            	return createIcon(new File(name));
            }
        }

        return icon;
    }

    /**
     * Returns instance of the icon. Icons are cached. If an icon with the same
     * name is requested several times, <code>createIcon</code> will return the
     * same instance each time. If icons will be manipulated individually, do
     * not use this method to load icons. If icon is not found in the classpath,
     * or the file with that name does not exist on the disk return value will be null.
     *
     * @param name String
     * @return Icon Instance of Icon or null if icon could not be found.
     */
    public static Icon createIcon(String name) {
        return createIcon(name, IconHelper.class.getClassLoader());
    }

    /**
     * Creates a new image loaded from the provided image file name.
     * This class' class loader is used to access resources.
     * 
     * @param name the filename of the image
     * @return Image
     */
    public static Image createImage(String name) {
        return createImage(name, IconHelper.class.getClassLoader());
    }

    /**
     * Creates a new image loaded from the provided image file. The image
     * is loaded using the given class loader. 
     * 
     * @param name the filename of the image
     * @param cl classloader to be used for accessing resources
     * @return Image
     */
    public static Image createImage(String name, ClassLoader cl) {
        ImageIcon icon = ((ImageIcon) createIcon(name, cl));

        if (icon == null) {
            return null;
        }

        return icon.getImage();
    }
    
    /**
     * Creates an Icon by reading the image stored in the given file. This file
     * can be anywhere on the computer as long as it is accessible by the current user
     * rights (the icon file does not need to be added to classpath).
     * 
     * @param file the image file
     * @return icon
     */
    public static Icon createIcon(File file) {
    	if (file == null) return null;
    	Icon icon = getCachedIcon(file.getAbsolutePath());
    	if (icon == null) {
    		icon = new ImageIcon(file.getAbsolutePath());
   			putCachedIcon(file.getAbsolutePath(), icon);
    	}
    	return icon;
    }
}
