/*
 * Copyright (c) 2003-2008 by Cosylab d. d.
 *
 * This file is part of CosyBeans-Common.
 *
 * CosyBeans-Common is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * CosyBeans-Common is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with CosyBeans-Common.  If not, see <http://www.gnu.org/licenses/>.
 */

package com.cosylab.gui.components.wheelswitch;

import java.awt.Color;
import java.awt.Dimension;
import java.awt.event.MouseListener;

import javax.swing.JComponent;
import javax.swing.event.EventListenerList;

import com.cosylab.gui.components.ArrowButton;
import com.cosylab.gui.components.ArrowButton.Orientation;
import com.cosylab.gui.components.util.ColorHelper;

/**
 * Descedant of <code>javax.swing.JComponent</code> that contains two 
 * <code>ArrowButton</code>s acting as a two-way (up/down) control. 
 * 
 * @author <a href="mailto:jernej.kamenik@cosylab.com">Jernej Kamenik</a>
 * @version $id$
 */
public class UpDownButton extends JComponent {

	private static final long serialVersionUID = 1L;

	private ArrowButton down;

	private ArrowButton up;
	
	private Color foregroundColor = Color.BLACK;
	
	private int outOfBounds = 0;

	/**
	 * Constructor for UpDownButton.
	 */			
	public UpDownButton() {
		super();
		listenerList = new EventListenerList();	
		setLayout(new UpDownLayout());
		add(getUpButton());
		add(getDownButton());
	}		

  /**
   * This method was overridden to implement correct 
   * <code>UpDownButton<code> placement inside the 
   * <code>Wheelswitch</code> container.
   * 
   * @see JComponent#getPreferredSize()
   */
	@Override
	public Dimension getPreferredSize() {
		return new Dimension(12,24);
	}
	
  /**
   * This method was overridden to implement correct 
   * <code>UpDownButton<code> placement inside the 
   * <code>Wheelswitch</code> container.
   * 
   * @see JComponent#getMinimumSize()
   */
	@Override
	public Dimension getMinimumSize() {
		return new Dimension(12,24);
	}	
	/* (non-Javadoc)
	 * @see java.awt.Component#setEnabled(boolean)
	 */
	@Override
	public void setEnabled(boolean arg0) {
		super.setEnabled(arg0);
		getDownButton().setEnabled(arg0);
		getUpButton().setEnabled(arg0);
	}

	/**
	 * Sets this button to out of bounds state. When in such state the button
	 * will be painted with alarm color foreground.
	 * 
	 * @param i
	 */
	public void setOutOfBounds(int i) {
		outOfBounds = i;
		if (i>0) {
			getUpButton().setEnabled(false);
			if (isEnabled())getDownButton().setEnabled(true);
			getUpButton().setForeground(foregroundColor);
			getDownButton().setForeground(ColorHelper.getAlarmOutline());
		} else if (i<0) {
			if (isEnabled())getUpButton().setEnabled(true);
			getDownButton().setEnabled(false);
			getUpButton().setForeground(ColorHelper.getAlarmOutline());
			getDownButton().setForeground(foregroundColor);
		} else {
			if (isEnabled()) {
				getUpButton().setEnabled(true);
				getDownButton().setEnabled(true);
			}
			getUpButton().setForeground(foregroundColor);
			getDownButton().setForeground(foregroundColor);
		}
	}
	
	/**
	 * Laziliy initializes the up arrow button and returns it.
	 * 
	 * @return up button
	 */
	public ArrowButton getUpButton() {
		if (up==null) {
			up = new ArrowButton(Orientation.UP);
			up.setFocusable(false);
			up.setBackground(ColorHelper.getCosyControl());
		}

		return up;
	}
	
	/**
	 * Laziliy initializes the down arrow button and returns it.
	 * 
	 * @return down button
	 */
	public ArrowButton getDownButton() {
		if (down==null) {
			down = new ArrowButton(Orientation.DOWN);
			down.setFocusable(false);
			down.setBackground(ColorHelper.getCosyControl());
		}
		
		return down;
	}
	
	/*
	 * (non-Javadoc)
	 * @see javax.swing.JComponent#setBackground(java.awt.Color)
	 */
	@Override
	public void setBackground(Color bg) {
		super.setBackground(bg);
		getUpButton().setBackground(bg);
		getDownButton().setBackground(bg);
	}
	
	/*
	 * (non-Javadoc)
	 * @see javax.swing.JComponent#setForeground(java.awt.Color)
	 */
	@Override
	public void setForeground(Color fg) {
		super.setForeground(fg);
		foregroundColor = fg;
		setOutOfBounds(outOfBounds);
	}
	
	@Override
	public synchronized void addMouseListener(MouseListener l) {
		super.addMouseListener(l);
		getUpButton().addMouseListener(l);
		getDownButton().addMouseListener(l);
	}

}

