/*
 * Copyright (c) 2003-2008 by Cosylab d. d.
 *
 * This file is part of CosyBeans.
 *
 * CosyBeans is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * CosyBeans is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with CosyBeans.  If not, see <http://www.gnu.org/licenses/>.
 */

package com.cosylab.gui.displayers;

import java.awt.Color;
import java.awt.Dimension;
import java.awt.event.ActionEvent;
import java.beans.PropertyVetoException;

import javax.swing.AbstractAction;
import javax.swing.JLabel;
import javax.swing.JPanel;

import com.cosylab.gui.adapters.Converter;
import com.cosylab.gui.components.AbstractDisplayerPanelCustomizer;
import com.cosylab.gui.components.ResizableTextLabel;
import com.cosylab.gui.components.customizer.AbstractCustomizerPanel;
import com.cosylab.gui.components.util.CosyUIElements;
import com.cosylab.gui.components.util.PopupManageable;
import com.cosylab.gui.components.util.PopupManager;
import com.cosylab.gui.components.util.ScreenCapturer;


/**
 * <code>AbstractDisplayerPanel</code> is a displayer that provides common
 * feature of a Displayer interface.
 *
 * @author <a href="mailto:jernej.kamenik@cosylab.com">Jernej Kamenik</a>
 * @version $id$
 */
public abstract class AbstractDisplayerPanel extends JPanel implements Displayer,
	PopupManageable, ConvertibleDisplayer
{
	/**
     * 
     */
    private static final long serialVersionUID = 2150144397643937575L;
    private DataState dataState = new DataState(DataState.UNDEFINED);
	private int suspendCount = 0;
	private PopupManager popupManager;
	private boolean popupEnabled;
	private DataSource dataSource;
	private Converter converter;
	
	private ResizableTextLabel titleLabel;
	private AbstractCustomizerPanel customizer;

	/**
	 * Constructor for AbstractDisplayerPanel.
	 */
	public AbstractDisplayerPanel()
	{
		super();
		setBorder(CosyUIElements.getPlainBorder(true));

		Dimension size = new Dimension(200, 50);
		setPreferredSize(size);
		setPopupEnabled(true);
	}
	
	/**
	 * Generates the customizer for the displayer.
	 * 
	 * @return
	 */
	public AbstractCustomizerPanel getCustomizer()
	{
		if (customizer == null) {
			customizer = AbstractCustomizerPanel.findCustomizer(this);
		}

		return customizer;
	}
	
	protected ResizableTextLabel getTitleLabel() {
		if (titleLabel == null) {
			titleLabel = new ResizableTextLabel();
			titleLabel.setHorizontalAlignment(JLabel.CENTER);
			titleLabel.addMouseListener(getPopupManager().getMouseHook());
			titleLabel.setEnabled(isEnabled());
		}

		return titleLabel;
	}

	/* (non-Javadoc)
	 * @see com.cosylab.gui.displayers.DataConsumer#getDefaultDataConsumer()
	 */
	public DataConsumer getDefaultDataConsumer()
	{
		return this;
	}

	/* (non-Javadoc)
	 * @see com.cosylab.gui.displayers.DataConsumer#getSupportedCharacteristics()
	 */
	public String[] getSupportedCharacteristics()
	{
		return DisplayerUtilities.COMMON_NUMERIC_DISPLAYER_CHARACTERISTICS;
	}

	/* (non-Javadoc)
	 * @see com.cosylab.gui.displayers.DataConsumer#updateDataState(com.cosylab.gui.displayers.DataState)
	 */
	public void updateDataState(DataState state)
	{
		DataState old = dataState;
		dataState = state;
		firePropertyChange(DATA_STATE, old, dataState);
	}

	/* (non-Javadoc)
	 * @see com.cosylab.gui.displayers.CommonDisplayer#cleanup()
	 */
	public void cleanup()
	{
		updateDataState(new DataState(DataState.NOT_INITIALIZED));
	}

	/* (non-Javadoc)
	 * @see com.cosylab.gui.displayers.CommonDisplayer#getDataState()
	 */
	public DataState getDataState()
	{
		return dataState;
	}

	/* (non-Javadoc)
	 * @see com.cosylab.gui.displayers.CommonDisplayer#isSuspended()
	 */
	public boolean isSuspended()
	{
		return suspendCount > 0;
	}

	/* (non-Javadoc)
	 * @see com.cosylab.gui.displayers.CommonDisplayer#resume()
	 */
	public void resume()
	{
		suspendCount--;
	}

	/* (non-Javadoc)
	 * @see com.cosylab.gui.displayers.CommonDisplayer#suspend()
	 */
	public void suspend()
	{
		suspendCount++;
	}

	/**
	 * Returns popup manager for adding popup actions.
	 *
	 * @see com.cosylab.gui.components.util.PopupManageable#getPopupManager()
	 */
	public PopupManager getPopupManager()
	{
		if (popupManager == null) {
			popupManager = new PopupManager(this, false);
			popupManager.addAction(new AbstractAction("Preferences...") {
				private static final long serialVersionUID = 1L;
					public void actionPerformed(ActionEvent e)
					{
						getCustomizer().showDialog();
					}
				});
			popupManager.addAction(new AbstractAction("Capture screen..."){
				private static final long serialVersionUID = 1L;

				public void actionPerformed(ActionEvent e){
				   	ScreenCapturer sc = new ScreenCapturer(AbstractDisplayerPanel.this);
					sc.showScreenDialog();
				   }
			});
		}

		return popupManager;
	}
	
	/**
	 * Return true if the popup menu is enabled or false otherwise.
	 * 
	 * @return true if popup is enabled
	 */
	public boolean isPopupEnabled() {
		return popupEnabled;
	}
	
	/**
	 * Enables or disables the popup menu.
	 * 
	 * @param enabled true if enable or false if disableds
	 */
	public void setPopupEnabled(boolean enabled) {
		if (popupEnabled == enabled) return;
		popupEnabled = enabled;
		if (enabled) {
			addMouseListener(getPopupManager().getMouseHook());
		} else {
			removeMouseListener(getPopupManager().getMouseHook());
		}
		firePropertyChange("popupEnabled",!popupEnabled,popupEnabled);
	}

	/* (non-Javadoc)
	 * @see com.cosylab.gui.core.CosyComponent#destroy()
	 */
	public void destroy()
	{
		popupManager = null;
		dataState = null;
	}

	/* (non-Javadoc)
	 * @see com.cosylab.gui.displayers.Displayer#getDataSource()
	 */
	public DataSource getDataSource() {
		return dataSource;
	}
	
	/* (non-Javadoc)
	 * @see com.cosylab.gui.displayers.Displayer#setDataSource(com.cosylab.gui.displayers.DataSource)
	 */
	public void setDataSource(DataSource dataSource)
		throws PropertyVetoException
	{
		DisplayerUtilities.prepareNewDataSource(dataSource,this);
		
		DataSource old= this.dataSource;
		this.dataSource = dataSource;
		
		firePropertyChange(DATA_SOURCE,old,dataSource);
	}
	
	/**
	 * Returns the converter.
	 * @return Returns the converter.
	 */
	public Converter getConverter() {
		return converter;
	}
	
	/**
	 * Sets new converter.
	 * @param converter new converter
	 * @throws PropertyVetoException if set fails
	 */
	public void setConverter(Converter converter) throws PropertyVetoException {
		DisplayerUtilities.prepareNewConverter(converter,this);
		
		Converter old= this.converter;
		this.converter = converter;
		
		firePropertyChange(CONVERTER_PROPERTY,old,dataSource);
	}
	
	/**
	 * Returns the maximum title font size.
	 *
	 * @return font size in pixels
	 */
	public int getTitleMaximumFontSize()
	{
		return getTitleLabel().getMaximumFontSize();
	}

	/**
	 * Returns the minimum title font size.
	 *
	 * @return font size in pixels
	 */
	public int getTitleMinimumFontSize()
	{
		return getTitleLabel().getMinimumFontSize();
	}

	/**
	 * Sets the maximum title font size.
	 *
	 * @param max new font size in pixels
	 */
	public void setTitleMaximumFontSize(int max)
	{
		int old= getTitleMaximumFontSize();
		getTitleLabel().setMaximumFontSize(max);
		firePropertyChange(AbstractDisplayerPanelCustomizer.MAX_TITLE_FONT_SIZE,
		    old, max);
	}

	/**
	 * Sets the minimum title font size.
	 *
	 * @param min new font size in pixels
	 */
	public void setTitleMinimumFontSize(int min)
	{
		int old= getTitleMinimumFontSize();
		getTitleLabel().setMaximumFontSize(min);
		firePropertyChange(AbstractDisplayerPanelCustomizer.MIN_TITLE_FONT_SIZE,
		    old, min);
	}
	/**
     * Sets visibility of title label.
     * 
     * @param value
     *        True if title should be visible.
     */
	public void setTitleVisible(boolean value) {
		getTitleLabel().setVisible(value);
	}

	/**
     * Returns visibility of title.
     * 
     * @return True if title is visible.
     */
	public boolean isTitleVisible() {
		return getTitleLabel().isVisible();
	}

	/**
     * Returns title of this component.
     * 
     * @see com.cosylab.gui.displayers.Displayer#getTitle()
     */
	public String getTitle() {
		return getTitleLabel().getText();
	}

	/**
     * Sets the title of component.
     * 
     * @see com.cosylab.gui.displayers.Displayer#setTitle(java.lang.String)
     */
	public void setTitle(String value) {
		getTitleLabel().setText(value);
	}
	
	/*
	 * (non-Javadoc)
	 * @see javax.swing.JComponent#setBackground(java.awt.Color)
	 */
	@Override
	public void setBackground(Color bg) {
		super.setBackground(bg);
		getTitleLabel().setBackground(bg);
		
	}
	
	/*
	 * (non-Javadoc)
	 * @see javax.swing.JComponent#setForeground(java.awt.Color)
	 */
	@Override
	public void setForeground(Color fg) {
		super.setForeground(fg);
		getTitleLabel().setForeground(fg);
	}


}

/* __oOo__ */
