/*
 * Copyright (c) 2003-2008 by Cosylab d. d.
 *
 * This file is part of CosyBeans.
 *
 * CosyBeans is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * CosyBeans is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with CosyBeans.  If not, see <http://www.gnu.org/licenses/>.
 */

package com.cosylab.gui.displayers;

import java.util.Map;

import com.cosylab.util.CommonException;


/**
 * <code>DoubleConsumerMulticaster</code> casts different data consumers to
 * DoubleConsumer. ObjectConsumer cast must receive <code>Number</code>
 * instances. Delegate must be set, before any other method can be used.
 *
 * @author <a href="mailto:igor.kriznar@cosylab.com">Igor Kriznar</a>
 * @version $Id: DoubleConsumerMulticaster.java,v 1.8 2008-04-22 12:31:02 jbobnar Exp $
 *
 * @since Dec 4, 2003.
 */
public class DoubleConsumerMulticaster implements DoubleConsumer,
	NumberConsumer, LongConsumer, StringConsumer, ObjectConsumer
{
	/** Types of data consumers, that are supported by this multicaster. */
	public static final Class[] SUPPORTED_CONSUMER_TYPES = {
		DoubleConsumer.class, LongConsumer.class, NumberConsumer.class,
		ObjectConsumer.class, StringConsumer.class
	};

	/**
	 * Types of data consumers, that should be prefered by DoubleConsumer
	 * parent.
	 */
	public static final Class[] PREFERED_CONSUMER_TYPES = {
		DoubleConsumer.class, LongConsumer.class, NumberConsumer.class
	};
	DoubleConsumer delegate;

	/**
	 * If multicaster supports type, new multicaster is created, otherwise
	 * <code>null</code>is returned.
	 *
	 * @param type the requested consumer type
	 * @param delegate the consumer to which all call sare delegated
	 *
	 * @return new multicaster if type is supported, otherwise
	 *         <code>null</code>
	 */
	public static final DoubleConsumer createDataConsumer(Class type,
	    DoubleConsumer delegate)
	{
		if (type == DoubleConsumer.class) {
			return delegate;
		}

		if (type == LongConsumer.class || type == NumberConsumer.class
		    || type == StringConsumer.class || type == ObjectConsumer.class) {
			return new DoubleConsumerMulticaster(delegate);
		}

		return null;
	}

	/**
	 * Creates a new DoubleConsumerMulticaster object.
	 *
	 * @param delegate the consumer to which all call sare delegated
	 */
	public DoubleConsumerMulticaster(DoubleConsumer delegate)
	{
		this.delegate = delegate;
	}

	/**
	 * Creates a new DoubleConsumerMulticaster object.
	 */
	public DoubleConsumerMulticaster()
	{
	}

	/**
	 * Returns the consumer to which all call sare delegated.
	 *
	 * @return the consumer to which all call sare delegated
	 */
	public DoubleConsumer getDelegate()
	{
		return delegate;
	}

	/**
	 * Sets the consumer to which all call sare delegated.
	 *
	 * @param delegate the consumer to which all calls are delegated
	 */
	public void setDelegate(DoubleConsumer delegate)
	{
		this.delegate = delegate;
	}

	/* (non-Javadoc)
	 * @see com.cosylab.gui.displayers.DoubleConsumer#updateValue(long, double)
	 */
	public void updateValue(long timestamp, double value)
		throws CommonException
	{
		delegate.updateValue(timestamp, value);
	}

	/* (non-Javadoc)
	 * @see com.cosylab.gui.displayers.NumberConsumer#updateValue(long, java.lang.Number)
	 */
	public void updateValue(long timestamp, Number value)
		throws CommonException
	{
		delegate.updateValue(timestamp, value.doubleValue());
	}

	/* (non-Javadoc)
	 * @see com.cosylab.gui.displayers.LongConsumer#updateValue(long, long)
	 */
	public void updateValue(long timestamp, long value)
		throws CommonException
	{
		delegate.updateValue(timestamp, value);
	}

	/* (non-Javadoc)
	 * @see com.cosylab.gui.displayers.DataConsumer#getDefaultDataConsumer()
	 */
	public DataConsumer getDefaultDataConsumer()
	{
		return delegate.getDefaultDataConsumer();
	}

	/* (non-Javadoc)
	 * @see com.cosylab.gui.displayers.DataConsumer#getName()
	 */
	public String getName()
	{
		return delegate.getName();
	}

	/* (non-Javadoc)
	 * @see DataConsumer#setCharacteristics(Map)
	 */
	public void setCharacteristics(Map characteristics)
	{
		delegate.setCharacteristics(characteristics);
	}

	/* (non-Javadoc)
	 * @see com.cosylab.gui.displayers.DataConsumer#updateDataState(com.cosylab.gui.displayers.DataState)
	 */
	public void updateDataState(DataState state)
	{
		delegate.updateDataState(state);
	}

	/* (non-Javadoc)
	 * @see com.cosylab.gui.displayers.StringConsumer#updateValue(long, java.lang.String)
	 */
	public void updateValue(long timestamp, String value)
		throws CommonException
	{
		delegate.updateValue(timestamp, Double.parseDouble(value));
	}

	/* (non-Javadoc)
	 * @see com.cosylab.gui.displayers.ObjectConsumer#updateValue(long, java.lang.Object)
	 */
	public void updateValue(long timestamp, Object value)
		throws CommonException
	{
		delegate.updateValue(timestamp, ((Number)value).doubleValue());
	}

	/* (non-Javadoc)
	 * @see com.cosylab.gui.displayers.DataConsumer#getSupportedCharacteristics()
	 */
	public String[] getSupportedCharacteristics()
	{
		return delegate.getSupportedCharacteristics();
	}
	
	/*
	 * (non-Javadoc)
	 * @see com.cosylab.gui.displayers.DataConsumer#getDataConsumer(java.lang.Class)
	 */
	public <D extends DataConsumer> D getDataConsumer(Class<D> type) {
		return delegate.getDataConsumer(type);
	}

	/*
	 * (non-Javadoc)
	 * @see com.cosylab.gui.displayers.DataConsumer#getSupportedConsumerTypes()
	 */
	@SuppressWarnings("unchecked")
	public Class<DataConsumer>[] getSupportedConsumerTypes() {
		return PREFERED_CONSUMER_TYPES;
	}
}

/* __oOo__ */
