/*
 * Copyright (c) 2003-2008 by Cosylab d. d.
 *
 * This file is part of CosyBeans.
 *
 * CosyBeans is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * CosyBeans is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with CosyBeans.  If not, see <http://www.gnu.org/licenses/>.
 */

package com.cosylab.gui.displayers;

import java.util.BitSet;
import java.util.Map;

import com.cosylab.util.CommonException;


/**
 * <code>ObjectConsumerMulticaster</code> delegates all defined data sources to
 * <code>ObjectConsumer</code> delegate.
 *
 * @author <a href="mailto:igor.kriznar@cosylab.com">Igor Kriznar</a>
 * @version $Id: ObjectConsumerMulticaster.java,v 1.5 2008-04-22 12:31:02 jbobnar Exp $
 *
 * @since Feb 5, 2004.
 */
public class ObjectConsumerMulticaster implements DoubleConsumer,
	DoubleSeqConsumer, LongConsumer, LongSeqConsumer, NumberConsumer,
	ObjectConsumer, ObjectSeqConsumer, PatternConsumer, StringConsumer,
	StringSeqConsumer
{
	/** Types of data consumers, that are supported by this multicaster. */
	public static final Class[] SUPPORTED_CONSUMER_TYPES = {
		DoubleConsumer.class, DoubleSeqConsumer.class, LongConsumer.class,
		LongSeqConsumer.class, NumberConsumer.class, ObjectConsumer.class,
		ObjectSeqConsumer.class, PatternConsumer.class, StringConsumer.class,
		StringSeqConsumer.class
	};

	/**
	 * If multicaster supports type, new multicaster is created, otherwise
	 * <code>null</code>is returned.
	 *
	 * @param type the requested consumer type
	 * @param delegate the consumer to which all call sare delegated
	 *
	 * @return new multicaster if type is supported, otherwise
	 *         <code>null</code>
	 */
	public static final ObjectConsumer createDataConsumer(Class type,
	    ObjectConsumer delegate)
	{
		if (type == ObjectConsumer.class) {
			return delegate;
		}

		return new ObjectConsumerMulticaster(delegate);
	}

	private ObjectConsumer delegate;

	/**
	 * Creates multicaster without delegate. Set delegate before usage.
	 */
	public ObjectConsumerMulticaster()
	{
		super();
	}

	/**
	 * Creates a new ObjectConsumerMulticaster object.
	 *
	 * @param delegate DOCUMENT ME!
	 */
	public ObjectConsumerMulticaster(ObjectConsumer delegate)
	{
		super();
		this.delegate = delegate;
	}

	/**
	 * Returns delegated consumer.
	 *
	 * @return delegated consumer
	 */
	public ObjectConsumer getDelegate()
	{
		return delegate;
	}

	/**
	 * Sets delegated consumer.
	 *
	 * @param consumer delegated
	 */
	public void setDelegate(ObjectConsumer consumer)
	{
		delegate = consumer;
	}

	/* (non-Javadoc)
	 * @see com.cosylab.gui.displayers.DoubleConsumer#updateValue(long, double)
	 */
	public void updateValue(long timestamp, double value)
		throws CommonException
	{
		delegate.updateValue(timestamp, new Double(value));
	}

	/* (non-Javadoc)
	 * @see com.cosylab.gui.displayers.DoubleSeqConsumer#updateValue(long, double[])
	 */
	public void updateValue(long timestamp, double[] value)
		throws CommonException
	{
		delegate.updateValue(timestamp, value);
	}

	/* (non-Javadoc)
	 * @see com.cosylab.gui.displayers.LongConsumer#updateValue(long, long)
	 */
	public void updateValue(long timestamp, long value)
		throws CommonException
	{
		delegate.updateValue(timestamp, new Long(value));
	}

	/* (non-Javadoc)
	 * @see com.cosylab.gui.displayers.LongSeqConsumer#updateValue(long, long[])
	 */
	public void updateValue(long timestamp, long[] value)
		throws CommonException
	{
		delegate.updateValue(timestamp, value);
	}

	/* (non-Javadoc)
	 * @see com.cosylab.gui.displayers.NumberConsumer#updateValue(long, java.lang.Number)
	 */
	public void updateValue(long timestamp, Number value)
		throws CommonException
	{
		delegate.updateValue(timestamp, value);
	}

	/* (non-Javadoc)
	 * @see com.cosylab.gui.displayers.ObjectConsumer#updateValue(long, java.lang.Object)
	 */
	public void updateValue(long timestamp, Object value)
		throws CommonException
	{
		delegate.updateValue(timestamp, value);
	}

	/* (non-Javadoc)
	 * @see com.cosylab.gui.displayers.ObjectSeqConsumer#updateValue(long, java.lang.Object[])
	 */
	public void updateValue(long timestamp, Object[] value)
		throws CommonException
	{
		delegate.updateValue(timestamp, value);
	}

	/* (non-Javadoc)
	 * @see com.cosylab.gui.displayers.PatternConsumer#updateValue(long, java.util.BitSet)
	 */
	public void updateValue(long timestamp, BitSet value)
		throws CommonException
	{
		delegate.updateValue(timestamp, value);
	}

	/* (non-Javadoc)
	 * @see com.cosylab.gui.displayers.StringConsumer#updateValue(long, java.lang.String)
	 */
	public void updateValue(long timestamp, String value)
		throws CommonException
	{
		delegate.updateValue(timestamp, value);
	}

	/* (non-Javadoc)
	 * @see com.cosylab.gui.displayers.StringSeqConsumer#updateValue(long, java.lang.String[])
	 */
	public void updateValue(long timestamp, String[] value)
		throws CommonException
	{
		delegate.updateValue(timestamp, value);
	}

	/* (non-Javadoc)
	 * @see com.cosylab.gui.displayers.DataConsumer#getDefaultDataConsumer()
	 */
	public DataConsumer getDefaultDataConsumer()
	{
		return delegate.getDefaultDataConsumer();
	}

	/* (non-Javadoc)
	 * @see com.cosylab.gui.displayers.DataConsumer#getName()
	 */
	public String getName()
	{
		return delegate.getName();
	}

	/* (non-Javadoc)
	 * @see com.cosylab.gui.displayers.DataConsumer#getSupportedCharacteristics()
	 */
	public String[] getSupportedCharacteristics()
	{
		return delegate.getSupportedCharacteristics();
	}

	/* (non-Javadoc)
	 * @see com.cosylab.gui.displayers.DataConsumer#getSupportedConsumerTypes()
	 */
	@SuppressWarnings("unchecked")
	public Class[] getSupportedConsumerTypes()
	{
		return SUPPORTED_CONSUMER_TYPES;
	}

	/* (non-Javadoc)
	 * @see com.cosylab.gui.displayers.DataConsumer#setCharacteristics(java.util.Map)
	 */
	public void setCharacteristics(Map characteristics)
	{
		delegate.setCharacteristics(characteristics);
	}

	/* (non-Javadoc)
	 * @see com.cosylab.gui.displayers.DataConsumer#updateDataState(com.cosylab.gui.displayers.DataState)
	 */
	public void updateDataState(DataState state)
	{
		delegate.updateDataState(state);
	}

	/*
	 * (non-Javadoc)
	 * @see com.cosylab.gui.displayers.DataConsumer#getDataConsumer(java.lang.Class)
	 */
	public <D extends DataConsumer> D getDataConsumer(Class<D> type) {
		return delegate.getDataConsumer(type);
	}
}

/* __oOo__ */
