/*
 * Copyright (c) 2003-2008 by Cosylab d. d.
 *
 * This file is part of CosyBeans.
 *
 * CosyBeans is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * CosyBeans is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with CosyBeans.  If not, see <http://www.gnu.org/licenses/>.
 */

package com.cosylab.gui.displayers;

import java.util.BitSet;
import java.util.HashMap;
import java.util.Map;

import com.cosylab.util.BitSetUtilities;
import com.cosylab.util.CommonException;


/**
 * <code>PatternConsumerMulticaster</code> casts different data consumers to
 * PatternConsumer. Delegate must be set, before any other method can be used.
 *
 * @author <a href="mailto:anze.zupanc@cosylab.com">Anze Zupanc</a>
 * @version $Id: PatternConsumerMulticaster.java,v 1.7 2008-04-22 12:31:02 jbobnar Exp $
 *
 * @since Jul 27, 2004.
 */
public class PatternConsumerMulticaster implements PatternConsumer,
	LongConsumer, StringConsumer, DoubleConsumer, NumberConsumer,
	ObjectConsumer
{
	/** Types of data consumers, that are supported by this multicaster. */
	public static final Class[] SUPPORTED_CONSUMER_TYPES = {
			PatternConsumer.class, LongConsumer.class, StringConsumer.class,
			DoubleConsumer.class, NumberConsumer.class, ObjectConsumer.class
		};

	/**
	 * If multicaster supports type, new multicaster is created, otherwise
	 * <code>null</code>is returned.
	 *
	 * @param type the requested consumer type
	 * @param delegate the consumer to which all call sare delegated
	 *
	 * @return new multicaster if type is supported, otherwise
	 *         <code>null</code>
	 */
	public static final PatternConsumer createDataConsumer(Class type,
	    PatternConsumer delegate)
	{
		if (type == PatternConsumer.class) {
			return delegate;
		}

		if (type == NumberConsumer.class || type == StringConsumer.class
		    || type == ObjectConsumer.class || type == DoubleConsumer.class
		    || type == LongConsumer.class) {
			return new PatternConsumerMulticaster(delegate);
		}

		return null;
	}

	PatternConsumer delegate;

	/**
	 * Creates a new PatternConsumerMulticaster object.
	 */
	public PatternConsumerMulticaster()
	{
		super();
	}

	/**
	 * Creates a new PatternConsumerMulticaster object.
	 *
	 * @param delegate the consumer to which all call sare delegated
	 */
	public PatternConsumerMulticaster(PatternConsumer delegate)
	{
		this.delegate = delegate;
	}

	/* (non-Javadoc)
	 * @see com.cosylab.gui.displayers.DataConsumer#getDefaultDataConsumer()
	 */
	public DataConsumer getDefaultDataConsumer()
	{
		return delegate.getDefaultDataConsumer();
	}

	/**
	 * Returns the consumer to which all call sare delegated.
	 *
	 * @return the consumer to which all call sare delegated
	 */
	public PatternConsumer getDelegate()
	{
		return delegate;
	}

	/* (non-Javadoc)
	 * @see com.cosylab.gui.displayers.DataConsumer#getName()
	 */
	public String getName()
	{
		return delegate.getName();
	}

	/* (non-Javadoc)
	 * @see com.cosylab.gui.displayers.DataConsumer#getSupportedCharacteristics()
	 */
	public String[] getSupportedCharacteristics()
	{
		return delegate.getSupportedCharacteristics();
	}

	/* (non-Javadoc)
	 * @see com.cosylab.gui.displayers.DataConsumer#getSupportedConsumerTypes()
	 */
	@SuppressWarnings("unchecked")
	public Class[] getSupportedConsumerTypes()
	{
		return SUPPORTED_CONSUMER_TYPES;
	}

	/* (non-Javadoc)
	 * @see DataConsumer#setCharacteristics(Map)
	 */
	public void setCharacteristics(Map characteristics)
	{
		String f = (String)characteristics.get(CommonDisplayer.C_FORMAT);

		if (f != null) {
			HashMap map = new HashMap(characteristics);
			map.put(CommonDisplayer.C_FORMAT, DisplayerUtilities.toLongFormat(f));
			delegate.setCharacteristics(map);
		} else {
			delegate.setCharacteristics(characteristics);
		}
	}

	/**
	 * Sets the consumer to which all call sare delegated.
	 *
	 * @param delegate the consumer to which all calls are delegated
	 */
	public void setDelegate(PatternConsumer delegate)
	{
		this.delegate = delegate;
	}

	/* (non-Javadoc)
	 * @see com.cosylab.gui.displayers.DataConsumer#updateDataState(com.cosylab.gui.displayers.DataState)
	 */
	public void updateDataState(DataState state)
	{
		delegate.updateDataState(state);
	}

	/* (non-Javadoc)
	 * @see com.cosylab.gui.displayers.PatternConsumer#updateValue(long, java.util.BitSet)
	 */
	public void updateValue(long timestamp, BitSet value)
		throws CommonException
	{
		delegate.updateValue(timestamp, value);
	}

	/* (non-Javadoc)
	 * @see com.cosylab.gui.displayers.DoubleConsumer#updateValue(long, double)
	 */
	public void updateValue(long timestamp, double value)
		throws CommonException
	{
		delegate.updateValue(timestamp, BitSetUtilities.forLong((long)value));
	}

	/* (non-Javadoc)
	 * @see com.cosylab.gui.displayers.LongConsumer#updateValue(long, long)
	 */
	public void updateValue(long timestamp, long value)
		throws CommonException
	{
		delegate.updateValue(timestamp, BitSetUtilities.forLong(value));
	}

	/* (non-Javadoc)
	 * @see com.cosylab.gui.displayers.NumberConsumer#updateValue(long, java.lang.Number)
	 */
	public void updateValue(long timestamp, Number value)
		throws CommonException
	{
		delegate.updateValue(timestamp,
		    BitSetUtilities.forLong(value.longValue()));
	}

	/* (non-Javadoc)
	 * @see com.cosylab.gui.displayers.ObjectConsumer#updateValue(long, java.lang.Object)
	 */
	public void updateValue(long timestamp, Object value)
		throws CommonException
	{
		delegate.updateValue(timestamp,
		    BitSetUtilities.forLong(((Number)value).longValue()));
	}

	/* (non-Javadoc)
	 * @see com.cosylab.gui.displayers.StringConsumer#updateValue(long, java.lang.String)
	 */
	public void updateValue(long timestamp, String value)
		throws CommonException
	{
		delegate.updateValue(timestamp,
		    BitSetUtilities.forLong(Long.parseLong(value)));
	}

	/*
	 * (non-Javadoc)
	 * @see com.cosylab.gui.displayers.DataConsumer#getDataConsumer(java.lang.Class)
	 */
	public <D extends DataConsumer> D getDataConsumer(Class<D> type) {
		return delegate.getDataConsumer(type);
	}
}

/* __oOo__ */
