/*
 * Copyright (c) 2003-2008 by Cosylab d. d.
 *
 * This file is part of CosyBeans-Common.
 *
 * CosyBeans-Common is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * CosyBeans-Common is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with CosyBeans-Common.  If not, see <http://www.gnu.org/licenses/>.
 */

package com.cosylab.gui.property.editors;

import java.awt.Color;
import java.awt.Insets;
import java.awt.event.FocusAdapter;
import java.awt.event.FocusEvent;
import java.awt.event.KeyAdapter;
import java.awt.event.KeyEvent;
import java.lang.reflect.Modifier;

import javax.swing.JTextField;

import com.cosylab.gui.components.util.ColorHelper;
import com.cosylab.util.StringUtilities;

/**
 * A component used for specifyind <code>Class<code> objects.
 * User can enter the name of the class and upon pressing
 * enter or on loosing focus the <code>Class</code> instance
 * is obtained and propertyChange events are fired. If the class name
 * is invalid, the text in the textField is colored red.
 * 
 * @author <a href="mailto:miha.kadunc@cosylab.com">Miha Kadunc</a> 
 * @version $id$
 */
public class ClassEditor extends JTextField implements PropertyEditor {
	/**
     * 
     */
    private static final long serialVersionUID = -3114142314491216392L;
    public static final String APPLY_MODE_PROPERTY="applyMode";
	public static final int APPLY_ON_CHANGE=4;
	public static final int APPLY_ON_FOCUS_LOST=2;
	public static final int APPLY_ON_ENTER=1;
	public static final int APPLY_NEVER=0;
	private java.awt.Color errorColor = ColorHelper.getAlarm();
	private java.lang.Class fieldUserClass = null;
	private java.lang.Class fieldSuperClass = null;
	private boolean fieldAllowInterface = true;
	/**
	 *  
	 */
	public ClassEditor() {
		super();
		initialize();
	}
	/**
	 *  
	 */
	public ClassEditor(Class superClass) {
		this();
		setSuperClass(superClass);
	}

	/**
	 * Method generated to support the promotion of the errorColor attribute.
	 * 
	 * @param arg1
	 *            java.awt.Color
	 */
	private void paintField(boolean isError) {
		if (isError)
			setForeground(getErrorColor());
		else
			setForeground(Color.black);
	}

	/**
	 * Creation date: (9.1.2002 23:37:31)
	 */
	private boolean checkClass(Class c) {
		try {
			if (c == null)
				return true;
			if ((fieldSuperClass != null)
				&& (!(fieldSuperClass.isAssignableFrom(c)))) {
				return false;
			}
			if (!(fieldAllowInterface)
				&& ((c.isInterface())
					|| (Modifier.isAbstract(c.getModifiers())))) {
				return false;
			}
		} catch (Exception e) {
			setUserClass(null);
			paintField(true);
		}
		return true;
	}

	private void checkText() {
		try {
			setUserClass(StringUtilities.classFromString(getText()));
		} catch (Exception e) {
			setUserClass(null);
			paintField(true);
		}
	}
	private void initialize() {
		addFocusListener(new FocusAdapter() {
			/**
			 * @see java.awt.event.FocusAdapter#focusLost(FocusEvent)
			 */
			@Override
            public void focusLost(FocusEvent e) {
				super.focusLost(e);
				if ((getApplyMode() & APPLY_ON_FOCUS_LOST)==APPLY_ON_FOCUS_LOST) {
					checkText();
				}
			}

		});
		addKeyListener(new KeyAdapter() {
			/**
			 * @see java.awt.event.KeyAdapter#keyTyped(KeyEvent)
			 */
			@Override
            public void keyPressed(KeyEvent e) {
				if ((e.getKeyCode() == KeyEvent.VK_ENTER)) {
					if ((getApplyMode() & APPLY_ON_ENTER)==APPLY_ON_ENTER) {
						checkText();
						return;
					}
				} else {
					if ((getApplyMode() & APPLY_ON_CHANGE)==APPLY_ON_CHANGE) {
						checkText();
						return;
					}
				}
				paintField(false);
			}
		});
		setApplyMode(APPLY_ON_ENTER | APPLY_ON_FOCUS_LOST);
	}
	
	/**
	 * Gets the allowAbstract.
	 * 
	 * @return Returns a boolean
	 */
	public boolean getAllowAbstract() {
		return fieldAllowInterface;
	}

	/**
	 * Sets the AllowInterface.
	 * 
	 * @param AllowInterface
	 *            The AllowInterface to set
	 */
	public void setAllowAbstract(boolean allowInterface) {
		fieldAllowInterface = allowInterface;
	}

	/**
	 * Gets the superClass.
	 * 
	 * @return Returns a java.lang.Class
	 */
	public java.lang.Class getSuperClass() {
		return fieldSuperClass;
	}

	/**
	 * Sets the superClass.
	 * 
	 * @param superClass
	 *            The superClass to set
	 */
	public void setSuperClass(java.lang.Class superClass) {
		fieldSuperClass = superClass;
		checkText();
	}

	/**
	 * Gets the UserClass.
	 * 
	 * @return Returns a java.lang.Class
	 */
	public java.lang.Class getUserClass() {
		return fieldUserClass;
	}

	/**
	 * Sets the userClass.
	 * 
	 * @param userClass
	 *            The userClass to set
	 */
	private void setUserClass(java.lang.Class userClass) {
		if (checkClass(userClass)) {
			Class oldValue = fieldUserClass;
			if (oldValue==null && userClass==null) return;
			if (userClass != null) {
				paintField(false);
			}
			if (oldValue==null || !((oldValue).equals(userClass))) {
				fieldUserClass = userClass;
				firePropertyChange("userClass", oldValue, fieldUserClass);
				firePropertyChange(PROPERTY_VALUE_NAME,oldValue, fieldUserClass);
			}
			if (userClass!=null) {
				internalSetText(userClass.getName(),false);
			} else {
				internalSetText("",false);
			}
		} else {
			if (userClass!=null) {
				internalSetText(userClass.getName(),false);
			}
			paintField(true);
			if (fieldUserClass != null)
				setUserClass(null);
		}
	}

	/**
	 * Gets the errorColor.
	 * 
	 * @return Returns a java.awt.Color
	 */
	public java.awt.Color getErrorColor() {
		return errorColor;
	}

	/**
	 * Sets the errorColor.
	 * 
	 * @param errorColor
	 *            The errorColor to set
	 */
	public void setErrorColor(java.awt.Color errorColor) {
		this.errorColor = errorColor;
	}

	/**
	 * @see JTextComponent#setText(String)
	 */
	@Override
    public void setText(String t) {
		internalSetText(t,true);
	}
	
	private void internalSetText(String t, boolean check) {
		super.setText(t);
		if (check) {
			checkText();
		}
	}

	/**
	 * @see ValueEditor#getValue()
	 */
	public Object getPropertyValue() {
		checkText();
		return getUserClass();
	}

	/**
	 * @see ValueEditor#setValue(Object)
	 */
	public boolean setPropertyValue(Object value) {
		if (value instanceof Class) {
			setUserClass((Class) value);
			return true;
		}
		return false;
	}

	/**
	 * Bugfix for bug 4243496 (versions 1.1.7 - 1.4)
	 * 
	 * @see JComponent#setMargin(Insets)
	 */
	@Override
    public void setMargin(Insets margin) {
		if (margin == null) {
			super.setMargin(new Insets(0, 2, 0, 2));
			return;
		}
		if (margin.left == 0) {
			super.setMargin(
				new Insets(margin.top, 2, margin.bottom, margin.right));
			return;
		}
		super.setMargin(margin);
	}

	/**
	 * @see com.cosylab.gui.property.editors.PropertyEditor#getDescription()
	 */
	public String getDescription() {
		return null;
	}

	/**
	 * @see com.cosylab.gui.property.editors.PropertyEditor#getEditor()
	 */
	public PropertyEditor getEditor() {
		return this;
	}

	/**
	 * @see com.cosylab.gui.property.editors.PropertyEditor#setDescription(String)
	 */
	public void setDescription(String description) {
	}
	
	public void setApplyMode(int mode) {
		putClientProperty(APPLY_MODE_PROPERTY,new Integer(mode));
	}
	
	public int getApplyMode(){
		return ((Integer)getClientProperty(APPLY_MODE_PROPERTY)).intValue();
	}
}
