/*
 * Copyright (c) 2003-2008 by Cosylab d. d.
 *
 * This file is part of CosyBeans.
 *
 * CosyBeans is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * CosyBeans is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with CosyBeans.  If not, see <http://www.gnu.org/licenses/>.
 */

package com.cosylab.gui.util;

import java.awt.Component;
import java.awt.datatransfer.DataFlavor;
import java.awt.datatransfer.Transferable;
import java.awt.datatransfer.UnsupportedFlavorException;
import java.beans.PropertyChangeEvent;
import java.beans.PropertyVetoException;
import java.io.IOException;

import javax.swing.JComponent;

import com.cosylab.gui.components.util.CosyTransferHandler;
import com.cosylab.gui.displayers.DisplayerParameters;
import com.cosylab.introspection.BeanIntrospector;

/**
 * <code>AbstractDisplayerParametersTransferHandler</code> is a TransferHandler
 * for all DisplayerParameters-type-displayer. 
 * 
 * @author tkusterle
 *
 */
public abstract class AbstractDisplayerParametersTransferHandler extends CosyTransferHandler {
	
	private static final long serialVersionUID = -4223198013311762602L;
	protected DisplayerParametersSelectorDialog dialog;
	protected JComponent displayer;
	protected DataFlavor[] flavors= {DisplayerParametersFlavor.FLAVOR,DataFlavor.stringFlavor};

	public AbstractDisplayerParametersTransferHandler(JComponent comp, boolean reciveEnabled, boolean exportEnabled) {
		super(reciveEnabled, exportEnabled);
		displayer= comp;
		hookTransferHandler(comp);
	}

	/**
	 * Creates new DnD transfer handler with modified flavors. 
	 * @param comp the component for which this handler si created
	 * @param reciveEnabled flag for recifving DnD gestures
	 * @param exportEnabled flag for exporting DnD gestures
	 * @param f the array of flavors, whih are inseted to this handler at beginning of default 
	 *          flavors. If null or 0 length then ignored.
	 * @param resetFlavors if <code>true</code>, then provided flavors replaces default 
	 *                     flavors, otherwise provided flavors are inserted ad beginning of 
	 *                     default flavor list.
	 */
	public AbstractDisplayerParametersTransferHandler(JComponent comp, boolean reciveEnabled, boolean exportEnabled, DataFlavor[] f, boolean resetFlavors) {
		this(comp, reciveEnabled, exportEnabled);
		if (f==null || f.length==0) {
			return;
		}
		if (resetFlavors) {
			flavors= new DataFlavor[f.length];
			System.arraycopy(f, 0, flavors, 0, f.length);
		} else {
			DataFlavor[] ff= new DataFlavor[f.length+flavors.length];
			System.arraycopy(f, 0, ff, 0, f.length);
			System.arraycopy(flavors, 0, ff, f.length, flavors.length);
		}
	}

	/* (non-Javadoc)
	 * @see com.cosylab.gui.components.util.CosyTransferHandler#getTransferData(java.awt.datatransfer.DataFlavor)
	 */
	@Override
	public Object getTransferData(DataFlavor flavor)
			throws UnsupportedFlavorException, IOException {
		if (DataFlavor.stringFlavor.equals(flavor)) {
			DisplayerParameters p= exportDisplayerParameters();
			if (p!=null) {
				return p.toString();
			}
		}
		if (flavor instanceof DisplayerParametersFlavor) {
			DisplayerParameters p= exportDisplayerParameters();
			return p;
		}		
		throw new UnsupportedFlavorException(flavor);
	}

	abstract protected DisplayerParameters exportDisplayerParameters();

	/* (non-Javadoc)
	 * @see com.cosylab.gui.components.util.CosyTransferHandler#getTransferDataFlavors()
	 */
	@Override
	public DataFlavor[] getTransferDataFlavors() {
		return flavors;
	}

	/* (non-Javadoc)
	 * @see javax.swing.TransferHandler#importData(javax.swing.JComponent, java.awt.datatransfer.Transferable)
	 */
	@Override
	public boolean importData(JComponent comp, Transferable t) {
		try {
			Object drop=null;
			if (t.isDataFlavorSupported(DisplayerParametersFlavor.FLAVOR)) {
				drop = t.getTransferData(DisplayerParametersFlavor.FLAVOR);
				DisplayerParameters[] dp = null;
				dp = castParameters(drop);
				if (dp!=null) {
					return importDisplayerParameters(dp);
				}
			}
			if (t.isDataFlavorSupported(DataFlavor.stringFlavor)) {
				drop = t.getTransferData(DataFlavor.stringFlavor);
				if (drop!=null) {
					return importString(drop.toString());
				}
			}			
		} catch (Exception e) {
			e.printStackTrace();
		}
		
		return false;
	}
	
	protected boolean importString(String string) throws PropertyVetoException {
		return false;
	}

	protected DisplayerParameters[] castParameters(Object... parameters) {
		if (parameters instanceof DisplayerParameters[]) {
			return (DisplayerParameters[])parameters.getClass().cast(parameters);
		} else if (parameters.length == 1) {
			if (parameters[0] instanceof DisplayerParameters[]) {
				return (DisplayerParameters[]) parameters[0];
			} else if (parameters[0] instanceof DisplayerParameters) {
				return new DisplayerParameters[] {(DisplayerParameters)parameters[0]};
			}
		}
		return null;
	}
	
	/**
	 * @param dp
	 * @throws PropertyVetoException 
	 */
	protected boolean importDisplayerParameters(DisplayerParameters[] dp) throws PropertyVetoException {
		int i= selectParameters(displayer,dp);
		if (i<0) {
			return false;
		}
		return setParametersToDisplayer(displayer,dp[i]);
	}

	protected DisplayerParametersSelectorDialog getParameterSelectionDialog() {
		if (dialog == null) {
			dialog = new DisplayerParametersSelectorDialog(displayer, "Adding Source", "Select Source:");
		}
		return dialog;
	}
	
	protected int selectParameters(Component c,DisplayerParameters[] dp) {
		return getParameterSelectionDialog().showSelectionDialog(c,dp);
	}

	protected int selectParameters(Component c,String[] dp) {
		return getParameterSelectionDialog().showSelectionDialog(c,dp);
	}
	
	/**
	 * Sets the supplied parameters to the displayer.
	 * 
	 * @param disp destination displayer
	 * @param parameters parameters to be set
	 * @return true if successful
	 * 
	 * @throws PropertyVetoException when setting of parameters failed
	 */
	public boolean setParametersToDisplayer(Object disp, DisplayerParameters parameters) throws PropertyVetoException {
		String[] n= parameters.getNames();
		Object[] o= parameters.getValues();
		
		for (int i = 0; i < n.length; i++) {
			if (o[i]!=null && (!(o[i] instanceof Double) || !((Double)o[i]).isNaN())) {
				try {
					BeanIntrospector.setPropertyValue(disp,n[i],o[i]);
				} catch(NoSuchMethodException e) {
					//do nothing
				} catch (Exception e) {
					throw new PropertyVetoException("Property '"+n[i]+"' not set: "+e.getMessage(),new PropertyChangeEvent(disp,n[i],null,o[i]));
				}
			}
		}
		return true;
	}
	
	/**
	 * Returns the displayer.
	 * @return Returns the displayer.
	 */
	public JComponent getComponent() {
		return displayer;
	}
	
	/*
	 * (non-Javadoc)
	 * @see com.cosylab.gui.components.util.CosyTransferHandler#canImport(javax.swing.JComponent, java.awt.datatransfer.DataFlavor[])
	 */
	@Override
	public boolean canImport(JComponent comp, DataFlavor[] transferFlavors) {
		return super.canImport(comp, transferFlavors) & !isDragging;
	}
}
