/*
 * Copyright (c) 2003-2008 by Cosylab d. d.
 *
 * This file is part of Java-Common.
 *
 * Java-Common is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Java-Common is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Java-Common.  If not, see <http://www.gnu.org/licenses/>.
 */

package com.cosylab.logging;

import java.io.ByteArrayInputStream;

import java.util.logging.Formatter;
import java.util.logging.Handler;
import java.util.logging.Level;
import java.util.logging.LogManager;
import java.util.logging.LogRecord;
import java.util.logging.Logger;


/**
 * Implementation of Java Logging API handler.
 *
 * @author <a href="mailto:matej.sekoranjaATcosylab.com">Matej Sekoranja</a>
 * @version $id$
 */
public class ConsoleLogHandler extends Handler
{
	/** Logging formatter. */
	private Formatter formatter;

	/**
	 * Default constructor.
	 */
	public ConsoleLogHandler()
	{
		this(new ConsoleLogFormatter());
	}

	/**
	 * Construct handler with using giver formatter.
	 *
	 * @param formatter console log formatter, non-<code>null</code>.
	 */
	public ConsoleLogHandler(Formatter formatter)
	{
		assert (formatter != null);
		this.formatter = formatter;
	}

	/**
	 * @see java.util.logging.Handler#close()
	 */
	@Override
    public void close() throws SecurityException
	{
		// noop
	}

	/**
	 * @see java.util.logging.Handler#flush()
	 */
	@Override
    public void flush()
	{
		System.out.flush();
	}

	/**
	 * Prints the log record to the console using the current formatter, if the
	 * log record is loggable.
	 *
	 * @param record the log record to publish
	 *
	 * @see java.util.logging.Handler#publish(java.util.logging.LogRecord)
	 */
	@Override
    public void publish(LogRecord record)
	{
		if (isLoggable(record)) {
			System.out.print(formatter.format(record));
		}
	}

	/**
	 * Overrides default JVM logging configuration. This is convenience method
	 * - there is no need for additional configuration file and system
	 * property.
	 */
	public static void initializeConsoleLogging()
	{
		try {
			/*ClassLoader.getSystemClassLoader().loadClass(ConsoleLogHandler.class
			    .getName());*/
			
			// to make sure class loaded, because of problems with JUnit class loader
			/*ConsoleLogHandler clh= new ConsoleLogHandler();
			clh.close();*/

			ByteArrayInputStream bais = new ByteArrayInputStream(("handlers="
				    + ConsoleLogHandler.class.getName()).getBytes());
			LogManager.getLogManager().readConfiguration(bais);
		} catch (Throwable th) {
			th.printStackTrace();
		}
	}

	/**
	 * Simple test.
	 *
	 * @param args
	 */
	public static void main(String[] args)
	{
		initializeConsoleLogging();

		Logger logger = Logger.getLogger("loggerName");
		logger.info("Testing: 1, 2, 3...");
		logger.info("... 4, 5, 6.");
		logger.setLevel(Level.ALL);
		logger.finest("... finest");
	}
}

/* __oOo__ */
