/*
 * Copyright (c) 2003-2008 by Cosylab d. d.
 *
 * This file is part of Java-Common.
 *
 * Java-Common is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Java-Common is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Java-Common.  If not, see <http://www.gnu.org/licenses/>.
 */

package com.cosylab.logging;

import java.util.logging.Handler;
import java.util.logging.Level;
import java.util.logging.LogManager;
import java.util.logging.LogRecord;
import java.util.logging.Logger;


/**
 * <code>NullLogger</code> extends Java <code>Logger</code> and overrides all
 * logging methods to do nothing. Using this class statically switches all
 * logging capabilities off, thus this method my be not callled at all by JVM
 * optimization. Similar as sending logs to the /dev/null device.
 * 
 * <p>
 * <b>Recomended usage</b> is to add final Logger field to a class and use
 * <code>NullLogger</code> instance when no debug information is necessary.
 * </p>
 *
 * @author <a href="mailto:igor.kriznar@cosylab.com">Igor Kriznar</a>
 * @version $Id: NullLogger.java,v 1.4 2008-04-22 12:26:29 jbobnar Exp $
 *
 * @since Jun 27, 2004.
 */
public final class NullLogger extends Logger
{
	/**
	 * Returns NullLogger registered at LogManager. IF Log manager does not
	 * have NullLogger, new one is created, registered at LogManager and
	 * returned.
	 *
	 * @return NullLogger instance registered at LogManager
	 */
	public synchronized static final NullLogger getNullLogger()
	{
		NullLogger r = (NullLogger)LogManager.getLogManager().getLogger(NullLogger.class
			    .getName());

		if (r == null) {
			r = new NullLogger();
			LogManager.getLogManager().addLogger(r);
		}

		return r;
	}

	private NullLogger()
	{
		super("NullLogger", null);

		//super.setUseParentHandlers(false);
	}

	/* (non-Javadoc)
	 * @see java.util.logging.Logger#addHandler(java.util.logging.Handler)
	 */
	@Override
    public synchronized void addHandler(Handler handler)
		throws SecurityException
	{
		// hopefully removed by JIT
	}

	/* (non-Javadoc)
	 * @see java.util.logging.Logger#config(java.lang.String)
	 */
	@Override
    public void config(String msg)
	{
		// hopefully removed by JIT
	}

	/* (non-Javadoc)
	 * @see java.util.logging.Logger#entering(java.lang.String, java.lang.String)
	 */
	@Override
    public void entering(String sourceClass, String sourceMethod)
	{
		// hopefully removed by JIT
	}

	/* (non-Javadoc)
	 * @see java.util.logging.Logger#entering(java.lang.String, java.lang.String, java.lang.Object)
	 */
	@Override
    public void entering(String sourceClass, String sourceMethod, Object param1)
	{
		// hopefully removed by JIT
	}

	/* (non-Javadoc)
	 * @see java.util.logging.Logger#entering(java.lang.String, java.lang.String, java.lang.Object[])
	 */
	@Override
    public void entering(String sourceClass, String sourceMethod,
	    Object[] params)
	{
		// hopefully removed by JIT
	}

	/* (non-Javadoc)
	 * @see java.util.logging.Logger#exiting(java.lang.String, java.lang.String)
	 */
	@Override
    public void exiting(String sourceClass, String sourceMethod)
	{
		// hopefully removed by JIT
	}

	/* (non-Javadoc)
	 * @see java.util.logging.Logger#fine(java.lang.String)
	 */
	@Override
    public void fine(String msg)
	{
		// hopefully removed by JIT
	}

	/* (non-Javadoc)
	 * @see java.util.logging.Logger#exiting(java.lang.String, java.lang.String, java.lang.Object)
	 */
	@Override
    public void exiting(String sourceClass, String sourceMethod, Object result)
	{
		// hopefully removed by JIT
	}

	/* (non-Javadoc)
	 * @see java.util.logging.Logger#finer(java.lang.String)
	 */
	@Override
    public void finer(String msg)
	{
		// hopefully removed by JIT
	}

	/* (non-Javadoc)
	 * @see java.util.logging.Logger#finest(java.lang.String)
	 */
	@Override
    public void finest(String msg)
	{
		// hopefully removed by JIT
	}

	/* (non-Javadoc)
	 * @see java.util.logging.Logger#info(java.lang.String)
	 */
	@Override
    public void info(String msg)
	{
		// hopefully removed by JIT
	}

	/* (non-Javadoc)
	 * @see java.util.logging.Logger#isLoggable(java.util.logging.Level)
	 */
	@Override
    public boolean isLoggable(Level level)
	{
		return false;
	}

	/* (non-Javadoc)
	 * @see java.util.logging.Logger#log(java.util.logging.Level, java.lang.String)
	 */
	@Override
    public void log(Level level, String msg)
	{
		// hopefully removed by JIT
	}

	/* (non-Javadoc)
	 * @see java.util.logging.Logger#log(java.util.logging.Level, java.lang.String, java.lang.Object)
	 */
	@Override
    public void log(Level level, String msg, Object param1)
	{
		// hopefully removed by JIT
	}

	/* (non-Javadoc)
	 * @see java.util.logging.Logger#log(java.util.logging.Level, java.lang.String, java.lang.Object[])
	 */
	@Override
    public void log(Level level, String msg, Object[] params)
	{
		// hopefully removed by JIT
	}

	/* (non-Javadoc)
	 * @see java.util.logging.Logger#log(java.util.logging.Level, java.lang.String, java.lang.Throwable)
	 */
	@Override
    public void log(Level level, String msg, Throwable thrown)
	{
		// hopefully removed by JIT
	}

	/* (non-Javadoc)
	 * @see java.util.logging.Logger#log(java.util.logging.LogRecord)
	 */
	@Override
    public void log(LogRecord record)
	{
		// hopefully removed by JIT
	}

	/* (non-Javadoc)
	 * @see java.util.logging.Logger#logp(java.util.logging.Level, java.lang.String, java.lang.String, java.lang.String)
	 */
	@Override
    public void logp(Level level, String sourceClass, String sourceMethod,
	    String msg)
	{
		// hopefully removed by JIT
	}

	/* (non-Javadoc)
	 * @see java.util.logging.Logger#logp(java.util.logging.Level, java.lang.String, java.lang.String, java.lang.String, java.lang.Object)
	 */
	@Override
    public void logp(Level level, String sourceClass, String sourceMethod,
	    String msg, Object param1)
	{
		// hopefully removed by JIT
	}

	/* (non-Javadoc)
	 * @see java.util.logging.Logger#logp(java.util.logging.Level, java.lang.String, java.lang.String, java.lang.String, java.lang.Object[])
	 */
	@Override
    public void logp(Level level, String sourceClass, String sourceMethod,
	    String msg, Object[] params)
	{
		// hopefully removed by JIT
	}

	/* (non-Javadoc)
	 * @see java.util.logging.Logger#logp(java.util.logging.Level, java.lang.String, java.lang.String, java.lang.String, java.lang.Throwable)
	 */
	@Override
    public void logp(Level level, String sourceClass, String sourceMethod,
	    String msg, Throwable thrown)
	{
		// hopefully removed by JIT
	}

	/* (non-Javadoc)
	 * @see java.util.logging.Logger#logrb(java.util.logging.Level, java.lang.String, java.lang.String, java.lang.String, java.lang.String)
	 */
	@Override
    public void logrb(Level level, String sourceClass, String sourceMethod,
	    String bundleName, String msg)
	{
		// hopefully removed by JIT
	}

	/* (non-Javadoc)
	 * @see java.util.logging.Logger#logrb(java.util.logging.Level, java.lang.String, java.lang.String, java.lang.String, java.lang.String, java.lang.Object)
	 */
	@Override
    public void logrb(Level level, String sourceClass, String sourceMethod,
	    String bundleName, String msg, Object param1)
	{
		// hopefully removed by JIT
	}

	/* (non-Javadoc)
	 * @see java.util.logging.Logger#logrb(java.util.logging.Level, java.lang.String, java.lang.String, java.lang.String, java.lang.String, java.lang.Object[])
	 */
	@Override
    public void logrb(Level level, String sourceClass, String sourceMethod,
	    String bundleName, String msg, Object[] params)
	{
		// hopefully removed by JIT
	}

	/* (non-Javadoc)
	 * @see java.util.logging.Logger#logrb(java.util.logging.Level, java.lang.String, java.lang.String, java.lang.String, java.lang.String, java.lang.Throwable)
	 */
	@Override
    public void logrb(Level level, String sourceClass, String sourceMethod,
	    String bundleName, String msg, Throwable thrown)
	{
		// hopefully removed by JIT
	}

	/* (non-Javadoc)
	 * @see java.util.logging.Logger#setLevel(java.util.logging.Level)
	 */
	@Override
    public void setLevel(Level newLevel) throws SecurityException
	{
		// hopefully removed by JIT
	}

	/* (non-Javadoc)
	 * @see java.util.logging.Logger#getLevel()
	 */
	@Override
    public Level getLevel()
	{
		return Level.OFF;
	}

	/* (non-Javadoc)
	 * @see java.util.logging.Logger#severe(java.lang.String)
	 */
	@Override
    public void severe(String msg)
	{
		// hopefully removed by JIT
	}

	/* (non-Javadoc)
	 * @see java.util.logging.Logger#throwing(java.lang.String, java.lang.String, java.lang.Throwable)
	 */
	@Override
    public void throwing(String sourceClass, String sourceMethod,
	    Throwable thrown)
	{
		// hopefully removed by JIT
	}

	/* (non-Javadoc)
	 * @see java.util.logging.Logger#getUseParentHandlers()
	 */
	@Override
    public synchronized boolean getUseParentHandlers()
	{
		return false;
	}

	/* (non-Javadoc)
	 * @see java.util.logging.Logger#setUseParentHandlers(boolean)
	 */
	@Override
    public synchronized void setUseParentHandlers(boolean useParentHandlers)
	{
		// hopefully removed by JIT
	}

	/* (non-Javadoc)
	 * @see java.util.logging.Logger#warning(java.lang.String)
	 */
	@Override
    public void warning(String msg)
	{
		// hopefully removed by JIT
	}
}

/* __oOo__ */
