/*
 * Copyright (c) 2003-2008 by Cosylab d. d.
 *
 * This file is part of Java-Common.
 *
 * Java-Common is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Java-Common is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Java-Common.  If not, see <http://www.gnu.org/licenses/>.
 */

package com.cosylab.util;

import java.awt.Image;
import java.beans.BeanDescriptor;
import java.beans.BeanInfo;
import java.beans.EventSetDescriptor;
import java.beans.IntrospectionException;
import java.beans.Introspector;
import java.beans.MethodDescriptor;
import java.beans.PropertyDescriptor;
import java.beans.SimpleBeanInfo;
import java.lang.reflect.Method;
import java.util.HashMap;
import java.util.Map;

/**
 * @author <a href="mailto:jernej.kamenik@cosylab.com">Jernej Kamenik</a>
 * @version $id$
 *
 */
public class BeanInfoSupport extends SimpleBeanInfo implements BeanInfo {

	private static HashMap delegates = new HashMap();

	public static BeanInfo getBeanInfo(Class beanType) {
		BeanInfo retVal = null;
		if (delegates.containsKey(beanType)) {
			retVal = (BeanInfo)delegates.get(beanType);
		} else {
			try {
				retVal = Introspector.getBeanInfo(beanType,Introspector.IGNORE_IMMEDIATE_BEANINFO);
				delegates.put(beanType,retVal);
			} catch (IntrospectionException e) {
				e.printStackTrace();
			}
		}	
		return retVal;
	}

	public static BeanDescriptor getBeanDescriptor(Class beanType) {
		BeanInfo delegate = getBeanInfo(beanType);
		if (delegate==null) return null;
		return delegate.getBeanDescriptor();
	}

	public static EventSetDescriptor[] getEventSetDescriptors(Class beanType) {
		BeanInfo delegate = getBeanInfo(beanType);
		if (delegate==null) return null;
		return delegate.getEventSetDescriptors();
	}

	/**
	 * @see BeanInfo#getDefaultEventIndex()
	 */
	public static int getDefaultEventIndex(Class beanType) {
		BeanInfo delegate = getBeanInfo(beanType);
		if (delegate==null) return -1;
		return delegate.getDefaultEventIndex();
	}

	/**
	 * @see BeanInfo#getPropertyDescriptors()
	 */
	public static PropertyDescriptor[] getPropertyDescriptors(Class beanType) {
		BeanInfo delegate = getBeanInfo(beanType);
		if (delegate==null) return null;
		return delegate.getPropertyDescriptors();
	}

	/**
	 * @see BeanInfo#getDefaultPropertyIndex()
	 */
	public static int getDefaultPropertyIndex(Class beanType) {
		BeanInfo delegate = getBeanInfo(beanType);
		if (delegate==null) return -1;
		return delegate.getDefaultPropertyIndex();
	}

	/**
	 * @see BeanInfo#getMethodDescriptors()
	 */
	public static MethodDescriptor[] getMethodDescriptors(Class beanType) {
		BeanInfo delegate = getBeanInfo(beanType);
		if (delegate==null) return null;
		return delegate.getMethodDescriptors();
	}

	/**
	 * @see BeanInfo#getAdditionalBeanInfo()
	 */
	public static BeanInfo[] getAdditionalBeanInfo(Class beanType) {
		BeanInfo delegate = getBeanInfo(beanType);
		if (delegate==null) return null;
		return delegate.getAdditionalBeanInfo();
	}

	/**
	 * @see BeanInfo#getIcon(int)
	 */
	public static Image getIcon(int arg0, Class beanType) {
		BeanInfo delegate = getBeanInfo(beanType);
		if (delegate==null) return null;
		return delegate.getIcon(arg0);
	}
	
	private Class beanClass;
	private Map<String,PropertyDescriptor> propertyDescriptors = new HashMap<>();
	private BeanDescriptor beanDescriptor;
	
	/**
	 * Creates new instance of BeanInfoSupport.
	 */
	protected BeanInfoSupport() {
		beanClass= getBeanClass();
	}
	/**
	 * Override this method, if default implementation is not suitable for you. Default implementation
	 * calculates bean class from bean info class name by stripping BeanInfo part.
	 * @return calculates bean class from bean info class name by stripping BeanInfo part
	 */
	protected Class getBeanClass() {
		String s= this.getClass().getName();
		s= s.substring(0,s.length()-"BeanInfo".length());
		try {
			return Class.forName(s);
		} catch (ClassNotFoundException e) {
			e.printStackTrace();
		}
		return null;
	}
	
	protected void setPropertyEditor(String property, Class editor) throws IntrospectionException {
		PropertyDescriptor p = getPropertyDescriptor(property);
		p.setPropertyEditorClass(editor);
	}
	
	protected PropertyDescriptor getPropertyDescriptor(String name) throws IntrospectionException {
		PropertyDescriptor p = propertyDescriptors.get(name);
		if (p==null) {
			p= new PropertyDescriptor(name,beanClass);
			propertyDescriptors.put(name,p);
		}
		return p;
	}

	protected PropertyDescriptor removePropertyDescriptor(String name) {
		return propertyDescriptors.remove(name);
	}
	
	protected void loadDefaultPropertyDescriptors() {
		PropertyDescriptor[] pd= getPropertyDescriptors(beanClass);
		
		for (PropertyDescriptor p : pd) {
			if (p.getWriteMethod() == null)
				continue;
			if (p.getReadMethod() == null)
				continue;
			if (!propertyDescriptors.containsKey(p.getName())) propertyDescriptors.put(p.getName(),p);
		}
		
	}
	
	protected void loadPropertyDescriptors(String... names) throws IntrospectionException {
		for (String s : names) {
			if (!propertyDescriptors.containsKey(s)) propertyDescriptors.put(s,new PropertyDescriptor(s,beanClass));
		}
	}

	protected void addPropertyDescriptor(PropertyDescriptor... descriptors) {
		for (PropertyDescriptor p : descriptors) {
			propertyDescriptors.put(p.getName(),p);
		}
	}

	protected void removePropertyDescriptors(String... names) {
		for (String s : names) {
			propertyDescriptors.remove(s);
		}
	}
	
	protected void loadPropertyDescriptor(String name, Class editor) throws IntrospectionException {
		PropertyDescriptor p= propertyDescriptors.get(name);
		if (p == null) {
			p= new PropertyDescriptor(name,beanClass);
		}
		p.setPropertyEditorClass(editor);
		if (!propertyDescriptors.containsKey(name)) propertyDescriptors.put(name,p);
	}
	
	protected void loadPropertyDescriptor(String name, Class editor, Method readMethod, Method writeMethod) throws IntrospectionException {
		PropertyDescriptor p= propertyDescriptors.get(name);
		if (p == null) {
			p= new PropertyDescriptor(name,beanClass, readMethod.getName(), writeMethod.getName());
		}
		p.setPropertyEditorClass(editor);
		p.setReadMethod(readMethod);
		p.setWriteMethod(writeMethod);
		if (!propertyDescriptors.containsKey(name)) propertyDescriptors.put(name,p);
	}
	
	protected void overridePropertyDescriptor(String name, Class editor, String readMethod, String writeMethod) throws IntrospectionException {
		PropertyDescriptor p = new PropertyDescriptor(name,beanClass, readMethod, writeMethod);
		p.setPropertyEditorClass(editor);
		propertyDescriptors.put(name,p);
	}
	
	protected void clearPropertyDescriptors() {
		propertyDescriptors.clear();
	}
	
	/**
	 * Sets new beanDescriptor.
	 */
	protected void setBeanDescriptor(Class customizer) {
		this.beanDescriptor = new BeanDescriptor(beanClass,customizer);
		this.beanDescriptor.setValue("EXPLICIT_PROPERTY_CHANGE", Boolean.TRUE);
	}
	
	/* (non-Javadoc)
	 * @see java.beans.SimpleBeanInfo#getPropertyDescriptors()
	 */
	@Override
	public PropertyDescriptor[] getPropertyDescriptors() {
		if (propertyDescriptors.size()>0) {
			return propertyDescriptors.values().toArray(new PropertyDescriptor[propertyDescriptors.size()]);
		}
		return super.getPropertyDescriptors();
	}
	
	/* (non-Javadoc)
	 * @see java.beans.SimpleBeanInfo#getBeanDescriptor()
	 */
	@Override
	public BeanDescriptor getBeanDescriptor() {
		if (beanDescriptor!=null) {
			return beanDescriptor;
		}
		return super.getBeanDescriptor();
	}

}
