/*
 * Copyright (c) 2003-2008 by Cosylab d. d.
 *
 * This file is part of Java-Common.
 *
 * Java-Common is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Java-Common is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Java-Common.  If not, see <http://www.gnu.org/licenses/>.
 */

package com.cosylab.util;

import java.io.IOException;
import java.io.InputStream;
import java.net.MalformedURLException;
import java.net.URL;
import java.util.Enumeration;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;
import java.util.Properties;


/**
 * DOCUMENT ME!
 *
 * @version $id$
 */
public class ExtendedProperties extends Properties
{
    /**
     * 
     */
    private static final long serialVersionUID = 6145123486153794332L;

    /**
     * Suffix used to speficy a linked properties file. The key
     * up to the string is used as the root for all the properties found in
     * the referenced file.
     * <p>e.g.: <code>icon.@link = gui/icons.properties</code> will look
     * for a file "gui/icons.properties" and a property named <code>text.edit</code>
     * from this file will be referenced in the parent properties as "icon.text.edit".
     * <p>
     * If multiple links are assigned to the same prefix, they should be numbered
     * so that their property keys remain unique, e.g.:<br/>
     * <code>icon.@link1 = gui/myIcons.properties</code>
     * <code>icon.@link2 = gui/otherIcons.properties</code>
     * </p>
     * 
     */
    public static final String LINK_SUFFIX="@link";
    
	/**
	 * The delimiter that is used to separate each part in the property name
	 * that makes the hierarchy. For example the property name 'windowX' under
	 * the property 'Panel1' becomes 'Panel1.windowX' and the dot between
	 * those names is the ExtendedProperties.DELIMITER
	 */
	public static final String DELIMITER = ".";

	/**
	 * Creates a new ExtendedProperties object.
	 */
	public ExtendedProperties()
	{
		super();
	}

	/**
	 * Creates a new ExtendedProperties object.
	 *
	 * @param defaults DOCUMENT ME!
	 */
	public ExtendedProperties(Properties defaults)
	{
		super(defaults);
	}

	/**
	 * Convenience method that tries to return int value of property.
	 *
	 * @param key Name of key to query value for
	 *
	 * @return int type value of property
	 *
	 * @throws NumberFormatException
	 *
	 * @see #getProperty(String)
	 */
	public int getIntProperty(String key) throws NumberFormatException
	{
		String value = getProperty(key);

		if (value == null) {
			throw new NumberFormatException("Cannot create int from null.");
		}

		return Integer.parseInt(value);
	}

	/**
	 * Convenience method that tries to return double value of property.
	 *
	 * @param key Name of key to query value for
	 *
	 * @return double type value of property
	 *
	 * @throws NumberFormatException
	 *
	 * @see #getProperty(String)
	 */
	public double getDoubleProperty(String key) throws NumberFormatException
	{
		String value = getProperty(key);

		if (value == null) {
			throw new NumberFormatException("Cannot create double from null.");
		}

		return Double.parseDouble(value);
	}

	/**
	 * Convenience method that tries to return <code>Class</code> instace
	 * specified by fully qualified string.
	 *
	 * @param key Name of key associated with value.
	 *
	 * @return Instance of <code>Class</code> or null if not found.
	 *
	 * @see #getProperty(String)
	 */
	public Class getClassProperty(String key)
	{
		String value = getProperty(key);

		if (value == null) {
			return null;
		}

		Class c;

		try {
			c = Class.forName(value);
		} catch (ClassNotFoundException e) {
			return null;
		}

		return c;
	}

	/**
	 * Returns an instance of the ExtendedProperties object which have only
	 * properties under the hierarchy given by <code>mementoName</code>. For
	 * example if this object has 'window.x', 'window.y' and  'something else'
	 * properties then invoking getProperties("window") will return
	 * ExtendedProperties with anly 'x' and 'y' properties.
	 *
	 * @param mementoName name to extract from this properties
	 *
	 * @return an ExtendedProperties object with properties matching
	 *         mementoName
	 */
	public ExtendedProperties getProperties(String mementoName)
	{
		assert (mementoName != null);

		// calculate susbstring length that we will skip on ech key 
		int stripLen = mementoName.length() + DELIMITER.length();
		ExtendedProperties ep = new ExtendedProperties();

		for (Enumeration e = keys(); e.hasMoreElements();) {
			String key = (String)e.nextElement();

			if (!key.startsWith(mementoName)) {
				continue;
			}

			ep.put(key.substring(stripLen), get(key));
		}

		return ep;
	}

	/**
	 * Inserts the content of the given ExtendedProperties to this properties
	 * by prefixing each key with given <code>mementoName</code>.
	 *
	 * @param mementoName
	 * @param ep
	 */
	public void setProperties(String mementoName, ExtendedProperties ep)
	{
		assert (ep != null);
        
        String prefix=mementoName+DELIMITER;
        if (mementoName==null || mementoName.length()<1) {
            prefix="";
        }

		for (Enumeration e = ep.keys(); e.hasMoreElements();) {
			String key = (String)e.nextElement();
			String newKey = prefix +  key;
			put(newKey, ep.get(key));
		}
	}
    
    /* (non-Javadoc)
     * @see java.util.Properties#load(java.io.InputStream)
     */
    public synchronized void load(URL resource) throws IOException, MalformedURLException {
        try (InputStream stream =resource.openStream()) {
            super.load(stream);
            checkForLinks(resource);
            stream.close();
        }
    }
    
    /**
     * @deprecated Use @link #load(URL) to process links properly
     */
    @Override
    @Deprecated
    public synchronized void load(InputStream inStream) throws IOException {
        super.load(inStream);
    }
    
    public synchronized void checkForLinks(URL resource) throws IOException, MalformedURLException{
            Map linksToAdd=null;
            Iterator iter=entrySet().iterator();
            while (iter.hasNext()) {
                Map.Entry entry = (Map.Entry) iter.next();
                String key=(String)entry.getKey();
                int lastDelIdx=key.lastIndexOf(ExtendedProperties.DELIMITER);
                String lastPart=key;
                if (lastDelIdx>=0) {
                    lastPart=key.substring(lastDelIdx+1);
                }
                if (lastPart.startsWith(LINK_SUFFIX)) {
                    String prefix="";
                    if (lastDelIdx>0) {
                        prefix=key.substring(0, lastDelIdx);
                    }
                
                    String linkPath=(String)entry.getValue();
                    URL childURL=new URL(resource,linkPath);
                    ExtendedProperties childProps=new ExtendedProperties();
                    childProps.load(childURL);
                    if (linksToAdd==null) {
                        linksToAdd=new HashMap();
                    }
                    linksToAdd.put(prefix,childProps);
                }
            }
            if (linksToAdd!=null) {
                iter=linksToAdd.keySet().iterator();
                while(iter.hasNext()) {
                    String prefix=(String)iter.next();
                    ExtendedProperties childProps=(ExtendedProperties)linksToAdd.get(prefix);
                    setProperties(prefix, childProps);
                }
            }
    }
}

/* __oOo__ */
