/*
 * Copyright (c) 2003-2008 by Cosylab d. d.
 *
 * This file is part of Java-Common.
 *
 * Java-Common is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Java-Common is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Java-Common.  If not, see <http://www.gnu.org/licenses/>.
 */

package com.cosylab.util;

import java.util.LinkedHashMap;
import java.util.Map;


/**
 * Helper class that caches <code>PrintfFormat</code> classes. Used by table
 * renderers, which need to switch between small number of different  formats.
 * 
 * @author <a href="mailto:ales.pucelj@cosylab.com">Ales Pucelj</a>  
 * @version $id$
 */
public final class FormatCache
{
	/** Default format string */
	private static final String DEFAULT_FORMAT = "%+4.2f";
	private static final int CACHE_LIMIT = 200;

	/** Static cache of formatters */
	private static final Map formatters = new LinkedHashMap();

	/**
	 * Returns instance of <code>PrintfFormat</code>. If formatter does not
	 * already exist in cache, new instance will be created. If format is
	 * invalid, null will be returned.
	 *
	 * @param format C-Style format string.
	 *
	 * @return Instance of formatter or null if format string is invalid.
	 */
	public static PrintfFormat getFormatter(String format)
	{
		Object o = formatters.get(format);

		if (o == null) {
			try {
				if (format == null) {
					o = getDefaultFormatter();
				} else {
					o = new PrintfFormat(format);
				}
			} catch (IllegalArgumentException e) {
				return null;
			}

			formatters.put(format, o);

			if (formatters.size() > CACHE_LIMIT) {
				flush();
			}
		}

		return (PrintfFormat)o;
	}

	/**
	 * Returns default formatter with format string %4.4f. This method will
	 * always return valid formatter.
	 *
	 * @return Instance of formatter with default format string %4.4f.
	 */
	public static PrintfFormat getDefaultFormatter()
	{
		return getFormatter(DEFAULT_FORMAT);
	}

	/**
	 * Clears cache. This method will be periodically called to prevent
	 * excessive cache growth.
	 */
	private static final void flush()
	{
		//formatters.clear();
		while(formatters.size()>CACHE_LIMIT) {
			formatters.remove(formatters.keySet().iterator().next());
		}
	}
}

/* __oOo__ */
