/*
 * Copyright (c) 2003-2008 by Cosylab d. d.
 *
 * This file is part of Java-Common.
 *
 * Java-Common is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Java-Common is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Java-Common.  If not, see <http://www.gnu.org/licenses/>.
 */

package com.cosylab.util;

/**
 * A listener list is an efficient data structure for holding listeners. Its
 * design is based on the fact that listeners are extensively accessed for
 * event dispatching, but the collection of listeners changes rarely. Under
 * these circumstances (especilly rare invocations of
 * <code>addListener()</code> and <code>removeListener</code> methods) this
 * class is a good choice. It holds listeners in an array that is trimmed
 * every time the listener membership changes (which is quite inefficient). On
 * the other hand, accessing the elements of the array is quick, <b>on the
 * condition and contract that the client does not modify the array being
 * returned</b>. Also note that the array is of the correct run-time type
 * specified by the constructor of this class, and can therefore be cast. For
 * example, if you speficy <code>WindowListener</code> as the RTT of this
 * list, you can perform <code>WindowListener[] wla =
 * (WindowListener[])toArray()</code> without raising a
 * <code>ClassCastException</code>. For cases where the listener membership
 * changes often, but events are dispatched rarely, use <code>ArrayList</code>
 * instead. <b>This class is thread safe, there is no need for additional
 * synchronization.</b> See also swing based listener registration for an
 * approach that is more time consuming during dispatching, but can store
 * different classes of listeners in a single array.
 *
 * @author <a href="mailto:gasper.tkacik@cosylab.com">Gasper Tkacik</a>
 * @version $id$
 *
 * @see javax.swing.event.EventListenerList
 */
public class ListenerList extends ObjectList
{
	/**
     * 
     */
    private static final long serialVersionUID = -5785487166622972775L;

    /**
	 * Constructs a new instance of the listener list that will be used to hold
	 * elements of type <code>type</code>.  The specified class should be a
	 * subclass of <code>java.util.EventListener</code>.
	 *
	 * @param type a run-time type of the elements in this list
	 *
	 * @throws NullPointerException DOCUMENT ME!
	 * @throws IllegalArgumentException DOCUMENT ME!
	 */
	public ListenerList(Class type)
	{
		super(type);

		if (type == null) {
			throw new NullPointerException("type");
		}

		if (!java.util.EventListener.class.isAssignableFrom(type)) {
			throw new IllegalArgumentException("Type '" + type
			    + "' does not extend 'java.util.EventListener'.");
		}
	}
}

/* __oOo__ */
