/*
 * Copyright (c) 2003-2008 by Cosylab d. d.
 *
 * This file is part of Java-Common.
 *
 * Java-Common is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Java-Common is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Java-Common.  If not, see <http://www.gnu.org/licenses/>.
 */

package com.cosylab.util;

/**
 * An object that is able of being recycled and stored in an object pool.
 * This object must support:
 * <ul>
 * 	<li>
 * 		<b>Resurrection.</b> When no other objects reference this object, this object
 * 		may be garbage collected. Usually - if this is a small instance that is used
 * 		(created) a lot of times, garbage collection will occur. In the <code>finalize</code>
 * 		method, this object must call its <code>objectPool</code> and invoke method
 * 		<code>returnToPool</code> on it. In this way, the object pool will contain
 * 		a reference to <code>this</code>, making this object strongly reachable 
 * 		(see Java reference for explanation) after finalization and consequently 
 * 		making this object resurrected.
 * 	</li>
 * 	<li>
 * 		<b>State reset.</code> In order to make this object useful again, it must support
 * 		<code>clear</code> method. This method must force the object into the same state
 * 		that it achieved during construction. Since object pools are mainly used for data
 * 		holder objects, this would imply a reset of all data fields to their default values.
 * 	</li>
 * 	<li>
 * 		<b>Holding a reference to the pool.</b> This reference is set through 
 * 		<code>setObjectPool</code> when the object is instantiated by the pool automatically,
 * 		<b>OR</b> must be set by hand when the object is constructed manually.
 * 	</li>
 * </ul>
 * <p>
 * Note that object pools are mostly usable for simple data holder objects, which are 
 * created in great amounts and are also frequently released. This means that the receivers
 * of these objects simply inspect some values (or even do nothing), after which they
 * have no more use for the object and do not store it in some sort of data container structure.
 * </p>
 */
public interface PoolableObject
{
	
	/**
	 * Clears the contents (data fields) of this object by resetting them to the 
	 * uninitialized (after construction) values.
	 */
	public void clear();
	
	/**
	 * Sets the object pool to which this object belongs. During finalization of this 
	 * object, it will attempt to resurrect itself by returning to the object pool.
	 * 
	 * @param pool the pool to which this object will return
	 */
	public void setObjectPool(ObjectPool pool);
}
