package de.desy.acop.displayers;

import java.beans.PropertyVetoException;

import javax.swing.Action;

import com.cosylab.gui.CheckBoxController;
import com.cosylab.gui.InfoDialog;
import com.cosylab.gui.displayers.DataConsumer;
import com.cosylab.gui.displayers.DataState;
import com.cosylab.gui.displayers.DisplayerUtilities;
import com.cosylab.gui.displayers.DoubleSeqConsumer;
import com.cosylab.gui.util.UserSettingsProtection;
import com.cosylab.util.CommonException;

import de.desy.acop.displayers.tools.AcopDisplayer;
import de.desy.acop.displayers.tools.AcopDisplayerTransferHandler;
import de.desy.acop.displayers.tools.AcopInfoDialog;
import de.desy.acop.transport.ConnectionParameters;
import de.desy.acop.transport.adapters.AcopTransportDataSource;
import de.desy.acop.transport.adapters.AdapterFactory;
import de.desy.acop.transport.adapters.AdapterFactoryService;

/**
 * AcopCheckBox is a displayer for displaying and switching between two
 * double values.
 * <p>
 * User can set two different values on <i>selectedValue</i>
 * and <i>deselectedValue</i> properties. When remote connection point has
 * value same as selectedValue, CheckBox appear selected; when remote value
 * is same as deselectedValue, CheckBox appear not selected. If remote value
 * is some else, than CheckBox switches to UNDEFINED state with cross drawn
 * over the component.
 * </p>
 * <p>
 * When user changes selection, than proper value is also sent to remote
 * connection point. If new state is selected, selectedValue is sent; if new
 * state is deselected, deselectedValue is sent.
 * </p>
 * <p>
 * The connection point for this displayer should be set using the 
 * <i>connectionParameters</i> property, where the remote name of the 
 * {@link ConnectionParameters} points to the desired property.
 * </p>
 * 
 * @author <a href="mailto:blaz.hostnik@cosylab.com">Blaz Hostnik, Cosylab</a>
 */
public class AcopCheckBox extends CheckBoxController implements AcopDisplayer {

	private static final long serialVersionUID = 1647491427169502213L;

	private int arrayIndex=0;

	private ConnectionParameters connectionParameters;

	private InfoDialog dialog;

	/**
	 * Default constructor: with "Enabled" text in front of the check box.
	 */
	public AcopCheckBox() {
		this(DEFAULT_TEXT);
	}
	
	/**
	 * Constructor with parameters.
	 * 
	 * @param text in front of the check box
	 */
	public AcopCheckBox(String title) {
		super(title);
		new AcopDisplayerTransferHandler(this);
		UserSettingsProtection.setProtection(this,DisplayerUtilities.COMMON_NUMERIC_DISPLAYER_PROPERTIES,false);
	}
	
	@Override
	public int getArrayIndex() {
		return arrayIndex;
	}

	@Override
	public void setArrayIndex(int index) {
		if (arrayIndex == index) return;
		int oldValue = arrayIndex;
		arrayIndex = index;
		firePropertyChange("arrayIndex", oldValue, arrayIndex);
	}

	@Override
	public InfoDialog getInfoDialog() {
		if (dialog == null) {
			dialog = new AcopInfoDialog(this);
		}
		return dialog;
	}

	@Override
	public void updateValue(long timestamp, double[] value)
			throws CommonException {
	    updateValue(timestamp, DisplayerUtilities.extract(arrayIndex,value));	
	}

	@Override
	public DataConsumer getDataConsumer(Class type) {
		if (type == DoubleSeqConsumer.class){
			return this;
		} else return super.getDataConsumer(type);
	}
	
	@Override
	public ConnectionParameters getConnectionParameters() {
		return connectionParameters;
	}

	@Override
	public void setConnectionParameters(ConnectionParameters param)
			throws CommonException, PropertyVetoException {
		if (param!=null && connectionParameters != null) {
			if(param.equals(connectionParameters)) return;
		}
		updateDataState(new DataState(DataState.UNDEFINED));
			
		ConnectionParameters old = connectionParameters;
		
		AdapterFactory factory = AdapterFactoryService.getInstance().getAdapterFactory();
		if (getDataSource() != null)
			factory.releaseDataSource((AcopTransportDataSource) getDataSource());
		if (param == null) {
			setDataSource(null);
			setTitle("No Connection");
		} else {
			if (param.getPropertySize() == ConnectionParameters.AUTO_PROPERTY_SIZE) {
				param = param.deriveWithPropertySize(1);
			}
			AcopTransportDataSource ds = factory.createDataSource(param);
			setDataSource(ds);
		}
		
		this.connectionParameters = param;
		firePropertyChange(CONNECTION_PARAMETERS_PROPERTY,old, connectionParameters);
	}
	
	/**
	 * Enables/disables Properties item in the popup menu.
	 * @param enable
	 */
	public void setPropertiesPopupEnabled(boolean enable) {
		Action[] actions = getPopupManager().getActions();
		for (Action a : actions) {
			if (a != null && "Preferences...".equals(a.getValue(Action.NAME))) {
				a.setEnabled(enable);
				return;
			}
		}
	}
	
	/**
	 * Returns true if Properties item is enabled in the popup menu.
	 * @return
	 */
	public boolean isPropertiesPopupEnabled() {
		Action[] actions = getPopupManager().getActions();
		for (Action a : actions) {
			if (a != null && "Preferences...".equals(a.getValue(Action.NAME))) {
				return a.isEnabled();
			}
		}
		return false;
	}


}
