/*
 * Copyright (c) 2006 Stiftung Deutsches Elektronen-Synchroton,
 * Member of the Helmholtz Association, (DESY), HAMBURG, GERMANY.
 *
 * THIS SOFTWARE IS PROVIDED UNDER THIS LICENSE ON AN "../AS IS" BASIS.
 * WITHOUT WARRANTY OF ANY KIND, EXPRESSED OR IMPLIED, INCLUDING BUT NOT LIMITED
 * TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR PARTICULAR PURPOSE AND
 * NON-INFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE
 * FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
 * TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR
 * THE USE OR OTHER DEALINGS IN THE SOFTWARE. SHOULD THE SOFTWARE PROVE DEFECTIVE
 * IN ANY RESPECT, THE USER ASSUMES THE COST OF ANY NECESSARY SERVICING, REPAIR OR
 * CORRECTION. THIS DISCLAIMER OF WARRANTY CONSTITUTES AN ESSENTIAL PART OF THIS LICENSE.
 * NO USE OF ANY SOFTWARE IS AUTHORIZED HEREUNDER EXCEPT UNDER THIS DISCLAIMER.
 * DESY HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS,
 * OR MODIFICATIONS.
 * THE FULL LICENSE SPECIFYING FOR THE SOFTWARE THE REDISTRIBUTION, MODIFICATION,
 * USAGE AND OTHER RIGHTS AND OBLIGATIONS IS INCLUDED WITH THE DISTRIBUTION OF THIS
 * PROJECT IN THE FILE LICENSE.HTML. IF THE LICENSE IS NOT INCLUDED YOU MAY FIND A COPY
 * AT HTTP://WWW.DESY.DE/LEGAL/LICENSE.HTM
 */

package de.desy.acop.displayers;

import java.beans.PropertyVetoException;
import java.util.Map;

import javax.swing.Action;

import com.cosylab.gui.InfoDialog;
import com.cosylab.gui.SliderDisplayer;
import com.cosylab.gui.components.SliderSetMode;
import com.cosylab.gui.components.util.IconHelper;
import com.cosylab.gui.displayers.DataConsumer;
import com.cosylab.gui.displayers.DataState;
import com.cosylab.gui.displayers.DisplayerUtilities;
import com.cosylab.gui.displayers.DoubleConsumer;
import com.cosylab.gui.displayers.DoubleConsumerMulticaster;
import com.cosylab.gui.displayers.DoubleSeqConsumer;
import com.cosylab.gui.util.UserSettingsProtection;
import com.cosylab.util.CommonException;

import de.desy.acop.displayers.tools.AcopDisplayer;
import de.desy.acop.displayers.tools.AcopDisplayerTransferHandler;
import de.desy.acop.displayers.tools.AcopInfoDialog;
import de.desy.acop.launcher.Launcher;
import de.desy.acop.transport.ConnectionParameters;
import de.desy.acop.transport.adapters.AcopTransportDataSource;
import de.desy.acop.transport.adapters.AdapterFactory;
import de.desy.acop.transport.adapters.AdapterFactoryService;

/**
 * <code>AcopSlider</code> is a displayer which can show a single 
 * double value. The slider can be used as a component which controls
 * settable connection points.
 * <p>
 * The component consists of a JSlider which has a user interface 
 * is sensitive to mouse movement. User can set a value by dragging the 
 * thumb of the JSlider. The thumb will always show the value that the
 * user has set. A black trailer on the same slider will show the actual 
 * value that is read from the remote property. The position of the trailer
 * cannot be set in runtime.
 * </p>
 * <p>
 * Value on AcopSlider can also be adjusted using the navigation buttons
 * or the number field on the top of the component. By setting the properties
 * {@link #setBitStep(double)}, {@link #setSmallStep(double)} and 
 * {@link #setLargeStep(double)} one can configure the size of the steps used
 * when pressing the navigations buttons. If using the number field the
 * value has to be confirmed by pressing enter key before it results in the 
 * movement of the thumb.
 * </p>
 * <p>
 * Slider has three different types of set modes (see {@link SliderSetMode}) which
 * specify at what actions the set value will be sent to the remote connection point.
 * </p>
 * <p>
 * The connection point for this displayer should be set using the 
 * <i>connectionParameters</i> property, where the remote name of the 
 * {@link ConnectionParameters} points to the desired property.
 * </p>
 * 
 * 
 * @author <a href="mailto:jaka.bobnar@cosylab.com">Jaka Bobnar</a>
 *
 */
public class AcopSlider extends SliderDisplayer implements DoubleSeqConsumer,  
	AcopDisplayer {
	 
	public static void main(String[] args) {
		
		Launcher.launch(AcopSlider.class, args, true);
		
	}
	
	private static final long serialVersionUID = -2503329416853153653L;
	private ConnectionParameters connectionParameters;
	private int arrayIndex=0;
	private AcopInfoDialog dialog;

	/**
	 * Constructs new AcopSlider.
	 *
	 */
	public AcopSlider() {
		super();
		setRestoreButtonTitle("Old");
		setStorageButtonsVisible(true);
		setAcopIcons();
		setContinuousControlVisible(true);
		new AcopDisplayerTransferHandler(this);
		UserSettingsProtection.setProtection(this,DisplayerUtilities.COMMON_NUMERIC_DISPLAYER_PROPERTIES,false);		
	}
	
	/**
	 * Apply navigation button icons demanded by ACOP
	 * 
	 * bitDecrement button uses ArrowLeft.gif instead of Minus16.gif
	 * bitIncrement button uses ArrowRight.gif instead of Plus16.gif
	 * slowDecrement button uses ArrowLeftLeft.gif instead of ArrowLeftLeft.gif
	 * slowIncrement button uses ArrowRightRight.gif instead of
	 * ArrowRightRight.gif fastDecrement button uses ArrowLeftLeftLeft.gif
	 * instead of ArrowLeftLeft.gif fastIncrement button uses
	 * ArrowRightRightRight.gif instead of ArrowRightRight.gif
	 * 
	 */
	public void setAcopIcons() {
		setButtonIcon(Button.BIT_DECREMENT, IconHelper.createIcon("icons/navigation/ArrowLeft16.gif"));
		setButtonIcon(Button.BIT_INCREMENT, IconHelper.createIcon("icons/navigation/ArrowRight16.gif"));

		setButtonIcon(Button.SLOW_DECREMENT, IconHelper.createIcon("icons/navigation/ArrowLeftLeft16.gif"));
		setButtonIcon(Button.SLOW_INCREMENT, IconHelper.createIcon("icons/navigation/ArrowRightRight16.gif"));

		setButtonIcon(Button.FAST_DECREMENT, IconHelper.createIcon("icons/navigation/ArrowLeftLeftLeft16.gif"));
		setButtonIcon(Button.FAST_INCREMENT, IconHelper.createIcon("icons/navigation/ArrowRightRightRight16.gif"));
	}

	/*
	 * (non-Javadoc)
	 * @see com.cosylab.gui.displayers.DoubleSeqConsumer#updateValue(long, double[])
	 */
	public void updateValue(long timestamp, double[] value) throws CommonException {
	    updateValue(timestamp, DisplayerUtilities.extract(arrayIndex,value));	
    }

	/* (non-Javadoc)
	 * @see com.cosylab.gui.displayers.DataConsumer#getDataConsumer(java.lang.Class)
	 */
	@Override
	public DataConsumer getDataConsumer(Class type)
	{
		if (type == DoubleConsumer.class || type == DoubleSeqConsumer.class) {
			return this;
		}
		
		return DoubleConsumerMulticaster.createDataConsumer(type, this);
	}
	
	/*
	 * (non-Javadoc)
	 * @see com.cosylab.gui.DialKnobDisplayer#setCharacteristics(java.util.Map)
	 */
	@Override
	public void setCharacteristics(Map characteristics) {
		super.setCharacteristics(characteristics);
		
		Object o  = characteristics.get("isChannel");
		if (o != null) {
			if (((Boolean)o)) {
				try {
					UserSettingsProtection.setUnprotected(this,"arrayIndex",0);
				} catch (Exception e) {
					e.printStackTrace();
				}
			}
		}
	}
	
	/*
	 * (non-Javadoc)
	 * @see de.desy.acop.displayers.AcopDisplayerConnector#getConnectionParameters()
	 */
	public ConnectionParameters getConnectionParameters() {
		return connectionParameters;
	}
	
	/*
	 * (non-Javadoc)
	 * @see de.desy.acop.displayers.AcopDisplayerConnector#setConnectionParameters(de.desy.acop.transport.ConnectionParameters)
	 */
	public void setConnectionParameters(ConnectionParameters param) throws CommonException, PropertyVetoException {

		if (param!=null && connectionParameters != null) {
			if(param.equals(connectionParameters)) return;
		}
		updateDataState(new DataState(DataState.UNDEFINED));
			
		ConnectionParameters old = connectionParameters;
		AdapterFactory factory = AdapterFactoryService.getInstance().getAdapterFactory();
		if (getDataSource() != null)
			factory.releaseDataSource((AcopTransportDataSource) getDataSource());
		if (param == null) {
			setDataSource(null);
			setTitle("No Connection");
		} else {
			if (param.getPropertySize() == ConnectionParameters.AUTO_PROPERTY_SIZE) {
				param = param.deriveWithPropertySize(1);
			}
			AcopTransportDataSource ds = factory.createDataSource(param);
			setDataSource(ds);
		}
		this.connectionParameters = param;
		firePropertyChange(CONNECTION_PARAMETERS_PROPERTY,old, connectionParameters);
	}

	/*
	 * (non-Javadoc)
	 * @see de.desy.acop.displayers.tools.AcopDisplayer#getArrayIndex()
	 */
	public int getArrayIndex() {
		return arrayIndex;
	}

	/*
	 * (non-Javadoc)
	 * @see de.desy.acop.displayers.tools.AcopDisplayer#setArrayIndex(int)
	 */
	public void setArrayIndex(int arrayIndex) {
		if (this.arrayIndex == arrayIndex) return;
		int oldValue = this.arrayIndex;
		this.arrayIndex = arrayIndex;
		firePropertyChange("arrayIndex", oldValue, this.arrayIndex);
	}

	/*
	 * (non-Javadoc)
	 * @see com.cosylab.gui.SliderDisplayer#getInfoDialog()
	 */
	@Override
	public InfoDialog getInfoDialog() {
		if (dialog == null) {
			dialog = new AcopInfoDialog(this);
		}
		return dialog;
	}
	
	/**
	 * Enables/disables Properties item in the popup menu.
	 * @param enable
	 */
	public void setPropertiesPopupEnabled(boolean enable) {
		Action[] actions = getPopupManager().getActions();
		for (Action a : actions) {
			if (a != null && "Preferences...".equals(a.getValue(Action.NAME))) {
				a.setEnabled(enable);
				return;
			}
		}
	}
	
	/**
	 * Returns true if Properties item is enabled in the popup menu.
	 * @return
	 */
	public boolean isPropertiesPopupEnabled() {
		Action[] actions = getPopupManager().getActions();
		for (Action a : actions) {
			if (a != null && "Preferences...".equals(a.getValue(Action.NAME))) {
				return a.isEnabled();
			}
		}
		return false;
	}
}

/* __oOo__ */
