/*
 * Copyright (c) 2006 Stiftung Deutsches Elektronen-Synchroton,
 * Member of the Helmholtz Association, (DESY), HAMBURG, GERMANY.
 *
 * THIS SOFTWARE IS PROVIDED UNDER THIS LICENSE ON AN "../AS IS" BASIS.
 * WITHOUT WARRANTY OF ANY KIND, EXPRESSED OR IMPLIED, INCLUDING BUT NOT LIMITED
 * TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR PARTICULAR PURPOSE AND
 * NON-INFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE
 * FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
 * TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR
 * THE USE OR OTHER DEALINGS IN THE SOFTWARE. SHOULD THE SOFTWARE PROVE DEFECTIVE
 * IN ANY RESPECT, THE USER ASSUMES THE COST OF ANY NECESSARY SERVICING, REPAIR OR
 * CORRECTION. THIS DISCLAIMER OF WARRANTY CONSTITUTES AN ESSENTIAL PART OF THIS LICENSE.
 * NO USE OF ANY SOFTWARE IS AUTHORIZED HEREUNDER EXCEPT UNDER THIS DISCLAIMER.
 * DESY HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS,
 * OR MODIFICATIONS.
 * THE FULL LICENSE SPECIFYING FOR THE SOFTWARE THE REDISTRIBUTION, MODIFICATION,
 * USAGE AND OTHER RIGHTS AND OBLIGATIONS IS INCLUDED WITH THE DISTRIBUTION OF THIS
 * PROJECT IN THE FILE LICENSE.HTML. IF THE LICENSE IS NOT INCLUDED YOU MAY FIND A COPY
 * AT HTTP://WWW.DESY.DE/LEGAL/LICENSE.HTM
 */

package de.desy.acop.displayers;

import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.datatransfer.DataFlavor;
import java.awt.event.ActionEvent;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;
import java.awt.event.MouseMotionListener;
import java.beans.PropertyVetoException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;

import javax.swing.AbstractAction;
import javax.swing.Action;
import javax.swing.JComponent;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.ListSelectionModel;
import javax.swing.TransferHandler;

import com.cosylab.gui.adapters.Converter;
import com.cosylab.gui.components.ObjectTable;
import com.cosylab.gui.components.customizer.AbstractCustomizerPanel;
import com.cosylab.gui.components.table.cells.TableCell;
import com.cosylab.gui.components.util.CosyTransferHandler;
import com.cosylab.gui.components.util.PopupManager;
import com.cosylab.gui.components.util.RunnerHelper;
import com.cosylab.gui.components.util.ScreenCapturer;
import com.cosylab.gui.displayers.DataConsumer;
import com.cosylab.gui.displayers.DataSource;
import com.cosylab.gui.displayers.DataState;
import com.cosylab.gui.displayers.DoubleSeqConsumer;
import com.cosylab.gui.util.DisplayerParametersFlavor;
import com.cosylab.gui.util.MultipleDisplayerParametersFlavor;
import com.cosylab.util.CommonException;

import de.desy.acop.displayers.selector.SelectorUtilities;
import de.desy.acop.displayers.table.AcopBitTableColumn;
import de.desy.acop.displayers.table.AcopNumericTableColumn;
import de.desy.acop.displayers.table.AcopStringTableColumn;
import de.desy.acop.displayers.table.AcopTableColumn;
import de.desy.acop.displayers.table.AcopTableModel;
import de.desy.acop.displayers.table.AcopTableParameters;
import de.desy.acop.displayers.table.AcopTableParametersFlavor;
import de.desy.acop.displayers.table.TableColumnType;
import de.desy.acop.displayers.tools.AcopMultipleDisplayerInfoDialog;
import de.desy.acop.displayers.tools.AcopTableTransferHandler;
import de.desy.acop.displayers.tools.ConnectionParametersFlavor;
import de.desy.acop.displayers.tools.MultipleAcopDisplayer;
import de.desy.acop.transport.AccessMode;
import de.desy.acop.transport.ConnectionFailed;
import de.desy.acop.transport.ConnectionParameters;
import de.desy.acop.transport.adapters.AcopTransportDataSource;
import de.desy.acop.transport.adapters.AdapterFactory;
import de.desy.acop.transport.adapters.AdapterFactoryService;

/**
 * <code>AcopTable</code> is a <code>MultipleAcopDisplayer</code> which is 
 * inteded to show values of multi-channel arrays or spectrum arrays. 
 * The values in the table are shown in columns, where each column
 * represent one multi channel. To connect the table to some remote property
 * one has to add the appropriate {@link AcopTableParameters} using the method
 * {@link #addDisplayerParameters(AcopTableParameters)} or 
 * {@link #setDisplayerParameters(AcopTableParameters[])}. Upon setting
 * the parameters the AcopTable will determine the type of values that needs
 * to be shown and created the appropriate {@link AcopTableColumn}. After
 * a monitor on one of the multi-channels is no longer needed, that particular
 * parameters should be removed from the table using the 
 * {@link #removeDisplayerParameters(AcopTableParameters)} method.
 * 
 *  
 * @author <a href="mailto:jaka.bobnar@cosylab.com">Jaka Bobnar</a>
 *
 */
public class AcopTable extends JPanel implements MultipleAcopDisplayer<AcopTableParameters> {
   
	private static final long serialVersionUID = -2993155391577813158L;
	private JScrollPane scrollPane;
	private ObjectTable table;
	private AcopTableModel model;
	private int suspended = 0;    
	private Map<String,AcopTableColumn> consumers= new HashMap<>();
	private HashMap<AcopTableParameters, AcopTableColumn> connections = new HashMap<>();
	private DataState dataState = new DataState(DataState.NORMAL);
	private PopupManager popupManager;
	private AbstractCustomizerPanel customizer;
	private AcopTableParameters selectedParameters;
	private boolean popupEnabled;
	
	private AcopMultipleDisplayerInfoDialog dialog;
	
	/**
	 * Constructs new AcopTable.
	 *
	 */
	public AcopTable() {
		super();
		initialize();
		DataFlavor[] flavor = new DataFlavor[]{AcopTableParametersFlavor.FLAVOR,
				MultipleDisplayerParametersFlavor.FLAVOR,
				DisplayerParametersFlavor.FLAVOR,ConnectionParametersFlavor.FLAVOR, 
				DataFlavor.stringFlavor};
		new AcopTableTransferHandler(this, true, true, flavor, false);
	}
	
	/**
	 * Returns the parameters that are selected in the table.
	 * 
	 * @return the selected parameters
	 */
	public AcopTableParameters getSelectedParameters() {
		return selectedParameters;
	}
	
	private void initialize() {
		this.setLayout(new BorderLayout());
		scrollPane = new JScrollPane(getTable());
		this.add(scrollPane, BorderLayout.CENTER);
		setPopupEnabled(true);
		scrollPane.addMouseListener(getPopupManager().getMouseHook());
		getTable().addMouseListener(getPopupManager().getMouseHook());
		setNumberOfRows(20);
		getTable().getSelectionModel().setSelectionMode(ListSelectionModel.SINGLE_SELECTION);
		getTable().addMouseListener(new MouseAdapter() {
			@Override
			public void mousePressed(MouseEvent e) {
				int row =getTable().getSelectedRow();
				int column = getTable().getSelectedColumn();
				if (row >= 0 && column >= 0) {
					Object o = getTable().getModel().getValueAt(row, column);
					if (o instanceof TableCell) {
						Object source = ((TableCell)o).getDataSource();
						if (source instanceof AcopTableColumn) {
							AcopTableParameters parameters = ((AcopTableColumn)source).getDisplayerParameters();
							if (SelectorUtilities.isChannel(parameters.getConnectionParameters())) {
								try {
									String device = ((String[])AdapterFactoryService.getInstance().getAdapterFactory().getCharacteristics(parameters.getConnectionParameters()).get("xLabels"))[row];
									ConnectionParameters cp = parameters.getConnectionParameters();
									ConnectionParameters newCP = new ConnectionParameters(cp.getAccessProtocol(), cp.getDeviceContext(), cp.getDeviceGroup(), device, cp.getDeviceProperty(), cp.getAccessMode(), cp.getAccessRate(), -1);
									selectedParameters = parameters.deriveWith(newCP);
									return;
								} catch (ConnectionFailed e1) {
									e1.printStackTrace();
								} 
							} 

							if (SelectorUtilities.getSequenceLength(parameters.getConnectionParameters()) > 1) {
								selectedParameters = new AcopTableParameters(parameters.getConnectionParameters().deriveWithPropertySize(-1), parameters.getColumnType(), row, parameters.getConverter(), parameters.getColumnName());
							} else {
								selectedParameters = parameters.deriveWith(parameters.getConnectionParameters().deriveWithPropertySize(-1));
							}

						}
					}
				}
			}
		});
	}
	
	/**
	 * Returns the JTable used by this component. User can change the table
	 * as he wishes and can replace any part of the table except for the
	 * TableModel. Thi JTable is not replaceable by other instances.
	 * 
	 * @return the JTable used by this component 
	 */
	public ObjectTable getTable() {
		if (table == null) {
			table = new ObjectTable();
			table.setModel(getTableModel());
		}
		return table;
	}
	
	private AcopTableModel getTableModel() {
		if (model == null) {
			model = new AcopTableModel();
		}
		return model;
	}
	
	/* (non-Javadoc)
	 * @see com.cosylab.gui.displayers.CommonDisplayer#cleanup()
	 */
	public void cleanup() {
		for(AcopTableColumn c: consumers.values()) {
			c.destroy();
		}
		consumers.clear();
	}

	/* (non-Javadoc)
	 * @see com.cosylab.gui.displayers.CommonDisplayer#suspend()
	 */
	public void suspend() {
		suspended++;
	}

	/* (non-Javadoc)
	 * @see com.cosylab.gui.displayers.CommonDisplayer#resume()
	 */
	public synchronized void resume() {
		if (suspended>0) {
			suspended--;
		}
	}

	/* (non-Javadoc)
	 * @see com.cosylab.gui.displayers.CommonDisplayer#isSuspended()
	 */
	public boolean isSuspended() {
		return suspended>0;
	}

	/* (non-Javadoc)
	 * @see com.cosylab.gui.displayers.DataStateProvider#getDataState()
	 */
	public DataState getDataState() {
		return dataState;
	}

	/* (non-Javadoc)
	 * @see com.cosylab.gui.displayers.MultipleDataConsumer#getConsumer(java.lang.String, java.lang.Class)
	 */
	public <D extends DataConsumer> D getConsumer(String name, Class<D> type) {
		return getConsumer(name, name, type);
	}
	
	private <D extends DataConsumer> D getConsumer(String uniqueName, String shortName, Class<D> type) {
				
		if (consumers.containsKey(uniqueName)) {
			return type.cast(consumers.get(uniqueName));
		}
		
		AcopTableColumn c = null;
		if (type.isAssignableFrom(AcopNumericTableColumn.class)) {
			c = new AcopNumericTableColumn(this, uniqueName, shortName);
		} else if (type.isAssignableFrom(AcopBitTableColumn.class)) {
			c = new AcopBitTableColumn(this, uniqueName, shortName);
		} else if (type.isAssignableFrom(AcopStringTableColumn.class)) {
			c = new AcopStringTableColumn(this, uniqueName, shortName);
		}
		
		if (c != null) {
			getTableModel().addColumn(c);
			consumers.put(uniqueName,c);
			return type.cast(c);
		}
		
		return null;
	}

	/* (non-Javadoc)
	 * @see com.cosylab.gui.displayers.MultipleDataConsumer#releaseConsumer(com.cosylab.gui.displayers.DataConsumer)
	 */
	public void releaseConsumer(DataConsumer consumer) {
		consumers.remove(consumer.getName()).destroy();
		getTableModel().removeColumn((AcopTableColumn)consumer);
	}

	/* (non-Javadoc)
	 * @see com.cosylab.gui.displayers.MultipleDataConsumer#getConsumers()
	 */
	public DataConsumer[] getConsumers() {
		return consumers.values().toArray(new DataConsumer[consumers.size()]);
	}

	/* (non-Javadoc)
	 * @see com.cosylab.gui.displayers.MultipleDataConsumer#getSupportedConsumerTypes()
	 */
	@SuppressWarnings("unchecked")
	public Class<DataConsumer>[] getSupportedConsumerTypes() {
		return new Class[]{DoubleSeqConsumer.class};
	}
	
	private static String[] getDeviceNames(AcopTableParameters parameters) throws ConnectionFailed {
		AdapterFactory factory = AdapterFactoryService.getInstance().getAdapterFactory();
		String[] names = (String[])factory.getCharacteristics(parameters.getConnectionParameters()).get("xLabels");
		int size = parameters.getConnectionParameters().getPropertySize();
		if (size <= 0 || size == names.length) {
			return names;
		}
		String[] retVal = new String[size];
		for (int i = 0; i < size; i++) {
			retVal[i] = names[i];
		}
		return retVal;
	}
	
	/*
	 * (non-Javadoc)
	 * @see de.desy.acop.displayers.AcopMultipleDisplayerConnector#addConnectionParameters(de.desy.acop.transport.ConnectionParameters)
	 */
	public void addDisplayerParameters(AcopTableParameters parameters) throws CommonException, PropertyVetoException {
		if (parameters == null) return;
		if (connections.get(parameters) != null) {
			return;         
		}  
		suspend();
		AcopTableParameters[] old = getDisplayerParameters();
		AcopTableColumn consumer = null;
		try {
    		AdapterFactory factory = AdapterFactoryService.getInstance().getAdapterFactory();
    		if (parameters.getColumnType() == TableColumnType.NAME) {
    			consumer = getConsumer(parameters.toString(), parameters.getName(), AcopStringTableColumn.class);
    			consumer.setDisplayerParameters(parameters);
    			((AcopStringTableColumn)consumer).updateValue(0,getDeviceNames(parameters));
    		} else {
    			if (parameters.getColumnType() == TableColumnType.STRING) {
    				consumer = getConsumer(parameters.toString(), parameters.getName(), AcopStringTableColumn.class);
    			} else if (parameters.getColumnType() == TableColumnType.BIT) {
    				consumer = getConsumer(parameters.toString(), parameters.getName(), AcopBitTableColumn.class);
    			} else {
    				consumer = getConsumer(parameters.toString(), parameters.getName(), AcopNumericTableColumn.class);
    			}
    			AcopTransportDataSource ds = factory.createDataSource(parameters.getConnectionParameters());
    			if (parameters.getConverter() != null && !parameters.getColumnType().equals(TableColumnType.STRING)){
    				Converter c = parameters.getConverter();
    				c.addConsumer(consumer);
    				ds.addConsumer(c);
    			}
    			else {
    				ds.addConsumer(consumer);
    			}
    			consumer.setDisplayerParameters(parameters);
    		}
		} catch (PropertyVetoException e) {
			releaseConsumer(consumer);
			throw e;
		} catch (CommonException e) {
			releaseConsumer(consumer);
			throw e;
		}

		connections.put(parameters, consumer);
		resume();
		firePropertyChange(MultipleAcopDisplayer.DISPLAYER_PARAMETERS_PROPERTY, old, getDisplayerParameters());
    }
 	
	/**
	 * 
	 * @see de.desy.acop.displayers.tools.MultipleAcopDisplayer#addConnectionParameters(de.desy.acop.transport.ConnectionParameters)
	 * @deprecated use {@link #addDisplayerParameters(AcopTableParameters)}
	 */
	@Deprecated
	public void addConnectionParameters(ConnectionParameters parameters) throws CommonException, PropertyVetoException {
		
		addDisplayerParameters(new AcopTableParameters(parameters, TableColumnType.VALUE, null));
    }
	
	/*
	 * (non-Javadoc)
	 * @see de.desy.acop.displayers.AcopMultipleDisplayerConnector#getConnectionParameters()
	 */
	public AcopTableParameters[] getDisplayerParameters() {
		AcopTableParameters[] params = new AcopTableParameters[connections.keySet().size()];
	    return connections.keySet().toArray(params);
    }
	
	/**
	 * (non-Javadoc)
	 * @see de.desy.acop.displayers.tools.MultipleAcopDisplayer#getConnectionParameters()
	 * @deprecated use {@link #getDisplayerParameters()}
	 */
	@Deprecated
	public ConnectionParameters[] getConnectionParameters() {
		
		AcopTableParameters[] agp = getDisplayerParameters();
		ConnectionParameters[] cp = new ConnectionParameters[agp.length];
		for (int i = 0; i < agp.length; i++) {
			cp[i] = agp[i].getConnectionParameters();
		}
		return cp;
	}
	
	/*
	 * (non-Javadoc)
	 * @see de.desy.acop.displayers.AcopMultipleDisplayerConnector#removeConnectionParameters(de.desy.acop.transport.ConnectionParameters)
	 */
	public DataSource removeDisplayerParameters(AcopTableParameters parameters) {
		AcopTableParameters[] old = getDisplayerParameters();
		AdapterFactory factory = AdapterFactoryService.getInstance().getAdapterFactory();
		AcopTableColumn consumer = connections.remove(parameters);
		if (consumer == null) return null;
		DataSource ds = null;
		if (parameters.getColumnType() != TableColumnType.NAME) {
			ds = factory.releaseDataSource(parameters.getConnectionParameters());
			if (parameters.getConverter() != null) {
				ds.removeConsumer(parameters.getConverter());
				parameters.getConverter().removeConsumer(consumer);
			} else {
				ds.removeConsumer(consumer);
			}
		}
		releaseConsumer(consumer);
		firePropertyChange(MultipleAcopDisplayer.DISPLAYER_PARAMETERS_PROPERTY, old, getDisplayerParameters());
	    return ds;
	    
    }
	
	/**
	 * 
	 * @see de.desy.acop.displayers.tools.MultipleAcopDisplayer#removeConnectionParameters(de.desy.acop.transport.ConnectionParameters)
	 * @deprecated use {@link #removeDisplayerParameters(AcopTableParameters)}
	 */
	@Deprecated
	public DataSource removeConnectionParameters(ConnectionParameters parameters) {
		AcopTableParameters[] old = getDisplayerParameters();
		DataSource ds = null;
		for (AcopTableParameters p : old) {
			if (p.getConnectionParameters().equals(parameters)) {
				AdapterFactory factory = AdapterFactoryService.getInstance().getAdapterFactory();
				AcopTableColumn consumer = connections.remove(p);
				if (consumer == null) continue;
				ds = factory.releaseDataSource(parameters);
				ds.removeConsumer(consumer);
				releaseConsumer(consumer);
			}
		}
		firePropertyChange(MultipleAcopDisplayer.DISPLAYER_PARAMETERS_PROPERTY, old, getDisplayerParameters());
		return ds;
	}
	
	/*
	 * (non-Javadoc)
	 * @see de.desy.acop.displayers.AcopMultipleDisplayerConnector#setConnectionParameters(de.desy.acop.transport.ConnectionParameters[])
	 */
	public void setDisplayerParameters(AcopTableParameters[] parameters) throws CommonException, PropertyVetoException {
		AcopTableParameters[] old = getDisplayerParameters();
		ArrayList<AcopTableParameters> toBeRemoved = new ArrayList<>();
		Iterator<AcopTableParameters> it = connections.keySet().iterator();
		while(it.hasNext()) {
			toBeRemoved.add(it.next());
		}
		
		ArrayList<AcopTableParameters> toBeAdded = new ArrayList<>();
	
		for (AcopTableParameters p : parameters) {
			if (toBeRemoved.contains(p)) {
				toBeRemoved.remove(p);
			} else {
				toBeAdded.add(p);
			}
		}
		
		for(AcopTableParameters p : toBeRemoved) {
			removeDisplayerParameters(p);
		}
		
		for(AcopTableParameters p : toBeAdded) {
			addDisplayerParameters(p);
		}
		
		firePropertyChange(MultipleAcopDisplayer.DISPLAYER_PARAMETERS_PROPERTY, old, parameters);
    }
	
	/*
	 * (non-Javadoc)
	 * @see de.desy.acop.displayers.tools.MultipleAcopDisplayer#setConnectionParameters(de.desy.acop.transport.ConnectionParameters[])
	 */
	@Deprecated
	public void setConnectionParameters(ConnectionParameters[] parameters) throws CommonException, PropertyVetoException {
		  
		AcopTableParameters[] agp = new AcopTableParameters[parameters.length];
		for (int i = 0; i < parameters.length; i++) {
			agp[i] = new AcopTableParameters(parameters[i], TableColumnType.VALUE, null);
		}
		setDisplayerParameters(agp);
	}
	
	/**
	 * Returns the Customizer intance for this displayer.
	 *
	 * @return the Customizer intance for this displayer
	 */
	public AbstractCustomizerPanel getCustomizer()
	{
		if (customizer == null) {
			customizer = AbstractCustomizerPanel.findCustomizer(this);
		}
		return customizer;
	}

	/*
	 * (non-Javadoc)
	 * @see com.cosylab.gui.components.util.PopupManageable#getPopupManager()
	 */
	public PopupManager getPopupManager() {
		if (popupManager == null) {
			popupManager = new PopupManager(this, false);
			popupManager.addAction(new AbstractAction("Preferences...") {
				private static final long serialVersionUID = 1L;
				public void actionPerformed(ActionEvent e)
				{
					getCustomizer().showDialog();
				}
			});
			
			popupManager.addAction(new AbstractAction("Capture screen..."){
				private static final long serialVersionUID = 1L;

				public void actionPerformed(ActionEvent e){
				   	ScreenCapturer sc = new ScreenCapturer(AcopTable.this);
					sc.showScreenDialog();
				   }
			});
			
			popupManager.addAction(new AbstractAction("Info...") {
				private static final long serialVersionUID = 1L;
				public void actionPerformed(ActionEvent e)
				{
					getInfoDialog().setVisible(true);
				}
			});
		}
		return popupManager;
	}
	
	/**
	 * Return true if the popup menu is enabled or false otherwise.
	 * 
	 * @return true if popup is enabled
	 */
	public boolean isPopupEnabled() {
		return popupEnabled;
	}
	
	/**
	 * Enables or disables the popup menu.
	 * 
	 * @param enabled true if enable or false if disableds
	 */
	public void setPopupEnabled(boolean enabled) {
		if (popupEnabled == enabled) return;
		popupEnabled = enabled;
		if (enabled) {
			addMouseListener(getPopupManager().getMouseHook());
		} else {
			removeMouseListener(getPopupManager().getMouseHook());
		}
		firePropertyChange("popupEnabled",!popupEnabled,popupEnabled);
	}
	
	/**
	 * Returns the number of rows displayed in the table.
	 * @return
	 */
	public int getNumberOfRows() {
		return getTableModel().getRowCount();
	}
	
	/**
	 * Sets the number of rows displayer in the table.
	 * @param numberOfRows
	 */
	public void setNumberOfRows(int numberOfRows) {
		getTableModel().setRowCount(numberOfRows);
	}
	
	/*
	 * (non-Javadoc)
	 * @see javax.swing.JComponent#setTransferHandler(javax.swing.TransferHandler)
	 */
	@Override
	public void setTransferHandler(TransferHandler newHandler) {
		super.setTransferHandler(newHandler);
		CosyTransferHandler.registerTransferHandler(this, newHandler, new JComponent[]{scrollPane, getTable()});
	}

	
	/*
	 * (non-Javadoc)
	 * @see java.awt.Component#removeMouseMotionListener(java.awt.event.MouseMotionListener)
	 */
	@Override
	public synchronized void removeMouseMotionListener(MouseMotionListener l) {
		super.removeMouseMotionListener(l);
		scrollPane.removeMouseMotionListener(l);
		getTable().removeMouseMotionListener(l);
	}
	
	/**
	 * Enables/disables mouse dragging. Dragging can only be enabled if
	 * this component uses CosyTransferHandler.
	 * 
	 * @param enabled
	 */
	public void setDragEnabled(boolean enabled) {
		if (getTransferHandler() instanceof CosyTransferHandler) {
			if (((CosyTransferHandler)getTransferHandler()).isExportEnabled() == enabled) return;
			
			((CosyTransferHandler)getTransferHandler()).setExportEnabled(enabled, this);
		}
	}
	
	/**
	 * Returns true if drag is enabled.
	 * 
	 * @return true if drag is enabled
	 */
	public boolean isDragEnabled() {
		if (getTransferHandler() instanceof CosyTransferHandler) {
			return ((CosyTransferHandler)getTransferHandler()).isExportEnabled();
		} else {
			return getTransferHandler() != null;
		}
	}
	
	/**
	 * Enable/disable the mouse drop. Drop can only be enabled if this component uses
	 * CosyTransferHandler.
	 * 
	 * @param enabled
	 */
	public void setDropEnabled(boolean enabled) {
		if (getTransferHandler() instanceof CosyTransferHandler) {
			if (((CosyTransferHandler)getTransferHandler()).isReceiveEnabled() == enabled) return;
			
			((CosyTransferHandler)getTransferHandler()).setReceiveEnabled(enabled, this);
		}
	}
	
	/**
	 * Returns true if drop is enabled.
	 * 
	 * @return true if drop is enabled
	 */
	public boolean isDropEnabled() {
		if (getTransferHandler() instanceof CosyTransferHandler) {
			return ((CosyTransferHandler)getTransferHandler()).isReceiveEnabled();
		} else {
			return getTransferHandler() != null;
		}
	}
	
	/**
	 * Returns the info dialog associated with this table.
	 * 
	 * @return the info dialog
	 */
	public AcopMultipleDisplayerInfoDialog getInfoDialog() {
		if (dialog == null) {
			dialog = new AcopMultipleDisplayerInfoDialog(this);
		}
		return dialog;
	}
	
	/**
	 * Enables/disables Properties item in the popup menu.
	 * @param enable
	 */
	public void setPropertiesPopupEnabled(boolean enable) {
		Action[] actions = getPopupManager().getActions();
		for (Action a : actions) {
			if (a != null && "Preferences...".equals(a.getValue(Action.NAME))) {
				a.setEnabled(enable);
				return;
			}
		}
	}
	
	/**
	 * Returns true if Properties item is enabled in the popup menu.
	 * @return
	 */
	public boolean isPropertiesPopupEnabled() {
		Action[] actions = getPopupManager().getActions();
		for (Action a : actions) {
			if (a != null && "Preferences...".equals(a.getValue(Action.NAME))) {
				return a.isEnabled();
			}
		}
		return false;
	}
	
	/**
	 * Enables/disables reordering of the columns in the table in runtime.
	 * 
	 * @param reorderingAllowed true if reordering should be allowed
	 */
	public void setReorderingAllowed(boolean reorderingAllowed) {
		if (getReorderingAllowed() == reorderingAllowed) return;
		getTable().getTableHeader().setReorderingAllowed(reorderingAllowed);
		firePropertyChange("reordeningAllowed", !reorderingAllowed, reorderingAllowed);
	}
	
	/**
	 * Returns true if reordering of columns in the table is allowed in runtime.
	 * 
	 * @return true if reordening is allowed
	 */
	public boolean getReorderingAllowed() {
		return getTable().getTableHeader().getReorderingAllowed();
	}	
	
	/*
	 * (non-Javadoc)
	 * @see javax.swing.JComponent#setBackground(java.awt.Color)
	 */
	@Override
	public void setBackground(Color bg) {
		super.setBackground(bg);
		getTable().setBackground(bg);
	}
	
	/*
	 * (non-Javadoc)
	 * @see javax.swing.JComponent#setForeground(java.awt.Color)
	 */
	@Override
	public void setForeground(Color fg) {
		super.setForeground(fg);
		getTable().setForeground(fg);
	}
	
	

	public static void main(String[] args) throws CommonException, PropertyVetoException {
		AcopTable table = new AcopTable();
		table.setForeground(Color.RED);
//		table.addDisplayerParameters(new AcopTableParameters(new ConnectionParameters("TINE", "WORKSHOP", "WKSineGen","TempModule0", "Temperature", AccessMode.POLL, 1000), TableDisplayType.NAME));
//		table.addDisplayerParameters(new AcopTableParameters(new ConnectionParameters("TINE", "WORKSHOP", "WKSineGen","TempModule0", "Temperature", AccessMode.POLL, 1000), TableDisplayType.BIT));
//		table.addDisplayerParameters(new AcopTableParameters(new ConnectionParameters("TINE", "WORKSHOP", "WKSineGen","TempModule0", "Temperature", AccessMode.POLL, 1000), TableDisplayType.VALUE));
//		table.addDisplayerParameters(new AcopTableParameters(new ConnectionParameters("TINE", "WORKSHOP", "WKSineGen","Device 0", "Sine", AccessMode.POLL, 1000), TableDisplayType.NAME));
//		table.addDisplayerParameters(new AcopTableParameters(new ConnectionParameters("TINE", "WORKSHOP", "WKSineGen","Device 0", "Sine", AccessMode.POLL, 1000), TableDisplayType.BIT));
		table.addDisplayerParameters(new AcopTableParameters(new ConnectionParameters("TINE", "WORKSHOP", "WKSineGen","Device 0", "Amplitude", AccessMode.POLL, 1000), TableColumnType.VALUE, "bla"));
		
		RunnerHelper.runComponent(table, 800, 600);
	}
	
}
