package de.desy.acop.displayers.chart;

import java.awt.Color;
import java.awt.Component;
import java.awt.Graphics;
import java.awt.GridBagConstraints;
import java.awt.GridBagLayout;
import java.awt.Insets;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;
import java.util.HashMap;

import javax.swing.DefaultListCellRenderer;
import javax.swing.DefaultListModel;
import javax.swing.Icon;
import javax.swing.JButton;
import javax.swing.JDialog;
import javax.swing.JLabel;
import javax.swing.JList;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.SwingUtilities;

import com.cosylab.application.AutoScalable;
import com.cosylab.gui.displayers.DataConsumer;

import de.desy.acop.displayers.AcopChartReorg;

/**
 * 
 * <code>LegendDialog</code> is a dialog which displays a legend of all traces in 
 * the chart.
 *
 * @author <a href="mailto:jaka.bobnar@cosylab.com">Jaka Bobnar</a>
 *
 */
@AutoScalable
public class LegendDialog extends JDialog {

	private class CellRenderer extends DefaultListCellRenderer {
		
		private class DotIcon implements Icon {
			private Color color;
			private DotIcon(Color color) {
				this.color = color;
			}
			public int getIconHeight() {
				return 16;
			}
			public int getIconWidth() {
				return 16;
			}
			public void paintIcon(Component c, Graphics g, int x, int y) {
				Color col = g.getColor();
				g.setColor(color);
				g.fillRect(x, y, 16, 16);
				g.setColor(col);
			}
		}
		private static final long serialVersionUID = 1L;
		private HashMap<AcopChartConsumer, Icon> cache = new HashMap<>();
		@Override
		public Component getListCellRendererComponent(JList<?> list, Object value, int index, boolean isSelected, boolean cellHasFocus) {
			Component c = super.getListCellRendererComponent(list, value, index, isSelected, cellHasFocus);
			if (value instanceof AcopChartConsumer && c instanceof JLabel) {
				AcopChartConsumer consumer = (AcopChartConsumer) value;
				JLabel l = (JLabel)c;
				
				l.setText(consumer.getName());
				l.setIcon(getCachedIcon(consumer));
			}
			return c;
			
		}
		
		private Icon getCachedIcon(AcopChartConsumer consumer) {
			Icon icon = cache.get(consumer);
			if (icon == null) {
				icon = new DotIcon(consumer.getColor());
				cache.put(consumer, icon);
			}
			return icon;
		}
	}
	
	
	private static final long serialVersionUID = 1L;
	private JPanel contentPanel;
	private JList<DataConsumer> legendList;
	private DefaultListModel<DataConsumer> model;
	private JButton closeButton;
	private AcopChartReorg chart;
	
	/**
	 * Constructs a new LegendDialog associated with the given chart.
	 * 
	 * @param chart
	 */
	public LegendDialog(AcopChartReorg chart) {
		super();
		this.chart = chart;
		chart.addPropertyChangeListener(AcopChartReorg.CONSUMERS, new PropertyChangeListener(){
			public void propertyChange(PropertyChangeEvent evt) {
				fillList(LegendDialog.this.chart.getConsumers());
				
			}
		});
		initialize();
		fillList(LegendDialog.this.chart.getConsumers());
	}
	
	private void initialize() {
		setSize(500,500);
		setLocationRelativeTo(chart);
		setDefaultCloseOperation(JDialog.DISPOSE_ON_CLOSE);
		setContentPane(getContentPanel());
	}
	
	private JPanel getContentPanel() {
		if (contentPanel == null) {
			contentPanel = new JPanel(new GridBagLayout());
			JScrollPane scroll = new JScrollPane(getLegendList());
			contentPanel.add(scroll, new GridBagConstraints(0,0,1,1,1,1,GridBagConstraints.CENTER,GridBagConstraints.BOTH, new Insets(5,5,5,5),0,0));
			closeButton = new JButton("Close");
			closeButton.addActionListener(new ActionListener(){
				public void actionPerformed(ActionEvent e) {
					dispose();
				}
			});
			contentPanel.add(closeButton, new GridBagConstraints(0,1,1,1,1,0,GridBagConstraints.EAST, GridBagConstraints.NONE, new Insets(5,5,5,5),0,0));			
		}
		return contentPanel;
	}
	
	private JList<DataConsumer> getLegendList() {
		if (legendList == null) {
			legendList = new JList<>(getListModel());
			legendList.setCellRenderer(new CellRenderer());
		}
		return legendList;
	}
	
	private DefaultListModel<DataConsumer> getListModel() {
		if (model == null) {
			model = new DefaultListModel<>();
		}
		return model;
	}
	
	private void fillList(final DataConsumer[] consumers) {
		SwingUtilities.invokeLater(new Runnable(){
			public void run() {
				getListModel().clear();
				for (DataConsumer c : consumers) {
					getListModel().addElement(c);
				}
			}
		});
	}
}
