/*
 * Copyright (c) 2006 Stiftung Deutsches Elektronen-Synchroton,
 * Member of the Helmholtz Association, (DESY), HAMBURG, GERMANY.
 *
 * THIS SOFTWARE IS PROVIDED UNDER THIS LICENSE ON AN "../AS IS" BASIS.
 * WITHOUT WARRANTY OF ANY KIND, EXPRESSED OR IMPLIED, INCLUDING BUT NOT LIMITED
 * TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR PARTICULAR PURPOSE AND
 * NON-INFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE
 * FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
 * TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR
 * THE USE OR OTHER DEALINGS IN THE SOFTWARE. SHOULD THE SOFTWARE PROVE DEFECTIVE
 * IN ANY RESPECT, THE USER ASSUMES THE COST OF ANY NECESSARY SERVICING, REPAIR OR
 * CORRECTION. THIS DISCLAIMER OF WARRANTY CONSTITUTES AN ESSENTIAL PART OF THIS LICENSE.
 * NO USE OF ANY SOFTWARE IS AUTHORIZED HEREUNDER EXCEPT UNDER THIS DISCLAIMER.
 * DESY HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS,
 * OR MODIFICATIONS.
 * THE FULL LICENSE SPECIFYING FOR THE SOFTWARE THE REDISTRIBUTION, MODIFICATION,
 * USAGE AND OTHER RIGHTS AND OBLIGATIONS IS INCLUDED WITH THE DISTRIBUTION OF THIS
 * PROJECT IN THE FILE LICENSE.HTML. IF THE LICENSE IS NOT INCLUDED YOU MAY FIND A COPY
 * AT HTTP://WWW.DESY.DE/LEGAL/LICENSE.HTM
 */

package de.desy.acop.displayers.chart;

import java.awt.GridBagConstraints;
import java.awt.GridBagLayout;
import java.awt.Insets;
import java.awt.event.KeyAdapter;
import java.awt.event.KeyEvent;

import javax.swing.JButton;
import javax.swing.JLabel;
import javax.swing.JPanel;
import javax.swing.JTextField;

import com.cosylab.gui.components.numberfield.DoubleDocument;

import de.desy.acop.displayers.tools.AcopUtilities;

/**
 * <code>MinMaxPanel</code> provides a user interface through which minimum
 * and maximum values can be entered. When values are entered into the field
 * the actual properties of this panel are not change until the user
 * presses the <i>Apply</i> button.
 * 
 * @author <a href="mailto:jaka.bobnar@cosylab.com">Jaka Bobnar</a>
 *
 */
public class MinMaxPanel extends JPanel {
		
	private static final long serialVersionUID = 1L;

	private JButton rescaleButton;
	
	private JTextField maxField;
	private JTextField minField;
	private DoubleDocument maxDocument;
	private DoubleDocument minDocument;
	private JLabel maxLabel;
	private JLabel minLabel;
	private double min = 0;
	private double max = 0;
	private boolean rescaleEnabled = false;
	private boolean extraDigits = false;
	private boolean useScientific = true;
	
	
	/**
	 * Creates a new MinMaxPanel.
	 *
	 */
	public MinMaxPanel() {
		super();
		initialize();
	}
	
	private void initialize() {
		setLayout(new GridBagLayout());	
		
		maxField = new JTextField();
		maxDocument = new DoubleDocument();
		maxField.setDocument(maxDocument);
		maxField.setText(String.valueOf(max));
		maxField.addKeyListener(new KeyAdapter() {
			@Override
			public void keyPressed(KeyEvent evt) {
				rescaleEnabled = true;
				rescaleButton.setEnabled(isEnabled());
			}
		});
		maxField.addActionListener(e -> rescaleValue());
		maxField.setToolTipText("Maximum Value");
		
		maxLabel = new JLabel("Max");
		this.add(maxField, new GridBagConstraints(1,0,1,1,1,0,GridBagConstraints.WEST, GridBagConstraints.BOTH, new Insets(0,2,2,0),0,0));
		this.add(maxLabel, new GridBagConstraints(0,0,1,1,0,0,GridBagConstraints.WEST, GridBagConstraints.NONE, new Insets(0,0,2,2),0,0));
		
		minField = new JTextField();
		minDocument = new DoubleDocument();
		minField.setDocument(minDocument);
		minField.setText(String.valueOf(min));
		minField.addKeyListener(new KeyAdapter() {
			@Override
			public void keyPressed(KeyEvent evt) {
				rescaleEnabled = true;
				rescaleButton.setEnabled(isEnabled());
			}
		});
		minField.addActionListener(e -> rescaleValue());
		minField.setToolTipText("Minimum Value");
		
		minLabel = new JLabel("Min");
		this.add(minField, new GridBagConstraints(1,1,1,1,1,0,GridBagConstraints.WEST, GridBagConstraints.HORIZONTAL, new Insets(2,2,2,0),0,0));
		this.add(minLabel, new GridBagConstraints(0,1,1,1,0,0,GridBagConstraints.WEST, GridBagConstraints.NONE, new Insets(0,2,2,2),0,0));
		
		rescaleButton = new JButton("Apply Scale");
		rescaleButton.addActionListener(e -> rescaleValue());
		rescaleButton.setEnabled(false);
		rescaleButton.setToolTipText("Apply New Scale Settings");
		rescaleButton.setEnabled(false);
		
		this.add(rescaleButton, new GridBagConstraints(0,2,2,1,1,0,GridBagConstraints.WEST, GridBagConstraints.HORIZONTAL, new Insets(2,0,0,0),0,0));
	}
	
	private void rescaleValue() {
	    rescaleEnabled = false;
        Double min = minDocument.getValue();
        Double max = maxDocument.getValue();
        if (min != null && max != null) { 
            setMin(minDocument.getValue().doubleValue());
            setMax(maxDocument.getValue().doubleValue());
            rescaleButton.setEnabled(false);
        }
	}
	
	/**
	 * Returns the maximum value.
	 * @return the max value
	 */
	public double getMax() {
		return max;
	}

	/**
	 * Sets maximum value and fires property change event.
	 * 
	 * @param max new maximum value
	 */
	public void setMax(double max) {
		setMax(max, true);
	}
	
	/**
	 * Sets new maximum value and fires a property change event if requested.
	 * 
	 * @param max new maximum value
	 * @param notify should property change event be triggered
	 */
	public void setMax(double max, boolean notify) {
		if (this.max == max) return;
		double oldValue = this.max;
		this.max = max;
		applyMax(max);
		if (notify) {
			firePropertyChange("max", oldValue, this.max);
		}
	}
		
	private void applyMax(double max) {
		boolean enabled = maxField.isEnabled();
		maxField.setEnabled(true);
		maxField.setText(AcopUtilities.formatValue(max, extraDigits));
		maxField.setEnabled(enabled);
		maxField.setCaretPosition(0);
	}
	
	/**
	 * Returns the minimum value.
	 * @return the minimum value
	 */
	public double getMin() {
		return min;
	}

	/**
	 * Sets the minimum and fires property change event.
	 * 
	 * @param min new minimum
	 */
	public void setMin(double min) {
		setMin(min, true);
	}
	
	/**
	 * Sets new minimum value and fires a property change event if requested.
	 * 
	 * @param min new minimum value
	 * @param notify should property change event be triggered
	 */
	public void setMin(double min, boolean notify) {
		if (this.min == min) return;
		double oldValue = this.min;
		this.min = min;
		applyMin(min);
		if (notify) {
			firePropertyChange("min", oldValue, this.min);
		}
	}
	
	private void applyMin(double min) {
		boolean enabled = minField.isEnabled();
		minField.setEnabled(true);
		minField.setText(AcopUtilities.formatValue(min, extraDigits));
		minField.setEnabled(enabled);
		minField.setCaretPosition(0);
	}
	
	/**
	 * Sets the apply button enabled/disabled.
	 * 
	 * @param enabled
	 */
	public void setRescaleEnabled(boolean enabled) {
		if (enabled == rescaleEnabled) return;
		this.rescaleEnabled = enabled;
		rescaleButton.setEnabled(enabled & isEnabled());
		firePropertyChange("rescaleEnabled", !enabled, enabled);
	}
	
	/**
	 * Returns true if the apply button is enabled.
	 * 
	 * @return the state of the apply button
	 */
	public boolean isRescaleEnabled() {
		return rescaleEnabled;
	}

	/*
	 * (non-Javadoc)
	 * @see javax.swing.JComponent#setEnabled(boolean)
	 */
	@Override
	public void setEnabled(boolean enabled) {
		super.setEnabled(enabled);
		minField.setEnabled(enabled);
		maxField.setEnabled(enabled);
		minLabel.setEnabled(enabled);
		maxLabel.setEnabled(enabled);
		rescaleButton.setEnabled(enabled & rescaleEnabled);
	}
	
	/**
	 * Sets the flag of extra digits. If extra digits is turned on 6-decimal
	 * format is used, otherwise 2-decimal format is used.
	 * 
	 * @param extraDigits
	 */
	public void setExtraDigits(boolean extraDigits) {
		if (this.extraDigits == extraDigits) return;
		this.extraDigits = extraDigits;
		firePropertyChange("extraDigits", !extraDigits, extraDigits);
		applyMax(max);
		applyMin(min);
	}
	
	/**
	 * Returns true if extra digits flag is true.
	 * 
	 * @return
	 */
	public boolean isExtraDigits() {
		return extraDigits;
	}
	
	/**
	 * If scientific is set to true, the value will be formatted using
	 * a scientific exponential format if value is less than 1e-2 or 
	 * greater or equal than 1e4.
	 * @param scientific
	 */
	public void setUseScientific(boolean scientific) {
		if (this.useScientific == scientific) return;
		this.useScientific = scientific;
		firePropertyChange("useScientific", !scientific, scientific);
		applyMax(max);
		applyMin(min);
	}
	
	/**
	 * Returns true if scientific formatting is used.
	 * @return
	 */
	public boolean isUseScientific() {
		return useScientific;
	}
}
