/*
 * Copyright (c) 2006 Stiftung Deutsches Elektronen-Synchroton,
 * Member of the Helmholtz Association, (DESY), HAMBURG, GERMANY.
 *
 * THIS SOFTWARE IS PROVIDED UNDER THIS LICENSE ON AN "../AS IS" BASIS.
 * WITHOUT WARRANTY OF ANY KIND, EXPRESSED OR IMPLIED, INCLUDING BUT NOT LIMITED
 * TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR PARTICULAR PURPOSE AND
 * NON-INFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE
 * FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
 * TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR
 * THE USE OR OTHER DEALINGS IN THE SOFTWARE. SHOULD THE SOFTWARE PROVE DEFECTIVE
 * IN ANY RESPECT, THE USER ASSUMES THE COST OF ANY NECESSARY SERVICING, REPAIR OR
 * CORRECTION. THIS DISCLAIMER OF WARRANTY CONSTITUTES AN ESSENTIAL PART OF THIS LICENSE.
 * NO USE OF ANY SOFTWARE IS AUTHORIZED HEREUNDER EXCEPT UNDER THIS DISCLAIMER.
 * DESY HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS,
 * OR MODIFICATIONS.
 * THE FULL LICENSE SPECIFYING FOR THE SOFTWARE THE REDISTRIBUTION, MODIFICATION,
 * USAGE AND OTHER RIGHTS AND OBLIGATIONS IS INCLUDED WITH THE DISTRIBUTION OF THIS
 * PROJECT IN THE FILE LICENSE.HTML. IF THE LICENSE IS NOT INCLUDED YOU MAY FIND A COPY
 * AT HTTP://WWW.DESY.DE/LEGAL/LICENSE.HTM
 */

package de.desy.acop.displayers.chart;

import com.cosylab.util.PrimitiveDoubleList;

import de.desy.tine.histUtils.THistory;
import de.desy.tine.types.DBLDBL;

/**
 * <code>THistoryConnector</code> is used to retrieve data from the 
 * history server.
 * 
 * @author Tilen Kusterle, Cosylab
 *
 */
public class THistoryConnector {
	
	private boolean simulate;
	private boolean checkForBadData;
	private int timeout = 3000;

	protected synchronized double[][] getHistoryData(String context, String server, String device, String property, int index, double start, double stop) {		
		int size = 1000;
		DBLDBL[] data = new DBLDBL[size];
		if (simulate) {
			double span = stop - start;
			for (int i = 0; i < size; i++) {
				data[i] = new DBLDBL(Math.random()*3 + 60, start + span/size * i);
			}
		} else {
			THistory.getArchivedData(context, server, property, device, index, start, stop, data, timeout);
		}
	
		PrimitiveDoubleList x = new PrimitiveDoubleList();
		PrimitiveDoubleList y = new PrimitiveDoubleList();
		if (isCheckForBadData()) {
			for (int i = 0; i < data.length; i++) {
				if (data[i].d2val == 0 || data[i].d1val > 1E30) continue;
				x.add(data[i].d2val);
				y.add(data[i].d1val);
			}
		} else {
			for (int i = 0; i < data.length; i++) {
				if (data[i].d2val == 0) continue;
				x.add(data[i].d2val);
				y.add(data[i].d1val);
			}
		}
		
		double[][] result = new double[2][x.size()];
		result[0] = x.toArray();
		result[1] = y.toArray();
		return result;		
	}
	
	/**
	 * Returns the history data for the specified remote point between 
	 * the given start and stop dates in TINE time.
	 * 
	 * @param hp history connection point
	 * @param start starting time
	 * @param stop end time
	 * @return array of history data of length 2: the first array are timestamps and
	 * 			the second are the values
	 */
	public double[][] loadHistory(HistoryParameters hp, double start, double stop) {
		if (hp == null) throw new IllegalArgumentException("History parameters can not be null.");
		return getHistoryData(hp.getHistoryContext(), hp.getHistoryServer(), hp.getHistoryDevice(), hp.getHistoryProperty(), hp.getHistoryIndex(), start, stop);
	}

	/**
	 * Returns true if check for bad data is turned on.
	 * 
	 * @return the checkForBadData
	 */
	public boolean isCheckForBadData() {
		return checkForBadData;
	}

	/**
	 * Enables/disables checking for bad data. Bad data is any value higher than 1E30.
	 * If enabled these values are filtered out.
	 * 
	 * @param checkForBadData true if check for bad data is turned on
	 */
	public void setCheckForBadData(boolean checkForBadData) {
		this.checkForBadData = checkForBadData;
	}

	/**
	 * Is history data simulated.
	 * 
	 * @return true if data is simulated
	 */
	public boolean isSimulate() {
		return simulate;
	}

	/**
	 * Simulates the history data.
	 * 
	 * @param simulate true if history data should be simulated
	 */
	public void setSimulate(boolean simulate) {
		this.simulate = simulate;
	}
	
	/**
	 * Returns the timeout.
	 * 
	 * @return the timeout
	 */
	public int getTimeout() {
		return timeout;
	}

	/**
	 * Sets the timeout for all remote calls.
	 * 
	 * @param timeout the timeout to set
	 */
	public void setTimeout(int timeout) {
		this.timeout = timeout;
	}
	
}
