/*
 * Copyright (c) 2006 Stiftung Deutsches Elektronen-Synchroton,
 * Member of the Helmholtz Association, (DESY), HAMBURG, GERMANY.
 *
 * THIS SOFTWARE IS PROVIDED UNDER THIS LICENSE ON AN "../AS IS" BASIS.
 * WITHOUT WARRANTY OF ANY KIND, EXPRESSED OR IMPLIED, INCLUDING BUT NOT LIMITED
 * TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR PARTICULAR PURPOSE AND
 * NON-INFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE
 * FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
 * TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR
 * THE USE OR OTHER DEALINGS IN THE SOFTWARE. SHOULD THE SOFTWARE PROVE DEFECTIVE
 * IN ANY RESPECT, THE USER ASSUMES THE COST OF ANY NECESSARY SERVICING, REPAIR OR
 * CORRECTION. THIS DISCLAIMER OF WARRANTY CONSTITUTES AN ESSENTIAL PART OF THIS LICENSE.
 * NO USE OF ANY SOFTWARE IS AUTHORIZED HEREUNDER EXCEPT UNDER THIS DISCLAIMER.
 * DESY HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS,
 * OR MODIFICATIONS.
 * THE FULL LICENSE SPECIFYING FOR THE SOFTWARE THE REDISTRIBUTION, MODIFICATION,
 * USAGE AND OTHER RIGHTS AND OBLIGATIONS IS INCLUDED WITH THE DISTRIBUTION OF THIS
 * PROJECT IN THE FILE LICENSE.HTML. IF THE LICENSE IS NOT INCLUDED YOU MAY FIND A COPY
 * AT HTTP://WWW.DESY.DE/LEGAL/LICENSE.HTM
 */

package de.desy.acop.displayers.info;

import java.awt.BorderLayout;
import java.awt.GridBagConstraints;
import java.awt.GridBagLayout;
import java.awt.Insets;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;

import javax.swing.BorderFactory;
import javax.swing.JCheckBox;
import javax.swing.JPanel;
import javax.swing.JTextArea;

import de.desy.acop.displayers.tools.AcopInfoDialog;
import de.desy.acop.transport.ConnectionParameters;
import de.desy.tine.client.TLinkFactory;
import de.desy.tine.queryUtils.TQuery;

/**
 * <code>AcopDebugPanel</code> shows debug data for certain connection point. 
 * This panel can be used in combination with <code>AcopInfoDialog</code>.
 * 
 * @author <a href="mailto:jaka.bobnar@cosylab.com">Jaka Bobnar</a>
 * @see AcopInfoDialog
 */
public class AcopDebugPanel extends JPanel {
		
	private static final long serialVersionUID = -1597262900075362868L;
	private JTextArea textArea;
	private ConnectionParameters connectionParameters;
	private JCheckBox debugCheckBox;
	private boolean showDebug = true;
	
	/**
	 * Constructs new AcopDebugPanel.
	 *
	 */
	public AcopDebugPanel() {
		super();
		initialize();
	}
	
	private void initialize() {
		this.setLayout(new BorderLayout());
		textArea = new JTextArea();
		textArea.setBorder(BorderFactory.createTitledBorder("Fec Information"));
		this.add(textArea, BorderLayout.CENTER);
		JPanel southPanel = new JPanel(new GridBagLayout());
		debugCheckBox = new JCheckBox("Debug");
		debugCheckBox.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent e) {
				TLinkFactory.getInstance().setDebugLevel(debugCheckBox.isSelected() ? 2 : 0);					
			}
		});
		southPanel.add(debugCheckBox, new GridBagConstraints(0,0,1,1,1,1,GridBagConstraints.WEST, GridBagConstraints.NONE, new Insets(0,0,0,0),0,0));
		this.add(southPanel, BorderLayout.NORTH);			
	}
	
	/**
	 * Sets the ConnectionParameters, which specify the connection point that
	 * debug data is shown for.
	 *  
	 * @param param the remote connection point
	 */
	public void setConnectionParameters(ConnectionParameters param) {
		if (this.connectionParameters != null && this.connectionParameters.equals(param)) return;
		this.connectionParameters = param;
		if (connectionParameters != null){
			textArea.setText("Updating ... Please wait.");
			new Thread(new Runnable() {
				public void run() {
					String srv = connectionParameters.getDeviceGroup();
			        String con = connectionParameters.getDeviceContext();
			        StringBuffer s = new StringBuffer(TQuery.getModuleAddressInfo(srv, con) +"\n");
			        s.append(TQuery.getTineVersion(con, srv) + "\n");
			        s.append(TQuery.getAppVersion(con, srv) + "\n");
			        s.append(TQuery.getAppDate(con, srv) + "\n");
			        textArea.setText(s.substring(0));
				}
			}).start(); 
		} else {
			textArea.setText("");
		}
	}
	
	/**
	 * Returns the ConnectionParameters that specify the remote connection point, which
	 * is being debugged.
	 * 
	 * @return
	 */
	public ConnectionParameters getConnectionParameters() {
		return connectionParameters;
	}
	
	/**
	 * Sets the debug level for TINE TLinkFactory. It always uses level 2.
	 * 
	 * @param debug new debug level (0 debug off)
	 * @see TLinkFactory#setDebugLevel(int)
	 */
	public void setDebug(int debug) {
		if (debug > 0) {
			debugCheckBox.setSelected(true);
		} else {
			debugCheckBox.setSelected(false);
		}
	}
	
	/**
	 * Returns the debug level (0 - debug is off, 2 - debug is on).
	 * 
	 * @return the debug level
	 */
	public int isDebug() {
		return debugCheckBox.isSelected() ? 2 : 0;
	}
	
	/**
	 * Sets the visibility of the debug checkbox.
	 * 
	 * @param showDebug true if the checkbox should be visible, false if not
	 */
	public void setShowDebug(boolean showDebug) {
		if (this.showDebug == showDebug) return;
		this.showDebug = showDebug;
		debugCheckBox.setVisible(showDebug);
		firePropertyChange("showDebug", !showDebug, showDebug);
	}
	
	/**
	 * Returns true if the debug check box is visible.
	 * @return the debug check box visibile
	 */
	public boolean isShowDebug() {
		return showDebug;
	}
}
