/*
 * Copyright (c) 2006 Stiftung Deutsches Elektronen-Synchroton,
 * Member of the Helmholtz Association, (DESY), HAMBURG, GERMANY.
 *
 * THIS SOFTWARE IS PROVIDED UNDER THIS LICENSE ON AN "../AS IS" BASIS.
 * WITHOUT WARRANTY OF ANY KIND, EXPRESSED OR IMPLIED, INCLUDING BUT NOT LIMITED
 * TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR PARTICULAR PURPOSE AND
 * NON-INFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE
 * FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
 * TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR
 * THE USE OR OTHER DEALINGS IN THE SOFTWARE. SHOULD THE SOFTWARE PROVE DEFECTIVE
 * IN ANY RESPECT, THE USER ASSUMES THE COST OF ANY NECESSARY SERVICING, REPAIR OR
 * CORRECTION. THIS DISCLAIMER OF WARRANTY CONSTITUTES AN ESSENTIAL PART OF THIS LICENSE.
 * NO USE OF ANY SOFTWARE IS AUTHORIZED HEREUNDER EXCEPT UNDER THIS DISCLAIMER.
 * DESY HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS,
 * OR MODIFICATIONS.
 * THE FULL LICENSE SPECIFYING FOR THE SOFTWARE THE REDISTRIBUTION, MODIFICATION,
 * USAGE AND OTHER RIGHTS AND OBLIGATIONS IS INCLUDED WITH THE DISTRIBUTION OF THIS
 * PROJECT IN THE FILE LICENSE.HTML. IF THE LICENSE IS NOT INCLUDED YOU MAY FIND A COPY
 * AT HTTP://WWW.DESY.DE/LEGAL/LICENSE.HTM
 */

package de.desy.acop.displayers.selector;

import java.awt.Dimension;
import java.awt.GridBagConstraints;
import java.awt.GridBagLayout;
import java.awt.Insets;
import java.beans.Customizer;
import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;

import javax.swing.JLabel;
import javax.swing.JPanel;
import javax.swing.JSlider;
import javax.swing.JSpinner;
import javax.swing.SpinnerNumberModel;
import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;

import com.cosylab.gui.components.util.RunnerHelper;

import de.desy.acop.displayers.tools.AcopDisplayer;
import de.desy.acop.transport.ConnectionParameters;

/**
 * <code>ArrayIndexCustomizer</code> is a customizer which enables selection
 * of a specific index in the array. When using spectrum properties on a single
 * value displayers one can use this customizer to select a specific index which
 * one wants to display on the widget. 
 * <p>
 * This component consist of a JSlider and JSpinner which allow selection of
 * a specific index within the size of the property.
 * </p>
 * 
 * @author <a href="mailto:jaka.bobnar@cosylab.com">Jaka Bobnar</a>
 *
 */
public class ArrayIndexCustomizer extends JPanel implements Customizer{
	
	private static final long serialVersionUID = -401693602095711422L;
	private AcopDisplayer disp;
	private int arrayIndex = 0;
	private JSlider slider;
	private ConnectionParameters connectionParameters;
	private JSpinner spinner;
	private SpinnerNumberModel spinnerModel;
	private int lastValue = 0;
	
	/**
	 * Constructs a new ArrayIndexCustomizer.
	 *
	 */
	public ArrayIndexCustomizer() {
		super();
		initialize();
	}
	
	private void initialize() {
		this.setLayout(new GridBagLayout());
		add(new JLabel("Array index"), new GridBagConstraints(0,0,1,1,1,0,GridBagConstraints.EAST,GridBagConstraints.NONE,new Insets(4,11,4,4),0,0));
		add(getSpinner(), new GridBagConstraints(1,0,1,1,1,0,GridBagConstraints.WEST,GridBagConstraints.NONE,new Insets(4,4,4,11),0,0));
		add(getSlider(), new GridBagConstraints(0,1,2,1,1,0,GridBagConstraints.CENTER,GridBagConstraints.HORIZONTAL,new Insets(4,11,4,11),0,0));

	}
	
	private JSlider getSlider() {
		if (slider == null) {
			slider = new JSlider();
	
			slider.setMinimum(0);
			slider.setValue(0);
			slider.setMaximum(100);
			slider.setMajorTickSpacing((int)Math.floor(100/4.0));
			slider.setMinorTickSpacing(1);
			slider.setPaintLabels(true);
			slider.setPaintTicks(true);
			slider.setPaintTrack(true);
			slider.setSnapToTicks(true);
			slider.addChangeListener(new ChangeListener() {
				
				public void stateChanged(ChangeEvent e) {
					//register only when value is changed, dont care about max/min
					if (slider.getValue() != lastValue) {
						lastValue = slider.getValue();
						setArrayIndex(lastValue);
						if (disp!=null) {
							disp.setArrayIndex(lastValue);
						}
						
					}
				}
			
			});
			
		}
		return slider;
	}
	
	private void refreshIndexList() {
		if (connectionParameters != null) {
			int length = connectionParameters.getPropertySize();
			if (length <= 0) {
				length = SelectorUtilities.getSequenceLength(connectionParameters);
			}
			if (length <= 1 || SelectorUtilities.isChannel(connectionParameters)) {
				slider.setEnabled(false);
				spinner.setEnabled(false);
				slider.setMaximum(1);
				spinnerModel.setMaximum(1);
			} else {
				slider.setEnabled(true);
				spinner.setEnabled(true);
				slider.setMaximum(length-1);
				spinnerModel.setMaximum(length-1);
				slider.setLabelTable(null);
				slider.setLabelTable(null);
				slider.setMajorTickSpacing((int)Math.floor(length/4.0));
			}
		} else {
			slider.setMaximum(1);
			spinnerModel.setMaximum(1);
			slider.setEnabled(false);
			spinner.setEnabled(false);
		}
	}
		
	/*
	 * (non-Javadoc)
	 * @see java.beans.Customizer#setObject(java.lang.Object)
	 */
	public void setObject(Object bean) {
		if (bean instanceof AcopDisplayer) {
			disp = (AcopDisplayer) bean;
			disp.addPropertyChangeListener("arrayIndex",
					new PropertyChangeListener(){
				public void propertyChange(PropertyChangeEvent evt) {
					setArrayIndex(disp.getArrayIndex());				
				}
			});
			disp.addPropertyChangeListener(AcopDisplayer.CONNECTION_PARAMETERS_PROPERTY,
					new PropertyChangeListener(){
				public void propertyChange(PropertyChangeEvent evt) {
					setConnectionParameters(disp.getConnectionParameters());
					setArrayIndex(disp.getArrayIndex());				
				}
			});
		} else {
			throw new IllegalArgumentException("Only AcopCommonDisplayer and" +
					" MultipleAcopDisplayer can use this Customizer.");
		}
		setConnectionParameters(disp.getConnectionParameters());
		setArrayIndex(disp.getArrayIndex());
	}
	
	/**
	 * Sets the selected array index.
	 * 
	 * @param index the selected index
	 */
	public void setArrayIndex(int index) {
		if (index == arrayIndex) return;
		int oldValue = arrayIndex;
		this.arrayIndex = index;
		getSlider().setValue(arrayIndex);
		getSpinner().setValue(arrayIndex);
		firePropertyChange("arrayIndex", oldValue, arrayIndex);
	}
	
	/**
	 * Returns the selected array index.
	 * 
	 * @return the selected index
	 */
	public int getArrayIndex() {
		return arrayIndex;
	}
	
	/**
	 * Sets the ConnectionParameters that points to the property, which index
	 * has to be selected for.
	 *   
	 * @param parameters remote connection point
	 */
	public void setConnectionParameters(ConnectionParameters parameters) {
		if (this.connectionParameters != null && this.connectionParameters.equals(parameters)) return;
		ConnectionParameters oldValue = this.connectionParameters;
		this.connectionParameters = parameters;
		refreshIndexList();
		firePropertyChange(AcopDisplayer.CONNECTION_PARAMETERS_PROPERTY, oldValue, this.connectionParameters);
	}
	
	/**
	 * Returns the ConnectionParameters that index is being selected for.
	 * 
	 * @return the connection parameters
	 */
	public ConnectionParameters getConnectionParameters() {
		return connectionParameters;
	}
	
	public static void main(String[] args) {
		RunnerHelper.runComponent(new ArrayIndexCustomizer(), 500, 420);
	}

	
	/**
	 * Returns the spinner.
	 * @return Returns the spinner.
	 */
	private JSpinner getSpinner() {
		if (spinner==null) {
			spinner= new JSpinner();
			spinner.setMinimumSize(new Dimension(50,21));
			spinner.setPreferredSize(new Dimension(50,21));
			spinner.setModel(spinnerModel= new SpinnerNumberModel());
			spinnerModel.setStepSize(1);
			spinnerModel.setMinimum(0);
			spinnerModel.setMaximum(100);
			spinnerModel.addChangeListener(new ChangeListener() {
			
				public void stateChanged(ChangeEvent e) {
					getSlider().setValue(spinnerModel.getNumber().intValue());
				}
			
			});
		}
		return spinner;
	}

}
