/*
 * Copyright (c) 2006 Stiftung Deutsches Elektronen-Synchroton,
 * Member of the Helmholtz Association, (DESY), HAMBURG, GERMANY.
 *
 * THIS SOFTWARE IS PROVIDED UNDER THIS LICENSE ON AN "../AS IS" BASIS.
 * WITHOUT WARRANTY OF ANY KIND, EXPRESSED OR IMPLIED, INCLUDING BUT NOT LIMITED
 * TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR PARTICULAR PURPOSE AND
 * NON-INFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE
 * FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
 * TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR
 * THE USE OR OTHER DEALINGS IN THE SOFTWARE. SHOULD THE SOFTWARE PROVE DEFECTIVE
 * IN ANY RESPECT, THE USER ASSUMES THE COST OF ANY NECESSARY SERVICING, REPAIR OR
 * CORRECTION. THIS DISCLAIMER OF WARRANTY CONSTITUTES AN ESSENTIAL PART OF THIS LICENSE.
 * NO USE OF ANY SOFTWARE IS AUTHORIZED HEREUNDER EXCEPT UNDER THIS DISCLAIMER.
 * DESY HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS,
 * OR MODIFICATIONS.
 * THE FULL LICENSE SPECIFYING FOR THE SOFTWARE THE REDISTRIBUTION, MODIFICATION,
 * USAGE AND OTHER RIGHTS AND OBLIGATIONS IS INCLUDED WITH THE DISTRIBUTION OF THIS
 * PROJECT IN THE FILE LICENSE.HTML. IF THE LICENSE IS NOT INCLUDED YOU MAY FIND A COPY
 * AT HTTP://WWW.DESY.DE/LEGAL/LICENSE.HTM
 */

package de.desy.acop.displayers.selector;

import java.util.Map;

import com.cosylab.gui.displayers.CommonDisplayer;

import de.desy.acop.displayers.tools.AcopDisplayer;
import de.desy.acop.transport.ConnectionFailed;
import de.desy.acop.transport.ConnectionParameters;
import de.desy.acop.transport.adapters.AdapterFactoryService;

/**
 * <code>SelectorUtilities</code> provides several utility methods used by the
 * <code>SelectorGUI</code>
 * 
 * @author <a href="mailto:jaka.bobnar@cosylab.com">Jaka Bobnar</a>
 */
public class SelectorUtilities {

	private static ConnectionParameters latestParameters;
	private static Map<String, Object> latestCharacteristics;
	
	/**
	 * Checks if the parameters define a multi-channel array.
	 * 
	 * @param param remote connection point
	 * @return true if this is a channel array
	 */
	public static boolean isChannel(ConnectionParameters param) {
		try {
			if (param == null) return false;
			if (!(param.equals(latestParameters))) {
				latestCharacteristics = AdapterFactoryService.getInstance().getAdapterFactory().getCharacteristics(param);
			} 
			latestParameters = param;
			return (Boolean)latestCharacteristics.get("isChannel");
//			return ((TArrayType) AdapterFactoryService.getInstance().getAdapterFactory().getCharacteristics(param).get("arrayType")).isChannel();
		} catch (ConnectionFailed e) {
			e.printStackTrace();
		}
		return false;
	}
	
	/**
	 * Checks if ConnectionParameters are valid. Parameters are valid if all
	 * values are non-null and are not empty strings.
	 * 
	 * @param param 
	 * @return true if parameters are valid
	 */
	public static boolean isConnectionParametersValid(ConnectionParameters param) {
    	if (param == null) return false;
    	if (param.getAccessProtocol() == null || param.getAccessMode() == null 
    			|| param.getDeviceContext() == null || param.getDeviceGroup() == null 
    			|| param.getDeviceName() == null || param.getDeviceProperty() == null ||
    			param.getAccessProtocol().trim().equals("") || param.getDeviceContext().trim().equals("") 
    			|| param.getDeviceGroup().trim().equals("") 
    			|| param.getDeviceName().trim().equals("") || param.getDeviceProperty().trim().equals("")) return false;
    	return true;
    }
	
	/**
	 * Returns the info about specific connection point. When requesting the
	 * same info for several times in a row, only the first time a remote server
	 * is asked. The last info value is always stored for later use or until
	 * info for a different parameters is requested.
	 * 
	 * @param param
	 * @return the info
	 */
	public static String getInfo(ConnectionParameters param) {
		try {
			if (param == null) return "";
			if (!(param.equals(latestParameters))) {
				latestCharacteristics = AdapterFactoryService.getInstance().getAdapterFactory().getCharacteristics(param);
			} 
			latestParameters = param;
			return (String)latestCharacteristics.get(CommonDisplayer.C_DESCRIPTION);
//			return (String) AdapterFactoryService.getInstance().getAdapterFactory().getCharacteristics(param).get(CommonDisplayer.C_DESCRIPTION);
		} catch (ConnectionFailed e) {
			e.printStackTrace();
		}
		return "N/A";
	}
	
	/**
	 * Returns the sequence length of a channel array.
	 * 
	 * @param param 
	 * @return the sequence length
	 */
	public static int getSequenceLength(ConnectionParameters param) {
		try {
			if (param == null) return 1;
			if (!(param.equals(latestParameters))) {
				latestCharacteristics = AdapterFactoryService.getInstance().getAdapterFactory().getCharacteristics(param);
			} 
			latestParameters = param;
			return ((Integer)latestCharacteristics.get(AcopDisplayer.C_SEQUENCE_LENGTH)).intValue();
//			return ((Integer) AdapterFactoryService.getInstance().getAdapterFactory().getCharacteristics(param).get(AcopDisplayer.C_SEQUENCE_LENGTH)).intValue();
		} catch (ConnectionFailed e) {
			System.out.println(e.getMessage());
		}
		
		return 0;
	}
	
	/**
	 * Returns the extended data, from the DynamicParameters in the 
	 * ConnetionParameters, as a string which exactly shows the type 
	 * and the values.
	 * 
	 * @param param
	 * @return extended data as String
	 */
	public static String getExtendedDataString(ConnectionParameters param){
		StringBuilder builder = new StringBuilder();
		Object extendedData = param.getDynamicParameters();
		if (extendedData != null) {
			if (extendedData.getClass() == double[].class) {
				double[] array = (double[]) extendedData;
				builder.append("new double[]{");
				for (int i = 0; i < array.length; i++) {
					if (i != 0) builder.append(',');
					builder.append(array[i]);
				}
				builder.append('}');
			} else if (extendedData.getClass() == float[].class) {
				float[] array = (float[]) extendedData;
				builder.append("new float[]{");
				for (int i = 0; i < array.length; i++) {
					if (i != 0) builder.append(',');
					builder.append(array[i]);
				}
				builder.append('}');
			} else if (extendedData.getClass() == int[].class) {
				int[] array = (int[]) extendedData;
				builder.append("new int[]{");
				for (int i = 0; i < array.length; i++) {
					if (i != 0) builder.append(',');
					builder.append(array[i]);
				}
				builder.append('}');
			} else if (extendedData.getClass() == short[].class) {
				short[] array = (short[]) extendedData;
				builder.append("new short[]{");
				for (int i = 0; i < array.length; i++) {
					if (i != 0) builder.append(',');
					builder.append(array[i]);
				}
				builder.append('}');
			} else if (extendedData.getClass() == byte[].class) {
				byte[] array = (byte[]) extendedData;
				builder.append("new byte[]{");
				for (int i = 0; i < array.length; i++) {
					if (i != 0) builder.append(',');
					builder.append(array[i]);
				}
				builder.append('}');
			} else if (extendedData.getClass() == long[].class) {
				long[] array = (long[]) extendedData;
				builder.append("new long[]{");
				for (int i = 0; i < array.length; i++) {
					if (i != 0) builder.append(',');
					builder.append(array[i]);
				}
				builder.append('}');
			} else if (extendedData.getClass() == char[].class) {
				char[] array = (char[]) extendedData;
				builder.append("new char[]{");
				for (int i = 0; i < array.length; i++) {
					if (i != 0)	builder.append(',');
					builder.append(array[i]);
				}
				builder.append('}');
			} else if (extendedData.getClass() == String.class) {
				builder.append("\"" + (String) extendedData + "\"");
			}
		} else
			builder.append("null");
		return builder.toString();
	}
}
