/*
 * Copyright (c) 2006 Stiftung Deutsches Elektronen-Synchroton,
 * Member of the Helmholtz Association, (DESY), HAMBURG, GERMANY.
 *
 * THIS SOFTWARE IS PROVIDED UNDER THIS LICENSE ON AN "../AS IS" BASIS.
 * WITHOUT WARRANTY OF ANY KIND, EXPRESSED OR IMPLIED, INCLUDING BUT NOT LIMITED
 * TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR PARTICULAR PURPOSE AND
 * NON-INFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE
 * FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
 * TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR
 * THE USE OR OTHER DEALINGS IN THE SOFTWARE. SHOULD THE SOFTWARE PROVE DEFECTIVE
 * IN ANY RESPECT, THE USER ASSUMES THE COST OF ANY NECESSARY SERVICING, REPAIR OR
 * CORRECTION. THIS DISCLAIMER OF WARRANTY CONSTITUTES AN ESSENTIAL PART OF THIS LICENSE.
 * NO USE OF ANY SOFTWARE IS AUTHORIZED HEREUNDER EXCEPT UNDER THIS DISCLAIMER.
 * DESY HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS,
 * OR MODIFICATIONS.
 * THE FULL LICENSE SPECIFYING FOR THE SOFTWARE THE REDISTRIBUTION, MODIFICATION,
 * USAGE AND OTHER RIGHTS AND OBLIGATIONS IS INCLUDED WITH THE DISTRIBUTION OF THIS
 * PROJECT IN THE FILE LICENSE.HTML. IF THE LICENSE IS NOT INCLUDED YOU MAY FIND A COPY
 * AT HTTP://WWW.DESY.DE/LEGAL/LICENSE.HTM
 */

package de.desy.acop.displayers.table;

import com.cosylab.gui.adapters.Converter;

import de.desy.acop.displayers.tools.AcopDisplayerParameters;
import de.desy.acop.transport.ConnectionParameters;

/**
 * <code>AcopTableParameters</code> are an instance of <code>AcopDisplayerParameters</code>
 * usid in combination with <code>AcopTable</code>. Parameters provide remote connetion
 * point as well as settings that define the <code>AcopTableColumn</code> type and
 * its properties.
 * 
 * @author Jaka Bobnar, Cosylab
 * @see AcopTable
 * @see AcopTableColumn
 */
public class AcopTableParameters extends AcopDisplayerParameters {
	
	private static final long serialVersionUID = -4689823006847926238L;

	public AcopTableParameters() {
		this(null, TableColumnType.VALUE, null);
	}

	/**
	 * Constructs new AcopTableParameters using the provided ConnectionParameters.
	 * 
	 * @param parameters remote connection point
	 */
	public AcopTableParameters(ConnectionParameters parameters) {
		this(parameters, TableColumnType.VALUE, "");
	}
	
	/**
	 * Constructs new AcopTableParameters using the provided parameters
	 * 
	 * @param parameters remote connection point
	 * @param type type of the column shown in the table
	 * @param name the name of the column
	 */
	public AcopTableParameters(ConnectionParameters parameters, TableColumnType type, String name) {
		this(parameters, type, null, name);
	}
	
	/**
	 * Constructs new AcopTableParameters using the provided parameters
	 * 
	 * @param parameters remote connection point
	 * @param type type of the column shown in the table
	 * @param arrayIndex the index in the array data which should be the first row in the
	 * 			table column
	 * @param name the name of the column
	 */
	public AcopTableParameters(ConnectionParameters parameters, TableColumnType type, int arrayIndex, String name) {
		this(parameters, type, arrayIndex, null, name);
	}

	/**
	 * Constructs new AcopTableParameters using the supplied parameters. 
	 * This constructor assumes the TableColumnType.VALUE.
	 * 
	 * @param parameters the remote connection point
	 * @param converter converter associated with the data source
	 * @param name the name of column
	 * @see TableColumnType
	 */
	public AcopTableParameters(ConnectionParameters parameters, Converter converter, String name) {
		this(parameters, TableColumnType.VALUE, converter, name);
	}
	
	/**
	 * Constructs new AcopTableParameters using the supplied parameters.
	 * 
	 * @param parameters the remote connection point
	 * @param type type of the column shown in the table
	 * @param converter converter associated with the data source
	 * @param name the name of the column
	 */
	public AcopTableParameters(ConnectionParameters parameters, TableColumnType type, Converter converter, String name) {
		this(parameters, type, 0, converter, name);
	}
	
	/**
	 * Constructs new AcopTableParameters using the supplied parameters.
	 * 
	 * @param parameters the remote connection point
	 * @param type type of the column shown in the table
	 * @param arrayIndex the index in the array data which should be the first row in the
	 * 			table column
	 * @param converter converter associated with the data source
	 * @param name the name of the column
	 */
	public AcopTableParameters(ConnectionParameters parameters, TableColumnType type, int arrayIndex, Converter converter, String name) {
		super(parameters, arrayIndex, converter);
		values.put("columnType", type);
		values.put("columnName", name);
	}
	
	/**
	 * Returns the type of the column.
	 * 
	 * @return the column type
	 */
	public TableColumnType getColumnType() {
		return (TableColumnType)values.get("columnType");
	}
	
	/**
	 * Returns the name of the column.
	 * 
	 * @return the name
	 */
	public String getColumnName() {
		return (String) values.get("columnName");
	}
	
	/**
	 * Construct new AcopTableParameters which resembles this object but has
	 * a different ConnectionParameters.
	 * 
	 * @param parameters new remote connection point
	 * @return new AcopTableParameters
	 */
	public AcopTableParameters deriveWith(ConnectionParameters parameters) {
		return new AcopTableParameters(parameters, getColumnType(), getArrayIndex(), getConverter(), getColumnName());
	}
		
}
