/**
 * File: ANode.java
 * Date: Aug 23, 2010
 */
package de.desy.acop.displayers.tarantula;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.Enumeration;
import java.util.List;
import java.util.Vector;
import java.util.concurrent.locks.Lock;
import java.util.concurrent.locks.ReentrantLock;
import java.util.logging.Logger;

import javax.swing.tree.DefaultMutableTreeNode;
import javax.swing.tree.MutableTreeNode;

import de.desy.tine.definitions.TErrorList;

/**
 * Abstract and general tree node. 
 * 
 * @author <a href="mailto:marcus.walla(at)desy.de">walla</a>
 * @version $Id: ANode.java 79375 2017-11-27 06:35:55Z bobnar $
 */
abstract class ANode extends DefaultMutableTreeNode {

	// --- CLASS FIELDS --------------------------------------------------------
	
	private static final long serialVersionUID = -894807825144405794L;
	private static final Logger log = Utility.getLogger(ANode.class);
	public static final List<ANode> EMPTY_LIST = new ArrayList<>();
	
	// --- INSTANCE FIELDS -----------------------------------------------------

	private final Lock lock = new ReentrantLock();
	
	// --- CONSTRUCTOR ---------------------------------------------------------

	/**
	 * Constructor.
	 */
	protected ANode() {
		this(null);
	}

	/**
	 * Constructor.
	 * @param userObject User object. Contains the information to be displayed
	 * 		in the UI layer.
	 */
	protected ANode(final Object userObject) {
		this(userObject, true);
	}

	/**
	 * Constructor.
	 * @param userObject User object. Contains the information to be displayed
	 * 		in the UI layer.
	 * @param allowsChildren Allows children, i.e. if this node is a lead 
	 * 		({@code false}) or brunch ({@code true}).
	 */
	protected ANode(final Object userObject, final boolean allowsChildren) {
		super(userObject, allowsChildren);
		super.children = new Vector<>();
	}
	
	// --- INSTANCE METHODS ----------------------------------------------------
    
	/**
	 * <p>Returns all child nodes as a list of {@link NodeConnection}.</p>
	 * <p><b>Remark</b>: This list is not a generic one, which, for 
	 * example, returns a collection of the type {@code T extends ANode}. This 
	 * results in warning by adding the node to the list. Possibly anybody
	 * else knows how to solve this problem.</p>
	 * @return Child nodes.
	 */
    public List<NodeConnection> getChildNodes() {
    	if (super.children == null || super.children.size() == 0)
    		return Collections.emptyList();
    	
		this.lock.lock();
    	try {
	    	final List<NodeConnection> nodes = new ArrayList<>();
	
	    	for(final Object node: super.children)
	    		if (node instanceof NodeConnection)
	    			nodes.add((NodeConnection) node);
	    	
	    	return Collections.unmodifiableList(nodes); 
    	}
    	finally {
    		this.lock.unlock();
    	}
    }
    
    /**
     * Returns the number child nodes.
     * @return Number of child nodes.
     */
    @Override public int getChildCount() {
    	this.lock.lock();
    	try {
    		return super.getChildCount();
    	}
    	finally {
    		this.lock.unlock();
    	}
    }

    /** 
     * This method is deprecated and should never be used! Instead use the 
     * method of {@link #getChildNodes()}.
     */
	@SuppressWarnings("rawtypes")
	@Deprecated
	@Override
	public Enumeration children() {
		return super.children();
	}

    // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
    
	/**
	 * Returns the parent node as an instance of {@link ANode}. For the root 
	 * node it returns a {@code null}.
	 * @return Parent node.
	 */
    @Override public final ANode getParent() { return (ANode) super.getParent(); }
    
    /**
     * Returns the full address of this node as an character collection.
     * @return Full address.
     */
    abstract String getFullAddress();
    
    /**
     * Returns a connection information instance.
     * @return Connection information.
     */
    abstract ConnectionInfo getInfo();
   
    /** 
     * Returns the connection information transformed into a string instance.
     * @return Connection information. 
     */
    abstract String getInfoString();

    /**
     * Sets the condition of the node. See {@link ENodeCondition} for more
     * information about the conditions.
     * @param condition Condition.
     */
    abstract void setCondition(final ENodeCondition condition);
    
    /**
     * Reflects the condition flag. See {@link ENodeCondition} for more information.
     * @return Condition flag.
     */
    abstract ENodeCondition getCondition();
    
    /**
     * Returns the status information transformed into a string instance.
     * @return Status information.
     */
    abstract String getStatusLink();
    
    /**
     * Returns the status information as an integer value. For more information
     * about the integer value you should have a look at {@link TErrorList} which
     * contains all TINE error codes beside the successful ones.
     * @return Returns the link status code.
     */
    abstract int getStatusLinkCode();
    
    // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 

    @Override public final void add(final MutableTreeNode node) {
    	this.lock.lock();
    	try {
    		super.add(node);
    	}
    	finally {
    		this.lock.unlock();
    	}
    }
    
    @Override public final void remove(final int index) {
    	this.lock.lock();
    	try {
    		super.remove(index);
    	}
    	finally {
    		this.lock.unlock();
    	}
    }

    @Override public final void remove(final MutableTreeNode node) {
    	this.lock.lock();
    	try {
    		super.remove(node);
    	}
    	finally {
    		this.lock.unlock();
    	}
    }
    
    @Override public final void insert(final MutableTreeNode node, final int index) {
    	this.lock.lock();
    	try {
    		super.insert(node, index);
    	}
    	finally {
    		this.lock.unlock();
    	}
    }
    /**
     * Adds a collection of {@link ANode}'s as children to this node.
     * @param nodes Collection of {@link ANode}'s.
     */
    public final void addAllChildren(final Collection<ANode> nodes) {
    	this.lock.lock();
    	try {
	    	for(final ANode node: nodes)
	    		this.add(node);
    	}
    	finally {
    		this.lock.unlock();
    	}
    }
    
    /** This method is deprecated and should never be called. There also 
     * doesn't exist a replacement. */
    @Deprecated @Override public final void removeFromParent() {
    	log.warning("method is deprecated - never use this. There also doesn't exist a replacement");
    }

    /** Returns the equality of this method by comparing the full address by
     * ignoring the characters case elements. */
    @Override public boolean equals(final Object obj) {
    	if (obj == null)
    		return false;
    	else if (obj instanceof ANode)
    		return ((ANode) obj).getFullAddress().equalsIgnoreCase(this.getFullAddress());
		else 
			return super.equals(obj);
    }
    
    @Override
    public int hashCode() {
        return getFullAddress().hashCode();
    }
}
/* end of file */