package de.desy.acop.displayers.tarantula;

/**
 * <p>Connection exception which represents an exception if something with the
 * remote connection failed. An instance of this class returns the error code
 * as an integer value and the full address information depending on the 
 * protocol used.</p>
 * 
 * <p>There also exists two methods which depends on the current protocol used
 * and therefore should be implemented in the special protocol exception 
 * class.</p>
 * 
 * @author <a href="mailto:marcus.walla(at)desy.de">walla</a>
 * @version $Id: ConnectionException.java 40186 2010-11-03 14:12:06Z walla $
 */
abstract class ConnectionException extends Exception {

	// --- CLASS FIELDS ----------------------------------------------------
	
	private static final long serialVersionUID = 9165224509169875311L;
	
	// --- INSTANCE FIELDS -------------------------------------------------
	
	private final int errorCode;
	private final String fullAddress;
	
	// --- CONSTRUCTOR -----------------------------------------------------
	
	/**
	 * Constructor.
	 * 
	 * @param errorCode Error code.
	 */
	ConnectionException(final int errorCode) { 
		this.errorCode = errorCode; 
		this.fullAddress = "unknown";
	}
	
	/**
	 * Constructor.
	 * 
	 * @param errorCode Error code.
	 * @param fullAddress Full address.
	 */
	ConnectionException(final int errorCode, final String fullAddress) { 
		this.errorCode = errorCode; 
		this.fullAddress = fullAddress;
	}
	
	/**
	 * Constructor.
	 * 
	 * @param errorCode Error code.
	 * @param fullAddress Full address.
	 * @param throwable Throwable.
	 */
	ConnectionException(final int errorCode, final String fullAddress, final Throwable throwable) { 
		super(throwable); 
		this.errorCode = errorCode; 
		this.fullAddress = fullAddress;
	}
	
	/**
	 * Constructor.
	 * 
	 * @param errorCode Error code.
	 * @param throwable Throwable.
	 */
	ConnectionException(final int errorCode, final Throwable throwable) { 
		super(throwable); 
		this.errorCode = errorCode; 
		this.fullAddress = "unknown";
	}
	
	// --- INSTANCE METHODS ------------------------------------------------
	
	/**
	 * Returns the error code.
	 * @return Error code.
	 */
	int getErrorCode() { return this.errorCode; }
	/**
	 * Returns the full address. Can also include the string {@code unknown} if
	 * the corresponding constructor without the full address is called. 
	 * @return Full address.
	 */
	String getFullAddress() { return this.fullAddress; }
	/**
	 * <p>Returns the transformed integer error code as a string value.</p>
	 * @return Transformed TINE error code.
	 */
	abstract String getError();
	/**
	 * <p>Returns the transformed integer error code as a string value surround 
	 * the error code by brackets at the end.</p>
	 * @return Transformed TINE error code.
	 */
	abstract String getErrorWithCode();
	
	@Override public String toString() { return this.getFullAddress() + " - " + this.getErrorWithCode(); }
}
/* end of file */