package de.desy.acop.displayers.tarantula;

import java.awt.Color;
import java.awt.Component;
import java.text.DateFormat;
import java.util.logging.Logger;

import javax.swing.BorderFactory;
import javax.swing.Icon;
import javax.swing.JTree;
import javax.swing.border.Border;
import javax.swing.tree.DefaultTreeCellRenderer;
import javax.swing.tree.TreePath;

/**
 * Connection renderer. Renders the node with an internal label and draws its
 * content in the surrounding component. Currently it will only draw an instance
 * of {@link NodeConnection} or {@link NodeRoot}, where each of the node returns
 * the current status, for example for the instance of {@code NodeRoot} their
 * exists {@link NodeRoot#getStatus()}. All other instances will fail and print
 * a corresponding warning on the logging console.
 * 
 * @author <a href="mailto:marcus.walla(at)desy.de">walla</a>
 * @version $Id: ConnectionRenderer.java 40598 2010-12-01 16:53:36Z walla $
 */
final class ConnectionRenderer extends DefaultTreeCellRenderer {
	// --- CLASS FIELDS --------------------------------------------------------
	
	private static final Logger log = Utility.getLogger(ConnectionRenderer.class);
	private static final long serialVersionUID = -827804325249028756L;
	private static final DateFormat DATE_FORMAT = DateFormat.getTimeInstance(DateFormat.LONG);
	
	private final static String STR_ROOT_WAIT = "Please wait! Refreshing the connected links...";
	private final static String STR_ROOT_UPDATE_FINISHED = "Application Connections / last update on ";
	private final static String STR_ROOT_PRESS_BUTTON_TO_REFRESH = "Press the button \"" + TarantulaPanel.STR_START_REFRESH + "\" to start refreshing!";
	private final static String STR_ROOT_WAIT_FOR_REFRESH = "The connection structure will immediately be updated!";
	private final static String STR_ROOT_TIME_EXCEEDED = "Loop time is exceeded - try to set a higher periode in '" + TarantulaPanel.STR_MAX_LOOP_TIME + "'!";
	private final static String STR_ROOT_ONLY_FAILED_CONNECTION = "Failed connections / last update on ";
	private final static String STR_ROOT_NO_FAILED_CONNECTION = "Non of the connections failed / last update on ";
	
	private final static String STR_LINK_OK = "Link OK - address: ";
	private final static String STR_LINK_FAILURE = "Link failure - address: <b>%s</b>, error: <b>%s</b> (depth %d)";
	private final static String STR_UNKNOWN_CONNECTION_STATUS = "UNKNOWN STATUS: <b>%s</b> - server: <b>%s</b> ( %d)";
	private final static String STR_UNKNOWN_ROOT_STATUS = "UNKNOWN ROOT: <b>%s</b>";
	private final static Border LINE_BORDER = BorderFactory.createCompoundBorder(BorderFactory.createLineBorder(Color.BLACK), BorderFactory.createEmptyBorder(5, 5, 5, 5));
	
	// --- INSTANCE METHODS ----------------------------------------------------
	
	@Override public Component getTreeCellRendererComponent(
			final JTree tree,
	        final Object value,
	        final boolean selected,
	        final boolean expanded,
	        final boolean leaf,
	        final int row,
	        final boolean hasFocus) {
		// --- ATTENTION --- ATTENTION --- ATTENTION --- ATTENTION --- ATTENTION
		// It seem to be important also to call the super method on the renderer
		// to draw the label with its appropriated text field, also in HTML 
		// style.
		// Without the super call the label can be curious in its drawing behavior.
		super.getTreeCellRendererComponent(
		                tree, value, selected,
		                expanded, leaf, row,
		                hasFocus);
		// --- ATTENTION --- ATTENTION --- ATTENTION --- ATTENTION --- ATTENTION
		
		final boolean isEnabled = tree.isEnabled();
		
		if (value instanceof NodeConnection) {
			final NodeConnection node = (NodeConnection) value;
			
			switch(node.getStatus()) {
			case SUCCESS: { // display as OK
				if (node.getChildCount() == 0)
					super.setText(STR_LINK_OK + node.getFullAddress() + " (depth " + node.getLevel() + ')');
				else if (node.getChildCount() == 1)
					super.setText(STR_LINK_OK + node.getFullAddress() + " (include one connection, depth " + node.getLevel() + ')');
				else
					super.setText(STR_LINK_OK + node.getFullAddress() + " (include " + node.getChildCount() + " connections, depth " + node.getLevel() + ')');
				super.setToolTipText(Utility.getInfo(node));
				this.setIconImpl(Utility.getStatusIcon(node.getStatus(), node.getCondition(), isEnabled), isEnabled);
				super.setBorder(null);
				break;
			}
			
			case PENDING: { // display as a warning
				super.setText("Link is pending - " + node.getFullAddress() + " (depth " + node.getLevel() + ')');
				super.setToolTipText(Utility.getInfo(node));
				this.setIconImpl(Utility.getStatusIcon(node.getStatus(), node.getCondition(), isEnabled), isEnabled);
				super.setBorder(null);
				break;
			}
			
			case ASCENDING_WARNING: { // display a ascending warning which reflects an error in some child nodes
				if (node.getChildCount() == 0)
					super.setText(STR_LINK_OK + node.getFullAddress() + " (depth " + node.getLevel() + ')');
				else if (node.getChildCount() == 1)
					super.setText(STR_LINK_OK + node.getFullAddress() + " (include one connection, depth " + node.getLevel() + ')');
				else
					super.setText(STR_LINK_OK + node.getFullAddress() + " (include " + node.getChildCount() + " connections, depth " + node.getLevel() + ')');
				super.setToolTipText(Utility.getInfo(node));
				this.setIconImpl(Utility.getStatusIcon(node.getStatus(), isEnabled), isEnabled);
				super.setBorder(null);
				break;
			}
			
			case FAILURE: { // display as an error
				super.setText(TineConnection.includeHTMLTag(String.format(ConnectionRenderer.STR_LINK_FAILURE, 
						node.getFullAddress(), 
						node.getStatusLink(),
						node.getLevel())));
				super.setToolTipText(Utility.getInfo(node));
				this.setIconImpl(Utility.getStatusIcon(node.getStatus(), node.getCondition(), isEnabled), isEnabled);
				super.setBorder(null);
				break;
			}
			
			case FAILURE_BLACKLISTED: { // display as an error
				super.setText("No connections delivered - " + node.getFullAddress() + " (depth " + node.getLevel() + ')');
				super.setToolTipText(Utility.getInfo(node));
				this.setIconImpl(Utility.getStatusIcon(node.getStatus(), node.getCondition(), isEnabled), isEnabled);
				super.setBorder(null);
				break;
			}
			
			case FAILURE_UNRESOLVED_ADDRESS: { // display as an error
				super.setText("Unresolved address - " + node.getFullAddress() + " (depth " + node.getLevel() + ')');
				super.setToolTipText(Utility.getInfo(node));
				this.setIconImpl(Utility.getStatusIcon(node.getStatus(), node.getCondition(), isEnabled), isEnabled);
				super.setBorder(null);
				break;
			}
			
			case FAILURE_INVALID_DATA_REFERENCE: { // display as an error
				super.setText("Invalid data reference - " + node.getFullAddress() + " (depth " + node.getLevel() + ')');
				super.setToolTipText(Utility.getInfo(node));
				this.setIconImpl(Utility.getStatusIcon(node.getStatus(), node.getCondition(), isEnabled), isEnabled);
				super.setBorder(null);
				break;
			}
			
			case FAILURE_ILLEGAL_PROPERTY: { // display as an error
				super.setText("No connections delivered - " + node.getFullAddress() + " (depth " + node.getLevel() + ')');
				super.setToolTipText(Utility.getInfo(node));
				this.setIconImpl(Utility.getStatusIcon(node.getStatus(), node.getCondition(), isEnabled), isEnabled);
				super.setBorder(null);
				break;
			}
			
			case STATUS_UNKNOWN: {
				super.setText("Unknown state - " + node.getFullAddress() + " (depth " + node.getLevel() + ')');
				super.setToolTipText(Utility.getInfo(node));
				this.setIconImpl(Utility.getStatusIcon(node.getStatus(), node.getCondition(), isEnabled), isEnabled);
				super.setBorder(null);
				break;
			}
			
			default: {
				super.setText(TineConnection.includeHTMLTag(String.format(ConnectionRenderer.STR_UNKNOWN_CONNECTION_STATUS,
						node.getStatus(), 
						node.getFullAddress(),
						node.getLevel())));
				super.setToolTipText(TineConnection.includeHTMLTag(node.getInfoString()));
				this.setIconImpl(Utility.getStatusIcon(node.getStatus(), node.getCondition(), isEnabled), isEnabled);
				super.setBorder(LINE_BORDER);
			}
			}
		}
		else if (value instanceof NodeRoot)
		{
			final NodeRoot node = (NodeRoot) value;
			switch(node.getStatus()) {
			
			case ROOT_REFRESH_FINISHED: { // updating is finished
				final TreePath path = new TreePath(node);
				if (node.getChildCount() == 0)
					super.setText(ConnectionRenderer.STR_ROOT_PRESS_BUTTON_TO_REFRESH);
				else if (tree.isCollapsed(path))
					super.setText("DOUBLE CLICK TO EXPAND! - " + 
							ConnectionRenderer.STR_ROOT_UPDATE_FINISHED + 
							ConnectionRenderer.DATE_FORMAT.format(node.getUpdateTime()));
				else
					super.setText(ConnectionRenderer.STR_ROOT_UPDATE_FINISHED + 
							ConnectionRenderer.DATE_FORMAT.format(node.getUpdateTime()));
				super.setSize(16, 600);
				super.setToolTipText(null);
				this.setIconImpl(Utility.getRootIcon(node.getStatus(), isEnabled), isEnabled);
				super.setBorder(null);
				break;
			}
				
			case ROOT_REFRESH_CONNECTIONS: { // updating connections {
				// ATTENTION: needs to add some spaces at the end to perform the
				// size of the inherited label to the right width otherwise the
				// text of updating finished won't fit in the label.
				super.setText(ConnectionRenderer.STR_ROOT_WAIT + "                                ");
				super.setToolTipText(null);
				this.setIconImpl(Utility.getRootIcon(node.getStatus(), isEnabled), isEnabled);
				super.setBorder(null);
				break;
			}
			
			case ROOT_REFRESH_PRESS_BUTTON_TO_REFRESH: {
				// ATTENTION: needs to add some spaces at the end to perform the
				// size of the inherited label to the right width otherwise the
				// text of updating finished won't fit in the label.
				super.setText(ConnectionRenderer.STR_ROOT_PRESS_BUTTON_TO_REFRESH + "                                ");
				super.setToolTipText(null);
				this.setIconImpl(Utility.getRootIcon(node.getStatus(), isEnabled), isEnabled);
				super.setBorder(null);
				break;
			}
			
			case ROOT_REFRESH_WAIT_FOR_REFRESH: {
				// ATTENTION: needs to add some spaces at the end to perform the
				// size of the inherited label to the right width otherwise the
				// text of updating finished won't fit in the label.
				super.setText(ConnectionRenderer.STR_ROOT_WAIT_FOR_REFRESH + "                                ");
				super.setToolTipText(null);
				this.setIconImpl(Utility.getRootIcon(node.getStatus(), isEnabled), isEnabled);
				super.setBorder(null);
				break;
			}
			case ROOT_TIME_EXCEEDED: {
				super.setText(ConnectionRenderer.STR_ROOT_TIME_EXCEEDED + "                                ");
				super.setToolTipText(null);
				this.setIconImpl(Utility.getRootIcon(node.getStatus(), isEnabled), isEnabled);
				super.setBorder(null);
				break;
			}
			case ROOT_ONLY_FAILED_CONNECTIONS: {
				super.setText(ConnectionRenderer.STR_ROOT_ONLY_FAILED_CONNECTION + 
						ConnectionRenderer.DATE_FORMAT.format(node.getUpdateTime()));
				super.setToolTipText(null);
				this.setIconImpl(Utility.getRootIcon(node.getStatus(), isEnabled), isEnabled);
				super.setBorder(null);
				break;
			}
			case ROOT_NON_CONNECTION_FAILED: {
				super.setText(ConnectionRenderer.STR_ROOT_NO_FAILED_CONNECTION  + 
						ConnectionRenderer.DATE_FORMAT.format(node.getUpdateTime()));
				super.setToolTipText(null);
				this.setIconImpl(Utility.getRootIcon(node.getStatus(), isEnabled), isEnabled);
				super.setBorder(null);
				break;
			}
			
			default:
			{
				super.setText(TineConnection.includeHTMLTag(String.format(STR_UNKNOWN_ROOT_STATUS,
						node.getStatus())));
				super.setToolTipText(null);
				super.setIcon(Utility.getStatusIcon(EStatusLink.INTERNAL_ERROR, tree.isEnabled()));
				super.setBorder(LINE_BORDER);
			}
			}
		}
		else
		{
			super.setText("UNKNOWN INSTANCE: " + value.getClass().getSimpleName());
			this.setIconImpl(Utility.getStatusIcon(EStatusLink.INTERNAL_ERROR, isEnabled), isEnabled);
			super.setToolTipText(null);
			log.warning("updating tree cell on unknown instance of " + value.getClass().getSimpleName());
		}
		
		return this;
	}

	private void setIconImpl(final Icon icon, final boolean isEnabled) {
		super.setEnabled(isEnabled);
		if (isEnabled)
			super.setIcon(icon);
		else
			super.setDisabledIcon(icon);
	}
}
/* end of file */