/**
 * File: NodeConnection.java
 * Date: Jul 23, 2010
 */
package de.desy.acop.displayers.tarantula;

import java.util.Collection;
import java.util.logging.Logger;

import de.desy.tine.client.TLink;
import de.desy.tine.definitions.TErrorList;

/**
 * Connection node.
 * 
 * @author <a href="mailto:marcus.walla(at)desy.de">walla</a>
 * @version $Id: NodeConnection.java 40565 2010-11-30 13:40:10Z walla $
 */
final class NodeConnection extends ANode {

	// --- CLASS FIELDS --------------------------------------------------------
	
	@SuppressWarnings("unused")
	private static final Logger log = Utility.getLogger(NodeConnection.class);
	private static final long serialVersionUID = 5436479509057431071L;
	
    // --- INSTANCE FIELDS -----------------------------------------------------
    
    private final String context;
    private final String server;
    private final String deviceName;
    private final String propertyName;
    private final String fullAddress;
    private final String strInfo;
    private final ConnectionInfo info;
    private int linkStatus = 0;
    protected EStatusLink status;
    private ENodeCondition condition;
	private String strStatus = "UNKNOWN";
	
    // --- CONSTRUCTOR ---------------------------------------------------------
	
	/**
	 * <p>Constructor. Initialize this instance with a TINE link object.</p>
	 * <p><b>Remark</b>: This constructor will possibly removed in future 
	 * version to make the whole tarantula framework independent from the 
	 * current used connection library.</p>
	 * @param The initial TINE link.
	 */
    NodeConnection(final TLink lnk) {
    	this(lnk.getContext(), lnk.getDeviceServer(), lnk.getDeviceName(), lnk.getProperty(), TineConnection.getConnectionInfo(lnk.getContext(), lnk.getDeviceServer()));
    }
    
    /**
     * <p>Constructor.</p>
     * @param node Node to copy.
     */
    NodeConnection(final NodeConnection node) {
    	this(node.getContext(), node.getServer(), node.getDeviceName(), node.getPropertyName(), node.getInfo());
    	this.setStatusLinkCode(node.getStatusLinkCode());
    	this.setCondition(node.getCondition());
    }
	
    /**
	 * <p>Constructor. Initializes this instance with all parts of the full TINE
	 * address.</p>
	 * <p><b>Remark</b>: This constructor will possibly removed in future 
	 * version to make the whole tarantula framework independent from the 
	 * current used connection library.</p>
     * 
     * @param context Context
     * @param server Server
     * @param deviceName Device name
     * @param propertyName Property name
     * @param info Connection information.
     */
    NodeConnection(
    		final String context, 
    		final String server,
    		final String deviceName,
    		final String propertyName,
    		final ConnectionInfo info) {
    	this.context = context;
    	this.server = server;
    	this.deviceName = deviceName;
    	this.propertyName = propertyName;
   		this.fullAddress = TineConnection.getFullAddress(context, server, deviceName, propertyName);
    	this.status = EStatusLink.STATUS_UNKNOWN;
    	this.info = info;
        this.strInfo = getInfo(this.info);
        this.condition = ENodeCondition.ORDINARY_NODE;
    }

    // --- INSTANCE METHODS ----------------------------------------------------
    
    private String getInfo(final ConnectionInfo info) { 
    	final String strInfo = 
    		"<table border=\"0\" cellpadding=\"1\">" 
        	+ "<tr bgcolor=\"#E0E0E0\"><td>" + EConnectionInfoMeta.DESCRIPTION.getDescription() + ":</td><td>" + info.getDescription() + "</td></tr>" 
        	+ "<tr bgcolor=\"#A0A0A0\"><td>" + EConnectionInfoMeta.RESPONSIBLE.getDescription() + ":</td><td>" + info.getResponsible() + "</td></tr>" 
        	+ "<tr bgcolor=\"#E0E0E0\"><td>" + EConnectionInfoMeta.LOCATION.getDescription()    + ":</td><td>" + info.getLocation()    + "</td></tr>" 
        	+ "<tr bgcolor=\"#A0A0A0\"><td>" + EConnectionInfoMeta.FEC_NAME.getDescription()    + ":</td><td>" + info.getFecName()     + "</td></tr>" 
        	+ "<tr bgcolor=\"#E0E0E0\"><td>" + EConnectionInfoMeta.EQM_NAME.getDescription()    + ":</td><td>" + info.getEqmName()     + "</td></tr>" 
        	+ "<tr bgcolor=\"#A0A0A0\"><td>" + EConnectionInfoMeta.OS.getDescription()          + ":</td><td>" + info.getOs()          + "</td></tr>" 
        	+ "<tr bgcolor=\"#E0E0E0\"><td>" + EConnectionInfoMeta.VERSION.getDescription()     + ":</td><td>" + info.getTineVersion() + "</td></tr>" 
        	+ "<tr bgcolor=\"#A0A0A0\"><td>" + EConnectionInfoMeta.HARDWARE.getDescription()    + ":</td><td>" + info.getHardware()    + "</td></tr>" 
        	+ "<tr bgcolor=\"#E0E0E0\"><td>" + EConnectionInfoMeta.HOST_NAME.getDescription()   + ":</td><td>" + info.getHostName()    + " (" +  info.getIpAddress() + ")</td></tr>" 
        	+ "<tr bgcolor=\"#A0A0A0\"><td>" + EConnectionInfoMeta.PORT_OFFSET.getDescription() + ":</td><td>" + info.getPortOffset()  + "</td></tr>" 
        	+ "</table>";
        return strInfo; 
    }

    // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
    
    /**
     * Fetches the child node depending on the full address information. If non
     * is equal to the to this address a {@code null} is returned otherwise the
     * node.
     * @param fullAddress Full address information. Can also be {@code null}
     * @return Return the corresponding node.
     */
    final NodeConnection getChildNode(final String fullAddress) {
    	final Collection<NodeConnection> nodes = this.getChildNodes();
    	for(final NodeConnection node: nodes)
    		if (node.getFullAddress().equalsIgnoreCase(fullAddress))
    			return node;
    	return null;
    }
    
    final String                    getContext()          { return this.context; }
    final String                    getServer()           { return this.server; }
    final String                    getDeviceName()       { return this.deviceName; }
    final String                    getPropertyName()     { return this.propertyName; }
    final EStatusLink               getStatus()           { return this.status; }
    @Override final ENodeCondition  getCondition()        { return this.condition; }
    @Override final String          getInfoString()       { return this.strInfo; }
    @Override final ConnectionInfo  getInfo()             { return this.info; }
    @Override final String          getFullAddress()      { return this.fullAddress; }
    /** The return string is never {@code null}. */
    @Override final String          getStatusLink()     { return this.strStatus; }
    @Override final int             getStatusLinkCode() { return this.linkStatus; }
    @Override final void            setCondition(final ENodeCondition condition) { this.condition = condition; }
    
    // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 

    boolean containsChild(final String fullAddress) {
    	final Collection<NodeConnection> nodes = this.getChildNodes();
    	for(final NodeConnection node: nodes)
    		if (node.getFullAddress().equals(fullAddress))
    			return true;
    	return false;
    }
    
    // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
    
    final void setStatusLinkCode(final int code) {
    	this.linkStatus = code;
    	this.status = TineConnection.convertToStatus(code);
    	this.strStatus = TErrorList.getErrorString(code) + " (" + code + ")";
    }
    
    final void setAscendingWarning() {
		if (this.getStatus() != EStatusLink.ASCENDING_WARNING) {
	    	this.status = EStatusLink.ASCENDING_WARNING;
	    	if (super.getParent() instanceof NodeConnection)
	    		((NodeConnection) super.getParent()).setAscendingWarning();
		}
    }
    
    // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 

    @Override public String toString() {
   		return this.getFullAddress();
    }
}
/* end of file */