/**
 * File: TineConnection.java
 * Date: Sep 17, 2010
 */
package de.desy.acop.displayers.tarantula;

import java.net.InetAddress;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.Collections;
import java.util.HashSet;
import java.util.Set;
import java.util.logging.Logger;

import de.desy.tine.addrUtils.FECInfo;
import de.desy.tine.addrUtils.TSrvEntry;
import de.desy.tine.client.TLink;
import de.desy.tine.dataUtils.TDataType;
import de.desy.tine.definitions.TAccess;
import de.desy.tine.definitions.TErrorList;
import de.desy.tine.definitions.TMode;
import de.desy.tine.exceptions.BoundToInactiveLinkException;
import de.desy.tine.exceptions.InvalidDataReferenceException;
import de.desy.tine.exceptions.UnresolvedAddressException;
import de.desy.tine.queryUtils.TQuery;
import de.desy.tine.server.connections.TConnectionStruct;

/**
 * <p>All sorts of method for a TINE connection environment.</p>
 * 
 * <p>Moreover the developer can also set the current use of the reflected TINE 
 * links, i.e. if it also should display links which are called only once but 
 * are initialized with a normal asynchronous one. This links have a special 
 * TINE mode called {@link TMode#CM_SINGLE}. This flag allows to distinguish 
 * between this special link behavior and the more general asynchronous 
 * links. The method to set the behavior of the mode is called 
 * {@link #setTINEMode(ETINEMode)}.</p>
 * 
 * @author <a href="mailto:marcus.walla(at)desy.de">walla</a>
 * @version $Id: TineConnection.java 82467 2018-08-22 13:37:17Z walla $
 */
final class TineConnection {

	// --- CLASS FIELDS --------------------------------------------------------
	
	private final static Logger log = Utility.getLogger(TineConnection.class);
	
	/**
	 * TINE mode for displaying different links depending on this mode. 
	 * Actually their exists two modes. One allows to include also asynchronous
	 * links which are used for only one synchronous connection, where the other
	 * only reflects all asynchronous links without the synchronous ones. 
	 * 
	 * @author <a href="mailto:marcus.walla(at)desy.de">walla</a>
	 * @version $Id: TineConnection.java 82467 2018-08-22 13:37:17Z walla $
	 */
	enum ETINEMode { 
		ASYNC_LINKS("Asynchronous"), 
		SYNC_AND_ASYNC_LINKS("Synchronous and Asynchronous"); 
	
		private final String title;
		
		ETINEMode(final String title) { this.title = title; }
		
		String getTitle() { return this.title; }
	}

	/** 
	 * <p>An important flag which controls if TINE modes should be taken under 
	 * control of the method {@link #isSingleLinkReflected(int)} or not. See 
	 * there for more information. If this flag is set to {@code true} then the 
	 * control will take place otherwise not. The setter method is here:
	 * {@link #setTINEMode(ETINEMode)}.</p>
	 */
	private static ETINEMode useTINEMode = ETINEMode.ASYNC_LINKS;
	
	/**
	 * Sets the current responsible TINE mode, where the displayed links depends
	 * on the used mode. See {@link ETINEMode} to get a more detailed 
	 * description of the different modes.
	 * @param tineMode Current TINE mode.
	 */
	static void setTINEMode(final ETINEMode tineMode) {
		TineConnection.useTINEMode = tineMode;
	}

	final static int IDX_CONTEXT  = 0;
	final static int IDX_SERVER   = 1;
	final static int IDX_DEVICE   = 2;
	final static int IDX_PROPERTY = 3;
	
	/**
	 * A set of servers which are ignored by the update task of the tarantula. 
	 * Currently this includes the following server names: 
	 * {@code ALARMSTATE} (possibly renamed by Phil, but included for saftey's sake),
	 * {@code ALMSTATE},
	 * {@code ARCHIVER},
	 * {@code CAS},
	 * {@code CSSPY},
	 * {@code EVENTS},
	 * {@code EVENTSTORE},
	 * {@code FECSTATS},
	 * {@code HISTORY},
	 * {@code ENS},
	 * {@code GENS},
	 * {@code CLOG},
	 * and {@code TIMESRV}.
	 */
	protected final static Set<String> IGNORE_SERVER = Collections.unmodifiableSet(new HashSet<>(
			Arrays.asList(
					"ALARMSTATE", // possibly renamed by Phil, but included for saftey's sake
					"ALMSTATE", 
					"ARCHIVER", 
					"CAS",
					"CSSPY",
					"EVENTS",
					"EVENTSTORE",
					"FECSTATS", 
					"HISTORY",
					
					"ENS",
					"GENS",
					"CLOG",
					"TIMESRV")
					));

	private static final Set<String> SPECIAL_CONTEXTS = new HashSet<>();
	private static final Set<String> VALID_CONTEXTS = new HashSet<>();

	static {
		final String[] tmpContexts = TQuery.getContexts();
		if (tmpContexts != null && tmpContexts.length != 0) {
			TineConnection.VALID_CONTEXTS.addAll(Arrays.asList(tmpContexts));
			/* FIXME Add here the special contexts which didn't need to be 
			 * controlled by the RefreshConnections class. 
			 * => So, I initialize here an empty list! */
			TineConnection.SPECIAL_CONTEXTS.addAll(Arrays.asList(new String[] {}));
		}
		else
			log.severe("Cannot find any TINE contexts -> return immediately with an empty set");

	}

	// --- CLASS METHODS -------------------------------------------------------

	/**
	 * Returns the connection information instance depending on the TINE context
	 * and server name (also known as the equipment name). Both are fetched from
	 * the full TINE address as the first two arguments.
	 * 
	 * @param fullAddress Full TINE address; at least including the context,
	 * 		server, and property name.
	 * @return Fetches via a TINE query and pushes the returned information into
	 * 		an instance of {@link ConnectionInfo}.
	 * @throws ConnectionException Throws a general connection exception if a 
	 * 		TINE error occurs.
	 * @see #getConnectionInfo(String, String)
	 */
	static ConnectionInfo getConnectionInfo(final String fullAddress) {
		final String[] address = TineConnection.parseFullAddress(fullAddress);
		if (address == null)
			return ConnectionInfo.UNKNOWN_CONNECTION_INFO;
		// return the FECInfo with a polymorphic method
		return TineConnection.getConnectionInfo(address[TineConnection.IDX_CONTEXT], address[TineConnection.IDX_SERVER]);
	}

	/**
	 * Returns the connection information instance depending on the TINE context
	 * and server name (also known as the equipment name).
	 * 
	 * @param context Context name.
	 * @param server Server name.
	 * @return Fetches via a TINE query and pushes the returned information into
	 * 		an instance of {@link ConnectionInfo}.
	 * @throws ConnectionException Throws a general connection exception if a 
	 * 		TINE error occurs.
	 */
	static ConnectionInfo getConnectionInfo(final String context, final String server) {
		try {
		    final TSrvEntry srvEntry = TQuery.getModuleAddress(server, context);
		    if (srvEntry != null) {
		    	final FECInfo info = TQuery.getServerInformation(srvEntry.fecName);
		    	final InetAddress ipAddress = srvEntry.fecAddr.fecHost;
		    	/*
				 * @param fecName FEC (Front End Computer) name
				 * @param eqmName Equipment name
				 * @param os Operation system
				 * @param responsible Who is responsible for the server
				 * @param location Where is the server located
				 * @param version Version of the server
				 * @param hostName Host name
				 * @param hostAddress Numerical host address
				 * @param portOffset Port offset
		    	 */
		        return new ConnectionInfo(
		        		srvEntry.fecName, 
		        		srvEntry.eqmName, 
		        		info.getOs(), 
		        		info.getResponsible(), 
		        		info.getLocation(), 
		        		info.getVersion(),
		        		info.getHardware(),
		        		info.getDescription(),
		        		ipAddress.getHostName(),
		        		ipAddress.getHostAddress(),
		        		srvEntry.fecAddr.fecPortOffset);
		    }
		}
	    catch(RuntimeException exception) {
	    	log.severe("error - server=" + '/' + context + '/' + server + ", exception=" + exception.getClass().getSimpleName());
	    }
	    
	    return ConnectionInfo.UNKNOWN_CONNECTION_INFO;
	}

	/**
	 * Returns all server connections represented by the client address. Can
	 * result with a TINE exception if invoking the appropriated methods fails.
	 * @param context Context
	 * @param server Server
	 * @return Collection of the fetched connection structure of TINE.
	 * @throws TineException Is thrown when achieving the appropriate result 
	 * 		fails.
	 */
	static Collection<TConnectionStruct> getConnections(
			final String context, 
			final String server) throws TineException {
	    
		final TConnectionStruct[] connections = new TConnectionStruct[50];
	    for(int i = 0; i < connections.length; i++) connections[i] = new TConnectionStruct();
	    
	    final Collection<TConnectionStruct> result = new ArrayList<>();
	    int completionLength = 0;
	    final String serverName = '/' + context + '/' + server;
	    try {
	        final TDataType outputType = new TDataType(connections);
	        final TLink lnk = new TLink('/' + context + '/' + server, 
	        		"CONNECTIONS", outputType, null, TAccess.CA_READ);
	        final int rc = lnk.execute(1000, true);
	        lnk.close();
	    
	        if (rc != 0) {
        		log.warning(TErrorList.getErrorString(rc) + " - " + serverName);

    	        // ATTENTION: also request if the executing is a blacklisted link
    	        // or an illegal property, which means that it is an older TINE 
    	        // version and hasn't the stock property of CONNECTIONS. In this 
    	        // case throw a TINE exception.
        		if (rc == TErrorList.illegal_property 
        				|| rc == TErrorList.link_blacklisted)
	        		throw new TineException(rc, '/' + context + '/' + server);
        		
        		completionLength = 0;
	        }
	        else if (outputType.getCompletionLength() != 0 
	        		&& TineConnection.hasErroneousContexts(connections)) {
	        	/* ATTENTION: TINE version 4.1.0 returns a faulty array by 
	        	 * fetching the connection values. To avoid this erroneous data 
	        	 * the first element of the connection is queried if the output 
	        	 * type returns one or more elements and the context is null or 
	        	 * empty. Under these conditions a TINE error of illegal 
	        	 * property is thrown, because this error is established with a 
	        	 * predefined icon in the UI context. */
	        	final int predefinedError = TErrorList.illegal_property;
        		log.warning("Faulty TINE version which fetches erroneous connections on server " + serverName + " -> throw a TINE exception with the error " + TErrorList.getErrorString(predefinedError));
        		throw new TineException(predefinedError, '/' + context + '/' + server);
	        }
	        else
	            completionLength = outputType.getCompletionLength();
	    }
	    catch(UnresolvedAddressException exception) {
	        // unresolved TINE address
    		log.warning(TErrorList.getErrorString(TErrorList.address_unresolved) + " - " + serverName);
    		completionLength = 0;
	    }
	    catch(InvalidDataReferenceException exception) {
	        // invalid data reference -> should never be thrown!
    		log.warning(TErrorList.getErrorString(TErrorList.invalid_reference) + " - " + serverName);
    		completionLength = 0;
	    }
	    catch(BoundToInactiveLinkException exception) {
	    	// link is bound to an inactive link
	    	log.warning(TErrorList.getErrorString(TErrorList.link_exists) + " - link is bound to an inactive link - " + serverName);
	    	completionLength = 0;
	    }
	    catch(RuntimeException exception) {
	    	log.severe("error - server=" + '/' + context + '/' + server + ", exception=" + exception.getClass().getSimpleName());
	    	completionLength = 0;
	    }
	    finally {
	        for(int i = 0; i < completionLength; i++) 
	        	if (!isSpecialConnection(connections[i]))
	        		result.add(connections[i]);
	    }
        return Collections.unmodifiableCollection(result);
	}

	/** 
	 * Examines the delivered TINE connection structure if one of the context
	 * contains nothing (i.e. is {@code null}), is empty, or contains no letter,
	 * digit, # or underline characters. In this case a {@code true} is returned
	 * otherwise {@code false}. 
	 * 
	 * @param connections TINE connection structure array.
	 * @return Depending on the result returns a boolean value. 
	 */
	private static boolean hasErroneousContexts(final TConnectionStruct[] connections) {
		if (connections != null && connections.length > 0 && connections[0] != null)
			return !TineConnection.VALID_CONTEXTS.contains(connections[0].getContext());
		return true;
	}
	
	/**
	 * Reflects if the connection is realized with a special link mode or the
	 * address has a special meaning. In both cases this method returns a 
	 * {@code true} otherwise {@code false}.
	 * 
	 * @param connection TINE Connection structure.
	 * @param useTINEModes Used TINE mode.
	 * @return Returns a boolean value depending if the connection reflects a
	 * 		special meaning.
	 */
	static boolean isSpecialConnection(final TConnectionStruct connection) {
		return isSpecialConnection(connection.getContext(), connection.getMode());
	}
	
	/**
	 * Reflects if the connection is realized with a special link mode or the
	 * address has a special meaning. In both cases this method returns a 
	 * {@code true} otherwise {@code false}.
	 * @param lnk TINE link.
	 * @return Returns a boolean value depending if the connection reflects a
	 * 		special meaning.
	 */
	static boolean isSpecialConnection(final TLink lnk) {
		return isSpecialConnection(lnk.getContext(), lnk.getLinkMode());
	}
	
	private static boolean isSpecialConnection(
			final String context, 
			final int mode) {
		// ATTENTION: will always skip because currently their doesn't 
		// exist any special context!
		if (TineConnection.containsSpecialAddress(context))
			return true;
		else if (TineConnection.isSingleLinkReflected(mode, TineConnection.useTINEMode))
			return true;
		return false;
	}
	
	/** 
	 * Request if the TINE mode of the link is reflected with a single link
	 * connection and therefore shown in the structure or not. This depends on 
	 * the current value of the used {@link ETINEMode#SYNC_AND_ASYNC_LINKS} and 
	 * if it is marked with {@link TMode#CM_SINGLE}. If one of both returns a 
	 * {@code true} the method returns a {@code true} otherwise {@code false}. 
	 * @param mode TINE mode value.
	 * @return If the corresponding link reflects a single link connection.
	 */
	private static boolean isSingleLinkReflected(final int mode, final ETINEMode useTINEModes) {
		// FIXME Possibly this is a mistake to connect both values with a
		// logical and operation. 
		return useTINEModes == ETINEMode.SYNC_AND_ASYNC_LINKS || mode == TMode.CM_SINGLE;
	}

	/**
	 * Checks if the node is a TINE failure.
	 * @param node Node.
	 * @return Depending on the failure state of the node returns a {@code true}
	 * 		if it is a failure otherwise {@code false}.
	 */
	static boolean isNodeFailure(final NodeConnection node) {
		return isNodeFailure(node.getStatus());
	}
	
	/**
	 * Checks if the status is a TINE failure.
	 * @param status Status enumeration of the TINE link.
	 * @return Depending on the failure state of the node returns a {@code true}
	 * 		if it is a failure otherwise {@code false}.
	 */
	static boolean isNodeFailure(final EStatusLink status) {
		if (status == EStatusLink.FAILURE
				|| status == EStatusLink.FAILURE_BLACKLISTED
				|| status == EStatusLink.FAILURE_ILLEGAL_PROPERTY
				|| status == EStatusLink.FAILURE_INVALID_DATA_REFERENCE
				|| status == EStatusLink.FAILURE_UNRESOLVED_ADDRESS
				|| status == EStatusLink.PENDING)
			return true;
		else
			return false;
	}

	/**
	 * <p>Checks if the context and/or server name is a special address.</p>
	 * 
	 * <p><b>FIXME</b>: Currently this is initialized with an empty list and 
	 * therefore will always return a {@code false}. If there exists different
	 * contexts you should add them to the internal set of contexts!.</p>
	 * @param fullAddress Full address
	 * @return Depending on the full address returns a {@code true} if it is
	 * 		a special address otherwise {@code false}.
	 */
	static boolean containsSpecialAddress(final String fullAddress) {
		return TineConnection.SPECIAL_CONTEXTS.contains(getContext(fullAddress).toUpperCase());
	}

	private static String[] parseFullAddress(final String address) {
		if (address == null || address.isEmpty())
			return null;

		final String upperServerAddress = address.toUpperCase();
		final int startDelimiter = upperServerAddress.indexOf('/');
		if (startDelimiter != -1)
			return address.substring(startDelimiter + 1).split("/");
		
		return null;
	}
	
	private static String getContext(final String fullAddress) {
		if (fullAddress == null || fullAddress.isEmpty())
			return Utility.EMPTY_STRING;
		final int startDelimiter = fullAddress.indexOf('/');
		if (startDelimiter == -1)
			return fullAddress;
		final int endDelimiter = fullAddress.indexOf('/', startDelimiter + 1);
		if (endDelimiter == -1)
			return fullAddress.substring(startDelimiter);
		return fullAddress.substring(startDelimiter + 1, endDelimiter);
	}

	/**
	 * Returns the full qualified TINE address from the {@link TLink} instance.
	 * This instance can also be {@code null} in which case an empty string is
	 * returned.
	 * 
	 * @param lnk Link instance.
	 * @return Full qualified TINE address including the context, server, 
	 * 		device, and property name and the slash as a delimiter.
	 * @see #getFullAddress(String, String, String, String)
	 */
	static String getFullAddress(final TLink lnk) {
		if (lnk == null)
			return "";
		return TineConnection.getFullAddress(lnk.getContext(), lnk.getDeviceServer(), lnk.getDeviceName(), lnk.getProperty());
	}

	/**
	 * <p>Returns the full qualified TINE address from the TINE connection 
	 * structure instance (is a self declared TINE data type in the TINE library
	 * itself). This includes all needed information to create the full 
	 * address.</p>
	 * <p>The TINE connection structure can also be {@code null} in which cases
	 * this method just returns an empty string instance.</p>
	 * 
	 * @param connection TINE connection structure.
	 * @return Full qualified TINE address including the context, server, 
	 * 		device, and property name and the slash as a delimiter.
	 * @see #getFullAddress(String, String, String, String)
	 */
	static String getFullAddress(final TConnectionStruct connection) {
		if (connection == null)
			return "";
		return TineConnection.getFullAddress(connection.getContext(), connection.getExpName(), connection.getDevName(), connection.getPrpName());
	}

	/**
	 * Returns the full qualified TINE address from the string instances. The
	 * device name can also be {@code null}, empty, or starting with a hash 
	 * sign, in which cases it is transformed to a null device, i.e. &quot;#0&quot;.
	 * 
	 * @param context Context.
	 * @param server Server name.
	 * @param deviceName Device name.
	 * @param propertyName Property name.
	 * @return Full qualified TINE address including the context, server, 
	 * 		device, and property name and the slash as a delimiter.
	 */
	static String getFullAddress(final String context, final String server, final String deviceName, final String propertyName) {
		if (deviceName == null || deviceName.isEmpty())
			return new StringBuilder().append('/').append(context).append('/').append(server).append("/#0/").append(propertyName).toString();	
		else
			return new StringBuilder().append('/').append(context).append('/').append(server).append('/').append(deviceName).append('/').append(propertyName).toString();	
	}

	/**
	 * From the internally created list of ignored TINE servers a boolean value
	 * is returned, where a {@code true} means that this server should be 
	 * discounted.
	 * @param server Server name.
	 * @return Returns a {@code true} if the server should be ignored otherwise
	 * 		a {@code false}.
	 */
	static boolean ignoreServer(final String server) { return TineConnection.IGNORE_SERVER.contains(server); }

	/**
	 * Surrounds the string instance with root HTML tag structure, so that it
	 * can be displayed in a label with HTML style.
	 * @param html Unfinished HTML styled text.
	 * @return Finished HTML styled text, enclosed with the root HTML tag.
	 */
	static String includeHTMLTag(final String html) {
		return new StringBuilder().append("<html>").append(html).append("</html>").toString();
	}

	/**
	 * <p>Converts the link status into the status enumeration. The link status 
	 * is a TINE code.</p>
	 * 
	 * <p>The link status information are take from the enumeration {@link EStatusLink}. 
	 * See there for an updated documentation of status information.</p>
	 * 
	 * <table cellpadding="4" border="0">
	 * <tr><td>{@link EStatusLink#SUCCESS}</td><td>Links are established and connections are OK.</td></tr>
	 * <tr><td>{@link EStatusLink#FAILURE}</td><td>Links are established and connections failed.</td></tr>
	 * <tr><td>{@link EStatusLink#FAILURE_BLACKLISTED}</td><td>Links aren't established because of a blacklisted link (link doesn't exist).</td></tr>
	 * <tr><td>{@link EStatusLink#FAILURE_UNRESOLVED_ADDRESS}</td><td>Links aren't established because of an unresolved address (address doesn't exist).</td></tr>
	 * <tr><td>{@link EStatusLink#FAILURE_INVALID_DATA_REFERENCE}</td><td>Links aren't established because of an invalid data reference (provided data on link differs).</td></tr>
	 * <tr><td>{@link EStatusLink#FAILURE_ILLEGAL_PROPERTY}</td><td>Links aren't established because of an invalid property. This can be hint that the stock property of CONNECTION don't exist.</td></tr>
	 * <tr><td>{@link EStatusLink#PENDING}</td><td>Links is pending (no TINE link status is returned due to a pending (German: schwebend} link).</td></tr>
	 * <tr><td>{@link EStatusLink#STATUS_UNKNOWN}</td><td>Link is established but the status is currently unknown.<b>Is currently not reflected by this method!</b></td></tr>
	 * <tr><td>{@link EStatusLink#EStatusLink#ASCENDING_WARNING}</td>
	 *   <td>Ascending warning, reflecting that one of the descending connections 
	 *	 returns a failure code (i.e. a TINE code number unequal to 0).  This 
	 * 	 enumeration is equal to that one from {@link EStatusRoot#ASCENDING_WARNING}.</td>
	 *   </tr>
	 * <tr><td>{@link EStatusLink#INTERNAL_ERROR}</td><td>Internal error occurred.</td></tr>
	 * </table>
	 * 
	 * @param lnkStatus Integer value of the TINE status code.
	 * @return Returns the transformed enumeration entry.
	 */
	static EStatusLink convertToStatus(final int lnkStatus) {
		switch(lnkStatus) {
		case TErrorList.success:
			return EStatusLink.SUCCESS;
		case TErrorList.link_blacklisted:
			return EStatusLink.FAILURE_BLACKLISTED;
		case TErrorList.address_unresolved:
			return EStatusLink.FAILURE_UNRESOLVED_ADDRESS;
		case TErrorList.invalid_reference:
			return EStatusLink.FAILURE_INVALID_DATA_REFERENCE;
		case TErrorList.illegal_property:
			return EStatusLink.FAILURE_ILLEGAL_PROPERTY;
		case TErrorList.code_failure:
			return EStatusLink.INTERNAL_ERROR;
		case -1:
			return EStatusLink.PENDING;
		default:
			return EStatusLink.FAILURE;
		}
	}
}
/* end of file */