/*
 * Copyright (c) 2006 Stiftung Deutsches Elektronen-Synchroton,
 * Member of the Helmholtz Association, (DESY), HAMBURG, GERMANY.
 *
 * THIS SOFTWARE IS PROVIDED UNDER THIS LICENSE ON AN "../AS IS" BASIS.
 * WITHOUT WARRANTY OF ANY KIND, EXPRESSED OR IMPLIED, INCLUDING BUT NOT LIMITED
 * TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR PARTICULAR PURPOSE AND
 * NON-INFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE
 * FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
 * TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR
 * THE USE OR OTHER DEALINGS IN THE SOFTWARE. SHOULD THE SOFTWARE PROVE DEFECTIVE
 * IN ANY RESPECT, THE USER ASSUMES THE COST OF ANY NECESSARY SERVICING, REPAIR OR
 * CORRECTION. THIS DISCLAIMER OF WARRANTY CONSTITUTES AN ESSENTIAL PART OF THIS LICENSE.
 * NO USE OF ANY SOFTWARE IS AUTHORIZED HEREUNDER EXCEPT UNDER THIS DISCLAIMER.
 * DESY HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS,
 * OR MODIFICATIONS.
 * THE FULL LICENSE SPECIFYING FOR THE SOFTWARE THE REDISTRIBUTION, MODIFICATION,
 * USAGE AND OTHER RIGHTS AND OBLIGATIONS IS INCLUDED WITH THE DISTRIBUTION OF THIS
 * PROJECT IN THE FILE LICENSE.HTML. IF THE LICENSE IS NOT INCLUDED YOU MAY FIND A COPY
 * AT HTTP://WWW.DESY.DE/LEGAL/LICENSE.HTM
 */

package de.desy.acop.displayers.tools;

import java.util.Map;

import com.cosylab.gui.adapters.Converter;
import com.cosylab.gui.displayers.DisplayerParameters;
import com.cosylab.gui.displayers.DoubleDisplayer;

import de.desy.acop.transport.ConnectionParameters;

/**
 * <code>AcopDisplayerParameters</code> are parameters displayer parameters intended
 * to be used in combination with <code>AcopDisplayer</code>. This parameters holds
 * reference to the remote connection point as well as other parameters that
 * define the look and presentation of the value on the displayer, such as minimum and
 * maximum value, format etc.
 * 
 * @author <a href="mailto:jaka.bobnar@cosylab.com">Jaka Bobnar</a>
 * @see AcopDisplayer
 */
public class AcopDisplayerParameters extends DisplayerParameters {

	private static final long serialVersionUID = 2758958154017521990L;

	/**
	 * Constructs new AcopDisplayerParameters.
	 *
	 */
	public AcopDisplayerParameters() {
		this(null, 0, Double.NaN, Double.NaN, null, null);
	}
	
	/**
	 * Constructs new AcopDisplayerParameters with the provided connection point.
	 * 
	 * @param parameters remote connection point
	 */
	public AcopDisplayerParameters(ConnectionParameters parameters) {
		this(parameters, 0, Double.NaN, Double.NaN, null, null);
	}
	
	/**
	 * Constructs new AcopDisplayerParameters with the provided connection point
	 * and index in the array of incomming data.
	 * 
	 * @param parameters remote connectin point
	 * @param index array data index
	 */
	public AcopDisplayerParameters(ConnectionParameters parameters, int index) {
		this(parameters, index, Double.NaN, Double.NaN, null, null);
	}
	
	/**
	 * Constructs new AcopDisplayerParameters using the provided parameters.
	 * 
	 * @param parameters remote connection point
	 * @param index array data index
	 * @param min minimum value
	 * @param max maximum value
	 * @param units physical units
	 * @param format C-style format
	 */
	public AcopDisplayerParameters(ConnectionParameters parameters, int index, double min, double max, String units, String format) {
		this(parameters.getRemoteName(), 6, parameters, index, min, max, units, format);
	}
	
	/**
	 * Constructs new AcopDisplayerParameters using the provided parameters.
	 * 
	 * @param name the name of the parameters
	 * @param size initial number of all parameters
	 * @param parameters remote connection point
	 * @param index array data index
	 * @param min maximum value
	 * @param max minimum value
	 * @param units physical units
	 * @param format C-style format
	 */
	protected AcopDisplayerParameters(String name, int size, ConnectionParameters parameters, int index, double min, double max, String units, String format) {
		super(parameters!=null ? parameters.getRemoteName(): name, size, min, max, units, format);
		
		values.put(AcopDisplayer.CONNECTION_PARAMETERS_PROPERTY,parameters);
		values.put(AcopDisplayer.ARRAY_INDEX_PROPERTY,index);
	}

	/**
	 * Constructs new AcopDisplayerParameters gathering the essential data from
	 * the provided displayer.
	 * 
	 * @param name the name of the parameters
	 * @param size the number of all parameters
	 * @param adisp source for connection related parameters
	 * @param ddisp source for display related parameters
	 */
	protected AcopDisplayerParameters(String name, int size, AcopDisplayer adisp, DoubleDisplayer ddisp) {
		super(name, size, ddisp);
		
		values.put(AcopDisplayer.CONNECTION_PARAMETERS_PROPERTY,adisp.getConnectionParameters());
		values.put(AcopDisplayer.ARRAY_INDEX_PROPERTY,adisp.getArrayIndex());
	}
	
	/**
	 * Constructs new AcopDisplayerParameters gathering the essential data from
	 * the provided displayer.
	 * 
	 * @param adisp source for connection related parameters
	 * @param ddisp source for display related parameters
	 */
	public AcopDisplayerParameters(AcopDisplayer adisp, DoubleDisplayer ddisp) {
		this(adisp.getConnectionParameters() != null ? adisp.getConnectionParameters().getRemoteName() : null, 6, adisp, ddisp);
	}

	/**
	 * Constructs new AcopDisplayerParameters using the provided parameters.
	 * 
	 * @param parameters remote connection point
	 * @param index array data index
	 * @param converter converter associated with the data source
	 */
	public AcopDisplayerParameters(ConnectionParameters parameters, int index, Converter converter) {
		this(parameters, index, Double.NaN, Double.NaN, null, null, converter);
	}
	
	/**
	 * Constructs new AcopDisplayerParameters using the provided parameters.
	 * 
	 * @param parameters remote connection point
	 * @param index array data index
	 * @param min minimum value
	 * @param max maximum value
	 * @param units physical units
	 * @param format C-style format
	 * @param converter converter associated with the data source
	 */
	public AcopDisplayerParameters(ConnectionParameters parameters, int index, double min, double max, String units, String format, Converter converter) {
		this(parameters!=null ? parameters.getRemoteName() : "", 6, parameters, index, min, max, units, format, converter);
	}
	
	/**
	 * Contructs new AcopDisplayerParameters using the provided parameters.
	 * 
	 * @param name the name of the parameters
	 * @param size number of all parameters
	 * @param parameters remote connection point
	 * @param index array data index
	 * @param min minimum value
	 * @param max maximum value
	 * @param units physical units
	 * @param format C-style format
	 * @param converter converter associated with the data source
	 */
	protected AcopDisplayerParameters(String name, int size, ConnectionParameters parameters, int index, double min, double max, String units, String format, Converter converter) {
		this(name, size, parameters, index, min, max, units, format);
		values.put(AcopDisplayer.CONVERTER_PROPERTY, converter);
	}
	
	/**
	 * Returns the converter associated with this parameters.
	 * 
	 * @return the converter
	 */
	public Converter getConverter() {
		return (Converter) values.get(AcopDisplayer.CONVERTER_PROPERTY);
	}
	
	
	
	/**
	 * Creates new instance of AcopDisplayerParameters.
	 * 
	 * @param name the name of this parameters
	 * @param vaues additional parameters - all necessary parameters should be included
	 * 		in this map
	 */
	protected AcopDisplayerParameters(String name, Map<String,Object> values) {
		super(name, values);
	}

	/* (non-Javadoc)
	 * @see java.lang.Object#hashCode()
	 */
	@Override
	public int hashCode() {
		return toString().hashCode();
	}

	/* (non-Javadoc)
	 * @see java.lang.Comparable#compareTo(java.lang.Object)
	 */
	public int compareTo(Object o) {
		if (!(o instanceof AcopDisplayerParameters)) return -1;
	      return this.toString().compareTo(o.toString());
	}

	/**
	 * Returns the connection parameters.
	 * @return the remote connection point
	 */
	public ConnectionParameters getConnectionParameters() {
		return (ConnectionParameters)values.get(AcopDisplayer.CONNECTION_PARAMETERS_PROPERTY);
	}

	/**
	 * Returns the array index.
	 * 
	 * @return the arrayIndex
	 */
	public int getArrayIndex() {
		return getInt(AcopDisplayer.ARRAY_INDEX_PROPERTY);
	}

	/**
	 * Returns new AcopDisplayerParameters which resebles this object but has
	 * a different array index.
	 * 
	 * @param arrayIndex new array index
	 * @return new acop displayer parameters
	 */
	public AcopDisplayerParameters deriveWith(int arrayIndex) {
		AcopDisplayerParameters a=new AcopDisplayerParameters(name,values);
		values.put(AcopDisplayer.ARRAY_INDEX_PROPERTY,arrayIndex);
		return a;
	}

}
