/*
 * Copyright (c) 2006 Stiftung Deutsches Elektronen-Synchroton,
 * Member of the Helmholtz Association, (DESY), HAMBURG, GERMANY.
 *
 * THIS SOFTWARE IS PROVIDED UNDER THIS LICENSE ON AN "../AS IS" BASIS.
 * WITHOUT WARRANTY OF ANY KIND, EXPRESSED OR IMPLIED, INCLUDING BUT NOT LIMITED
 * TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR PARTICULAR PURPOSE AND
 * NON-INFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE
 * FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
 * TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR
 * THE USE OR OTHER DEALINGS IN THE SOFTWARE. SHOULD THE SOFTWARE PROVE DEFECTIVE
 * IN ANY RESPECT, THE USER ASSUMES THE COST OF ANY NECESSARY SERVICING, REPAIR OR
 * CORRECTION. THIS DISCLAIMER OF WARRANTY CONSTITUTES AN ESSENTIAL PART OF THIS LICENSE.
 * NO USE OF ANY SOFTWARE IS AUTHORIZED HEREUNDER EXCEPT UNDER THIS DISCLAIMER.
 * DESY HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS,
 * OR MODIFICATIONS.
 * THE FULL LICENSE SPECIFYING FOR THE SOFTWARE THE REDISTRIBUTION, MODIFICATION,
 * USAGE AND OTHER RIGHTS AND OBLIGATIONS IS INCLUDED WITH THE DISTRIBUTION OF THIS
 * PROJECT IN THE FILE LICENSE.HTML. IF THE LICENSE IS NOT INCLUDED YOU MAY FIND A COPY
 * AT HTTP://WWW.DESY.DE/LEGAL/LICENSE.HTM
 */

package de.desy.acop.displayers.tools;

import java.beans.PropertyVetoException;

import com.cosylab.gui.displayers.DataSource;
import com.cosylab.gui.displayers.MultipleDisplayer;
import com.cosylab.util.CommonException;

import de.desy.acop.transport.ConnectionParameters;
import de.desy.acop.transport.adapters.AdapterFactory;

/**
 * <code>AcopMultipleDisplayer</code> should be implemented by those displayers that
 * connects to multiple data sources. The class should accept multiple 
 * <code>AcopDisplayerParameters</code> and create <code>DataSource</code>s for 
 * those parameters and connect the DataSources with GUI. DataSources 
 * should be obtained via <code>AdapterFactory</code> and released in the manner 
 * prescribed by AdapterFactory. 
 * 
 * @author Jaka Bobnar, Cosylab
 * @see AcopDisplayerParameters
 * @see AdapterFactory
 */
public interface MultipleAcopDisplayer<T extends AcopDisplayerParameters> extends MultipleDisplayer {
	
	/** DisplayerParameters property name tag */
	public static final String DISPLAYER_PARAMETERS_PROPERTY = "displayerParameters";
	
	/**
	 * Returns all ConnectionParameters attached to this displayer. 
	 * 
	 * @return an array of all connection parameters
	 * @deprecated use {@link #getDisplayerParameters()}
	 */
	@Deprecated
	public ConnectionParameters[] getConnectionParameters();
	
	/**
	 * Adds ConnectionParameters to the displayer. Upon adding ConnectionParameters
	 * the displayer should create a DataSource and connect it to the remote channe
	 * as specified by the parameters. 
	 *   
	 * @param parameters 
	 * @throws CommonException if setting failed due to connection problem
	 * @throws PropertyVetoException if setting the property failed because of improper
	 * 				parameters
	 * @deprecated use {@link #addDisplayerParameters(AcopDisplayerParameters)}
	 */
	@Deprecated
	public void addConnectionParameters(ConnectionParameters parameters) throws CommonException, PropertyVetoException;
	
	/**
	 * Sets ConnectionParameter to the displayer. All previous connections
	 * should be terminated and then the new added. If the new array includes
	 * parameters that are already connected with this displayer, those connections
	 * do not need to be terminated. At the end of procedure the displayer should
	 * have only connections specified by the array <code>parameters</code>.
	 *  
	 * @param parameters remote connections points
	 * @throws CommonException if setting failed due to connection problem
	 * @throws PropertyVetoException if setting the property failed because of improper
	 * 				parameters
	 * @deprecated use {@link #setDisplayerParameters(AcopDisplayerParameters[])}
	 */
	@Deprecated
	public void setConnectionParameters(ConnectionParameters[] parameters) throws CommonException, PropertyVetoException;
	
	/**
	 * Adds <code>AcopDisplayerParameters</code> to this displayer. All previous 
	 * connections are kept and this one is added to the list. Upon adding the 
	 * displayer should connect to the remote point specified by this 
	 * <code>parameters</code> and adapt the screen according to other supplied 
	 * parameters. 
	 * 
	 * @param parameters remote connection point and screen parameters
	 * @throws CommonException if setting failed due to connection problem
	 * @throws PropertyVetoException if setting the property failed because of improper
	 * 				parameters
	 */
	public void addDisplayerParameters(T parameters) throws CommonException, PropertyVetoException;
	
	/**
	 * Returns an array of all <code>AcopDisplayerParameters</code> associated
	 * with this displayer. 
	 * 
	 * @return an array of all parameters
	 */
	public T[] getDisplayerParameters();
	
	/**
	 * Sets <code>AcopDisplayerParameters</code> to this displayer. All previous 
	 * connections should be terminated and then the new added. If the new array 
	 * includes parameters that are already connected with this displayer, those 
	 * connections do not need to be terminated. At the end of procedure the 
	 * displayer should have only connections specified by the array 
	 * <code>parameters</code>. Displayer should not allow two have two identical
	 * parameters. Parameters are identical if the 
	 * {@link AcopDisplayerParameters#equals(Object)} returns true.
	 * 
	 * @param parameters remote connections points
	 * @throws CommonException if setting failed due to connection problem
	 * @throws PropertyVetoException if setting the property failed because of improper
	 * 				parameters
	 */
	public void setDisplayerParameters(T[] parameters) throws CommonException, PropertyVetoException;
	
	/**
	 * Removes a <code>ConnectionParameters</code> from this displayer. DataSource that
	 * belongs to this parameters should be released (via AdapterFactory) and all
	 * resource associated with this parameters in the displayer should be released.
	 * 
	 * @param parameters connection to be removed
	 * @return DataSource associated with this parameters
	 * @deprecated use {@link #removeDisplayerParameters(AcopDisplayerParameters)}
	 */
	@Deprecated
	public DataSource removeConnectionParameters(ConnectionParameters parameters);
	
	/**
	 * Removes a <code>AcopDisplayerParameters</code> from this displayer. DataSource 
	 * that belongs to this parameters should be released (via AdapterFactory) and all
	 * resource associated with this parameters in the displayer should be released.
	 * 
	 * @param parameters parameters to be removed
	 * @return DataSource associated with this parameters
	 */
	public DataSource removeDisplayerParameters(T parameters);
}


/* __oOo__ */
