package de.desy.acop.launcher;

import java.io.IOException;
import java.io.InputStream;
import java.util.Iterator;
import java.util.LinkedHashMap;
import java.util.Map;
import java.util.Properties;

/**
 * 
 * <code>Utilities</code> offers a set of static methods to load the acop configurations.
 *
 * @author <a href="mailto:jaka.bobnar@cosylab.com">Jaka Bobnar</a>
 *
 */
public class Utilities {
	
	public static final String EMPTY_STRING = "";
	private static final String FILE_NAME = "acop/acop.properties";
	public static final String SPIDER_CUSTOMACTION = "spider.customAction";
	public static final String WEBPAGE_LOCATION = "about.webpage";
	
	private static Properties properties;
	
	/**
	 * Returns the property defined by the given name.
	 * 
	 * @param name the name of the property
	 * @return the property value
	 */
	public static String getAcopProperty(String name) {
		return getProperties().getProperty(name);
	}
	
	/**
	 * Reads and returns the properties from the acop/acop.properties file.
	 * 
	 * @return the acop properties
	 */
	public static Properties getProperties() {
		if (properties == null) {
			properties = System.getProperties();
			
			try(InputStream stream = Utilities.class.getClassLoader().getResourceAsStream(FILE_NAME)) {
				properties.load(stream);
			} catch (IOException e) {
				e.printStackTrace();
			}
		}
		return properties;
	}
	
	/**
	 * If arguments contains strings in format KEY=VAL or KEY="VAL with spaces" then they are added to
	 * properties as key/VALUE pairs. If argument string does not contain = sigh, 
	 * then it is added as key with empty string for value.
	 * @param args command line arguments
	 * @return parameters
	 */
	public static Map<String, String> argsToParameters(String[] args) {
		return argsToParameters(args, (Properties)null);
	}
	
	/**
	 * If arguments contains strings in format KEY=VAL or KEY="VAL with spaces" then they are added to
	 * properties as key/VALUE pairs. If argument string does not contain = sigh, 
	 * then it is added as key with empty string for value.
	 * @param args command line arguments
	 * @param prop a Properties, it is not modified, their values are copied to returned Properties
	 * @return parameters new Properties object which contains copy of provided properties plus parsed arguments 
	 */
	public static Map<String,String> argsToParameters(String[] args, Properties prop) {
		//preserver the order
		LinkedHashMap<String,String> p = new LinkedHashMap<>();
		
		if (args!=null) {
			for (int i = 0; i < args.length; i++) {
				int e= args[i].indexOf('=');
				if (e>-1) {
					String key= args[i].substring(0, e); 
					String val= args[i].substring(e+1);
					p.put(key, val);
					if (key.startsWith("jnlp.")) {
					    p.put(key.substring(5),val);
					}
				} else {
					p.put(args[i], EMPTY_STRING);
					if (args[i].startsWith("jnlp.")) {
                        p.put(args[i].substring(5),EMPTY_STRING);
                    }
				}
			}
		}
		//override the program arguments with the system properties. 
		//this may not seem logical but system properties is the only
		//way to supply startup parameters to a jnlp (jnlp supports program
		//arguments, but they have to be listed in the jnlp file and cannot
		//be given at javaws call, which resulted in default settings
		//always overriding the settings that one wanted to use)
		if (prop!=null) {
			Iterator<Object> it = prop.keySet().iterator();
			String key,val;
			while(it.hasNext()) {
				key = (String)it.next();
				val = prop.getProperty(key);
				p.put(key,val);
				if (key.startsWith("jnlp.")) {
                    p.put(key.substring(5),val);
                }
			}
		}
		return p;
	}
	
	/**
	 * If arguments contains strings in format KEY=VAL or KEY="VAL with spaces" then they are added to
	 * properties as key/VALUE pairs. If argument string does not contain = sigh, 
	 * then it is added as key with empty string for value.
	 * @param args command line arguments
	 * @param prop a Properties, it is not modified, their values are copied to returned Properties
	 * @return parameters new Properties object which contains copy of provided properties plus parsed arguments 
	 */
	public static Map<String,String> argsToParameters(String[] args, Map<String,String> prop) {
		//preserver the order
		LinkedHashMap<String,String> p = new LinkedHashMap<>();
		
		if (args!=null) {
			for (int i = 0; i < args.length; i++) {
				int e= args[i].indexOf('=');
				if (e>-1) {
					String key= args[i].substring(0, e); 
					String val= args[i].substring(e+1);
					p.put(key, val);
					if (key.startsWith("jnlp.")) {
                        p.put(key.substring(5),val);
                    }
				} else {
					p.put(args[i], EMPTY_STRING);
					if (args[i].startsWith("jnlp.")) {
                        p.put(args[i].substring(5),EMPTY_STRING);
                    }
				}
			}
		}
		//override the program arguments with the system properties. 
		//this may not seem logical but system properties is the only
		//way to supply startup parameters to a jnlp (jnlp supports program
		//arguments, but they have to be listed in the jnlp file and cannot
		//be given at javaws call, which resulted in default settings
		//always overriding the settings that one wanted to use)
		if (prop!=null) {
			Iterator<String> it = prop.keySet().iterator();
			String key,val;
			while(it.hasNext()) {
				key = it.next();
				val = prop.get(key);
				p.put(key,val);
				if (key.startsWith("jnlp.")) {
                    p.put(key.substring(5),val);
                }
			}
		}
		return p;
	}
	
}
