package de.desy.acop.transport.adapters;

/*
 * Copyright (c) 2006 by Cosylab d.o.o.
 *
 * The full license specifying the redistribution, modification, usage and other
 * rights and obligations is included with the distribution of this project in
 * the file license.html. If the license is not included you may find a copy at
 * http://www.cosylab.com/legal/abeans_license.htm or may write to Cosylab, d.o.o.
 *
 * THIS SOFTWARE IS PROVIDED AS-IS WITHOUT WARRANTY OF ANY KIND, NOT EVEN THE
 * IMPLIED WARRANTY OF MERCHANTABILITY. THE AUTHOR OF THIS SOFTWARE, ASSUMES
 * _NO_ RESPONSIBILITY FOR ANY CONSEQUENCE RESULTING FROM THE USE, MODIFICATION,
 * OR REDISTRIBUTION OF THIS SOFTWARE.
 */

import java.util.Map;

import com.cosylab.gui.displayers.DataConsumer;
import com.cosylab.gui.displayers.DataState;
import com.cosylab.gui.displayers.NonblockingNumberConsumer;
import com.cosylab.gui.displayers.SynchronousNumberConsumer;
import com.cosylab.util.CommonException;

import de.desy.acop.transport.AccessMode;
import de.desy.acop.transport.ConnectionParameters;
import de.desy.tine.client.TCallback;
import de.desy.tine.client.TLink;
import de.desy.tine.dataUtils.TDataType;
import de.desy.tine.definitions.TAccess;
import de.desy.tine.definitions.TErrorList;
import de.desy.tine.definitions.TFormat;
import de.desy.tine.definitions.TMode;


/**
 * <code>ReverseConsumer</code> is an implementation consumer that is attached
 * to those consumers which also act as data sources. This consumers receive
 * data from those sources and forward it to the TINE channel.
 * 
 * @author Jaka Bobnar, Cosylab
 *
 */
public class ReverseConsumer implements SynchronousNumberConsumer, NonblockingNumberConsumer, TCallback
	
{
	/** Supporter consumer types */
	public static final Class<?>[] SUPPORTED_CONSUMER_TYPES = {
		SynchronousNumberConsumer.class, NonblockingNumberConsumer.class};
	
	private final ConnectionParameters parameters;
	private TLink asyncLink;
	private final TFormat format;
	private final AcopTransportDataSource dataSource;
	private static final int TIMEOUT = 2000;
	
	public ReverseConsumer(ConnectionParameters param, TFormat format, AcopTransportDataSource dataSource)
	{
		this.format = format;
        this.parameters = param.deriveWith(AccessMode.WRITE);
        this.dataSource = dataSource;
	}

	/* (non-Javadoc)
	 * @see com.cosylab.gui.displayers.DataConsumer#getDataConsumer(java.lang.Class)
	 */
	@SuppressWarnings({"unchecked","rawtypes"})
    public DataConsumer getDataConsumer(Class type)
	{
		if (type.isAssignableFrom(this.getClass())) {
			return this;
		}

		return null;
	}

	/* (non-Javadoc)
	 * @see com.cosylab.gui.displayers.DataConsumer#getDefaultDataConsumer()
	 */
	public DataConsumer getDefaultDataConsumer()
	{
		return this;
	}

	/* (non-Javadoc)
	 * @see com.cosylab.gui.displayers.DataConsumer#updateDataState(com.cosylab.gui.displayers.DataState)
	 */
	public void updateDataState(DataState state)
	{
	    dataSource.setWriteDataState(state);
	}
	
	private void updateState(int linkStatus) {
	    if (linkStatus == 0) {
            updateDataState(new DataState(DataState.NORMAL,System.currentTimeMillis(),"Write Success"));
        } else {
            String type = DataState.ERROR;
            if (linkStatus == TErrorList.connection_timeout) {
                type = DataState.TIMEOUT;
            } else if (linkStatus == TErrorList.link_blacklisted || linkStatus == TErrorList.link_timeout) {
                type = DataState.DISCONNECTED;
            }
            updateDataState(new DataState(type,System.currentTimeMillis(),TErrorList.getErrorString(linkStatus)));
        }
	}

	/* (non-Javadoc)
	 * @see com.cosylab.gui.displayers.DataConsumer#setCharacteristics(java.util.Map)
	 */
	@SuppressWarnings("rawtypes")
	public void setCharacteristics(Map attributes)
	{
	}

	/* (non-Javadoc)
	 * @see com.cosylab.gui.displayers.DataConsumer#getName()
	 */
	public String getName()
	{
		return "WriteAdapter";
	}

	/* (non-Javadoc)
	 * @see com.cosylab.gui.displayers.DataConsumer#getSupportedCharacteristics()
	 */
	public String[] getSupportedCharacteristics()
	{
		return null;
	}

	/* (non-Javadoc)
	 * @see com.cosylab.gui.displayers.DataConsumer#getSupportedConsumerTypes()
	 */
	@SuppressWarnings({"unchecked","rawtypes"})
	public Class[] getSupportedConsumerTypes()
	{
		return SUPPORTED_CONSUMER_TYPES;
	}

	/*
	 * (non-Javadoc)
	 * @see com.cosylab.gui.displayers.SynchronousNumberConsumer#updateSynchronously(long, java.lang.Number)
	 */
	public void updateSynchronously(long timestamp, Number value) throws CommonException {
	    if (asyncLink == null) {
			createLink();
		}
	    TDataType data = TDataType.toTDataType(transformToString(value),format.getValue());
	    asyncLink.setInputDataObject(data);
	    synchronized(this) {
	        int cc = asyncLink.execute(TIMEOUT);
	        updateState(cc);
	    }
    }

	/*
	 * (non-Javadoc)
	 * @see com.cosylab.gui.displayers.NonblockingNumberConsumer#updateNonblocking(java.lang.Number)
	 */
	public void updateNonblocking(Number value) {
		if (asyncLink == null) {
			createLink();
		}
		TDataType data = TDataType.toTDataType(transformToString(value),format.getValue());
		asyncLink.setInputDataObject(data);	
		synchronized(this) {
    		int cc = asyncLink.attach(TMode.CM_SINGLE,this,TIMEOUT);
    		if (cc < 0) {
    		    updateState(-cc);
    		} else {
    		    updateState(0);
    		}
		}
    }
	
	private String transformToString(Number input) {
		int value = format.getValue();
		if (value == TFormat.CF_DLONG 
				|| value == TFormat.CF_INT16 
				|| value == TFormat.CF_LONG 
				|| value == TFormat.CF_SHORT 
				|| value == TFormat.CF_BYTE
				|| value == TFormat.CF_INT8 
				|| value == TFormat.CF_INT32 
				|| value == TFormat.CF_INT64 
				|| value == TFormat.CF_BIT 
				|| value == TFormat.CF_BOOLEAN) {
			return String.valueOf(input.longValue());		
		} else if (value == TFormat.CF_LNGINT 
				|| value == TFormat.CF_INTINT 
				|| value == TFormat.CF_SHTINT 
				|| value == TFormat.CF_INTDBL) {
			return String.valueOf(input.longValue()) + "\n0";
		} else if (value == TFormat.CF_FLTINT
				|| value == TFormat.CF_DBLINT
				|| value == TFormat.CF_FLTFLT 
				|| value == TFormat.CF_FLTDBL 
				|| value == TFormat.CF_DBLDBL) {
			return input.toString() + "\n0";
		} else if (value == TFormat.CF_INTFLTINT
				|| value == TFormat.CF_INTINTINT) {
			return String.valueOf(input.longValue()) + "\n0\n0";
		} else if (value == TFormat.CF_FLTFLTINT 
				|| value == TFormat.CF_FLTINTINT) {
			return input.toString() + "\n0\n0";
		} else if (value == TFormat.CF_INTFLTFLTFLT
				|| value == TFormat.CF_INTINTINTINT
				|| value == TFormat.CF_INTFLTSHTSHT 
				|| value == TFormat.CF_INTINTFLTFLT) {
			return String.valueOf(input.longValue()) + "\n0\n0\n0";
		} else if (value == TFormat.CF_FLTINTFLTINT) {
			return input.toString() + "\n0\n0\n0";
		} else {
			return input.toString();
		}
	}
	 	
	private void createLink() {
		String deviceName = "/" + parameters.getDeviceContext() +"/" + parameters.getDeviceGroup() +"/" + parameters.getDeviceName();
		asyncLink = new TLink(deviceName, parameters.getDeviceProperty(), new TDataType(), new TDataType(), TAccess.CA_WRITE);
	}	
	
	/**
	 * Releases this consumer.
	 */
	public void release() {
		try {
			if (asyncLink != null)
				asyncLink.close();
		} catch (Exception e) {
			//ignore
		}
		asyncLink = null;
	}
	
	@Override
	public void callback(int LinkIndex, int LinkStatus) {
	    synchronized(this) {
	        updateState(LinkStatus);
	    }
	}
}
/* __oOo__ */
