package de.desy.acop.video.analysis;

import de.desy.acop.video.Definitions;
import de.desy.tine.structUtils.TTaggedStructure;
import de.desy.tine.types.IMAGE;

/**
 *
 * <code>AnalysisImage</code> is the object that is transferred across the
 * system from the video analysis server to the video client.
 * This object carries with it the raw tine IMAGE and analysis results
 * and parameters. For details on parameters see the constructor
 * documentation.
 *
 * @author <a href="mailto:jaka.bobnar@cosylab.com">Jaka Bobnar</a>
 *
 */
public abstract class AnalysisImage extends TTaggedStructure {

	private final IMAGE[] image = new IMAGE[1];
	private final int[] imageW = new int[1];
	private final int[] imageH = new int[1];
	private final int[] roiX = new int[1];
	private final int[] roiY = new int[1];
	private final int[] roiW = new int[1];
	private final int[] roiH = new int[1];
	private final int[] roi2X = new int[1];
	private final int[] roi2Y = new int[1];
	private final int[] roi2W = new int[1];
	private final int[] roi2H = new int[1];
	private final int[] thresholdPoints = new int[1];
	private final double[] threshold = new double[1];
	private final double[] sideViewX = new double[Definitions.MAX_WIDTH];
	private final double[] sideViewY = new double[Definitions.MAX_HEIGHT];
	private final byte[] calculateThreshold = new byte[1];
	private final byte[] performFit = new byte[1];
	private final byte[] performSmoothing = new byte[1];

	private int width;
	private int height;
	/**
	 * Constructs a new AnalysisImage, with default values (empty image and 0 value parameters
	 * and results).
	 */
	protected AnalysisImage(String name) {
		super(name);
		image[0] = new IMAGE();
		width = -1;
		height = -1;
		initStructure();
	}

	/**
	 * Constructs a new AnalysisImage from the given AnalysisImage. This method creates a clone
	 * of the given image.
	 *
	 * @param image the image
	 */
	protected AnalysisImage(String name, AnalysisImage image) {
		this(name, image.getImage(),
				image.getRoiX(),image.getRoiY(),image.getRoiW(),image.getRoiH(),
				image.getRoi2X(), image.getRoi2Y(), image.getRoi2W(), image.getRoi2H(), image.getThreshold(),
				image.getSideViewX(), image.getSideViewY(),
				image.getThresholdPoints(),image.isCalculateThreshold(),image.isPerformFit(),image.isPerformSmoothing());
	}

	/**
	 * Constructs a new AnalysisImage.
	 *
	 * @param image the raw image, which was analyzed and all data was obtained from
	 * @param roiX the region of interest right coordinate (image is analyzed inside region of interest)
	 * @param roiY the region of interest upper coordinate (image is analyzed inside region of interest)
	 * @param roiW the width of region of interest (image is analyzed inside region of interest)
	 * @param roiH the height of region of interest (image is analyzed inside region of interest)
	 * @param roi2X the region of interest used for automatic threshold calculation right coordinate
	 * @param roi2Y the region of interest used for automatic threshold calculation upper coordinate
	 * @param roi2W the region of interest used for automatic threshold calculation width
	 * @param roi2H the region of interest used for automatic threshold calculation height
	 * @param threshold the used threshold value (calculated or user defined)
	 * @param sideViewX horizontal side view values
	 * @param sideViewY vertical side view values
	 * @param thresholdPoints the number of points on which the threshold was calculated
	 * @param calculateThreshold flag indicating if threshold calculation is automatic
	 * @param performFit flag indicating if fitting was performed (if false the 'fit' parameters
	 * 			are a-priori invalid)
	 * @param performSmoothing flag indicating if smoothing should be performed
	 */
	protected AnalysisImage(String name, IMAGE image,
			int roiX, int roiY, int roiW, int roiH, int roi2X, int roi2Y,int roi2W, int roi2H, double threshold,
			double[] sideViewX, double[] sideViewY,
			int thresholdPoints, boolean calculateThreshold, boolean performFit, boolean performSmoothing) {
		super(name);

		int width = image.getFrameHeader().sourceWidth;
		int height = image.getFrameHeader().sourceHeight;
		int aoiWidth = image.getFrameHeader().aoiWidth;
		int aoiHeight = image.getFrameHeader().aoiHeight;

		if (aoiWidth > 0) {
			if (sideViewX.length != aoiWidth) throw new IllegalArgumentException("Incompatible size of X side view array.");
			if (roiX < 0 || roiW < 0 || roiX+roiW > aoiWidth) throw new IllegalArgumentException("Illegal region of interest specified.");
			if (roi2X < 0 || roi2W < 0 || roi2X+roi2W > aoiWidth) throw new IllegalArgumentException("Illegal region of interest 2 specified.");
			this.width = aoiWidth;
		} else {
		    if (sideViewX.length != width) throw new IllegalArgumentException("Incompatible size of X side view array.");
            if (roiX < 0 || roiW < 0 || roiX+roiW > width) throw new IllegalArgumentException("Illegal region of interest specified.");
            if (roi2X < 0 || roi2W < 0 || roi2X+roi2W > width) throw new IllegalArgumentException("Illegal region of interest 2 specified.");
            this.width = width != 0 ? width : -1;
		}

		if (aoiHeight > 0) {
            if (sideViewY.length != aoiHeight) throw new IllegalArgumentException("Incompatible size of Y side view array.");
            if (roiY < 0 || roiH < 0 || roiY+roiH > aoiHeight) throw new IllegalArgumentException("Illegal region of interest specified.");
            if (roi2Y < 0 || roi2H < 0 || roi2Y+roiH > aoiHeight) throw new IllegalArgumentException("Illegal region of interest 2 specified.");
            this.height = aoiHeight;
        } else {
			if (sideViewY.length != height) throw new IllegalArgumentException("Incompatible size of Y side view array.");
			if (roiY < 0 || roiH < 0 || roiY+roiH > height) throw new IllegalArgumentException("Illegal region of interest specified.");
			if (roi2Y < 0 || roi2H < 0 || roi2Y+roi2H > height) throw new IllegalArgumentException("Illegal region of interest 2 specified.");
			this.height = height != 0 ? height : -1;
		}

		System.arraycopy(sideViewX,0,this.sideViewX,0,sideViewX.length);
		System.arraycopy(sideViewY,0,this.sideViewY,0,sideViewY.length);
		this.roiX[0] = roiX;
		this.roiY[0] = roiY;
		this.roiW[0] = roiW;
		this.roiH[0] = roiH;
		this.roi2X[0] = roi2X;
		this.roi2Y[0] = roi2Y;
		this.roi2W[0] = roi2W;
		this.roi2H[0] = roi2H;
		this.imageW[0] = width;
		this.imageH[0] = height;
		this.threshold[0] = threshold;
		this.thresholdPoints[0] = thresholdPoints;
		this.calculateThreshold[0] = calculateThreshold ? (byte)1 : (byte)0;
		this.performFit[0] = performFit ? (byte)1 : (byte)0;
		this.performSmoothing[0] = performSmoothing ? (byte)1 : (byte)0;
		//assign directly, but make sure you pass in a clean copy
		this.image[0] = image;
//		loadImage(this.image, image);

		initStructure();
	}

	private void initStructure() {
		addField(image,"image");
		addField(imageW,"imageW");
	    addField(imageH,"imageH");
		addField(roiX,"roiX");
		addField(roiY,"roiY");
		addField(roiW,"roiW");
	    addField(roiH,"roiH");
	    addField(roi2X,"roi2X");
		addField(roi2Y,"roi2Y");
		addField(roi2W,"roi2W");
	    addField(roi2H,"roi2H");
	    addField(threshold,"threshold");
	    addField(sideViewX, "sideViewX");
	    addField(sideViewY, "sideViewY");
	    addField(thresholdPoints, "thresholdPoints");
	    addField(performFit, "performFit");
	    addField(performSmoothing, "performSmoothing");
	    addField(calculateThreshold, "calculateThreshold");

	}

	/**
	 * Returns the image which was analyzed.
	 *
	 * @return the original image
	 */
	public IMAGE getImage() {
		return image[0];
	}

	/**
	 * The width of the image.
	 *
	 * @return the width
	 *
	 * @see {@link #getImage()}
	 */
	public int getImageW() {
		return imageW[0];
	}

	/**
	 * The height of the image.
	 *
	 * @return the height
	 *
	 * @see {@link #getImage()}
	 */
	public int getImageH() {
		return imageH[0];
	}

	/**
	 * Left coordinate of the region of interest.
	 *
	 * @return the left coordinate
	 */
	public int getRoiX() {
		return roiX[0];
	}

	/**
	 * Upper coordinate of the region of interest.
	 *
	 * @return the upper coordinate
	 */
	public int getRoiY() {
		return roiY[0];
	}

	/**
	 * The width of the region of interest.
	 *
	 * @return the width
	 */
	public int getRoiW() {
		return roiW[0];
	}

	/**
	 * The height of the region of interest.
	 *
	 * @return the height
	 */
	public int getRoiH() {
		return roiH[0];
	}

	/**
	 * Left coordinate of the second region of interest.
	 *
	 * @return the left coordinate
	 */
	public int getRoi2X() {
		return roi2X[0];
	}

	/**
	 * Upper coordinate of the second region of interest.
	 *
	 * @return the upper coordinate
	 */
	public int getRoi2Y() {
		return roi2Y[0];
	}

	/**
	 * The width of the second region of interest.
	 *
	 * @return the width
	 */
	public int getRoi2W() {
		return roi2W[0];
	}

	/**
	 * The height of the second region of interest.
	 *
	 * @return the height
	 */
	public int getRoi2H() {
		return roi2H[0];
	}

	/**
	 * The threshold value used for analysis of the image.
	 *
	 * @return the threshold value
	 */
	public double getThreshold() {
		return threshold[0];
	}

	/**
	 * Returns the array representing the side projection of the intensity image
	 * on the X axis.
	 *
	 * @return the projection on the x axis
	 *
	 * @see {@link #getImage()}
	 */
	public double[] getSideViewX() {
	    if (width == -1 && image[0].getFrameHeader().sourceWidth != 0) {
	        width = image[0].getFrameHeader().sourceWidth;
	    }
	    if (width == -1) {
            return new double[0];
        }
	    double[] result = new double[width];
        System.arraycopy(sideViewX, 0, result, 0, result.length);
        return result;
	}

	/**
	 * Returns the array representing the side projection of the intensity image
	 * on the Y axis.
	 *
	 * @return the projection on the y axis
	 *
	 * @see {@link #getImage()}
	 */
	public double[] getSideViewY() {
	    if (height == -1 && image[0].getFrameHeader().sourceHeight != 0) {
            height = image[0].getFrameHeader().sourceHeight;
        }
	    if (height == -1) {
            return new double[0];
        }
		double[] result = new double[height];
		System.arraycopy(sideViewY, 0, result, 0, result.length);
		return result;
	}

	/**
	 * Returns true if the calculation of threshold was automatic or false if
	 * threshold was user defined.
	 *
	 * @return true if calculation of threshold is automatic
	 */
	public boolean isCalculateThreshold() {
		return calculateThreshold[0] == 1;
	}

	/**
	 * Returns true if the fitting algorithm was used during analysis.
	 *
	 * @return true if fitting was performed
	 */
	public boolean isPerformFit() {
		return performFit[0] == 1;
	}

	/**
	 * Returns true if the smoothing algorithm was used during analysis.
	 *
	 * @return true if fitting was performed
	 */
	public boolean isPerformSmoothing() {
		return performSmoothing[0] == 1;
	}

	/**
	 * Returns the number of points on which the threshold was calculated.
	 *
	 * @return the number of points threshold was calculated on
	 */
	public int getThresholdPoints() {
		return thresholdPoints[0];
	}

}
