package de.desy.acop.video.analysis;

import java.awt.image.BufferedImage;

import de.desy.acop.video.displayer.ColorMap;
import de.desy.acop.video.timageio.TBufferedImage;
import de.desy.tine.types.IMAGE;

/**
 * <code>ImageAnalyzer</code> implements the image analysis. There can be different implementation of image analysis,
 * depending on the source and type of the image (e.g. beam cross section, interferometer image).
 *
 * @author <a href="mailto:jaka.bobnar@cosylab.com">Jaka Bobnar</a>
 */
public abstract class ImageAnalyzer {

    /**
     * Clear cached values and recalculate the fit from scratch
     */
    public abstract void clear();

    /**
     * Finds the minimum and maximum value in the array looking at the elements between start and start+length.
     *
     * @param values the values to check
     * @param start the starting index
     * @param length the length of the check
     * @return the triple min/max/index
     */
    protected static double[] findMinMax(double[] values, int start, int length) {
        double min = Double.MAX_VALUE;
        double max = -Double.MAX_VALUE;
        int index = 0;
        for (int i = start; i < start + length; i++) {
            if (values[i] > max) {
                max = values[i];
                index = i;
            }
            if (values[i] < min) min = values[i];
        }
        return new double[]{min,max,index};
    }

    /**
     * Perform the analysis and return the analysis image.
     *
     * @param colorMap the color map used for transforming the image to RGB values
     * @param decoder decoder changes the RGB values to a single double value
     * @param image the image to perform analysis on
     * @param preciseBackground the background image to subtract from the image
     * @param imageWidth the image width
     * @param imageHeight the image height
     * @param roiX region of interest left x coordinate
     * @param roiY region of interest upper y coordinate
     * @param roiW region of interest width
     * @param roiH region of interest height
     * @param roi2x threshold calculation region left x coordinate
     * @param roi2y threshold calculation region upper y coordinate
     * @param roi2w threshold calculation region width
     * @param roi2h threshold calculation region height
     * @param threshold the threshold value (only pixel values above threshold will be used in analysis)
     * @param calculateThreshold true if the threshold should be calculated as the average value in the threshold region
     * @param performFit true if side view fitting should be performed
     * @param performSmoothing true if smoothing of values should be performed prior to attempt fitting
     * @return the result of the analysis
     */
    public AnalysisImage performAnalysis(ColorMap colorMap, ColorDecoder decoder, IMAGE image,
            double[] preciseBackground, int imageWidth, int imageHeight, int roiX, int roiY, int roiW, int roiH,
            int roi2x, int roi2y, int roi2w, int roi2h, double threshold, boolean calculateThreshold,
            boolean performFit, boolean performSmoothing) {
        IMAGE newImage = image;
        double[] values = null;
        int aH = image.getFrameHeader().aoiHeight;
        int aW = image.getFrameHeader().aoiWidth;
        if (aH != -1) {
            imageWidth = aW;
            imageHeight = aH;
        }
        if (preciseBackground != null) {
            BufferedImage bufferedImage = ImageAnalysisEngine.toBufferedImage(image,ColorMap.NONE);
            if (bufferedImage == null) throw new IllegalStateException("No BufferedImage!");
            values = ImageAnalysisEngine.imageToDoubleValues(bufferedImage,
                    image.getFrameHeader().effectiveBitsPerPixel,decoder);
            try {
                int i = 0;
                //try catch is actually faster than anything else
                for (;;) {
                    values[i] -= preciseBackground[i];
                    i++;
                }
            } catch (ArrayIndexOutOfBoundsException e) {
                //ignore
            }
            bufferedImage = ImageAnalysisEngine.toBufferedImage(values,bufferedImage.getWidth(),image);
            newImage = TBufferedImage.toIMAGE(bufferedImage);
            newImage.getFrameHeader().frameNumber = image.getFrameHeader().frameNumber;
        }

        BufferedImage bufferedImage = ImageAnalysisEngine.toBufferedImage(newImage,ColorMap.NONE);
        if (bufferedImage == null) throw new IllegalStateException("No BufferedImage!");
        values = ImageAnalysisEngine.imageToDoubleValues(bufferedImage,
                newImage.getFrameHeader().effectiveBitsPerPixel,decoder);
        //calculate the threshold value, which is a constant offset. values below this are not taken into account in
        //calculation. sum up the values in the threshold ROI and outside ROI at the same time and calculate the average
        //value. That is the new threshold, if there wasn't any specified by the user
        //logger.log(Level.FINE, "Analysis: Calculating threshold.");
        int thresholdPointsCount = 0;
        if (calculateThreshold) {
            double sum = 0;
            int r = roiX + roiW, b = roiY + roiH;
            boolean inside = false;
            for (int j = 0; j < roi2h; j++) {
                inside = roiY <= roi2y + j && roi2y + j < b;
                for (int i = 0; i < roi2w; i++) {
                    //do not include indexes, which are in normal ROI
                    if (roiX <= roi2x + i && roi2x + i < r && inside) continue;
                    sum += values[i + roi2x + imageWidth * (j + roi2y)];
                    thresholdPointsCount++;
                }
            }
            if (thresholdPointsCount != 0) {
                threshold = sum / thresholdPointsCount;
            } else {
                threshold = Double.NaN;
            }
        }
        return doAnalysis(newImage,values,imageWidth,imageHeight,roiX,roiY,roiW,roiH,roi2x,roi2y,roi2w,roi2h,
                thresholdPointsCount,threshold,calculateThreshold,performFit,performSmoothing);
    }

    /**
     * Perform the actual analysis of the image. The threshold was already calculated and the background image was
     * already subtracted.
     *
     * @param newImage the image to perform the analysis of
     * @param values the image transformed and decoded to a single array of numbers between 0 and 255. The background
     *        image was already subtracted. The values array has length iW*iH and the data are ordered row after row,
     *        starting from the top.
     * @param iW image width
     * @param iH image height
     * @param rX region of interest left x coordinate
     * @param rY region of interest upper y coordinate
     * @param rW region of interest width
     * @param rH region of interest height
     * @param r2x threshold calculation region left x coordinate
     * @param r2y threshold calculation region upper y coordinate
     * @param r2w threshold calculation region width
     * @param r2h threshold calculation region height
     * @param thresholdPointsCount the number of points used in threshold calculation
     * @param threshold the threshold to use (already calculated if so requested by the next parameter)
     * @param calculateThreshold true if threshold was calculated or false otherwise
     * @param performFit true if fitting should be performed or false otherwise
     * @param smoothing true if smoothing should be performed prior to attempting the fit
     * @return the analysis result
     */
    protected abstract AnalysisImage doAnalysis(IMAGE newImage, double[] values, int iW, int iH, int rX, int rY, int rW,
            int rH, int r2X, int r2Y, int r2W, int r2H, int thresholdPointsCount, double threshold,
            boolean calculateThreshold, boolean performFit, boolean smoothing);

    /**
     * Return the type of the analysis performed by the analyzer.
     *
     * @return the analysis type
     */
    public abstract AnalysisType getAnalysisType();
}
