package de.desy.acop.video.analysis.gaussian;

import de.desy.tine.types.IMAGE;

/**
 * <code>AImage</code> is the object that is transferred across the system from the video analysis server to the video
 * client. This object carries with it the raw tine IMAGE and analysis results and parameters. For details on parameters
 * see the constructor documentation.
 *
 * @author <a href="mailto:jaka.bobnar@cosylab.com">Jaka Bobnar</a>
 */
public class AImage extends AbstractGaussianImage {

    private static final String NAME = "AISr4";
    private final double[] sideViewAmplitudeX = new double[1];
    private final double[] sideViewAmplitudeY = new double[1];
    private final double[] sideViewConstX = new double[1];
    private final double[] sideViewConstY = new double[1];
    private final double[] stdA = new double[1];
    private final double[] stdB = new double[1];
    private final double[] meanXFit = new double[1];
    private final double[] stdXFit = new double[1];
    private final double[] amplitudeXFit = new double[1];
    private final double[] constXFit = new double[1];
    private final double[] slopeXFit = new double[1];
    private final double[] meanYFit = new double[1];
    private final double[] stdYFit = new double[1];
    private final double[] amplitudeYFit = new double[1];
    private final double[] constYFit = new double[1];
    private final double[] slopeYFit = new double[1];

    /**
     * Constructs a new AImage, with default values (empty image and 0 value parameters and results).
     */
    public AImage() {
        super(NAME);
        initStructure();
        initDone();
    }

    /**
     * Constructs a new AImage from the given AImage. This method creates a clone of the given image.
     *
     * @param image the image
     */
    public AImage(AImage image) {
        this(image.getImage(),image.getRoiX(),image.getRoiY(),image.getRoiW(),image.getRoiH(),image.getRoi2X(),
                image.getRoi2Y(),image.getRoi2W(),image.getRoi2H(),image.getThreshold(),image.getSideViewX(),
                image.getMeanX(),image.getStdX(),image.getSideViewAmplitudeX(),image.getSideViewConstX(),
                image.getMeanXFit(),image.getStdXFit(),image.getAmplitudeXFit(),image.getConstXFit(),
                image.getSlopeXFit(),image.getSideViewY(),image.getMeanY(),image.getStdY(),
                image.getSideViewAmplitudeY(),image.getSideViewConstY(),image.getMeanYFit(),image.getStdYFit(),
                image.getAmplitudeYFit(),image.getConstYFit(),image.getSlopeYFit(),image.getStdA(),image.getStdB(),
                image.getAngle2D(),image.getAmplitude2D(),image.getConst2D(),image.getThresholdPoints(),
                image.isCalculateThreshold(),image.isPerformFit(),image.isPerformSmoothing());
    }

    /**
     * Constructs a new AImage.
     *
     * @param image the raw image, which was analyzed and all data was obtained from
     * @param roiX the region of interest right coordinate (image is analyzed inside region of interest)
     * @param roiY the region of interest upper coordinate (image is analyzed inside region of interest)
     * @param roiW the width of region of interest (image is analyzed inside region of interest)
     * @param roiH the height of region of interest (image is analyzed inside region of interest)
     * @param roi2X the region of interest used for automatic threshold calculation right coordinate
     * @param roi2Y the region of interest used for automatic threshold calculation upper coordinate
     * @param roi2W the region of interest used for automatic threshold calculation width
     * @param roi2H the region of interest used for automatic threshold calculation height
     * @param threshold the used threshold value (calculated or user defined)
     * @param sideViewX horizontal side view values
     * @param meanX horizontal statistical mean value
     * @param stdX horizontal statistical standard deviation value
     * @param sideAmplitudeX horizontal statistical amplitude value
     * @param sideConstX horizontal statistical constant value
     * @param meanXFit horizontal fitted mean value
     * @param stdXFit horizontal fitted standard deviation value
     * @param sideAmplitudeXFit horizontal fitted amplitude value
     * @param sideConstXFit horizontal fitted constant value
     * @param slopeXFit the slope of the linear function in the fit
     * @param sideViewY vertical side view values
     * @param meanY vertical statistical mean value
     * @param stdY vertical statistical standard deviation value
     * @param sideAmplitudeY vertical statistical amplitude value
     * @param sideConstY vertical statistical constant value
     * @param meanYFit vertical fitted mean value
     * @param stdYFit vertical fitted standard deviation value
     * @param sideAmplitudeYFit vertical fitted amplitude value
     * @param sideConstYFit vertical fitted constant value
     * @param slopeYFit the slope of the linear function in the fit
     * @param stdA main 2D axis statistical standard deviation
     * @param stdB secondary 2D axis statistical standard deviation
     * @param angle2D the statistical rotation angle of the main axis according to x-y coordinate system
     * @param amplitude2D the statistical 2D amplitude
     * @param const2D the statistical 2D constant (minimum value)
     * @param thresholdPoints the number of points on which the threshold was calculated
     * @param calculateThreshold flag indicating if threshold calculation is automatic
     * @param performFit flag indicating if fitting was performed (if false the 'fit' parameters are a-priori invalid)
     * @param performSmoothing flag indicating if smoothing should be performed
     */
    public AImage(IMAGE image, int roiX, int roiY, int roiW, int roiH, int roi2X, int roi2Y, int roi2W, int roi2H,
            double threshold, double[] sideViewX, double meanX, double stdX, double sideAmplitudeX, double sideConstX,
            double meanXFit, double stdXFit, double sideAmplitudeXFit, double sideConstXFit, double slopeXFit,
            double[] sideViewY, double meanY, double stdY, double sideAmplitudeY, double sideConstY, double meanYFit,
            double stdYFit, double sideAmplitudeYFit, double sideConstYFit, double slopeYFit, double stdA, double stdB,
            double angle2D, double amplitude2D, double const2D, int thresholdPoints, boolean calculateThreshold,
            boolean performFit, boolean performSmoothing) {
        super(NAME,image,roiX,roiY,roiW,roiH,roi2X,roi2Y,roi2W,roi2H,threshold,sideViewX,sideViewY,
                meanX, stdX, meanY, stdY, amplitude2D, const2D, angle2D,thresholdPoints,
                calculateThreshold,performFit,performSmoothing);
        this.stdA[0] = stdA;
        this.stdB[0] = stdB;
        this.sideViewAmplitudeX[0] = sideAmplitudeX;
        this.sideViewAmplitudeY[0] = sideAmplitudeY;
        this.sideViewConstX[0] = sideConstX;
        this.sideViewConstY[0] = sideConstY;
        this.stdXFit[0] = stdXFit;
        this.stdYFit[0] = stdYFit;
        this.meanXFit[0] = meanXFit;
        this.meanYFit[0] = meanYFit;
        this.amplitudeXFit[0] = sideAmplitudeXFit;
        this.amplitudeYFit[0] = sideAmplitudeYFit;
        this.constXFit[0] = sideConstXFit;
        this.constYFit[0] = sideConstYFit;
        this.slopeXFit[0] = slopeXFit;
        this.slopeYFit[0] = slopeYFit;
        initStructure();
        initDone();
    }

    @Override
    protected void initStructure() {
        //reinit here, to keep a different order (backwards compatible
        addField(sideViewAmplitudeX,"sideViewAmplitudeX");
        addField(sideViewAmplitudeY,"sideViewAmplitudeY");
        addField(sideViewConstX,"sideViewConstX");
        addField(sideViewConstY,"sideViewConstY");
        addField(stdX,"stdX");
        addField(stdY,"stdY");
        addField(meanX,"meanX");
        addField(meanY,"meanY");
        addField(stdA,"stdA");
        addField(stdB,"stdB");
        addField(amplitude2D,"amplitude2D");
        addField(angle2D,"angle2D");
        addField(const2D,"const2D");
        addField(meanXFit,"meanXFit");
        addField(stdXFit,"stdXFit");
        addField(amplitudeXFit,"amplitudeXFit");
        addField(constXFit,"constXFit");
        addField(slopeXFit,"slopeXFit");
        addField(meanYFit,"meanYFit");
        addField(stdYFit,"stdYFit");
        addField(amplitudeYFit,"amplitudeYFit");
        addField(constYFit,"constYFit");
        addField(slopeYFit,"slopeYFit");
    }

    /**
     * Returns the standard deviation of the 2-dimensional intensity image in the direction of the "a" axis of the
     * profile (assuming an elliptical profile).
     *
     * @return the standard deviation in a direction
     * @see {@link #getImage()}
     */
    public double getStdA() {
        return stdA[0];
    }

    /**
     * Returns the standard deviation of the 2-dimensional intensity image in the direction of the "b" axis of the
     * profile (assuming an elliptical profile).
     *
     * @return the standard deviation in b direction
     * @see {@link #getImage()}
     */
    public double getStdB() {
        return stdB[0];
    }

    /**
     * Returns the amplitude (maximum minus minimum) value of the side view projection of the image on the x axis. This
     * is calculated from the values returned by the {@link #getSideViewX()}.
     *
     * @return the amplitude
     */
    public double getSideViewAmplitudeX() {
        return sideViewAmplitudeX[0];
    }

    /**
     * Returns the amplitude (maximum minus minimum) value of the side view projection of the image on the y axis. This
     * is calculated from the values returned by the {@link #getSideViewY()}.
     *
     * @return the amplitude
     */
    public double getSideViewAmplitudeY() {
        return sideViewAmplitudeY[0];
    }

    /**
     * Returns the constant (minimum) value of the side view projection of the image on the x axis. This is calculated
     * from the values returned by the {@link #getSideViewX()}.
     *
     * @return the constant
     */
    public double getSideViewConstX() {
        return sideViewConstX[0];
    }

    /**
     * Returns the constant (minimum) value of the side view projection of the image on the y axis. This is calculated
     * from the values returned by the {@link #getSideViewY()}.
     *
     * @return the constant
     */
    public double getSideViewConstY() {
        return sideViewConstY[0];
    }

    /**
     * Returns the fitted amplitude in x axis.
     *
     * @return
     */
    public double getAmplitudeXFit() {
        return amplitudeXFit[0];
    }

    /**
     * Returns the fitted constant in x axis.
     *
     * @return
     */
    public double getConstXFit() {
        return constXFit[0];
    }

    /**
     * Returns the fitted mean value in x axis.
     *
     * @return
     */
    public double getMeanXFit() {
        return meanXFit[0];
    }

    /**
     * Returns the fitted std value in x axis.
     *
     * @return
     */
    public double getStdXFit() {
        return stdXFit[0];
    }

    /**
     * Returns the fitted amplitude in y axis.
     *
     * @return
     */
    public double getAmplitudeYFit() {
        return amplitudeYFit[0];
    }

    /**
     * Returns the slope of the linear function in the fit.
     *
     * @return the slope
     */
    public double getSlopeXFit() {
        return slopeXFit[0];
    }

    /**
     * Returns the fitted constant in y axis.
     *
     * @return
     */
    public double getConstYFit() {
        return constYFit[0];
    }

    /**
     * Returns the fitted mean value in y axis.
     *
     * @return
     */
    public double getMeanYFit() {
        return meanYFit[0];
    }

    /**
     * Returns the fitted std value in y axis.
     *
     * @return
     */
    public double getStdYFit() {
        return stdYFit[0];
    }

    /**
     * Returns the slope of the linear function in the fit.
     *
     * @return
     */
    public double getSlopeYFit() {
        return slopeYFit[0];
    }
}
