package de.desy.acop.video.analysis.gaussian;

import de.desy.acop.video.analysis.AnalysisImage;
import de.desy.tine.types.IMAGE;

/**
 * <code>Gaussian2DImage</code> is the object that is transferred across the system from the video analysis server to
 * the video client. This object carries with it the raw tine IMAGE and analysis results and parameters. For details on
 * parameters see the constructor documentation.
 *
 * @author <a href="mailto:jaka.bobnar@cosylab.com">Jaka Bobnar</a>
 */
public abstract class AbstractGaussianImage extends AnalysisImage {

    protected final double[] stdX = new double[1];
    protected final double[] stdY = new double[1];
    protected final double[] meanX = new double[1];
    protected final double[] meanY = new double[1];
    protected final double[] amplitude2D = new double[1];
    protected final double[] const2D = new double[1];
    protected final double[] angle2D = new double[1];

    /**
     * Constructs a new AbstractGaussianImage, with default values (empty image and 0 value parameters and results).
     */
    public AbstractGaussianImage(String name) {
        super(name);
        initStructure();
    }

    /**
     * Constructs a new AbstractGaussianImage from the given AbstractGaussianImage. This method creates a clone of the
     * given image.
     *
     * @param image the image
     */
    public AbstractGaussianImage(AbstractGaussianImage image) {
        this(image.getName(),image.getImage(),image.getRoiX(),image.getRoiY(),image.getRoiW(),image.getRoiH(),
                image.getRoi2X(),image.getRoi2Y(),image.getRoi2W(),image.getRoi2H(),image.getThreshold(),
                image.getSideViewX(),image.getSideViewY(),image.getMeanX(),image.getStdX(),image.getMeanY(),
                image.getStdY(),image.getAmplitude2D(),image.getConst2D(),image.getAngle2D(),image.getThresholdPoints(),
                image.isCalculateThreshold(),image.isPerformFit(),image.isPerformSmoothing());
    }

    /**
     * Constructs a new AbstractGaussianImage.
     *
     * @param name the name of the structure
     * @param image the raw image, which was analyzed and all data was obtained from
     * @param roiX the region of interest right coordinate (image is analyzed inside region of interest)
     * @param roiY the region of interest upper coordinate (image is analyzed inside region of interest)
     * @param roiW the width of region of interest (image is analyzed inside region of interest)
     * @param roiH the height of region of interest (image is analyzed inside region of interest)
     * @param roi2X the region of interest used for automatic threshold calculation right coordinate
     * @param roi2Y the region of interest used for automatic threshold calculation upper coordinate
     * @param roi2W the region of interest used for automatic threshold calculation width
     * @param roi2H the region of interest used for automatic threshold calculation height
     * @param threshold the used threshold value (calculated or user defined)
     * @param sideViewX horizontal side view values
     * @param sideViewY vertical side view values
     * @param meanX horizontal statistical mean value
     * @param stdX horizontal statistical standard deviation value
     * @param meanY vertical statistical mean value
     * @param stdY vertical statistical standard deviation value
     * @param amplitude vertical statistical amplitude value
     * @param const vertical statistical constant value
     * @param angle the rotation angle of the main axes
     * @param thresholdPoints the number of points on which the threshold was calculated
     * @param calculateThreshold flag indicating if threshold calculation is automatic
     * @param performFit flag indicating if fitting was performed (if false the 'fit' parameters are a-priori invalid)
     * @param performSmoothing flag indicating if smoothing should be performed
     */
    public AbstractGaussianImage(String name, IMAGE image, int roiX, int roiY, int roiW, int roiH, int roi2X, int roi2Y,
            int roi2W, int roi2H, double threshold, double[] sideViewX, double[] sideViewY, double meanX, double stdX,
            double meanY, double stdY, double amplitude, double constant, double angle, int thresholdPoints,
            boolean calculateThreshold, boolean performFit, boolean performSmoothing) {
        super(name,image,roiX,roiY,roiW,roiH,roi2X,roi2Y,roi2W,roi2H,threshold,sideViewX,sideViewY,thresholdPoints,
                calculateThreshold,performFit,performSmoothing);
        this.stdX[0] = stdX;
        this.meanX[0] = meanX;
        this.stdY[0] = stdY;
        this.meanY[0] = meanY;
        this.amplitude2D[0] = amplitude;
        this.const2D[0] = constant;
        this.angle2D[0] = angle;
    }

    protected void initStructure() {
        addField(stdX,"stdX");
        addField(stdY,"stdY");
        addField(meanX,"meanX");
        addField(meanY,"meanY");
        addField(amplitude2D,"amplitude2D");
        addField(const2D,"const2D");
        addField(angle2D,"angle2D");
    }

    /**
     * Returns the amplitude of the 2D fit on the data. The data are decoded according to the image and color decoder,
     * then normalized to 256.
     *
     * @return the amplitude
     */
    public double getAmplitude2D() {
        return amplitude2D[0];
    }

    /**
     * Returns the angle of rotation of the 2-dimensional intensity profile around its center.
     *
     * @return the angle of rotation of the 2-dimensional intensity profile
     */
    public double getAngle2D() {
        return angle2D[0];
    }

    /**
     * Returns the constant of the 2D fit on the data. This value indicates the background level of the image.
     *
     * @return the constant value of the intensity profile
     */
    public double getConst2D() {
        return const2D[0];
    }

    /**
     * Returns the standard deviation of x axis obtained from the 2D fit results.
     *
     * @return the standard deviation
     */
    public double getStdX() {
        return stdX[0];
    }

    /**
     * Returns the standard deviation of y axis obtained from the 2D fit results.
     *
     * @return the standard deviation
     */
    public double getStdY() {
        return stdY[0];
    }

    /**
     * Returns the mean x value obtained from the fit results. This is the x coordinate of the center of the intensity
     * profile
     *
     * @return the mean value
     */
    public double getMeanX() {
        return meanX[0];
    }

    /**
     * Returns the mean y value obtained from the fit results. This is the x coordinate of the center of the intensity
     * profile
     *
     * @return the mean value
     */
    public double getMeanY() {
        return meanY[0];
    }
}
