package de.desy.acop.video.analysis.gaussian;

import java.awt.image.BufferedImage;
import java.util.Arrays;
import java.util.function.Predicate;

import de.desy.acop.video.analysis.AnalysisImage;
import de.desy.acop.video.analysis.AnalysisType;
import de.desy.acop.video.analysis.FittingUtilities;
import de.desy.acop.video.analysis.ImageAnalysisEngine;
import de.desy.acop.video.analysis.ImageAnalyzer;
import de.desy.acop.video.timageio.TBufferedImage;
import de.desy.tine.types.IMAGE;

public class Gaussian2DImageAnalyzer extends ImageAnalyzer {

    private double[] previousStartValues = null;
    private double[] x = new double[1];
    private double[] y = new double[1];
    private double[] weights = new double[1];
    private int prevNRH = -1;
    private int prevNRW = -1;
    private int prevNRX = -1;
    private int prevNRY = -1;

    @Override
    public void clear() {
        previousStartValues = null;
        prevNRH = -1;
        prevNRW = -1;
        prevNRX = -1;
        prevNRY = -1;
    }

    @Override
    protected AnalysisImage doAnalysis(IMAGE newImage, double[] values, int iW, int iH, int rX, int rY, int rW, int rH,
            int r2X, int r2Y, int r2W, int r2H, int thresholdPointsCount, double threshold, boolean calculateThreshold,
            boolean performFit, boolean smoothing) {
        int i, j;
        performFit = true;
//        smoothing = true;
        if (smoothing) {
            // smooth, nearest neighbour
            values = FittingUtilities.smooth(values,iW);
            BufferedImage im = ImageAnalysisEngine.toBufferedImage(values,iW,newImage);
            newImage = TBufferedImage.toIMAGE(im);
        }
        double[] sideViewX = new double[iW];
        double[] sideViewY = new double[iH];
        double maxValue = -Double.MAX_VALUE;
        double minValue = Double.MAX_VALUE;
        double value;
        // if threshold is less than 0 (that can happen if such background is substracted, that the values of background
        // are larger than the image values), do not subtract the threshold.
        if (Double.isNaN(threshold)) {
            for (i = 0; i < iW; ++i) {
                for (j = 0; j < iH; ++j) {
                    value = values[i + j * iW];
                    if (minValue > value) minValue = value;
                    if (maxValue < value) maxValue = value;
                    sideViewX[i] += (value);
                    sideViewY[j] += (value);
                }
            }
        } else {
            for (i = 0; i < iW; ++i) {
                for (j = 0; j < iH; ++j) {
                    value = values[i + j * iW];
                    if (minValue > value) minValue = value;
                    if (value < threshold) continue;
                    if (maxValue < value) maxValue = value;
                    sideViewX[i] += (value - threshold);
                    sideViewY[j] += (value - threshold);
                }
            }
        }

        for (i = 0; i < values.length; i++) {
            if (values[i] < 0) values[i] = 0;
        }

        double[] xMinMax = findMinMax(sideViewX,0,rW);
        double[] yMinMax = findMinMax(sideViewY,0,rH);
        // perform the statistics calculation on the sideview data
        double xM = 0;
        double yM = 0;
        double suma = 0;
        double xSTD = 0;
        double ySTD = 0;
        double constX = xMinMax[0];
        double constY = yMinMax[0];
        int ii;
        double ivalue;
        for (ii = 0; ii < rW; ii++) {
            ivalue = sideViewX[ii+rX] - constX;
            if (ivalue < 0) ivalue = 0;
            xM += ii * ivalue;
            xSTD += ii * ii * ivalue;
            suma += ivalue;
        }
        xM = xM / suma;
        xSTD = Math.sqrt(xSTD / suma - xM * xM);
        xM += rX;
        suma = 0;
        for (ii = 0; ii < rH; ii++) {
            ivalue = sideViewY[ii + rY] - constY;
            if (ivalue < 0) ivalue = 0;
            yM += ii * ivalue;
            ySTD += ii * ii * ivalue;
            suma += ivalue;
        }
        yM = yM / suma;
        ySTD = Math.sqrt(ySTD / suma - yM * yM);
        yM += rY;
        double[] startValues = new double[] { 255, xM, xSTD, yM, ySTD, 0, 0 };

        int nrW = rW;
        int nrH = rH;
        int nrX = rX;
        int nrY = rY;
        if (nrW > 500 && nrW == iW) {
            // ROI should be 5 sigma, 2.5 in each direction
            nrW = (int)(5*startValues[2]);
            nrX = (int)(startValues[1] - 2.5*startValues[2]);
            if (nrX < 0) nrX = 0;
            if (nrX + nrW > iH) {
                nrW = iW - nrX;
            }
        }
        if (nrH > 500 && nrH == iH) {
            // ROI should be 5 sigma, 2.5 in each direction
            nrH = (int)(5*startValues[4]);
            nrY = (int)(startValues[3] - 2.5*startValues[4]);
            if (nrY < 0) nrY = 0;
            if (nrY + nrH > iH) {
                nrH = iH - nrY;
            }
        }
        if (nrW > rW) {
            nrX = rX;
            nrW = rW;
        }
        if (nrH > rH) {
            nrY = rY;
            nrH = rH;
        }

        boolean changeROI = Math.abs(nrW - prevNRW) > 10 ||
                Math.abs(nrH - prevNRH) > 10 ||
                Math.abs(nrX - prevNRX) > 10 ||
                Math.abs(nrY - prevNRY) > 10;

        if (changeROI && (nrW != iW || nrH != iH)) {
            double[] newValues = new double[nrW * nrH];
            int m = 0;
            for (j = nrY; j < nrY + nrH; j++) {
                for (i = nrX; i < nrX + nrW; i++) {
                    newValues[m++] = values[j * iW + i];
                }
            }
            values = newValues;
        }
        startValues[1] -= nrX;//+Math.random()*15-7.5;
        startValues[3] -= nrY;//+Math.random()*15-7.5;

        if (performFit) {
            if (changeROI && (x.length != nrW || x[0] != nrX)) {
                x = new double[nrW];
                for (i = 0; i < nrW; i++) {
                    x[i] = i;
                }
            }
            if (changeROI && (y.length != nrH || y[0] != nrY)) {
                y = new double[nrH];
                for (i = 0; i < nrH; i++) {
                    y[i] = i;
                }
            }
            if (changeROI && (weights.length != values.length)) {
                weights = new double[values.length];
                Arrays.fill(weights,1);
            }

            Predicate<double[]> faultyResultTester = s -> s[0] > 400 || s[0] <= 1 || s[1] <= 1 || s[1] >= iW
                    || s[2] <= 1 || s[2] >= iW / 2 || s[3] <= 1 || s[3] >= iH || s[4] <= 1 || s[4] >= iH / 2
                    || s[5] <= -Math.PI / 2 - 1e-8 || s[5] >= Math.PI / 2 + 1e-8 || s[6] >= 250;
            boolean fail = false;
            try {
                startValues = FittingUtilities.lmGauss2DR(x,y,values,weights,startValues);
            } catch (Exception e) {
                System.out.println(e);
                fail = true;
            }
            startValues[1] += nrX;
            startValues[3] += nrY;
            // fallback: try again with different starting vector
            if (fail || faultyResultTester.test(startValues)) {
                fail = false;
                if (previousStartValues != null) {
                    startValues = previousStartValues;
                    startValues[1] -= nrX;
                    startValues[3] -= nrY;
                    try {
                        startValues = FittingUtilities.lmGauss2DR(x,y,values,weights,startValues);
                    } catch (Exception e) {
                        System.out.println(e);
                        fail = true;
                    }
                    startValues[1] += nrX;
                    startValues[3] += nrY;
                }
                if (fail || faultyResultTester.test(startValues)) {
                    Arrays.fill(startValues,Double.NaN);
                    previousStartValues = null;
                } else {
                    previousStartValues = startValues;
                }
            } else {
                previousStartValues = startValues;
            }
        } else {
            Arrays.fill(startValues,Double.NaN);
            previousStartValues = null;
        }

        double angle = startValues[5] * 180 / Math.PI;
        if (angle < -45) {
            double temp = startValues[2];
            startValues[2] = startValues[4];
            startValues[4] = temp;
            angle = angle + 90;
        } else if (angle > 45) {
            double temp = startValues[2];
            startValues[2] = startValues[4];
            startValues[4] = temp;
            angle = angle - 90;
        }
        Gaussian2DImage newAImage = new Gaussian2DImage(newImage,rX,rY,rW,rH,r2X,r2Y,r2W,r2H,threshold,sideViewX,
                sideViewY,startValues[1],startValues[2],startValues[3],startValues[4],startValues[0],
                startValues[6],angle,thresholdPointsCount,calculateThreshold,performFit,smoothing);
        return newAImage;
    }

    @Override
    public AnalysisType getAnalysisType() {
        return AnalysisType.GAUSSIAN2D;
    }
}
