package de.desy.acop.video.analysis.gaussian;

import de.desy.acop.video.analysis.AnalysisImage;
import de.desy.acop.video.analysis.AnalysisType;
import de.desy.acop.video.analysis.FittingUtilities;
import de.desy.acop.video.analysis.ImageAnalyzer;
import de.desy.tine.types.IMAGE;

public class GaussianImageAnalyzer extends ImageAnalyzer {

    private double[] previousStartValuesX = null;
    private double[] previousStartValuesY = null;

    @Override
    public void clear() {
        previousStartValuesX = null;
        previousStartValuesY = null;
    }

    @Override
    protected AnalysisImage doAnalysis(IMAGE newImage, double[] values, int iW, int iH, int rX, int rY, int rW, int rH,
            int r2X, int r2Y, int r2W, int r2H, int thresholdPointsCount, double threshold, boolean calculateThreshold,
            boolean performFit, boolean smoothing) {
        
        int i, j;
        double[] sideViewX = new double[iW];
        double[] sideViewY = new double[iH];
        double maxValue = -Double.MAX_VALUE;
        double minValue = Double.MAX_VALUE;
        double value;
        //if threshold is less than 0 (that can happen if such background is substracted, that the values of background
        //are larger than the image values), do not subtract the threshold.
        if (Double.isNaN(threshold)) {
            for (i = 0; i < iW; ++i) {
                for (j = 0; j < iH; ++j) {
                    value = values[i + j * iW];
                    if (minValue > value) minValue = value;
                    if (maxValue < value) maxValue = value;
                    sideViewX[i] += (value);
                    sideViewY[j] += (value);
                }
            }
        } else {
            for (i = 0; i < iW; ++i) {
                for (j = 0; j < iH; ++j) {
                    value = values[i + j * iW];
                    if (minValue > value) minValue = value;
                    if (value < threshold) continue;
                    if (maxValue < value) maxValue = value;
                    sideViewX[i] += (value - threshold);
                    sideViewY[j] += (value - threshold);
                }
            }
        }
        //Do not do these: request from Gero
        //		for (i = 0; i < iW; i++) {
        //			if (sideViewX[i] < 0) {
        //				sideViewX[i] = 0;
        //			}
        //		}
        //		for (j = 0; j < iH; j++) {
        //			if (sideViewY[j] < 0) {
        //				sideViewY[j] = 0;
        //			}
        //		}
        //smoothing
        if (smoothing) {
            sideViewX = FittingUtilities.smooth(sideViewX);
            sideViewY = FittingUtilities.smooth(sideViewY);
        }
        //end of smoothing
        double[] roiSideViewX = new double[rW];
        double[] roiSideViewY = new double[rH];
        System.arraycopy(sideViewX,rX,roiSideViewX,0,rW);
        System.arraycopy(sideViewY,rY,roiSideViewY,0,rH);
        for (i = 0; i < values.length; i++) {
            if (values[i] < 0) values[i] = 0;
        }
        double sum = 0;
        double sumX = 0;
        double sumY = 0;
        double sumXX = 0;
        double sumYY = 0;
        double sumXY = 0;
        if (Double.isNaN(threshold)) {
            for (j = 0; j < rH; j++) {
                for (i = 0; i < rW; i++) {
                    value = values[i + rX + iW * (j + rY)];
                    sum += value;
                    sumX += value * (i + rX);
                    sumY += value * (j + rY);
                }
            }
        } else {
            for (j = 0; j < rH; j++) {
                for (i = 0; i < rW; i++) {
                    value = values[i + rX + iW * (j + rY)];
                    if (threshold < value) {
                        value -= threshold;
                        sum += value;
                        sumX += value * (i + rX);
                        sumY += value * (j + rY);
                    }
                }
            }
        }
        double Y = sumY / sum;
        double X = sumX / sum;
        double argx, argy;
        if (Double.isNaN(threshold)) {
            for (j = 0; j < rH; j++) {
                for (i = 0; i < rW; i++) {
                    value = values[i + rX + iW * (j + rY)];
                    argx = i + rX - X;
                    argy = j + rY - Y;
                    sumXX += value * argx * argx;
                    sumXY += value * argx * argy;
                    sumYY += value * argy * argy;
                }
            }
        } else {
            for (j = 0; j < rH; j++) {
                for (i = 0; i < rW; i++) {
                    value = values[i + rX + iW * (j + rY)];
                    if (threshold < value) {
                        value -= threshold;
                        argx = i + rX - X;
                        argy = j + rY - Y;
                        sumXX += value * argx * argx;
                        sumXY += value * argx * argy;
                        sumYY += value * argy * argy;
                    }
                }
            }
        }
        double XX = sumXX / sum;
        double XY = sumXY / sum;
        double YY = sumYY / sum;
        double b = -2 * (XX + YY);
        double c = (XX * YY) - (XY * XY);
        double D = Math.sqrt((b * b) - (4 * c));
        double A = (-b - D) / 2;
        double B = (-b + D) / 2;
        double phase = Math.atan2(2 * XX,YY - XX) * (90 / Math.PI);
        double[] xMinMax = findMinMax(roiSideViewX,0,rW);
        double[] yMinMax = findMinMax(roiSideViewY,0,rH);
        //perform the statistics calculation on the sideview data
        double xM = 0;
        double yM = 0;
        double suma = 0;
        double xSTD = 0;
        double ySTD = 0;
        double constX = xMinMax[0];
        double constY = yMinMax[0];
        for (i = 0; i < rW; i++) {
            value = roiSideViewX[i] - constX;
            if (value < 0) value = 0;
            xM += i * value;
            xSTD += i * i * value;
            suma += value;
        }
        xM = xM / suma;
        xSTD = Math.sqrt(xSTD / suma - xM * xM);
        suma = 0;
        for (i = 0; i < rH; i++) {
            value = roiSideViewY[i] - constY;
            if (value < 0) value = 0;
            yM += i * value;
            ySTD += i * i * value;
            suma += value;
        }
        yM = yM / suma;
        ySTD = Math.sqrt(ySTD / suma - yM * yM);
        double[] startValuesX = new double[]{xMinMax[1] - constX,xSTD,xM,constX,0};
        double[] startValuesY = new double[]{yMinMax[1] - constY,ySTD,yM,constY,0};
        if (performFit) {
            int l = roiSideViewX.length;
            double[] steps = new double[l];
            double[] weights = new double[l];
            for (i = 0; i < l; i++) {
                steps[i] = i;
                weights[i] = 1;
            }
            try {
                startValuesX = FittingUtilities.lmLinear(steps,roiSideViewX,weights,startValuesX);
                //fallback: try again with different starting vector
                if ((startValuesX[0] > 7000000 || startValuesX[0] <= 1 || startValuesX[1] <= 1)) {
                    if (previousStartValuesX != null) {
                        startValuesX = previousStartValuesX;
                        i = 0;
                        do {
                            startValuesX = FittingUtilities.lmLinear(steps,roiSideViewX,weights,startValuesX);
                            i++;
                        } while ((startValuesX[0] > 7000000 || startValuesX[0] <= 1 || startValuesX[1] <= 1) && i < 3);
                        if ((startValuesX[0] > 7000000 || startValuesX[0] <= 1 || startValuesX[1] <= 1)) {
                            startValuesX = new double[]{Double.NaN,Double.NaN,Double.NaN,Double.NaN,Double.NaN};
                            previousStartValuesX = null;
                        } else {
                            previousStartValuesX = startValuesX;
                        }
                    } else {
                        startValuesX = new double[]{Double.NaN,Double.NaN,Double.NaN,Double.NaN,Double.NaN};
                        previousStartValuesX = null;
                    }
                } else {
                    //in case there is not enough tail, the fit will over/undershot in the constant
                    //we need to add some tail and refit
                    if (rW < iW) {
                        double slope = startValuesX[4];
                        startValuesX = new double[]{xMinMax[1] - constX,xSTD,rX + xM,constX,0};
                        double[] roiSX = new double[iW];
                        for (i = 0; i < iW; i++) {
                            roiSX[i] = sideViewX[i] - slope * i;
                        }
                        steps = new double[iW];
                        weights = new double[iW];
                        for (i = 0; i < iW; i++) {
                            steps[i] = i;
                            weights[i] = 1;
                        }
                        startValuesX = FittingUtilities.lmLinear(steps,roiSX,weights,startValuesX);
                        if ((startValuesX[0] > 7000000 || startValuesX[0] <= 1 || startValuesX[1] <= 1)
                                && previousStartValuesX != null) {
                            startValuesX = previousStartValuesX;
                            do {
                                startValuesX = FittingUtilities.lmLinear(steps,roiSX,weights,startValuesX);
                                i++;
                            } while ((startValuesX[0] > 7000000 || startValuesX[0] <= 1 || startValuesX[1] <= 1)
                                    && i < 3);
                        }
                        startValuesX[4] += slope;
                        if ((startValuesX[0] > 7000000 || startValuesX[0] <= 1 || startValuesX[1] <= 1)) {
                            startValuesX = new double[]{Double.NaN,Double.NaN,Double.NaN,Double.NaN,Double.NaN};
                            previousStartValuesX = null;
                        } else {
                            previousStartValuesX = startValuesX;
                        }
                    } else {
                        previousStartValuesX = startValuesX;
                    }
                }
            } catch (Exception e) {
                startValuesX = new double[]{Double.NaN,Double.NaN,Double.NaN,Double.NaN,Double.NaN};
                previousStartValuesX = null;
            }
            try {
                l = roiSideViewY.length;
                steps = new double[l];
                weights = new double[l];
                for (i = 0; i < l; i++) {
                    steps[i] = i;
                    weights[i] = 1;
                }
                startValuesY = FittingUtilities.lmLinear(steps,roiSideViewY,weights,startValuesY);
                //fallback: try again with different starting vector
                if ((startValuesY[0] > 7000000 || startValuesY[0] <= 1 || startValuesY[1] <= 1)) {
                    if (previousStartValuesY != null) {
                        startValuesY = previousStartValuesY;
                        i = 0;
                        do {
                            startValuesY = FittingUtilities.lmLinear(steps,roiSideViewY,weights,startValuesY);
                            i++;
                        } while ((startValuesY[0] > 7000000 || startValuesY[0] <= 1 || startValuesY[1] <= 1) && i < 3);
                        if ((startValuesY[0] > 7000000 || startValuesY[0] <= 1 || startValuesY[1] <= 1)) {
                            startValuesY = new double[]{Double.NaN,Double.NaN,Double.NaN,Double.NaN,Double.NaN};
                            previousStartValuesY = null;
                        } else {
                            previousStartValuesY = startValuesY;
                        }
                    } else {
                        startValuesY = new double[]{Double.NaN,Double.NaN,Double.NaN,Double.NaN,Double.NaN};
                        previousStartValuesY = null;
                    }
                } else {
                    //in case there is not enough tail, the fit will over/undershot in the constant
                    //we need to add some tailvalues
                    if (rH < iH) {
                        double slope = startValuesY[4];
                        startValuesY = new double[]{yMinMax[1] - constY,ySTD,rY + yM,constY,0};
                        double[] roiSY = new double[iH];
                        for (i = 0; i < iH; i++) {
                            roiSY[i] = sideViewY[i] - slope * i;
                        }
                        steps = new double[iH];
                        weights = new double[iH];
                        for (i = 0; i < iH; i++) {
                            steps[i] = i;
                            weights[i] = 1;
                        }
                        i = 0;
                        startValuesY = FittingUtilities.lmLinear(steps,roiSY,weights,startValuesY);
                        if ((startValuesY[0] > 7000000 || startValuesY[0] <= 1 || startValuesY[1] <= 1)
                                && previousStartValuesY != null) {
                            startValuesY = previousStartValuesY;
                            do {
                                startValuesY = FittingUtilities.lmLinear(steps,roiSY,weights,startValuesY);
                                i++;
                            } while ((startValuesY[0] > 7000000 || startValuesY[0] <= 1 || startValuesY[1] <= 1)
                                    && i < 3);
                        }
                        startValuesY[4] += slope;
                        if ((startValuesY[0] > 7000000 || startValuesY[0] <= 1 || startValuesY[1] <= 1)) {
                            startValuesY = new double[]{Double.NaN,Double.NaN,Double.NaN,Double.NaN,Double.NaN};
                            previousStartValuesY = null;
                        } else {
                            previousStartValuesY = startValuesY;
                        }
                    } else {
                        previousStartValuesY = startValuesY;
                    }
                }
            } catch (Exception e) {
                startValuesY = new double[]{Double.NaN,Double.NaN,Double.NaN,Double.NaN,Double.NaN};
                previousStartValuesY = null;
            }
            if (FittingUtilities.isNaN(startValuesY)) {
                startValuesY = new double[]{yMinMax[1] - constY,ySTD,rY + yM,constY,0};
            }
            if (FittingUtilities.isNaN(startValuesX)) {
                startValuesX = new double[]{xMinMax[1] - constX,ySTD,rX + xM,constX,0};
            }
        } else {
            startValuesX = new double[]{Double.NaN,Double.NaN,Double.NaN,Double.NaN,Double.NaN};
            startValuesY = new double[]{Double.NaN,Double.NaN,Double.NaN,Double.NaN,Double.NaN};
            previousStartValuesY = null;
            previousStartValuesX = null;
        }
        try {
            if (!Double.isNaN(threshold)) {
                maxValue = maxValue - threshold;
            }
            AImage newAImage = new AImage(newImage,rX,rY,rW,rH,r2X,r2Y,r2W,r2H,threshold,sideViewX,rX + xM,xSTD,
                    xMinMax[1] - constX,constX,startValuesX[2],startValuesX[1],startValuesX[0],startValuesX[3],
                    startValuesX[4],sideViewY,rY + yM,ySTD,yMinMax[1] - constY,constY,startValuesY[2],startValuesY[1],
                    startValuesY[0],startValuesY[3],startValuesY[4],Math.sqrt(A),Math.sqrt(B),phase,maxValue,minValue,
                    thresholdPointsCount,calculateThreshold,performFit,smoothing);
            return newAImage;
        } catch (Exception e) {
            e.printStackTrace();
        }
        return null;
    }

    @Override
    public AnalysisType getAnalysisType() {
        return AnalysisType.GAUSSIAN;
    }
}
