package de.desy.acop.video.analysis.interferometer;

import de.desy.acop.video.analysis.AnalysisImage;
import de.desy.tine.types.IMAGE;

/**
 * <code>AbstractIFMImage</code> is the base class for interferometer image analysis.
 * 
 * @author <a href="mailto:jaka.bobnar@cosylab.com">Jaka Bobnar</a>
 */
public abstract class AbstractIFMImage extends AnalysisImage {

    private final double[] constantY = new double[1];
    private final double[] amplitudeY = new double[1];
    private final double[] asincY = new double[1];
    private final double[] psincY = new double[1];
    private final double[] acosY = new double[1];
    private final double[] pcosY = new double[1];
    private final double[] visY = new double[1];

    /**
     * Constructs a new IFMImage, with default values (empty image and 0 value parameters and results).
     */
    protected AbstractIFMImage(String name) {
        super(name);
        initStructure();
    }

    /**
     * Constructs a new IFMImage from the given IFMImage. This method creates a clone of the given image.
     * 
     * @param image the image
     */
    public AbstractIFMImage(AbstractIFMImage image) {
        this(image.getName(),image.getImage(),image.getRoiX(),image.getRoiY(),image.getRoiW(),image.getRoiH(),
                image.getRoi2X(),image.getRoi2Y(),image.getRoi2W(),image.getRoi2H(),image.getThreshold(),
                image.getSideViewX(),image.getSideViewY(),image.getConstantY(),image.getAmplitudeY(),image.getAsincY(),
                image.getPsincY(),image.getVisY(),image.getAcosY(),image.getPcosY(),image.getThresholdPoints(),
                image.isCalculateThreshold(),image.isPerformFit(),image.isPerformSmoothing());
    }

    /**
     * Constructs a new IFMImage.
     * 
     * @param image the raw image, which was analyzed and all data was obtained from
     * @param roiX the region of interest right coordinate (image is analyzed inside region of interest)
     * @param roiY the region of interest upper coordinate (image is analyzed inside region of interest)
     * @param roiW the width of region of interest (image is analyzed inside region of interest)
     * @param roiH the height of region of interest (image is analyzed inside region of interest)
     * @param roi2X the region of interest used for automatic threshold calculation right coordinate
     * @param roi2Y the region of interest used for automatic threshold calculation upper coordinate
     * @param roi2W the region of interest used for automatic threshold calculation width
     * @param roi2H the region of interest used for automatic threshold calculation height
     * @param threshold the used threshold value (calculated or user defined)
     * @param sideViewX horizontal side view values
     * @param sideViewY vertical side view values
     * @param constantY the background constant of the Y fit
     * @param amplitudeY the amplitude of the Y fit
     * @param asincY the asinc parameter of the Y fit
     * @param psincY the psinc parameter of the Y fit
     * @param acosY the acos parameter of the Y fit
     * @param pcosY the pcos parameter of the Y fit
     * @param visY the visibility parameter of the Y fit
     * @param thresholdPoints the number of points on which the threshold was calculated
     * @param calculateThreshold flag indicating if threshold calculation is automatic
     * @param performFit flag indicating if fitting was performed (if false the 'fit' parameters are a-priori invalid)
     * @param performSmoothing flag indicating if smoothing should be performed
     * @param vertical true if this image represents vertically oriented interferometer or false if it represents
     *        horizontally oriented interferometer
     */
    public AbstractIFMImage(String name, IMAGE image, int roiX, int roiY, int roiW, int roiH, int roi2X, int roi2Y,
            int roi2W, int roi2H, double threshold, double[] sideViewX, double[] sideViewY, double constantY,
            double amplitudeY, double asincY, double psincY, double visY, double acosY, double pcosY,
            int thresholdPoints, boolean calculateThreshold, boolean performFit, boolean performSmoothing) {
        super(name,image,roiX,roiY,roiW,roiH,roi2X,roi2Y,roi2W,roi2H,threshold,sideViewX,sideViewY,thresholdPoints,
                calculateThreshold,performFit,performSmoothing);
        this.constantY[0] = constantY;
        this.amplitudeY[0] = amplitudeY;
        this.pcosY[0] = pcosY;
        this.acosY[0] = acosY;
        this.asincY[0] = asincY;
        this.psincY[0] = psincY;
        this.visY[0] = visY;
        initStructure();
    }

    private void initStructure() {
        addField(constantY,"constantY");
        addField(amplitudeY,"amplitudeY");
        addField(asincY,"asincY");
        addField(psincY,"psincY");
        addField(acosY,"acosY");
        addField(pcosY,"pcosY");
        addField(visY,"visY");
    }

    /**
     * @return the constantY
     */
    public double getConstantY() {
        return constantY[0];
    }

    /**
     * @return the amplitudeY
     */
    public double getAmplitudeY() {
        return amplitudeY[0];
    }

    /**
     * @return the asincY
     */
    public double getAsincY() {
        return asincY[0];
    }

    /**
     * @return the psincY
     */
    public double getPsincY() {
        return psincY[0];
    }

    /**
     * @return the acosY
     */
    public double getAcosY() {
        return acosY[0];
    }

    /**
     * @return the pcosY
     */
    public double getPcosY() {
        return pcosY[0];
    }

    /**
     * @return the visY
     */
    public double getVisY() {
        return visY[0];
    }
}
