package de.desy.acop.video.analysis.interferometer;

import de.desy.tine.types.IMAGE;

/**
 * <code>DIFMImage</code> is video analysis object representing a two sided interferometer image.
 *
 * @author <a href="mailto:jaka.bobnar@cosylab.com">Jaka Bobnar</a>
 */
public class DIFMImage extends AbstractIFMImage {

    private static final String NAME = "DIFMISr4";
    private final double[] constantX = new double[1];
    private final double[] amplitudeX = new double[1];
    private final double[] asincX = new double[1];
    private final double[] psincX = new double[1];
    private final double[] acosX = new double[1];
    private final double[] pcosX = new double[1];
    private final double[] visX = new double[1];
    private final int[] originalRoi = new int[4];

    /**
     * Constructs a new DIFMImage, with default values (empty image and 0 value parameters and results).
     */
    public DIFMImage() {
        this(NAME);
    }

    /**
     * Constructs a new DIFMImage, with default values (empty image and 0 value parameters and results).
     *
     * @param name the name of the structure
     */
    protected DIFMImage(String name) {
        super(name);
        initStructure();
        initDone();
    }

    /**
     * Constructs a new DIFMImage from the given DIFMImage. This method creates a clone of the given image.
     *
     * @param image the image
     */
    public DIFMImage(DIFMImage image) {
        this(image.getImage(),image.getRoiX(),image.getRoiY(),image.getRoiW(),image.getRoiH(),image.getRoi2X(),
                image.getRoi2Y(),image.getRoi2W(),image.getRoi2H(),image.getThreshold(),image.getSideViewX(),
                image.getSideViewY(),image.getConstantY(),image.getAmplitudeY(),image.getAsincY(),image.getPsincY(),
                image.getVisY(),image.getAcosY(),image.getPcosY(),image.getConstantX(),image.getAmplitudeX(),
                image.getAsincX(),image.getPsincX(),image.getVisX(),image.getAcosX(),image.getPcosX(),
                image.getThresholdPoints(),image.isCalculateThreshold(),image.isPerformFit(),
                image.isPerformSmoothing());
    }

    /**
     * Constructs a new DIFMImage.
     *
     * @param image the raw image, which was analyzed and all data was obtained from
     * @param roiX the region of interest right coordinate (image is analyzed inside region of interest)
     * @param roiY the region of interest upper coordinate (image is analyzed inside region of interest)
     * @param roiW the width of region of interest (image is analyzed inside region of interest)
     * @param roiH the height of region of interest (image is analyzed inside region of interest)
     * @param roi2X the region of interest used for automatic threshold calculation right coordinate
     * @param roi2Y the region of interest used for automatic threshold calculation upper coordinate
     * @param roi2W the region of interest used for automatic threshold calculation width
     * @param roi2H the region of interest used for automatic threshold calculation height
     * @param threshold the used threshold value (calculated or user defined)
     * @param sideViewX horizontal side view values
     * @param sideViewY vertical side view values
     * @param constantY the background constant of the Y fit
     * @param amplitudeY the amplitude of the Y fit
     * @param asincY the asinc parameter of the Y fit
     * @param psincY the psinc parameter of the Y fit
     * @param acosY the acos parameter of the Y fit
     * @param pcosY the pcos parameter of the Y fit
     * @param visY the visibility parameter of the Y fit
     * @param constantX the background constant of the X fit
     * @param amplitudeX the amplitude of the X fit
     * @param asincX the asinc parameter of the X fit
     * @param psincX the psinc parameter of the X fit
     * @param acosX the acos parameter of the X fit
     * @param pcosX the pcos parameter of the X fit
     * @param visX the visibility parameter of the X fit
     * @param thresholdPoints the number of points on which the threshold was calculated
     * @param calculateThreshold flag indicating if threshold calculation is automatic
     * @param performFit flag indicating if fitting was performed (if false the 'fit' parameters are a-priori invalid)
     * @param performSmoothing flag indicating if smoothing should be performed
     * @param vertical true if this image represents vertically oriented interferometer or false if it represents
     *        horizontally oriented interferometer
     */
    public DIFMImage(IMAGE image, int roiX, int roiY, int roiW, int roiH, int roi2X, int roi2Y, int roi2W, int roi2H,
            double threshold, double[] sideViewX, double[] sideViewY, double constantY, double amplitudeY,
            double asincY, double psincY, double visY, double acosY, double pcosY, double constantX, double amplitudeX,
            double asincX, double psincX, double visX, double acosX, double pcosX, int thresholdPoints,
            boolean calculateThreshold, boolean performFit, boolean performSmoothing) {
        this(NAME,image,roiX,roiY,roiW,roiH,roi2X,roi2Y,roi2W,roi2H,threshold,sideViewX,sideViewY,constantY,amplitudeY,
                asincY,psincY,visY,acosY,pcosY,constantX,amplitudeX,asincX,psincX,visX,acosX,pcosX,thresholdPoints,
                calculateThreshold,performFit,performSmoothing);
    }

    protected DIFMImage(String name, IMAGE image, int roiX, int roiY, int roiW, int roiH, int roi2X, int roi2Y,
            int roi2W, int roi2H, double threshold, double[] sideViewX, double[] sideViewY, double constantY,
            double amplitudeY, double asincY, double psincY, double visY, double acosY, double pcosY, double constantX,
            double amplitudeX, double asincX, double psincX, double visX, double acosX, double pcosX,
            int thresholdPoints, boolean calculateThreshold, boolean performFit, boolean performSmoothing) {
        super(name,image,roiX,roiY,roiW,roiH,roi2X,roi2Y,roi2W,roi2H,threshold,sideViewX,sideViewY,constantY,amplitudeY,
                asincY,psincY,visY,acosY,pcosY,thresholdPoints,calculateThreshold,performFit,performSmoothing);
        this.constantX[0] = constantX;
        this.amplitudeX[0] = amplitudeX;
        this.pcosX[0] = pcosX;
        this.acosX[0] = acosX;
        this.asincX[0] = asincX;
        this.psincX[0] = psincX;
        this.visX[0] = visX;
        initStructure();
        initDone();
    }

    private void initStructure() {
        addField(constantX,"constantX");
        addField(amplitudeX,"amplitudeX");
        addField(asincX,"asincX");
        addField(psincX,"psincX");
        addField(acosX,"acosX");
        addField(pcosX,"pcosX");
        addField(visX,"visX");
    }

    /**
     * @return the constantX
     */
    public double getConstantX() {
        return constantX[0];
    }

    /**
     * @return the amplitudeX
     */
    public double getAmplitudeX() {
        return amplitudeX[0];
    }

    /**
     * @return the asincX
     */
    public double getAsincX() {
        return asincX[0];
    }

    /**
     * @return the psincX
     */
    public double getPsincX() {
        return psincX[0];
    }

    /**
     * @return the acosX
     */
    public double getAcosX() {
        return acosX[0];
    }

    /**
     * @return the pcosX
     */
    public double getPcosX() {
        return pcosX[0];
    }

    /**
     * @return the visX
     */
    public double getVisX() {
        return visX[0];
    }

    /**
     * Original region of intereset (x,y,w,h). This the region of intereset, which is not truncated to the width and
     * height of the strip that was used to calculate
     *
     * @return
     */
    public int[] getOriginalROI() {
        return originalRoi;
    }
}
