package de.desy.acop.video.analysis.interferometer;

import de.desy.tine.types.IMAGE;

/**
 * <code>IFMImage</code> is a single sided interferometer image analysis result.
 *
 * @author <a href="mailto:jaka.bobnar@cosylab.com">Jaka Bobnar</a>
 */
public class IFMImage extends AbstractIFMImage {

    private static final String NAME = "IFMISr4";

    private final byte[] vertical = new byte[1];

    /**
     * Constructs a new IFMImage, with default values (empty image and 0 value parameters and results).
     */
    public IFMImage() {
        super(NAME);
        initStructure();
    }

    /**
     * Constructs a new IFMImage, with default values (empty image and 0 value parameters and results).
     *
     * @param name the name of the structure
     */
    protected IFMImage(String name) {
        super(name);
        initStructure();
        initDone();
    }

    /**
     * Constructs a new IFMImage from the given IFMImage. This method creates a clone of the given image.
     *
     * @param image the image
     */
    public IFMImage(IFMImage image) {
        this(image.getImage(),image.getRoiX(),image.getRoiY(),image.getRoiW(),image.getRoiH(),image.getRoi2X(),
                image.getRoi2Y(),image.getRoi2W(),image.getRoi2H(),image.getThreshold(),image.getSideViewX(),
                image.getSideViewY(),image.getConstantY(),image.getAmplitudeY(),image.getAsincY(),image.getPsincY(),
                image.getVisY(),image.getAcosY(),image.getPcosY(),image.getThresholdPoints(),
                image.isCalculateThreshold(),image.isPerformFit(),image.isPerformSmoothing(),image.isVertical());
    }

    /**
     * Constructs a new IFMImage.
     *
     * @param name the name of the structure
     * @param image the raw image, which was analyzed and all data was obtained from
     * @param roiX the region of interest right coordinate (image is analyzed inside region of interest)
     * @param roiY the region of interest upper coordinate (image is analyzed inside region of interest)
     * @param roiW the width of region of interest (image is analyzed inside region of interest)
     * @param roiH the height of region of interest (image is analyzed inside region of interest)
     * @param roi2X the region of interest used for automatic threshold calculation right coordinate
     * @param roi2Y the region of interest used for automatic threshold calculation upper coordinate
     * @param roi2W the region of interest used for automatic threshold calculation width
     * @param roi2H the region of interest used for automatic threshold calculation height
     * @param threshold the used threshold value (calculated or user defined)
     * @param sideViewX horizontal side view values
     * @param sideViewY vertical side view values
     * @param constantY the background constant of the Y fit
     * @param amplitudeY the amplitude of the Y fit
     * @param asincY the asinc parameter of the Y fit
     * @param psincY the psinc parameter of the Y fit
     * @param acosY the acos parameter of the Y fit
     * @param pcosY the pcos parameter of the Y fit
     * @param visY the visibility parameter of the Y fit
     * @param thresholdPoints the number of points on which the threshold was calculated
     * @param calculateThreshold flag indicating if threshold calculation is automatic
     * @param performFit flag indicating if fitting was performed (if false the 'fit' parameters are a-priori invalid)
     * @param performSmoothing flag indicating if smoothing should be performed
     * @param vertical true if this image represents vertically oriented interferometer or false if it represents
     *        horizontally oriented interferometer
     */
    protected IFMImage(String name, IMAGE image, int roiX, int roiY, int roiW, int roiH, int roi2X, int roi2Y, int roi2W,
            int roi2H, double threshold, double[] sideViewX, double[] sideViewY, double constantY, double amplitudeY,
            double asincY, double psincY, double visY, double acosY, double pcosY, int thresholdPoints,
            boolean calculateThreshold, boolean performFit, boolean performSmoothing, boolean vertical) {
        super(name,image,roiX,roiY,roiW,roiH,roi2X,roi2Y,roi2W,roi2H,threshold,sideViewX,sideViewY,constantY,
                amplitudeY,asincY,psincY,visY,acosY,pcosY,thresholdPoints,calculateThreshold,performFit,
                performSmoothing);
        this.vertical[0] = vertical ? (byte)1 : (byte)0;
        initStructure();
        initDone();
    }

    /**
     * Constructs a new IFMImage.
     *
     * @param image the raw image, which was analyzed and all data was obtained from
     * @param roiX the region of interest right coordinate (image is analyzed inside region of interest)
     * @param roiY the region of interest upper coordinate (image is analyzed inside region of interest)
     * @param roiW the width of region of interest (image is analyzed inside region of interest)
     * @param roiH the height of region of interest (image is analyzed inside region of interest)
     * @param roi2X the region of interest used for automatic threshold calculation right coordinate
     * @param roi2Y the region of interest used for automatic threshold calculation upper coordinate
     * @param roi2W the region of interest used for automatic threshold calculation width
     * @param roi2H the region of interest used for automatic threshold calculation height
     * @param threshold the used threshold value (calculated or user defined)
     * @param sideViewX horizontal side view values
     * @param sideViewY vertical side view values
     * @param constantY the background constant of the Y fit
     * @param amplitudeY the amplitude of the Y fit
     * @param asincY the asinc parameter of the Y fit
     * @param psincY the psinc parameter of the Y fit
     * @param acosY the acos parameter of the Y fit
     * @param pcosY the pcos parameter of the Y fit
     * @param visY the visibility parameter of the Y fit
     * @param thresholdPoints the number of points on which the threshold was calculated
     * @param calculateThreshold flag indicating if threshold calculation is automatic
     * @param performFit flag indicating if fitting was performed (if false the 'fit' parameters are a-priori invalid)
     * @param performSmoothing flag indicating if smoothing should be performed
     * @param vertical true if this image represents vertically oriented interferometer or false if it represents
     *        horizontally oriented interferometer
     */
    public IFMImage(IMAGE image, int roiX, int roiY, int roiW, int roiH, int roi2X, int roi2Y, int roi2W,
            int roi2H, double threshold, double[] sideViewX, double[] sideViewY, double constantY, double amplitudeY,
            double asincY, double psincY, double visY, double acosY, double pcosY, int thresholdPoints,
            boolean calculateThreshold, boolean performFit, boolean performSmoothing, boolean vertical) {
        this(NAME,image,roiX,roiY,roiW,roiH,roi2X,roi2Y,roi2W,roi2H,threshold,sideViewX,sideViewY,constantY,
                amplitudeY,asincY,psincY,visY,acosY,pcosY,thresholdPoints,calculateThreshold,performFit,
                performSmoothing,vertical);
    }

    private void initStructure() {
        addField(vertical,"vertical");
    }

    /**
     * @return true if this image is vertical or false if horizontal
     */
    public boolean isVertical() {
        return vertical[0] == 1;
    }

    /**
     * @return 1 if the image is vertical or 0 if horizontal
     */
    public byte[] getVertical() {
        return vertical;
    }
}
