package de.desy.acop.video.displayer;

/**
 * <code>HuffmanDecompression</code> is a wrapper class around HuffYUV open source
 * video codec v1.2.3 huffman-based video data decompression.<br>
 * <br>
 * It basically is a pure Java rewrite of simple C++ source code of HuffYUV 1.2.3.
 * The HuffYUV site where the original C++ code was taken from in 2002 was situated at 
 * Berkeley University, but is no longer available. A archived version of it can be 
 * found at http://neuron2.net/www.math.berkeley.edu/benrg/huffyuv.html 
 * (last checked July 19, 2008). The original author of the C++ source is 
 * Ben Rudiak-Gould.<br>
 * <br>
 *  This so-called HuffYUV compression is lossless and was adapted and used at 
 *  Video System v2 in order to compress/uncompress grayscale data before/after 
 *  transferring it over the network.<br>
 *  <br>
 *  Speed of pure Java implementation was (after optimization) much better than thought. 
 *  I could not believe how fast Java code can get. Only necessary method to call is 
 *  ({@link DecompressHuffYUV( byte[], int, byte [], int, int)}).
 *  
 * @author <a href="mailto:stefan.weisse@desy.de">Stefan Weisse</a>
 * @version $Id: Templates.xml,v 1.10 2008/06/19 17:40:13 sweisse Exp $
 *
 */

public final class HuffmanDecompression {

	/** HuffYUV table taken from C++ source */
	/* 2 512 */ 
	private static final int[][] decode_luma9 = {
			 { 0,  0},{ 0,  0},{ 0,  0},{ 0,  0},{ 0,  0},{ 0,  0},{ 0,  0},{ 0,  0},{ 0,  0},{ 0,  0},{ 0,  0},{ 0,  0},{ 0,  0},{ 0,  0},{ 0,  0},{ 0,  0},
			 { 0,  0},{ 0,  0},{ 0,  0},{ 0,  0},{ 0,  0},{ 0,  0},{ 0,  0},{ 0,  0},{ 0,  0},{ 0,  0},{ 0,  0},{ 0,  0},{ 0,  0},{ 0,  0},{ 0,  0},{ 0,  0},
			 { 0,  0},{ 0,  0},{ 0,  0},{ 9,213},{ 9, 31},{ 9,214},{ 9, 30},{ 9,215},{ 9,216},{ 9, 29},{ 9,217},{ 9,218},{ 9, 28},{ 9,219},{ 9, 27},{ 9,220},
			 { 9,221},{ 9, 26},{ 9, 25},{ 9,222},{ 9,223},{ 9, 24},{ 9,224},{ 9,225},{ 9, 23},{ 9,226},{ 9, 22},{ 9,227},{ 9,228},{ 9, 21},{ 9,229},{ 9,230},
			 { 9, 20},{ 9, 19},{ 9,231},{ 9,232},{ 9, 18},{ 9,233},{ 9, 17},{ 9,234},{ 9,235},{ 9, 16},{ 8, 15},{ 8, 15},{ 8,236},{ 8,236},{ 8, 14},{ 8, 14},
			 { 8,237},{ 8,237},{ 8, 13},{ 8, 13},{ 8,238},{ 8,238},{ 8,239},{ 8,239},{ 8, 12},{ 8, 12},{ 8, 11},{ 8, 11},{ 8,240},{ 8,240},{ 8,241},{ 8,241},
			 { 8, 10},{ 8, 10},{ 8,242},{ 8,242},{ 8,  9},{ 8,  9},{ 8,243},{ 8,243},{ 7,244},{ 7,244},{ 7,244},{ 7,244},{ 7,  8},{ 7,  8},{ 7,  8},{ 7,  8},
			 { 7,245},{ 7,245},{ 7,245},{ 7,245},{ 7,  7},{ 7,  7},{ 7,  7},{ 7,  7},{ 7,246},{ 7,246},{ 7,246},{ 7,246},{ 7,247},{ 7,247},{ 7,247},{ 7,247},
			 { 7,  6},{ 7,  6},{ 7,  6},{ 7,  6},{ 7,248},{ 7,248},{ 7,248},{ 7,248},{ 7,249},{ 7,249},{ 7,249},{ 7,249},{ 7,  5},{ 7,  5},{ 7,  5},{ 7,  5},
			 { 6,250},{ 6,250},{ 6,250},{ 6,250},{ 6,250},{ 6,250},{ 6,250},{ 6,250},{ 6,251},{ 6,251},{ 6,251},{ 6,251},{ 6,251},{ 6,251},{ 6,251},{ 6,251},
			 { 5,  4},{ 5,  4},{ 5,  4},{ 5,  4},{ 5,  4},{ 5,  4},{ 5,  4},{ 5,  4},{ 5,  4},{ 5,  4},{ 5,  4},{ 5,  4},{ 5,  4},{ 5,  4},{ 5,  4},{ 5,  4},
			 { 5,252},{ 5,252},{ 5,252},{ 5,252},{ 5,252},{ 5,252},{ 5,252},{ 5,252},{ 5,252},{ 5,252},{ 5,252},{ 5,252},{ 5,252},{ 5,252},{ 5,252},{ 5,252},
			 { 5,  3},{ 5,  3},{ 5,  3},{ 5,  3},{ 5,  3},{ 5,  3},{ 5,  3},{ 5,  3},{ 5,  3},{ 5,  3},{ 5,  3},{ 5,  3},{ 5,  3},{ 5,  3},{ 5,  3},{ 5,  3},
			 { 5,253},{ 5,253},{ 5,253},{ 5,253},{ 5,253},{ 5,253},{ 5,253},{ 5,253},{ 5,253},{ 5,253},{ 5,253},{ 5,253},{ 5,253},{ 5,253},{ 5,253},{ 5,253},
			 { 4,254},{ 4,254},{ 4,254},{ 4,254},{ 4,254},{ 4,254},{ 4,254},{ 4,254},{ 4,254},{ 4,254},{ 4,254},{ 4,254},{ 4,254},{ 4,254},{ 4,254},{ 4,254},
			 { 4,254},{ 4,254},{ 4,254},{ 4,254},{ 4,254},{ 4,254},{ 4,254},{ 4,254},{ 4,254},{ 4,254},{ 4,254},{ 4,254},{ 4,254},{ 4,254},{ 4,254},{ 4,254},
			 { 4,255},{ 4,255},{ 4,255},{ 4,255},{ 4,255},{ 4,255},{ 4,255},{ 4,255},{ 4,255},{ 4,255},{ 4,255},{ 4,255},{ 4,255},{ 4,255},{ 4,255},{ 4,255},
			 { 4,255},{ 4,255},{ 4,255},{ 4,255},{ 4,255},{ 4,255},{ 4,255},{ 4,255},{ 4,255},{ 4,255},{ 4,255},{ 4,255},{ 4,255},{ 4,255},{ 4,255},{ 4,255},
			 { 4,  1},{ 4,  1},{ 4,  1},{ 4,  1},{ 4,  1},{ 4,  1},{ 4,  1},{ 4,  1},{ 4,  1},{ 4,  1},{ 4,  1},{ 4,  1},{ 4,  1},{ 4,  1},{ 4,  1},{ 4,  1},
			 { 4,  1},{ 4,  1},{ 4,  1},{ 4,  1},{ 4,  1},{ 4,  1},{ 4,  1},{ 4,  1},{ 4,  1},{ 4,  1},{ 4,  1},{ 4,  1},{ 4,  1},{ 4,  1},{ 4,  1},{ 4,  1},
			 { 3,  2},{ 3,  2},{ 3,  2},{ 3,  2},{ 3,  2},{ 3,  2},{ 3,  2},{ 3,  2},{ 3,  2},{ 3,  2},{ 3,  2},{ 3,  2},{ 3,  2},{ 3,  2},{ 3,  2},{ 3,  2},
			 { 3,  2},{ 3,  2},{ 3,  2},{ 3,  2},{ 3,  2},{ 3,  2},{ 3,  2},{ 3,  2},{ 3,  2},{ 3,  2},{ 3,  2},{ 3,  2},{ 3,  2},{ 3,  2},{ 3,  2},{ 3,  2},
			 { 3,  2},{ 3,  2},{ 3,  2},{ 3,  2},{ 3,  2},{ 3,  2},{ 3,  2},{ 3,  2},{ 3,  2},{ 3,  2},{ 3,  2},{ 3,  2},{ 3,  2},{ 3,  2},{ 3,  2},{ 3,  2},
			 { 3,  2},{ 3,  2},{ 3,  2},{ 3,  2},{ 3,  2},{ 3,  2},{ 3,  2},{ 3,  2},{ 3,  2},{ 3,  2},{ 3,  2},{ 3,  2},{ 3,  2},{ 3,  2},{ 3,  2},{ 3,  2},
			 { 2,  0},{ 2,  0},{ 2,  0},{ 2,  0},{ 2,  0},{ 2,  0},{ 2,  0},{ 2,  0},{ 2,  0},{ 2,  0},{ 2,  0},{ 2,  0},{ 2,  0},{ 2,  0},{ 2,  0},{ 2,  0},
			 { 2,  0},{ 2,  0},{ 2,  0},{ 2,  0},{ 2,  0},{ 2,  0},{ 2,  0},{ 2,  0},{ 2,  0},{ 2,  0},{ 2,  0},{ 2,  0},{ 2,  0},{ 2,  0},{ 2,  0},{ 2,  0},
			 { 2,  0},{ 2,  0},{ 2,  0},{ 2,  0},{ 2,  0},{ 2,  0},{ 2,  0},{ 2,  0},{ 2,  0},{ 2,  0},{ 2,  0},{ 2,  0},{ 2,  0},{ 2,  0},{ 2,  0},{ 2,  0},
			 { 2,  0},{ 2,  0},{ 2,  0},{ 2,  0},{ 2,  0},{ 2,  0},{ 2,  0},{ 2,  0},{ 2,  0},{ 2,  0},{ 2,  0},{ 2,  0},{ 2,  0},{ 2,  0},{ 2,  0},{ 2,  0},
			 { 2,  0},{ 2,  0},{ 2,  0},{ 2,  0},{ 2,  0},{ 2,  0},{ 2,  0},{ 2,  0},{ 2,  0},{ 2,  0},{ 2,  0},{ 2,  0},{ 2,  0},{ 2,  0},{ 2,  0},{ 2,  0},
			 { 2,  0},{ 2,  0},{ 2,  0},{ 2,  0},{ 2,  0},{ 2,  0},{ 2,  0},{ 2,  0},{ 2,  0},{ 2,  0},{ 2,  0},{ 2,  0},{ 2,  0},{ 2,  0},{ 2,  0},{ 2,  0},
			 { 2,  0},{ 2,  0},{ 2,  0},{ 2,  0},{ 2,  0},{ 2,  0},{ 2,  0},{ 2,  0},{ 2,  0},{ 2,  0},{ 2,  0},{ 2,  0},{ 2,  0},{ 2,  0},{ 2,  0},{ 2,  0},
			 { 2,  0},{ 2,  0},{ 2,  0},{ 2,  0},{ 2,  0},{ 2,  0},{ 2,  0},{ 2,  0},{ 2,  0},{ 2,  0},{ 2,  0},{ 2,  0},{ 2,  0},{ 2,  0},{ 2,  0},{ 2,  0}
			};

	/** HuffYUV table taken from C++ source */
	/* 2 512 */
	private static final int[][] decode_luma12 = {
			 { 0,  0},{ 0,  0},{ 0,  0},{ 0,  0},{ 0,  0},{ 0,  0},{ 0,  0},{ 0,  0},{ 0,  0},{ 0,  0},{ 0,  0},{ 0,  0},{ 0,  0},{ 0,  0},{ 0,  0},{12, 94},
			 {12,177},{12, 93},{12, 92},{12,178},{12, 91},{12, 90},{12,179},{12, 89},{12, 88},{12,180},{12, 87},{12, 86},{12,181},{12, 85},{12, 84},{12, 83},
			 {12,182},{12, 82},{12, 81},{12, 80},{12,183},{12, 79},{11, 78},{11, 78},{11, 77},{11, 77},{11,184},{11,184},{11, 76},{11, 76},{11, 75},{11, 75},
			 {11, 74},{11, 74},{11,185},{11,185},{11, 73},{11, 73},{11, 72},{11, 72},{11, 71},{11, 71},{11,186},{11,186},{11, 70},{11, 70},{11, 69},{11, 69},
			 {11, 68},{11, 68},{11,187},{11,187},{11, 67},{11, 67},{11, 66},{11, 66},{11, 65},{11, 65},{11, 64},{11, 64},{11,188},{11,188},{11, 63},{11, 63},
			 {11, 62},{11, 62},{11, 61},{11, 61},{11,189},{11,189},{11, 60},{11, 60},{11, 59},{11, 59},{11,190},{11,190},{11, 58},{11, 58},{11, 57},{11, 57},
			 {11, 56},{11, 56},{11,191},{11,191},{10, 55},{10, 55},{10, 55},{10, 55},{10, 54},{10, 54},{10, 54},{10, 54},{10, 53},{10, 53},{10, 53},{10, 53},
			 {10,192},{10,192},{10,192},{10,192},{10, 52},{10, 52},{10, 52},{10, 52},{10, 51},{10, 51},{10, 51},{10, 51},{10,193},{10,193},{10,193},{10,193},
			 {10, 50},{10, 50},{10, 50},{10, 50},{10, 49},{10, 49},{10, 49},{10, 49},{10,194},{10,194},{10,194},{10,194},{10, 48},{10, 48},{10, 48},{10, 48},
			 {10, 47},{10, 47},{10, 47},{10, 47},{10,195},{10,195},{10,195},{10,195},{10, 46},{10, 46},{10, 46},{10, 46},{10, 45},{10, 45},{10, 45},{10, 45},
			 {10,196},{10,196},{10,196},{10,196},{10, 44},{10, 44},{10, 44},{10, 44},{10,197},{10,197},{10,197},{10,197},{10, 43},{10, 43},{10, 43},{10, 43},
			 {10,198},{10,198},{10,198},{10,198},{10, 42},{10, 42},{10, 42},{10, 42},{10, 41},{10, 41},{10, 41},{10, 41},{10,199},{10,199},{10,199},{10,199},
			 {10, 40},{10, 40},{10, 40},{10, 40},{10,200},{10,200},{10,200},{10,200},{10,201},{10,201},{10,201},{10,201},{10, 39},{10, 39},{10, 39},{10, 39},
			 {10,202},{10,202},{10,202},{10,202},{10, 38},{10, 38},{10, 38},{10, 38},{10,203},{10,203},{10,203},{10,203},{10, 37},{10, 37},{10, 37},{10, 37},
			 {10,204},{10,204},{10,204},{10,204},{10,205},{10,205},{10,205},{10,205},{10, 36},{10, 36},{10, 36},{10, 36},{10,206},{10,206},{10,206},{10,206},
			 {10,207},{10,207},{10,207},{10,207},{10, 35},{10, 35},{10, 35},{10, 35},{10,208},{10,208},{10,208},{10,208},{10, 34},{10, 34},{10, 34},{10, 34},
			 {10,209},{10,209},{10,209},{10,209},{10, 33},{10, 33},{10, 33},{10, 33},{10,210},{10,210},{10,210},{10,210},{10,211},{10,211},{10,211},{10,211},
			 {10, 32},{10, 32},{10, 32},{10, 32},{10,212},{10,212},{10,212},{10,212},{ 9,213},{ 9,213},{ 9,213},{ 9,213},{ 9,213},{ 9,213},{ 9,213},{ 9,213},
			 { 9, 31},{ 9, 31},{ 9, 31},{ 9, 31},{ 9, 31},{ 9, 31},{ 9, 31},{ 9, 31},{ 9,214},{ 9,214},{ 9,214},{ 9,214},{ 9,214},{ 9,214},{ 9,214},{ 9,214},
			 { 9, 30},{ 9, 30},{ 9, 30},{ 9, 30},{ 9, 30},{ 9, 30},{ 9, 30},{ 9, 30},{ 9,215},{ 9,215},{ 9,215},{ 9,215},{ 9,215},{ 9,215},{ 9,215},{ 9,215},
			 { 9,216},{ 9,216},{ 9,216},{ 9,216},{ 9,216},{ 9,216},{ 9,216},{ 9,216},{ 9, 29},{ 9, 29},{ 9, 29},{ 9, 29},{ 9, 29},{ 9, 29},{ 9, 29},{ 9, 29},
			 { 9,217},{ 9,217},{ 9,217},{ 9,217},{ 9,217},{ 9,217},{ 9,217},{ 9,217},{ 9,218},{ 9,218},{ 9,218},{ 9,218},{ 9,218},{ 9,218},{ 9,218},{ 9,218},
			 { 9, 28},{ 9, 28},{ 9, 28},{ 9, 28},{ 9, 28},{ 9, 28},{ 9, 28},{ 9, 28},{ 9,219},{ 9,219},{ 9,219},{ 9,219},{ 9,219},{ 9,219},{ 9,219},{ 9,219},
			 { 9, 27},{ 9, 27},{ 9, 27},{ 9, 27},{ 9, 27},{ 9, 27},{ 9, 27},{ 9, 27},{ 9,220},{ 9,220},{ 9,220},{ 9,220},{ 9,220},{ 9,220},{ 9,220},{ 9,220},
			 { 9,221},{ 9,221},{ 9,221},{ 9,221},{ 9,221},{ 9,221},{ 9,221},{ 9,221},{ 9, 26},{ 9, 26},{ 9, 26},{ 9, 26},{ 9, 26},{ 9, 26},{ 9, 26},{ 9, 26},
			 { 9, 25},{ 9, 25},{ 9, 25},{ 9, 25},{ 9, 25},{ 9, 25},{ 9, 25},{ 9, 25},{ 9,222},{ 9,222},{ 9,222},{ 9,222},{ 9,222},{ 9,222},{ 9,222},{ 9,222},
			 { 9,223},{ 9,223},{ 9,223},{ 9,223},{ 9,223},{ 9,223},{ 9,223},{ 9,223},{ 9, 24},{ 9, 24},{ 9, 24},{ 9, 24},{ 9, 24},{ 9, 24},{ 9, 24},{ 9, 24},
			 { 9,224},{ 9,224},{ 9,224},{ 9,224},{ 9,224},{ 9,224},{ 9,224},{ 9,224},{ 9,225},{ 9,225},{ 9,225},{ 9,225},{ 9,225},{ 9,225},{ 9,225},{ 9,225},
			 { 9, 23},{ 9, 23},{ 9, 23},{ 9, 23},{ 9, 23},{ 9, 23},{ 9, 23},{ 9, 23},{ 9,226},{ 9,226},{ 9,226},{ 9,226},{ 9,226},{ 9,226},{ 9,226},{ 9,226},
			 { 9, 22},{ 9, 22},{ 9, 22},{ 9, 22},{ 9, 22},{ 9, 22},{ 9, 22},{ 9, 22},{ 9,227},{ 9,227},{ 9,227},{ 9,227},{ 9,227},{ 9,227},{ 9,227},{ 9,227},
			 { 9,228},{ 9,228},{ 9,228},{ 9,228},{ 9,228},{ 9,228},{ 9,228},{ 9,228},{ 9, 21},{ 9, 21},{ 9, 21},{ 9, 21},{ 9, 21},{ 9, 21},{ 9, 21},{ 9, 21},
			 { 9,229},{ 9,229},{ 9,229},{ 9,229},{ 9,229},{ 9,229},{ 9,229},{ 9,229},{ 9,230},{ 9,230},{ 9,230},{ 9,230},{ 9,230},{ 9,230},{ 9,230},{ 9,230}
			};

	/** HuffYUV table taken from C++ source */
	/* 2 512 */	
	private static final int[][] decode_luma17 = {
			 {17,139},{17,136},{17,134},{17,141},{17,132},{17,146},{17,143},{17,137},{17,138},{17,133},{17,145},{17,135},{17,144},{17,131},{17,142},{17,147},
			 {17,140},{17,129},{17,148},{17,130},{16,150},{16,150},{16,149},{16,149},{16,128},{16,128},{16,151},{16,151},{16,127},{16,127},{16,152},{16,152},
			 {16,153},{16,153},{16,126},{16,126},{16,125},{16,125},{16,154},{16,154},{16,155},{16,155},{16,124},{16,124},{16,156},{16,156},{16,123},{16,123},
			 {16,157},{16,157},{16,122},{16,122},{16,121},{16,121},{16,158},{16,158},{15,120},{15,120},{15,120},{15,120},{15,159},{15,159},{15,159},{15,159},
			 {15,119},{15,119},{15,119},{15,119},{15,118},{15,118},{15,118},{15,118},{15,160},{15,160},{15,160},{15,160},{15,117},{15,117},{15,117},{15,117},
			 {15,161},{15,161},{15,161},{15,161},{15,116},{15,116},{15,116},{15,116},{15,162},{15,162},{15,162},{15,162},{15,115},{15,115},{15,115},{15,115},
			 {15,163},{15,163},{15,163},{15,163},{15,114},{15,114},{15,114},{15,114},{15,164},{15,164},{15,164},{15,164},{15,113},{15,113},{15,113},{15,113},
			 {14,165},{14,165},{14,165},{14,165},{14,165},{14,165},{14,165},{14,165},{14,112},{14,112},{14,112},{14,112},{14,112},{14,112},{14,112},{14,112},
			 {14,111},{14,111},{14,111},{14,111},{14,111},{14,111},{14,111},{14,111},{14,166},{14,166},{14,166},{14,166},{14,166},{14,166},{14,166},{14,166},
			 {14,110},{14,110},{14,110},{14,110},{14,110},{14,110},{14,110},{14,110},{14,109},{14,109},{14,109},{14,109},{14,109},{14,109},{14,109},{14,109},
			 {14,167},{14,167},{14,167},{14,167},{14,167},{14,167},{14,167},{14,167},{14,108},{14,108},{14,108},{14,108},{14,108},{14,108},{14,108},{14,108},
			 {14,168},{14,168},{14,168},{14,168},{14,168},{14,168},{14,168},{14,168},{14,107},{14,107},{14,107},{14,107},{14,107},{14,107},{14,107},{14,107},
			 {14,169},{14,169},{14,169},{14,169},{14,169},{14,169},{14,169},{14,169},{14,106},{14,106},{14,106},{14,106},{14,106},{14,106},{14,106},{14,106},
			 {14,105},{14,105},{14,105},{14,105},{14,105},{14,105},{14,105},{14,105},{14,170},{14,170},{14,170},{14,170},{14,170},{14,170},{14,170},{14,170},
			 {13,104},{13,104},{13,104},{13,104},{13,104},{13,104},{13,104},{13,104},{13,104},{13,104},{13,104},{13,104},{13,104},{13,104},{13,104},{13,104},
			 {13,171},{13,171},{13,171},{13,171},{13,171},{13,171},{13,171},{13,171},{13,171},{13,171},{13,171},{13,171},{13,171},{13,171},{13,171},{13,171},
			 {13,103},{13,103},{13,103},{13,103},{13,103},{13,103},{13,103},{13,103},{13,103},{13,103},{13,103},{13,103},{13,103},{13,103},{13,103},{13,103},
			 {13,102},{13,102},{13,102},{13,102},{13,102},{13,102},{13,102},{13,102},{13,102},{13,102},{13,102},{13,102},{13,102},{13,102},{13,102},{13,102},
			 {13,172},{13,172},{13,172},{13,172},{13,172},{13,172},{13,172},{13,172},{13,172},{13,172},{13,172},{13,172},{13,172},{13,172},{13,172},{13,172},
			 {13,101},{13,101},{13,101},{13,101},{13,101},{13,101},{13,101},{13,101},{13,101},{13,101},{13,101},{13,101},{13,101},{13,101},{13,101},{13,101},
			 {13,173},{13,173},{13,173},{13,173},{13,173},{13,173},{13,173},{13,173},{13,173},{13,173},{13,173},{13,173},{13,173},{13,173},{13,173},{13,173},
			 {13,100},{13,100},{13,100},{13,100},{13,100},{13,100},{13,100},{13,100},{13,100},{13,100},{13,100},{13,100},{13,100},{13,100},{13,100},{13,100},
			 {13, 99},{13, 99},{13, 99},{13, 99},{13, 99},{13, 99},{13, 99},{13, 99},{13, 99},{13, 99},{13, 99},{13, 99},{13, 99},{13, 99},{13, 99},{13, 99},
			 {13,174},{13,174},{13,174},{13,174},{13,174},{13,174},{13,174},{13,174},{13,174},{13,174},{13,174},{13,174},{13,174},{13,174},{13,174},{13,174},
			 {13, 98},{13, 98},{13, 98},{13, 98},{13, 98},{13, 98},{13, 98},{13, 98},{13, 98},{13, 98},{13, 98},{13, 98},{13, 98},{13, 98},{13, 98},{13, 98},
			 {13,175},{13,175},{13,175},{13,175},{13,175},{13,175},{13,175},{13,175},{13,175},{13,175},{13,175},{13,175},{13,175},{13,175},{13,175},{13,175},
			 {13, 97},{13, 97},{13, 97},{13, 97},{13, 97},{13, 97},{13, 97},{13, 97},{13, 97},{13, 97},{13, 97},{13, 97},{13, 97},{13, 97},{13, 97},{13, 97},
			 {13, 96},{13, 96},{13, 96},{13, 96},{13, 96},{13, 96},{13, 96},{13, 96},{13, 96},{13, 96},{13, 96},{13, 96},{13, 96},{13, 96},{13, 96},{13, 96},
			 {13,176},{13,176},{13,176},{13,176},{13,176},{13,176},{13,176},{13,176},{13,176},{13,176},{13,176},{13,176},{13,176},{13,176},{13,176},{13,176},
			 {13, 95},{13, 95},{13, 95},{13, 95},{13, 95},{13, 95},{13, 95},{13, 95},{13, 95},{13, 95},{13, 95},{13, 95},{13, 95},{13, 95},{13, 95},{13, 95},
			 {12, 94},{12, 94},{12, 94},{12, 94},{12, 94},{12, 94},{12, 94},{12, 94},{12, 94},{12, 94},{12, 94},{12, 94},{12, 94},{12, 94},{12, 94},{12, 94},
			 {12, 94},{12, 94},{12, 94},{12, 94},{12, 94},{12, 94},{12, 94},{12, 94},{12, 94},{12, 94},{12, 94},{12, 94},{12, 94},{12, 94},{12, 94},{12, 94}
			};

	/**
	 * Decompresses HuffYUV compressed data.
	 * 
	 * @param aCompressedBuffer byte array containing the compressed bits, see Remark
	 * @param aCompressedOffset offset of first byte where compressed bits start
	 * @param aUncompressedBuffer see Remark, byte array of exact size to hold uncompressed frame bits
	 * @param aUncompressedOffset offset in uncompressed buffer where uncompressed bytes should start
	 * @param aSizeImage size of uncompressed frame bits (usually pxwidth*pxheight*bytes_per_px)
	 * <br>
	 * <br>
	 * <b>Remarks:</b> 
	 * <ul>
	 * <li>aCompressedBuffer must be at least [aConmpressedOffset+size_compressed_bits+4] 
	 * bytes big because it is read 32-bit wise so under rare occassions the loop needs 
	 * up to four extra bytes at end of buffer.
	 * <li>
	 * aUncompressedBuffer must be exactly [aUncompressedOffset+aSizeImage] bytes big! 
	 * This is due to optimisation. As a side-effect of this, aSizeImage is not used at the moment.
	 * </ul> 
	 */
	 public static final void decompressHuffYUV( final byte[] aCompressedBuffer, int aCompressedOffset, byte [] aUncompressedBuffer, int aUncompressedOffset, int aSizeImage)
	 {
		 // TODO: return boolean and check properly for good decompression, account aSizeImage
		 
		 // comments further down usually refer to debug output that was necessary during 
		 // development
		 
		 // some comments refer to commented out things because of code speed optimisation
		 
		 int bitpos;
		 int outpos=aUncompressedOffset;
		 int inpos=aCompressedOffset;
		 
		 bitpos = 32+(aCompressedOffset*8);

		 long fc =	(((long)(aCompressedBuffer[inpos+3]&0xFF)<<16))|
		 			(((long)(aCompressedBuffer[inpos+2]&0xFF)<<8))|
		 			(aCompressedBuffer[inpos+1]&0xFF); 
		 
		 //int fco = aCompressedBuffer[inpos] >>> 8;
		 
		 aUncompressedBuffer[outpos+3] = (byte) ((fc >> 24));
		 aUncompressedBuffer[outpos+2] = (byte) ((fc >> 16));
		 aUncompressedBuffer[outpos+1] = (byte) ((fc >> 8));
		 aUncompressedBuffer[outpos] = (byte) (fc);
		 
		 outpos += 3;
		 
		 try {
			 // while( outpos < (aSizeImage+aUncompressedOffset) )
			 for(;;outpos++)
			 {
				 // SW! signed, unsigned
				 //System.out.println("bitpos="+bitpos[0]);

				//int wordt4o = (bitpos >>> 5)<<2;
				int wordt4 = ((bitpos>>>3)&0xFFFFFFFC);
				int bit = bitpos & 31;
				int retval;
				
				//System.out.println("wordt4o="+wordt4o+" wordt4="+wordt4);
						
				long val = 	(((long)(aCompressedBuffer[wordt4+3]&0xFF))<<24)|
							(((long)(aCompressedBuffer[wordt4+2]&0xFF))<<16)|
							(((long)(aCompressedBuffer[wordt4+1]&0xFF))<<8)|
							( (long)(aCompressedBuffer[wordt4])&0xFF);
						
				if (bit != 0) 
				{
					wordt4+=4;
					long val1 = (((long)(aCompressedBuffer[wordt4+3]&0xFF))<<24)|
								(((long)(aCompressedBuffer[wordt4+2]&0xFF))<<16)|
								(((long)(aCompressedBuffer[wordt4+1]&0xFF))<<8)|
								( (long)(aCompressedBuffer[wordt4])&0xFF);
							
					//val = (val << bit) | (in[word+1] >> (32-bit));
							
					val = ((val << bit)) | ((val1 >> (32-bit)));
					val &= 0xFFFFFFFFL;
				}
						
				if (val >= (35 << 23))
				{
					final int pos23 = (int) (val >> 23);
					
					bitpos += decode_luma9[pos23][0];
					retval = decode_luma9[pos23][1];
				}
				else if (val >= (15 << 20))
				{
					final int pos20 = (int) (val >> 20);
						
					bitpos += decode_luma12[pos20][0];
					retval = decode_luma12[pos20][1];
				}
				else
				{
					final int pos15 = (int) (val >> 15);
						
					bitpos += decode_luma17[pos15][0];
					retval = decode_luma17[pos15][1];
				}
				 
				aUncompressedBuffer[outpos] = (byte) (aUncompressedBuffer[outpos-3] + retval );

				//aUncompressedBuffer[outpos] = (byte) 
				//( ((((int)aUncompressedBuffer[outpos-3])&0xFF)) + retval );
			
				 //System.out.println("bitpos="+bitpos[0]);
				//outpos++;
			 }
		 }
		 catch( java.lang.ArrayIndexOutOfBoundsException ex ) {
		 	// performance & loop stop condition
		 }
			 
	 }

	 /*
		static final int UnhuffLuma(final byte[] aCompressedBuffer, long[] pos)
		{
			int word = (int) (pos[0] >>> 5);
			int bit = (int) (pos[0] & 31);
			
			long val = 	(((long)(aCompressedBuffer[4*word+3]&0xFF))<<24)|
						(((long)(aCompressedBuffer[4*word+2]&0xFF))<<16)|
						(((long)(aCompressedBuffer[4*word+1]&0xFF))<<8)|
						( (long)(aCompressedBuffer[4*word])&0xFF);
			
			if (bit != 0) 
			{
				long val1 = (((long)(aCompressedBuffer[4*(word+1)+3]&0xFF))<<24)|
							(((long)(aCompressedBuffer[4*(word+1)+2]&0xFF))<<16)|
							(((long)(aCompressedBuffer[4*(word+1)+1]&0xFF))<<8)|
							( (long)(aCompressedBuffer[4*(word+1)])&0xFF);
				
				//val = (val << bit) | (in[word+1] >> (32-bit));
				
				val = ((val << bit)&0xFFFFFFFFL) | ((val1 >>> (32-bit))&0xFFFFFFFFL);
			}
			
			if (val >= (35L << 23))
			{
				final int pos23 = (int) (val >> 23);
			
				pos[0] += decode_luma9[pos23][0];
				return decode_luma9[pos23][1];
			}
			
			if (val >= (15L << 20))
			{
				final int pos20 = (int) (val >> 20);
				
				pos[0] += decode_luma12[pos20][0];
				return decode_luma12[pos20][1];
			}

			final int pos15 = (int) (val >> 15);
			
			pos[0] += decode_luma17[pos15][0];
			return decode_luma17[pos15][1];
		}
		*/
		/*	C code
		 * static inline unsigned char UnhuffLuma(const DWORD* in, unsigned& pos) {
			  unsigned word = pos >> 5;
			  unsigned bit = pos & 31;
			  unsigned val = in[word];
			  if (bit)    // the conditional is necessary because shifting by 32 doesn't work correctly on x86 processors
			    val = (val << bit) | (in[word+1] >> (32-bit));
//			  // figure out the appropriate lookup table based on the number of leading zeros
			  if (val >= (35 << 23)) {
			    pos += decode_luma9[val >> 23].shift;
			    return decode_luma9[val >> 23].add;
			  }
			  if (val >= (15 << 20)) {
			    pos += decode_luma12[val >> 20].shift;
			    return decode_luma12[val >> 20].add;
			  }
			  pos += decode_luma17[val >> 15].shift;
			  return decode_luma17[val >> 15].add;
			}
		 */
	 
	 
	 
// old:
	 /*
	static final int UnhuffLuma(final byte[] aCompressedBuffer, int[] pos)
	{
		int word = pos[0] >>> 5;
		int bit = pos[0] & 31;
		
		long val = 	(((long)(aCompressedBuffer[4*word+3]&0xFF))<<24)|
					(((long)(aCompressedBuffer[4*word+2]&0xFF))<<16)|
					(((long)(aCompressedBuffer[4*word+1]&0xFF))<<8)|
					( (long)(aCompressedBuffer[4*word])&0xFF);
		
		if (bit != 0) 
		{
			long val1 = (((long)(aCompressedBuffer[4*(word+1)+3]&0xFF))<<24)|
						(((long)(aCompressedBuffer[4*(word+1)+2]&0xFF))<<16)|
						(((long)(aCompressedBuffer[4*(word+1)+1]&0xFF))<<8)|
						( (long)(aCompressedBuffer[4*(word+1)])&0xFF);
			
			//val = (val << bit) | (in[word+1] >> (32-bit));
			
			val = ((val << bit)&0xFFFFFFFFL) | ((val1 >>> (32-bit))&0xFFFFFFFFL);
		}
		
		if (val >= (35 << 23))
		{
			final int pos23 = (int) (val >> 23);
		
			pos[0] += decode_luma9[pos23][0];
			return decode_luma9[pos23][1];
		}
		
		if (val >= (15 << 20))
		{
			final int pos20 = (int) (val >> 20);
			
			pos[0] += decode_luma12[pos20][0];
			return decode_luma12[pos20][1];
		}

		final int pos15 = (int) (val >> 15);
		
		pos[0] += decode_luma17[pos15][0];
		return decode_luma17[pos15][1];
	}
	*/
	/*
	 // aCompressedBuffer needs 4 bytes more than the compressed bytes contained inside
	 static final void DecompressHuffYUV( final byte[] aCompressedBuffer, int aCompressedOffset, byte [] aUncompressedBuffer, int aUncompressedOffset, int aSizeImage)
	 {
		 long[] bitpos = new long[1];
		 int outpos=aUncompressedOffset;
		 int inpos=aCompressedOffset;
		 
		 bitpos[0] = 32+(aCompressedOffset*8);

		 long fc =	(((long)(aCompressedBuffer[inpos+3]&0xFF)<<24))|
		 			(((long)(aCompressedBuffer[inpos+2]&0xFF)<<16))|
		 			(((long)(aCompressedBuffer[inpos+1]&0xFF)<<8))|
		 			((long)(aCompressedBuffer[inpos]&0xFF)); 
		 
		 fc >>>= 8;
		                            
		 //int fc = aCompressedBuffer[inpos] >>> 8;
		 
		 // SW! might be reverse order
		 aUncompressedBuffer[outpos+3] = (byte) ((fc >> 24) & 0xFF);
		 aUncompressedBuffer[outpos+2] = (byte) ((fc >> 16) & 0xFF);
		 aUncompressedBuffer[outpos+1] = (byte) ((fc >> 8) & 0xFF);
		 aUncompressedBuffer[outpos+0] = (byte) (fc & 0xFF);
		 
		 outpos += 3;
		 
		 //try {
			 while( outpos < aSizeImage )
			 {
				 // SW! signed, unsigned
				 //System.out.println("bitpos="+bitpos[0]);
				 
				 				 
				 aUncompressedBuffer[outpos] = (byte) (((((int)aUncompressedBuffer[outpos-3])&0xFF)) + UnhuffLuma(aCompressedBuffer, bitpos));
				 //System.out.println("bitpos="+bitpos[0]);
				 outpos++;
			 }
		  //}
		 //catch( java.lang.ArrayIndexOutOfBoundsException ex )
		 //{
			 //int dbgbrkpt = 1;
			 //System.out.println("AIOOBex outpos="+outpos);
		// }
			 
			 
		 }
	 
	 */
	 
	 
	 /*	
//		 This function decompresses the raw image bits. It is directly taken out of the HufYUV v1.2.3 source
		void CCompression::DecompressRGB( void *compressed, void *uncompressed, DWORD dwSize, int sizeimage) 
		{
		  int size = sizeimage;
		  
		  const DWORD* const in = (DWORD*) compressed;
		  unsigned char* const end = (unsigned char*)uncompressed + size;
		  unsigned char* out = (unsigned char*)uncompressed;

		  *(DWORD*)out = in[0] >> 8;
		  unsigned pos = 32;

		  // RGB 24-bit
		  out += 3;
		  while (out < end) // main decompression loop
		  {
		    out[0] = out[-3] + UnhuffLuma(in, pos);
		    ++out;
		  }

		}
	
	*/

	 // goal is to read compressed image blob 
	// write out imm file at the end

/*	public static void main(String[] args) {
		
		final int width = 768;
		final int height = 574;
		final double scale = 1.2345;
		int sizeImage = width*height;
		
		byte[] uncompressedBuffer = new byte[sizeImage];
		byte[] compressedBuffer;
		
		java.io.FileInputStream fis = null;
		java.io.FileOutputStream fos = null;
		
		try {
		
			fis = new java.io.FileInputStream(new java.io.File("input.huffyuv.blob"));

			long fsize = fis.getChannel().size();
		
			compressedBuffer = new byte[(int) fsize+4];
			compressedBuffer[(int) fsize-1] = 0;
			compressedBuffer[(int)fsize-2] = 0;
			compressedBuffer[(int)fsize-3] = 0;
			compressedBuffer[(int)fsize-4] = 0;
		
			int ret = fis.read(compressedBuffer, 0, (int) fsize);

			if (ret == fsize)
			{
				
				long start = System.currentTimeMillis();
				
				System.out.println("successfully read file, size read "+ret+" bytes.");
				for (int i=0;i<1000;i++)
				{
					DecompressHuffYUV(compressedBuffer, uncompressedBuffer, sizeImage);
				}
				
				long stop = System.currentTimeMillis();
				
				System.out.println("time for 1000 runs="+(stop-start)+" msec tpr="+((stop-start)/1000.0)+" msec.");
				
				
				try {
					fos = new java.io.FileOutputStream("outfile.imm", false);
					
					java.nio.ByteBuffer bb = java.nio.ByteBuffer.allocate(sizeImage+16);
					bb.order(java.nio.ByteOrder.LITTLE_ENDIAN);
					
					bb.putInt(width);
					bb.putInt(height);
					bb.put(uncompressedBuffer);
					bb.putDouble(scale);
					
					bb.flip();

					//buf.flip();
		            //
		            //// Write the bytes to the channel
		            //int numWritten = channel.write(buf);

					
					//java.nio.channels.WritableByteChannel wbc(fos.getChannel()); 
					
				    int ret1 = fos.getChannel().write( bb );
				    
				    System.out.println("successfully wrote "+ret1+" bytes to outfile.");
					fos.close();
				}
				catch( java.io.IOException e )
				{
					System.out.println("IOEx: "+e.getMessage()+"\n");
				}
				finally
				{
					try { if (fos != null) fos.close(); } catch(java.io.IOException e ){}
				}

				
			}
			else System.out.println("Error reading full input file.");
		}
		catch( java.io.FileNotFoundException e )
		{
			System.out.println("FNFex: "+e.getMessage()+"\n");
		}
		catch( java.io.IOException e )
		{
			System.out.println("IOEx: "+e.getMessage()+"\n");
		}
		finally
		{
			try { if (fis != null) fis.close(); } catch(java.io.IOException e ){}
		}
		
		// read blob, write imm
	}
*/
}
