package de.desy.acop.video.displayer;

import java.util.HashMap;
import java.util.Map;

/**
 * Image format enumeration
 * 
 * @author mdavid
 * 
 */
public enum ImageFormat {
	/**
	 * Grayscale format
	 */
	IMAGE_FORMAT_GRAY(0, true),

	/**
	 * RGB (24bpp r-g-b) format
	 */
	IMAGE_FORMAT_RGB(1, true),

	/**
	 * RGB (32bpp a-r-g-b) format
	 */
	IMAGE_FORMAT_RGBA(2, true),

	/**
	 * RGB (16bpp 0-5r-5g-5b) format
	 */
	IMAGE_FORMAT_RGB15(3),

	/**
	 * RGB (16bpp 5r-6g-5b)) format
	 */
	IMAGE_FORMAT_RGB16(4),

	/**
	 * JPEG file
	 */
	IMAGE_FORMAT_JPEG(5, true),

	/**
	 * TIFF file
	 */
	IMAGE_FORMAT_TIFF(6),

	/**
	 * YUV411 format
	 */
	IMAGE_FORMAT_YUV411(7),

	/**
	 * YUV422 format
	 */
	IMAGE_FORMAT_YUV422(8),

	/**
	 * YUV444 format
	 */
	IMAGE_FORMAT_YUV444(9),

	/**
	 * HuffYUV (compressed grayscale)
	 */
	IMAGE_FORMAT_HUFFYUV(10, true),

	/**
	 * Windows BMP file
	 */
	IMAGE_FORMAT_BMP(11),

	/**
	 * Bayer pattern
	 */
	IMAGE_FORMAT_BAYER(12);

	private final int id;
	private final boolean isSupported;

	private static Map<Integer, ImageFormat> map = new HashMap<>(values().length);
	static {
		for (ImageFormat e : values())
			if (map.put(e.id, e) != null)
				throw new InternalError("duplicate id: " + e.id);
	}

	private ImageFormat(int id) {
		this(id, false);
	}

	private ImageFormat(int id, boolean isSupported) {
		this.id = id;
		this.isSupported = isSupported;
	}

	/**
	 * Returns integer representation of the TINE image format
	 * 
	 * @return integer representation of the TINE image format
	 */
	public int getId() {
		return id;
	}

	/**
	 * Returns true if format is supported.
	 * 
	 * @return true if TINE image format is supported, otherwise false
	 */
	public boolean isSupported() {
		return isSupported;
	}

	public boolean equals(int id) {
		return this.id == id;
	}

	public static ImageFormat valueOf(int id) throws IllegalArgumentException {
		ImageFormat e = map.get(id);
		if (e == null)
			throw new IllegalArgumentException("non-existing image format id: " + id);
		return e;
	}

}
