package de.desy.acop.video.timageio;

import java.awt.Graphics2D;
import java.awt.Transparency;
import java.awt.color.ColorSpace;
import java.awt.image.BufferedImage;
import java.awt.image.ColorModel;
import java.awt.image.ComponentColorModel;
import java.awt.image.DataBuffer;
import java.awt.image.DataBufferByte;
import java.awt.image.DataBufferInt;
import java.awt.image.DataBufferUShort;
import java.awt.image.DirectColorModel;
import java.awt.image.Raster;
import java.awt.image.SampleModel;
import java.awt.image.WritableRaster;
import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.util.Arrays;
import java.util.Hashtable;

import javax.imageio.ImageIO;
import javax.imageio.ImageReader;
import javax.imageio.ImageWriter;
import javax.imageio.stream.ImageInputStream;
import javax.imageio.stream.ImageOutputStream;

import de.desy.acop.video.displayer.ColorMap;
import de.desy.acop.video.displayer.HuffmanDecompression;
import de.desy.acop.video.displayer.ImageFormat;
import de.desy.acop.video.displayer.ImageOrientation;
import de.desy.tine.types.IMAGE;
import de.desy.tine.types.IMAGE.FrameHeader;

/**
 * The <code>TBufferedImage</code> comprised of a {@link TImageMetadata} of image metadata and a {@link BufferedImage}
 * of image data.
 * 
 * @author David Melkumyan, DESY Zeuthen
 * @version $Revision: 91076 $ : $LastChangedDate: 2021-01-19 20:37:15 +0100 (Tue, 19 Jan 2021) $
 */
public class TBufferedImage {

	/**
	 * Java <code>BufferedImage</code> with an accessible buffer of image data.
	 */
	private BufferedImage bufferedImage;

	/**
	 * Image metadata (non-image data) associated with TINE images.
	 */
	private TImageMetadata tiMetadata;

	/**
	 * Constructs a <code>TBufferedImage</code> from <code>BufferedImage</code> and <code>TImageMetadata</code>.
	 * 
	 * @note only for internal use
	 * 
	 * @param bi
	 *            - Java <code>BufferedImage</code>
	 * @param tiMetadata
	 *            - <code>TImageMetadata</code>
	 * 
	 * @see TImageMetadata
	 */
	public TBufferedImage(BufferedImage bi, TImageMetadata tiMetadata) {
		if (bi == null) {
			throw new NullPointerException("bufferedImage == null!");
		}
		this.bufferedImage = bi;
		this.tiMetadata = tiMetadata;
	}

	/**
	 * Creates <code>TBufferedImage</code> instance from Tine IMAGE (in 'archival' mode).
	 * 
	 * @param ti
	 *            - TINE Image
	 */
	public TBufferedImage(IMAGE ti) {
		this(TBufferedImage.toBufferedImage(ti), new TImageMetadata(ti));
	}

	/**
	 * Creates <code>TBufferedImage</code> instance from Tine IMAGE (in 'archival' mode).
	 * 
	 * @param ti
	 *            - TINE Image
	 */
	public TBufferedImage(IMAGE ti, ColorMap colorMap, boolean normalize, boolean aoiExpansion) {
		this(TBufferedImage.toBufferedImage(ti, colorMap, normalize, aoiExpansion), new TImageMetadata(ti));
	}

	/**
	 * Returns Java BufferedImage with an accessible buffer of image data
	 * 
	 * @return BufferedImage
	 */
	public BufferedImage getBufferedImage() {
		return bufferedImage;
	}

	public void setBufferedImage(BufferedImage image) {
		this.bufferedImage = image;
	}

	/**
	 * Returns image metadata (non-image data) associated with TINE images.
	 * 
	 * @return TImageMetadata
	 */
	public TImageMetadata getMetadata() {
		return tiMetadata;
	}

	/**
	 * Sets the TINE image metadata
	 * 
	 * @param metadata
	 *            - TINE image metadata
	 */
	public void setMetadata(TImageMetadata metadata) {
		this.tiMetadata = metadata;
	}

	/**
	 * Creates <code>BufferedImage</code> from Tine IMAGE in 'archival' mode.
	 * 
	 * @param ti
	 *            - input Tine IMAGE
	 * @return Java Buffered Image
	 */
	public static BufferedImage toBufferedImage(IMAGE ti) {
		return toBufferedImage(ti, null, false, false);
	}

	/**
	 * Creates <code>BufferedImage</code> from Tine IMAGE in 'archival'/'export' mode.
	 * 
	 * @param ti
	 *            - input Tine IMAGE
	 * 
	 * @param colorMap
	 *            - apply color map
	 * @param normalize
	 *            - apply the histogram equalization will be applied
	 * @param aoiExpansion
	 *            - apply the AOI expansion
	 * 
	 * @return Java Buffered Image
	 */
	public static BufferedImage toBufferedImage(IMAGE ti, ColorMap colorMap, boolean normalize, boolean aoiExpansion) {
		return toBufferedImage(ti, colorMap, normalize, -1, -1, aoiExpansion, ImageOrientation.NoChange);
	}

	public static BufferedImage toBufferedImage(IMAGE ti, ColorMap colorMap, //
			boolean normalize, int min, int max, //
			boolean aoiExpansion, //
			ImageOrientation orientation) {

		if (ti == null) {
			throw new NullPointerException("ti == null!");
		}

		BufferedImage destImage = null;

		final FrameHeader frmHdr = ti.getFrameHeader();
		final boolean isAOI = (frmHdr.aoiWidth > 0) || (frmHdr.aoiHeight > 0);

		final ImageFormat format = ImageFormat.valueOf(frmHdr.imageFormat);
		final int width = ((frmHdr.aoiWidth > 0) ? frmHdr.aoiWidth : frmHdr.sourceWidth);
		final int height = ((frmHdr.aoiHeight > 0) ? frmHdr.aoiHeight : frmHdr.sourceHeight);

		final int bpp = frmHdr.bytesPerPixel;
		final int bits = frmHdr.effectiveBitsPerPixel;
		final int appendedSize = frmHdr.appendedFrameSize;
		final int wh = width * height;
		final byte[] imageData = ti.getImageFrameBuffer();

		if (bits < 8) {
			throw new IllegalArgumentException("bits=" + bits);
		}
		if (imageData.length < appendedSize) {
			throw new IllegalArgumentException("format=" + format + ", size=" + appendedSize //
					+ ", w=" + width + ", h=" + height + ", bpp=" + bpp + ", data.size=" + imageData.length);
		}

		final Hashtable<?, ?> props = TImageMetadata.createProperties(ti);

		// Swap width and height if image was rotated
		int orientedW = width;
		int orientedH = height;
		if (orientation != null && //
				(ImageOrientation.RotationBy90CW.equals(orientation) || ImageOrientation.RotationBy90CCW.equals(orientation))) {
			orientedW = height;
			orientedH = width;
		}

		DataBuffer dataBuffer = null;

		switch (format) {
		case IMAGE_FORMAT_GRAY:
			if (appendedSize != wh * bpp) {
				throw new IllegalArgumentException("format=" + format + ", size=" + appendedSize + ", w=" + width + ", h=" + height + ", bpp=" + bpp);
			}

			switch (bpp) {
			case 1: // 1 bytes per pixel
				// created buffered image: TYPE_BYTE_INDEXED / TYPE_BYTE_GRAY
				if (bits != 8) {
					throw new IllegalArgumentException("format=" + format + ", bits=" + bits);
				}

				// Rotate image
				byte[] byteArray = rotateByteArray(imageData.length == wh ? //
						imageData : Arrays.copyOf(imageData, wh), width, height, orientation);

				// Normalize image
				if (normalize) {
					if (byteArray == imageData) {
						byteArray = Arrays.copyOf(imageData, wh);
					}
					byteArray = normalizeGray(byteArray, min == -1 ? 0x0 : min, max == -1 ? 0xff : max);
				}

				// Apply color map and create buffered image
				dataBuffer = new DataBufferByte(byteArray, wh);
				if (colorMap == null || ColorMap.NONE.equals(colorMap)) {
					destImage = createImageGray(orientedW, orientedH, bits, dataBuffer, props);
				} else {
					destImage = createImageIndexed(colorMap.getLUT(bits), orientedW, orientedH, dataBuffer, props);
				}
				break;

			case 2: // 2 bytes per pixel
				// created buffered image: TYPE_CUSTOM / TYPE_USHORT_GRAY
				if (bits > 16) {
					throw new IllegalArgumentException("format=" + format + ", bits=" + bits);
				}

				// Rotate image
				short[] shortArray = createShortArray(imageData, width, height, bits, orientation);

				// Normalize image
				if (normalize) {
					normalizeGray(shortArray, (1 << bits) - 1, min == -1 ? 0x0 : min, max == -1 ? 0xffff : max);
				}

				// Apply color map and create buffered image
				dataBuffer = new DataBufferUShort(shortArray, wh);
				if (colorMap == null || ColorMap.NONE.equals(colorMap) || ColorMap.GRAYSCALE.equals(colorMap)) {
					// #pixelBits = 16 numComponents = 1 maxBits = [1-16]
					destImage = createImageGray(orientedW, orientedH, bits, dataBuffer, props);
				} else {
					// #pixelBits = [1-16] numComponents = 3
					destImage = createImageIndexed(colorMap.getLUT(bits), orientedW, orientedH, dataBuffer, props);
				}
				break;

			default:
				throw new IllegalArgumentException("format=" + format + ", bpp=" + bpp);
			}
			break;

		case IMAGE_FORMAT_RGB:
		case IMAGE_FORMAT_RGBA:
			boolean isRGB = (format == ImageFormat.IMAGE_FORMAT_RGB);

			if ((isRGB && (bpp != 3 || bits != 24)) || (!isRGB && (bpp != 4 || bits != 32))) {
				throw new IllegalArgumentException("format=" + format + ", bpp=" + bpp + ", bits=" + bits);
			}
			if (appendedSize != wh * bpp) {
				throw new IllegalArgumentException("format=" + format + ", size=" + appendedSize + ", w=" + width + ", h=" + height + ", bpp=" + bpp);
			}

			// Rotate image
			int[] intArray = createArrayColor(imageData, bpp, width, height, true, orientation);

			// Normalize image
			if (normalize) {
				normalizeRGB(intArray, min == -1 ? 0x0 : min, max == -1 ? 0xff : max);
			}

			// Create buffered image
			destImage = createImageRGB(orientedW, orientedH, new DataBufferInt(intArray, wh), props);
			break;

		case IMAGE_FORMAT_HUFFYUV:
			// created buffered image: TYPE_BYTE_INDEXED / TYPE_BYTE_GRAY
			if (bpp != 1 || bits != 8) {
				throw new IllegalArgumentException("format=" + format + ", bpp=" + bpp + ", bits=" + bits);
			}

			if (frmHdr.sourceFormat != 0) { // IMAGE_FORMAT_GRAY
				throw new IllegalArgumentException("format=" + format + ", srcFormat=" + ImageFormat.valueOf(frmHdr.sourceFormat));
			}

			// Decompress HuffYUV compressed data image
			byte[] compressed = Arrays.copyOf(imageData, appendedSize);
			byte[] uncompressed = new byte[wh * bpp];
			HuffmanDecompression.decompressHuffYUV(compressed, 0, uncompressed, 0, uncompressed.length);

			// Rotate image
			byte[] byteArray = rotateByteArray(uncompressed, width, height, orientation);

			// Normalize image
			if (normalize) {
				normalizeGray(byteArray, min == -1 ? 0x0 : min, max == -1 ? 0xff : max);
			}

			// Apply color map and create buffered image
			dataBuffer = new DataBufferByte(byteArray, wh);
			if (colorMap == null || colorMap.equals(ColorMap.NONE)) {
				destImage = createImageGray(orientedW, orientedH, bits, dataBuffer, props);
			} else {
				destImage = createImageIndexed(colorMap.getLUT(bits), orientedW, orientedH, dataBuffer, props);
			}
			break;

		case IMAGE_FORMAT_JPEG:
			// created buffered image:
			// TYPE_BYTE_INDEXED / TYPE_BYTE_GRAY / TYPE_3BYTE_BGR /
			// TYPE_INT_RGB
			if ((bpp != 1 && bpp != 3) || (bpp == 1 && bits != 8) || (bpp == 3 && bits != 24)) {
				throw new IllegalArgumentException("format=" + format + ", bpp=" + bpp + ", bits=" + bits);
			}

			BufferedImage bi = createImageJPEG(Arrays.copyOf(imageData, appendedSize), props);

			switch (bi.getType()) {
			case BufferedImage.TYPE_BYTE_GRAY:
				byte[] byteArrayJpegGray = ((DataBufferByte) bi.getRaster().getDataBuffer()).getData();

				// Rotate image
				if (orientation != null && !ImageOrientation.NoChange.equals(orientation)) {
					switch (orientation) {
					case RotationBy180:
						// must be same array
						TImageUtil.reverseArray(byteArrayJpegGray, true);
						break;

					case RotationBy90CW:
						bi = TImageUtil.rotateImageCW(bi);
						byteArrayJpegGray = ((DataBufferByte) bi.getRaster().getDataBuffer()).getData();
						break;

					case RotationBy90CCW:
						bi = TImageUtil.rotateImageCCW(bi);
						byteArrayJpegGray = ((DataBufferByte) bi.getRaster().getDataBuffer()).getData();
						break;

					default:
						throw new IllegalArgumentException("unsupported image orientation: " + orientation);
					}
				}

				// Normalize image
				if (normalize) {
					normalizeGray(byteArrayJpegGray, min == -1 ? 0x0 : min, max == -1 ? 0xff : max);
				}
				if (colorMap != null && !colorMap.equals(ColorMap.NONE)) {
					destImage = new BufferedImage(colorMap.getLUT(bits), bi.getRaster(), false, props);
				} else {
					destImage = new BufferedImage(bi.getColorModel(), bi.getRaster(), false, props);
				}
				break;

			case BufferedImage.TYPE_3BYTE_BGR:
				// Rotate image
				if (orientation != null && !ImageOrientation.NoChange.equals(orientation)) {
					switch (orientation) {
					case RotationBy180:
						// must be same array
						TImageUtil.reverseColorArray(((DataBufferByte) bi.getRaster().getDataBuffer()).getData(), true);
						break;

					case RotationBy90CW:
						bi = TImageUtil.rotateImageCW(bi);
						bi = new BufferedImage(bi.getColorModel(), bi.getRaster(), false, props);
						break;

					case RotationBy90CCW:
						bi = TImageUtil.rotateImageCCW(bi);
						bi = new BufferedImage(bi.getColorModel(), bi.getRaster(), false, props);
						break;

					default:
						throw new IllegalArgumentException("unsupported image orientation: " + orientation);
					}
				}

				// Normalize image
				if (normalize) {
					normalizeBGR(((DataBufferByte) bi.getRaster().getDataBuffer()).getData(), //
							min == -1 ? 0x0 : min, max == -1 ? 0xff : max);
				}
				destImage = bi;
				break;

			case BufferedImage.TYPE_INT_RGB:
			case BufferedImage.TYPE_INT_ARGB:
				// Rotate image
				if (orientation != null && !ImageOrientation.NoChange.equals(orientation)) {
					switch (orientation) {
					case RotationBy180:
						TImageUtil.reverseArray(((DataBufferInt) bi.getRaster().getDataBuffer()).getData(), true);
						bi = new BufferedImage(bi.getColorModel(), bi.getRaster(), false, props);
						break;

					case RotationBy90CW:
						bi = TImageUtil.rotateImageCW(bi);
						bi = new BufferedImage(bi.getColorModel(), bi.getRaster(), false, props);
						break;

					case RotationBy90CCW:
						bi = TImageUtil.rotateImageCCW(bi);
						break;

					default:
						throw new IllegalArgumentException("unsupported image orientation: " + orientation);
					}
				}

				// Normalize image
				if (normalize) {
					normalizeRGB(((DataBufferInt) bi.getRaster().getDataBuffer()).getData(), //
							min == -1 ? 0x0 : min, max == -1 ? 0xff : max);
				}
				destImage = bi;
				break;

			default:
				throw new IllegalArgumentException("Unsupported image format: " + bi.getType());
			}
			break;

		default:
			throw new IllegalArgumentException("tiFormat=" + format);
		}
		return (aoiExpansion && isAOI) ? expanseAOI(destImage, props) : destImage;
	}

	public static BufferedImage toBufferedImage(IMAGE ti, ColorMap colorMap, boolean normalize, int min, int max, boolean aoiExpansion) {
		return toBufferedImage(ti, colorMap, normalize, min, max, aoiExpansion, ImageOrientation.NoChange);
	}

	/**
	 * Converts Java <code>BufferedImage</code> to TINE image
	 * 
	 * @param bi
	 *            - Java BufferedImage with/without TINE image metadata
	 * @return TINE image
	 */
	public static IMAGE toIMAGE(BufferedImage bi) {
		return toIMAGE(bi, false);
	}

	/**
	 * Converts Java <code>BufferedImage</code> to TINE image
	 * 
	 * @param bi
	 *            - Java BufferedImage with/without TINE image metadata
	 * 
	 * @param rejectNonArchival
	 *            - if <code>true</code> and <code>bi</code> does not contain TINE image metadata throws
	 *            <code>IllegalArgumentException</code>
	 * @return TINE image
	 */
	@SuppressWarnings("deprecation")
	public static IMAGE toIMAGE(BufferedImage bi, boolean rejectNonArchival) {
		TImageMetadata tiMetadata = new TImageMetadata(bi, rejectNonArchival);
		IMAGE ti = new IMAGE(0);
		tiMetadata.write(ti);
		if (tiMetadata.imageFormat == ImageFormat.IMAGE_FORMAT_JPEG) {
		    byte[] data = createBytesFromJPEG(bi);
		    ti.setImageFrameBuffer(data);
		    ti.getFrameHeader().appendedFrameSize = data.length;
		} else {
		    ti.setImageFrameBuffer(getData(bi));
		}

		return ti;
	}
	
    private static byte[] createBytesFromJPEG(BufferedImage image) {
        ImageIO.setUseCache(false); // don`t use cache file
        ImageWriter writer = ImageIO.getImageWritersByFormatName("JPG").next();
        ByteArrayOutputStream byteOut = new ByteArrayOutputStream();
        try (ImageOutputStream out = ImageIO.createImageOutputStream(byteOut)) {
            writer.setOutput(out);
            writer.write(image);
            return byteOut.toByteArray();
        } catch (Exception e) {
            throw new IllegalArgumentException(e);
        }
    }

	/**
	 * Represents <code>TBufferedImage</code> as TINE image
	 * 
	 * @return TINE image
	 */
	public IMAGE toIMAGE() {
		return toIMAGE(bufferedImage);
	}

	/**
	 * Represents <code>TBufferedImage</code> as TINE image
	 * 
	 * @param rejectNonArchival
	 *            - if <code>true</code> and <code>TBufferedImage</code> does not contain TINE image metadata throws
	 *            <code>IllegalArgumentException</code>
	 * @return TINE image
	 */
	public IMAGE toIMAGE(boolean rejectNonArchival) {
		return toIMAGE(bufferedImage, rejectNonArchival);
	}

	/**
	 * Returns true if TINE image format is supported.
	 * 
	 * @param format
	 *            - an Image Format
	 * @return true if image format is supported, otherwise - false
	 * @throws IllegalArgumentException
	 *             if format is invalid
	 */
	public static boolean isSupported(int format) {
		return ImageFormat.valueOf(format).isSupported();
	}

	/**
	 * Returns true if TINE image contains AOI, otherwise false.
	 * 
	 * @param ti
	 *            - TINE image
	 * @return true if AOI is present in the image, otherwise false
	 * @note TODO: have to be a part of the TINE Image interface
	 */
	public static boolean containsAOI(IMAGE ti) {
		return (ti.getFrameHeader().aoiHeight > 0 || ti.getFrameHeader().aoiWidth > 0);
	}

	private static BufferedImage createImageGray(int w, int h, int bits, DataBuffer db, Hashtable<?, ?> props) {
		WritableRaster raster = Raster.createInterleavedRaster(db, w, h, w, 1, new int[] { 0 }, null);
		ColorModel cm = new ComponentColorModel(ColorSpace.getInstance(ColorSpace.CS_GRAY), //
				new int[] { bits }, false, true, Transparency.OPAQUE, db.getDataType());
		return new BufferedImage(cm, raster, false, props);
	}

	private static BufferedImage createImageIndexed(ColorModel cm, int w, int h, DataBuffer db, Hashtable<?, ?> props) {
		WritableRaster raster = Raster.createInterleavedRaster(db, w, h, w, 1, new int[] { 0 }, null);
		return new BufferedImage(cm, raster, false, props);
	}

	private static BufferedImage createImageRGB(int w, int h, DataBuffer db, Hashtable<?, ?> props) {
		ColorModel cm = new DirectColorModel(24, 0x00ff0000, 0x0000ff00, 0x000000ff, 0x0);
		SampleModel sm = cm.createCompatibleSampleModel(w, h);
		WritableRaster raster = WritableRaster.createWritableRaster(sm, db, null);
		return new BufferedImage(cm, raster, false, props);
	}

	private static BufferedImage expanseAOI(BufferedImage srcImage, Hashtable<?, ?> props) {
		if (srcImage == null)
			throw new NullPointerException("srcImage == null!");

		int srcWidth = (Integer) props.get(TImageMetadata.KEY_SOURCE_WIDTH);
		int srcHeight = (Integer) props.get(TImageMetadata.KEY_SOURCE_HEIGHT);
		int xStart = (Integer) props.get(TImageMetadata.KEY_XSTART);
		int yStart = (Integer) props.get(TImageMetadata.KEY_YSTART);

		ColorModel cm = new DirectColorModel(24, 0x00ff0000, 0x0000ff00, 0x000000ff, 0x0);
		WritableRaster raster = cm.createCompatibleWritableRaster(srcWidth, srcHeight);
		BufferedImage bi = new BufferedImage(cm, raster, false, props);

		Graphics2D g2d = bi.createGraphics();
		g2d.drawImage(srcImage, xStart, yStart, null);
		srcImage.flush();
		g2d.dispose();

		return bi;
	}

	/**
	 * @param bi
	 *            - BufferedImage
	 * @return always <code>null</code> (!!!) while <code>BufferedImage#getPropertyNames()</code> returns always
	 *         <code>null</code>
	 * @deprecated
	 * @see <code>BufferedImage#getPropertyNames()</code>
	 */
	@Deprecated
	public static Hashtable<String, Object> getProperties(BufferedImage bi) {
		String[] propertyNames = bi.getPropertyNames();
		if (propertyNames == null)
			return null;

		Hashtable<String, Object> properties = new Hashtable<>(32);
		for (int i = 0; i < propertyNames.length; i++) {
			String name = propertyNames[i];
			properties.put(name, bi.getProperty(name));
		}
		return properties;
	}

	private static BufferedImage createImageJPEG(byte[] dataArray, Hashtable<?, ?> properties) {
		ImageIO.setUseCache(false); // don`t use cache file
		ImageReader reader = ImageIO.getImageReadersByFormatName("JPG").next();
		try (ImageInputStream iis = ImageIO.createImageInputStream(new ByteArrayInputStream(dataArray))) {
			reader.setInput(iis, true);
			BufferedImage bi = reader.read(0);
			ColorModel cm = bi.getColorModel();
			return new BufferedImage(cm, bi.getRaster(), cm.isAlphaPremultiplied(), properties);

		} catch (Exception e) {
			throw new IllegalArgumentException(e);
		}
	}

	private static short[] createShortArray(byte[] byteArray, int w, int h, int bits, ImageOrientation orientation) {
		short[] res = new short[w * h];
		if (orientation != null && ImageOrientation.RotationBy180.equals(orientation)) {
			for (int i = res.length, j = 0; i > 0;) {
				int gray0 = byteArray[j++] & 0xff;
				int gray1 = byteArray[j++] & 0xff;
				int pixel = ((gray1 << 8) | gray0);
				res[--i] = (short) pixel;
			}
			return res;
		}

		// Convert to short array
		for (int i = 0, j = 0; i < res.length;) {
			int gray0 = byteArray[j++] & 0xff;
			int gray1 = byteArray[j++] & 0xff;
			int pixel = ((gray1 << 8) | gray0);
			res[i++] = (short) pixel;
		}
		if (orientation == null || ImageOrientation.NoChange.equals(orientation)) {
			return res;
		}
		if (ImageOrientation.RotationBy90CW.equals(orientation)) {
			return TImageUtil.rotateBitmapCW(res, w, h);
		}
		if (ImageOrientation.RotationBy90CCW.equals(orientation)) {
			return TImageUtil.rotateBitmapCCW(res, w, h);
		}
		throw new IllegalArgumentException("unsupported image orientation: " + orientation);
	}

	private static int[] createArrayColor(byte[] bitmap, int bpp, int w, int h, boolean isRGB, ImageOrientation orien) {
		int[] res = new int[w * h];
		if (orien != null && ImageOrientation.RotationBy180.equals(orien)) {
			for (int i = res.length, j = 0; i > 0; j += bpp) {
				if (isRGB) {
					res[--i] = TImageUtil.getColor(bitmap[j], bitmap[j + 1], bitmap[j + 2]);
				} else { // BGR
					res[--i] = TImageUtil.getBGR(bitmap[j + 2], bitmap[j + 1], bitmap[j]);
				}
			}
			return res;
		}

		// Convert to integer array
		for (int i = 0, j = 0; i < res.length; j += bpp) {
			if (isRGB) {
				res[i++] = TImageUtil.getRGB(bitmap[j], bitmap[j + 1], bitmap[j + 2]);
			} else {
				res[i++] = TImageUtil.getRGB(bitmap[j + 2], bitmap[j + 1], bitmap[j]);
			}

		}
		if (orien == null || ImageOrientation.NoChange.equals(orien)) {
			return res;
		}
		if (ImageOrientation.RotationBy90CW.equals(orien)) {
			return TImageUtil.rotateBitmapCW(res, w, h);
		}
		if (ImageOrientation.RotationBy90CCW.equals(orien)) {
			return TImageUtil.rotateBitmapCCW(res, w, h);
		}
		throw new IllegalArgumentException("unsupported image orientation: " + orien);
	}

	private static byte[] rotateByteArray(byte[] byteArray, int w, int h, ImageOrientation orientation) {
		if (orientation == null || ImageOrientation.NoChange.equals(orientation)) {
			return byteArray;
		}
		if (ImageOrientation.RotationBy180.equals(orientation)) {
			return TImageUtil.reverseArray(byteArray, true);
		}
		if (ImageOrientation.RotationBy90CW.equals(orientation)) {
			return TImageUtil.rotateBitmapCW(byteArray, w, h);
		}
		if (ImageOrientation.RotationBy90CCW.equals(orientation)) {
			return TImageUtil.rotateBitmapCCW(byteArray, w, h);

		}
		throw new IllegalArgumentException("unsupported image orientation: " + orientation);
	}

	static void reverse(int[] input) {
		int len = input.length >> 1;
		for (int i = 0; i < len; i++) {
			int temp = input[i];
			input[i] = input[input.length - 1 - i];
			input[input.length - 1 - i] = temp;
		}
	}

	private static int[] getExtrema(byte[] array) {
		int max = 0x0;
		int min = 0xFF;
		for (int i : array) {
			i = (i & 0xFF);
			max = (max >= i) ? max : i;
			min = (min <= i) ? min : i;
		}
		return (new int[] { min, max });
	}

	private static int[] getExtrema(short[] array) {
		int max = 0x0;
		int min = 0xFFFFF;
		for (int i : array) {
			i = (i & 0xFFFF);
			max = (max >= i) ? max : i;
			min = (min <= i) ? min : i;
		}
		return (new int[] { min, max });
	}

	private static int[] getExtrema(int[] array) {
		int max = 0x0;
		int min = 0xFFFFFFFF;
		for (int pixel : array) {
			pixel = (int) (0.299 * ((pixel >> 16) & 0xff) + 0.587 * ((pixel >> 8) & 0xff) + 0.114 * (pixel & 0xff));
			max = (max >= pixel) ? max : pixel;
			min = (min <= pixel) ? min : pixel;
		}
		return (new int[] { min, max });
	}

	private static byte[] normalizeGray(byte[] dataArray, final int lowest, final int highest) {

		if (highest <= lowest)
			throw new IllegalArgumentException(highest + " <= " + lowest);

		final int[] extrema = getExtrema(dataArray);
		// System.out.println("TBufferedImage.normalizeGray(), min = " +
		// extrema[0] + ", max = " + extrema[1]);

		// no amplitude rescaling needed, return source image
		if (extrema[0] == extrema[1])
			return dataArray;

		int min = Math.max(extrema[0], lowest);
		int max = Math.min(extrema[1], highest);

		// calculate factor and offset of the amplitude rescaling
		double constant = 255D / (max - min);
		double offset = constant * min;

		// amplitude rescaling of the image data
		int size = dataArray.length;
		for (int i = 0, pixel; i < size; i++) {
			pixel = (dataArray[i] & 0xff);
			if (pixel < min)
				pixel = min;
			if (pixel > max)
				pixel = max;
			dataArray[i] = (byte) (constant * pixel - offset);
		}
		return dataArray;
	}

	private static short[] normalizeGray(short[] dataArray, int colors, int lowest, int highest) {

		// System.out.println("TBufferedImage.normalizeGray(), colors = " +
		// colors);

		if (highest <= lowest)
			throw new IllegalArgumentException(highest + " <= " + lowest);

		final int[] extrema = getExtrema(dataArray);
		// System.out.println("TBufferedImage.normalizeGray(), extrema.min = " +
		// extrema[0] + ", extrema.max = " + extrema[1]);

		// no amplitude rescaling needed, return source image
		if (extrema[0] == extrema[1])
			return dataArray;

		int min = Math.max(extrema[0], lowest);
		int max = Math.min(extrema[1], Math.min(colors, highest));

		// System.out.println("TBufferedImage.normalizeGray(), min = " + min +
		// ", max = " + max);

		// calculate factor and offset for amplitude rescaling
		double constant = 1D * colors / (max - min);
		double offset = constant * min;

		// amplitude rescaling of the image data
		int size = dataArray.length;
		for (int i = 0, pixel; i < size; i++) {
			pixel = (dataArray[i] & colors);
			if (pixel < min)
				pixel = min;
			if (pixel > max)
				pixel = max;
			dataArray[i] = (short) (constant * pixel - offset);
		}
		return dataArray;
	}

	private static int[] normalizeRGB(int[] dataArray, final int lowest, final int highest) {

		// calculate min/max
		final int[] extrema = getExtrema(dataArray);
		// System.out.println("TBufferedImage.normalizeGray(), min = " +
		// extrema[0] + ", max = " + extrema[1]);

		int min = Math.max(extrema[0], lowest);
		int max = Math.min(extrema[1], highest);

		// create scaling factor and offset for stretching
		final double constant = 255D / (max - min);
		final double offset = constant * min;

		// normalize image data
		final int size = dataArray.length;
		for (int i = 0, pixel; i < size; i++) {
			pixel = dataArray[i];
			int r = (pixel >> 16) & 0xff; // red component
			int g = (pixel >> 8) & 0xff; // green component
			int b = pixel & 0xff; // blue component

			double dY = constant - (offset / (0.299 * r + 0.587 * g + 0.114 * b));
			r *= dY;
			g *= dY;
			b *= dY;

			if (r < 0)
				r = 0;
			if (g < 0)
				g = 0;
			if (b < 0)
				b = 0;
			if (r > 0xff)
				r = 0xff;
			if (g > 0xff)
				g = 0xff;
			if (b > 0xff)
				b = 0xff;

			dataArray[i] = 0xff000000 | (r << 16) | (g << 8) | (b);
		}
		return dataArray;
	}

	private static byte[] normalizeBGR(byte[] dataArray, int lowest, int highest) {
		// calculate min/max
		final int[] extrema = getExtrema(dataArray);
		// System.out.println("TBufferedImage.normalizeGray(), min = " +
		// extrema[0] + ", max = " + extrema[1]);

		int min = Math.max(extrema[0], lowest);
		int max = Math.min(extrema[1], highest);

		// create scaling factor and offset for stretching
		final double constant = 255D / (max - min);
		final double offset = constant * min;

		// normalize image data
		final int size = dataArray.length;
		for (int i = 0; i < size; i += 3) {
			int r = (dataArray[i + 2] & 0xff);
			int g = (dataArray[i + 1] & 0xff);
			int b = (dataArray[i] & 0xff);

			double dY = constant - (offset / (0.299 * r + 0.587 * g + 0.114 * b));
			r *= dY;
			g *= dY;
			b *= dY;

			if (r < 0)
				r = 0;
			if (g < 0)
				g = 0;
			if (b < 0)
				b = 0;
			if (r > 0xff)
				r = 0xff;
			if (g > 0xff)
				g = 0xff;
			if (b > 0xff)
				b = 0xff;

			dataArray[i + 2] = (byte) r; // R
			dataArray[i + 1] = (byte) g; // G
			dataArray[i] = (byte) b; // B
		}
		return dataArray;
	}
	
	/**
	 * Returns TINE image buffer byte-array data
	 * 
	 * @return image data
	 */
	private static byte[] getData(BufferedImage bi) {
		int[] intArray;
		byte[] byteArray;
		byte[] data;
		int length;
		DataBuffer dataBuffer = bi.getRaster().getDataBuffer();
		int imageType = bi.getType();
		switch (imageType) {
		case BufferedImage.TYPE_BYTE_GRAY:
			return ((DataBufferByte) dataBuffer).getData();
			
		case BufferedImage.TYPE_USHORT_GRAY:
			return getDataUShort((DataBufferUShort) dataBuffer);

		case BufferedImage.TYPE_INT_RGB:
		case BufferedImage.TYPE_INT_ARGB:
			intArray = ((DataBufferInt) dataBuffer).getData();
			length = intArray.length;
			data = new byte[3 * length];
			for (int i = 0, pos = 0; i < length; i++) {
				data[pos++] = (byte) (intArray[i] >>> 16); // red
				data[pos++] = (byte) (intArray[i] >>> 8); // green
				data[pos++] = (byte) (intArray[i] & 0xff); // blue
			}
			return data;

		case BufferedImage.TYPE_INT_BGR:
			intArray = ((DataBufferInt) dataBuffer).getData();
			length = intArray.length;
			data = new byte[3 * length];
			for (int i = 0, pos = 0; i < length; i++) {
				data[pos++] = (byte) (intArray[i] & 0xff); // red
				data[pos++] = (byte) (intArray[i] >>> 8); // green
				data[pos++] = (byte) (intArray[i] >>> 16); // blue
			}
			return data;

		case BufferedImage.TYPE_3BYTE_BGR:
			byteArray = ((DataBufferByte) dataBuffer).getData();
			length = byteArray.length;
			data = new byte[length];
			for (int i = 0, pos = 0; pos < length; i += 3) {
				data[pos++] = byteArray[i + 2]; // red
				data[pos++] = byteArray[i + 1]; // green
				data[pos++] = byteArray[i]; // blue
			}
			return data;

		case BufferedImage.TYPE_4BYTE_ABGR:
			byteArray = ((DataBufferByte) dataBuffer).getData();
			length = byteArray.length / 4 * 3;
			data = new byte[length];
			for (int i = 0, pos = 0; pos < length; i += 4) {
				data[pos++] = byteArray[i + 3]; // red
				data[pos++] = byteArray[i + 2]; // green
				data[pos++] = byteArray[i + 1]; // blue
			}
			return data;

		case BufferedImage.TYPE_CUSTOM:
			if (dataBuffer instanceof DataBufferUShort) {
				return getDataUShort((DataBufferUShort) dataBuffer);
			}

		default:
			intArray = bi.getRGB(0, 0, bi.getWidth(), bi.getHeight(), null, 0, bi.getWidth());
			length = intArray.length;
			data = new byte[3 * length];
			for (int i = 0, pos = 0; i < length; i++) {
				data[pos++] = (byte) (intArray[i] >>> 16); // red
				data[pos++] = (byte) (intArray[i] >>> 8); // green
				data[pos++] = (byte) (intArray[i] & 0xff); // blue
			}
			return data;
		}
	}

	/**
	 * Returns TINE image buffer byte-array data
	 * 
	 * @return image data
	 */
	private static byte[] getDataUShort(DataBufferUShort dataBuffer) {
		short[] shortArray = dataBuffer.getData();
		int length = shortArray.length;
		byte[] data = new byte[2 * length];
		for (int i = 0, pos = 0; i < length; i++) {
			data[pos++] = (byte) (shortArray[i]);
			data[pos++] = (byte) (shortArray[i] >>> 8);
		}
		return data;
	}

	public static BufferedImage toImageRGB(BufferedImage src) {
		final int type = src.getType();
		if (type == BufferedImage.TYPE_INT_RGB)
			return src;
		BufferedImage dest = new BufferedImage(src.getWidth(), src.getHeight(), BufferedImage.TYPE_INT_RGB);
		dest.getGraphics().drawImage(src, 0, 0, null);
		src.flush();
		return dest;
	}

}