/*
 * Created on Jun 13, 2007
 *
 */
package de.desy.mst.auxiliary.devl.deviceserverlib.config;

import javax.xml.bind.annotation.XmlRootElement;
import javax.xml.bind.annotation.XmlTransient;
import javax.xml.bind.annotation.XmlElement;

import org.apache.log4j.Logger;

import de.desy.mst.auxiliary.devl.deviceserverlib.tools.StringUtils;


/**
 * Configuration for input and output data types of operations.
 * An operation can have both fixed data types and user-defined types.
 * This class is some kind of a wrapper which can include either
 * a property type or a user-defined type.
 * @author jwlg
 */
@XmlRootElement
public class CfgOperationDataType {

    private CfgPropertyType propertyType = CfgPropertyType.INT;
    private CfgUserType userType = new CfgUserType();
    private boolean isArray = false;
    private int arraySize = 1;
    private boolean isUserDefined = false;
    
    private static Logger log = Logger.getLogger(CfgOperationDataType.class);

    /**
     * Returns the array size.
     */
    @XmlElement
    public int getArraySize() {
        return arraySize;
    }

    /**
     * Sets the array size.
     * @param arraySize New array size.
     */
    public void setArraySize(int arraySize) {
        this.arraySize = arraySize;
    }

    /**
     * Returns true if the data type is an array.
     */
    @XmlElement
    public boolean isArray() {
         return isArray;
    }

    /**
     * Specifies if the data type is an array.
     * @param isArray New value.
     */
    public void setArray(boolean isArray) {
        this.isArray = isArray;
    }

    /**
     * Returns the property type.
     */
    @XmlElement
    public CfgPropertyType getPropertyType() {
        return propertyType;
    }

    /**
     * Sets the property type.
     * Automatically sets the user-defined type to null.
     * @param propertyType
     */
    public void setPropertyType(CfgPropertyType propertyType) {
        if (propertyType != null) {
            this.propertyType = propertyType;
            this.userType = null;
            this.isUserDefined = false;
        } else {
            log.error("No property type specified (null)!");
        }
    }

    /**
     * Returns the user-defined type.
     */
    @XmlElement
    public CfgUserType getUserType() {
        return userType;
    }

    /**
     * Sets the user-defined type.
     * Automatically sets the property type to null.
     * @param userDefinedType
     */
    public void setUserType(CfgUserType userDefinedType) {
        if (userDefinedType != null) {
            if (!StringUtils.isEmpty(userDefinedType.getClassName())) {
                this.propertyType = null;
                this.userType = userDefinedType;
                this.isUserDefined = true;
            }
        } 
    }

    /**
     * Returns true if this is a user-defined type.
     */
    @XmlElement
    public boolean isUserDefined() {
        return isUserDefined && (this.userType != null);
    }


    /**
     * Specifies if this is a user-defiend type.
     * TODO Is this really needed?
     * @param isUserDefined New value.
     */
    public void setUserDefined(boolean isUserDefined) {
        this.isUserDefined = isUserDefined;
    }
    
    /**
     * Returns true if this is a property type.
     */
    @XmlTransient
    public boolean isProperty() {
        return (!isUserDefined) && (this.propertyType != null);
    }

    /**
     * Returns true if this is a simple type.
     * @see CfgPropertyType#isSimple
     */
    @XmlTransient
    public boolean isSimple() {
        return (isProperty() && getPropertyType().isSimple()); 
    }

    /**
     * Returns true if this type is a boolean.
     * @see CfgPropertyType#isBoolean
     */
    @XmlTransient
    public boolean isBoolean() {
        return ((!isUserDefined()) && getPropertyType().isBoolean()); 
    }

    /**
     * Returns true if this type is void.
     * @see CfgPropertyType#isVoid
     */
    @XmlTransient
    public boolean isVoid() {
        return (isProperty() && getPropertyType().isVoid()); 
    }

    /**
     * Returns true if this type is a string.
     * @see CfgPropertyType#isString
     */
    @XmlTransient
    public boolean isString() {
        return (isProperty() && getPropertyType().isString()); 
    }

    /**
     * Returns true if this type is a TINE type.
     * @see CfgPropertyType#isTineType
     */
    @XmlTransient
    public boolean isTineType() {
        return (isProperty() && getPropertyType().isTineType()); 
    }

    /**
     * Returns true if this type is a user-defined type
     * or a property type which can instantiate objects.
     */
    @XmlTransient
    public boolean isObject() {
        return (isUserDefined() || (isProperty() && getPropertyType().isObject()));
    }

}
