/*
 * Created on Mar 7, 2007
 *
  */
package de.desy.mst.auxiliary.devl.deviceserverlib.config;

import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.net.URL;

import javax.xml.bind.JAXBContext;
import javax.xml.bind.JAXBException;
import javax.xml.bind.Unmarshaller;

/**
 * A generalized JAXB reader.
 * 
 * @author jwlg
 */
public abstract class JaxbReader<T> {
    
    JAXBContext context = null;
    Class resultClass = null;
    
    /**
     * Constructor.
     * Builds a JaxbReader which reads an initialization
     * file and constructs an object tree with a 
     * specified root class.
     * @param clazz Class of the root element.
     */
    public JaxbReader(Class clazz) {
        super();
        initialize(clazz);
    }
    
    /**
     * Returns the JAXB context.
     * @throws JAXBException
     */
    private JAXBContext getContext() throws JAXBException {
        if (context == null) {    
            context = JAXBContext.newInstance(resultClass);
        }
        return context;
    }

    /**
     * Common initializer.
     * Called by constructors.
     * @param clazz
     */
    private void initialize(Class clazz) {
        resultClass = clazz;
    }
    
    /**
     * Reads an object tree from a file.
     * @param file The file to be read.
     * @return Object tree as read from a file.
     * @throws JAXBException 
     * @throws IOException 
     */
    public T read(File file) throws JAXBException, IOException {
        return readAndClose(new FileInputStream(file));
    }
    
    /**
     * Reads an object tree from an URL.
     * @param url The URL to be read from.
     * @return Read Object tree.
     * @throws JAXBException
     * @throws IOException
     */
    public T read(URL url) throws JAXBException, IOException {
        return readAndClose(url.openStream());
    }
    
    /**
     * Reads from a file or URL.
     * Interprets the parameter as a file first. If that
     * doesn't work, interprets it as an URL. 
     * @param fileOrUrl File name or URL.
     * @return Read Object tree.
     * @throws JAXBException
     * @throws IOException
     */
    public T read(String fileOrUrl) throws JAXBException, IOException {
        File file = new File(fileOrUrl);
        if (file.exists()) {
            return readAndClose(new FileInputStream(file));
        } else {
            URL url = new URL(fileOrUrl);
            return readAndClose(url.openStream());
        }
    }
    
    /**
     * Reads from an input stream.
     * @param inputStream
     * @return Read object tree.
     * @throws JAXBException
     */
    @SuppressWarnings("unchecked")
    public T read(InputStream inputStream) throws JAXBException {
        Unmarshaller um = getContext().createUnmarshaller(); 
        T result = (T) um.unmarshal(inputStream);
        return result;
    }
    
    /**
     * Reads from an input stream and closes the stream.
     * For internal use.
     * @param inputStream INput stream.
     * @throws JAXBException
     * @throws IOException
     */
    private T readAndClose(InputStream inputStream) throws JAXBException, IOException {
        T result = read(inputStream);
        inputStream.close();
        return result;
    }
    
}
