/*
 * Created on Mar 27, 2007
 *
 */
package de.desy.mst.auxiliary.devl.deviceserverlib.initialization;

import java.sql.Connection;
import java.sql.DriverManager;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Statement;
import java.util.HashMap;
import java.util.Properties;

import de.desy.mst.auxiliary.devl.deviceserverlib.config.CfgInitType;

/**
 * A Generic JDBC Table driver.
 * This class is used internally for JDBC drivers.
 * In opposite to JdbcDriver, the driver class, 
 * prefix and driver properties cannot be configured
 * externally, but only via the constructor. The 
 * GenericTableDriver is especially used for all Stels 
 * JDBC drivers. 
 * 
 * @author jwlg
 */
public class GenericDriver implements IDatabaseDriver {

    String driverClassName = "";
    String urlPrefix = "";
    Properties driverProperties = null;
    Connection connection = null;

    /**
     * Constructor.
     * Useful for drivers without specific driver properties.
     * @param driverClassName The class name of the driver.
     * @param urlPrefix The URL prefix.
     */
    public GenericDriver(String driverClassName, String urlPrefix) {
        this.driverClassName = driverClassName;
        this.urlPrefix = urlPrefix;
        this.driverProperties = null;
    }

    /**
     * Constructor.
     * @param driverClassName  The class name of the driver.
     * @param urlPrefix The URL prefix.
     * @param driverProperties The driver properties. These are passed to DriverManager.openConnection(). Null is allowed.
     */
    public GenericDriver(String driverClassName, String urlPrefix, Properties driverProperties) {
        this.driverClassName = driverClassName;
        this.urlPrefix = urlPrefix;
        this.driverProperties = driverProperties;
    }       

    /**
     * Returns the default properties.
     * Returns a hash map with the keys "dataSource" and "sqlQuery"
     * and empty strings as values. These properties are needed
     * by the getResult() method.
     * @see #getResult(HashMap)
     */
    public HashMap<String, String> getDefaultProperties() {
        HashMap<String,String> params = new HashMap<String,String>();
        params.put("dataSource", "");
        params.put("sqlQuery", "");
        return params;
    }

    public CfgInitType getInitType() {
        return null;
    }

    /**
     * Returns a result set of an SQL query.
     * @param properties A hash map of properties. Must contain the keys "dataSource" and "sqlQuery".
     */
    public ResultSet getResult(HashMap<String, String> properties)
    throws ClassNotFoundException, SQLException {

        ResultSet resultSet = null;
        String directory = properties.get("dataSource");
        String sqlQuery = properties.get("sqlQuery");
        
        /*
         * If no data source is given, use the device server's
         * database directory.
         */
        if ((directory == null) || directory.equals("")) {
            directory = properties.get("system.defaultDirectory");
        }

        if ((sqlQuery == null) || sqlQuery.equals("")) {
            sqlQuery = properties.get("table");
        } 
        sqlQuery = SelectUtil.extendSelect(sqlQuery);

        String url = urlPrefix + directory;
        Class.forName(driverClassName);
        
        if (driverProperties == null)
            connection = DriverManager.getConnection(url);
        else 
            connection = DriverManager.getConnection(url, driverProperties);

        Statement statement = connection.createStatement();
        resultSet = statement.executeQuery(sqlQuery);
        connection.close();
        return resultSet;

    }
    
    public void close() {
        if (connection != null) {
            try {
                connection.close();
            } catch (SQLException e) {
                e.printStackTrace();
            }
        }
    }

    @Override
    protected void finalize() throws Throwable {
        super.finalize();
        close();
    }

}
