package de.desy.mst.auxiliary.devl.deviceserverlib.server;

import java.util.Set;
import java.util.TreeSet;
import java.util.concurrent.Semaphore;

import de.desy.mst.auxiliary.devl.deviceserverlib.alarms.Alarm;

/**
 * Abstract ABasicDevice.
 * An anstract base class of all device classes.
 * ABasicDevice implements basic functionality which is 
 * use by all devices. It supplies a device context,
 * a set of alarms, and a semaphore which can be used
 * to lock the device while a sensitive operation 
 * is running.
 * 
 * @author jwlg
 *
 * @param A The initial record class.
 */
public abstract class ABasicDevice implements IBasicDevice {

    Semaphore semaphore = new Semaphore(1);
    DeviceContext deviceContext = null;
	TreeSet<Alarm> alarms = new TreeSet<Alarm>();
    private String deviceName = "";
    
    /**
     * Default Constructor.
     * Note: Device context needs to be set after contruction.
     */
    public ABasicDevice() {
        super();
    }

    public ABasicDevice(String deviceName) {
        this.deviceName = deviceName;
    }
    
    /**
     * Constructor.
     * @param deviceContext The device context.
     */
	public ABasicDevice(DeviceContext deviceContext) {
		super();
        this.deviceContext = deviceContext;
	}
		
    /**
     * Tries to lock the device.
     * If the device is locked, this method waits until
     * the device is available.
     */
    public synchronized void acquireLock() throws InterruptedException {
        semaphore.acquire();
    }

    /**
     * Releases the device lock.
     * This method unocks the device. It has to be called
     * after an acquireLock() and after a successful
     * tryAcquireLock() in order to release the lock
     * and make the device available for other clients. 
     */
    public synchronized void releaseLock() {
        semaphore.release();
    }

    /**
     * Tries to lock the device.
     * If the device is already locked, this method
     * immediately returnes false. Otherwise it locks
     * the device and returns true. Note that 
     * releaseLock() has to be called in order to 
     * release the obtained lock.
     * 
     */
    public synchronized boolean tryAcquireLock() {
        return semaphore.tryAcquire();
    }
    
    /**
     * Returns the device context.
     * @see DeviceContext
     */
    public DeviceContext getContext() {
        return deviceContext;
    }

    /**
     * Sets an alarm.
     * A device can set alarms. An alarm
     * is an object which is put into the 
     * device's alarm set and remains there
     * until clearAlarms() or clearAlarm()
     * has been called. 
     * Each alarm is unique in the alarm list.
     * Alarms are cleared and set in the update() method.
     * @param alarm The alarm which is to be set.
     */
    protected void setAlarm(Alarm alarm) {
        alarms.add(alarm);
    }
    
    /**
     * Removes a single alarm from the alarm list.
     * 
     * @param alarm The alarm which is to be removed.
     */
    protected void clearAlarm(Alarm alarm) {
        alarms.remove(alarm);
    }
    
    /**
     * Removes all alarms from the alarm list.
     * In the update method, each device has to call clearAlarms()
     * first and then sets its alarms.
     */
    public void clearAlarms() {
        alarms.clear();
    }
    
    /**
     * Returns the set of active alarms.
     */
    public Set<Alarm> getAlarms() {
        return alarms;
    }

    /**
     * Returns the device name.
     */
    public String getDeviceName() {
        return deviceName;
    }
    
    /**
     * Assigns the device name.
     * <P>
     * <B>NOTE:</B> The device name has to be set only once 
     * when the device is constructed. It must not be changed
     * afterwards!
     * @param deviceName New device name.
     */
    public void setDeviceName(String deviceName) {
        this.deviceName = deviceName;
    }
    
    protected final DeviceContext getDeviceContext() {
        return deviceContext;
    }

    public final void setDeviceContext(DeviceContext deviceContext) {
        this.deviceContext = deviceContext;
    }

    
}
