/*
 * Created on Apr 5, 2007
 *
  */
package de.desy.mst.auxiliary.devl.deviceserverlib.server;

import de.desy.tine.server.connections.TClient;

/**
 * Base class for property handlers.
 * This base class combines basic data and methods
 * used by APropertyHandler and AOperationHandler.
 * 
 * Each device group must have a property handler for each property
 * of its device type..
 * A property handler takes a TINE request, selects the addressed
 * devices, marshals and unmarshals the data, and calls the 
 * devices' method corresponding to the property.
 * 
 *  The implementation of property handlers is done by the 
 *  device server wizard's code generator.
 * 
 * @author jwlg
 */
public abstract class ABasicPropertyHandler<D extends IBasicDevice> implements IPropertyHandler {
    
    private DeviceGroup<D> deviceGroup;
    private String propertyName = "";
    private EquipmentModule<D> equipmentModule;
    
    /**
     * Constructor.
     * 
     * 
     * @param propertyName Name of the property.
     * @param deviceGroup Device group to which the property belongs.
     */
    public ABasicPropertyHandler(String propertyName, DeviceGroup<D> deviceGroup) {
        super();
        this.deviceGroup = deviceGroup;
        this.propertyName = propertyName;
    }
    
    /**
     * Returns the device group.
     * Used by derived classes to get the device group.
     */
    protected final DeviceGroup<D> getDeviceGroup() {
        return deviceGroup;
    }
    
    /**
     * Returns true if an index corresponds to a device in the device group.
     * Used by derived classes in TINE calls.
     * @param number The index.
     */
    protected final boolean isValidDeviceIndex(int number) {
        return ((number >= 0) && (number < deviceGroup.size()));
    }
    
    /**
     * Returns true if the supplied number is less than or equal to the number of devices.
     * Used by derived classes to ensure that a given number of devices
     * is 
     * @param nDevices Number of devices.
     */
    protected final boolean isValidDimension(int nDevices) {
        return ((nDevices > 0) && (nDevices <=deviceGroup.size()));
    }
    
    /**
     * Returns tue if a given device number and a given number of devices
     * are valid as a combination.
     * Used by derived classes to verify TINE addressing parameters.
     * Internally uses isValidDeviceIndex() and isValidDimension().
     * 
     * @param deviceIndex The index of the device.
     * @param nDevices The number if devices.
     */
    protected final boolean isValidSubsetDimension(int deviceIndex, int nDevices) {
        return (isValidDeviceIndex(deviceIndex) && isValidDimension(deviceIndex + nDevices));
    }
    
    /**
     * Returns a device specified by an index.
     * A convenience method which wraps the DeviceGroup method of the same name. 
     * @param deviceIndex The index.
     * @see DeviceGroup#getDeviceByIndex
     */
    public final D getDeviceByIndex(int deviceIndex) {
        return deviceGroup.getDeviceByIndex(deviceIndex);
    }

    /**
     * Returns a device specified by a name.
     * A convenience method which wraps the DeviceGroup method of the same name. 
     * @param deviceName
     * @see DeviceGroup#getDeviceByName
     */
    public final D getDeviceByName(String deviceName) {
        return deviceGroup.getDeviceByName(deviceName);
    }

    /**
     * Returns a device specified by a TINE name.
     * A convenience method which wraps the DeviceGroup method of the same name. 
     * @param tineDeviceName The device name or device number preceded by a hash.
     * @see DeviceGroup#getDeviceByTineName
     */
    public final D getDeviceByTineName(String tineDeviceName) {
        return deviceGroup.getDeviceByTineName(tineDeviceName);
    }

    /**
     * Returns the device index which correspnds to a device name.
     * A convenience method which wraps the DeviceGroup method of the same name. 
     * @param deviceName The device name.
     * @see DeviceGroup#getDeviceIndexByName
     */
    public final int getDeviceIndexByName(String deviceName) {
        return deviceGroup.getDeviceIndexByName(deviceName);
    }

    /**
     * Returns the device index which corresponds to a TINE device name.
     * A convenience method which wraps the DeviceGroup method of the same name. 
     * 
     * @param tineDeviceName The device name or device number preceded by a hash.
     * @see DeviceGroup#getDeviceIndexByName
     */
    public final int getDeviceIndexByTineName(String tineDeviceName) {
        return deviceGroup.getDeviceIndexByTineName(tineDeviceName);
    }
    
    /**
     * Returns the property name.
     */
    public final String getPropertyName() {
        return propertyName;
    }
    
    /**
     * Returns the client.
     * This is useful if the derived property handler needs information
     * about the caller. Currently used for logging purposes only.
     * 
     */
    protected TClient getClient() {
        return equipmentModule.getCaller();
    }

    /**
     * Returns the equipment module.
     * This method gives derivec classes access to the equipment module.
     */
    protected EquipmentModule<D> getEquipmentModule() {
    
        return equipmentModule;
    }
    
    /**
     * Assigns the equipment module.
     * @param equipmentModule The equipment module.
     */
    protected void setEquipmentModule(EquipmentModule<D> equipmentModule) {
    
        this.equipmentModule = equipmentModule;
    }
    
    
}
