package de.desy.mst.auxiliary.devl.deviceserverlib.server;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;

import org.apache.log4j.Logger;

/**
 * A group of devices.
 * 
 * A device group is a set of devices which have the 
 * same interface. That interface is derived from 
 * IBasicDevice.
 * 
 * The devices contained in the group are indexed by name
 * and by number. The number corresponds to the order in 
 * which the devices have been added to the group.
 * 
 * Furthermore, DeviceGroup provides methods to lock 
 * devices in common or as a subset. 
 * 
 * @author jwlg
 */
public class DeviceGroup<T extends IBasicDevice> implements Iterable<T> {

    @SuppressWarnings("unused")
    private static Logger log = Logger.getLogger(DeviceGroup.class);
    private DeviceMap<T> deviceMap;
    private DeviceLockManager deviceLockManager = null;
    private String name = ""; /* A device group knows its name; important for logging/debugging!" */
    
    public DeviceGroup() {
        deviceMap = new DeviceMap<T>();
        deviceLockManager = new DeviceLockManager(deviceMap);
    }

    public DeviceLockSet acquireAllDeviceLocks() {

        return deviceLockManager.acquireDeviceLocks();
    }

    public DeviceLockSet acquireDeviceLocks(int[] deviceIndices) {

        return deviceLockManager.acquireDeviceLocks(deviceIndices);
    }

    public DeviceLockSet acquireDeviceLocks(List<String> deviceNames) {

        return deviceLockManager.acquireDeviceLocks(deviceNames);
    }

    public DeviceLockSet acquireLocks(List<IBasicDevice> devices) {

        return deviceLockManager.acquireLocks(devices);
    }

    public DeviceLockSet tryAcquireAllDeviceLocks() {

        return deviceLockManager.tryAcquireDeviceLocks();
    }

    public DeviceLockSet tryAcquireDeviceLocks(int[] deviceIndices) {

        return deviceLockManager.tryAcquireDeviceLocks(deviceIndices);
    }

    public DeviceLockSet tryAcquireDeviceLocks(List<String> deviceNames) {

        return deviceLockManager.tryAcquireDeviceLocks(deviceNames);
    }

    public DeviceLockSet tryAcquireLocks(List<IBasicDevice> devices) {

        return deviceLockManager.tryAcquireLocks(devices);
    }

    public void addDevice(T device) {

        deviceMap.addDevice(device);
    }

    public ArrayList<T> getDeviceArrayList() {

        return deviceMap.getDeviceArrayList();
    }

    public T getDeviceByIndex(int deviceIndex) {

        return deviceMap.getDeviceByIndex(deviceIndex);
    }

    public T getDeviceByName(String deviceName) {

        return deviceMap.getDeviceByName(deviceName);
    }

    public T getDeviceByNameIgnoreCase(String deviceName) {

        return deviceMap.getDeviceByNameIgnoreCase(deviceName);
    }

    public T getDeviceByTineName(String tineDeviceName) {

        return deviceMap.getDeviceByTineName(tineDeviceName);
    }

    public int getDeviceIndexByName(String deviceName) {

        return deviceMap.getDeviceIndexByName(deviceName);
    }

    public int getDeviceIndexByNameIgnoreCase(String deviceName) {

        return deviceMap.getDeviceIndexByNameIgnoreCase(deviceName);
    }

    public int getDeviceIndexByTineName(String tineDeviceName) {
        return deviceMap.getDeviceIndexByTineName(tineDeviceName);
    }

    public List<T> getDeviceList() {
        return deviceMap.getDeviceList();
    }

    public List<T> getDeviceListByNames(List<String> deviceNames) {
        return deviceMap.getDeviceListByNames(deviceNames);
    }

    public String getDeviceNameByIndex(int deviceIndex) {
        return deviceMap.getDeviceNameByIndex(deviceIndex);
    }

    public List<String> getDeviceNames() {
        return deviceMap.getDeviceNames();
    }

    public void removeAllDevices() {
        deviceMap.removeAllDevices();
    }

    public void removeDevice(int deviceIndex) {
        deviceMap.removeDevice(deviceIndex);
    }

    public void removeDevice(String deviceName) {
        deviceMap.removeDevice(deviceName);
    }

    public void removeDevice(T device) {
        deviceMap.removeDevice(device);
    }

    public int size() {
        return deviceMap.size();
    }

    public Iterator<T> iterator() {
        return deviceMap.iterator();
    }
    
    public ArrayList<T> getDevices() {
        ArrayList<T> result = new ArrayList<T>();
        result.addAll(deviceMap.getDeviceArrayList());
        return result;
    }

    public LinkedList<T> getDeviceSubset(String deviceName, int nDevices) {
        return deviceMap.getDeviceSubset(deviceName, nDevices);
    }

    public LinkedList<T> getSubset(int deviceIndex, int nDevices) {
        return deviceMap.getSubset(deviceIndex, nDevices);
    }
    
    /**
     * 
     * @return The device map.
     */
    DeviceMap<T> getDeviceMap() {
        return deviceMap;
    }
    
    /**
     * Returns the device group name.
     * This is also the TINE export name.
     */
    public String getName() {
        return name;
    }

    /**
     * Assigns the device group name.
     * This method is called by the equipment module builder.
     * @param name The name of the device group.
     */
    public void setName(String name) {
        this.name = name;
    }
    
    public void setDeviceContext(DeviceContext context) {
        for (T device : deviceMap) {
            if (device instanceof ABasicDevice) { 
                device.setDeviceContext(context);
            }
        }
    }

}
