/*
 * Created on Sep 24, 2007
 *
 */
package de.desy.mst.auxiliary.devl.deviceserverlib.server;

import org.apache.log4j.Logger;

/**
 * A class which performs and monitors a device lock.
 * 
 * This is a helper class for device server programmers.
 * The direct way to use device locking is not quit safe 
 * and has some pitfalls, e.g. it is possible to unlock 
 * a device which somebody else has locked, and it is easy
 * to forget to unlock a locked device.
 * 
 *  The DeviceLock unlocks a device only if it has locked 
 *  it itself. A finalizer cares for an unreleased lock 
 *  when the DeviceLock is deleted. It checks if the lock is set
 *  lock and releases it if necessary. 
 *  
 * @author jwlg
 */
public class DeviceLock {

    IBasicDevice device = null;
    boolean isLocked = false;
    
    private static Logger log = Logger.getLogger(DeviceLock.class);

    
    /**
     * Constructor.
     * @param device The device to lock/unlock.
     */
    public DeviceLock(IBasicDevice device) {
        this.device = device;
    }


    /**
     * Tries to lock a device.
     * Waits until the device is locked.
     * 
     * @throws InterruptedException
     */
    public synchronized void acquireLock() throws InterruptedException {
        try {
            this.device.acquireLock();
            isLocked = true;
        } catch (InterruptedException e) {
            isLocked = false;
            throw e;
        }
    }

    /**
     * Tries to lock the device.
     * Returns immediately.
     * 
     * @return True if the device is locked
     * or was already locked by this DeviceLock.
     * False if the device cannot be locked.
     */
    public synchronized boolean tryAcquireLock() {
        if (!isLocked) {
            isLocked = device.tryAcquireLock();
        }
        return isLocked;
    }

    /**
     * Releases the lock.
     */
    public synchronized void release() {
        if (isLocked) {
            device.releaseLock();
            isLocked = false;
        }
    }

    /**
     * Unlocks the device if necessary.
     */
    @Override
    protected void finalize() throws Throwable {
        super.finalize();
        if ((device != null) && (isLocked)) {
            log.warn("Releasing device " + device.getDeviceName() + "in finalizer!");
            release();
        }
    }

}
