/*
 * Created on Mar 21, 2007
 *
 */
package de.desy.mst.auxiliary.devl.deviceserverlib.server;

import java.io.File;
import java.util.prefs.BackingStoreException;
import java.util.prefs.Preferences;

import org.apache.log4j.Logger;

/**
 * A Directory set factory which creates a standard directory set.
 * A standard directory set has a home directory and subdirectories
 * as specified in DirectorySet.
 *  
 * @author jwlg
 * @see DirectorySet
 * @see #createDirectorySet(String)
 */
public class DirectorySetFactory implements IDirectorySetFactory {

    private static DirectorySetFactory theInstance = new DirectorySetFactory();
    private Logger log = Logger.getLogger(DirectorySetFactory.class);

    /**
     * The root node in the preferences.
     */
    public static String ROOT_NODE = "/de/desy/mst/deviceserverwizard/server";

    /**
     * Private constructor. 
     * Ensures that this class can have only one instance.
     * @see #getInstance()
     */
    private DirectorySetFactory() {
    }

    /**
     * Creates a directory set from a given application name.
     * 
     * 
     * The method searches for a root node in the user preferences.
     * In the root node, a common root directory set be specified
     * for all applications, or root and subdirectories can be
     * specified for any particular application.
     * <p>
     * If the root node contains a subnode with the application name 
     * as its key, an individual configuration for this application is
     * expected in that node. At minimum, the key "baseDirectory" must be
     * specified. All subdirectories can be specified separately here, 
     * although it is not necessary. By default, a standard structure is 
     * assumed under the base directory. But each of the subdirectories
     * can be specified individually as well. The node names are 
     * "logDirectory", "configDirectory" ,"historyDirectory", 
     * "workDirectory", and "databaseDirectory".
     * <p> 
     * If the root node contains no application-specific configuration, 
     * the key "applicationRoot" in the root node must be specified. 
     * The factory then creates a directory set with a top directory
     * and subdirectories under the application root. The topdirectory's
     * name is the application name. The subdirectories are named
     * as specified in createDefaultDirectorySet().
     * <p>
     * If nothing exists in the user preferences, the current directory
     * is taken as the base directory. All subdirectories are direct
     * subdirectories of the current directory.
     *  
     * 
     * @param applicationName
     * @see #ROOT_NODE
     */
    public DirectorySet createDirectorySet(String applicationName) {

        DirectorySet result = null;
        Preferences userPrefs = Preferences.userRoot().node(ROOT_NODE);
        Preferences appPrefs = null;
        String prefsApplicationRoot = userPrefs.get("applicationRoot", "");
        String systemUserDir = System.getProperty("user.dir");
        String systemBaseDir = System.getProperty("base.dir");
        String appPrefsBaseDir = null;

        try {
            if (userPrefs.nodeExists(applicationName)) {
                appPrefs = userPrefs.node(applicationName);
                appPrefsBaseDir = appPrefs.get("baseDirectory", "");
            }
        } catch (BackingStoreException e) {
            log.error("Cannot find user preferences node for " + applicationName + ": " + e, e);
        }

        log.info("Determining base directory for " + applicationName);

        // 1. Determine the Base directory
        if (!(systemBaseDir == null || systemBaseDir.equals(""))) {
            log.info("Using base directory -Dbase.dir=\"" + systemBaseDir + "\".");
            result = createDefaultDirectorySet(systemBaseDir);
        } else if ((appPrefsBaseDir != null) && (!appPrefsBaseDir.equals(""))) {
            log.info("Using base directory from user preferences: \"" + appPrefsBaseDir + "\"" );
            result = createDefaultDirectorySet(appPrefsBaseDir);
        } else if (!prefsApplicationRoot.equals("")) {
            log.info("Using subdirectory " + applicationName + " under application root as specified in user preferences .");
            result = createDefaultDirectorySet(prefsApplicationRoot + File.separator + applicationName);
        } else {
            log.warn("No base directory specified. Using current directory!");
            result = createDefaultDirectorySet(systemUserDir);
        }

        // 2. Override default settings with application-specific settings
        if (appPrefs != null) {
            String overrideDir = null;

            overrideDir = appPrefs.get("logDirectory", result.getLogDirectory());
            if (!overrideDir.equals(result.getLogDirectory())) {
                log.info("Overriding Log Directory: " + overrideDir);
                result.setLogDirectory(overrideDir);
            }

            overrideDir = appPrefs.get("historyDirectory", result.getHistoryDirectory());
            if (!overrideDir.equals(result.getHistoryDirectory())) {
                log.info("Overriding History Directory: " + overrideDir);
                result.setHistoryDirectory(overrideDir);
            }

            overrideDir = appPrefs.get("workDirectory", result.getWorkDirectory());
            if (!overrideDir.equals(result.getWorkDirectory())) {
                log.info("Overriding Work Directory: " + overrideDir);
                result.setWorkDirectory(overrideDir);
            }

            overrideDir = appPrefs.get("configDirectory", result.getConfigDirectory());
            if (!overrideDir.equals(result.getConfigDirectory())) {
                log.info("Overriding Config Directory: " + overrideDir);
                result.setConfigDirectory(overrideDir);
            }

            overrideDir = appPrefs.get("databaseDirectory", result.getDatabaseDirectory());
            if (!overrideDir.equals(result.getDatabaseDirectory())) {
                log.info("Overriding Database Directory: " + overrideDir);
                result.setDatabaseDirectory(overrideDir);
            }
        } 

        return result;
    }

    /**
     * Creates a directory set under a specified home directory.
     * The directory set consists of the four subdirectories
     * "log", "history", "work", and 'config".
     * This methods does not create anything in the file system.
     * It only creates the path strings.
     * @param homeDir Full path of the home directory.
     * @return A directory set.
     */
    public DirectorySet createDefaultDirectorySet(String homeDir) {
        DirectorySet result = new DirectorySet();
        result.setBaseDirectory(homeDir);
        result.setLogDirectory(homeDir + File.separator + "log");
        result.setHistoryDirectory(homeDir + File.separator + "history");
        result.setWorkDirectory(homeDir + File.separator + "work");
        result.setConfigDirectory(homeDir + File.separator + "config");
        result.setDatabaseDirectory(homeDir + File.separator + "database");
        return result;
    }

    /**
     * Returns the instance.
     */
    public static DirectorySetFactory getInstance() {
        return theInstance;
    }

}
