/*
 * Created on May 11, 2006
 *
 */
package de.desy.mst.auxiliary.devl.deviceserverlib.server;

import java.util.HashMap;
import java.util.Set;

import org.apache.log4j.Logger;

import de.desy.mst.auxiliary.devl.deviceserverlib.alarms.Alarm;
import de.desy.mst.auxiliary.devl.deviceserverlib.alarms.TineAlarmFactory;
import de.desy.mst.auxiliary.devl.deviceserverlib.config.CfgDeviceGroup;
import de.desy.mst.auxiliary.devl.deviceserverlib.config.CfgDeviceServer;
import de.desy.mst.auxiliary.devl.deviceserverlib.config.CfgDeviceType;
import de.desy.mst.auxiliary.devl.deviceserverlib.config.CfgOperation;
import de.desy.mst.auxiliary.devl.deviceserverlib.config.CfgProperty;
import de.desy.mst.auxiliary.devl.deviceserverlib.history.HistoryMap;
import de.desy.mst.auxiliary.devl.deviceserverlib.tools.ModelUtils;
import de.desy.mst.auxiliary.devl.deviceserverlib.tools.tine.LocalNameGenerator;
import de.desy.mst.auxiliary.devl.deviceserverlib.tools.tine.TineClassFactory;
import de.desy.tine.definitions.TArrayType;
import de.desy.tine.server.alarms.TAlarmDefinition;
import de.desy.tine.server.devices.TDevice;
import de.desy.tine.server.devices.TDeviceList;
import de.desy.tine.server.equipment.TEquipmentModule;
import de.desy.tine.server.histories.THistorySpecification;
import de.desy.tine.server.properties.TExportProperty;

/**
 * An equipment module.
 * 
 * This class represents a TINE equipment module.
 * It connects a device group with a TINE equipment module and
 * facilitates the installation of all properties, operations,
 * alarms, etc.
 * 
 * For every device group, an EquipmentModule is created by the
 * Module factory.
 * 
 * @author jwlg
 * @see AModuleFactory
 * 
 * TODO better INCLUDE a TEquipmentModule instead of extending it!!!
 */
public class EquipmentModule<D extends IBasicDevice> extends TEquipmentModule {

    private DeviceGroup<D> deviceGroup = null;
    private PropertySet<D> propertySet = null;
    
    private CfgDeviceServer cfgDeviceServer = null;
    private CfgDeviceGroup cfgDeviceGroup = null;
    private CfgDeviceType cfgDeviceType = null;
    private ACommonData<D> commonData = null;
    
    
    private static Logger log = Logger.getLogger(EquipmentModule.class);

    /**
     * Constructor.
     * The constructor needs the export name as an input parameter.
     * It generates a local name internally. The local name is only 
     * needed by TINE.
     * 
     * @param exportName The export name. This is the name
     * which in TINE is also called "Device Server".
     */
    public EquipmentModule(String exportName) {
        super(getNextLocalName(exportName));
        setExportName(exportName);
    }
    
    /**
     * Assigns the device server configuration.
     * @param cfgDeviceServer The device server configuration.
     */
    public void setDeviceServerConfig(CfgDeviceServer cfgDeviceServer) {
        this.cfgDeviceServer = cfgDeviceServer;
    }
    
    
    /** 
     * Assigns the device group configuration.
     * 
     * An equipment module either belongs to a control group or to
     * a device group. If the controller group configuration is set,
     * the device group configuration is set to null, and vice versa.
     * 
     * @param cfgDeviceGroup The device group configuration.
     */
    public void setDeviceGroupConfig(CfgDeviceGroup cfgDeviceGroup) {
        this.cfgDeviceGroup = cfgDeviceGroup;
        this.cfgDeviceType = cfgDeviceGroup.getDeviceType();
    }
    
    /**
     * Creates a local name.
     * @param exportName The export name.
     * @see LocalNameGenerator
     */
    private static String getNextLocalName(String exportName) {
        return LocalNameGenerator.getLocalName(exportName);
    }

    /**
     * Registers an operation.
     * 
     * Registers a property with TINE and 
     * assigns an operation handler to it.
     * 
     * @param operation The operation handler.
     */
    private void registerOperation(AOperationHandler<D> operation) {
        try {
            String operationName = operation.getOperationName();
            TExportProperty tExportProperty= null;
            CfgOperation cfgOperation = cfgDeviceType.findOperation(operationName);
            if (cfgOperation == null) {
                log.error("Cannot register operation " + operation.getOperationName() + ": No configuration available.");
            } else {
                tExportProperty = 
                    TineClassFactory.getInstance().getTExportProperty(
                        cfgDeviceServer, cfgOperation);
                registerProperty(tExportProperty,
                    new PropertyHandlerBridge<D>(this, operation));
                log.info("Operation " + operation.getOperationName() + " registered.");
            }    
        } catch (Exception e) {
            log.error("Registering Operation " + operation.getOperationName(),e);
        }
    }
    
    /**
     * Registers a property.
     * 
     * Registers a property with TINE and
     * assigns a property handler to it.
     * 
     * @param property The property handler.
     */
    private void registerProperty(APropertyHandler<D> property) {
        
        TExportProperty tExportProperty= null;
        String propertyName = property.getPropertyName();
        
        /* Find property configuration from device type. */
        CfgProperty cfgProperty = cfgDeviceType.findProperty(propertyName);
        
        /* Not found? Maybe it's a standard property?? */
        if (cfgProperty == null) {
            cfgProperty = ModelUtils.getInstance().findStandardControllerProperty(propertyName);
        }
        
        if (cfgProperty == null) {
            log.error("Cannot register property " + property.getPropertyName() + ": No property configuration available.");
        } else {

            tExportProperty = TineClassFactory.getInstance().getTExportProperty(
                cfgProperty,
                deviceGroup.size());
            registerProperty(tExportProperty,
                new PropertyHandlerBridge<D>(this, property));
            log.info("Property " + property.getPropertyName() + " registered.");
        }
    }
    
    /**
     * Registers alarms.
     * RegisterAlarms must be called from OUTSIDE the
     * equipment module because of TINE oddites.
     * @param alarmSystemID Alarm System ID.
     * @param alarmModuleDescription The description of the alarm module.
     * @param alarms The set of alarms to be registered.
     * 
     * TODO Remove the alarmSystemID. A default of 0 should be sufficient.
     * 
     */
    @SuppressWarnings("unchecked")  // For TINE getAlarmDefinitionList()
    public void registerAlarms(
            int alarmSystemID,
            String alarmModuleDescription,
            Set<Alarm> alarms) {
        TineAlarmFactory aFactory = new TineAlarmFactory(
            alarmSystemID, 
            alarmModuleDescription);
        log.info("Registering alarms. (n=" + alarms.size() + ")");
        for (Alarm alarm : alarms) {
            TAlarmDefinition alarmDef = aFactory.getAlarmDefinition(alarm);
            log.info("Adding Alarm Definition [" + alarmDef.getAlmCode() + "]: " + alarmDef.getAlmText() );
            addAlarmDefinition(alarmDef);
        }
    }
 
    /**
     * Assigns a device group and registers its devices in TINE.
     * @param deviceGroup The device group.
     */
    public void registerDevices(DeviceGroup<D> deviceGroup) {
        this.deviceGroup = deviceGroup;
        for (D currentDevice : this.deviceGroup ) {
            registerDeviceName(currentDevice.getDeviceName());
        }
    }
    
    /**
     * Registers the properties which belong to the device set.
     * 
     * Registers all property handlers and operation handlers
     * in the property set.
     * 
     * @param properties A property set.
     */
    public void registerProperties(PropertySet<D> properties) {
        
        this.propertySet = properties;
 
        for (APropertyHandler<D> property : properties.getProperties()) {
                 registerProperty(property);
        }
        
        for (AOperationHandler<D> operation : properties.getOperations()) {
            registerOperation(operation);
        }
        
    }
    
    /**
     * Registers all local histories as specified in the history map.
     * 
     * @param historyMap The history map.
     */
    public void registerLocalHistories(HistoryMap historyMap) {
        HashMap<String, CfgProperty> propertyLookupList = new HashMap<String,CfgProperty>();
        
        if (cfgDeviceGroup != null) {
            for (CfgProperty property : cfgDeviceGroup.getDeviceType().getProperties()) {
                propertyLookupList.put(property.getName(), property);
            }
        } 
        
        if (propertySet != null) {
            for (APropertyHandler<D> property : propertySet.getProperties())
            {
                CfgProperty propertyConfig = propertyLookupList.get(property.getPropertyName());
                
                if (propertyConfig == null) {
                    propertyConfig = ModelUtils.getInstance().findStandardControllerProperty(property.getPropertyName());
                }
                
                if (propertyConfig != null) {
                    if (propertyConfig.hasLocalArchive()) {
                        int archiveIndex = 0;
                        THistorySpecification tHistorySpecification = TineClassFactory.getInstance().getTHistorySpecification(propertyConfig);
                        if (propertyConfig.isArray()) {
                            log.info("Creating history records for array property " + propertyConfig.getName());
                            for (D device : deviceGroup) {
                                archiveIndex = historyMap.getIndexOf(
                                    this.getExportName(),
                                    device.getDeviceName(),
                                    propertyConfig.getName());
                                if (archiveIndex >= 0) {
                                    log.debug("Adding history record: " + device.getDeviceName() + "/" + property.getPropertyName()); 
                                    this.addLocalHistoryRecord(
                                        archiveIndex,
                                        device.getDeviceName(),
                                        property.getPropertyName(),
                                        propertyConfig.getArrayLength(),
                                        propertyConfig.getType().getTineFormatKey(),
                                        TArrayType.AT_UNKNOWN,
                                        tHistorySpecification);
                                }
                            }
                        } else {
                            log.info("Creating multichannel local history record for property " + property.getPropertyName());
                            archiveIndex = historyMap.getIndexOf(
                                this.getExportName(),
                                "",
                                propertyConfig.getName());
                            if (archiveIndex >= 0) {
                                log.debug("Adding history record: /" + property.getPropertyName()); 
                                // TODO DeviceName "#0" is not really a good idea!
                                this.addLocalHistoryRecord(
                                    archiveIndex,
                                    "#0",
                                    property.getPropertyName(),
                                    deviceGroup.size(),
                                    propertyConfig.getType().getTineFormatKey(),
                                    TArrayType.AT_CHANNEL,
                                    tHistorySpecification);
                            }
                        }
                    } else {
                        log.info("Local Archiving disabled for property " + property.getPropertyName());
                    }
                } else {
                    log.error("Property " + property.getPropertyName() + " not found in configuration!");
                }
            }
        } else {
            log.error("Cannot register local history: Properties not registered.");
        }
        
    }
    
    /**
     * Returns the device group.
     */
    public DeviceGroup<D> getDeviceGroup() {
        return deviceGroup;
    }
    
    /**
     * Updates all devices in the device group.
     */
    public void updateDevices() {
        
        log.trace("Updating equipment module " + this.getExportName());
     
        
        if (deviceGroup == null) {
            log.error("Cannot update devices. Device group not set.");
            return;
        }
        
        try {
        if (commonData != null) {
            try {
                commonData.update();
            } catch (Exception e) {
                log.error("Updating common data of " 
                    + this.getExportName(), e);
                e.printStackTrace();
            }
        }
        for (D currentDevice : deviceGroup) {
            try {
                currentDevice.update();
            } catch (Exception e) {
                log.error("Updating device " 
                    + currentDevice.getDeviceName() 
                    + " of " 
                    + this.getExportName(), e);
                e.printStackTrace();
            }
        }
        } catch (Exception e) {
            log.error("Cannot update devices: " + e, e);
            e.printStackTrace();
        }
    }
    
    /**
     * Calls the setup routine of all devoce in the device group.
     */
    public void setupDevices() {
 
        if (commonData != null) {
            try {
                commonData.setup();
            } catch (Exception e) {
                log.error("Setting up common data of " 
                    + this.getExportName(), e);
                e.printStackTrace();
            }
        }
        if (deviceGroup != null) {
            for (D currentDevice : deviceGroup) {
                try {
                    currentDevice.setup();
                } catch (Exception e) {
                    log.error("Setting up device " 
                        + currentDevice.getDeviceName() 
                        + " of " 
                        + this.getExportName(), e);
                    e.printStackTrace();
                }
            }
            if (deviceGroup.size() == 0) {
                log.warn("No devices in device group!");
            }
        } else {
            log.error("Cannot setup devices. No device group set.");
        }

    }
    
    /**
     * Transmits device alarms to TINE.
     */
    public void updateAlarms() {
        TDeviceList tineDevices = getDeviceList();
        TDevice tineDevice = null;
        
        if (deviceGroup == null) {
            log.error("Cannot update alarms. Device group not set.");
            return;
        }
        
        for (IBasicDevice device : deviceGroup) {
            tineDevice = tineDevices.getDevice(device.getDeviceName());
            if (tineDevice == null) {
                log.warn("updateAlarms(): TINE device not found: " + device.getDeviceName());
            } else {
                tineDevice.clearAlarms();
                for (Alarm alarm : device.getAlarms()) {
                    tineDevice.setAlarm(TineAlarmFactory.getTineCode(alarm));
                    log.info("Setting TINE alarm: " + alarm.getName() + " " + TineAlarmFactory.getTineCode(alarm));
                }
            }
        }
    }

    /**
     * Updates all buffered properties.
     */
    public void updatePropertyBuffers() {
        if (propertySet == null) {
            log.error("Cannot update property buffers: No property set");
            return;
        }
        try {
            propertySet.updateProperties();
        } catch(Exception e) {
            log.error("Property buffer update: " + e.toString(), e);
        }
    }
    
    /**
     * Assigns the common data.
     * Called once by AModuleFactory after creation of a device group.
     * 
     * @param commonData
     */
    public void setCommonData(ACommonData<D> commonData) {
        this.commonData = commonData;
    }

    /**
     * Returns the common data.
     */
    public ACommonData<D> getCommonData() {
        return this.commonData;
    }


    
}
