/*
 * Created on Sep 19, 2007
 *
  */
package de.desy.mst.auxiliary.devl.deviceserverlib.server;



/**
 * The common interface for equipment module builders.
 * 
 * An equipment module builder is used by the EquipmentModuleFactory.
 * The builder knows how to build parts of the equipment module
 * and is driven by the factory.
 * 
 * The purpose of the IEquipment module builder is to hide the 
 * single steps of building an equipment module, so that derived
 * classes can implement them in their special way, depending on
 * what kind of device group an equipment module is build.
 * 
 *  The basic steps are always the same. They are called by 
 *  the EquipmentModuleFactory in the following order:
 *  
 * <nl>
 * <li> initializeEquipmentModule(); </li>
 * <li> createDeviceContext(); </li>
 * <li> createDeviceGroup(); </li>
 * <li> createCommonData(); </li>
 * <li> registerDevices(); </li>
 * <li> registerProperties(); </li>
 * <li> registerLocalHistories(); </li>
 * <li> getEquipmentModule(); </li>
 * </nl>
 * 
 * The methods have to be called in the following order:
 * @author jwlg
 */
public interface IEquipmentModuleBuilder<T extends IBasicDevice> {
    
    /**
     * Creates and initializes the equipment module.
     * Creates an equipment module and a device context.
     */
    public void initializeEquipmentModule() throws EquipmentModuleBuilderException;
    
    /**
     * Creates a device context.
     * This method creates the device context which is
     * assigned to all devices.
     * The device context does not yet contain a device group
     * or common data.
     */
    public void createDeviceContext() throws EquipmentModuleBuilderException;
    
    /**
     * Creates and populates the device group.
     * Creates the device group, creates the devices,
     * assigns the group to the device context, 
     * assigns the context to the devices.
     */
    public void createDeviceGroup() throws EquipmentModuleBuilderException;
    
    /**
     * Creates the common data.
     * Creates the common data, assigns the common data
     * to the device context.
     */
    public void createCommonData() throws EquipmentModuleBuilderException;
    
    /**
     * Registers the devices.
     */
    public void registerDevices() throws EquipmentModuleBuilderException;
    
    /**
     * Registers the properties.
     */
    public void registerProperties() throws EquipmentModuleBuilderException;
    
    /**
     * Registers the alarms.
     */
    public void registerAlarms() throws EquipmentModuleBuilderException;
    
    /**
     * Registers the local histories.
     */
    public void registerLocalHistories() throws EquipmentModuleBuilderException;

    /**
     * Returns the Equipment module.
     * This method can be called after initializeEquipmentModule();
     */
    public EquipmentModule<T> getEquipmentModule();

}
