/*
 * Created on Jun 9, 2006
 *
 */
package de.desy.mst.auxiliary.devl.deviceserverlib.tools;

import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.util.LinkedList;
import java.util.List;
import java.util.Set;
import java.util.TreeSet;

import org.apache.log4j.Logger;

import de.desy.mst.auxiliary.devl.deviceserverlib.config.CfgOperationDataType;
import de.desy.mst.auxiliary.devl.deviceserverlib.config.CfgDeviceServer;
import de.desy.mst.auxiliary.devl.deviceserverlib.config.CfgDeviceType;
import de.desy.mst.auxiliary.devl.deviceserverlib.config.CfgOperation;
import de.desy.mst.auxiliary.devl.deviceserverlib.config.CfgProperty;
import de.desy.mst.auxiliary.devl.deviceserverlib.config.CfgPropertyType;
import de.desy.mst.auxiliary.devl.deviceserverlib.config.CfgTypeDefinition;
import de.desy.mst.auxiliary.devl.deviceserverlib.config.CfgTypeField;
import de.desy.mst.auxiliary.devl.deviceserverlib.config.CfgUserType;
import de.desy.tine.definitions.TFormat;


/**
 * A class which provides useful routines for device server models.
 * 
 * ModelUtils is designed as a singleton, not a utility class, because
 * it can be passed by reference like any other object. This feature is heavily 
 * used by velocity templates.
 * 
 * @author jwlg
 */
public class ModelUtils {

    private static ModelUtils instance = new ModelUtils();
    private static Logger log = Logger.getLogger(ModelUtils.class);
    LinkedList<CfgProperty> defaultAttributes = null;
    LinkedList<CfgProperty> standardControllerProperties = null;

    /**
     * Private constuctor.
     */
    private ModelUtils() {
        super();
    }

    /**
     * Returns the instance.
     */
    public static ModelUtils getInstance() {
        return instance;
    }


    /**
     * Returns the list of default attributes.
     * 
     * For every device type, an initial record is to be generated.
     * An initial record contains attributes, and there are
     * default attributes which all devices have. These
     * default attributes are created by this method.
     * 
     * Currently there are only two default attributes,
     * deviceName and deviceNumber.
     * @deprecated
     */
    public List<CfgProperty> getDefaultAttributes() {
        if (defaultAttributes == null) {
            //CfgProperty newAttrib = null;
            defaultAttributes = new LinkedList<CfgProperty>();
            /*
            newAttrib = new CfgProperty("deviceName", String.class.getSimpleName());
            newAttrib.setExported(false);
            defaultAttributes.add(newAttrib);
            newAttrib = new CfgProperty("deviceNumber", int.class.getSimpleName());
            newAttrib.setExported(false);
            defaultAttributes.add(newAttrib);
             */
        }
        return defaultAttributes;
    }

    /**
     * Returns default attributes and specific attributes.
     * @param deviceType A device type.
     * @deprecated
     */
    public List<CfgProperty> getAllAttributes(CfgDeviceType deviceType) {
        LinkedList<CfgProperty> result = new LinkedList<CfgProperty>();
        /*
        result.addAll(getDefaultAttributes());
        result.addAll(deviceType.getAttributeList());
         */
        return result;

    }

    /**
     * Returns a list of classes which are used as property types
     * or input/output types of operations.
     * 
     * In the returned set, all class names are contained only once.
     * Simple data types are not included.
     * This method is needed for generating import statements.
     * 
     * @param deviceType A device type.
     */
    public Set<String> getPropertyImports(CfgDeviceType deviceType) {

        
        TreeSet<String> result = new TreeSet<String>();
        
        if (deviceType == null) { 
            log.warn("NULL specified as device type");
            return result;
        }

        for (CfgProperty p : deviceType.getProperties()) {
            if (p.getType().isObject()) {
                result.add(p.getType().getFullClassName());
            }
        }
        
        for (CfgOperation o : deviceType.getOperations()) {

            CfgOperationDataType inputData = o.getInputData();
            CfgOperationDataType outputData = o.getOutputData();

            // FIXME Auch benutzerdefinierte Typen beruecksichtigen!
            // (Pfade des Projektes werden bentigt!)

            if (inputData.isProperty()) {
                if (inputData.getPropertyType().isObject()) {
                    result.add(inputData.getPropertyType().getFullClassName());
                }
            }

            if (outputData.isProperty()) {
                if (outputData.getPropertyType().isObject()) {
                    result.add(outputData.getPropertyType().getFullClassName());
                }
            }

        }

        return result;
    }

    /**
     * Returns a set of all classes that an operation
     * needs to import.
     * @param server The device server project.
     * @param operation The operation.
     */
    public Set<String> getOperationImports(CfgDeviceServer server, CfgOperation operation) {

        TreeSet<String> result = new TreeSet<String>();
        NameFactory nameFactory = NameFactory.getInstance();
        CfgOperationDataType inputData = operation.getInputData();
        CfgOperationDataType outputData = operation.getOutputData();

        if (inputData.isUserDefined()) {
            result.add(
                nameFactory.getUserTypesPackage(server) 
                + "." 
                + inputData.getUserType().getClassName()); 
        } else if (inputData.isProperty()) {
            if (inputData.getPropertyType().isObject()) {
                result.add(inputData.getPropertyType().getFullClassName());
            }
        } else {
            log.warn("Input data of " + operation.getName() + " is no property and not user-defined.");
        }

        if (outputData.isUserDefined()) {
            result.add(
                nameFactory.getUserTypesPackage(server) 
                + "." 
                + outputData.getUserType().getClassName()); 
        } else if (outputData.isProperty()) {
            if (outputData.getPropertyType().isObject()) {
                result.add(outputData.getPropertyType().getFullClassName());
            }
        } else {
            log.warn("Input data of " + operation.getName() + " is no property and not user-defined.");
        }

        return result;
    }

    /** 
     * Returns a unique set of all user-defined 
     * types which are used in operations. 
     * Only simple class names are listed.
     * 
     * @param deviceType A device type.
     */
    public Set<String> getOperationTypeList(CfgDeviceType deviceType) {
        TreeSet<String> result = new TreeSet<String>();
        for (CfgOperation operation : deviceType.getOperations()) {
            if (operation.getInputData().isUserDefined()) {
                result.add(operation.getInputData().getUserType().getClassName());
            } 
            if (operation.getOutputData().isUserDefined()) {
                result.add(operation.getOutputData().getUserType().getClassName());
            }
        }
        return result;
    }

    /**
     * Returns a unique set of user-defined data types which one
     * operation uses.
     * Only simple class names are listed.
     * @param operation An operation.
     */
    public Set<String> getTypeListOfOperation(CfgOperation operation) {
        TreeSet<String> result = new TreeSet<String>();
        if (operation.getInputData().isUserDefined()) {
            result.add(operation.getInputData().getUserType().getClassName());
        } 
        if (operation.getOutputData().isUserDefined()) {
            result.add(operation.getOutputData().getUserType().getClassName());
        }
        return result;
    }


    /**
     * Checks if a type name is used as an input or output type in an operation.
     * Before a type name can be removed, this method can be used to examine
     * if is used somewhere.
     * @param deviceServer A device server.
     * @param typeName A type name.
     * @return True if the type is in use.
     */
    public boolean isTypeUsedInOperations(CfgDeviceServer deviceServer, String typeName) {

        boolean isTypeUsed = false;
        if (typeName == null) return false;

        LinkedList<CfgDeviceType> deviceTypeList = deviceServer.getDefinition().getDeviceTypes();
        for (CfgDeviceType deviceType : deviceTypeList) {
            for (CfgOperation operation : deviceType.getOperations()) {
                if (operation.getInputData().isUserDefined()) {
                    isTypeUsed |= (typeName.equals(
                        operation.getInputData().getUserType().getClassName()));
                } 
                if (operation.getOutputData().isUserDefined()) {
                    isTypeUsed |= (typeName.equals(
                        operation.getOutputData().getUserType().getClassName()));
                }
                if (isTypeUsed) break;
            }
            if (isTypeUsed) break;
        }

        return isTypeUsed;

    }

    /**
     * Renames a type in all operations of a device server.
     * This method is used when a type should be renamed.
     * If the old type name is used as an input type or 
     * output type of any operation, that type name  
     * has to be changed as well.
     * @param deviceServer A device server.
     * @param oldName The old type name.
     * @param newName The new type name.
     */
    public void renameTypeInOperations(CfgDeviceServer deviceServer, String oldName, String newName) {

        if ((oldName == null) || (newName == null)) return;
        LinkedList<CfgDeviceType> deviceTypeList = deviceServer.getDefinition().getDeviceTypes();
        for (CfgDeviceType deviceType : deviceTypeList) {
            for (CfgOperation operation : deviceType.getOperations()) {
                if (operation.getInputData().isUserDefined() &&
                        oldName.equals(operation.getInputData().getUserType().getClassName())) {
                    operation.getInputData().getUserType().setClassName(newName);
                }
                if (operation.getOutputData().isUserDefined() &&
                        oldName.equals(operation.getOutputData().getUserType().getClassName())) {
                    operation.getOutputData().getUserType().setClassName(newName);
                }
            }
        }
    }


    /**
     * Calculates the size in bytes of an object
     * of the specfied property type.
     * Creates an instance of the specified object and 
     * calls getSizeInBytes().
     * @param type
     * @return Size in bytes or 0 if the size could not be determined.
     * @throws ClassNotFoundException
     * @throws InstantiationException
     * @throws IllegalAccessException
     * @throws SecurityException
     * @throws NoSuchMethodException
     * @throws IllegalArgumentException
     * @throws InvocationTargetException
     */
    @SuppressWarnings("unchecked")
    public int getObjectSizeInBytes(CfgPropertyType type) throws ClassNotFoundException, InstantiationException, IllegalAccessException, SecurityException, NoSuchMethodException, IllegalArgumentException, InvocationTargetException {
        Class typeClass = Class.forName(type.getFullClassName());
        Object obj = typeClass.newInstance();
        Method method = typeClass.getMethod("getSizeInBytes", (Class[]) null);
        Integer returnValue = (Integer)method.invoke(obj, (Object[]) null);
        return returnValue.intValue();
    }
    
    /**
     * Calculates the size of a user-defined type.
     * This method creates an instance of the specified type
     * and calls getSizeInBytes().
     * @param deviceServer The device server project.
     * @param userType The user-defined type.
     */
    @SuppressWarnings("unchecked")
    public int getUserTypeSizeInBytes(CfgDeviceServer deviceServer, CfgUserType userType) {
        int result = 0;
        try {
            String typePackage = NameFactory.getInstance().getUserTypesPackage(deviceServer);
            Class typeClass = Class.forName(typePackage + "." + userType.getClassName());
            Object obj = typeClass.newInstance();
            Method method = typeClass.getMethod("getSizeInBytes", (Class[]) null);
            Integer returnValue = (Integer)method.invoke(obj, (Object[]) null);
            result = returnValue.intValue();
        } catch (Exception e) {
            //log.error("Cannot calculate data size: " + e);
            result = 0;
        }
        return result;
    }

    /**
     * Returns the size in bytes as calculated by TINE.
     * @param tineFormat a TFormat key, e.g TFormat.INT.
     * @return The size in bytes or 0 if it could not be determined.
     */
    public int getTineSizeInBytes(int tineFormat) {
        int result = TFormat.formatSizeOf(TFormat.CF_NAME32);
        if (result < 0) result = 0;
        return result;
    }

    /**
     * Returns the size in bytes as calculated by TINE.
     * @param type A property type.
     * @return The size in bytes or 0 if it could not be determined.
     */
    public int getTineSizeInBytes(CfgPropertyType type) {
        return getTineSizeInBytes(type.getTineFormatKey());
    }

    /**
     * Calculates the size in bytes of an instance of a property type.
     * @param type
     * @return Size in bytes or 0 if hte size could not be calculated.
     */
    public int getTypeSizeInBytes(CfgPropertyType type) {
        int result = 0;
        if (type.isVoid()) {
            result = 0;
        } else if (type.isString()) {
            result = TFormat.formatSizeOf(TFormat.CF_NAME32);
        } else if  (type.isSimple()) {
            result = TFormat.formatSizeOf((short)type.getTineFormatKey());
        } else {
            result = 0;
            try {
                result = getObjectSizeInBytes(type);
            } catch (Exception e) {
                result = 0;
            } 
            if (result <= 0) {
                result = getTineSizeInBytes(type);
            }
        }  
        return result;
    }
    
    /**
     * Calculates the size in bytes of a User-defined type.
     * If a field is a string, the array length is interpreted
     * as the string length, and the string length is interpreted
     * as a number of bytes (not chars). TINE does it the same way.
     * @param typeDef The definition of a user-defined type.
     * @return Size in bytes or 0.
     */
    public int getStructSizeInBytes(CfgTypeDefinition typeDef) {
        int result = 0;
        for (CfgTypeField field : typeDef.getFields()) {
            if (field.getType().isString()) {
                result += field.getArraySize() * TFormat.formatSizeOf(TFormat.CF_BYTE);
            } else {
                result += field.getArraySize() * getTypeSizeInBytes(field.getType());
            }
        }
        return result;
    }

    /**
     * Returns a unique set of all classes which need 
     * to be imported in the fields of a CfgTypeDefinition.
     * @param def The type definition.
     */
    public Set<String> getTypeImports(CfgTypeDefinition def) {
        TreeSet<String> result = new TreeSet<String>();
        for (CfgTypeField field : def.getFields()) {
            if (field.getType().isObject()) {
                result.add(field.getType().getFullClassName());
            }
        }
        return result;
    }

    
    public List<CfgProperty> getStandardControllerProperties() {
        if (standardControllerProperties == null) {
            standardControllerProperties = new LinkedList<CfgProperty>();

            CfgProperty standardPropertyNMembers = new CfgProperty("NMembers", CfgPropertyType.INT);
            standardPropertyNMembers.setReadable(true);
            standardPropertyNMembers.setWriteable(false);
            standardPropertyNMembers.setBuffered(false);

            CfgProperty standardPropertyMembers = new CfgProperty("Members", CfgPropertyType.STRING);
            standardPropertyMembers.setArray(true);
            standardPropertyMembers.setArrayLength(1024);
            standardPropertyMembers.setReadable(true);
            standardPropertyMembers.setWriteable(false);
            standardPropertyMembers.setBuffered(false);

            standardControllerProperties.add(standardPropertyNMembers);
            standardControllerProperties.add(standardPropertyMembers);
        }
        return standardControllerProperties;

 
    }
    
    public CfgProperty findStandardControllerProperty(String propertyName) {
        CfgProperty result = null;
        
        if (propertyName == null) return null;
        
        for (CfgProperty property : getStandardControllerProperties()) {
            if (propertyName.equals(property.getName())) {
                result = property;
                break;
            }
        }
        return result;
    }

}


