/*
 * Created on Mar 8, 2006
 *
  */
package de.desy.mst.auxiliary.devl.deviceserverlib.tools;

/**
 * Name Utilities.
 * 
 * Utilities to check and create names.
 * This class mainly contains validity checks which are used
 * by the device server wizard and the generated classes.
 * 
 * NameUtils is designed as a singleton, not a utility class, because
 * it can be passed by reference like any other object. This feature is heavily 
 * used by velocity templates.
 * 
 * @author jwlg
 */
public class NameUtils {
    
    public static final int MAX_PROPERTY_LENGTH = 32;
    public static final int MAX_MODULE_LENGTH = 16;
    public static final int MAX_ALARM_NAME_LENGTH = 16;
    
    private static NameUtils theInstance = new NameUtils();
    
    /**
     * Private constructor.
     * This is a singleton.
     */
    private NameUtils() {
        super();
    }
    
    /**
     * Returns the instance.
     */
    public static NameUtils getInstance() {
        return theInstance;
    }

    /**
     * Returns true if a specified name is a valid Java Class name. 
     * @param name The name to be examined.
     */
    public boolean isValidClassName(String name) {
        return name.matches("[A-Z]\\w*");
    }

    
    /**
     * Returns true if a specified name is a valid alarm name. 
     * @param name The name to be examined.
     */
    public boolean isValidAlarmName(String name) {
        return name.matches("[A-Z]\\w*");
    }
    
    /**
     * Returns true if a specified name is a valid module name. 
     * @param name The name to be examined.
     */
    public boolean isValidModuleName(String name) {
        return name.matches("[A-Za-z]\\w*");
    }
    
    /**
     * Returns true if a specified name is a valid constant name. 
     * @param name The name to be examined.
     */
    public boolean isValidConstantName(String name) {
        return name.matches("[A-Z][A-Z0-9_]*");
    }
    
    /**
     * Returns true if a specified name is a valid property name. 
     * @param name The name to be examined.
     */
    public boolean isValidPropertyName(String name) {
        return name.matches("[A-Za-z]\\w*");
    }
    
    /**
     * Returns true if a specified name is a valid attribute name. 
     * @param name The name to be examined.
     */
    public boolean isValidAttributeName(String name) {
        return isValidPropertyName(name);
    }
    
    /**
     * Returns true if the size of the property name
     * is permitted of TINE.
     * @param name The name to be examined.
     */
    public boolean isPropertyNameSizeOK(String name) {
        return ((name.length() > 0 ) && (name.length() <= MAX_PROPERTY_LENGTH)); 
    }

    /**
     * Returns true if the size of the alarm name
     * is permitted of TINE.
     * @param name The name to be examined.
     */
     public boolean isAlarmNameSizeOK(String name) {
        return ((name.length() > 0 ) && (name.length() <= MAX_ALARM_NAME_LENGTH)); 
    }

     /**
      * Returns true if the size of the attribute name
      * is permitted of TINE.
      * @param name The name to be examined.
      */
    public boolean isAttributeNameSizeOK(String name) {
        return isPropertyNameSizeOK(name); 
    }
    
    /**
     * Returns true if the size of the module name
     * is permitted of TINE.
     * @param name The name to be examined.
     */
    public boolean isModuleNameSizeOK(String name) {
        return ((name.length() > 0 ) && (name.length() <= MAX_MODULE_LENGTH)); 
    }
    
    /**
     * Returns true if a specified name is a valid package name. 
     * @param name The name to be examined.
     */
    public boolean isValidPackageName(String name) {
        return name.matches("[a-z]+[A-Za-z0-9_.]*\\w");
    }
    
    /**
     * Returns true if a specified name is a valid name for a package section. 
     * @param name The name to be examined.
     */
    public boolean isValidPackageSectionName(String name) {
        return name.matches("^[a-z]+[a-z0-9_]*\\w$");
    }
    
    /**
     * Changes a name into a valid class name.
     * Turns the first letter to uppercase and
     * removes all whitespaces.
     * @param name The class name.
     */
    public String toValidClassName(String name) {
        String result = name;
        char firstChar = name.charAt(0);
        if (Character.isLowerCase(firstChar)) {
            result = Character.toUpperCase(firstChar) + name.substring(1);
        }
        result = result.replaceAll("\\W","");
        return result;
    }

    /** 
     * Creates a TINE propery name.
     * Ensures that a name is short enough. Cuts off
     * characters which exceed the maximum size.
     * @param propertyName A property name.
     */
    public String generateTineProperty(String propertyName) {
        String result;
        result = new String(propertyName);
        if (result.length() > MAX_PROPERTY_LENGTH) {
            result = result.substring(0, MAX_PROPERTY_LENGTH -1);
        }
        return result;
    }

}
