/*
 * Created on Aug 29, 2007
 *
  */
package de.desy.mst.auxiliary.devl.deviceserverlib.tools.array;


public class LongArrayUtils {
    
    /**
     * Private Constructor.
     * This is a utility class. No instances can be made.
     */
    private LongArrayUtils() {
    }
    
    /**
     * Rotates an array one position right.
     * The element at the last index is 
     * copied to the first index.
     * @param array  The array.
     */
    public static void rotateRight(long[] array) {
        prepend(array, array[array.length]);
    }
    
    /**
     * Rotates an array one position left.
     * The element at the first index is 
     * copied to the last index.
     * @param array The array.
     */
    public static void rotateLeft(long[] array) {
        append(array, array[0]);
    }
    
    /**
     * Duplicates an array.
     * @param source Source array.
     * @return Copy of the array.
     */
    public static long[] clone(long[] source) {

        long[] result = new long[source.length];
        copy(source, result);
        return result;
    }
    
    /**
     * Copies a slice of one array into a slice of another array.
     * @param source Source array.
     * @param sourceOffset Offset of first element to be copied.
     * @param nElements Number of elements to be copied.
     * @param dest Destination array.
     * @param destOffset Index where the first copied element goes.
     * @throws ArrayIndexOutOfBoundsException
     * @return Number of copied elements.
     * Caution: source and dest must not be the same array.
     */
    public static int copySlice(long[] source, 
            int sourceOffset,
            int nElements, 
            long[] dest, 
            int destOffset) {

            int effectiveLength = nElements;
            effectiveLength = Math.min(effectiveLength, source.length
                - sourceOffset);
            effectiveLength = Math.min(effectiveLength, dest.length
                - destOffset);

            for (int i = 0; i < effectiveLength; i++)
                dest[destOffset + i] = source[sourceOffset + i];

            return effectiveLength;
    }
    
    /**
     * Copies values from a source array to a destination array.
     * The algorithm copies the element at the source offset to
     * the place at the destination offset, then increments both
     * and copies the next element. If the last element in an array
     * is exceeded, the next element is the first element.
     * 
     * @param source Source array.
     * @param sourceOffset Index of the first element to be copied.
     * @param nElements Number of elements to be copied.
     * @param dest Destination array.
     * @param destOffset Index where the first element (sourceOffset) is copied.
     * @return Number of copied elements.
     */
    public static int copyRotate(long[] source, int sourceOffset, int nElements, long[] dest, int destOffset) {
        int copiedElements = Math.min(nElements, source.length);
        int sourceIndex = sourceOffset;
        int destIndex = destOffset;
        for (int i = 0; i< copiedElements; i++) {
            dest[destIndex] = source[sourceIndex];
            sourceIndex++;
            if (sourceIndex >= source.length) {
                sourceIndex = 0;
            }
            destIndex++;
            if (destIndex >= dest.length) {
                destIndex = 0;
            }
        }
        return copiedElements;
    }
    
    /**
     * Copies and rotates an array.
     * @param source The source array.
     * @param sourceOffset Index of the first element to be copied.
     * @param dest The destination array.
     * @throws ArrayIndexOutOfBoundsException;
     */
    public static int copyRotate(long[] source, int sourceOffset,
            long[] dest) {
        return copyRotate(source, sourceOffset, source.length, dest, 0);
    }
    
    /**
     * Copies one array into another.
     * @param source The source array.
     * @param dest The destination array.
     * @return Number of copied elements.
     */
    public static int copy(long[] source, long[] dest) {

        int minLength = Math.min(source.length, dest.length);
        for (int i = 0; i < minLength; i++) {
            dest[i] = source[i];
        }
        return minLength;
    }

    /**
     * Prepends a value to an array.
     * Rotates an array to the right, deletes the last element
     * and adds the value at the first index.
     * @param array The array.
     * @param value The new value on position 0.
     */
    public static void prepend(long[] array, long value) {

        int imax = array.length - 1;
        for (int i = imax; i > 0; i--) {
            array[i] = array[i - 1];
        }
        array[0] = value;
    }
    
    /**
     * Appends a value to an array.
     * Rotates an array to the left, deletes the first element
     * and adds the value at the last index.
     * @param array The array. 
     * @param value The value to be appended.
     */
    public static void append(long[] array, long value) {

        int imax = array.length - 1;
        for (int i = 0; i < imax; i++) {
            array[i] = array[i + 1];
        }
        array[imax] = value;
    }

}
