package de.desy.tine.accesslayer;

import java.util.Date;
import java.util.HashMap;
import java.util.Map;
import java.util.concurrent.CancellationException;
import java.util.concurrent.Delayed;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.Future;
import java.util.concurrent.RejectedExecutionException;
import java.util.concurrent.RejectedExecutionHandler;
import java.util.concurrent.RunnableScheduledFuture;
import java.util.concurrent.ScheduledExecutorService;
import java.util.concurrent.ScheduledThreadPoolExecutor;
import java.util.concurrent.ThreadPoolExecutor;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.TimeoutException;
import java.util.concurrent.atomic.AtomicBoolean;

import de.desy.tine.dataUtils.TDataType;
import de.desy.tine.definitions.TErrorList;
import de.desy.tine.definitions.TFormat;
import de.desy.tine.queryUtils.TPropertyQuery;
import de.desy.tine.queryUtils.TQuery;

/**
 *
 * <code>ChannelFactory</code> provides access to tine channels, by using a common tine layer for
 * all requests that are targeted at the same address.
 *
 * @author <a href="mailto:jaka.bobnar@cosylab.com">Jaka Bobnar</a>
 *
 */
public final class ChannelFactory {

    public static void main(String[] args) throws ConnectionException, Exception {
        ChannelFactory factory = ChannelFactory.getInstance();
        //String address = "/TEST/WinSineServer/SineGen0/Amplitude";
        String address = "/TEST/SineServer/SineGen0/Amplitude";
        //make a channel that periodically receives new values, using default data type (e.g. double)
        Object data = ChannelFactory.getInstance().getValueSync(address);
        System.out.println(data);
        
        String address2 = "/TEST/SineServer/SineGen1/Amplitude";
        Object data2 = ChannelFactory.getInstance().getValueSync(address2);
        System.out.println(data2);
                        
        @SuppressWarnings("unused")
        Channel channel = factory.getChannel(address,TFormat.valueOf(TFormat.CF_DOUBLE), 1,
                ConnectionMode.SYNC_SINGLE,1000,new ChannelCallbackAdapter() {

            @Override
            public void updateValue(Channel channel) {
                System.out.println("Update received: " + new Date((long)channel.getRawValue().getTimeStamp() * 1000) + " " + channel.getValue());
            }
            @Override
            public void channelError(String message, Throwable cause) {
                cause.printStackTrace();
            }

            @Override
            public void updateValue(ChannelReadEvent event) {
//                System.out.println(event.getStatusString());
            }

            @Override
            public void updateError(ChannelReadEvent event) {
//                System.out.println(event.getStatusString());
            }
        });


//        Thread.sleep(30000000L);
//        channel.stop(); 
        factory.close();
    }

    static int WAIT_TIME = 1000; // 2000; < P.D. try this for a while ...
    static int DEFAULT_SIZE = 128;
    static int DEFAULT_USTRING_SIZE = 2048; // P.D. try to cut to the chase with worst case XFEL num devices

    /** The time for how long single read links are kept alive after no longer needed. */
    public static final String LINK_KEEP_ALIVE_MSEC = "tine.layer.keepAlive";
    /** The maximum time how long an links are opened after no longer needed. */
    public static final String MAX_KEEP_ALIVE_MSEC = "tine.layer.maxKeepAlive";

    /** The time in milliseconds how long single links are kept alive, before being terminated if noone uses them */
    private static final long DEFAULT_KEEP_ALIVE_SINGLE_LINKS = Long.parseLong(
            System.getProperty(LINK_KEEP_ALIVE_MSEC,"30000"));
    /** The maximum alive time: anything more than 5 minutes does not really need to be monitored */
    private static final long MAX_KEEP_ALIVE_SINGLE_LINKS = Long.parseLong(System.getProperty(
            MAX_KEEP_ALIVE_MSEC, String.valueOf(10*DEFAULT_KEEP_ALIVE_SINGLE_LINKS)));

    private static class ChannelFuture<V> implements RunnableScheduledFuture<V> {

        final Runnable task;
        final RunnableScheduledFuture<V> future;
        ChannelFuture(RunnableScheduledFuture<V> future, Runnable task) {
            this.task = task;
            this.future = future;
        }
        @Override
        public void run() {
            future.run();
        }
        @Override
        public boolean cancel(boolean mayInterruptIfRunning) {
            return future.cancel(mayInterruptIfRunning);
        }
        @Override
        public boolean isCancelled() {
            return future.isCancelled();
        }
        @Override
        public boolean isDone() {
            return future.isDone();
        }
        @Override
        public V get() throws InterruptedException, ExecutionException {
            return future.get();
        }
        @Override
        public V get(long timeout, TimeUnit unit) throws InterruptedException, ExecutionException, TimeoutException {
            return future.get(timeout,unit);
        }
        @Override
        public long getDelay(TimeUnit unit) {
            return future.getDelay(unit);
        }
        @Override
        public int compareTo(Delayed o) {
            return future.compareTo(o);
        }
        @Override
        public boolean isPeriodic() {
            return future.isPeriodic();
        }
    }

    private static class Counter {
        long lastRequestedTime = System.currentTimeMillis();
        long currentKeepAliveTime = DEFAULT_KEEP_ALIVE_SINGLE_LINKS;

        void usedNow(boolean changeKeepAlive) {
            long now = System.currentTimeMillis();
            if (changeKeepAlive) {
                long time = now - lastRequestedTime;
                if (time > currentKeepAliveTime) {
                    currentKeepAliveTime = Math.min(1000L*(long)(time*0.0015),MAX_KEEP_ALIVE_SINGLE_LINKS);
                }
            }
            lastRequestedTime = now;
        }
    }

    private static ChannelFactory INSTANCE;

    /**
     * @return the singleton instance of the factory
     */
    public static synchronized ChannelFactory getInstance() {
        if (INSTANCE == null) {
            INSTANCE = new ChannelFactory();
        }
        return INSTANCE;
    }

    private Map<String,Counter> loadCache;
    private Map<String,MasterChannel> channelCache;
    private ScheduledExecutorService executorService;
    private ExecutorService callbackExecutor;
//    private ExecutorService toTineExecutor;

    private ChannelFactory() {
        channelCache = new HashMap<>();
        loadCache = new HashMap<>();
        Thread th = new Thread(new Runnable(){
            public void run() {
                close();
            }
        },"ChannelFactory-cleanup");
        th.setDaemon(false);
        Runtime.getRuntime().addShutdownHook(th);
    }

    /**
     * Returns the property query for the given property. If the query is overloaded, only the first instance is returned.
     *
     * @param device the full device name
     * @param property the property name
     * @return the property query
     * @throws ConnectionException if the query loading failed or it does not exist
     */
    TPropertyQuery getPropertyQuery(String address) throws ConnectionException {
        int idx = address.lastIndexOf('/');
        String property = address.substring(idx+1);
        address = address.substring(0,idx);
        String[] devs = address.charAt(0) == '/' ? address.substring(1).split("/") : address.split("/");
        idx = address.indexOf(devs[1]);
        String dd = address.substring(address.indexOf('/',idx)+1);
        TPropertyQuery[] tpq = TQuery.getPropertyInformation(devs[0],devs[1],dd,property);
        if (tpq == null) {
            throw new ConnectionException("Cannot read property information.");
        }
        return tpq[0];
    }

    /**
     * Synchronously sets the value to the given property. Upon success method does not return anything. If the
     * write fails an exception si thrown, describing the error.
     *
     * @param address the proeprty to write to
     * @param value the value to write
     * @throws ConnectionException if write failed for any reason
     */
    public void setValue(String address, Object value) throws ConnectionException {
        setValue(address, value, null, WAIT_TIME);
    }

    /**
     * Synchronously sets the value to the given property. Upon success method does not return anything, but
     * it does read the value and applies it to the given outValue parameter (the action is atomic).  If the
     * write fails an exception is thrown, describing the error.
     *
     * @param address the proeprty to write to
     * @param value the value to write
     * @param outValue the data type that will receive the readback value after write is completed
     * @param timeout the timeout in milliseconds
     * @throws ConnectionException if write failed for any reason
     */
    public void setValue(String address, Object value, TDataType outValue, int timeout) throws ConnectionException {
        MasterChannel holder;
        synchronized(channelCache)  {
            String key = getChannelKey(address,TFormat.CF_DEFAULT,DEFAULT_SIZE,null);
            holder = channelCache.get(key);
            if (holder == null) {
                holder = new MasterChannel(address,null,-1,null,DEFAULT_KEEP_ALIVE_SINGLE_LINKS);
            }
        }
        holder.setValue(value, outValue, timeout);
    }

    /**
     * Asynchronously sets the value to the given property. Upon completed write the callback is notified, either through
     * {@link ChannelCallback#writeCompleted(Object)} or {@link ChannelCallback#writeFailed(Object, int)}.
     *
     * @param address the property to connect to
     * @param value the value to set
     * @param callback the callback to be notified
     * @throws ConnectionException if establishing the connection failed
     */
    public void setValueAsync(String address, Object value, ChannelCallback callback) throws ConnectionException {
       setValueAsync(address,value,null,WAIT_TIME,callback);
    }

    /**
     * Asynchronously sets the value to the given property. Upon completed write the callback is notified, either through
     * {@link ChannelCallback#writeCompleted(Object)} or {@link ChannelCallback#writeFailed(Object, int)}. The outValue also
     * receives the readback value.
     *
     * @param address the property to connect to
     * @param value the value to set
     * @param outValue the data type that will receive the readback value after write is completed
     * @param callback the callback to be notified
     * @param timeout the timeout in milliseconds
     * @throws ConnectionException if establishing the connection failed
     */
    public void setValueAsync(String address, Object value, TDataType outValue, int timeout, ChannelCallback callback) throws ConnectionException {
        MasterChannel holder;
        synchronized(channelCache)  {
            String key = getChannelKey(address,TFormat.CF_DEFAULT,DEFAULT_SIZE,null);
            holder = channelCache.get(key);
            if (holder == null) {
                holder = new MasterChannel(address,null,-1,null,DEFAULT_KEEP_ALIVE_SINGLE_LINKS);
            }
        }
        holder.setValue(value,outValue,timeout,callback);
    }

    /**
     * Synchronously sets the value to the given property. Upon success method does not return anything. If the
     * write fails an exception si thrown, describing the error.
     *
     * @param address the proeprty to write to
     * @param value the value to write
     * @throws ConnectionException if write failed for any reason
     */
    public void setRawValue(String address, TDataType value) throws ConnectionException {
        setRawValue(address,value,null,WAIT_TIME);
    }

    /**
     * Synchronously sets the value to the given property. Upon success method does not return anything; however the
     * outValue does receive the readback value. If the write fails an exception si thrown, describing the error.
     *
     * @param address the proeprty to write to
     * @param value the value to write
     * @param outValue the data type that will receive the readback value after write is completed
     * @param timeout the timeout in milliseconds
     * @throws ConnectionException if write failed for any reason
     */
    public void setRawValue(String address, TDataType value, TDataType outValue, int timeout) throws ConnectionException {
        MasterChannel holder;
        synchronized(channelCache)  {
            String key = getChannelKey(address,TFormat.CF_DEFAULT,DEFAULT_SIZE,null);
            holder = channelCache.get(key);
            if (holder == null) {
                holder = new MasterChannel(address,null,-1,null,DEFAULT_KEEP_ALIVE_SINGLE_LINKS);
            }
        }
        holder.setValue(value,outValue,timeout);
    }

    /**
     * Asynchronously sets the value to the given property. Upon completed write the callback is notified, either through
     * {@link ChannelCallback#writeCompleted(Object)} or {@link ChannelCallback#writeFailed(Object, int)}.
     *
     * @param address the property to connect to
     * @param value the value to set
     * @param callback the callback to be notified
     * @throws ConnectionException if establishing the connection failed
     */
    public void setRawValueAsync(String address, TDataType value, ChannelCallback callback) throws ConnectionException {
        setRawValueAsync(address,value,null,WAIT_TIME,callback);
    }

    /**
     * Asynchronously sets the value to the given property. Upon completed write the callback is notified, either through
     * {@link ChannelCallback#writeCompleted(Object)} or {@link ChannelCallback#writeFailed(Object, int)}. The readback value
     * is also applied to the given outValue parameter.
     *
     * @param address the property to connect to
     * @param value the value to set
     * @param outValue the data type that will receive the readback value after write is completed
     * @param timeout the timeout in milliseconds
     * @param callback the callback to be notified
     * @throws ConnectionException if establishing the connection failed
     */
    public void setRawValueAsync(String address, TDataType value, TDataType outValue, int timeout, ChannelCallback callback) throws ConnectionException {
        MasterChannel holder;
        synchronized(channelCache)  {
            String key = getChannelKey(address,TFormat.CF_DEFAULT,DEFAULT_SIZE,null);
            holder = channelCache.get(key);
            if (holder == null) {
                holder = new MasterChannel(address,null,-1,null,DEFAULT_KEEP_ALIVE_SINGLE_LINKS);
            }
        }
        holder.setValue(value,outValue,timeout,callback);
    }

    private boolean useSynchronousGetValue = false;
    public void setUseSynchronousGetValue(boolean value)
    {
    	useSynchronousGetValue = value;
    }
    /**
     * Makes a single synchronous connection to the address specified by parameters and returns the value.
     * If the value read failed a ConnectionException is thrown.
     *
     * @param address property name to connect to
     * @return the value of the property
     * @throws ConnectionException if connection failed
     */
    public Object getValue(String address) throws ConnectionException {
        return getValue(address,null,null,-1,null,useSynchronousGetValue,WAIT_TIME);
    }

    @Deprecated
    public Object getValue(String address, TDataType dataOut, Object dataIn) throws ConnectionException {
        return getValue(address,dataOut,dataIn,1000);
    }

    /**
     * Makes a single synchronous connection to the address specified by parameters and returns the value.
     * If the value read failed a ConnectionException is thrown.
     *
     * @param address property name to connect to
     * @param dataout the object to received the output data
     * @param dataIn input data of the read request
     * @param timeout the timeout in milliseconds after which the call will return terminated
     * @return the value of the property
     * @throws ConnectionException if connection failed
     */
    public Object getValue(String address, TDataType dataOut, Object dataIn, int timeout) throws ConnectionException {
        return getValue(address,dataOut,null,-1,dataIn,useSynchronousGetValue,timeout);
    }
    
    /**
     * Makes a true single synchronous connection to the address specified by parameters and returns the value.
     * If the value read failed a ConnectionException is thrown.
     *
     * @param address property name to connect to
     * @return the value of the property
     * @throws ConnectionException if connection failed
     */
    public Object getValueSync(String address) throws ConnectionException {
        return getValue(address,null,null,-1,null,true,WAIT_TIME);
    }
    
    /**
     * Makes a true single synchronous connection to the address specified by parameters and returns the value.
     * If the value read failed a ConnectionException is thrown.
     *
     * @param address property name to connect to
     * @param dataout the object to received the output data
     * @param dataIn input data of the read request
     * @param timeout the timeout in milliseconds after which the call will return terminated
     * @return the value of the property
     * @throws ConnectionException if connection failed
     */
    public Object getValueSync(String address, TDataType dataOut, Object dataIn, int timeout) throws ConnectionException {
        return getValue(address,dataOut,null,-1,dataIn,true,timeout);
    }

    /**
     * Makes a single synchronous connection to the address specified by parameters and returns the value.
     * If the value read failed a ConnectionException is thrown.
     *
     * @param address property name to connect to
     * @param dataout the object to received the output data
     * @param outFormat requested data out format
     * @param outSize requested data out size
     * @param dataIn input data of the read request
     * @param timeout the timeout in milliseconds
     * @param synchronousCall true if a synchronous TINE link is used or false if asnyc link is 
     *          used (blocking this thread)
     * @return the value of the property
     * @throws ConnectionException if connection failed
     */
    private Object getValue(String address, TDataType dataOut, TFormat outFormat, int outSize,
            Object dataIn, boolean synchronousCall, int timeout) throws ConnectionException {

        final Object mutex = new Object();
        Channel channel;
        if (dataOut != null) {
            outFormat = TFormat.valueOf(dataOut.getFormat());
            outSize = dataOut.getArrayLength();
        }
        if (outFormat == null) {
            outFormat = TFormat.valueOf(TFormat.CF_DEFAULT);
        }
        if (outSize <= 0) {
            outSize = DEFAULT_SIZE;
        }
        final AtomicBoolean wait = new AtomicBoolean(true);
        synchronized (mutex) {
            if (timeout < 1000) timeout = 1000;
            channel = getChannel(address,outFormat,outSize,dataOut, dataIn,
                    synchronousCall ? ConnectionMode.SYNC_SINGLE : ConnectionMode.SINGLE,
                            timeout,new ChannelCallbackAdapter() {
                @Override
                public void updateValue(Channel channel) {
                    wait.set(false);
                    synchronized (mutex) {
                        mutex.notifyAll();
                    }
                }
                @Override
                public void updateErrorStatus(Channel channel) {
                    wait.set(false);
                    synchronized (mutex) {
                        mutex.notifyAll();
                    }
                }
                @Override
                public void channelError(String message, Throwable cause) {
                    wait.set(false);
                    synchronized (mutex) {
                        mutex.notifyAll();
                    }
                }
            });
            if (channel != null) {
                if (!TErrorList.hasData(channel.getStatus())) {
                    long newtimeout = Math.max(30000,3*timeout); //half a minute minimum
                    try {
                        long time = System.currentTimeMillis();
                        while(wait.get() && (System.currentTimeMillis() - time < newtimeout)) {
                            mutex.wait(10);
                            if (TErrorList.hasData(channel.getStatus())) break;
                            if (!TErrorList.isTimeout(channel.getStatus())) break;
                        }
                    } catch (InterruptedException e) {
                        //ignore
                    }
                }
            } else {
                throw new ConnectionException("Could not establish connection.");
            }
        }
        if (TErrorList.hasData(channel.getStatus())) {
            return channel.getValue();
        } else {
            if (wait.get()) {
                throw new ConnectionException("Could not read the value: " + TErrorList.toString(TErrorList.connection_timeout),TErrorList.connection_timeout);
            } else {
                throw new ConnectionException("Could not read the value: " + channel.getStatusString(), channel.getStatus());
            }
        }
    }

    /**
     * Makes an asynchronous connection to the tine property specified by the device and property. When the value
     * is received it is forwarded to the given callback. If the connection could not be established an exception is
     * thrown.
     *
     * @param address the property name
     * @param callback callback to receive notifications
     * @throws ConnectionException if there was an error establishing the connection
     */
    public void getValueAsync(String address, ChannelCallback callback) throws ConnectionException {
        getChannel(address,null,-1,null,ConnectionMode.SINGLE,1000,callback);
    }

    /**
     * Makes an asynchronous connection to the tine property specified by the device and property. When the value
     * is received it is forwarded to the given callback. If the connection could not be established an exception is
     * thrown.
     *
     * @param address the full property name
     * @param dataIn the input object of the read request
     * @param callback callback to receive notifications
     * @throws ConnectionException if there was an error establishing the connection
     */
    public void getValueAsync(String address, ChannelCallback callback, Object dataIn) throws ConnectionException {
        getChannel(address,null,-1,dataIn,ConnectionMode.SINGLE,1000,callback);
    }

    /**
     * Makes a single synchronous connection to the address specified by parameters and returns the value.
     * If the value read failed a ConnectionException is thrown.
     *
     * @param device device name to connect to
     * @param property property name to connect to
     * @return the value of the property
     * @throws ConnectionException if connection failed
     */
    public TDataType getRawValue(String address) throws ConnectionException {
        return getRawValue(address,null,-1,null,false);
    }

    /**
     * Makes a single synchronous connection to the address specified by parameters and returns the value.
     * If the value read failed a ConnectionException is thrown.
     *
     * @param address property name to connect to
     * @param outFormat requested data out format
     * @param outSize requested data out size
     * @param dataIn input data of the read request
     * @param mode the connection mode used
     * @return the value of the property
     * @throws ConnectionException if connection failed
     */
    public TDataType getRawValue(String address, TFormat outFormat, int outSize,
            Object dataIn) throws ConnectionException {
        return getRawValue(address, outFormat, outSize, dataIn, false);
    }
    
    /**
     * Makes a true single synchronous connection (TINE synchronous link) to the address specified by parameters 
     * and returns the value. If the value read failed a ConnectionException is thrown.
     *
     * @param device device name to connect to
     * @param property property name to connect to
     * @return the value of the property
     * @throws ConnectionException if connection failed
     */
    public TDataType getRawValueSync(String address) throws ConnectionException {
        return getRawValue(address,null,-1,null,true);
    }

    /**
     * Makes a true single synchronous connection (TINE synchronous link) to the address specified by parameters 
     * and returns the value. If the value read failed a ConnectionException is thrown.
     *
     * @param address property name to connect to
     * @param outFormat requested data out format
     * @param outSize requested data out size
     * @param dataIn input data of the read request
     * @param mode the connection mode used
     * @return the value of the property
     * @throws ConnectionException if connection failed
     */
    public TDataType getRawValueSync(String address, TFormat outFormat, int outSize,
            Object dataIn) throws ConnectionException {
        return getRawValue(address, outFormat, outSize, dataIn, true);
    }
    
    /**
     * Makes a single synchronous connection to the address specified by parameters and returns the value.
     * If the value read failed a ConnectionException is thrown.
     *
     * @param address property name to connect to
     * @param outFormat requested data out format
     * @param outSize requested data out size
     * @param dataIn input data of the read request
     * @param mode the connection mode used
     * @param synchronousCall true to use a synchronous TINE link or false to use asynchronous link but 
     *          still blocking this thread
     * @return the value of the property
     * @throws ConnectionException if connection failed
     */
    public TDataType getRawValue(String address, TFormat outFormat, int outSize,
            Object dataIn, boolean synchronousCall) throws ConnectionException {
        final Object mutex = new Object();
        Channel channel;
        int timeout = WAIT_TIME;
        final boolean[] wait = new boolean[]{true};
        synchronized (mutex) {
            channel = getChannel(address,outFormat,outSize,dataIn,
                    synchronousCall ? ConnectionMode.SYNC_SINGLE : ConnectionMode.SINGLE,
                            timeout,new ChannelCallbackAdapter() {
                @Override
                public void updateValue(Channel channel) {
                    wait[0] = false;
                    synchronized (mutex) {
                        mutex.notifyAll();
                    }
                }
                @Override
                public void updateErrorStatus(Channel channel) {
                    wait[0] = false;
                    synchronized (mutex) {
                        mutex.notifyAll();
                    }
                }
                @Override
                public void channelError(String message, Throwable cause) {
                    wait[0] = false;
                    synchronized (mutex) {
                        mutex.notifyAll();
                    }
                }
            });
            if (channel != null) {
                if (!TErrorList.hasData(channel.getStatus())) {
                    long newtimeout = Math.max(30000,3*timeout); //half a minute minimum
                    try {
                        long time = System.currentTimeMillis();
                        while(wait[0] && (System.currentTimeMillis() - time < newtimeout)) {
                            mutex.wait(10);
                        }
                    } catch (InterruptedException e) {
                        //ignore
                    }
                }
            } else {
                throw new ConnectionException("Could not establish connection.");
            }
        }
        if (TErrorList.hasData(channel.getStatus())) {
            return channel.getRawValue();
        } else {
            if (wait[0]) {
                throw new ConnectionException("Could not read the value: " + TErrorList.toString(TErrorList.connection_timeout),TErrorList.connection_timeout);
            }
            throw new ConnectionException("Could not read the value: " + channel.getStatusString(), channel.getStatus());
        }
    }

    /**
     * Creates a channel that receives notifications according to the given connection mode and rate.
     * If connection mode is CHANGE the callback will receive a notification every time the value changes,
     * if the connection mod is POLL the callback will receive notifications at the specified rate, if the mode
     * is single the callback will receive a single notification. The connection is created using the default
     * data format and size.
     *
     * @param address the property to connect to
     * @param mode the connection mode
     * @param rate the notification rate in milliseconds
     * @param callback the callback to receive notifications
     *
     * @return channel connected to the given property that receives notification according to the policy
     *
     * @throws ConnectionException if there was an error establishing connection
     */
    public Channel getChannel(String address, ConnectionMode mode, int rate,
            ChannelCallback callback) throws ConnectionException {
        return getChannel(address,null,-1,null,mode,rate,callback);
    }

    /**
     * Creates a channel that receives notifications according to the given connection mode and rate.
     * If connection mode is CHANGE the callback will receive a notification every time the value changes,
     * if the connection mod is POLL the callback will receive notifications at the specified rate, if the mode
     * is single the callback will receive a single notification. The connection is created using the default
     * data format and size.
     *
     * @param address the property to connect to
     * @param mode the connection mode
     * @param rate the notification rate in milliseconds
     * @param callback the callback to receive notifications
     * @param dataIn the input data for the monitor
     *
     * @return channel connected to the given property that receives notification according to the policy
     *
     * @throws ConnectionException if there was an error establishing connection
     */
    public Channel getChannel(String address, ConnectionMode mode, int rate,
            ChannelCallback callback, Object dataIn) throws ConnectionException {
        return getChannel(address,null,-1,dataIn,mode,rate,callback);
    }

    /**
     * Creates a channel that receives notifications according to the given connection mode and rate.
     * If connection mode is CHANGE the callback will receive a notification every time the value changes,
     * if the connection mod is POLL the callback will receive notifications at the specified rate, if the mode
     * is single the callback will receive a single notification. The connection is created using the given format
     * and size. Depending on the server, the request might fail if the format and size are not compatible.
     *
     * @param address the property to connect to
     * @param dataOutFormat the requested output format; channel will provide data in this format
     * @param dataOutSize the requested output size; channel will provide data of this size
     * @param mode the connection mode
     * @param rate the notification rate in milliseconds
     * @param callback the callback to receive notifications
     *
     * @return channel connected to the given property that receives notification according to the policy
     *
     * @throws ConnectionException if there was an error establishing connection
     */
    public Channel getChannel(String address, TFormat dataOutFormat, int dataOutSize,
            ConnectionMode mode, int rate, ChannelCallback callback)
            throws ConnectionException {
        return getChannel(address,dataOutFormat,dataOutSize,null,mode,rate,callback);
    }

    /**
     * Creates a channel that receives notifications according to the given connection mode and rate.
     * If connection mode is CHANGE the callback will receive a notification every time the value changes,
     * if the connection mod is POLL the callback will receive notifications at the specified rate, if the mode
     * is single the callback will receive a single notification. The connection is created using the given format
     * and size. Depending on the server, the request might fail if the format and size are not compatible with
     * property type.
     *
     * @param address the property to connect to
     * @param dataOutFormat the requested output format; channel will provide data in this format
     * @param dataOutSize the requested output size; channel will provide data of this size
     * @param dataIn some channels require input data to successfully read their values; the data are provided
     *          by the given parameter
     * @param mode the connection mode
     * @param rate the notification rate in milliseconds
     * @param callback the callback to receive notifications
     *
     * @return channel connected to the given property that receives notification according to the policy
     *
     * @throws ConnectionException if there was an error establishing connection
     */
    public Channel getChannel(String address, TFormat dataOutFormat, int dataOutSize,
            Object dataIn, ConnectionMode mode, int rate, ChannelCallback callback)
            throws ConnectionException {
        if (dataOutFormat == null) {
            short fmt = address.indexOf('*') > 0 ? TFormat.CF_USTRING : TFormat.CF_DEFAULT;
            dataOutFormat = TFormat.valueOf(fmt);
        }
        if (dataOutSize <= 0) {
            dataOutSize = (dataOutFormat.getValue() == TFormat.CF_USTRING) ?
                    DEFAULT_USTRING_SIZE : DEFAULT_SIZE;
        }
        return getChannel(address,dataOutFormat,dataOutSize,null,dataIn,mode,rate,callback);
    }

    /**
     * Creates a channel that receives notifications according to the given connection mode and rate.
     * If connection mode is CHANGE the callback will receive a notification every time the value changes,
     * if the connection mod is POLL the callback will receive notifications at the specified rate, if the mode
     * is single the callback will receive a single notification. The connection is created using the format
     * and size given in the <code>dataOutObject</code>, which will be assigned the output value upon completion.
     *
     * @param address the property to connect to
     * @param dataOutObject the object into which the output data will be written
     * @param dataIn some channels require input data to successfully read their values; the data are provided
     *          by the given parameter
     * @param mode the connection mode
     * @param rate the notification rate in milliseconds
     * @param callback the callback to receive notifications
     *
     * @return channel connected to the given property that receives notification according to the policy
     *
     * @throws ConnectionException if there was an error establishing connection
     */
    public Channel getChannel(String address, TDataType dataOutObject,
            Object dataIn, ConnectionMode mode, int rate, ChannelCallback callback)
            throws ConnectionException {
        short format = TFormat.CF_DEFAULT;
        int size = DEFAULT_SIZE;
        if (dataOutObject != null) {
            format = dataOutObject.getFormat();
            size = dataOutObject.getArrayLength();
        }
        return getChannel(address,TFormat.valueOf(format),size,dataOutObject,dataIn,mode,rate,callback);
    }

    private Channel getChannel(String address, TFormat format, int size,
            final TDataType dataOutObject, Object dataIn, ConnectionMode mode,
            int rate, ChannelCallback callback)
            throws ConnectionException {
        if (rate == 0) rate = 1000; //the client doesn't care about the rate
        if (dataOutObject != null) {
            if (dataOutObject.getFormat() != format.getValue()) {
                throw new IllegalArgumentException("Invalid parameters. The input data doesn't match the input format.");
            }
            if (dataOutObject.getArrayLength() != size) {
                throw new IllegalArgumentException("Invalid parameters. The input data size doesn't match the input size.");
            }
        }

        MasterChannel holder;
        synchronized(channelCache)  {
            String keyDefault = getChannelKey(address,TFormat.CF_DEFAULT,DEFAULT_SIZE,dataIn);
            String key = getChannelKey(address,format.getValue(),size,dataIn);
            holder = channelCache.get(key);
            if (holder == null) {
                holder = channelCache.get(keyDefault);
                if (holder != null) {
                    TFormat fmt = holder.getActualFormat();
                    int sz = holder.getActualSize();
                    if (!((fmt == null && sz == DEFAULT_SIZE) || (fmt != null &&
                            fmt.getValue() == format.getValue() && size == sz))) {
                        holder = null;
                    }
                }
            }
            Counter c = loadCache.get(key);
            if (c == null) {
                c = new Counter();
                if (mode != ConnectionMode.SYNC_SINGLE) {
                    loadCache.put(key,c);
                }
            }
            c.usedNow(holder == null);
            if (holder == null) {
                holder = new MasterChannel(address,format,size,dataIn,c.currentKeepAliveTime);
                if (mode != ConnectionMode.SYNC_SINGLE) {
                    channelCache.put(key,holder);
                }
            }
            holder.increment();
        }
        final MasterChannel master = holder;
//        Future<Channel> future = getToTineExecutor().submit(new Callable<Channel>() {
//            public Channel call() throws Exception {
//                return master.with(mode,rate,dataOutObject,callback);
//            }
//        });
//        try {
//            return future.get();
//        } catch (Exception e) {
//            throw new ConnectionException("Could not create a channel for " + address + ".",e);
//        }
        //we can let the client into tine, because the underlying problem has been fixed.
        return master.with(mode,rate,dataOutObject,callback);
    }

    /**
     * Release the master channel and destroy it. The channel is only destroyed if it is no longer used.
     *
     * @param channel the channel to destroy
     * @return true if the channel was release or false otherwise
     */
    boolean releaseChannel(MasterChannel channel) {
        synchronized(channelCache) {
            if (channel.getCounter() == 0) {
                String key = getChannelKey(channel.getName(),
                        channel.getFormat().getValue(),channel.getSize(),channel.getDataIn());
                channelCache.remove(key);
                channel.destroy();
                return true;
            }
            return false;
        }
    }

    /**
     * @return the shared executor service
     */
    synchronized ScheduledExecutorService getSchedulerService() {
        if (executorService == null) {
            executorService = new ScheduledThreadPoolExecutor(10,new ChannelThreadFactory("TINE-Data-Layer"),new RejectedExecutionHandler() {
                @Override
                public void rejectedExecution(Runnable r, ThreadPoolExecutor executor) {
                    if (r instanceof TINEChannel) {
                        //should never ever really happen, except maybe during clumsy application shutdown
                        TINEChannel c =  ((TINEChannel)r);
                        c.getCallback().channelError("Cannot poll values for the channel " +c.getName() + ". Please try again.",
                                new RejectedExecutionException());
                        c.stop();
                    }
                }
            }){
                @Override
                protected void afterExecute(Runnable r, Throwable t) {
                    if (r instanceof Future<?>) {
                        if (((Future<?>)r).isDone()) {
                            try {
                                ((Future<?>)r).get(1, TimeUnit.MILLISECONDS);
                            } catch (ExecutionException e) {
                                t = e.getCause();
                            } catch (CancellationException e) {
                                //ignore
                            } catch (InterruptedException e) {
                                //ignore
                            } catch (TimeoutException e) {
                                //ignore
                            }
                        }
                    }
                    if (t != null) t.printStackTrace();
                    if (t != null && r instanceof ChannelFuture) {
                        Runnable runnable = ((ChannelFuture<?>)r).task;
                        if (runnable instanceof TINEChannel.Update) {
                            TINEChannel c =  ((TINEChannel.Update)runnable).getChannel();
                            if (c != null) {
                                c.getCallback().channelError("There was an error in the channel callback for " +c.getName() + ".",t);
                                return;
                            }
                        }
                    }
                }

                @Override
                protected <V> RunnableScheduledFuture<V> decorateTask(Runnable runnable, RunnableScheduledFuture<V> task) {
                    return new ChannelFuture<>(task,runnable);
                }
            };
        }
        return executorService;
    }

    synchronized ExecutorService getCallbackExecutor() {
        if (callbackExecutor == null) {
            callbackExecutor = Executors.newSingleThreadExecutor(new ChannelThreadFactory("TINE Layer Callback"));
        }
        return callbackExecutor;
    }

//    private synchronized ExecutorService getToTineExecutor() {
//        if (toTineExecutor == null) {
//            toTineExecutor = new ThreadPoolExecutor(10,10,0L,TimeUnit.SECONDS,new LinkedBlockingQueue<Runnable>(), new ChannelThreadFactory("TINE-Access"),new RejectedExecutionHandler() {
//                @Override
//                public void rejectedExecution(Runnable r, ThreadPoolExecutor executor) {
//                    if (r instanceof TINEChannel) {
//                        //should never ever really happen, except maybe during clumsy application shutdown
//                        TINEChannel c =  ((TINEChannel)r);
//                        c.getCallback().channelError("Cannot poll values for the channel " +c.getName() + ". Please try again.",
//                                new RejectedExecutionException());
//                        c.stop();
//                    }
//                }
//            }){
//                @Override
//                protected void afterExecute(Runnable r, Throwable t) {
//                    super.afterExecute(r,t);
//                    if (t != null && r instanceof TINEChannel) {
//                        TINEChannel c =  ((TINEChannel)r);
//                        c.getCallback().channelError("There was an error in the channel callback for " +c.getName() + ".",t);
//                    }
//                }
//            };
//        }
//        return toTineExecutor;
//    }

    /**
     * Creates the hash key used to distinguish between different requests. Two requests that have the same
     * key, are bound to the same master channel. The key is created by concatenating all parameters separated
     * by '/'.
     *
     * @param address requested property
     * @param dataOutFormat requested output format
     * @param dataOutSize requested output data size
     * @param dataIn the input data
     * @return the hash key
     */
    private static String getChannelKey(String address, short dataOutFormat, int dataOutSize, Object dataIn) {

        StringBuilder dataInStuff = new StringBuilder();
        if (dataIn != null) {
            if (!(dataIn instanceof TDataType)) {
                dataIn = TDataType.toTDataType(dataIn);
            }
            if (dataIn != null) {
                short format = ((TDataType)dataIn).getFormat();
                int size = ((TDataType)dataIn).getArrayLength();
                if (format != TFormat.CF_NULL || size != 0) {
                    dataInStuff = new StringBuilder(50).append(format).append('/').append(size).append('/')
                            .append(String.valueOf(dataIn));
                }
            }
        }
        return new StringBuilder(address.length() + dataInStuff.length() + 10).append(address).append('/')
                .append(dataOutFormat).append('/').append(dataOutSize).append('/').append(dataInStuff).toString();
    }

    /*
     * (non-Javadoc)
     * @see java.lang.Object#finalize()
     */
    @Override
    protected void finalize() throws Throwable {
        close();
        super.finalize();
    }

    /**
     * Close all active links, interrupt all running threads.
     */
    public void close() {
        synchronized(channelCache) {
            for (MasterChannel c : channelCache.values()) {
                c.destroy();
            }
            channelCache.clear();
            loadCache.clear();
        }
        if (executorService != null) {
            executorService.shutdownNow();
        }
        if (callbackExecutor != null) {
            callbackExecutor.shutdownNow();
        }
//        if (toTineExecutor != null) {
//            toTineExecutor.shutdownNow();
//        }
    }
}
