package de.desy.tine.accesslayer;

import java.util.EventObject;

import de.desy.tine.dataUtils.TDataType;

/**
 *
 * <code>ChannelEvent</code> describes a read event on the channel. It contains the reference to the source channel
 * as well as the read value and status.
 *
 * @author <a href="mailto:jaka.bobnar@cosylab.com">Jaka Bobnar</a>
 *
 */
public class ChannelReadEvent extends EventObject {

    private static final long serialVersionUID = -56765816451952596L;
    private final TDataType rawValue;
    private final int status;
    private final String statusString;
    private final boolean successful;

    /**
     * Construct a new read event.
     *
     * @param source the source of the event
     * @param rawValue the value
     * @param status the status code
     * @param statusString the human readable representation of the status code
     * @param successful indicates if this event describes a success (true) or an error (false)
     */
    public ChannelReadEvent(Channel source, TDataType rawValue, int status, String statusString, boolean successful) {
        super(source);
        this.rawValue = rawValue;
        this.status = status;
        this.statusString = statusString;
        this.successful = successful;
    }

    /**
     * Returns the channel, which is always the source of this event.
     *
     * @return the source channel
     */
    public Channel getChannel() {
        return (Channel)getSource();
    }

    /**
     * Returns the raw value that was read from the channel.
     *
     * @return the raw value
     */
    public TDataType getRawValue() {
        return rawValue;
    }

    /**
     * Returns the status code of the read action.
     *
     * @return the status code
     */
    public int getStatus() {
        return status;
    }

    /**
     * Returns the status code in a human readable text.
     *
     * @return the status code as a string
     */
    public String getStatusString() {
        return statusString;
    }

    /**
     * Returns true if this event describes an action completed successfully, or false if the action was completed
     * with an error
     *
     * @return true if successful or false if an error
     */
    public boolean isSuccessful() {
        return successful;
    }
}
