/*
 * Created on Aug 16, 2007
 *
  */
package de.desy.tine.client;

import java.lang.reflect.InvocationTargetException;

import javax.swing.SwingUtilities;

import de.desy.tine.definitions.TErrorList;
import de.desy.tine.server.logger.MsgLog;

/**
 * A TLink callback which is executed synchronously on the AWT 
 * event dispatching thread.
 * <p>
 * This callback is useful for programs with an AWT or Swing GUI.
 * It is not advisable to make any GUI modifications directly
 * from TLinkCallback.callback(), since different threads are
 * involved. An indirection via the AWT event queue has to 
 * be implemented.   
 * <p>
 * This class simplifies GUI callbacks for client programmers.
 * The programmer simply replaces TLinkCallback with TAwtSyncCallback
 * and implements onCallback() instead of callback. From onCallback(),
 * GUI components can be accessed safely.
 * <p>
 * In this class, the inherited TINE callback() method calls 
 * SwingUtilities.invokeAndWait(). It does not return 
 * until onCallback() and the subsequent AWT events have 
 * been processed. 
 * Compared to TAwtAsyncCallback, this is the safer, 
 * but slower way to use TINE callbacks in GUI applications. 
 *
 * @author jwlg
 */
public abstract class TAwtSyncCallback implements TLinkCallback, Runnable {

    private TLink storedLink = null;
    
    /**
     * The original TINE callback method.
     */
    public final void callback(TLink link) {
        this.storedLink = link;
        try {
            SwingUtilities.invokeAndWait(this);
        } catch (InterruptedException e) {
          MsgLog.log("TAwtSyncCallback", e.getMessage(), TErrorList.io_error, e, 1);
        } catch (InvocationTargetException e) {
          MsgLog.log("TAwtSyncCallback", e.getMessage(), TErrorList.code_failure, e, 1);
        }
    }

    /**
     * Called from the AWT event thread.
     */
    public final void run() {
        onCallback(storedLink);
    }
    
    /**
     * Returns the link status.
     * @return
     */
    public final int getLinkStatus() {
        return storedLink.getLinkStatus();
    }
    
    /**
     * Returns the error string.
     */
    public final String getErrorString() {
        return storedLink.getLastError();
    }

    /**
     * Replaces TLinkCallback.callback().
     * This method to be implemented by derived classes. 
     * 
     * @param link The TLink which fired this callback.
     */
    public abstract void onCallback(TLink link);

}
