package de.desy.tine.definitions;

/**
 * Data access flags
 * 
 * The access word in a TLink will contain information pertaining
 * to the type of access requested by the caller or by the subsystem. A
 * client can specify 'CA_READ' or 'CA_WRITE' and in addition ask for a
 * data link over a connected (TCP/IP) socket by specifying
 * 'CA_CONNECT'. A server can examine the access word when handling the
 * call. A server can also determine whether the call is coming from the
 * local alarm server or local history server. Likewise, a server can
 * examine the scope of the call, that is whether the link is starting
 * up (the 'CA_FIRST' bit is set) or going out of scope (the 'CA_LAST'
 * bit is set).
 */
public final class TAccess
{
  private short mode;
  public static final short CA_NULL = 0x00;
  public static final short CA_READ = 0x01;
  public static final short CA_WRITE = 0x02;
  public static final short CA_REPEAT = 0x04; // applied to access in call to property handler (following 'not_ready')
  public static final short CA_SYNC = 0x08;
  public static final short CA_ALARM = 0x10; // applied to access in call to property handler (from local alarm sub system)
  public static final short CA_HIST = 0x20; // applied to access in call to property handler (from local history sub system)
  public static final short CA_FIRST = 0x40; // applied to access in call to property handler (coming into scope) 
  public static final short CA_LAST = 0x80; // applied to access in call to property handler  (going out of scope)
  public static final short CA_STATIC = 0x80; // property registration contains static information
  public static final short CA_RETRY = 0x100; // retry call before issuing a link_timeout
  public static final short CA_FORCEOUTPUT = 0x100; // property registration for contract coercion
  public static final short CA_NETWORK = 0x200; // property registration signals NETWORK access required
  public static final short CA_QUERY = 0x400;
  public static final short CA_SINGLE = 0x400;// used as signal during contract examination
  public static final short CA_LOCKED = 0x800;// applied to access in call to property handler
  public static final short CA_XREAD = 0x800; // used as signal during property registration
  public static final short CA_NORETRY = 0x1000; // call should NOT retry on timeout (cln: passed to link call)
  public static final short CA_NOSYNC = 0x1000; // used as signal during property registration
  public static final short CA_CONNECT = 0x2000; // sync call should use CM_CONNECT access (cln: passed to link call)
  public static final short CA_MUTABLE = 0x4000; // sync call can change the input parameters (cln: passed to link call)
  public static final short CA_SAVERESTORE = 0x4000; // property registration signals save-and-restore property
//  public static final short CA_SYNCNOTIFY = 0x8000; <doesn't work as a short in java !
  /**
   * Constructor. Used for when CF-constants are OR-ed
   * 
   * @param newMode
   */
  public TAccess(int newMode)
  {
    mode = (short) newMode;
  }
  public TAccess()
  {
    this(CA_NULL);
  }
  /**
   * Sets the access mode to CA_NULL
   */
  public void clear()
  {
    mode = CA_NULL;
  }
  /**
   * Gets the TINE access code for the given access string
   * 
   * @param accessString is the access string of the form "READ", "READ|WRITE",
   * "XREAD|WRITE", etc.
   * 
   * @return the TINE access code
   */
  public static short getAccessCode(String accessString)
  {
    short access = 0;
    if (accessString != null)
    {
      try 
      {
        int a = Integer.parseInt(accessString);
        if (a == 1) access |= CA_READ;
        if (a == 2) access |= CA_WRITE;
        if (a == 3) access |= CA_READ|CA_WRITE;
      }
      catch (Exception ignore) {}
      String s = accessString.toUpperCase();
      if (s.contains("READ")) access |= CA_READ;
      if (s.contains("WRITE")) access |= CA_WRITE;
      if (s.contains("NETWORK")) access |= CA_NETWORK;
      if (s.contains("SYNC")) access |= CA_SYNC;
      if (s.contains("CONNECT")) access |= CA_CONNECT;
      if (s.contains("SAVERESTORE")) access |= CA_SAVERESTORE;
      if (s.contains("FORCEOUTPUT")) access |= CA_FORCEOUTPUT;
    }
    return access;
  }
  /**
   * Determines whether the given input access string
   * has 'exclusive read' access
   * 
   * @param accessString is the access string of the form "READ", "READ|WRITE",
   * "XREAD|WRITE", etc.
   * 
   * @return <code>true</code> or <code>false</code>
   */
  public static boolean hasExclusiveRead(String accessString)
  {
    if (accessString != null)
    {
      String s = accessString.toUpperCase();
      if (s.contains("XREAD")) return true;
    }
    return false;
  }
  /**
   * Determines whether the given input property access string
   * enforces the output data type and length
   * 
   * @param accessString is the access string of the form "READ", "READ|WRITE",
   * "XREAD|WRITE", etc.
   * 
   * @return <code>true</code> or <code>false</code>
   */
  public static boolean isEnforceOutput(String accessString)
  {
    if (accessString != null)
    {
      String s = accessString.toUpperCase();
      if (s.contains("FORCEOUTPUT")) return true;
    }
    return false;
  }
  /**
   * Determines whether the given input access string
   * has 'exclusive read' access even without an access lock
   *
   * @param accessString is the access string of the form "READ", "READ|WRITE",
   * "XREAD|WRITE", etc.
   * 
   * @return true of false
   */
  public static boolean hasUnlockedExclusiveRead(String accessString)
  {
    if (accessString == null) return false;
    int idx = -1;
    String s = accessString.toUpperCase();
    if ((idx=s.indexOf("XREAD")) >= 0) 
    { // so far : true; now see if "READ" shows up alone ...
      String s1 = s.substring(0, idx);
      s1 += s.substring(idx+5, s.length());
      if (s1.indexOf("READ") < 0) return true;
    }
    return false;
  }
  /**
   * converts to the principal access required 
   * (combination of CA_READ or CA_WRITE).
   * @param access is the access mode for which the base mode is desired
   * @return the base access mode 
   * (one of CA_NULL, CA_READ, CA_WRITE, or CA_READ|CA_WRITE)
   */
  public static short toBase(short access)
  {
    return (short) (access & 0x03);
  }
  
  /**
   * Converts the given input to a TAccess object.
   * @param mode is the access mode for which a TAccess object is desired
   * @return a TAccess Object representative of the input mode
   */
  public static TAccess valueOf(short mode) 
  {
	return new TAccess(mode);
  }
  /** 
   * Returns a string representation of the access mode
   * @return the string representation of the access mode 
   */
  public String toString()
  {
    return TAccess.toString(mode);
  }
  /** 
   * Returns a string representation of the access mode given
   * @param access is the access mode for which the string representation is desired
   * @return the string representation of the access mode 
   */
  public static String toString(short access)
  {
    String add = "";
    if ((access & CA_SYNC) == CA_SYNC) add = "sync ";
    if ((access & CA_NETWORK) == CA_NETWORK) add = "network ";
    switch (access & 0x03)
    {
      case CA_NULL:
        return "none";
      case CA_READ:
        return add + "read only";
      case CA_WRITE:
        return add + "write";
      default:
        return add + "read/write";
    }
  }
  /**
   * Returns the access mode bits as a short integer.
   * 
   * @return
   */
  public short toShort()
  {
    return this.mode;
  }
  /**
   * Returns true if all bits which are set in pattern are also set in this
   * object.
   * 
   * @param pattern
   *          A Bit pattern.
   * @return
   */
  public boolean includes(short pattern)
  {
    return ((this.mode & pattern) == pattern);
  }
  /**
   * Returns true if read bit is set.
   * @return <code>true</code> if read bit is set.
   */
  public boolean isRead()
  {
    return ((this.mode & CA_READ) != 0);
  }
  /**
   * Returns true if read bit is set.
   * @param access is the access byte to test
   * @return <code>true</code> if read bit is set.
   */
  public static boolean isRead(byte access)
  {
    short uaccess = access;
    if (uaccess < 0) uaccess += 256;
    return (access & CA_READ) != 0;
  }
  /**
   * Returns true if write bit is set.
   * @return <code>true</code> if write bit is set.
   */
  public boolean isWrite()
  {
    return ((this.mode & CA_WRITE) != 0);
  }
  /**
   * Returns true if write bit is set.
   * @param access is the access byte to test
   * @return <code>true</code> if write bit is set.
   */
  public static boolean isWrite(byte access)
  {
    short uaccess = access;
    if (uaccess < 0) uaccess += 256;
    return (access & CA_WRITE) != 0;
  }
  /**
   * Returns true if network bit is set.
   * @return true if read bit is set.
   */
  public boolean isNetwork()
  {
    return ((this.mode & CA_NETWORK) != 0);
  }
  /**
   * Returns true if network bit is set.
   * @param access is the access byte to test
   * @return true if read bit is set.
   */
  public static boolean isNetwork(short access)
  {
    short uaccess = access;
    if (uaccess < 0) uaccess += 256;
    return (access & CA_NETWORK) != 0;
  }
  public static boolean isConnected(short access)
  {
    short uaccess = access;
    if (uaccess < 0) uaccess += 256;
    return (access & CA_CONNECT) != 0;
  }
  /**
   * Returns true if FIRST bit is set.
   * @return true if FIRST bit is set.
   */
  public boolean isFirst()
  {
    return ((this.mode & CA_FIRST) != 0);
  }
  /**
   * Returns true if FIRST bit is set.
   * @param access is the access byte to test
   * @return true if FIRST bit is set.
   */
  public static boolean isFirst(byte access)
  {
    short uaccess = access;
    if (uaccess < 0) uaccess += 256;
    return (access & CA_FIRST) != 0;
  }
  /**
   * Returns true if LAST bit is set.
   * @return true if LAST bit is set.
   */
  public boolean isLast()
  {
    return ((this.mode & CA_LAST) != 0);
  }
  /**
   * Returns true if LAST bit is set.
   * @param access is the access byte to test
   * @return true if LAST bit is set.
   */
  public static boolean isLast(byte access)
  {
    short uaccess = access;
    if (uaccess < 0) uaccess += 256;
    return (access & CA_LAST) != 0;
  }
  /**
   * Returns the input access byte with the appropriate bits removed
   * @param access is the access byte from which to remove the bit
   * @param bits are the access bits to remove
   * @return the input access byte with the appropriate bit removed
   */
  public static byte removeBits(byte access,short bits)
  {
    short uaccess = access;
    if (uaccess < 0) uaccess += 256;
    uaccess &= ~(bits);
    return (byte)uaccess;
  }
  /**
   * Returns true if both the FIRST and LAST bits are set.
   * @param access is the access byte to test
   * @return true if both the FIRST and LAST bits are set.
   */
  public static boolean isSynchronous(byte access)
  {
    short uaccess = access;
    if (uaccess < 0) uaccess += 256;
    return (access & (CA_LAST|CA_FIRST)) == (CA_FIRST|CA_LAST);
  }
  /**
   * Returns true if no bit is set.
   * 
   * @return
   */
  public boolean isNull()
  {
    return (this.mode == CA_NULL);
  }
  /**
   * @return true if the access contains the CA_RETRY flag
   */
  public boolean canRetry()
  {
    return ((this.mode & CA_RETRY) != 0);
  }
  /**
   * Returns true if all bits are equal.
   * 
   * @param acc
   * @return
   */
  public boolean equals(TAccess acc)
  {
    return (this.mode == acc.mode);
  }
  /**
   * Returns true if all mode bits which are set in another object are also set
   * in this object.
   * 
   * @param acc
   * @return
   */
  public boolean includes(TAccess acc)
  {
    return includes(acc.mode);
  }
  /**
   * Appends the input access mode flag
   */
  public void add(short mode)
  {
    this.mode |= mode;
  }
  /**
   * Appends the input access mode flag
   */
  public void add(TAccess acc)
  {
    this.mode |= acc.mode;
  }
  // TODO More methods (getSync etc)
}