package de.desy.tine.example.sine;

import java.util.LinkedList;
import de.desy.tine.dataUtils.TDataType;
import de.desy.tine.definitions.TErrorList;
import de.desy.tine.definitions.TFormat;
import de.desy.tine.server.alarms.TAlarm;
import de.desy.tine.server.alarms.TAlarmDefinition;
import de.desy.tine.server.equipment.*;

/**
 * This is an example of a de.desy.tine.client-based server. The device server consists of 10
 * devices. Each device generates data of a propagating sine wave. For each
 * device, the amplitude and the frequency of the sine wave can be set and read.
 */
public class SineDeviceServer
{
  private static SineEquipmentModule sineEqpModule;
  private static CoSineEquipmentModule cosineEqpModule;
  private static TEquipmentBackgroundTask sineBkgFcn;
  private TEquipmentModuleFactory thisEqmFactory;
  private LinkedList sineDeviceSet = new LinkedList();
  private LinkedList cosineDeviceSet = new LinkedList();
  /**
   * In the constructor we instantiate our devices and
   * initialize the device server.
   */
  public SineDeviceServer()
  {
    initializeDevices(); // should create a device list (eqm not yet known !)
    initializeDeviceServer(); // will create an eqm, passing a device list
   // optional: spit these out just to confirm they have been registered ...
    sineEqpModule.dumpProperties(); 
    sineEqpModule.dumpDevices();
  }
  private void initializeDevices()
  {
    // create the devices and add them to a device set
    for (int i = 0; i < 10; i++)
    {
      // add new devices; assign them only a device number :
      sineDeviceSet.add(new SineDevice(i));
      // The device names will be assigned through the configuration
      // file <local-name>-devices.csv, in this case: "SINEQM-devices.csv"
      // Alternative: fill in via your own database or just fake it:
      // e.g. sineDeviceSet.add(new SineDevice("SINE_DEV_" + i,i));
      cosineDeviceSet.add(new CoSineDevice("CosineGen" + i,i));
    }
  }
  private static final int TOO_MUCH_NOISE = 512;
  private void initializeDeviceServer()
  {
    // Create Equipment Module(s) (this example has only one)
    sineEqpModule = new SineEquipmentModule("SINEQM",(SineDevice[])sineDeviceSet.toArray(new SineDevice[0]));
    sineEqpModule.setUseMSecHistoryTimestamps(true);
    // Create an alarm definition for 'TOO_MUCH_NOISE'
    // give it a severity of 7 and put it in the alarm system designated by 300 (these need to be enumerated somewhere)
    sineEqpModule.addAlarmDefinition(
        new TAlarmDefinition(
            "Too much noise",
            "The sine curve has too much noise",
            "sine curve tag",
            TOO_MUCH_NOISE,7,300,0xffff,
            TFormat.CF_FLOAT,1,"noise level that caused the alarm",
            "http://www.malfunctions.org")
        );
    //Create Background Task(s) (hardware IO, middle-layer activities, etc.)
    sineBkgFcn = new TEquipmentBackgroundTask()
    { // must implement the 'call' method :
      public void call()
      {
        SineDevice dev = null;
        float[] noiseLevel = new float[1];
        noiseLevel[0] = (float)50.0; // just a test example
        TDataType v = new TDataType(noiseLevel);
        for (int i=0; i<sineDeviceSet.size(); i++)
        {
          dev = (SineDevice)sineDeviceSet.get(i);
          dev.update();
          dev.clearAlarms(); // clear all alarms and see if they come back ...
        }
        // set alarm value_too_high for device # 0
   //     dev = (SineDevice)sineDeviceSet.get(0);
   //     dev.setAlarm(TOO_MUCH_NOISE,v.getDataBuffer());
      }
    };
    // how often should the background task be called :
    sineBkgFcn.setBackgroundTaskInterval(200);  // msec

    cosineEqpModule = new CoSineEquipmentModule("COSEQM",(CoSineDevice[])cosineDeviceSet.toArray(new CoSineDevice[0]));
  
  }
  public void activate()
  {
    // record a reference to the equipment module factory (all modules will use this one)
    thisEqmFactory = sineEqpModule.getTEqmFactory();
    // add the background task(s) to the factory
    if (sineBkgFcn != null) 
      thisEqmFactory.addEquipmentBackgroundTask(sineBkgFcn);
    // initialize the FEC (starts all services, reads configuration databases, etc.)
    thisEqmFactory.systemInit();   // initialize the factory
    // alternative: thisEqmFactory.systemInit(myFecName, myTinePort) hardcodes 
    // the FECNAME instead of using fecid.csv
    thisEqmFactory.systemWait(-1); // wait here forever ... (or loop here and check command line?)
  }
  public static void main(String[] args)
  {
    // create the device server; export name read from exports.csv file 
    SineDeviceServer theServer = new SineDeviceServer();
    // optional: sineEqpModule.setExportName(myExportName) hardcodes 
    // the exported device server name instead of reading from exports.csv
    
    // Now start serving ...
    theServer.activate();
  }
}