package de.desy.tine.headers;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.IOException;
import java.util.Arrays;

import de.desy.tine.dataUtils.UnsignedUtils;
import de.desy.tine.definitions.TEncoding;
import de.desy.tine.definitions.TEndianness;
import de.desy.tine.definitions.TErrorList;
import de.desy.tine.definitions.TTransport;
import de.desy.tine.endianUtils.Swap;
import de.desy.tine.server.logger.MsgLog;

public final class TRspHdr
{
  // header bytes:
  private int tineProtocol;
  private int totalSizeInBytes;
  private int headerSizeInBytes;
  private short numberSegments;
  private byte endianness;
  private byte encoding;
  private String fecname = null; // 16 char
  // end of header bytes
  static public final int hdrSizeInBytes = 32; // network header size
  private boolean isjumbo = false; // for legacy STREAM request
  public boolean isJumbo() { return isjumbo; }
  ByteArrayOutputStream dBuffer;
  public TRspHdr(int sizeInBytes,String fn,int protocolLevel)
  {
    this(sizeInBytes,fn,1,protocolLevel);
  }
  public TRspHdr(int sizeInBytes,String fn,int segments,int protocolLevel)
  {
    this(sizeInBytes,fn,segments,protocolLevel,false);
  }
  public TRspHdr(int sizeInBytes,String fn,int segments,int protocolLevel,boolean isJumbo)
  {
    totalSizeInBytes = sizeInBytes;
    fecname = fn.toUpperCase();
    tineProtocol = protocolLevel;
    dBuffer = new ByteArrayOutputStream(hdrSizeInBytes);
    headerSizeInBytes = getHeaderSizeInBytes(protocolLevel,isJumbo);
    endianness = (byte)TEndianness.BO_LITTLE_ENDIAN; // for the moment
    encoding = (byte)TEncoding.CE_ASCII;
    numberSegments = (short)segments;
    isjumbo = isJumbo;
  }
  public int getHeaderSizeInBytes()
  {
    if (tineProtocol <= TTransport.legacyProtocolLevel)
    {
      return isjumbo ? 4 : 2;
    }
    return headerSizeInBytes;
  }
  public static int getHeaderSizeInBytes(int protocol,boolean isJumbo)
  {
    if (protocol <= TTransport.legacyProtocolLevel)
    {
      return isJumbo ? 4 : 2;
    }
    return hdrSizeInBytes; 
  }
  public int getNumberSegments() { return numberSegments; }
  public void setNumberSegments(int value)
  {
    numberSegments = (short)(value > 1 ? value : 1);
  }
  public int getEndianness() { return endianness; }
  public int getEncoding() { return encoding; }
  public String getFecName() { return fecname; }
  public void setFecName(String name) 
  { 
    if (name == null) return;
    if (fecname != name && fecname.compareToIgnoreCase(name) != 0)
    {
      fecname = name.toUpperCase();
    }
  }
  public int getTineProtocol()
  {
    return tineProtocol;
  }
  public void setTineProtocol(int protocolLevel)
  {
    tineProtocol = protocolLevel;
  }
  public int getTotalSizeInBytes() { return totalSizeInBytes; }
  private byte[] bstr = new byte[16];
  public byte[] toByteArray()
  { // prepare outgoing bytestream (e.g. server sending response to client)
    try
    {
      dBuffer.reset();
      DataOutputStream ds = new DataOutputStream(dBuffer);
      if (tineProtocol <= TTransport.legacyProtocolLevel)
      {
        if (isjumbo)
        { // legacy CM_STREAM request
          ds.writeInt(Swap.Int(totalSizeInBytes));
        }
        else
        {
          ds.writeShort(Swap.Short(totalSizeInBytes));
        }
      }
      else
      {
        ds.writeInt(Swap.Int(tineProtocol));
        ds.writeInt(Swap.Int(totalSizeInBytes));
        ds.writeInt(Swap.Int(headerSizeInBytes));
        ds.writeShort(Swap.Short(numberSegments));
        ds.write(endianness);
        ds.write(encoding);   
        Arrays.fill(bstr, (byte)0);
        byte[] tmp  = fecname.getBytes();
        for (int i=0; i<16 && i<tmp.length; i++) bstr[i] = tmp[i];
        ds.write(bstr,0,16);
      }
      ds.close();
      return dBuffer.toByteArray();
    }
    catch (IOException e)
    {
      e.printStackTrace();
      MsgLog.log("TRspHdr.toByteArray",e.toString(),TErrorList.code_failure,e,0);
    }
    return null;  
  }
  public TRspHdr(byte[] data,boolean isJumbo)  
  { // prepare from incoming bytestream (e.g. client getting response from server)
    try
    {
      ByteArrayInputStream instream = new ByteArrayInputStream(data);
      DataInputStream ds = new DataInputStream(instream);
      tineProtocol = Swap.Int(ds.readInt());
      if (tineProtocol < 0 || tineProtocol > TTransport.thisProtocolLevel)
      { // indicative of response from legacy server
        // get to position 18 in the stream :
        ds.skipBytes(14);
        //for (int i=0; i<14; i++) ds.readByte();
        tineProtocol = Swap.Short(ds.readShort());
        if (tineProtocol != TTransport.legacyProtocolLevel) 
        { // release 3.xx -> really old!
          tineProtocol = TTransport.legacyProtocolLevel - 1;  
        }
        ds.reset();
      }
      if (tineProtocol <= TTransport.legacyProtocolLevel)
      { // we only read either 2 or 4 bytes from the stream
        totalSizeInBytes = isJumbo ? Swap.Int(ds.readInt()) : UnsignedUtils.toUnsignedShort(Swap.Short(ds.readShort()));
        headerSizeInBytes = isJumbo ? 4 : 2;
        endianness = (byte)TEndianness.BO_LITTLE_ENDIAN;
        encoding = (byte)TEncoding.CE_ASCII;
        isjumbo = isJumbo;
        // TODO: worry about fecname here ?
      }
      else
      { // we read the entire response header from the stream (tineProtocol already read)
        totalSizeInBytes = Swap.Int(ds.readInt());
        headerSizeInBytes = Swap.Int(ds.readInt());
        numberSegments = Swap.Short(ds.readShort());
        endianness = ds.readByte();
        encoding = ds.readByte();      
        Arrays.fill(bstr, (byte)0);
        ds.read(bstr,0,16);
        fecname = (new String(bstr)).trim();
      }
      ds.close();
      instream.close();
    }
    catch (IOException e)
    {
      e.printStackTrace();
      MsgLog.log("TRspHdr",e.toString(),TErrorList.code_failure,e,0);
    }
  }      

}
