package de.desy.tine.histUtils;

import java.io.FileWriter;
import java.io.PrintWriter;
import java.text.DecimalFormat;
import java.text.DecimalFormatSymbols;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Comparator;
import java.util.HashMap;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.List;
import java.util.Locale;
import java.util.Set;
import java.util.Vector;
import java.util.concurrent.locks.Lock;
import java.util.concurrent.locks.ReentrantLock;

import de.desy.tine.addrUtils.TSrvEntry;
import de.desy.tine.client.TLink;
import de.desy.tine.client.TLinkFactory;
import de.desy.tine.client.TLinkFactory.RedirectedItem;
import de.desy.tine.dataUtils.TDataTime;
import de.desy.tine.dataUtils.TDataType;
import de.desy.tine.definitions.TAccess;
import de.desy.tine.definitions.TErrorList;
import de.desy.tine.definitions.TFormat;
import de.desy.tine.exceptions.TineRuntimeErrorException;
import de.desy.tine.exceptions.UnresolvedAddressException;
import de.desy.tine.queryUtils.TPropertyQuery;
import de.desy.tine.queryUtils.TQuery;
import de.desy.tine.server.histories.THistoryRecordStruct;
import de.desy.tine.server.histories.THistorySpecification;
import de.desy.tine.server.logger.DbgLog;
import de.desy.tine.server.logger.MsgLog;
import de.desy.tine.stringUtils.NameToString;
import de.desy.tine.stringUtils.Str;
import de.desy.tine.stringUtils.StringToName;
import de.desy.tine.types.*;

/**
 * THistory is a utility class with numerous static methods for access
 * archive data from either the central archive system of the local
 * history subsystems of targeted servers.  
 * 
 * @author duval 
 */
public final class THistory
{
  private static final int HEARTBEAT = 60 * 15;
  private static final int MAX_HSTCFG_SIZE = 500;
  private static final int MAX_STATES_SIZE = 256;
  private static HstCfg[] cfgs = new HstCfg[MAX_HSTCFG_SIZE];
  private static McaCfg[] mcacfgs = new McaCfg[MAX_HSTCFG_SIZE];
  private static PrfCfgEx[] prfcfgs = new PrfCfgEx[MAX_HSTCFG_SIZE];
  private static HstView[] views = new HstView[MAX_HSTCFG_SIZE];
  private static DclState[] dclstates = new DclState[MAX_STATES_SIZE];
  private static DclProc[] dclprocs = new DclProc[MAX_STATES_SIZE];
  private static boolean isInitialized = false;
  private static boolean isStatesInitialized = false;
  private static void resetCfgs()
  {
    if (!isInitialized) return;
    for (int i=0; i<MAX_HSTCFG_SIZE; i++)
    {
      cfgs[i].clear();
      mcacfgs[i].clear();
      prfcfgs[i].clear();
      views[i].clear();
    }    
  }
  @SuppressWarnings("unused")
  private static void resetStates()
  {
    if (!isStatesInitialized) return;
    for (int i=0; i<MAX_STATES_SIZE; i++)
    {
      dclstates[i].clear();
      dclprocs[i].clear();
    }        
  }
  private static void initCfgs()
  {
    if (isInitialized) return;
    for (int i=0; i<MAX_HSTCFG_SIZE; i++)
    {
      cfgs[i] = new HstCfg();
      mcacfgs[i] = new McaCfg();
      prfcfgs[i] = new PrfCfgEx();
      views[i] = new HstView();
    }
    isInitialized = true;
  }
  private static void initStates()
  {
    if (isStatesInitialized) return;
    for (int i=0; i<MAX_STATES_SIZE; i++)
    {
      dclstates[i] = new DclState();
      dclprocs[i] = new DclProc();
    } 
    isInitialized = true;
  }
  public static boolean isMonitoredByCentralArchiver(String context, String server)
  {
    int[] n = new int[1];
    if (context == null || context.length() == 0) return false;
    if (server == null || server.length() == 0) return false;
    if (isCentralArchiveServer(server) || isCentralArchiver(server)) return false;
    try
    {
      TLink lnk = new TLink("/"+context+"/ARCHIVER","NKEYWORDS",new TDataType(n),new TDataType(server),TAccess.CA_READ);
      if (lnk.executeAndClose() != 0) return false;
      return n[0] > 0  && n[0] < 100;
    }
    catch (Exception any)
    {
      return false;
    }
  }
  public static boolean isCentralArchiveServer(String server)
  {
    return server.endsWith("HISTORY");
  }
  public static boolean isCentralArchiver(String server)
  {
    return server.endsWith("ARCHIVER");
  }
  private static String getQueryPropertyRoot(String property)
  {
    int slen = property.length();
    if (property.endsWith(".HIST") || property.endsWith(".ARCH")) 
      return property.substring(0, slen-5);
    if (property.endsWith(".HST") || property.endsWith(".ARC")) 
      return property.substring(0, slen-4);
    if (property.endsWith(".AR")) 
      return property.substring(0, slen-3);
    return property;
  }
  private static boolean isScheduledProperty(String property)
  {
    int idx = property.lastIndexOf('.');
    if (idx < 0) return false;
    String dec = property.substring(idx+1);
    if (dec == null) return false;
    return dec.toUpperCase().startsWith("SCH");
  }
  private static boolean isArchiveQueryProperty(String property)
  {
    if (property.endsWith(".HIST")) return true;
    if (property.endsWith(".HST")) return true;
    if (property.endsWith(".ARCH")) return true;
    if (property.endsWith(".ARC")) return true;
    if (property.endsWith(".AR")) return true;
    return false;
  }
  private static boolean preferCentralSource = false; //TODO: this of course not thread-safe
  private static boolean preferLocalSource = false;
  /**
   * Establishes a preference for central archive data over local history data in archive calls if 
   * this setting is 'true'.
   * @param value is the desired setting
   */
  public static void setPreferCentralSource(boolean value)
  {
    preferCentralSource = value;
  }
  /**
   * Establishes a preference for local history data over central archive data in archive calls if 
   * this setting is 'true'.
   * @param value is the desired setting
   * 
   * @note this setting takes precedence over setPreferCentralSource
   */
  public static void setPreferLocalSource(boolean value)
  {
    preferLocalSource = value;
  }
  private static HashMap<String, Integer> hbtTbl = new HashMap<String,Integer>();
  public static int getArchivedHeartbeat(String context, String server, String property)
  {
    if (context == null || property == null) return TErrorList.argument_list_error;
    if (server == null) server = "HISTORY"; // assume the standard central archive server for the given context
    String key = "/"+context+"/"+server+"/"+property;
    if (hbtTbl.containsKey(key)) return hbtTbl.get(key).intValue();
    if (isCentralArchiveServer(server))
    {
      try
      {
        DAQRec res = TArchiver.getRecordFromKeyword(context, property);
        Integer hbt = res.getHeartbeat();
        hbtTbl.put(key,hbt);
        return hbt.intValue();
      }
      catch (Exception any) 
      {
        any.printStackTrace(); 
      }
    }
    else
    {
      THistoryRecordStruct[] hrs = new THistoryRecordStruct[1];
      TLink lnk = new TLink("/"+context+"/"+server+"/"+property,"HISTORIES",new TDataType(hrs),null,TAccess.CA_READ);
      if (lnk.execute() == 0)
      {
        Integer hbt = hrs[0].getHeartbeat();
        hbtTbl.put(key,hbt);
        return hbt.intValue();
      }
    }
    return HEARTBEAT;
  }
  private static int getSuggestedTimeoutMilliseconds(double start, double stop)
  {
    int timeout = 1000;
    int nmonths = (int)((stop - start)/2592000.0);
    if (nmonths > 1) timeout = nmonths * 4000;
    if (timeout > 30000) timeout = 30000;
    return timeout;
  }
  private static synchronized int getArchivedNumPointsInInterval(String context,String server, String property, String device, double start, double stop,int timeout)
  {
    int cc = 0;
    int[] startstop = new int[2];
    int[] npts = new int[1];
    TDataType dout, din;
    
    preferCentralSource = false;
    if (server == null && context == null) return TErrorList.argument_list_error;
    if (context == null) context = "DEFAULT";
    if (server == null) server = "HISTORY"; // assume the standard central archive server for the given context
    startstop[0] = (int)start;
    startstop[1] = (int)Math.ceil(stop);
    din = new TDataType(startstop);
    dout = new TDataType(npts);
    
    if (isCentralArchiver(server))
    { // trap a potential redirection problem here:
      if (device.compareToIgnoreCase("keyword") == 0) device = "#0";
    }
    String dname = new String("/" + context + "/" + server + "/" + device);
    String pname = new String(property);
    if (!isCentralArchiveServer(server))
    { // not a call to a central archive server
      if (!isArchiveQueryProperty(property))
        pname = pname + ".HIST"; // meta extension not given, try this ...
    }
    try
    {
      TLink tl = new TLink(dname,pname,dout,din,TAccess.CA_READ);
      long tnow = System.currentTimeMillis();
      cc = tl.execute(timeout);
      while (cc == TErrorList.operation_busy)
      {
        timeout = getSuggestedTimeoutMilliseconds(start,stop);
        cc = tl.execute(timeout);
        TLinkFactory.systemWait(10);
        if (System.currentTimeMillis() > tnow + timeout) break;
      }
      tl.close();
      if (TErrorList.hasData(cc))
      {
        if (TErrorList.getErrorCode(cc) == TErrorList.at_limit) 
        { // signal from a local history that there isn't any long term storage
          preferCentralSource = true;
        }
        cc = 0;
      }
    }
    catch (UnresolvedAddressException e)
    {
      cc = TErrorList.address_unresolved;
    }
    return cc > 0 ? -cc : npts[0];
  }
  public static String[] getArchiveSource(String context,String server, String property, String device,int timeout)
  {
    String[] src = null;
    
    int cc = 0;
    TDataType dout;
    
    if (server == null && context == null) return src;
    if (context == null) context = "DEFAULT";
    if (server == null) server = "HISTORY"; // assume the standard central archive server for the given context
    if (srcTbl != null)
    {
      String value = device+"["+property+"]";
      if (srcTbl.containsValue(value))
      {
        Set<String> kset = srcTbl.keySet();  
        Iterator<String> i = kset.iterator();
        String key;
        while (i.hasNext())
        {
          key = i.next();
          if (srcTbl.get(key).compareToIgnoreCase(value) == 0)
            return new String[]{key};
        }
      }
    }
    String dname = new String("/" + context + "/" + server + "/" + property);
    try
    {
      if (isCentralArchiveServer(server) || isCentralArchiver(server))
      { // call to a central archive server
        String[] keysrcs = new String[100]; 
        dout = new TDataType(keysrcs);
        TLink tl = new TLink(dname,"KEYWORD.SOURCE",dout,null,TAccess.CA_READ);
        cc = tl.executeAndClose(timeout);
        if (cc == 0) 
        {
          int len =dout.getCompletionLength();
          if (len < 100) keysrcs = Arrays.copyOf(keysrcs, len);
          src = keysrcs;
          if (srcTbl != null && selectedContext != null && 
              context.compareToIgnoreCase(selectedContext) == 0)
          { // if there's a hash table going, add to it 
            String dev;
            String key = "keyword["+property+"]";
            String[] devs = null;
            if (keysrcs.length > 1)
            {
              devs = TQuery.getDeviceNames(context, server, property);
            }
            for (int k=0; k<keysrcs.length; k++)
            { // key is the src string (per device)
              if (keysrcs[k] == null) break;
              if (keysrcs.length > 1) 
              {
                dev = (devs == null || k >= devs.length) ? "#"+k : devs[k]; 
                key = dev+"["+property+"]";
              }
              srcTbl.put(keysrcs[k],key);
            }
          }
        }
      }
      else
      {
        src = new String[]{"/"+context+"/"+server+"/"+device+"["+property+"]"};
      }
    }
    catch (UnresolvedAddressException e)
    {
      cc = TErrorList.address_unresolved;
    }
    if (cc != 0) MsgLog.log("getArchiveSource", TErrorList.toString(cc),cc,null,1);
    
    return src;
  }
  public static synchronized FLTFLT getToleranceOverInterval(String context,String server, String property, double start, double stop,int timeout)
  {
    int cc = 0;
    int[] startstop = new int[2];
    TDataType dout, din;
    FLTFLT tols = new FLTFLT(0,0);
    
    if (server == null && context == null) return tols;
    if (context == null) context = "DEFAULT";
    if (server == null) server = "HISTORY"; // assume the standard central archive server for the given context
    startstop[0] = (int)start;
    startstop[1] = (int)Math.ceil(stop);
    
    String dname = new String("/" + context + "/" + server + "/" + property);
    try
    {
      if (isCentralArchiveServer(server))
      { // call to a central archive server
    	FLTFLT[] dtols = new FLTFLT[1];
        dtols[0] = tols;
        din = new TDataType(startstop);
        dout = new TDataType(dtols);
        TLink tl = new TLink(dname,"KEYWORD.TOLERANCE",dout,din,TAccess.CA_READ);
        cc = tl.executeAndClose(timeout);
      }
      else
      {
        THistoryRecordStruct[] hrs = new THistoryRecordStruct[1];
        hrs[0] = new THistoryRecordStruct();
        din = new TDataType(property);
        dout = new TDataType(hrs);
        TLink tl = new TLink(dname,"HISTORIES",dout,din,TAccess.CA_READ);
        cc = tl.executeAndClose(timeout);
        if (cc == 0)
        {
          tols.f1val = hrs[0].getAbsoluteTolerance();
          tols.f2val = hrs[0].getPercentTolerance();
        }
      }
    }
    catch (UnresolvedAddressException e)
    {
      cc = TErrorList.address_unresolved;
    }
    if (cc != 0) MsgLog.log("getToleranceOverInterval", TErrorList.toString(cc),cc,null,1);
    return tols;
  }
  public static double[] getToleratedUpperValues(double[] data,float tolAbsolute,float tolRelative)
  {
    if (data == null) return null;
    double[] d = new double[data.length];
    for (int i=0; i<data.length; i++)
    {
      d[i] = data[i]*(1.0 + tolRelative) + tolAbsolute;
    }
    return d;
  }
  public static double[] getToleratedLowerValues(double[] data,float tolAbsolute,float tolRelative)
  {
    if (data == null) return null;
    double[] d = new double[data.length];
    for (int i=0; i<data.length; i++)
    {
      d[i] = data[i]*(1.0 - tolRelative) - tolAbsolute;
    }
    return d;
  }
  public static int getArchivedStatus(String context,String server, String property, String device, double start, double stop,NAME32DBLDBL[] data)
  {
    return getStatusArray(context,server,property,device,start,stop,data,1000);
  }  
  public static int getArchivedStatus(String context,String server, String property, String device, double start, double stop,NAME32DBLDBL[] data,int timeout)
  {
    return getStatusArray(context,server,property,device,start,stop,data,timeout);
  }  
  private static synchronized int getStatusArray(String context,String server, String property, String device, double start, double stop, TCompoundDataObject[] data, int timeout)
  {
    if (context == null || property == null || data == null) return TErrorList.invalid_parameter;
    if (data.length < 1) return -TErrorList.dimension_error;
    if (timeout < 100) timeout = 100;
    int idx;
    String s1 = server;
    String p1 = property;
    if (s1 == null) s1 = "HISTORY";
    if ((idx=property.indexOf(".HIST")) >= 0 ||
        (idx=property.indexOf(".HST")) >= 0 ||
        (idx=property.indexOf(".AR")) >= 0)
    { // passed with .HIST appended, remove it!
      property = property.substring(0, idx);
    }
    if (!isCentralArchiveServer(s1)) 
    { // a specific server was passed
      p1 = property + ".ARCH"; // redirect to same channel at the central archiver
    }
    if (!(data instanceof NAME32DBLDBL[]))
    { // CF_NAME32DBLDBL and ONLY CF_NAME32DBLDBL
      return -TErrorList.illegal_format;
    }
    int nret = getArchivedDataArray(context,s1,p1,device,0,start,stop,data,timeout);
    if (nret == -TErrorList.illegal_format) nret = 0; // old archive server
    return nret;
  }
  /**
   * @param context is the context of the targeted central archive server 
   * @param property is the targeted archive keyword
   * @param start is a UTC double giving the start of the desired range
   * @param stop is a UTC double giving the end of the desired range
   * @param data is an array of UTC doubles giving the time stamps of any annotations 
   * for in the range given.  These timestamps can then be used in calls to getAnnotation().
   * @param timeout is the allowed duration of the annotation call before issuing
   * a timeout message.
   * @return 0 upon success or a TINE error code.
   */
  public static synchronized int getAnnotationTimes(String context, String property, double start, double stop, double[] data, int timeout)
  {
    return getAnnotationTimes(context, "HISTORY", property, start, stop, data, timeout);
  }
  /**
   * @param context is the context of the target server 
   * @param server is the targeted device server name.  If null then the 
   * central archive server is assumed.
   * @param property is the targeted archive property
   * @param start is a UTC double giving the start of the desired range
   * @param stop is a UTC double giving the end of the desired range
   * @param data is an array of UTC doubles giving the time stamps of any annotations 
   * for in the range given.  These timestamps can then be used in calls to getAnnotation().
   * @param timeout is the allowed duration of the annotation call before issuing
   * a timeout message.
   * @return 0 upon success or a TINE error code.
   */
  public static synchronized int getAnnotationTimes(String context, String server, String property, double start, double stop, double[] data, int timeout)
  {
    return getAnnotationTimes(context,server,property,start,stop,data,null,timeout);
  }
  /**
   * @param context is the context of the targeted central archive server 
   * @param property is the targeted archive keyword.
   * @param start is a UTC double giving the start of the desired range
   * @param stop is a UTC double giving the end of the desired range
   * @param data is an array of UTC doubles giving the time stamps of any annotations 
   * for in the range given.  These timestamps can then be used in calls to getAnnotation().
   * @param isActive is a boolean array whose contents determine whether the annotation in question
   * is an 'active' annotation or not.
   * @param timeout is the allowed duration of the annotation call before issuing
   * a timeout message.
   * @return 0 upon success or a TINE error code.
   */
  public static synchronized int getAnnotationTimes(String context, String property, double start, double stop, double[] data, boolean[] isActive, int timeout)
  {
    return getAnnotationTimes(context, "HISTORY", property, start, stop, data, isActive, timeout); 
  }
  /**
   * @param context is the context of the target server 
   * @param server is the targeted device server name.  If null then the 
   * central archive server is assumed.
   * @param property is the targeted archive property
   * @param start is a UTC double giving the start of the desired range
   * @param stop is a UTC double giving the end of the desired range
   * @param data is an array of UTC doubles giving the time stamps of any annotations 
   * for in the range given.  These timestamps can then be used in calls to getAnnotation().
   * @param isActive is a boolean array whose contents determine whether the annotation in question
   * is an 'active' annotation or not.
   * @param timeout is the allowed duration of the annotation call before issuing
   * a timeout message.
   * @return 0 upon success or a TINE error code.
   */
  public static synchronized int getAnnotationTimes(String context, String server, String property, double start, double stop, double[] data, boolean[] isActive, int timeout)
  {
    if (context == null || property == null || data == null) return -TErrorList.invalid_parameter;
    if (data.length < 1) return -TErrorList.dimension_error;
    if (server == null || server.length() == 0) server = "HISTORY";
    if (server.compareToIgnoreCase("ARCHIVER") == 0) server = "HISTORY"; // internal redirection to central archive HISTORY server
    boolean isCentralArchive = server.compareToIgnoreCase("HISTORY") == 0;
    if (timeout < 100) timeout = 100;
    INTINT[] iidata = null;
    int[] startstop = new int[2];
    startstop[0] = (int)start;
    startstop[1] = (int)Math.ceil(stop);
    TDataType din, dout;
    if (startstop[0] == 0 && startstop[1] == 0)
    {
      din = new TDataType();
    }
    else
    {
      din = new TDataType(startstop);
    }
    String dname, pname;
    if (isCentralArchive)
    {
      dname = "/"+context+"/"+server+"/"+property;
      pname = "RECORD.CMTS";
    }
    else
    {
      dname = "/"+context+"/"+server;
      pname = "HISTORY.CMTS";
      isActive = null; // local histories do not (yet) support live annotations
    }
    if (isActive != null && isActive.length >= data.length)
    {
      iidata = new INTINT[data.length];
      dout = new TDataType(iidata);
    }
    else
    {
      isActive = null;
      dout = new TDataType(data);
    }
    TLink tl = new TLink(dname,pname,dout,din,TAccess.CA_READ);
    int cc = tl.executeAndClose(timeout);
    if (cc != 0) return -cc;
    int n = dout.getCompletionLength();
    if (isActive != null)
    {
      for (int i=0; i<n; i++)
      {
        data[i] = iidata[i].i1val;
        isActive[i] = iidata[i].i2val != 0;
      }
    }
    else
    {
      if (n > 0) Arrays.sort(data,0,n);
    }
    return n;
  }
  /**
   * Gets an annotation from the central archive at a selected time
   * 
   * @param context is the context of the Central Archive Server 
   * @param property is the targeted record keyword name
   * @param target is the desired annotation time 
   * @param data is the annotation text as a string array
   * @param timeout is the allowed duration of the annotation call before issuing
   * a timeout message.
   * 
   * @return 0 upon success or a TINE error code
   */
  public static synchronized int getAnnotation(String context, String property, double target, String[] data, int timeout)
  {
    return getAnnotation(context, "HISTORY",property, target, data, timeout);
  }
  /**
   * Gets an annotation at a selected time
   * 
   * @param context is the context of the target server 
   * @param server is the targeted device server name.  If null then the 
   * central archive server is assumed.
   * @param property is the targeted archive property
   * @param target is the desired annotation time 
   * @param data is the annotation text as a string array
   * @param timeout is the allowed duration of the annotation call before issuing
   * a timeout message.
   * 
   * @return 0 upon success or a TINE error code
   */
  public static synchronized int getAnnotation(String context, String server,String property, double target, String[] data, int timeout)
  {
    if (context == null || property == null || data == null) return TErrorList.invalid_parameter;
    if (server == null || server.length() == 0) server = "HISTORY";
    if (server.compareToIgnoreCase("ARCHIVER") == 0) server = "HISTORY"; // internal redirection to central archive HISTORY server
    if (data.length < 1) return TErrorList.dimension_error;
    if (timeout < 100) timeout = 100;
    TDataType dout;
    char[] cmt = new char[1024];
    dout = new TDataType(cmt);

    boolean isCentralArchive = server.compareToIgnoreCase("HISTORY") == 0;
    String dname, pname;
    if (isCentralArchive)
    {
      dname = "/"+context+"/"+server+"/"+property+"@"+(int)target;
      pname = "RECORD.CMT";
    }
    else
    {
      dname = "/"+context+"/"+server+"/"+(int)target;
      pname = "HISTORY.CMT";      
    }
    TLink tl = new TLink(dname,pname,dout,null,TAccess.CA_READ);
    int cc = tl.executeAndClose(timeout);
    if (cc != 0) return cc;
    int len = dout.getCompletionLength();
    data[0] = new String(cmt,0,len);
    return 0;
  }
  /**
   * Annotates the designated context in the central archive at the current
   * time.
   * 
   * @param context is the context of the Central Archive Server 
   * @param text is the desired annotation text
   * 
   * @return 0 upon success or a TINE error code
   */
  public static synchronized int setAnnotation(String context, String text)
  {
    return setAnnotation(context,"State",0,text,false,1000);
  }
  /**
   * Annotates the designated record property in the central archive at the current
   * time.
   * 
   * @param context is the context of the Central Archive Server 
   * @param property is the targeted record keyword name
   * @param text is the desired annotation text
   * 
   * @return 0 upon success or a TINE error code
   */
  public static synchronized int setAnnotation(String context, String property, String text)
  {
    return setAnnotation(context,property,0,text,false,1000);
  }
  /**
   * Annotates the designated record property in the central archive
   * 
   * @param context is the context of the Central Archive Server 
   * @param property is the targeted record keyword name
   * @param targetTime is the desired annotation time (if <= 0 then the
   * current time will be used)
   * @param text is the desired annotation text
   * @param timeout is the allowed duration of the annotation call before issuing
   * a timeout message.
   * 
   * @return 0 upon success or a TINE error code
   */
  public static synchronized int setAnnotation(String context, String property, double targetTime, String text, int timeout)
  {
    return setAnnotation(context,property,targetTime,text,false,timeout);
  }
  /**
   * Annotates the designated record property in the central archive
   * 
   * @param context is the context of the Central Archive Server 
   * @param property is the targeted record keyword name
   * @param targetTime is the desired annotation time (if <= 0 then the
   * current time will be used)
   * @param text is the desired annotation text
   * @param keepAlive if 'true' this will flag the annotation as 'active'
   * (i.e. valid until cleared by the next annotation).  Only valid for central
   * archive annotations.
   * @param timeout is the allowed duration of the annotation call before issuing
   * a timeout message.
   * 
   * @return 0 upon success or a TINE error code
   */
  public static synchronized int setAnnotation(String context, String property, double targetTime, String text, boolean keepAlive, int timeout)
  {
    return setAnnotation(context, "HISTORY", property, targetTime, text, keepAlive, timeout);
  }
  /**
   * Annotates the designated record property in the central archive
   * 
   * @param context is the context of the Central Archive Server 
   * @param server is the targeted device server name.  If null then the 
   * central archive server is assumed.
   * @param property is the targeted record keyword name
   * @param targetTime is the desired annotation time (if <= 0 then the
   * current time will be used)
   * @param text is the desired annotation text
   * @param keepAlive if 'true' this will flag the annotation as 'active'
   * (i.e. valid until cleared by the next annotation)
   * @param timeout is the allowed duration of the annotation call before issuing
   * a timeout message.
   * 
   * @return 0 upon success or a TINE error code
   */
  public static synchronized int setAnnotation(String context, String server, String property, double targetTime, String text, boolean keepAlive, int timeout)
  {
    if (context == null || text == null) return TErrorList.invalid_parameter;
    if (server == null || server.length() == 0) server = "HISTORY";
    if (timeout < 100) timeout = 100;
    if (targetTime <= 0) targetTime = TDataTime.getDataTimeStamp();
    TDataType din = null;
    if (text != null && text.length() > 0)
    {
      din = new TDataType(text.toCharArray());
    }
    boolean isCentralArchive = server.compareToIgnoreCase("HISTORY") == 0;
    String keepString = keepAlive ? "[KEEPALIVE]" : "";
    String dname, pname;
    if (isCentralArchive)
    {
      if (property == null) property = "State";
      dname = "/"+context+"/"+server+"/"+property+"@"+(int)targetTime+keepString;
      pname = "RECORD.CMT";
    }
    else
    {
      dname = "/"+context+"/"+server+"/"+(int)targetTime;
      pname = "HISTORY.CMT";      
    }
    TLink tl = new TLink(dname,pname,null,din,TAccess.CA_WRITE);
    return tl.executeAndClose(timeout);
  }
  public enum ExclusiveHistorySource
  {
    none, central, local    
  }
  private static ExclusiveHistorySource exHstSrc = ExclusiveHistorySource.none;
  public static ExclusiveHistorySource getExclusiveHistorySource()
  {
    return exHstSrc;
  }
  public static void setExclusiveHistorySource(ExclusiveHistorySource value)
  {
    exHstSrc = value;
  }
  private static HashMap<String, TrendData> trendMap = new HashMap<String,TrendData>();
  /**
   * Gets the current trend of the targeted address.
   * 
   * The call will start an asynchronous listener to obtain the current value and append it to
   * the current trend data.  Thus, the call can be made repeatedly (logically, not more often than
   * the given update interval).  When the call is NOT made repeatedly then after 5 minutes of idle
   * time (i.e. without an additional call) then the listener will be stopped.
   * 
   * @param context is the targeted address context
   * @param server is the targeted address server
   * @param property is the property whose trend is desired
   * @param device is the targeted address device
   * @param index is the desired index (in case the target property represents an array)
   * @param fieldIndex is the desired field index (in case the target property is of a complex
   * or structured format)
   * @param depthInSeconds is the desired trend depth in seconds (e.g. 3600 for 1 hour, 86400 for 1 day)
   * @param interval is the desired update interval (passed to the asynchronous listener)
   * @param data is the data object to hold the trend data (must be of data type DBLDBL (i.e. DBLTIME) or 
   * DBLDBLDBL
   * @return the number of data points in the 'data' array element with non-zero values (maximum is the
   * array length of 'data')
   */
  public static int getTrendData(String context,String server, String property, String device, int index, int fieldIndex, int depthInSeconds, int interval, TCompoundDataObject[] data)
  {
    THistorySource src = new THistorySource();
    return getTrendData(context,server,property,device,index,fieldIndex,depthInSeconds,interval,data,src);    
  }
  public static int getTrendData(String context,String server, String property, String device, int index, int fieldIndex, int depthInSeconds, int interval, TCompoundDataObject[] data,THistorySource src)
  {
    int cc = 0;
    double stop = TDataTime.getDataTimeStamp();
    double start = stop - (double)depthInSeconds;
    TrendData trend = null;
    String key = "/"+context+"/"+server+"/"+device+"["+property+"]";
    int nmonths = (int)((stop - start)/2592000.0);
    int timeout = interval;
    if (nmonths > 1) timeout = Math.min(nmonths * 4000, 30000);
    trend = trendMap.get(key);
    if (trend == null || !trend.IsActive())
    { // get archive data 
      if (trend != null) trendMap.remove(key); // no longer active => remove it
      TCompoundDataObject[] tdata = data;
      if (TLinkFactory.isJddd())
      {
        tdata = new DBLTIME[2000];
        for (int i=0; i<2000; i++) tdata[i] = new DBLTIME();
      }
      trend = new TrendData(context,server,device,property,tdata,depthInSeconds,interval);
      trend.src = src;
      for (int i=0; i<2; i++)
      {
        cc = getDataArray(context,server,property,device,index,fieldIndex,start,stop,tdata,src,timeout);
        if (cc != TErrorList.not_posted && cc != TErrorList.connection_timeout) break;
      }
      if (cc == 0) trend.length = src.getNumberOfPointsReturned();
      trendMap.put(key, trend);
    }
    else
    {
      src.copyFrom(trend.src);
      long st = (long)start;
      long tst = trend.getTrendStart();
      if (Math.abs(tst - st) > 3600)
      { // sudden interest in an earlier time or another time frame ?
        TCompoundDataObject[] tdata = data;
        if (TLinkFactory.isJddd())
        {
          tdata = new DBLTIME[2000];
          for (int i=0; i<2000; i++) tdata[i] = new DBLTIME();
        }
        //System.out.println("getTrendData start changed from "+TDataTime.toString(tst*1000)+" to "+TDataTime.toString(st*1000));
        for (int i=0; i<2; i++)
        {
          cc = getDataArray(context,server,property,device,index,fieldIndex,start,stop,tdata,src,timeout);
          if (cc != TErrorList.not_posted && cc != TErrorList.connection_timeout) break;
        }
        if (cc == 0) 
        {
          trend.length = src.getNumberOfPointsReturned();
          trend.setTrendStart(st);
          trend.setDepthInSeconds(depthInSeconds);
          trend.setTrendData(tdata);
          trend.src = src;
        }
      }
      else if (st > tst)
      { // moving along ...
        trend.shift(st-tst); // shift seconds
      }
    }
    trend.touch();
    int npts = trend.length;
    if (trend.data != data) 
    { // different buffer provided as trend buffer ... 
      synchronized(trend.data) 
      { // don't get mixed up with an append or prune ...
        int pstart = trend.indexAt(start);
        int pstop = trend.indexAt(stop);
        npts = pstop - pstart;
        if (npts < 0) npts = 0;
        if (npts > data.length) 
        { // more points in trend than length of given buffer
          int raster = npts/data.length + (npts%data.length == 0 ? 0 : 1);
          npts /= raster;
          for (int i=0; i<npts; i++) data[i] = trend.data[i*raster]; 
        }
        else if (npts > 0)
        {
          System.arraycopy(trend.data, pstart, data, 0, npts);
        }
        if (npts < data.length) Arrays.fill(data, npts, data.length - 1, trend.nulldata);
      } 
    }
    return npts;
  }
  public static void removeTrend(String context,String server, String property, String device)
  {
    String key = "/"+context+"/"+server+"/"+device+"["+property+"]";
    TrendData trend = trendMap.get(key);
    if (trend == null) return;
    trendMap.remove(key);
  }
  private static String lastDataKey = null;
  private static NAME32DBLDBL[] ndd = null;
  private static double[] ats = null;
  private static boolean[] atsActive = null;
  private static synchronized int getDataArray(String context,String server, String property, String device, int index, double start, double stop, TCompoundDataObject[] data, THistorySource src, int timeout)
  {
    return getDataArray(context,server,property,device,index,-1,start,stop,data,src,timeout);
  }
  private static Boolean callTrend = false;
  private static boolean checkCallTrendDone = false;
  private static synchronized int getDataArray(String context,String server, String property, String device, int index, int fieldindex, double start, double stop, TCompoundDataObject[] data, THistorySource src, int timeout)
  {
    boolean jdddDebug = TLinkFactory.isJddd() && TLinkFactory.debugLevel > 0;
    synchronized (callTrend)
    {
      if (!checkCallTrendDone)
      {
        checkCallTrendDone = true;
        callTrend = TLinkFactory.isJddd();
      }
      boolean isTrendCall = stop >= (TDataTime.getDataTimeStamp() - 900.0);
      if (isTrendCall && callTrend)
      {
        callTrend = false;
        try
        {
          double now = TDataTime.getDataTimeStamp();
          int rng = (int)(now - start);  // range was: (stop - start)
          src.setNumberOfPointsReturned(getTrendData(context, server, property, device, index, 0, rng, 1000, data,src));
        } 
        catch (Exception x) 
        {
          x.printStackTrace();
        }
        callTrend = true;
        return 0;
      }
    }
    int idx, cc = 0, cc1=0, cc2=0;
    HstSrc[] srvs = new HstSrc[2];
    // s1, p1 will refer to the central archive repository
    // s2, p2 will refer to the local history at the source server
    String s1 = server;
    String p1 = property;
    if (s1 == null) s1 = "HISTORY";
    if ((idx=property.indexOf(".HIST")) >= 0 ||
        (idx=property.indexOf(".HST")) >= 0 ||
        (idx=property.indexOf(".AR")) >= 0)
    { // passed with .HIST appended, remove it!
      property = property.substring(0, idx);
    }
    String s2 = s1; // server 2nd source
    if (!isCentralArchiveServer(s1)) 
    { // a specific server was passed
      p1 = property + ".ARCH"; // redirect to same channel at the central archiver
    }
    else
    { // 2nd source must redirect back to find the target server's local histories
      s2 = "ARCHIVER"; 
    }
    double npstart, npstop;
    boolean extendedRange = (src != null && src.hasExtendedTimeRange());
    if (src != null && src.getHeartbeat() == 0)
    {
      src.setHeartbeat(getArchivedHeartbeat(context, server, property));
    }
    int hbt = src == null ? HEARTBEAT : src.getHeartbeat();
    // try to get the number of points in the 'desired' range correct:
    npstart = extendedRange ? start + hbt*2 : start;
    npstop = extendedRange ? stop - hbt*2 : stop;
    int npts1 = -TErrorList.not_applicable;
    if (exHstSrc != ExclusiveHistorySource.local)
      npts1 = getArchivedNumPointsInInterval(context, s1, p1, device, npstart, npstop, timeout);
    if (jdddDebug) System.out.println("num points from central : "+npts1);
    if (npts1 < 0) 
    { // a problem with this entry
    	cc1 = -npts1;
      npts1 = -1; // was 0
      lastDataKey = null;
      ndd = null;
      ats = null;
      atsActive = null;
    }
    else
    { // central archive is alive and knows about the record
      String thisDataKey = "/"+context+"["+property+"]"+start+"->"+stop;
      if (lastDataKey == null || thisDataKey.compareToIgnoreCase(lastDataKey) != 0)
      { // status info is only at the Central archiver       
        int ndays = (int)((stop - start)) / 86400;
        if (ndays < 1) ndays = 1;
        ndd = new NAME32DBLDBL[ndays*100];
        String stsprp = p1.endsWith(".ARCH") ? server+"."+property : property;
        int nr = getArchivedStatus(context, "HISTORY", stsprp, device, start, stop, ndd, timeout);
        ndd = nr <= 0 ? null : Arrays.copyOf(ndd, nr);
        ats = new double[ndays*10];
        atsActive = new boolean[ndays*10];
        nr = getAnnotationTimes(context, server, property, start, stop, ats, atsActive, timeout);
        ats = nr <= 0 ? null : Arrays.copyOf(ats, nr);
        atsActive = nr <= 0 ? null : Arrays.copyOf(atsActive, nr);
        lastDataKey = thisDataKey;
      }
    }
    srvs[0] = new HstSrc("Central Archive",s1,npts1);
    String p2 = property + ".HIST";
    String dev2 = getTargetDevice(context, server, device, property);
    int npts2 = -TErrorList.not_applicable;
    if (exHstSrc != ExclusiveHistorySource.central)
    { // local history: don't let the timeout be too big here !
      npts2 = getArchivedNumPointsInInterval(context, s2, p2, dev2, npstart, npstop, 2000);
    }
    boolean preferCentral = preferCentralSource;
    if (npts2 < 0 && !isCentralArchiver(s2)) 
    { // a problem with this entry
      if (npts2 == -TErrorList.un_allocated)
      {
        String[] hists = getArchivedProperties(context,s2);
        if (hists != null)
        { // there are histories available ....
          if (isScheduledProperty(property))
          { // look for unscheduled alternative
            for (String h : hists)
            {
              if (property.startsWith(h))
              {
                p2 = h+".HIST";
                npts2 = getArchivedNumPointsInInterval(context, s2, p2, dev2, npstart, npstop, 500);
                break;
              }
            }
          }
          else
          { // look for scheduled alternative
            for (String h : hists)
            {
              if (isScheduledProperty(h))
              {
                if (h.startsWith(property))
                {
                  p2 = h+".HIST";
                  npts2 = getArchivedNumPointsInInterval(context, s2, p2, dev2, npstart, npstop, 500);
                  break;             
                }
              }
            }        
          }
        }
      }
      if (npts2 < 0) 
      {
      	cc2 = -npts2;
      	npts2 = -1; // was 0
      }
    }
    if (jdddDebug) System.out.println("num points from local : "+npts2);
    srvs[1] = new HstSrc("Local History",s2,npts2);
    if (npts1 <= 0) preferCentral = false;
    boolean done = false;
    int nret = 0;
    boolean useCentral = preferCentral;
    if (npts1 > 500 || npts1 > npts2) useCentral = true; // central archive has 'enough' points or at least more than the local history
    if (preferLocalSource && npts2 > npts1) useCentral = false;
    boolean needBoundary = true;
    double dvstart = start;
    while (needBoundary)
    {
      if (!extendedRange) needBoundary = false;
      if (useCentral)
      { // central archive is requested
        nret = getArchivedDataArray(context,s1,p1,device,index,fieldindex,0,dvstart,stop,data,timeout);      
        if (nret > 0)
        { // everything's okay
          srvs[0].setAsDefault(true);
          srvs[0].setNumberReturned(nret);
          done = true;
        }
        else
        {
          if (jdddDebug) System.out.println("data from central archive returned "+nret);
          cc = -nret;
          cc1 = cc;
          nret = 0;
        }
      }
      if (!done && npts2 > 0)
      { // too few (or no) points upstairs, so try the local history ...
        if (jdddDebug) System.out.println("must use local history points !!!!!");
        if (dev2.startsWith("#") && !device.startsWith("#") && index == 0) index = -1;
        nret = getArchivedDataArray(context,s2,p2,dev2,index,fieldindex,0,dvstart,stop,data,timeout);
        srvs[1].setAsDefault(true);
        if (nret > 0)
        {
          srvs[1].setNumberReturned(nret);
          done = true;
        }
        else
        {
          cc = -nret;
          nret = 0;
        }
        if (cc != 0 || nret == 0)
        { // error (local or remote -> did not get back any points after all!)
          cc2 = cc;
          if (npts1 > 0)
          { // would have been a nice place for a 'goto' ....
            nret = getArchivedDataArray(context,s1,p1,device,index,fieldindex,0,dvstart,stop,data,timeout);      
            if (nret > 0)
            { // everything's okay
              srvs[1].setAsDefault(false);
              srvs[0].setAsDefault(true);
              srvs[0].setNumberReturned(nret);
              done = true;
            }
          }
        }
      }
      if (containsBoundary(data,npstart)) 
      { // we're good
        needBoundary = false;
      }
      if (needBoundary)
      { // adjust the data value start time ...
        dvstart -= (86400) * 7; // seconds in a week
      }
      if (start > dvstart + (86400 * 60))
      { // call it quits after looking 2 months back ...
        needBoundary = false;
      }
    }
    
    if (!done && npts2 <= 0 && 
        !(data instanceof INTFLTINT[]) &&
        !isCentralArchiveServer(server) &&
        !isCentralArchiver(server) &&
        exHstSrc != ExclusiveHistorySource.central)
    { // last desperate attempt (possible legacy doocs server ?)
      int ifi_len = data.length;
      if (ifi_len < 2000) ifi_len = 2000;
      INTFLTINT[] ifi = new INTFLTINT[ifi_len];
      int rst = 1, dn = 0, n = 0;
      double interim = 0;
      while (start < stop)
      {
        nret = getArchivedDataArray(context,s2,p2,device,index,start,stop,ifi,timeout);
        if (nret > 0)
        {
          if (dn == 0)
          {
            dn = (ifi[nret-1].i1val-(int)start);
            if (dn <= 0) dn = 1;
            rst = (int)(stop-start) / dn;
            if (rst < 1) rst = 1;
          }
          for (int i=0; n<data.length && i<nret; i += rst)
          {
            if (ifi[i] == null)
            {
              nret = i;
              break;
            }
            if (data instanceof FLTINT[])
              ((FLTINT[])data)[n] = new FLTINT(ifi[i].fval,ifi[i].i1val);
            else if (data instanceof DBLTIME[])
              ((DBLTIME[])data)[n] = new DBLTIME((double)ifi[i].fval,(double)ifi[i].i1val);
            else if (data instanceof DBLDBL[])
              ((DBLDBL[])data)[n] = new DBLDBL((double)ifi[i].fval,(double)ifi[i].i1val);
            n++;
          }
          interim = (double)(ifi[nret-1].i1val + 1);
          if (interim <= start) break;
          start = interim;
          done = true;
          if (rst == 1) break;
        }
        else
        {
          nret = 0;
          break;
        }
      }
      srvs[1].setAsDefault(true);
      srvs[1].setNumberPoints(n);
      srvs[1].setNumberReturned(n);
    }
    if (src != null)
    { // was passed, fill it in ...
      src.CentralArchive = srvs[0];
      src.LocalHistory = srvs[1];
      src.setNumberOfPointsReturned(nret);
      src.setStatusList(ndd);
      src.setAnnotationTimes(ats);
      src.setAnnotationActive(atsActive);
    }
    if (nret == 0)
    {
    	if (cc1 == TErrorList.connection_timeout || cc1 == TErrorList.link_timeout) 
    		cc = cc1; 
    	else if (cc2 != 0 && cc2 != TErrorList.un_allocated) 
    		cc = cc2;
    	else if (cc2 == 0)
    		cc = 0;
    	else
    		cc = cc1;
    }
    return cc;
  }
  private static boolean containsBoundary(TCompoundDataObject[] d, double boundaryStart)
  {
    if (d == null || d.length == 0 || d[0] == null) return false;
    double t = 0.0;
    if (d instanceof DBLDBL[])
    {
      t = ((DBLDBL[])d)[0].d2val;
    }
    else if (d instanceof DBLDBLDBL[])
    {
      t = ((DBLDBLDBL[])d)[0].d2val;
    }
    else if (d instanceof DBLTIME[])
    {
      t = ((DBLTIME[])d)[0].d2val;
    }
    else if (d instanceof NAME64DBL[])
    {
      t = ((NAME64DBL[])d)[0].dval;
    }
    else if (d instanceof NAME64DBLDBL[])
    {
      t = ((NAME64DBLDBL[])d)[0].d1val;
    }
    else
    {
      return true; // punt
    }
    return t > 0 && t <= boundaryStart;
  }
  private static synchronized int getArchivedDataArray(String context,String server, String property, String device, int index, double start, double stop, TCompoundDataObject[] data, int timeout)
  {
    return getArchivedDataArray(context,server,property,device,index,0,start,stop,data,timeout);
  }
  private static HashMap<String, String[]> dcsHsts = new HashMap<String, String[]>();
  private static boolean doocsSrvHasTineHst(String context, String server, String property)
  {
  	if (context == null || server == null || property == null) return false;
  	String key = "/"+context+"/"+server;
  	if (dcsHsts.containsKey(key)) 
  	{
  		for (String s : dcsHsts.get(key)) if (s.compareToIgnoreCase(property) == 0) return true;
  		return false;
  	}
  	try
  	{
	  	NAME64[] hprps = new NAME64[32];
	  	TDataType dt = new TDataType(hprps);
	  	TLink lnk = new TLink("/"+context+"/"+server,"HISTORIES",dt,null,TAccess.CA_READ);
	  	if (lnk.executeAndClose() != 0) return false;
	  	dcsHsts.put(key, NameToString.nameArrayToString(hprps));
	  	for (int i=0; i<dt.getCompletionLength(); i++)
	  	{
	  		if (hprps[i].getName().compareToIgnoreCase(property) == 0) return true;
	  	}
  	} catch (Exception x) {}
  	return false;
  }
  public static String getTargetDevice(String context, String server, String device, String property)
  {
    String ctx = context, srv = server, prp = property, dev = device;
    if (TLinkFactory.isRedirected(context, server, device, property))
    {
      RedirectedItem ri = TLinkFactory.getRedirectionInformation(context, server, device, property);
      ctx = ri.getDstContext();
      srv = ri.getDstServer();
      prp = ri.getDstProperty();
      dev = ri.getDstDevice();
    }
    int idx = -1;
    if ((idx=prp.lastIndexOf(".HIST")) > 0 || (idx=prp.lastIndexOf(".HST")) > 0)
    {
      prp = prp.substring(0, idx);
    }
    if (TSrvEntry.isDoocsSrv(ctx, srv) && doocsSrvHasTineHst(ctx,srv,prp))
    {
      String[] mems = TQuery.getDeviceGroupMembers(ctx, srv, dev, prp, 500);
      if (mems != null)
      {
        for (int i=0; i<mems.length; i++) 
          if (mems[i].compareToIgnoreCase(dev) == 0)
        {
          dev = "#"+i;
          break;
        }
      }
    }
    return dev;
  }
  private static synchronized int getArchivedDataArray(String context,String server, String property, String device, int index, int sample, double start, double stop, TCompoundDataObject[] data, int timeout)
  {
  	return getArchivedDataArray(context,server,property,device,index,-1,sample,start,stop,data,timeout);
  }
  private static int makeStartStopInput(double start,double stop,int index,int fieldindex,int sample,boolean iscentral,int[] startstop,double[] dstartstop)
  {
    int inptLength = 0;
    if (startstop != null && startstop.length >= 8)
    {
      startstop[0] = (int)start;
      startstop[1] = (int)Math.ceil(stop);
      startstop[2] = index;
      startstop[3] = sample;
      inptLength = (index == 0 && sample == 0) ? 2 : 4;
      if (fieldindex >= 0)
      {
        if (iscentral)
        {
          startstop[4] = fieldindex;  
          inptLength = 5;
        }
        else
        {
          startstop[6] = startstop[5] = startstop[4] = 0;
          startstop[7] = fieldindex;
          inptLength = 8;
        }
      }
    }
    if (dstartstop != null && dstartstop.length >= 8)
    {
      dstartstop[0] = start;
      dstartstop[1] = stop;
      dstartstop[2] = (double)index;
      dstartstop[3] = (double)sample;
      inptLength = (index == 0 && sample == 0) ? 2 : 4;
      if (fieldindex >= 0)
      {
        if (iscentral)
        {
          dstartstop[4] = (double)fieldindex;  
          inptLength = 5;
        }
        else
        {
          dstartstop[6] = dstartstop[5] = dstartstop[4] = 0;
          dstartstop[7] = (double)fieldindex;
          inptLength = 8;
        }
      }
    }
    return inptLength;
  }
  private static synchronized int getArchivedDataArray(String context,String server, String property, String device, int index, int fieldindex, int sample, double start, double stop, TCompoundDataObject[] data, int timeout)
  {
    int cc = 0, inptLength = 4;
    int[] startstop = new int[8];
    TDataType dout, din;

    if (server == null && context == null) return -TErrorList.argument_list_error;
    if (data == null) return -TErrorList.argument_list_error;
    if (context == null) context = "DEFAULT";
    if (server == null) server = "HISTORY"; // assume the standard central archive server for the given context
    if (timeout < 100) timeout = 100;
    if (device.startsWith("#"))
    {
      if (index == 0)
      {
        try
        {
          index = Integer.parseInt(device.substring(1));
          device = "#0";
        }
        catch (Exception x) { }
      }
      if (index < 0) index = 0;
    }
    inptLength = makeStartStopInput(start,stop,index,fieldindex,sample,isCentralArchiveServer(server),startstop,null);
    if (startstop[0] == 0 && startstop[1] == 0)
    {
      din = new TDataType();
    }
    else if (data instanceof INTFLTINT[])
    {
      WINDOW[] twnd = new WINDOW[1];
      twnd[0] = new WINDOW(startstop[0],startstop[1],0,0);
      din = new TDataType(twnd);
    }
    else
    {
      din = new TDataType(startstop,inptLength);
    }

    dout = new TDataType(data);
    
    if (isCentralArchiver(server))
    { // trap a potential redirection problem here:
      if (device.compareToIgnoreCase("keyword") == 0) device = "#0";
    }
    	
    String pname = new String(property);
    String rname = getQueryPropertyRoot(property);
    boolean tryAlternateProperty = false;
    if (!isCentralArchiveServer(server))
    { // not a call to a central archive server
      device = getTargetDevice(context,server,device,property);
      tryAlternateProperty = true;
      if (!isArchiveQueryProperty(property))
        pname = pname + ".HIST"; // meta extension not given, try this ...
    }
    String dname = new String("/" + context + "/" + server + "/" + device);
    TLink tl = new TLink(dname,pname,dout,din,TAccess.CA_READ);
    cc = tl.execute(timeout);
    long tnow = System.currentTimeMillis();
    while (cc == TErrorList.operation_busy)
    { // from a multi-threaded server ? ...
      timeout = getSuggestedTimeoutMilliseconds(start,stop);
      TLinkFactory.systemWait(10);
      cc = tl.execute(timeout);
      if (System.currentTimeMillis() > tnow + timeout) break;
    }
    tl.close();
    if ((cc == TErrorList.un_allocated || cc == TErrorList.device_not_connected) && tryAlternateProperty)
    {
      String[] hists = getArchivedProperties(context,server);
      if (hists != null)
      { // there are histories available ....
        if (isScheduledProperty(rname))
        { // look for unscheduled alternative
          for (String h : hists)
          {
            if (rname.startsWith(h))
            {
              pname = h+".HIST";
              tl = new TLink(dname,pname,dout,din,TAccess.CA_READ);
              cc = tl.executeAndClose(timeout);
              break;
            }
          }
        }
        else
        { // look for scheduled alternative
          for (String h : hists)
          {
            if (isScheduledProperty(h))
            {
              if (h.startsWith(rname))
              {
                pname = h+".HIST";
                tl = new TLink(dname,pname,dout,din,TAccess.CA_READ);
                cc = tl.executeAndClose(timeout);
                break;             
              }
            }
          }        
        }
      }
    }
    int npts = cc == 0 ? dout.getCompletionLength() : 0;
    if (npts == 0 && isCentralArchiver(server))
    { // special 2nd chance !
      if (isArchiveQueryProperty(property))
      {
        pname = rname + ".ARCH";
      }
      else
      {
        pname = new String(property) + ".ARCH";
      }
      tl = new TLink(dname,pname,dout,din,TAccess.CA_READ);
      cc = tl.executeAndClose(timeout);
    }
    return cc == 0 ? npts : -cc;
  }  
  private static int getArchivedDataArray(String context,String server, String property, String device, int index, double start, double stop,TCompoundDataObject[] data)
  {
    return getArchivedDataArray(context,server,property,device,index,start,stop,data,2000);
  }
  private static synchronized int getArchivedDataSnapshot(String context,String server, String property, String device, double[] targetTime,float[] data,THistorySource src)
  {
    return getArchivedDataSnapshot(context,server,property,device,targetTime,data,-1,src);
  }
  private static synchronized int getArchivedDataSnapshot(String context,String server, String property, String device, double[] targetTime,float[] data,int fieldindex,THistorySource src)
  {
    int cc = 0, inptLength = 2;
    double[] dstartstop = new double[8];
    TDataType dout, din;

    if (server == null && context == null) return TErrorList.argument_list_error;
    if (targetTime == null || data == null) return TErrorList.argument_list_error;
    if (context == null) context = "DEFAULT";
    if (server == null) server = "HISTORY"; // assume the standard central archive server for the given context
    inptLength = makeStartStopInput(targetTime[0],targetTime[0],0,fieldindex,0,isCentralArchiveServer(server),null,dstartstop);
    din = new TDataType(dstartstop,inptLength);
    dout = new TDataType(data);
    if (device == null || device.trim().length() == 0) device = "#0";
    String dname = new String("/" + context + "/" + server + "/" + device);
    String[] pname = new String[2];
    pname[0] = new String(property);
    int ntries = 1;
    if (!server.endsWith("HISTORY"))
    { // not a call to a central archive server
      if (!isArchiveQueryProperty(property))
      { // meta extension not given, try this ...
        pname[0] = property + ".HIST";
        pname[1] = property + ".ARCH";
        ntries = 2;
      }
    }
    for (int i=0; i<ntries; i++)
    {
      TLink tl = new TLink(dname,pname[i],dout,din,TAccess.CA_READ);
      cc = tl.execute(2000, true);
      targetTime[0] = tl.getLastDataTimeStamp();
      tl.close();
      if (cc == 0 && dout.getCompletionLength() > 0) break;
    }
    if (src != null) {
    	src.setNumberOfPointsReturned(dout.getCompletionLength());
    }
    return cc;
  }  
  private static synchronized int getArchivedDataSnapshot(String context,String server, String property, String device, double[] targetTime,Object data, boolean image,THistorySource src)
  {
    return getArchivedDataSnapshot(context,server,property,device,targetTime,data,-1,image,src);
  }  
  private static synchronized int getArchivedDataSnapshot(String context,String server, String property, String device, double[] targetTime,Object data, int fieldindex,boolean image,THistorySource src)
  {
    int cc = 0, inptLength=2;
    double[] dstartstop = new double[8];
    TDataType dout, din;

    if (server == null && context == null) return TErrorList.argument_list_error;
    if (targetTime == null || data == null) return TErrorList.argument_list_error;
    if (targetTime.length < 1) return TErrorList.argument_list_error;
    if (context == null) context = "DEFAULT";
    if (server == null) server = "HISTORY"; // assume the standard central archive server for the given context
    boolean iscentral = isCentralArchiveServer(server);
    inptLength = makeStartStopInput(targetTime[0],targetTime[0],0,fieldindex,0,iscentral,null,dstartstop);
    din = new TDataType(dstartstop,inptLength);
    if (image) 
    {
      dout = new TDataType((IMAGE)data);
    } 
    else if (data instanceof String[]) 
    {
      dout = new TDataType((String[])data);
    }
    else if (data instanceof TCompoundDataObject[]) 
    {
      dout = new TDataType((TCompoundDataObject[])data);
    }
    else if (data instanceof int[]) 
    {
      dout = new TDataType((int[])data);
    }
    else if (data instanceof short[]) 
    {
      dout = new TDataType((short[])data);
    }
    else if (data instanceof float[]) 
    {
      dout = new TDataType((float[])data);
    }
    else if (data instanceof double[]) 
    {
      dout = new TDataType((double[])data);
    }
    else if (data instanceof long[]) 
    {
      dout = new TDataType((long[])data);
    }
    else
    {
      return TErrorList.illegal_format;
    }
    
    String dname = new String("/" + context + "/" + server + "/" + device);
    String[] pname = new String[2];
    pname[0] = new String(property);
    int ntries = 1;
    if (!iscentral)
    { // not a call to a central archive server
      boolean isqp = isArchiveQueryProperty(property);
      if (exHstSrc == ExclusiveHistorySource.central)
      {
        if (!isqp)
        {
          pname[0] = property + ".ARCH";
        }
        else if (property.endsWith(".HIST"))
        {
          return TErrorList.not_applicable;
        }
      }
      else if (!isqp)
      { // meta extension not given, try this ...
        pname[0] = property + ".HIST";
        pname[1] = property + ".ARCH";
        ntries = 2;
      }
    }
    for (int i=0; i<ntries; i++)
    {
      TLink tl = new TLink(dname,pname[i],dout,din,TAccess.CA_READ);
      cc = tl.executeAndClose(2000);
      targetTime[0] = tl.getLastDataTimeStamp();
      if (targetTime.length > 1) targetTime[1] = dout.getSystemDataStamp();
      if (cc == 0 && dout.getCompletionLength() > 0) break;
    }
    if (src != null) src.setNumberOfPointsReturned(dout.getCompletionLength());
    return cc;
  }  
  
  /**
   * Retrieves an archive data set.
   * 
   * Retrieve an archive data set (value - timestamp pairs) for the target property and device
   * given and over the time range specified. If server is null, an empty string, or "HISTORY",
   * the central archive server is queried for the data. Otherwise the local history 
   * subsystem of the device server specified is queried.  
   * 
   * @param context is the desired context  
   * @param server is the device server which manages the archived parameter. If null
   * then the central archive will be queried for the property and device
   * @param property is the desired archived property
   * @param device is the desired device name
   * @param start is the start time in seconds (UTC double, a TINE timestamp) 
   * @param stop is the stop time in seconds (UTC double, a TINE timestamp)
   * @param data is a reference to a (doublet) data array object to contain the returned data.
   * This can be one of FLTINT[], DBLDBL[], NAME64I[], and so on.  The array should be 
   * dimensioned to the maximum number of points desired (effectively determining a raster
   * over which the archived data will be distributed.  
   * The reference object can also be an array of HISTORY[] object instances, which have been
   * constructed with the desired data type (usually a compound data type) to be returned.    
   * The number of points returned in the call will often be less than array dimension
   * so one should pay attention to the return value of the call.
   * 
   * @return if positive, the number of points returned in the archive call; if negative,
   * the negative of a TINE error code.
   */
  public static int getArchivedData(String context,String server, String property, String device, double start, double stop,TCompoundDataObject[] data)
  {
    return getArchivedDataArray(context,server,property,device,0,start,stop,data);
  }
  public static int getArchivedData(String context,String server, String property, String device, double start, double stop,DBLDBL[] data)
  {
    return getArchivedDataArray(context,server,property,device,0,start,stop,data);
  }
  /**
   * Retrieves an archive data set.
   * 
   * Retrieve an archive data set (value - timestamp pairs) for the target property and device
   * given and over the time range specified.  Both the central archive and local archive are 
   * queried for the number of stored points over the interval. Data are usually
   * retrieved from the central archive system if there are at least 500 points over the
   * time interval in question.  Otherwise the local archive system is used.  By examining
   * the THistorySource object one can see which source was chosen and how many points both
   * sources have over the interval.  For dedicated retrieval from one source or the
   * other, please use an overloaded method which does not take the THistorySource parameter!
   * 
   * @param context is the desired context  
   * @param server is the device server which manages the archived parameter. If null
   * then the central archive will be queried for the property and device
   * @param property is the desired archived property
   * @param device is the desired device name
   * @param start is the start time in seconds (UTC double, a TINE timestamp) 
   * @param stop is the stop time in seconds (UTC double, a TINE timestamp)
   * @param data is a reference to a (doublet) data array object to contain the returned data.
   * This can be one of FLTINT[], DBLDBL[], NAME64I[], and so on.  The array should be 
   * dimensioned to the maximum number of points desired (effectively determining a raster
   * over which the archived data will be distributed. 
   * The reference object can also be an array of HISTORY[] object instances, which have been
   * constructed with the desired data type (usually a compound data type) to be returned.    
   * The src parameter will give the
   * exact number of points actually stored over the interval specified.
   * The number of points returned in the call will often be less than array dimension and
   * less than the exact number of points stored over the interval. 
   * @param src is a reference to a THistorySource object to receive the history source information.
   * @param timeout gives the number of milliseconds to wait for the call to complete 
   * (usually 1000 is sufficient).
   * @return 0 upon success or a TINE error code.
   */
  public static int getArchivedData(String context,String server, String property, String device, double start, double stop,TCompoundDataObject[] data,THistorySource src,int timeout)
  {
    return getDataArray(context,server,property,device,0,start,stop,data,src,timeout);
  }
  public static int getArchivedData(String context,String server, String property, String device, double start, double stop,DBLDBL[] data,THistorySource src,int timeout)
  {
    return getDataArray(context,server,property,device,0,start,stop,data,src,timeout);
  }
  public static int getArchivedData(String context,String server, String property, String device, double start, double stop,TCompoundDataObject[] data,THistorySource src,int timeout,boolean extendedRange)
  {
    return getArchivedData(context,server,property,device,0,start,stop,data,src,timeout,extendedRange);
  }
  public static int getArchivedData(String context,String server, String property, String device, double start, double stop,DBLDBL[] data,THistorySource src,int timeout,boolean extendedRange)
  {
    return getArchivedData(context,server,property,device,0,start,stop,data,src,timeout,extendedRange);
  }
  public static int getArchivedData(String context,String server, String property, String device, int index, double start, double stop,TCompoundDataObject[] data,THistorySource src,int timeout,boolean extendedRange)
  {  	
  	return getArchivedData(context,server,property,device,index,-1,start,stop,data,src,timeout,extendedRange);
  }
  /**
   * Retrieves an archive data set.
   * 
   * Retrieve an archive data set (value - timestamp pairs) for the target property and device
   * given and over the time range specified. If server is null, an empty string, or "HISTORY",
   * the central archive server is queried for the data. Otherwise the local history 
   * subsystem of the device server specified is queried.  
   * 
   * @param context is the desired context  
   * @param server is the device server which manages the archived parameter. If null
   * then the central archive will be queried for the property and device
   * @param property is the desired archived property
   * @param device is the desired device name
   * @param index is the desired array element index (if the data source represents an array)
   * @param fieldindex is the desired data type field index (if the stored data is a compound data type
   * or a structure.
   * @param start is the start time in seconds (UTC double, a TINE timestamp) 
   * @param stop is the stop time in seconds (UTC double, a TINE timestamp)
   * @param data is a reference to a (doublet) data array object to contain the returned data.
   * This can be one of FLTINT[], DBLDBL[], NAME64I[], and so on.  The array should be 
   * dimensioned to the maximum number of points desired (effectively determining a raster
   * over which the archived data will be distributed.  
   * The reference object can also be an array of HISTORY[] object instances, which have been
   * constructed with the desired data type (usually a compound data type) to be returned.    
   * The number of points returned in the call will often be less than array dimension
   * so one should pay attention to the return value of the call.
   * @param src is the THistorySource object reference, which gives information about the history source used
   * in accessing and returning data.
   * @param timeout is the amount of time (milliseconds) to wait for the history call to complete.
   * @param extendedRange if 'true' adds an amount of time (twice the archive heartbeat) to 
   * each end of the given time range. This will usually ensure that a plotted trend will be able
   * to extend itself to the boundaries of a graphical display. 
   * 
   * @return if positive, the number of points returned in the archive call; if negative,
   * the negative of a TINE error code.
   */
  public static int getArchivedData(String context,String server, String property, String device, int index, int fieldindex, double start, double stop,TCompoundDataObject[] data,THistorySource src,int timeout,boolean extendedRange)
  {
    boolean lockAccess = false;
    boolean jdddDebug = false;
    ReentrantLock alck = new ReentrantLock();
    String tid = "";
    if (TLinkFactory.isJddd()) 
    {
      jdddDebug = TLinkFactory.debugLevel > 0;
      if (jdddDebug)
      {
        long tnow = System.currentTimeMillis();
        tid = new String("thread "+Thread.currentThread().getId()+ " (" + Thread.currentThread().getName()+")");
        System.out.println( tid +": jddd calling getArchivedData @ "+TDataTime.toString(tnow)+
            " for "+property+"("+device+") with start = "+TDataTime.toString(start)+
            ", stop = "+TDataTime.toString(stop));
      }
      lockAccess = true;
      int nmonths = (int)((stop - start)/2592000.0);
      if (nmonths > 1) timeout = Math.min(nmonths * 4000, 30000);
      if (jdddDebug) System.out.println("timeout in "+timeout+" msecs");
    }
    if (extendedRange)
    {
      int hbt = getArchivedHeartbeat(context, server, property);
      if (hbt > 3600) hbt = 3600;
      if (jdddDebug) System.out.println(tid +": apply heartbeat of "+hbt+" seconds to start and stop");
      start -= hbt*2;
      stop  += hbt*2;
      if (src != null) 
      {
        src.setHeartbeat(hbt);
        src.setExtendedTimeRange(true);
      }
    }
    double tnow = (double)System.currentTimeMillis()/1000.0;
    if (stop > tnow) stop = tnow;
    if (lockAccess) alck.lock();
    int rc = getDataArray(context,server,property,device,index,fieldindex,start,stop,data,src,timeout);
    if (lockAccess) alck.unlock();
    if (jdddDebug) 
      System.out.println(tid +": getDataArray returned status "+rc+" and "+
         src.getNumberOfPointsReturned()+" points from "+(src.CentralArchive.isDefault() ? "central":"local"));
    return rc;
  }
  /**
   * Retrieves an archive data set.
   * 
   * Retrieve an archive data set (value - timestamp pairs) for the target property and device
   * given and over the time range specified. If server is null, an empty string, or "HISTORY",
   * the central archive server is queried for the data. Otherwise the local history 
   * subsystem of the device server specified is queried.  
   * 
   * @param context is the desired context  
   * @param server is the device server which manages the archived parameter. If null
   * then the central archive will be queried for the property and device
   * @param property is the desired archived property
   * @param device is the desired device name
   * @param index is the desired array element index (if the data source represents an array)
   * @param start is the start time in seconds (UTC double, a TINE timestamp) 
   * @param stop is the stop time in seconds (UTC double, a TINE timestamp)
   * @param data is a reference to a (doublet) data array object to contain the returned data.
   * This can be one of FLTINT[], DBLDBL[], NAME64I[], and so on.  The array should be 
   * dimensioned to the maximum number of points desired (effectively determining a raster
   * over which the archived data will be distributed.  
   * The reference object can also be an array of HISTORY[] object instances, which have been
   * constructed with the desired data type (usually a compound data type) to be returned.    
   * The number of points returned in the call will often be less than array dimension
   * so one should pay attention to the return value of the call.
   * 
   * @return if positive, the number of points returned in the archive call; if negative,
   * the negative of a TINE error code.
   */
  public static int getArchivedData(String context,String server, String property, String device, int index,double start, double stop,TCompoundDataObject[] data)
  {
    return getArchivedDataArray(context,server,property,device,index,start,stop,data);
  }
  /**
   * Retrieves an archive data set.
   * 
   * Retrieve an archive data set (value - timestamp pairs) for the target property and device
   * given and over the time range specified. If server is null, an empty string, or "HISTORY",
   * the central archive server is queried for the data. Otherwise the local history 
   * subsystem of the device server specified is queried.  
   * 
   * @param context is the desired context  
   * @param server is the device server which manages the archived parameter. If null
   * then the central archive will be queried for the property and device
   * @param property is the desired archived property
   * @param device is the desired device name
   * @param index is the desired array element index (if the data source represents an array)
   * @param start is the start time in seconds (UTC double, a TINE timestamp) 
   * @param stop is the stop time in seconds (UTC double, a TINE timestamp)
   * @param data is a reference to a (doublet) data array object to contain the returned data.
   * This can be one of FLTINT[], DBLDBL[], NAME64I[], and so on.  The array should be 
   * dimensioned to the maximum number of points desired (effectively determining a raster
   * over which the archived data will be distributed.  
   * The reference object can also be an array of HISTORY[] object instances, which have been
   * constructed with the desired data type (usually a compound data type) to be returned.    
   * The number of points returned in the call will often be less than array dimension
   * so one should pay attention to the return value of the call.
   * @param timeout gives the number of milliseconds to wait for the call to complete 
   * (usually 1000 is sufficient).
   * 
   * @return if positive, the number of points returned in the archive call; if negative,
   * the negative of a TINE error code.
   */
    public static int getArchivedData(String context,String server, String property, String device, int index,double start, double stop,TCompoundDataObject[] data,int timeout)
    {
      return getArchivedDataArray(context,server,property,device,index,start,stop,data,timeout);
    }
  /**
   * Retrieves an archive data set.
   * 
   * Retrieve an archive data set (value - timestamp pairs) for the target property and device
   * given and over the time range specified. If server is null, an empty string, or "HISTORY",
   * the central archive server is queried for the data. Otherwise the local history 
   * subsystem of the device server specified is queried.  
   * 
   * @param context is the desired context  
   * @param server is the device server which manages the archived parameter. If null
   * then the central archive will be queried for the property and device
   * @param property is the desired archived property
   * @param device is the desired device name
   * @param start is the start time in seconds (UTC double, a TINE timestamp) 
   * @param stop is the stop time in seconds (UTC double, a TINE timestamp)
   * @param sample is the desired raster sample.  Passing a '0' will instruct
   * the targeted server to select a raster which will best fill in the 
   * desired time range.  If (e.g.) a '1' is passed then no raster will be
   * used.  This means that the desired 'stop' time may not have been reached.
   * Thus the returned data need to be examined in order to determine whether the
   * call should be repeated (after adjusting the 'start' time). 
   * @param data is a reference to a HISTORY[] data array object to contain the returned data.
   * This object passed should be an array of HISTORY instances, which have been
   * constructed with the desired data type (usually a compound data type) to be
   * returned. The array should be 
   * dimensioned to the maximum number of points desired (effectively determining a raster
   * over which the archived data will be distributed.  
   * The number of points returned in the call will often be less than array dimension
   * so one should pay attention to the return value of the call.
   * @param timeout gives the number of milliseconds to wait for the call to complete 
   * (usually 1000 is sufficient).
   * 
   * @return if positive, the number of points returned in the archive call; if negative,
   * the negative of a TINE error code.
   */
  public static int getArchivedData(String context,String server, String property, String device, int index,int sample,double start, double stop,TCompoundDataObject[] data,int timeout)
  {
    return getArchivedDataArray(context,server,property,device,index,sample,start,stop,data,timeout);
  }
  /**
   * Retrieves an archive data set.
   * 
   * Retrieve an archive data set (value - timestamp pairs) for the target property and device
   * given and over the time range specified. If server is null, an empty string, or "HISTORY",
   * the central archive server is queried for the data. Otherwise the local history 
   * subsystem of the device server specified is queried.  
   * 
   * @param context is the desired context  
   * @param server is the device server which manages the archived parameter. If null
   * then the central archive will be queried for the property and device
   * @param property is the desired archived property
   * @param device is the desired device name
   * @param start is the start time in seconds (UTC double, a TINE timestamp) 
   * @param stop is the stop time in seconds (UTC double, a TINE timestamp)
   * @param sample is the desired raster sample.  Passing a '0' will instruct
   * the targeted server to select a raster which will best fill in the 
   * desired time range.  If (e.g.) a '1' is passed then no raster will be
   * used.  This means that the desired 'stop' time may not have been reached.
   * Thus the returned data need to be examined in order to determine whether the
   * call should be repeated (after adjusting the 'start' time). 
   * @param data is a reference to a HISTORY[] data array object to contain the returned data.
   * This object passed should be an array of HISTORY instances, which have been
   * constructed with the desired data type (usually a compound data type) to be
   * returned. The array should be 
   * dimensioned to the maximum number of points desired (effectively determining a raster
   * over which the archived data will be distributed.  
   * The number of points returned in the call will often be less than array dimension
   * so one should pay attention to the return value of the call.
   * @param timeout gives the number of milliseconds to wait for the call to complete 
   * (usually 1000 is sufficient).
   * @param extendedRange if 'true' signals that the call should begin searching for points 2
   * heartbeat intervals prior to the given start time and end 2 heartbeat intervals past the
   * given stop time.
   * 
   * @return if positive, the number of points returned in the archive call; if negative,
   * the negative of a TINE error code.
   */
  public static int getArchivedData(String context,String server, String property, String device, int index,int sample,double start, double stop,TCompoundDataObject[] data,int timeout, boolean extendedRange)
  {
    if (extendedRange)
    {
      int hbt = getArchivedHeartbeat(context, server, property);
      start -= hbt*2;
      stop  += hbt*2;
    }
    return getArchivedDataArray(context,server,property,device,index,sample,start,stop,data,timeout);
  }
  public static int writeArchivedDataToFile(String filename,String context,String server, String property, String device,double start, double stop)
  {
    return writeArchivedDataToFile(filename,context,server,property,device,start,stop,false);
  }
  public static int writeArchivedDataToFile(String filename,String context,String server, String property, String device,double start, double stop, boolean filterOutEqual)
  {
      return writeArchivedDataToFile(filename,context,server,property,device,start,stop,filterOutEqual,null);
  }
  public static int writeArchivedDataToFile(String filename,String context,String server, String property, String device,double start, double stop, boolean filterOutEqual, Locale locale)
  {
    PrintWriter fw = null;
    try
    {
      //column delimiter is comma if it is not used as a decimal separator, if comma is a dec separator, use semicolon
      char delimiter = locale == null ? ',' : DecimalFormatSymbols.getInstance(locale).getDecimalSeparator() == ',' ? ';' : ',';
      TPropertyQuery[] tpq = TQuery.getPropertyInformation(context, server, device, property);
      if (tpq == null) return TErrorList.non_existent_property;
      // THistory.getNumPointsInInterval(context, server, property, device, start, stop, 2000);
      // n.b. determine num points in interval with another call and warm user acquisition time if too many
      fw = new PrintWriter(new FileWriter(filename));
      StringBuilder sb = new StringBuilder(50);
      sb.append("#/").append(context).append('/').append(server).append('/').append(device)
          .append('[').append(property).append(']').append(delimiter)
          .append(" (").append(tpq[0].prpUnits).append(')').append(delimiter)
          .append(" start ").append(TDataTime.toString(start)).append(delimiter)
          .append(" stop ").append(TDataTime.toString(stop)).append("\r\n");
      fw.write(sb.toString());
      sb = new StringBuilder(50);
      sb.append("time string").append(delimiter).append(" time stamp").append(delimiter).append(" system stamp").append(delimiter).append(" value\r\n");
      fw.write(sb.toString());
      double ts=0, srtt = start;
      int ntot = 0, npts = 1000;
      HISTORY[] hsts = new HISTORY[1000];
      short format = tpq[0].prpFormat;
      for (int i=0; i<1000; i++)
      {
        hsts[i] = new HISTORY(new TDataType(1,format));
      }
      TDataType tdt;
      String lnout;
      long t = System.currentTimeMillis();
      String lastValue = "", currentValue = "";
      DecimalFormat decFormat = locale == null ? null : new DecimalFormat("#.##################", DecimalFormatSymbols.getInstance(locale));
      while (npts == 1000)
      {
        npts = THistory.getArchivedData(context, server, property,device, 0, 1,srtt, stop, hsts, 1000);
        for (int k=0; k<npts; k++)
        {
          tdt = hsts[k].getDataObject();
          if (decFormat == null) currentValue = tdt.toString().trim();
          else
          {
            if (format == TFormat.CF_DOUBLE) currentValue = decFormat.format(((double[])tdt.getDataObject())[0]);
            else if (format == TFormat.CF_FLOAT) currentValue = decFormat.format(((float[])tdt.getDataObject())[0]);
            else currentValue = tdt.toString().trim();
          }
          if (lastValue.equals(currentValue) && filterOutEqual) continue;
          ts = tdt.getDataTimeStamp()/1000.0;
          lnout = new StringBuilder(50 + currentValue.length()).append(TDataTime.toString(ts)).append(delimiter).append(String.format(locale,"%10.3f", ts)).append(delimiter).append(tdt.getSystemDataStamp()).append(delimiter).append(currentValue).toString();
          fw.println(lnout);
          lastValue = currentValue;
        }
        ntot += npts;
        if (ts <= srtt) break;
        srtt = ts;
      }
      t = System.currentTimeMillis() - t;
      fw.write("#access time:, "+t+" (msec), number points in interval:, "+ntot+"\n");
      fw.close();
    }
    catch (Exception x)
    {
      System.out.println("cannot write to "+filename+" : "+x.getMessage());
      return TErrorList.code_failure;
    } finally {
      try {
          if (fw != null)
              fw.close();
      } catch (Exception e) {
          System.out.println("cannot write to "+filename+" : "+e.getMessage());
          return TErrorList.code_failure;
      }
    }
    return 0;
  }
  /**
   * Retrieves the total number of stored data points in the interval specified.
   * 
   * Retrieve the number of points stored for the given parameters
   * and over the time range specified. If server is null, an empty string, or "HISTORY",
   * the central archive server is queried for the data. Otherwise the local history 
   * subsystem of the device server specified is queried.  
   * 
   * @param context is the desired context  
   * @param server is the device server which manages the archived parameter. If null
   * then the central archive will be queried for the property and device
   * @param property is the desired archived property
   * @param device is the desired device name
   * @param start is the start time in seconds (UTC double, a TINE timestamp) 
   * @param stop is the stop time in seconds (UTC double, a TINE timestamp)
   * 
   * @return if positive, the number of points returned in the archive call; if negative,
   * the negative of a TINE error code.
   */
  public static int getNumPointsInInterval(String context,String server, String property, String device, double start, double stop)
  {
    return getArchivedNumPointsInInterval(context, server, property, device, start, stop, 1000);
  }
  /**
   * Retrieves the total number of stored data points in the interval specified.
   * 
   * Retrieve the number of points stored for the given parameters
   * and over the time range specified. If server is null, an empty string, or "HISTORY",
   * the central archive server is queried for the data. Otherwise the local history 
   * subsystem of the device server specified is queried.  
   * 
   * @param context is the desired context  
   * @param server is the device server which manages the archived parameter. If null
   * then the central archive will be queried for the property and device
   * @param property is the desired archived property
   * @param device is the desired device name
   * @param start is the start time in seconds (UTC double, a TINE timestamp) 
   * @param stop is the stop time in seconds (UTC double, a TINE timestamp)
   * @param timeout gives the number of milliseconds to wait for the call to complete 
   * (usually 1000 is sufficient).
   * 
   * @return if positive, the number of points returned in the archive call; if negative,
   * the negative of a TINE error code.
   */
  public static int getNumPointsInInterval(String context,String server, String property, String device, double start, double stop,int timeout)
  {
    return getArchivedNumPointsInInterval(context, server, property, device, start, stop, timeout);
  }
  
  /**
   * Retrieves an archived data array 'snapshot' at the time specified.
   * 
   * Retrieve an archive data set for the target property and device
   * given at the time specified. If server is null, an empty string, or "HISTORY",
   * the central archive server is queried for the data. Otherwise the local history 
   * subsystem of the device server specified is queried. The returned data set
   * represents the entire data array, be it a multi-channel array, trace array, or
   * scalar at a stored time nearest (but not greater than) the given time.  
   * 
   * @param context is the desired context  
   * @param server is the device server which manages the archived parameter. If null
   * then the central archive will be queried for the property and device
   * @param property is the desired archived property
   * @param device is the desired device name
   * @param targetTime is the targetted time in seconds (UTC double, a TINE timestamp) 
   * This is passed as a single-valued array, so that the targetted time can be 
   * altered by the call to reflect the time found.
   * @param data is a reference to a String data array object to contain the returned data.
   * The array should be dimensioned to the number of points in the data record.
   * 
   * @return 0 if success or a TINE error code.
   */
  public static int getArchivedSnapshot(String context,String server, String property, String device, double[] targetTime,String[] data)
  {
    return getArchivedDataSnapshot(context,server,property,device,targetTime,data,false,null);
  }
  /**
   * Retrieves an archived data array 'snapshot' at the time specified.
   * 
   * Retrieve an archive data set for the target property and device
   * given at the time specified. If server is null, an empty string, or "HISTORY",
   * the central archive server is queried for the data. Otherwise the local history 
   * subsystem of the device server specified is queried. The returned IMAGE
   * represents the image at a stored time nearest (but not greater than) the given time.  
   * 
   * @param context is the desired context  
   * @param server is the device server which manages the archived parameter. If null
   * then the central archive will be queried for the property and device
   * @param property is the desired archived property
   * @param device is the desired device name
   * @param targetTime is the targetted time in seconds (UTC double, a TINE timestamp) 
   * This is passed as a single-valued array, so that the targetted time can be 
   * altered by the call to reflect the time found.
   * @param data is a reference to an IMAGE data array object to contain the returned data. 
   * The array should have a length 1
   * 
   * @return 0 if success or a TINE error code.
   */
  public static int getArchivedSnapshot(String context,String server, String property, String device, double[] targetTime,IMAGE data)
  {
    return getArchivedDataSnapshot(context,server,property,device,targetTime,data,true,null);
  }
  /**
   * Retrieves an archived data array 'snapshot' at the time specified.
   * 
   * Retrieve an archive data set for the target property and device
   * given at the time specified. If server is null, an empty string, or "HISTORY",
   * the central archive server is queried for the data. Otherwise the local history 
   * subsystem of the device server specified is queried. The returned data set
   * represents the entire data array, be it a multi-channel array, trace array, or
   * scalar at a stored time nearest (but not greater than) the given time.  
   * 
   * @param context is the desired context  
   * @param server is the device server which manages the archived parameter. If null
   * then the central archive will be queried for the property and device
   * @param property is the desired archived property
   * @param device is the desired device name
   * @param targetTime is the targetted time in seconds (UTC double, a TINE timestamp) 
   * This is passed as a single-valued array, so that the targetted time can be 
   * altered by the call to reflect the time found.
   * @param data is a reference to a float data array object to contain the returned data.
   * The array should be dimensioned to the number of points in the data record.
   * 
   * @return 0 if success or a TINE error code.
   */
  public static int getArchivedSnapshot(String context,String server, String property, String device, double[] targetTime,float[] data)
  {
    return getArchivedDataSnapshot(context,server,property,device,targetTime,data,null);
  }
  public static int getArchivedSnapshot(String context,String server, String property, String device, double[] targetTime,float[] data,THistorySource src)
  {
    return getArchivedDataSnapshot(context,server,property,device,targetTime,data,src);
  }
  public static int getArchivedSnapshot(String context,String server, String property, String device, double[] targetTime,TCompoundDataObject[] data,THistorySource src)
  {
    return getArchivedDataSnapshot(context,server,property,device,targetTime,data,false,src);
  }
  public static int getArchivedSnapshot(String context,String server, String property, String device, double[] targetTime,Object data,int fieldindex,THistorySource src)
  {
    return getArchivedDataSnapshot(context,server,property,device,targetTime,data,fieldindex,false,src);
  }
  public static int addToLocalHistory(
      String context,String server, String property, String device,
      int datasiz,int datafmt,int pollInterval,int archInterval,
      int depthShort,int depthLong,int heartBeat,float pTolerance,float aTolerance)
  {
    THistoryRecordStruct[] hrs = new THistoryRecordStruct[1];
    THistorySpecification hsp = new THistorySpecification(pollInterval, archInterval, depthShort, depthLong, heartBeat, pTolerance, aTolerance, null);
    hrs[0] = new THistoryRecordStruct(device, property, datasiz, datafmt, hsp);
    String dev = "/"+context+"/"+server;
    TLink lnk = new TLink(dev,"ADDHISTORY",null,new TDataType(hrs),TAccess.CA_WRITE);
    return lnk.executeAndClose();
  }
  public static String[] getLocalHistoryProperties(String context,String server,String device)
  {
    int[] n = new int[1];
    String dev = "/"+context+"/"+server;
    TLink lnk = new TLink(dev,"NHISTORIES",new TDataType(n),null,TAccess.CA_READ);
    if (lnk.executeAndClose() != 0) return null;
    if (n[0] == 0)
    { // could be a doocs server
      String eqm = TSrvEntry.getSrvEntryEqmName(context, server);
      if (eqm == null || eqm.compareTo(TSrvEntry.DOOCSEQM) != 0) return null;
      String[] prps = TQuery.getDeviceProperties(context, server, device);
      if (prps == null) return null;
      ArrayList<String>al = new ArrayList<String>();
      for (String p : prps)
      {
        if (p.endsWith(".HIST")) al.add(p.substring(0, p.lastIndexOf('.')));  
      }
      if (al.isEmpty()) return null;
      return al.toArray(new String[0]);
    }
    NAME64[] nlst = new NAME64[n[0]];
    lnk = new TLink(dev,"HISTORIES",new TDataType(nlst),null,TAccess.CA_READ);
    if (lnk.executeAndClose() != 0) return null;
    return NameToString.nameArrayToString(nlst);
  }
  public static THistoryRecordStruct getLocalHistoryRecord(
      String context,String server, String property, String device)
  {
    THistoryRecordStruct[] hrs = new THistoryRecordStruct[1];
    hrs[0] = new THistoryRecordStruct();
    String dev = "/"+context+"/"+server;
    TDataType dout = new TDataType(hrs);
    TDataType din = new TDataType(property.toCharArray());
    TLink lnk = new TLink(dev,"HISTORIES",dout,din,TAccess.CA_READ);
    int cc = lnk.executeAndClose();
    if (cc != 0) return null;
    if (dout.getCompletionLength() < 1) return null;
    return hrs[0];
  }
  /**
   * Retrieve a list of properties archived by the central archive server for the context given.
   * 
   * @param context is the targeted context
   *
   * @return a list of available archived properties
   */
  public static String[] getArchivedProperties(String context)
  {
    return getArchivedProperties(context,null,null);
  }
  /**
   * Retrieve a list of properties archived by the given server.
   * 
   * @param context is the targeted server context
   * @param server is the targeted server name
   *
   * @return a list of available archived properties
   */
  public static String[] getArchivedProperties(String context,String server)
  {
    return getArchivedProperties(context,server,null);
  }
  /**
   * Retrieve a list of properties archived by the given server.
   * 
   * @param context is the targeted server context
   * @param server is the targeted server name
   * @param subsystem is the targeted subsystem 
   * (applicable when the server = "HISTORY", i.e. the central archive server)
   *
   * @return a list of available archived properties
   */
  public static String[] getArchivedProperties(String context,String server,String subsystem)
  {
    if (server == null && context == null) return null;
    if (context == null) context = "DEFAULT";
    if (server == null) server = "HISTORY";
    boolean isCentral = isCentralArchiveServer(server);
    TDataType dtsubs;
    if (subsystem != null && subsystem.length() > 0 && 
        !subsystem.startsWith("ALL") && !subsystem.startsWith("All"))
    {
      dtsubs = new TDataType(subsystem);
    }
    else
    { 
      dtsubs = new TDataType();
    }
    NAME32[] nlist = null; //TODO:  why not NAME64 ?
    int[] nprps = new int[1];
    String ntagsString, tagsString;
    TDataType dtnprps = new TDataType(nprps);
    TDataType dtprps;
    tagsString = isCentral ? "TAGS" : "HISTORIES";
    ntagsString = "N" + tagsString;
    
    String dname = "/" + context + "/" + server;
    TLink tl = new TLink(dname,ntagsString,dtnprps,dtsubs,TAccess.CA_READ);
    int cc = tl.execute(800, true);
    tl.close();
    if (cc != 0)
    { // TODO: throw an exception here
      return null;
    }
    if (nprps[0] > 0)
    {
      nlist = new NAME32[nprps[0]];
      dtprps = new TDataType(nlist);
      tl = new TLink(dname,tagsString,dtprps,dtsubs,TAccess.CA_READ);
      cc = tl.execute(800, true);
      tl.close();
      if (cc != 0)
      { // TODO: throw an exception here
        return null;
      }
    }
    if (nlist == null) return null;
    String nam;
    Vector<String> v = new Vector<String>(nlist.length);
    int n = 0;
    for (int i=0; i<nlist.length; i++)
    {
      nam = nlist[i].getName();
      if (nam.endsWith(".NAM")) continue;
      if (nam.endsWith(".ENUM")) continue;
      if (nam.endsWith(".SRCADDR")) continue;
      if (nam.endsWith(".DESC")) continue;
      v.add(nam);
      n++;
    }
    return v.toArray(new String[n]);
  }
  public static String[] getAnnotatedProperties(String context,double start, double stop, int timeout)
  {
    if (context == null) return null;
    NAME64[] nlist = null;    
    double[] ststp = new double[]{start,stop};
    String dname = "/"+context+"/HISTORY";
    nlist = new NAME64[100];
    TLink tl = new TLink(dname,"RECORDS.WITH.CMTS",new TDataType(nlist),new TDataType(ststp),TAccess.CA_READ);
    int cc = tl.executeAndClose(timeout);
    if (cc != 0)
    { // TODO: throw an exception here
      return null;
    }
    return NameToString.nameArrayToString(nlist,tl.getOutputDataObject().getCompletionLength());
  }
  private static void fillinHstPrpVectors(NAME64[] plist,int len,Vector<String> nv,Vector<String> ev,Vector<String> sav, Vector<String> dv)
  {
    String p, ucnam;
    for (int i=0; i<len; i++)
    {
      ucnam = plist[i].getName().toUpperCase();
      if (ucnam.endsWith(".NAM")) 
      { 
        p = ucnam.substring(0, ucnam.lastIndexOf(".NAM"));
        if (!nv.contains(p)) nv.add(p); continue; 
      }
      if (ucnam.endsWith(".ENUM"))
      { 
        p = ucnam.substring(0, ucnam.lastIndexOf(".ENUM"));
        if (!ev.contains(p)) ev.add(p); continue; 
      }
      if (ucnam.endsWith(".SRCADDR"))
      { 
        p = ucnam.substring(0, ucnam.lastIndexOf(".SRCADDR"));
        if (!sav.contains(p)) sav.add(p); continue; 
      }
      if (ucnam.endsWith(".DESC"))
      { 
        int idx = ucnam.lastIndexOf(".NAM.DESC");
        if (idx < 0) idx = ucnam.lastIndexOf(".DESC");
        p = ucnam.substring(0, idx);
        if (!dv.contains(p)) dv.add(p); continue; 
      }          
    }
  }
  /**
   * Retrieve a list of properties archived by the given server as a list of
   * HstPrp instances, where additional information is returned along with the 
   * property name.
   * 
   * @param context is the targeted server context
   * @param server is the targeted server name
   * @param subsystem is the targeted subsystem 
   * (applicable when the server = "HISTORY", i.e. the central archive server)
   *
   * @return a list of available archived properties
   */
  public static HstPrp[] getArchivedPropertyList(String context,String server,String subsystem)
  {
    if (server == null && context == null) return null;
    if (context == null) context = "DEFAULT";
    if (server == null) server = "HISTORY";
    boolean isCentral = isCentralArchiveServer(server);
    TDataType dtsubs;
    if (subsystem != null && subsystem.length() > 0 && 
        !subsystem.startsWith("ALL") && !subsystem.startsWith("All"))
    {
      dtsubs = new TDataType(subsystem);
    }
    else
    { 
      dtsubs = new TDataType();
    }
    NAME64[] nlist = null; 
    int[] nprps = new int[1];
    String ntagsString, tagsString;
    tagsString = isCentral ? "TAGS" : "HISTORIES";
    ntagsString = "N" + tagsString;
    
    String dname = "/" + context + "/" + server;
    TLink tl = new TLink(dname,ntagsString,new TDataType(nprps),dtsubs,TAccess.CA_READ);
    int cc = tl.executeAndClose(800);
    if (cc != 0)
    { // TODO: throw an exception here
      return null;
    }
    if (nprps[0] > 0)
    { // get OFFICIAL list of archived elements
      nlist = new NAME64[nprps[0]];
      tl = new TLink(dname,tagsString, new TDataType(nlist),dtsubs,TAccess.CA_READ);
      cc = tl.executeAndClose(800);
      if (cc != 0)
      { // TODO: throw an exception here
        return null;
      }
    }
    if (nlist == null) return null;
    String ucnam, nam;
    Vector<String> nv = new Vector<String>(nlist.length);
    Vector<String> ev = new Vector<String>(25);
    Vector<String> sav = new Vector<String>(nlist.length);
    Vector<String> dv = new Vector<String>(25);
    Hashtable<String,HstPrp> pl = new Hashtable<String,HstPrp>();
    for (int i=0; i<nlist.length; i++)
    {
      nam = nlist[i].getName();
      if (nam.trim().length() == 0) continue;
      ucnam = nam.toUpperCase();
      if (ucnam.endsWith(".NAM")) 
      { 
        nv.add(ucnam.substring(0, ucnam.indexOf(".NAM"))); continue; 
      }
      if (ucnam.endsWith(".ENUM"))
      { 
        ev.add(ucnam.substring(0, ucnam.indexOf(".ENUM"))); continue; 
      }
      if (ucnam.endsWith(".SRCADDR"))
      { 
        sav.add(ucnam.substring(0, ucnam.indexOf(".SRCADDR"))); continue; 
      }
      if (ucnam.endsWith(".DESC"))
      { 
        int idx = ucnam.indexOf(".NAM.DESC");
        if (idx < 0) idx = ucnam.indexOf(".DESC");
        dv.add(ucnam.substring(0, idx)); continue; 
      }
      pl.put(ucnam, new HstPrp(nam));
    }
    if (pl.isEmpty()) return new HstPrp[0];
    if (isCentral)
    { // take care of other contingencies at the Central Archive Server
      NAME64[] plist = new NAME64[nlist.length*2];
      TDataType tdt = new TDataType(plist);
      tl = new TLink(dname,"PROPERTIES",tdt,null,TAccess.CA_READ);
      if (tl.executeAndClose(800) == 0)
      {
        fillinHstPrpVectors(plist,tdt.getCompletionLength(),nv,ev,sav,dv);
      }
      tl = new TLink(dname,"SRVALIASLIST",tdt,null,TAccess.CA_READ);
      if (tl.executeAndClose(800) == 0)
      {
        fillinHstPrpVectors(plist,tdt.getCompletionLength(),nv,ev,sav,dv);
      }      
    }    
    HstPrp hp;
    for (String s : nv) if ((hp=pl.get(s)) != null) hp.hasNameList = true;
    for (String s : ev) if ((hp=pl.get(s)) != null) hp.hasEnumList = true;
    for (String s : sav) if ((hp=pl.get(s)) != null) hp.hasSrcList = true;
    for (String s : dv) if ((hp=pl.get(s)) != null) hp.hasDescList = true;
    return pl.values().toArray(new HstPrp[pl.size()]);
  }
  private static String selectedContext = null;
  private static Hashtable<String, String> srcTbl = null;
  public static String getHistoryServerDropString(String context,String sourceDropString)
  {
    if (sourceDropString == null) return sourceDropString;
    if (sourceDropString.startsWith("TINE/")) return sourceDropString;
    if (selectedContext == null || selectedContext.compareToIgnoreCase(context) != 0)
    {
      selectedContext = context;
      srcTbl = getArchivedSources(context,"HISTORY",1000);
      if (srcTbl == null) 
        return sourceDropString.startsWith("/") ? "TINE"+sourceDropString : "TINE/"+sourceDropString;
    }
    String key;
    int idx = sourceDropString.indexOf('[');
    if (idx > 0)
    {
      key = sourceDropString;
    }
    else
    {
      idx = sourceDropString.lastIndexOf('/');
      if (idx < 0) // punt
        return sourceDropString.startsWith("/") ? "TINE"+sourceDropString : "TINE/"+sourceDropString;
      String prp = sourceDropString.substring(idx+1);
      key = sourceDropString.substring(0,idx)+"["+prp+"]";
    }
    if (srcTbl == null) // punt
      return sourceDropString.startsWith("/") ? "TINE"+sourceDropString : "TINE/"+sourceDropString;
    if (!key.startsWith("/")) key = "/"+key;
    if (srcTbl.containsKey(key))
    {
      String v = srcTbl.get(key);
      key = "/"+context+"/ARCHIVER/"+v;
    }
    // re-form to the viewers expectation
    idx = key.indexOf('[');
    if (idx < 0) return "TINE"+key; // punt
    int lidx = key.indexOf(']');
    String prp = key.substring(idx+1,lidx);
    key = key.substring(0,idx)+"/"+prp;      
    return "TINE"+key;
  }
  public static Hashtable<String, String> getArchivedSources(String context,String server,int timeout)
  {
    Hashtable<String, String> srcTbl = new Hashtable<String, String>();
    if (server == null && context == null) return null;
    if (context == null) context = "DEFAULT";
    if (server == null) server = "HISTORY";
    NAME64[] nlist = new NAME64[1000];
    TDataType dtprps = new TDataType(nlist);
    
    String dname = "/" + context + "/" + server;
    TLink tl = new TLink(dname,"*.SRCADDR",dtprps,null,TAccess.CA_READ);
    int cc = tl.executeAndClose(800);
    if (cc != 0)
    { // TODO: throw an exception here
      return null;
    }
    int listsize = dtprps.getCompletionLength();
    if (listsize > 0)
    {
      String prp;
      String[] srcs;
      for (int i=0; i<listsize; i++)
      {
        prp = nlist[i].getName();
        prp = prp.substring(0, prp.length()-8);
        srcs = getArchiveSource(context, server, prp, "#0", timeout);
        if (srcs != null)
        {
          String key = "keyword["+prp+"]";
          String dev;
          String[] devs = null;
          if (srcs.length > 1)
          {
            devs = TQuery.getDeviceNames(context, server, prp);
          }
          for (int k=0; k<srcs.length; k++)
          { // key is the src string (per device)
            if (srcs[k] == null) break;
            if (srcs.length > 1) 
            {
              dev = (devs == null || k >= devs.length) ? "#"+k : devs[k]; 
              key = dev+"["+prp+"]";
            }
            srcTbl.put(srcs[k],key);
          }
        }
      }
      return srcTbl;
    }
    return null;
  }
  /**
   * Retrieve a list of archived devices for the given server and property.
   * 
   * @param context is the targeted server context
   * @param server is the targeted server name
   * @param property is the targeted property 
   *
   * @return a list of available archived device names
   */
  public static String[] getArchivedDevices(String context,String server, String property)
  {
    return TQuery.getDeviceNames(context, server, property);
  }
  public static String[] getArchivedDevices(String context,String server, String property,double timestamp)
  {
    if (context == null) context = "DEFAULT";
    if (server == null) server = "HISTORY"; // assume the standard central archive server for the given context
    if (server.compareTo("HISTORY") != 0) return TQuery.getDeviceNames(context, server, property);
    if (property == null) return null;
    if (timestamp <= 0) timestamp = (double)(System.currentTimeMillis()/1000);
    double[] tt = new double[]{timestamp};
    NAME64[] lst = new NAME64[1024];
    //for (int i=0; i<1024; i++) lst[i] = new NAME64();
    THistorySource src = new THistorySource();
    int cc = getArchivedSnapshot(context,server,property+".NAM","#0",tt,lst,src);
    int alen = src.getNumberOfPointsReturned();
    if (alen < 1024) lst = Arrays.copyOf((NAME64[])lst, alen);   
    return TErrorList.hasData(cc) ? NameToString.nameArrayToString(lst) : null;
  }
  /**
   * Retrieve a list of available subsystems at the central archive server for the given context
   * 
   * @param context is the targeted server context
   *
   * @return a list of available subsystems
   */
  public static String[] getArchivedSubsystems(String context)
  {
    return getArchivedSubsystems(context,null);
  }
  public static String[] getArchivedSubsystems(String context,String server)
  {
    if (server == null && context == null) return null;
    if (context == null) context = "DEFAULT";
    if (server == null) server = "HISTORY"; // assume the standard central archive server for the given context
    if (!isCentralArchiveServer(server)) return null;
    int cc;
    NAME32[] nlist = null;
    int[] nsubs = new int[1];
    String ntagsString, tagsString;
    TDataType dtnsubs = new TDataType(nsubs);
    TDataType dtsubs;
    tagsString = "SUBSYSTEMS";
    ntagsString = "N" + tagsString;
    
    String dname = "/" + context + "/" + server;
    TLink tl = new TLink(dname,ntagsString,dtnsubs,null,TAccess.CA_READ);
    cc = tl.execute(800, true);
    tl.close();
    if (cc != 0)
    { // TODO: throw an exception here
      return null;
    }
    if (nsubs[0] > 0)
    {
      nlist = new NAME32[nsubs[0]];
      dtsubs = new TDataType(nlist);
      tl = new TLink(dname,tagsString,dtsubs,null,TAccess.CA_READ);
      cc = tl.execute(800, true);
      tl.close();
      if (cc != 0)
      { // TODO: throw an exception here
        return null;
      }
    }
    else
    {
      return new String[] { "" };
    }
    String nam;
    Vector<String> v = new Vector<String>(nlist.length);
    int n = 0;
    for (int i=0; i<nlist.length; i++)
    {
      nam = nlist[i].getName();
      if (nam.length() == 0) continue;
      v.add(nam);
      n++;
    }
    return v.toArray(new String[n]);
  }
  public static String[] getOperationHistoryContexts()
  {
    return TQuery.getContexts("STATE");
  }
  public static String[] getOperationHistoryContexts(String importance)
  {
    return TQuery.getContexts("STATE",importance);
  }
  /**
   * Retrieve a list of available contexts where a central archive server exists
   * 
   * @return a list of available contexts
   */
  public static String[] getArchiveServerContexts()
  {
    return TQuery.getContexts("HISTORY");
  }
  /**
   * Retrieve a list of available contexts where a central archive server exists
   * and has the importance specified
   * 
   * @param importance is the targeted importance level
   * (one of "ALL", "IMPORTANT", "CRITICAL", "ESSENTIAL")
   * @return a list of available contexts
   */
  public static String[] getArchiveServerContexts(String importance)
  {
    return TQuery.getContexts("HISTORY",importance);
  }
  public static String[] getOperationHistoryConfigurationNames(String context)
  {
    if (context == null) return null;
    return TQuery.getDeviceNames(context, "STATE", "CONFIGURATION");
  }
  public static String[] getViewerMcaConfigurationNames(String context)
  {
    if (context == null) return null;
    return TQuery.getDeviceNames(context, "HISTORY", "CONFIGURATION.MCA");
  }
  public static String[] getViewerConfigurationNames(String context)
  {
    if (context == null) return null;
    return TQuery.getDeviceNames(context, "HISTORY", "CONFIGURATION");
  }
  public static String[] getViewerTraceConfigurationNames(String context)
  {
    if (context == null) return null;
    String[] nms = TQuery.getDeviceNames(context, "HISTORY", "CONFIGURATION.TRACE");
    if (nms == null || nms.length < 2) return nms;
    List<String> nmsr = new ArrayList<>();
    for (int i=0; i<nms.length; i++)
    {
      if (nms[i].equalsIgnoreCase("properties") || nms[i].equalsIgnoreCase("traces")) 
      {
        continue;
      }
      nmsr.add(nms[i]);
    }
    return nmsr.toArray(new String[nmsr.size()]);
  }
  public static String[] getViewerConfigurationNames(String context,String group)
  {
    if (context == null) return null;
    String prp = "CONFIGURATION";
    if (group != null) prp += "." + group;
    String[] nms = TQuery.getDeviceNames(context, "HISTORY", prp);
    if (nms == null || nms.length < 2) return nms;
    String[] nmsr = new String[nms.length-1];
    boolean hasProperties = false;
    for (int i=0, k=0; i<nms.length; i++)
    {
      if (nms[i].compareToIgnoreCase("properties") == 0) 
      {
        hasProperties = true;
        continue;
      }
      if (k<nms.length-1) nmsr[k++] = nms[i];
    }
    return hasProperties ? nmsr : nms;
  }
  public static HstView[] getViewerConfigurationList(String context,String group)
  {
    if (context == null) return null;
    if (!isInitialized) initCfgs();
    String prp = "CONFIGURATION";
    if (group != null) prp += "." + group;
    prp += ".NAM";
    TDataType dout = new TDataType(views);
    String dname = "/" + context + "/HISTORY";
    TLink tl = new TLink(dname,prp,dout,null,TAccess.CA_READ);
    if (tl.executeAndClose(1000) != 0) return null;
    int len = dout.getCompletionLength();
    if (len == 0) return null;
    HstView[] list = new HstView[len];
    for (int i=0; i<len; i++) list[i] = new HstView(views[i]);
    return list;
  }
  public static int setViewerConfigurationList(String context,String group,HstView[] list)
  {
    if (context == null || list == null) return TErrorList.argument_list_error;
    if (!isInitialized) initCfgs();
    String prp = "CONFIGURATION";
    if (group != null) prp += "." + group;
    prp += ".NAM";
    TDataType din = new TDataType(list);
    String dname = "/" + context + "/HISTORY";
    TLink tl = new TLink(dname,prp,null,din,TAccess.CA_WRITE);
    return tl.executeAndClose(800);
  }
  public static String[] getViewerConfigurationGroups(String context)
  {
    if (context == null) return null;
    NAME64[] prps = new NAME64[100];
    TDataType dtprps = new TDataType(prps);
    String dname = "/" + context + "/HISTORY";
    String dprop = "CONFIGURATION.*";
    TLink tl = new TLink(dname,dprop,dtprps,null,TAccess.CA_READ);
    int cc = tl.execute(800, true);
    tl.close();
    if (cc != 0)
    { // TODO: throw an exception here
      return null;
    }
    int len = dtprps.getCompletionLength();
    String nam;
    Vector<String> v = new Vector<String>(len);
    int n = 0;
    for (int i=0; i<len; i++)
    {
      nam = prps[i].getName();
      if (nam.compareToIgnoreCase("CONFIGURATION") == 0) continue;
      if (nam.length() == 0) continue;
      if (nam.endsWith(".NAM")) continue;
      if (nam.endsWith(".MCA")) continue;
      if (nam.endsWith(".TRACE")) continue;
      v.add(nam.substring(14));
      n++;
    }
    return n > 0 ? v.toArray(new String[n]) : null;    
  }
  public static HstCfg[] getSelectableProptiesConfiguration(String context,String group)
  {
    if (context == null || group == null) return null;
    return getViewerConfiguration(context,"properties",group);
  }
  public static HstCfg[] getSelectableTracesConfiguration(String context)
  {
    if (context == null) return null;
    return getViewerConfiguration(context,"traces","TRACE");
  }
  public static HstCfg[] getViewerConfiguration(String context,String name)
  {
    return getViewerConfiguration(context,name,null);
  }
  public static synchronized HstCfg[] getViewerConfiguration(String context,String name,String group)
  {
    if (context == null || name == null) return null;
    if (!isInitialized) initCfgs();
    TDataType dtcfg = new TDataType(cfgs);
    String dname = "/" + context + "/HISTORY/" + name;
    String dprop = "CONFIGURATION";
    if (group != null && group.length() > 0) dprop += "." + group;
    TLink tl = new TLink(dname,dprop,dtcfg,null,TAccess.CA_READ);
    int cc = tl.execute(800, true);
    tl.close();
    if (cc != 0)
    { // TODO: throw an exception here
      DbgLog.log("getViewerConfiguration",TErrorList.getErrorString(cc));
      return null;
    }
    int num = dtcfg.getCompletionLength();
    if (num == 0) return null;
    boolean hasPropertiesList = false;
    for (int i=0; i<num; i++)
    { // is 'properties' in the list ?
      if (cfgs[i].getDescription().compareToIgnoreCase("properties") == 0)
      {
        hasPropertiesList = true;
        break;
      }
    }
    int ncfg = hasPropertiesList ? num - 1 : num;
    HstCfg[] cfg = new HstCfg[ncfg];
    for (int i=0, n=0; i<num; i++) 
    {
      if (cfgs[i].getDescription().compareToIgnoreCase("properties") != 0)
      {
        cfg[n++] = new HstCfg(cfgs[i]);        
      }
    }
    return cfg;
  }
  public static synchronized int setViewerConfiguration(String context,String name,String group,HstCfg[] hstCfgs)
  {
    if (context == null || name == null) return TErrorList.argument_list_error;
    if (!isInitialized) initCfgs();
    TDataType dtcfg = new TDataType(hstCfgs);
    String dname = "/" + context + "/HISTORY/" + name;
    String dprop = "CONFIGURATION";
    if (group != null && group.length() > 0) dprop += "." + group;
    TLink tl = new TLink(dname,dprop,null,dtcfg,TAccess.CA_WRITE);
    return tl.executeAndClose(800);
  }
  public static synchronized int setTraceViewerConfiguration(String context,String name,HstCfg[] hstCfgs)
  {
    if (context == null || name == null) return TErrorList.argument_list_error;
    if (!isInitialized) initCfgs();
    TDataType dtcfg = new TDataType(hstCfgs);
    String dname = "/" + context + "/HISTORY/" + name;
    String dprop = "CONFIGURATION.TRACE";
    TLink tl = new TLink(dname,dprop,null,dtcfg,TAccess.CA_WRITE);
    return tl.executeAndClose(800);
  }
  public static synchronized McaCfg[] getMcaViewerConfiguration(String context,String name)
  {
    if (context == null || name == null) return null; 
    if (!isInitialized) initCfgs();
    TDataType dtcfg = new TDataType(mcacfgs);
    String dname = "/" + context + "/HISTORY/" + name;
    String dprop = "CONFIGURATION.MCA";
    TLink tl = new TLink(dname,dprop,dtcfg,null,TAccess.CA_READ);
    int cc = tl.execute(800, true);
    tl.close();
    if (cc != 0)
    { // TODO: throw an exception here
      DbgLog.log("getMcaViewerConfiguration",TErrorList.getErrorString(cc));
      return null;
    }
    int num = dtcfg.getCompletionLength();
    if (num == 0) return null;
    McaCfg[] cfg = new McaCfg[num];
    for (int i=0; i<num; i++) 
    {
      cfg[i] = new McaCfg(mcacfgs[i]);
    }
    return cfg;
  }
  public static synchronized int setMcaViewerConfiguration(String context,String name,McaCfg[] mcaCfgs)
  {
    if (context == null || name == null) return TErrorList.argument_list_error; 
    if (!isInitialized) initCfgs();
    TDataType dtcfg = new TDataType(mcaCfgs);
    String dname = "/" + context + "/HISTORY/" + name;
    String dprop = "CONFIGURATION.MCA";
    TLink tl = new TLink(dname,dprop,null,dtcfg,TAccess.CA_WRITE);
    return tl.executeAndClose(800);
  }
  public static synchronized McaCfg[] getOperationHistoryConfiguration(String context,String name)
  {
    if (context == null || name == null) return null; 
    if (!isInitialized) initCfgs();
    TDataType dtcfg = new TDataType(mcacfgs);
    String dname = "/" + context + "/STATE/" + name;
    String dprop = "CONFIGURATION";
    TLink tl = new TLink(dname,dprop,dtcfg,null,TAccess.CA_READ);
    int cc = tl.execute(800, true);
    tl.close();
    if (cc != 0)
    { // TODO: throw an exception here
      DbgLog.log("getOperationHistoryConfiguration",TErrorList.getErrorString(cc));
      return null;
    }
    int num = dtcfg.getCompletionLength();
    if (num == 0) return null;
    McaCfg[] cfg = new McaCfg[num];
    for (int i=0; i<num; i++) 
    {
      cfg[i] = new McaCfg(mcacfgs[i]);
    }
    return cfg;
  }
  public static synchronized int setOperationHistoryConfiguration(String context,String name,McaCfg[] configs)
  {
    if (context == null || name == null || configs == null) return TErrorList.invalid_parameter; 
    TDataType dtcfg = new TDataType(configs);
    String dname = "/" + context + "/STATE/" + name;
    String dprop = "CONFIGURATION";
    TLink tl = new TLink(dname,dprop,null,dtcfg,TAccess.CA_WRITE);
    int cc = tl.execute(800, true);
    tl.close();
    if (cc != 0)
    { // TODO: throw an exception here
      DbgLog.log("setOperationHistoryConfiguration",TErrorList.getErrorString(cc));
    }
    return cc;
  }
  public static synchronized PrfCfg[] getOperationHistoryProfile(String context,String name)
  {
    if (context == null || name == null) return null; 
    if (!isInitialized) initCfgs(); else resetCfgs();
    TDataType dtcfg = new TDataType(prfcfgs);
    String dname = "/" + context + "/STATE/" + name;
    String dprop = "PROFILE";
    TLink tl = new TLink(dname,dprop,dtcfg,null,TAccess.CA_READ);
    int cc = tl.executeAndClose(800);
    if (cc != 0)
    { // TODO: throw an exception here
      DbgLog.log("getOperationHistoryProfile",TErrorList.getErrorString(cc));
      return null;
    }
    int num = dtcfg.getCompletionLength();
    if (num == 0) return null;
    Arrays.sort(prfcfgs, 0, num, prfcmp);
    PrfCfg[] cfg = new PrfCfg[num];
    for (int i=0; i<num; i++) 
    {
      cfg[i] = new PrfCfg(prfcfgs[i]);
    }
    return cfg;
  }
  private static class PrfCmp implements Comparator<PrfCfgEx>
  {
    @Override
    public int compare(PrfCfgEx p1, PrfCfgEx p2)
    {
      int m1 = p1.getMetric();
      int m2 = p2.getMetric();
      return (m1 < m2) ? -1 : (m1 == m2) ? 0 : 1; 
    }    
  }
  private static PrfCmp prfcmp = new PrfCmp();
  public static synchronized PrfCfgEx[] getOperationHistoryProfiles(String context,String name)
  {
    if (context == null || name == null) return null; 
    if (!isInitialized) initCfgs();
    TDataType dtcfg = new TDataType(prfcfgs);
    String dname = "/" + context + "/STATE/" + name;
    String dprop = "PROFILE";
    TLink tl = new TLink(dname,dprop,dtcfg,null,TAccess.CA_READ);
    int cc = tl.executeAndClose(800);
    if (cc != 0)
    { // TODO: throw an exception here
      DbgLog.log("getOperationHistoryProfile",TErrorList.getErrorString(cc));
      return null;
    }
    int num = dtcfg.getCompletionLength();
    if (num == 0) return null;
    PrfCfgEx[] cfg = new PrfCfgEx[num];
    for (int i=0; i<num; i++) 
    {
      cfg[i] = new PrfCfgEx(prfcfgs[i]);
    }
    Arrays.sort(cfg, prfcmp);
    return cfg;
  }
  public static synchronized DclState[] getOperationHistoryStates(String context)
  {
    if (context == null) return null; 
    if (!isStatesInitialized) initStates();
    TDataType dtstate = new TDataType(dclstates);
    String dname = "/" + context + "/STATE/#0";
    String dprop = "STATELIST";
    TLink tl = new TLink(dname,dprop,dtstate,null,TAccess.CA_READ);
    int cc = tl.executeAndClose(800);
    if (cc != 0)
    { // TODO: throw an exception here
      DbgLog.log("getOperationHistoryStates",TErrorList.getErrorString(cc));
      return null;
    }
    int num = dtstate.getCompletionLength();
    if (num == 0) return null;
    DclState[] st = new DclState[num];
    for (int i=0; i<num; i++) 
    {
      st[i] = new DclState(dclstates[i]);
    }
    return st;
  }
  public static synchronized DclProc[] getOperationHistoryProcedures(String context)
  {
    if (context == null) return null; 
    if (!isStatesInitialized) initStates();
    TDataType dtproc = new TDataType(dclprocs);
    String dname = "/" + context + "/STATE/#0";
    String dprop = "PROCLIST";
    TLink tl = new TLink(dname,dprop,dtproc,null,TAccess.CA_READ);
    int cc = tl.executeAndClose(800);
    if (cc != 0)
    { // TODO: throw an exception here
      DbgLog.log("getOperationHistoryProcedures",TErrorList.getErrorString(cc));
      return null;
    }
    int num = dtproc.getCompletionLength();
    if (num == 0) return null;
    DclProc[] pr = new DclProc[num];
    for (int i=0; i<num; i++) 
    {
      pr[i] = new DclProc(dclprocs[i]);
    }
    return pr;
  }
  public static synchronized String[] getOperationHistoryStatesInProcedure(String context,String procedure)
  {
    if (context == null || procedure == null) return null; 
    NAME64[] states = new NAME64[16];
    TDataType dtstates = new TDataType(states);
    String dname = "/" + context + "/STATE/" + procedure;
    String dprop = "PROCSTATES";
    TLink tl = new TLink(dname,dprop,dtstates,null,TAccess.CA_READ);
    int cc = tl.executeAndClose(800);
    if (cc != 0)
    { // TODO: throw an exception here
      DbgLog.log("getOperationHistoryStates",TErrorList.getErrorString(cc));
      return null;
    }
    int num = dtstates.getCompletionLength();
    if (num == 0) return null;
    return NameToString.nameArrayToString(Arrays.copyOf(states, num));
  }
  public static synchronized int setOperationHistoryStates(String context,DclState[] states)
  {
    if (context == null || states == null) return TErrorList.invalid_parameter; 
    TDataType dtstate = new TDataType(states);
    String dname = "/" + context + "/STATE/#0";
    String dprop = "STATELIST";
    TLink tl = new TLink(dname,dprop,null,dtstate,TAccess.CA_WRITE);
    int cc = tl.execute(800, true);
    tl.close();
    if (cc != 0)
    { // TODO: throw an exception here
      DbgLog.log("setOperationHistoryStates",TErrorList.getErrorString(cc));
    }
    return cc;
  }
  public static synchronized int setOperationHistoryStatesInProcedure(String context,String procedure,String[] states)
  {
    if (context == null || procedure == null) return TErrorList.argument_list_error; 
    NAME64[] names = StringToName.stringArrayToName64(states);
    TDataType dtstates = new TDataType(names);
    String dname = "/" + context + "/STATE/" + procedure;
    String dprop = "PROCSTATES";
    TLink tl = new TLink(dname,dprop,null,dtstates,TAccess.CA_WRITE);
    int cc = tl.executeAndClose(800);
    if (cc != 0)
    { // TODO: throw an exception here
      DbgLog.log("getOperationHistoryStates",TErrorList.getErrorString(cc));
    }
    return cc;
  }
  public static synchronized int setOperationHistoryProcedures(String context,DclProc[] procs)
  {
    if (context == null || procs == null) return TErrorList.invalid_parameter; 
    TDataType dtproc = new TDataType(procs);
    String dname = "/" + context + "/STATE/#0";
    String dprop = "PROCLIST";
    TLink tl = new TLink(dname,dprop,null,dtproc,TAccess.CA_WRITE);
    int cc = tl.execute(800, true);
    tl.close();
    if (cc != 0)
    { // TODO: throw an exception here
      DbgLog.log("setOperationHistoryProcedures",TErrorList.getErrorString(cc));
    }
    return cc;
  }
  public static synchronized int setOperationHistoryProfile(String context,String name,PrfCfg[] configs)
  {
    if (context == null || name == null || configs == null) return TErrorList.invalid_parameter; 
    TDataType dtcfg = new TDataType(configs);
    String dname = "/" + context + "/STATE/" + name;
    String dprop = "PROFILE";
    TLink tl = new TLink(dname,dprop,null,dtcfg,TAccess.CA_WRITE);
    int cc = tl.execute(800, true);
    tl.close();
    if (cc != 0)
    { // TODO: throw an exception here
      DbgLog.log("setOperationHistoryProfile",TErrorList.getErrorString(cc));
    }
    return cc;
  }
  public static synchronized int setOperationHistoryProfile(String context,String name,PrfCfgEx[] configs)
  {
    if (context == null || name == null || configs == null) return TErrorList.invalid_parameter; 
    TDataType dtcfg = new TDataType(configs);
    String dname = "/" + context + "/STATE/" + name;
    String dprop = "PROFILE";
    TLink tl = new TLink(dname,dprop,null,dtcfg,TAccess.CA_WRITE);
    int cc = tl.execute(800, true);
    tl.close();
    if (cc != 0)
    { // TODO: throw an exception here
      DbgLog.log("setOperationHistoryProfile",TErrorList.getErrorString(cc));
    }
    return cc;
  }
  public static synchronized int setStateCorrection(String context,StateCor correction)
  {
    return setStateCorrection(context,correction,null);
  }
  public static synchronized int setStateCorrection(String context,StateCor correction,String[] status)
  {
    if (context == null || correction == null) return TErrorList.invalid_parameter; 
    String dname = "/" + context + "/STATE/#0";
    String dprop = "Corrections";
    TLink tl = new TLink(dname,dprop,null,new TDataType(new StateCor[]{correction}),TAccess.CA_WRITE);
    int cc = tl.executeAndClose(800);
    String sts = tl.getLastError();
    if (cc != 0)
    { // TODO: throw an exception here
      DbgLog.log("setStateCorrection",sts);
    }
    if (status != null && status.length > 0) status[0] = sts;
    return cc;
  }
  public static synchronized StateCor[] getStateCorrections(String context,long startTime,long stopTime)
  {
    if (context == null) context = "TEST"; 
    String dname = "/" + context + "/STATE/#0";
    String dprop = "Corrections";
    StateCor[] corrections = new StateCor[1024];
    for (int i=0; i<1024; i++) corrections[i] = new StateCor();
    int[] range = new int[]{(int)(startTime/1000),(int)(stopTime/1000)};
    TLink tl = new TLink(dname,dprop,new TDataType(corrections),new TDataType(range),TAccess.CA_READ);
    int cc = tl.executeAndClose(800);
    if (cc != 0) throw new TineRuntimeErrorException(cc);
    int len = tl.getOutputDataObject().getCompletionLength();
    return len == corrections.length ? corrections : Arrays.copyOf(corrections, len);
  }
  private static synchronized int[] getCorrectionCounts(String context,String server,String type,long startTime,long stopTime)
  {
    if (context == null) context = "TEST"; 
    String dname = "/"+context+"/"+server+"/#0";
    String dprop = type.contains(".") ? type : "CorrectionCounts."+type;
    int[] range = new int[]{(int)(startTime/1000),(int)(stopTime/1000)};
    int[] cnts = new int[100];
    TLink tl = new TLink(dname,dprop,new TDataType(cnts),new TDataType(range),TAccess.CA_READ);
    int cc = tl.executeAndClose(800);
    if (cc != 0) throw new TineRuntimeErrorException(cc);
    int len = tl.getOutputDataObject().getCompletionLength();
    return len == 100 ? cnts : Arrays.copyOf(cnts, len);
  }
  public static synchronized int[] getStateCorrectionCounts(String context,long startTime,long stopTime)
  {
  	return getCorrectionCounts(context,"STATE","States",startTime,stopTime);
  }
  public static synchronized int[] getProcedureCorrectionCounts(String context,long startTime,long stopTime)
  {
  	return getCorrectionCounts(context,"STATE","Procedures",startTime,stopTime);
  }
  private static synchronized int setAvailabilityCorrection(String context,String property,AvailCor correction, String[] status)
  {
    if (context == null || correction == null) return TErrorList.invalid_parameter; 
    String dname = "/" + context + "/ALARMSTATE/#0";
    TLink tl = new TLink(dname,property,null,new TDataType(new AvailCor[]{correction}),TAccess.CA_WRITE);
    int cc = tl.executeAndClose(800);
    String cstr = tl.getLastError();
    if (cc != 0)
    { // TODO: throw an exception here
      DbgLog.log("setAvailabilityCorrection",cstr);
    }
    if (status != null && status.length > 0) status[0] = cstr;
    return cc;
  }
  public static synchronized int setAvailabilityCorrection(String context,AvailCor correction)
  {
    return setAvailabilityCorrection(context,"Corrections",correction, null);
  }
  public static synchronized int setAvailabilityCorrection(String context,AvailCor correction,String[] status)
  {
    return setAvailabilityCorrection(context,"Corrections",correction, status);
  }
  public static synchronized int setAssignedBlame(String context,AvailCor correction)
  {
    if (!correction.getIsAvailable()) correction.setState("Blamed");
    return setAvailabilityCorrection(context,"AssignedBlame",correction, null);
  }
  public static synchronized int setAssignedBlame(String context,AvailCor correction,String[] status)
  {
    if (!correction.getIsAvailable()) correction.setState("Blamed");
    return setAvailabilityCorrection(context,"AssignedBlame",correction, status);
  }
  private static synchronized AvailCor[] getAvailabilityCorrections(String context,String property,long startTime,long stopTime)
  {
    if (context == null) context = "TEST"; 
    String dname = "/" + context + "/ALARMSTATE/#0";
    AvailCor[] corrections = new AvailCor[1024];
    for (int i=0; i<1024; i++) corrections[i] = new AvailCor();
    int[] range = new int[]{(int)(startTime/1000),(int)(stopTime/1000)};
    TLink tl = new TLink(dname,property,new TDataType(corrections),new TDataType(range),TAccess.CA_READ);
    int cc = tl.executeAndClose(800);
    if (cc != 0) throw new TineRuntimeErrorException(cc);
    int len = tl.getOutputDataObject().getCompletionLength();
    return len == corrections.length ? corrections : Arrays.copyOf(corrections, len);
  }
  public static synchronized AvailCor[] getAvailabilityCorrections(String context,long startTime,long stopTime)
  {
    return getAvailabilityCorrections(context,"Corrections",startTime,stopTime);
  }
  public static synchronized AvailCor[] getAssignedBlame(String context,long startTime,long stopTime)
  {
    return getAvailabilityCorrections(context,"AssignedBlame",startTime,stopTime);
  }
  public static synchronized int[] getAvailabilityCorrectionCounts(String context,long startTime,long stopTime)
  {
    return getCorrectionCounts(context,"ALARMSTATE","Systems",startTime,stopTime);
  }
  public static synchronized int[] getAvailabilityGroupCorrectionCounts(String context,long startTime,long stopTime)
  {
    return getCorrectionCounts(context,"ALARMSTATE","Groups",startTime,stopTime);
  }
  public static synchronized int[] getAssignedBlameCounts(String context,long startTime,long stopTime)
  {
    return getCorrectionCounts(context,"ALARMSTATE","BlameCounts.Systems",startTime,stopTime);
  }
  public static long getDepthInSeconds(String depth)
  {
    long d  = 0;
    try
    {
      depth = Str.trim(depth).toUpperCase();
      if (depth.length() > 0)
      {
        String ss[] =  depth.split("[a-zA-Z ]", 2);
        d = Long.parseLong(ss[0]);
        if (ss.length < 2) return d;
        if (!depth.contains(" ")) ss[1] = depth.substring(ss[0].length());
        if (ss[1].equals("MILLISECONDS") || ss[1].contains("MSEC"))
        {
          return d/1000;
        }
        if (ss[1].contains("SEC"))
        {
          return d;
        }
        if (ss[1].contains("MIN"))
        {
          return d * 60;
        }
        if (ss[1].contains("HOUR") || ss[1].contains("HR"))
        {
          return d * 3600;
        }
        if (ss[1].contains("DAY"))
        {
          return d * 3600 * 24;
        }
      }
    }
    catch (Exception whatever) {};
    return d;
  }
}
