package de.desy.tine.queryUtils;

import java.io.ByteArrayOutputStream;
import java.io.DataInput;
import java.io.DataOutput;
import java.io.IOException;
import de.desy.tine.definitions.TStrings;
import de.desy.tine.server.properties.TExportProperty;
import de.desy.tine.stringUtils.StringToBytes;
import de.desy.tine.structUtils.TStructBase;

/**
 * The basic property query structure
 *
 * non-static public fields in exact order of structure
 * (this would be a blitable structure in C# 
 */
public class TPropertyQuery extends TStructBase
{ // Modern (Release 4.0) Property Query structure
  public String prpName; 
  public String prpDescription;
  public String prpRedirection;
  public String prpTag; 
  public String prpTagIn;
  public String prpUnits;
  public float prpMinValue;
  public float prpMaxValue;
  public int prpSize; 
  public int prpSizeIn;
  public int prpNumOverloads;
  public short prpHistoryDepthShort;// Short term history depth (ring buffer size)
  public short prpHistoryDepthLong; // Long term history depth in months
  public byte prpFormat; // Datatype (outbound)
  public byte prpFormatIn; // Datatype (inbound)
  public byte prpAccess; // Access mode(s) for Proptery
  public byte prpGraphType; // Graph stype for Property
  public String rngUnits; // 16 character engineering units for range
  public float rngMinValue; // range minimum
  public float rngMaxValue; // range maximum
  public short numRows; // number of rows in double array
  public short rowSize; // row size in double array
  public short prpArrayType; // property array type
  public byte[] reserved = new byte[6];
  public static final int sizeInBytes = 
    TStrings.PROPERTY_NAME_SIZE +
    TStrings.PROPERTY_DESC_SIZE +
    TStrings.PROPERTY_REDIR_SIZE +
    TStrings.TAG_NAME_SIZE * 2 +
    TStrings.UNITS_SIZE * 2 +
    4 * 12; 
  ByteArrayOutputStream dBuffer = new ByteArrayOutputStream(sizeInBytes);
  public TPropertyQuery()
  {
    prpName = new String(new StringBuffer(TStrings.PROPERTY_NAME_SIZE));
    prpDescription = new String(new StringBuffer(TStrings.PROPERTY_DESC_SIZE));
    prpRedirection = new String(new StringBuffer(TStrings.PROPERTY_REDIR_SIZE));
    prpTag = new String(new StringBuffer(TStrings.TAG_NAME_SIZE));
    prpTagIn = new String(new StringBuffer(TStrings.TAG_NAME_SIZE));
    prpUnits = new String(new StringBuffer(TStrings.UNITS_SIZE));
    rngUnits = new String(new StringBuffer(TStrings.UNITS_SIZE));
  }
  /**
   * Creates a PropertyQuery from a TExportProperty
   * 
   * @param prp the export property 
   * @Param numOverloads the number of overloads for the given property
   */
  public TPropertyQuery(TExportProperty prp, int numOverloads)
  {
    numRows = (short) prp.getDescription().getArrayNumRows();
    this.prpAccess = (byte) prp.getAccessMode();
    this.prpArrayType = prp.getDescription().getArrayType();
    this.prpDescription = prp.getDescription().getText();
    this.prpFormat = (byte) prp.getOutputFormat();
    this.prpFormatIn = (byte) prp.getInputFormat();
    this.prpGraphType = (byte) prp.getDescription().getYRange().getGraphType();
    this.prpHistoryDepthLong = 0; // fill this in later
    this.prpHistoryDepthShort = 0;
    this.prpMaxValue = prp.getDescription().getYRange().getMaxValue();
    this.prpMinValue = prp.getDescription().getYRange().getMinValue();
    this.prpName = prp.getName();
    this.prpNumOverloads = numOverloads;
    this.prpRedirection = prp.getDescription().getRedirection();
    if (this.prpRedirection == null) this.prpRedirection = "";
    this.prpSize = prp.getOutputSize();
    this.prpSizeIn = prp.getInputSize();
    this.prpTag = prp.getOutputTag();
    this.prpTagIn = prp.getInputTag();
    this.prpUnits = prp.getDescription().getYRange().getUnits();
    this.rngMaxValue = prp.getDescription().getXRange().getMaxValue();
    this.rngMinValue = prp.getDescription().getXRange().getMinValue();
    this.rngUnits = prp.getDescription().getXRange().getUnits();
    this.rowSize = (short) prp.getDescription().getArrayRowLength();
  }
  public void writeData(DataOutput dout) throws IOException
  { // NOTE No byte swapping here - use TOutputDataStream!
    dout.write(StringToBytes.stringToFixedBytes(prpName, TStrings.PROPERTY_NAME_SIZE));
    dout.write(StringToBytes.stringToFixedBytes(prpDescription, TStrings.PROPERTY_DESC_SIZE));
    dout.write(StringToBytes.stringToFixedBytes(prpRedirection, TStrings.PROPERTY_REDIR_SIZE));
    dout.write(StringToBytes.stringToFixedBytes(prpTag, TStrings.TAG_NAME_SIZE));
    dout.write(StringToBytes.stringToFixedBytes(prpTagIn, TStrings.TAG_NAME_SIZE));
    dout.write(StringToBytes.stringToFixedBytes(prpUnits, TStrings.UNITS_SIZE));
    dout.writeFloat(prpMinValue);
    dout.writeFloat(prpMaxValue);
    dout.writeInt(prpSize);
    dout.writeInt(prpSizeIn);
    dout.writeInt(prpNumOverloads);
    dout.writeShort(prpHistoryDepthShort);
    dout.writeShort(prpHistoryDepthLong);
    dout.writeByte(prpFormat);
    dout.writeByte(prpFormatIn);
    dout.writeByte(prpAccess);
    dout.writeByte(prpGraphType);
    dout.write(StringToBytes.stringToFixedBytes(rngUnits, TStrings.UNITS_SIZE));
    dout.writeFloat(rngMinValue);
    dout.writeFloat(rngMaxValue);
    dout.writeShort(numRows);
    dout.writeShort(rowSize);
    dout.writeShort(prpArrayType);
    dout.write(reserved);
  }
  public boolean isHistoryProperty()
  {
    return isHistoryProperty(prpName);
  }
  public static boolean isHistoryProperty(String prp)
  {
    if (prp == null) return false;
    if (prp.endsWith(".HIST") || prp.endsWith(".HST") ||
        prp.endsWith(".ARCH") ||prp.endsWith(".ARC") ||
        prp.endsWith(".AR"))
      return true;
    return false;
  }
  /*
   * (non-Javadoc)
   * 
   * @see de.desy.tine.client.TStructIfc#readData(java.io.DataInput)
   */
  public void readData(DataInput din) throws IOException
  {
    prpName = StringToBytes.readFixedString(din, TStrings.PROPERTY_NAME_SIZE);
    prpDescription = StringToBytes.readFixedString(din, TStrings.PROPERTY_DESC_SIZE);
    prpRedirection = StringToBytes.readFixedString(din, TStrings.PROPERTY_REDIR_SIZE);
    prpTag = StringToBytes.readFixedString(din, TStrings.TAG_NAME_SIZE);
    prpTagIn = StringToBytes.readFixedString(din, TStrings.TAG_NAME_SIZE);
    prpUnits = StringToBytes.readFixedString(din, TStrings.UNITS_SIZE);
    prpMinValue = din.readFloat();
    prpMaxValue = din.readFloat();
    prpSize = din.readInt();
    prpSizeIn = din.readInt();
    prpNumOverloads = din.readInt();
    prpHistoryDepthShort = din.readShort();
    prpHistoryDepthLong = din.readShort();
    prpFormat = din.readByte();
    prpFormatIn = din.readByte();
    prpAccess = din.readByte();
    prpGraphType = din.readByte();
    rngUnits = StringToBytes.readFixedString(din, TStrings.UNITS_SIZE);
    rngMinValue = din.readFloat();
    rngMaxValue = din.readFloat();
    numRows = din.readShort();
    rowSize = din.readShort();
    prpArrayType = din.readShort();
    din.readFully(reserved, 0, 6);
  }
}
