package de.desy.tine.server.equipment;

import de.desy.tine.headers.TContract;

/**
 * The property signal handler interface.
 * 
 * A property signal handler can be registered for a given equipment
 * module in order to intercept and react to 
 * various processing signals relevant to the property being accessed.
 * 
 * @author duval
 *
 */
public interface TPropertySignalHandler
{
  /**
   * The property signal handler function
   * 
   * @param signal is the specific signal bit(s) causing the signal to
   * be raised. 
   * @param property is the property string being accessed.
   * @param con is the contract being accessed (if known at the time of
   * the signal, otherwise a null variable).
   * @param status is the call status at the time of the signal.
   * 
   * @see TEquipmentModule.registerPropertySignalHandler()
   * @see TPropertySignal for definitions of the property signal bits
   */
  void handler(int signal,String property,TContract con,int status);
  /**
   * Sets the mask of property signals which should be used to raise the
   * signal.
   * 
   * This should be one of or a combination of the signal bit definitions 
   * in TPropertySignal:
   * <p>TPropertySignal.PS_ACCESS,
   * <br>TPropertySignal.PS_RETRY,
   * <br>TPropertySignal.PS_LATE,
   * <br>TPropertySignal.PS_PENDING,
   * <br>TPropertySignal.PS_SENT,
   * <br>TPropertySignal.PS_CALLED,
   * <br>TPropertySignal.PS_PROCESSED,
   * <br>TPropertySignal.PS_SCHEDULED,
   * <br>or TPropertySignal.PS_ALL
   * 
   * @param mask is a mask containing any of the allowed property signal
   * bits. The mask value of 0 is equivalent to PS_ALL. In order to turn
   * off the property signal dispatch one should set the handler to 'null'.
   */
  void setMask(int mask);
  /**
   * @return the mask of property signals which should be used to raise the
   * signal.
   * This should be one of or a combination of the signal bit definitions 
   * in TPropertySignal:
   * <p>TPropertySignal.PS_ACCESS,
   * <br>TPropertySignal.PS_RETRY,
   * <br>TPropertySignal.PS_LATE,
   * <br>TPropertySignal.PS_PENDING,
   * <br>TPropertySignal.PS_SENT,
   * <br>TPropertySignal.PS_CALLED,
   * <br>TPropertySignal.PS_PROCESSED,
   * <br>TPropertySignal.PS_SCHEDULED,
   * <br>or TPropertySignal.PS_ALL
   * 
   */
  int getMask();
}
