/*
 * CHANGES
 * 
 */
package de.desy.tine.server.properties;

import de.desy.tine.dataUtils.TDataType;
import de.desy.tine.definitions.TAccess;
import de.desy.tine.definitions.TErrorList;
import de.desy.tine.server.logger.MsgLog;

/**
 * An abstract handler for property requests. 
 *  
 * When a client requests access for
 * a certain property, the equipment module calls that property's handler. A
 * property handler has to implement the abstract method call().
 */
public abstract class TPropertyHandler
{
  /**
   * Public interface for call(). Verifies parameters and calls the abstract
   * method call(). This method is called by the equipment module.
   * 
   * @param devName
   *          Device name.
   * @param dout
   *          Output data buffer.
   * @param din
   *          Input data buffer.
   * @param devAccess
   *          Access specification.
   * @return Returns illegal_format if dout/din are null while the
   *         corresponding access is requested. Returns code_failure if an
   *         exception occurs in call(). Redirects the return value of call()
   *         otherwise. TODO change short devAccess to TAccess devAccess (See
   *         TAccess)
   */
  public final int callHandler(String devName, TDataType dout, TDataType din, TAccess devAccess)
  {
    int rv = 0;
    try
    {
      // maybe check isWrite() == true vs registered as a read-only property ?
      // can only do this in the factory thread before calling the handler !
      // range checking, bounds checking can also only be done versus what
      // was registered or let the call() method handle it.
      // illegal_property is already trapped.
      // trapping illegal_device should be left as optional? or strictly
      // enforced ? if so then again in the factory thread.
      if (!devAccess.isWrite() && (dout == null))
        rv = TErrorList.illegal_format;
      else
        rv = call(devName, dout, din, devAccess);
    }
    catch (Exception e)
    { // Catches unknown Exceptions in derived call method.
      e.printStackTrace();
      MsgLog.log("TPropertyHandler","unhandled exception",TErrorList.code_failure,e,0);
      rv = TErrorList.code_failure;
    }
    return rv;
  }
  /**
   * The method which handles all property requests. This method must be
   * implemented by derived classes.
   * 
   * @param devName
   *          Device name.
   * @param dout
   *          Output data buffer.
   * @param din
   *          Input data buffer.
   * @param devAccess
   *          Access specification.
   * @return 0 or some TErrorList constant.
   */
  protected abstract int call(String devName, TDataType dout, TDataType din, TAccess devAccess);
}