/*
 * Created on Feb 10, 2005
 *
 * CHANGES
 * 2005-02-15 JW * Better overload support.
 * 13.7.05 PD added setHandler() method.
 */
package de.desy.tine.server.properties;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.Set;
import java.util.TreeMap;

import de.desy.tine.bitfieldUtils.*;
import de.desy.tine.dataUtils.TDataType;
import de.desy.tine.definitions.*;
import de.desy.tine.server.equipment.TEquipmentModule;
import de.desy.tine.server.histories.THistoryRecord;
import de.desy.tine.server.logger.DbgLog;
import de.desy.tine.server.logger.MsgLog;
import de.desy.tine.server.logger.MsgLog.MsgEntry;
import de.desy.tine.structUtils.TStructRegistry;

/**
 * \internal
 * 
 * A Container for property data. For every property name, TPropertyList
 * contains a list of property data. Each element in that list represents an
 * instance of the property. Multiple properties can have the same name. Every
 * property has a TPropertyHandler.
 */
public class TPropertyList
{
  private int debugLevel = 0;
  private TEquipmentModule eqm = null; // parent container
  public TEquipmentModule getEquipmentModule() { return eqm; }
  //private HashMap<String, LinkedList<PropertyData>> myPropertyData = new HashMap<String, LinkedList<PropertyData>>();
  private TreeMap<String, LinkedList<PropertyData>> myPropertyData = new TreeMap<String, LinkedList<PropertyData>>(String.CASE_INSENSITIVE_ORDER);
  private boolean deviceOriented = false;
  public boolean isDeviceOriented() { return deviceOriented; }
  public void setDeviceOriented(boolean value) { deviceOriented = value; }
  /**
   * \internal
   * 
   * This class holds the data contained in a TPropertyList.
   * 
   * @author jwlg 
   */
  public class PropertyData
  {
    private TExportProperty property;
    private TPropertyHandler handler = null;
    public PropertyData(TExportProperty property, TPropertyHandler propertyHandler)
    {
      this.property = property;
      this.handler = propertyHandler;
    }
    /**
     * @return Returns the property handler.
     */
    public TPropertyHandler getHandler()
    {
      return handler;
    }
    /**
     * @return Returns the property.
     */
    public TExportProperty getProperty()
    {
      return property;
    }
    /**
     * @param handler
     *          The handler to set.
     */
    public void setHandler(TPropertyHandler handler)
    {
      this.handler = handler;
    }
  }
  /**
   * Returns a linked list of all properties which share the same name.
   * 
   * @param propertyName
   * @return
   */
  public LinkedList<TExportProperty> getEqualProperties(String propertyName)
  {
    LinkedList<PropertyData> datalist;
    LinkedList<TExportProperty> rv = new LinkedList<TExportProperty>();
    datalist = (LinkedList<PropertyData>) myPropertyData.get(propertyName);
    if (datalist != null)
    {
      for (Iterator<PropertyData> it = datalist.iterator(); it.hasNext();)
      {
        rv.add(((PropertyData) it.next()).getProperty());
      }
    }
    return rv;
  }
  /**
   * Returns the number of properties which share the same name.
   * 
   * @param propertyName
   * @return
   */
  public int getNumInstances(String propertyName)
  {
    int rv = 0;
    LinkedList<PropertyData> propList = (LinkedList<PropertyData>) myPropertyData.get(propertyName);
    if (propList != null)
    {
      rv = propList.size();
    }
    return rv;
  }
  /**
   * Returns the number of overloads. Provided for compatibility with
   * TExportPropertyTable.
   * 
   * @param propertyName
   * @return
   */
  public int getNumOverloads(String propertyName)
  {
    return (getNumInstances(propertyName) - 1);
  }
  /**
   * Returns the number of different property names. Number of properties, not
   * counting overloads.
   * 
   * @return
   */
  public int countUniqueProperties()
  {
    return myPropertyData.size();
  }
  /**
   * Returns the total number of properties, including overloads
   * 
   * @return
   */
  public int countAllProperties()
  {
    int rv = 0;
    Iterator<LinkedList<PropertyData>> itProps;
    LinkedList<PropertyData> currentData;
    itProps = myPropertyData.values().iterator();
    while (itProps.hasNext())
    {
      currentData = itProps.next();
      if (currentData != null) // null impossible
          rv += currentData.size();
    }
    return rv;
  }
  /**
   * Return a list of PropertyData which contains all properties with a
   * specified name.
   * 
   * @param propertyName
   * @return
   */
  public LinkedList<PropertyData> getPropertyDataList(String propertyName)
  {
    return myPropertyData.get(propertyName);
  }
  /**
   * Returns the property with the specified name
   * 
   * @param propertyName
   * @return
   */
  public TExportProperty getFirstProperty(String propertyName)
  {
    if (propertyName == null) return null;
    LinkedList<PropertyData> lst;
    TExportProperty rv = null;
    PropertyData theData;
    lst = myPropertyData.get(propertyName);
    if (lst != null)
    {
      theData = (PropertyData) lst.getFirst();
      if (theData != null)
      { // not possible!
        rv = (TExportProperty) theData.property;
      }
    }
    return rv;
  }
  public TExportProperty getFirstWriteProperty(String propertyName)
  {
    if (propertyName == null) return null;
    LinkedList<PropertyData> lst;
    TExportProperty rv = null;
    PropertyData theData;
    try 
    {
      if ((lst=myPropertyData.get(propertyName)) != null)
      {
        Iterator<PropertyData> it = lst.iterator();
        while (it.hasNext())
        {
          theData = it.next();
          if (TAccess.isWrite((byte)theData.getProperty().getAccessMode()))
          {
            rv = (TExportProperty) theData.property;
            break;
          }
        }
      }
    }
    catch (Exception ignore) {}
    return rv;
  }
  public LinkedList<TExportProperty> getFilledMetaProperties(String propertyName,int filter)
  {
    LinkedList<TExportProperty> lst = new LinkedList<TExportProperty>();
    TExportProperty p = getFirstProperty(propertyName);
    if (p != null && eqm != null)
    {
      boolean chkDsc = (filter & TMetaProperties.DSC_FLTR) == TMetaProperties.DSC_FLTR;
      boolean chkHst = (filter & TMetaProperties.HST_FLTR) == TMetaProperties.HST_FLTR;
      boolean chkEgu = (filter & TMetaProperties.EGU_FLTR) == TMetaProperties.EGU_FLTR;
      boolean chkXEgu = (filter & TMetaProperties.XEGU_FLTR) == TMetaProperties.XEGU_FLTR;
      boolean chkMax = (filter & TMetaProperties.MAX_FLTR) == TMetaProperties.MAX_FLTR;
      boolean chkXMax = (filter & TMetaProperties.XMAX_FLTR) == TMetaProperties.XMAX_FLTR;
      boolean chkOnl = (filter & TMetaProperties.ONL_FLTR) == TMetaProperties.ONL_FLTR;
      TPropertyDescription pd = p.getDescription();
      THistoryRecord lhr = eqm.getLocalHistoryRecord(p.getName(),"#0");
      if (pd != null)
      {
        String dsc = pd.getText();
        if (chkDsc && dsc != null && dsc.length() > 0)
        {
          lst.add(new TExportProperty(propertyName+".DESC",64,TFormat.CF_TEXT));
        }
        if (chkOnl && TArrayType.isChannel(pd.getArrayType()))
        {
          ArrayList<String> dal = p.getDeviceList();
          if (dal == null || dal.isEmpty())
          {
            lst.add(new TExportProperty(propertyName+".ONLINE",p.getOutputSize(),p.getOutputFormat()));
          }
        }
        TPropertyEGU egu = pd.getYRange();
        String u = egu.getUnits();
        if (chkEgu && u != null && u.compareTo("none") != 0 && u.length() > 0)
        { 
          lst.add(new TExportProperty(propertyName+".EGU",1,TFormat.CF_USTRING));
          if (chkMax && egu.getMaxValue() > 0)
          {
            lst.add(new TExportProperty(propertyName+".MAX",1,TFormat.CF_FLOAT));
            lst.add(new TExportProperty(propertyName+".MIN",1,TFormat.CF_FLOAT));
          }
        }
        egu = pd.getXRange();
        u = egu.getUnits();
        if (chkXEgu && u != null && u.compareTo("none") != 0 && u.length() > 0)
        { 
          lst.add(new TExportProperty(propertyName+".XEGU",1,TFormat.CF_USTRING));
          if (chkXMax && egu.getMaxValue() > 0)
          {
            lst.add(new TExportProperty(propertyName+".XMAX",1,TFormat.CF_FLOAT));
            lst.add(new TExportProperty(propertyName+".XMIN",1,TFormat.CF_FLOAT));
          }
        }
      }
      if (chkHst && lhr != null)
      {
        if (lhr.getRngBufferSize() > 0)
          lst.add(new TExportProperty(propertyName+".HIST",1000,TFormat.CF_DBLDBL));
      }
    }    
    return lst;
  }
  public int setHandler(String propertyName,TPropertyHandler handler)
  {
    LinkedList<PropertyData> lst;
    PropertyData pData;
    if ((lst=myPropertyData.get(propertyName)) == null) return TErrorList.illegal_property;
    if ((pData=(PropertyData)lst.getFirst()) == null) return TErrorList.code_failure;
    pData.setHandler(handler);
    return 0;
  }
  /**
   * Returns the handler of a specified property.
   * 
   * @param propertyName
   * @return
   */
  public TPropertyHandler getFirstPropertyHandler(String propertyName)
  {
    TPropertyHandler rv = null;
    LinkedList<PropertyData> lst = getPropertyDataList(propertyName);
    PropertyData theData = (PropertyData) lst.getFirst();
    if (theData != null)
    {
      rv = (TPropertyHandler) theData.handler;
    }
    return rv;
  }
  /**
   * Find a property handler for a given input and output format.
   * 
   * @param propertyName
   *          Name of the property
   * @param outputFormat
   *          Output format. TFormat.CF_NULL means any format.
   * @see TFormat
   * @return A property handler or null if no property or no suitable property
   *         handler could be found.
   */
  public TPropertyHandler findPropertyHandler(String propertyName, short outputFormat, short inputFormat)
  {
    TPropertyHandler rv = null;
    boolean outputOK, inputOK;
    PropertyData currentData;
    LinkedList<PropertyData> propList = getPropertyDataList(propertyName);
    if (propList != null)
    { // property found
      for (Iterator<PropertyData> it = propList.iterator(); it.hasNext() && (rv == null);)
      {
        currentData = (PropertyData) it.next();
        outputOK = ((outputFormat == TFormat.CF_NULL) || 
                    TFormat.formatsCanMap(currentData.getProperty().getOutputFormat(),outputFormat,false));
        inputOK = ((inputFormat == TFormat.CF_NULL) || 
                   TFormat.formatsCanMap(currentData.getProperty().getInputFormat(),inputFormat,true));
        
        if (outputOK && inputOK)
        {
          rv = currentData.getHandler();
          if (debugLevel > 1)
              DbgLog.log("findPropertyHandler","Output Format: " + currentData.getProperty().getOutputFormat()
                  + "==" + outputFormat);
          break;
        }
        else
        {
          if (debugLevel > 1)
              DbgLog.log("findPropertyHandler","Output Format: " + currentData.getProperty().getOutputFormat()
                  + "!=" + outputFormat);
        }
      }
    }
    else
    {
      DbgLog.log("findPropertyHandler","Property not found");
    }
    return rv;
  }
  public int getAccessDeadband(String propertyName)
  {
    TExportProperty p = getFirstProperty(propertyName);
    if (p != null)
    { // the first registered overload will be sufficient
      return p.getAccessDeadband();
    }
    return 0;    
  }
  public boolean requiresNetworkAccess(String propertyName)
  {
    TExportProperty p = getFirstProperty(propertyName);
    if (p != null)
    { // the first registered overload will be sufficient
      return TAccess.isNetwork(p.getAccessMode());
    }
    return false;
  }
  public boolean enforcesOutput(String propertyName)
  {
    TExportProperty p = getFirstProperty(propertyName);
    if (p != null)
    { // the first registered overload will be sufficient
      return p.isEnforceOutput;
    }
    return false;
  }
  public boolean requiresAsynchronousAccess(String propertyName)
  {
    TExportProperty p = getFirstProperty(propertyName);
    if (p != null)
    { // the first registered overload will be sufficient
      short acc = p.getAccessMode();
      return (acc & TAccess.CA_NOSYNC) == 0 ? false : true;
    }
    return false;
  }
  public boolean requiresConnectedAccess(String propertyName)
  {
    TExportProperty p = getFirstProperty(propertyName);
    if (p != null)
    { // the first registered overload will be sufficient
      return TAccess.isConnected(p.getAccessMode());
    }
    return false;
  }
  /**
   * Returns true if the container contains the specified property.
   * 
   * @param propertyName
   * @return
   */
  public boolean hasProperty(String propertyName)
  {
    return myPropertyData.containsKey(propertyName);
  }
  public boolean isBitfieldProperty(String propertyName,String[] field,String tag[])
  {
    if (propertyName == null) return false;
    int p = propertyName.lastIndexOf('.');
    if (p < 1 || p == propertyName.length()-1) return false;
    String prp = propertyName.substring(0, p);
    if (!hasProperty(prp)) return false;
    String fld = propertyName.substring(p+1);
    TExportProperty xp = getFirstProperty(prp);
    if (!TFormat.isBitfield(xp.getOutputFormat())) return false;
    String t = xp.getOutputTag();
    if (t == null || t.length() == 0) return false;
    if (TBitfieldRegistry.getBitfield(t) == null) return false;
    if (field != null && field.length > 0) field[0] = fld;
    if (tag != null && tag.length > 0) tag[0] = t; 
    return true;
  }
  public boolean isStructProperty(String propertyName,String[] field,String tag[])
  {
    if (propertyName == null) return false;
    int p = propertyName.lastIndexOf('.');
    if (p < 1 || p == propertyName.length()-1) return false;
    String prp = propertyName.substring(0, p);
    if (!hasProperty(prp)) return false;
    String fld = propertyName.substring(p+1);
    TExportProperty xp = getFirstProperty(prp);
    if (xp.getOutputFormat() != TFormat.CF_STRUCT) return false;
    String t = xp.getOutputTag();
    if (t == null || t.length() == 0) return false;
    if (TStructRegistry.get(t) == null) return false;
    if (field != null && field.length > 0) field[0] = fld;
    if (tag != null && tag.length > 0) tag[0] = t; 
    return true;
  }
  /**
   * Returns a Set of all contained property names.
   * 
   * @return
   */
  public Set<String> getPropertyNames()
  {
    return myPropertyData.keySet();
  }
  public void removeProperty(String prop)
  {
	if (myPropertyData.containsKey(prop))
	{
	  myPropertyData.remove(prop);
	}
  }
  /**
   * Adds or overloads a property.
   * 
   * @param newProperty
   * @param newHandler
   */
  public void addProperty(TExportProperty newProperty, TPropertyHandler newHandler)
  {
    addProperty(null, newProperty, newHandler);
  }
  public void addProperty(TEquipmentModule parentEqm, TExportProperty newProperty, TPropertyHandler newHandler)
  {
    if (eqm == null) eqm = parentEqm;
    String propertyName = newProperty.getName();
    LinkedList<PropertyData> propertyList;
    if (myPropertyData.containsKey(propertyName))
    { // This property already exists; overload.
      propertyList = getPropertyDataList(propertyName);
      if (propertyList != null)
      { // null impossible
        propertyList.add(new PropertyData(newProperty, newHandler));
      }
    }
    else
    { // New property
      propertyList = new LinkedList<PropertyData>();
      propertyList.add(new PropertyData(newProperty, newHandler));
      myPropertyData.put(propertyName, propertyList);
      //if ((newProperty.getDescription().getArrayType() & (TArrayType.AT_CHANNEL|TArrayType.AT_STRUCTURED)) != 0)
      { // a CHANNEL or STRUCTURE property => check for specific device names list ...
        if (eqm != null) 
        {
          eqm.getRegisteredNamesFromPropertyFile(newProperty.getName());
          if (!eqm.getRegisteredUsersFromFile(newProperty.gRegisteredUsersList,newProperty.gRegisteredGroupsList,newProperty.getName(),null))
          {
            eqm.getRegisteredUsersFromFile(newProperty.gRegisteredUsersList,newProperty.gRegisteredGroupsList,"property-default",null);
          }
          if (!eqm.getRegisteredNetsFromFile(newProperty.gRegisteredNetsList,newProperty.getName(),null))
          {
            eqm.getRegisteredNetsFromFile(newProperty.gRegisteredNetsList,"property-default",null);
          }
        }
      }
    }
  }
  public short callPropertyHandler(String propertyName, String devName, TDataType dout, TDataType din, TAccess devAccess)
  { // if the property is redirected, it won't have a handler !
    short cc = 0;
    short outputFormat = (dout == null) ? TFormat.CF_NULL : dout.getFormat();
    short inputFormat = (din == null) ? TFormat.CF_NULL : din.getFormat();
    TPropertyHandler theHandler = findPropertyHandler(propertyName, outputFormat, inputFormat);
    if (theHandler != null)
    { // Call handler
      if (debugLevel > 1) DbgLog.log("callPropertyHandler","Calling Property Handler for [" + propertyName + "]");
      try
      {
        TExportProperty p = getFirstProperty(propertyName);
        if (p.isBusy(50)) 
        { // don't re-enter a property handler ...
          cc = TErrorList.operation_busy;
          MsgEntry msg = MsgLog.getLastMessage();
          if (msg != null && msg.getCode() != TErrorList.operation_busy)
          { // only record if it's new
            MsgLog.log("callPropertyHandler", "block "+propertyName+" handler re-entry",cc,null,0);
          }
        }
        else synchronized (theHandler)
        {
          p.setBusy(true);
          cc = (short)theHandler.callHandler(devName, dout, din, devAccess);
          p.setBusy(false);
        }
      }
      catch (Exception e)
      {
        e.printStackTrace();
        MsgLog.log("callPropertyHandler", "unhandled exception "+e.toString(),TErrorList.code_failure,e,0);
        cc = TErrorList.code_failure;
      }
    }
    else
    {
      if (this.hasProperty(propertyName))
      { 
        if (debugLevel > 1) DbgLog.log("callPropertyHandler","handler not implemented for [" + propertyName + "]");
        cc = TErrorList.not_implemented;
      }
      else
      {
        if (debugLevel > 1) DbgLog.log("callPropertyHandler","Illegal Property [" + propertyName + "]");
        cc = TErrorList.illegal_property;
      }
    }
    return cc;
  }
}