package de.desy.tine.dataUtils;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.IOException;
import java.lang.reflect.Array;
import java.lang.reflect.Constructor;
import java.util.Arrays;
import java.util.Vector;
import de.desy.tine.bitfieldUtils.*;
import de.desy.tine.client.*;
import de.desy.tine.definitions.*;
import de.desy.tine.endianUtils.Swap;
import de.desy.tine.exceptions.InputDataLockedException;
import de.desy.tine.exceptions.TineRuntimeErrorException;
import de.desy.tine.headers.*;
import de.desy.tine.server.logger.*;
import de.desy.tine.structUtils.*;
import de.desy.tine.types.*;

/**
 * TINE Data Type object All data transfered client <-> server must be
 *        contained in a TDataType object. Furthermore all constructors of
 *        TDataType take an array of tine data types as input argument, i.e. a
 *        single element should be passed as an array of length 1
 */
public final class TDataType implements Cloneable
{
  /**
   * Converts a value transmitted as an 'unsigned' integer to a java
   *        signed long
   * @param ival
   *          an encoded unsigned integer value
   * @return the corresponding java signed long
   */
  public static long UnsignedInt(int ival)
  {
    long lval = (long) ival;
    if (lval < 0) lval += 4294967296L;
    return lval;
  }
  /**
   * Forms an 'equivalent' unsigned integer for transmission
   * @param lval
   *          a java long value (signed) to be encoded as an unsigned integer
   * @return the encoded unsigned integer
   */
  public static int UnsignedInt(long lval)
  {
    if (lval > 2147483647) lval -= 4294967296L;
    return (int) lval;
  }
  /**
   * Converts a value transmitted as an 'unsigned' short to a java signed
   *        integer
   * @param sval
   *          an encoded unsigned short value
   * @return the corresponding java signed integer
   */
  public static int UnsignedShort(short sval)
  {
    int ival = (int) sval;
    if (ival < 0) ival += 65536;
    return ival;
  }
  public static short UnsignedShort(byte bval)
  {
    short sval = (short) bval;
    if (sval < 0) sval += 256;
    return sval;
  }
  /**
   * Forms an 'equivalent' unsigned short for transmission
   * @param ival
   *          a java integer value (signed) to be encoded as an unsigned short
   * @return the encoded unsigned short
   */
  public static short UnsignedShort(int ival)
  {
    if (ival > 32767) ival -= 65536;
    return (short) ival;
  }
  public static byte UnsignedByte(short sval)
  {
    if (sval > 127) sval -= 256;
    return (byte) sval;
  }
  public static int toInt32(String v)
  {
    try
    {
      return Integer.parseInt(v);
    } catch (Exception ignore) {};
    return 0;
  }
  public static double toDouble(String v)
  {
    try
    {
      return Double.parseDouble(v);
    } catch (Exception ignore) {};
    return 0;    
  }
  public static final TDataType nulltype = new TDataType();
  public boolean acquireDefaultFormat = false;
  public boolean isLocked = false;
  public boolean isDataObjectInSync = false;
  public String lockedMessage = null;
  private boolean buffersBusy = false;
  public static int RPCERR_SIZE = TStrings.STATUS_SIZE;
  private static int dBufferSegSize = TTransport.UDP_BUFFER_SIZE;
  public int dCompletionLength;
  public int dArrayLength = 0;
  public short dFormat = TFormat.CF_NULL;
  private String structKey = null;
  public void setStructureKey(String key) { structKey = key; }
  public String getStructureKey() { return structKey; }
  protected Vector<Object> dVector;
  private boolean removeTagAtLinkTime = false;
  private String arrayDelimiter = new String("\n");
  private String[] arrayElementNames = null;
  public String[] getArrayElementNames()
  {
    return arrayElementNames;
  }
  public void setArrayElementNames(String[] elementNames)
  {
    arrayElementNames = elementNames;
  }
  public String dTag = new String(new StringBuffer(TStrings.TAG_NAME_SIZE));
  private String dTagKey = null; // what was this ?
  public void setTagKey(String key) { dTagKey = key; }
  public String getTagKey() { return dTagKey; }
  public String getDataTypeKey(boolean includeData)
  {
  	String key = "S"+dArrayLength+"F"+dFormat+"T"+dTag;
  	if (includeData) key += "D"+this.toString();
  	return key;
  }
  private TBitfield[] dBitField = null;
  private String dField = null;
  private int dSizeInBytes;
  private ByteArrayOutputStream dBuffer;
  private ByteArrayOutputStream[] dBuffers;
  private boolean[] dBuffersReady;
  private int waitForBuffers()
  {
    for (int counts = 0; counts < 10; counts++)
    {
      try
      {
        Thread.sleep(10);
      }
      catch (InterruptedException e)
      {
      }
      if (numblks == 1) return 0; // don't wait if there's a single MTU
      if (!buffersBusy) return 0;
    }
    MsgLog.log("waitForBuffers", "data stream busy during call to getData()", TErrorList.data_stale, null, 1);
    return TErrorList.data_stale;
  }
  public void resetBuffersReady()
  {
    buffersBusy = false;
    if (dBuffersReady == null && dArrayLength == 0)
    { // a command coming in with an error code ?
      dBuffersReady = new boolean[1];
    }
    if (dBuffersReady == null || dBuffersReady.length < numblks)
    {
      if (numblks > 0)
      {
        dBuffersReady = new boolean[numblks];
      }
      else
      {
        // String msg = dBuffersReady == null ? "no packet management buffers"
        // : "too few packet management buffers";
        // MsgLog.log("resetBuffersReady", msg, TErrorList.code_failure, null, 0);
        return;
      }
    }
    int nresets = 0;
    for (int i = 0; i < numblks; i++)
    {
      if (dBuffersReady[i]) nresets++;
      dBuffersReady[i] = false;
    }
    if (TLinkFactory.debugLevel > 1)
    {
      if (nresets > 0 && nresets < numblks)
        DbgLog.log("resetBuffersReady", "reset " + nresets + " of " + numblks + " rcv buffers");
    }
  }
  public void pushBytes(byte[] bytestream)
  {
    if (dBuffers == null || dBuffers[0] == null || bytestream == null) return;
    try
    {
      dBuffers[0].reset();
      dBuffers[0].write(bytestream);
    }
    catch (IOException e)
    {
      MsgLog.log("TDataType.pushBytes", e.toString(), TErrorList.code_failure, e, 1);
    }
    return;
  }
  public long dTimestamp;
  public int timestamp;
  public int timestampMSEC;
  public int timestampUSEC;
  public int usrDataStamp;
  public int getUserDataStamp()
  {
    return usrDataStamp;
  }
  public void setUserDataStamp(int dataStamp)
  {
    usrDataStamp = dataStamp;
  }
  public int sysDataStamp;
  public int getSystemDataStamp()
  {
    return sysDataStamp;
  }
  public void setSystemDataStamp(int dataStamp)
  {
    sysDataStamp = dataStamp;
  }
  private int syncOffset = 0;
  public int getSynchronizationOffset() { return syncOffset; }
  public void setSynchronizationOffset(int value) { syncOffset = value; }
  public int xferReason;
  public int getUpdateReason() { return xferReason; }
  public int numblks;
  public int blknum;
  public int blkid; 
  private static int blkidSeed = 0;
  public int blkcnt;
  private int mtu = (TTransport.UDP_BUFFER_SIZE - TContract.hdrSizeInBytes - TSubReqHdr.hdrSizeInBytes - TReqHdr.hdrSizeInBytes);
  public int getMtu()
  {
    return mtu;
  }
  public int getMtu(int tineProtocol)
  {
    int lmtu = 0;
    if (tineProtocol > TTransport.legacyProtocolLevel)
    {
      lmtu = (TTransport.UDP_BUFFER_SIZE - TContract.hdrSizeInBytes - TSubReqHdr.hdrSizeInBytes - TReqHdr.hdrSizeInBytes);
    }
    else if (tineProtocol == TTransport.legacyProtocolLevel)
    {
      lmtu = (TTransport.UDP_BUFFER_SIZE - TContract.hdrSizeInBytes - TSubReqHdr.hdrSizeInBytesR4 - TReqHdr.hdrSizeInBytesL);
    }
    else
    {
      lmtu = (TTransport.UDP_BUFFER_SIZE - TContractR3.hdrSizeInBytes - TSubReqHdr.hdrSizeInBytesR3 - TReqHdr.hdrSizeInBytesL);
    }
    int fsiz = TFormat.formatSizeOf(dFormat);
    if (fsiz > 0) lmtu -= lmtu % fsiz;
    return lmtu;
  }
  public void setMtu(int newMtu)
  {
    if (newMtu == mtu) return;
    if (newMtu >= 1000 && newMtu <= 64000)
    {
      mtu = newMtu;
      initCounters();
    }
  }
  public int blksin;
  public int bytesin;
  public boolean hasBeenUpdated = false;
  public boolean hasExtendedBytes = false;
  private Object hDataObject;
  private Object hStructObject;
  public boolean isArrayOfPrimitives()
  {
    if (hDataObject == null) return false;
    try 
    { // the one or two objects which aren't arrays aren't allowed S/R properties
      if (Array.getLength(hDataObject) > 1 && TFormat.isNumberFormat(dFormat))
      {  // a real array of primitive types ...
        return true;
      }
    } catch (Exception ignore) {};  
    return false;
  }
  public Object clone()
  {
    Object o = null;
    try
    {
      o = super.clone();
    }
    catch (CloneNotSupportedException e)
    {
      MsgLog.log("TDataType.clone", e.getMessage(), TErrorList.code_failure, e, 0);
    }
    return o;
  }
  public void resetCounters(int tineProtocol)
  {
    if (tineProtocol > TTransport.legacyProtocolLevel)
    {
      mtu = (TTransport.UDP_BUFFER_SIZE - TContract.hdrSizeInBytes - TSubReqHdr.hdrSizeInBytes - TReqHdr.hdrSizeInBytes);
    }
    if (tineProtocol == TTransport.legacyProtocolLevel)
    {
      mtu = (TTransport.UDP_BUFFER_SIZE - TContract.hdrSizeInBytes - TSubReqHdr.hdrSizeInBytesR4 - TReqHdr.hdrSizeInBytesL);
    }
    if (tineProtocol < TTransport.legacyProtocolLevel)
    {
      mtu = (TTransport.UDP_BUFFER_SIZE - TContractR3.hdrSizeInBytes - TSubReqHdr.hdrSizeInBytesR3 - TReqHdr.hdrSizeInBytesL);
    }
    repackageDataForOutput();
  }
  private void initCounters()
  {
    if (dFormat != TFormat.CF_NULL) mtu -= (mtu % TFormat.formatSizeOf(dFormat));
    numblks = dSizeInBytes / mtu;
    if (dSizeInBytes == 0 || (dSizeInBytes % mtu) > 0) numblks += 1;
    blkidSeed = (blkidSeed+1) % 1000; // make new static seed
    blkid = blkidSeed; // this TDataType is tagged with this blkid
    blknum = 1;
    blkcnt = 0;
  }
  public boolean equals(TDataType dt)
  {
    return equals(true, dt);
  }
  public boolean equals(boolean compareContents, TDataType dt)
  {
    return equals(compareContents,false,dt);
  }
  public boolean equals(boolean compareContents, boolean fixDefault, TDataType dt)
  {
    if (dt == null)
    {
      return dFormat == TFormat.CF_NULL ? true : false;
    }
    if (dt.dFormat == TFormat.CF_NULL && dFormat == TFormat.CF_NULL) return true;
    if (dt.dFormat == TFormat.CF_DEFAULT && fixDefault)
    { // CF_DEFAULT will not know the data type and size and NOT have dBuffers !
      dt.dFormat = dFormat;
      int fsiz = TFormat.formatSizeOf(dFormat);
      if (fsiz > 0) dt.dArrayLength /= fsiz;
      dt.dArrayLength = dArrayLength; // always do this !
      dt.setTag(dTag); // but the tag cannot be returned to the caller via CF_DEFAULT!
      dt.setField(dField);
      dt.setDataObject(dArrayLength, dFormat);
      if (TFormat.isBitfield(dFormat)) dt.setBitField(dBitField);
    }
    if (dt.dFormat == TFormat.CF_DEFAULT && acquireDefaultFormat) return true;
    if (dt.dFormat != dFormat) return false;
    if (dt.dArrayLength == 0) return false;
    if (!TFormat.isAdjustableLength(dt.dFormat) && dt.dArrayLength != dArrayLength) return false;
    if (dt.getTag().compareTo(getTag()) != 0) return false;
    if (compareContents)
    {
      if (dt.dBuffers == null || dBuffers == null) return false;
      int fsiz = TFormat.formatSizeOf(dt.dFormat);
      for (int i = 0; i < dt.dBuffers.length; i++)
      {
        if (dBufferSegSize * i > dArrayLength*fsiz) break;
        if (dBuffers[i] == null || dt.dBuffers[i] == null) return false;
        if (!Arrays.equals(dt.dBuffers[i].toByteArray(), dBuffers[i].toByteArray())) return false;
      }
    }
    return true;
  }
  public int copy(TDataType dt)
  {
    if (dt == null) return TErrorList.argument_list_error;
    if (dt.dFormat == TFormat.CF_NULL && dFormat == TFormat.CF_NULL) return 0;
    if (dt.dFormat != dFormat) return TErrorList.illegal_format;
    if (dt.numblks > this.numblks)
    { // possible if CF_STRING (or other?) variable length format types are used
      // and the capacity needs to be increased on the fly
      this.numblks = dt.numblks;
      reassignBuffers();
    }
    if (TFormat.isAdjustableLength(dt.dFormat) || dt.hasExtendedBytes)
    {
      dSizeInBytes = dt.dSizeInBytes;
    }
    else
    {
      if (dt.dArrayLength > dArrayLength) return TErrorList.dimension_error;
    }
    if (dt.dBuffers == null || dBuffers == null) return TErrorList.un_allocated;
    dCompletionLength = dt.dCompletionLength;
    if (dBuffers == dt.dBuffers) return 0; // no need to do anything
    try
    {
      for (int i=0; i<dt.dBuffers.length && i<dBuffers.length; i++)
      {
        if (dBuffers[i] == null || dt.dBuffers[i] == null) break;
        dBuffers[i].reset();
        dt.dBuffers[i].writeTo(dBuffers[i]);
      }
    }
    catch (Exception e)
    {
      e.printStackTrace();
      MsgLog.log("TDataType.copy", e.toString(), TErrorList.code_failure, e, 0);
    }
    return 0;
  }
  private int copy(TDataType dt, int offset)
  {
    if (dt == null) return TErrorList.argument_list_error;
    if (dArrayLength == dt.dArrayLength && offset == 0) return copy(dt);
    if (dt.dFormat == TFormat.CF_NULL && dFormat == TFormat.CF_NULL) return 0;
    if (dt.dFormat != dFormat) return TErrorList.illegal_format;
    if (dArrayLength == 0 || dt.dArrayLength == 0) return TErrorList.dimension_error;
    if (TFormat.isAdjustableLength(dt.dFormat) || dt.hasExtendedBytes)
    {
      dSizeInBytes = dt.dSizeInBytes;
    }
    if (dt.dBuffers == null || dBuffers == null) 
    {
      return TErrorList.un_allocated;
    }
    int off = offset * dSizeInBytes;
    int len = dArrayLength * dSizeInBytes;
    int nbytes = len;
    int siz;
    dCompletionLength = dArrayLength;
    if (dBuffers == dt.dBuffers) return 0; // no need to do anything
    try
    {
      for (int i = 0, n = 0; i < dt.dBuffers.length && n < dBuffers.length; i++)
      {
        if (dBuffers[n] == null || dt.dBuffers[i] == null) break;
        siz = dt.dBuffers[i].size();
        if (siz <= off)
        {
          off -= siz;
          continue;
        }
        dBuffers[n].reset();
        if (nbytes > siz - off) nbytes = siz - off;
        dBuffers[n].write(dt.dBuffers[i].toByteArray(), off, nbytes);
        if ((nbytes = len - nbytes) <= 0) break;
        off = 0;
        n++;
      }
    }
    catch (Exception e)
    {
      e.printStackTrace();
      MsgLog.log("TDataType.copy", e.toString(), TErrorList.code_failure, e, 0);
    }
    return 0;
  }
  private static double getDouble(String s)
  {
    if (s.compareToIgnoreCase("true") == 0) return 1.0;
    if (s.compareToIgnoreCase("false") == 0) return 0.0;
    return Double.parseDouble(s);    
  }
  private static int getInteger(String s) 
  {
    if (s.compareToIgnoreCase("true") == 0) return 1;
    if (s.compareToIgnoreCase("false") == 0) return 0;
    if (s.startsWith("0x") || s.startsWith("0X") || s.startsWith("&H") || s.startsWith("&h"))
    {
      long lval = Long.parseLong(s.substring(2),16);
      return UnsignedInt(lval);
    }
    return Integer.parseInt(s);
  }
  private static short getShort(String s) 
  {
    if (s.startsWith("0x") || s.startsWith("0X") || s.startsWith("&H") || s.startsWith("&h"))
    {
      int ival = Integer.parseInt(s.substring(2),16);
      return UnsignedShort(ival);
    }
    return Short.parseShort(s);
  }
  private static long getLong(String s) 
  {
    if (s.startsWith("0x") || s.startsWith("0X") || s.startsWith("&H") || s.startsWith("&h"))
      return Long.parseLong(s.substring(2),16);
    return Long.parseLong(s);
  }
  private static byte getByte(String s) 
  {
    if (s.startsWith("0x") || s.startsWith("0X") || s.startsWith("&H") || s.startsWith("&h"))
    {
      short sval = Short.parseShort(s.substring(2),16);
      return UnsignedByte(sval);
    }
    return Byte.parseByte(s);
  }
  /**
   * parse input string into a TDataType
   * 
   * @param text is the text string which is to be parsed into data
   * @param fmtstring is the TINE format code (in string representation)
   * defining the parsing rules
   * 
   * @return a valid TDataType object or null if the input text cannot
   * be parsed into a TDataType
   */
  public static TDataType toTDataType(String text,String fmtstring)
  {
    return toTDataType(text,TFormat.getFormatCode(fmtstring)); 
  }
  //TODO there's an awful lot of format case switching in various methods here.
  // maybe these could be consolidated ?
  /**
   * parse input string into a TDataType
   * 
   * @param text is the text string which is to be parsed into data
   * @param fmtcode is the TINE format code defining the parsing rules
   * 
   * @return a valid TDataType object or null if the input text cannot
   * be parsed into a TDataType
   */
  public static TDataType toTDataType(String text,short fmtcode)
  {
    return toTDataType(text, fmtcode,false);
  }
  public static TDataType toTDataType(String text,short fmtcode,boolean parseLines)
  {
    if (text == null || text.length() == 0) return new TDataType();
    if (text.contains("%20")) text = text.replace("%20", " ");
    String del = " ";
    if (!text.startsWith("\"") && !parseLines)
    { // if quoted -> then there are probably embedded commas
      if (text.indexOf(',') != -1) del = ",";
    }
    String[] vars = new String[text.length()]; // worst case
    String[] v = text.split("\n");
    int siz = 0, len = 0;
    if (TFormat.isVariableLengthString(fmtcode))
    {
      for (int p = 0; p < v.length; p++) vars[siz++] = v[p].trim();
    }
    else if (fmtcode == TFormat.CF_TEXT || fmtcode == TFormat.CF_XML)
    {
      siz = text.length();
    }
    else
    {
      String ssv;
      int segsiz = -1;
      if (parseLines && TFormat.containsString(fmtcode))
      { // instructions from above ...
        segsiz = TFormat.getSegments(fmtcode);
      }
      for (int n = 0; n < v.length; n++)
      { // run thru the lines ...
        String[] sv = v[n].trim().split(del);
        for (int p = 0; p < sv.length; p++)
        {
          if (sv.length < segsiz)
          {
            MsgLog.log("TDataTyp.toTDataType", "input does not have enough arguments to parse data type",TErrorList.parsing_error,null,0);
            return null;              
          }
          if (segsiz < 0)
          { // just scan thru, white space is white space
            ssv = sv[p];
            while (ssv.endsWith("\\") && p < sv.length-1)
            { // the 'delimiter' has been 'escaped' => not a delimiter!)
              p++;
              ssv = ssv.substring(0, ssv.length()-1) + del + sv[p];
            }
          }
          else if (sv.length > segsiz)
          {
            ssv = "";
            if (TFormat.endsWithString(fmtcode))
            { 
              while (p > segsiz && p < sv.length) ssv += del + sv[p++];
            }
            else
            {
              while (p < sv.length-segsiz) ssv += sv[p++] + del;
            }
            if (p < sv.length) ssv += sv[p];
          }
          else
          { // segsiz == sv.length
            ssv = sv[p];
          }
          vars[siz++] = ssv.trim();
        }
      }
    }
    if (siz == 0) return null;
    int i = 0, retrycnt = 2;
    while (retrycnt-- > 0)
    {
      try
      {
        switch (fmtcode)
        {
          case TFormat.CF_DOUBLE:
          {
            double[] vals = new double[siz];
            for (i = 0; i < siz; i++) vals[i] = Double.valueOf(vars[i]);
            return new TDataType(vals);
          }
          case TFormat.CF_INT16:
          {
            short[] vals = new short[siz];
            for (i = 0; i < siz; i++) vals[i] = getShort(vars[i]);
            return new TDataType(vals);
          }
          case TFormat.CF_UINT16:
          {
            short[] vals = new short[siz];
            for (i = 0; i < siz; i++) vals[i] = getShort(vars[i]);
            TDataType tdt = new TDataType(vals);
            tdt.dFormat = TFormat.CF_UINT16;
            return tdt;
          }
          case TFormat.CF_BOOLEAN:
          {
            boolean[] vals = new boolean[siz];
            for (i = 0; i < siz; i++) vals[i] = getInteger(vars[i]) == 0 ? false : true;
            return new TDataType(vals);
          }
          case TFormat.CF_UINT32:
          {
            int[] vals = new int[siz];
            for (i = 0; i < siz; i++) vals[i] = UnsignedInt(getLong(vars[i]));
            TDataType tdt = new TDataType(vals);
            tdt.dFormat = TFormat.CF_UINT32;
            return tdt;
          }
          case TFormat.CF_INT32:
          {
            int[] vals = new int[siz];
            for (i = 0; i < siz; i++) vals[i] = getInteger(vars[i]);
            return new TDataType(vals);
          }
          case TFormat.CF_UINT64: // just fall thru
          case TFormat.CF_INT64:
          {
            long[] vals = new long[siz];
            for (i = 0; i < siz; i++) vals[i] = getLong(vars[i]);
            TDataType tdt = new TDataType(vals);
            if (fmtcode == TFormat.CF_UINT64) tdt.dFormat = fmtcode;
            return tdt;
          }
          case TFormat.CF_FLOAT:
          {
            float[] vals = new float[siz];
            for (i = 0; i < siz; i++) vals[i] = Float.valueOf(vars[i]);
            return new TDataType(vals);
          }
          case TFormat.CF_BYTE:
          {
            if (siz == 1 && vars[0].length() > 1)
            { // single string to be converted to bytes
              try
              {
                return new TDataType(vars[0].getBytes("UTF-8"));
              } catch (Exception e) { throw new NumberFormatException(); }
            }
            byte[] vals = new byte[siz];
            for (i = 0; i < siz; i++) vals[i] = getByte(vars[i]);
            return new TDataType(vals);
          }
          case TFormat.CF_TEXT:
          {
            return new TDataType(text);
          }
          case TFormat.CF_XML:
          {
            TDataType tdt = new TDataType(text);
            tdt.dFormat = TFormat.CF_XML;
            return tdt;
          }
          case TFormat.CF_NAME8:
          {
            NAME8[] vals = new NAME8[siz];
            for (i = 0; i < siz; i++) vals[i] = new NAME8(vars[i]);
            return new TDataType(vals);
          }
          case TFormat.CF_NAME16:
          {
            NAME16[] vals = new NAME16[siz];
            for (i = 0; i < siz; i++) vals[i] = new NAME16(vars[i]);
            return new TDataType(vals);
          }
          case TFormat.CF_NAME32:
          {
            NAME32[] vals = new NAME32[siz];
            for (i = 0; i < siz; i++) vals[i] = new NAME32(vars[i]);
            return new TDataType(vals);
          }
          case TFormat.CF_NAME48:
          {
            NAME48[] vals = new NAME48[siz];
            for (i = 0; i < siz; i++) vals[i] = new NAME48(vars[i]);
            return new TDataType(vals);
          }
          case TFormat.CF_NAME64:
          {
            NAME64[] vals = new NAME64[siz];
            for (i = 0; i < siz; i++) vals[i] = new NAME64(vars[i]);
            return new TDataType(vals);
          }
          case TFormat.CF_NAME8I:
          {
            len = siz / 2;
            if (len == 0) return null;
            NAME8I[] vals = new NAME8I[len];
            for (i = 0; i < len; i++) vals[i] = new NAME8I(vars[i * 2], getInteger(vars[i * 2 + 1]));
            return new TDataType(vals);
          }
          case TFormat.CF_NAME16I:
          {
            len = siz / 2;
            if (len == 0) return null;
            NAME16I[] vals = new NAME16I[len];
            for (i = 0; i < len; i++) vals[i] = new NAME16I(vars[i * 2], getInteger(vars[i * 2 + 1]));
            return new TDataType(vals);
          }
          case TFormat.CF_NAME32I:
          {
            len = siz / 2;
            if (len == 0) return null;
            NAME32I[] vals = new NAME32I[len];
            for (i = 0; i < len; i++) vals[i] = new NAME32I(vars[i * 2], getInteger(vars[i * 2 + 1]));
            return new TDataType(vals);
          }
          case TFormat.CF_NAME48I:
          {
            len = siz / 2;
            if (len == 0) return null;
            NAME48I[] vals = new NAME48I[len];
            for (i = 0; i < len; i++) vals[i] = new NAME48I(vars[i * 2], getInteger(vars[i * 2 + 1]));
            return new TDataType(vals);
          }
          case TFormat.CF_NAME64I:
          {
            len = siz / 2;
            if (len == 0) return null;
            NAME64I[] vals = new NAME64I[len];
            for (i = 0; i < len; i++) vals[i] = new NAME64I(vars[i * 2], getInteger(vars[i * 2 + 1]));
            return new TDataType(vals);
          }
          case TFormat.CF_NAME64DBL:
          {
            len = siz / 2;
            if (len == 0) return null;
            NAME64DBL[] vals = new NAME64DBL[len];
            for (i = 0; i < len; i++) vals[i] = new NAME64DBL(vars[i * 2], getDouble(vars[i * 2 + 1]));
            return new TDataType(vals);
          }
          case TFormat.CF_NAME64DBLDBL:
          {
            len = siz / 3;
            if (len == 0) return null;
            NAME64DBLDBL[] vals = new NAME64DBLDBL[len];
            for (i = 0; i < len; i++) vals[i] = new NAME64DBLDBL(vars[i * 3], getDouble(vars[i * 3 + 1]), getDouble(vars[i * 3 + 2]));
            return new TDataType(vals);
          }
          case TFormat.CF_NAME32DBLDBL:
          {
            len = siz / 3;
            if (len == 0) return null;
            NAME32DBLDBL[] vals = new NAME32DBLDBL[len];
            for (i = 0; i < len; i++) vals[i] = new NAME32DBLDBL(vars[i * 3], getDouble(vars[i * 3 + 1]), getDouble(vars[i * 3 + 2]));
            return new TDataType(vals);
          }
          case TFormat.CF_NAME16DBLDBL:
          {
            len = siz / 3;
            if (len == 0) return null;
            NAME16DBLDBL[] vals = new NAME16DBLDBL[len];
            for (i = 0; i < len; i++) vals[i] = new NAME16DBLDBL(vars[i * 3], getDouble(vars[i * 3 + 1]), getDouble(vars[i * 3 + 2]));
            return new TDataType(vals);
          }
          case TFormat.CF_STRING:
          {
            return new TDataType(vars, siz);
          }
          case TFormat.CF_KEYVALUE:
          {
            KEYVALUE[] vals = new KEYVALUE[siz];
            for (i = 0; i < siz; i++) vals[i] = new KEYVALUE(vars[i]);
            return new TDataType(vals, siz);
          }
          case TFormat.CF_UNAME:
          {
            len = siz / 5;
            if (len == 0) return null;
            UNAME[] vals = new UNAME[len];
            for (i = 0; i < len; i++)
              vals[i] = new UNAME(getInteger(vars[i * 5]), Float.valueOf(vars[i * 5 + 1]),
                  Float.valueOf(vars[i * 5 + 2]), Float.valueOf(vars[i * 5 + 3]),
                  vars[i * 5 + 4]);
            return new TDataType(vals);
          }
          case TFormat.CF_USTRING:
          {
            if (siz == 4)
            { // special case: empty string passed in a single USTRING
              siz++;
              vars = Arrays.copyOf(vars, siz);
              vars[4] = "";
            }
            len = siz / 5;
            if (len == 0) return null;
            USTRING[] vals = new USTRING[len];
            for (i = 0; i < len; i++)
              vals[i] = new USTRING(getInteger(vars[i * 5]), Float.valueOf(vars[i * 5 + 1]),
                  Float.valueOf(vars[i * 5 + 2]), getInteger(vars[i * 5 + 3]), vars[i * 5 + 4]);
            return new TDataType(vals);
          }
          case TFormat.CF_INTINT:
          {
            len = siz / 2;
            if (len == 0) return null;
            INTINT[] vals = new INTINT[len];
            for (i = 0; i < len; i++)
              vals[i] = new INTINT(getInteger(vars[i * 2]), getInteger(vars[i * 2 + 1]));
            return new TDataType(vals);
          }
          case TFormat.CF_FLTINT:
          {
            len = siz / 2;
            if (len == 0) return null;
            FLTINT[] vals = new FLTINT[len];
            for (i = 0; i < len; i++)
              vals[i] = new FLTINT(Float.valueOf(vars[i * 2]), getInteger(vars[i * 2 + 1]));
            return new TDataType(vals);
          }
          case TFormat.CF_FLTFLT:
          {
            len = siz / 2;
            if (len == 0) return null;
            FLTFLT[] vals = new FLTFLT[len];
            for (i = 0; i < len; i++)
              vals[i] = new FLTFLT(Float.valueOf(vars[i * 2]), Float.valueOf(vars[i * 2 + 1]));
            return new TDataType(vals);
          }
          case TFormat.CF_DBLDBL:
          {
            len = siz / 2;
            if (len == 0) return null;
            DBLDBL[] vals = new DBLDBL[len];
            for (i = 0; i < len; i++)
              vals[i] = new DBLDBL(Double.valueOf(vars[i * 2]), Double.valueOf(vars[i * 2 + 1]));
            return new TDataType(vals);
          }
          case TFormat.CF_DBLTIME:
          {
            len = siz / 2;
            if (len == 0) return null;
            DBLTIME[] vals = new DBLTIME[len];
            for (i = 0; i < len; i++)
              vals[i] = new DBLTIME(Double.valueOf(vars[i * 2]), Double.valueOf(vars[i * 2 + 1]));
            return new TDataType(vals);
          }
          case TFormat.CF_NAME16FI:
          {
            len = siz / 3;
            if (len == 0) return null;
            NAME16FI[] vals = new NAME16FI[len];
            for (i = 0; i < len; i++)
              vals[i] = new NAME16FI(vars[i * 3], getInteger(vars[i * 3 + 2]), Float.valueOf(vars[i * 3 + 1]));
            return new TDataType(vals);
          }
          case TFormat.CF_NAME16II:
          {
            len = siz / 3;
            if (len == 0) return null;
            NAME16II[] vals = new NAME16II[len];
            for (i = 0; i < len; i++)
              vals[i] = new NAME16II(vars[i * 3], getInteger(vars[i * 3 + 1]), getInteger(vars[i * 3 + 2]));
            return new TDataType(vals);
          }
          case TFormat.CF_FLTFLTINT:
          {
            len = siz / 3;
            if (len == 0) return null;
            FLTFLTINT[] vals = new FLTFLTINT[len];
            for (i = 0; i < len; i++)
              vals[i] = new FLTFLTINT(Float.valueOf(vars[i * 3]), Float.valueOf(vars[i * 3 + 1]), getInteger(vars[i * 3 + 2]));
            return new TDataType(vals);
          }
          case TFormat.CF_FLTINTINT:
          {
            len = siz / 3;
            if (len == 0) return null;
            FLTINTINT[] vals = new FLTINTINT[len];
            for (i = 0; i < len; i++)
              vals[i] = new FLTINTINT(Float.valueOf(vars[i * 3]), getInteger(vars[i * 3 + 1]),
                  getInteger(vars[i * 3 + 2]));
            return new TDataType(vals);
          }
          case TFormat.CF_DBLDBLDBL:
          {
            len = siz / 3;
            if (len == 0) return null;
            DBLDBLDBL[] vals = new DBLDBLDBL[len];
            for (i = 0; i < len; i++)
              vals[i] = new DBLDBLDBL(Double.valueOf(vars[i * 3]), Double.valueOf(vars[i * 3 + 1]),Double.valueOf(vars[i * 3 + 2]));
            return new TDataType(vals);
          }
          case TFormat.CF_INTFLTFLTFLT:
          {
            len = siz / 4;
            if (len == 0) return null;
            INTFLTFLTFLT[] vals = new INTFLTFLTFLT[len];
            for (i = 0; i < len; i++)
              vals[i] = new INTFLTFLTFLT(getInteger(vars[i * 4]), Float.valueOf(vars[i * 4 + 1]),
                  Float.valueOf(vars[i * 4 + 2]), Float.valueOf(vars[i * 4 + 3]));
            return new TDataType(vals);
          }
          case TFormat.CF_INTFLTINT:
          {
            len = siz / 3;
            if (len == 0) return null;
            INTFLTINT[] vals = new INTFLTINT[len];
            for (i = 0; i < len; i++)
              vals[i] = new INTFLTINT(getInteger(vars[i * 3]), Float.valueOf(vars[i * 3 + 1]),
                  getInteger(vars[i * 3 + 2]));
            return new TDataType(vals);
          }
          case TFormat.CF_INTINTFLTFLT:
          {
            len = siz / 4;
            if (len == 0) return null;
            INTINTFLTFLT[] vals = new INTINTFLTFLT[len];
            for (i = 0; i < len; i++)
              vals[i] = new INTINTFLTFLT(Float.valueOf(vars[i * 4 + 2]), Float.valueOf(vars[i * 4 + 3]),
                  getInteger(vars[i * 4]), getInteger(vars[i * 4 + 1]));
            return new TDataType(vals);
          }
          case TFormat.CF_INTINTINTINT:
          {
            len = siz / 4;
            if (len == 0) return null;
            INTINTINTINT[] vals = new INTINTINTINT[len];
            for (i = 0; i < len; i++)
              vals[i] = new INTINTINTINT(getInteger(vars[i * 4]), getInteger(vars[i * 4 + 1]),
                  getInteger(vars[i * 4 + 2]), getInteger(vars[i * 4 + 3]));
            return new TDataType(vals);
          }
          case TFormat.CF_FLTINTFLTINT:
          {
            len = siz / 4;
            if (len == 0) return null;
            FLTINTFLTINT[] vals = new FLTINTFLTINT[len];
            for (i = 0; i < len; i++)
              vals[i] = new FLTINTFLTINT(Float.valueOf(vars[i * 4]), getInteger(vars[i * 4 + 1]),
                  Float.valueOf(vars[i * 4 + 2]), getInteger(vars[i * 4 + 3]));
            return new TDataType(vals);
          }
          case TFormat.CF_WINDOW:
          {
            len = siz / 4;
            if (len == 0) return null;
            WINDOW[] vals = new WINDOW[len];
            for (i = 0; i < len; i++)
              vals[i] = new WINDOW(getInteger(vars[i * 4]), getInteger(vars[i * 4 + 1]),
                  getInteger(vars[i * 4 + 2]), getInteger(vars[i * 4 + 3]));
            return new TDataType(vals);
          }
        }
        return null;
      }
      catch (NumberFormatException e)
      {
        siz = i;
        MsgLog.log("TDataType.toTDataType", "unable to parse input",TErrorList.parsing_error,e,0);
      }
    }
    return null;
  }
  
  /**
   * Create a TDataType object that contains the given value. The data type is created based
   * on the type of the given object, by calling one of the appropriate constructors.
   * 
   * @param value the value to insert into the data type
   * @return data type
   */
  public static TDataType toTDataType(Object value)
  {
      if (value instanceof ASPECTRUM) {
          return new TDataType(new ASPECTRUM[]{(ASPECTRUM)value});
      } else if (value instanceof ASPECTRUM[]) {
          return new TDataType((ASPECTRUM[])value);
      } else if (value instanceof boolean[]) {
          return new TDataType((boolean[])value);
      } else if (value instanceof Boolean) {
          return new TDataType(new boolean[]{(boolean)value});
      } else if (value instanceof byte[]) {
          return new TDataType((byte[])value);
      } else if (value instanceof Byte) {
          return new TDataType(new byte[]{(byte)value});
      } else if (value instanceof char[]) {
          return new TDataType((char[])value);
      } else if (value instanceof Character) {
          return new TDataType(new char[]{(char)value});
      } else if (value instanceof double[]) {
          return new TDataType((double[])value);
      } else if (value instanceof Double) {
          return new TDataType(new double[]{(double)value});
      } else if (value instanceof float[]) {
          return new TDataType((float[])value);
      } else if (value instanceof Float) {
          return new TDataType(new float[]{(float)value});
      } else if (value instanceof IMAGE[]) {
          return new TDataType((IMAGE[])value);
      } else if (value instanceof IMAGE) {
          return new TDataType((IMAGE)value);
      } else if (value instanceof int[]) {
          return new TDataType((int[])value);
      } else if (value instanceof Integer) {
          return new TDataType(new int[]{(int)value});
      } else if (value instanceof KEYVALUE[]) {
          return new TDataType((KEYVALUE[])value);
      } else if (value instanceof KEYVALUE) {
          return new TDataType(new KEYVALUE[]{(KEYVALUE)value});
      } else if (value instanceof long[]) {
          return new TDataType((long[])value);
      } else if (value instanceof Long) {
          return new TDataType(new long[]{(long)value});
      } else if (value instanceof MDA) {
          return new TDataType((MDA)value);
      } else if (value instanceof short[]) {
          return new TDataType((short[])value);
      } else if (value instanceof Short) {
          return new TDataType(new short[]{(short)value});
      } else if (value instanceof SPECTRUM) {
          return new TDataType((SPECTRUM)value);
      } else if (value instanceof GSPECTRUM) {
        return new TDataType((GSPECTRUM)value);
      } else if (value instanceof String[]) {
          return new TDataType((String[])value);
      } else if (value instanceof String) {
          return new TDataType((String)value);
      } else if (value instanceof StringBuffer) {
          return new TDataType((StringBuffer)value);
      } else if (value instanceof TCompoundDataObject[]) {
          return new TDataType((TCompoundDataObject[])value);
      } else if (value instanceof TCompoundDataObject) {
          return new TDataType(new TCompoundDataObject[]{(TCompoundDataObject)value});
      } else if (value instanceof TTaggedStructure[]) {
          return new TDataType((TTaggedStructure[])value);
      } else if (value instanceof TTaggedStructure) {
          return new TDataType(new TTaggedStructure[]{(TTaggedStructure)value});
      }
      return null;
  }
  
  /**
   * Sets the delimiter used in a string representation of an array
   */
  public void setArrayDelimiter(String delimiter)
  {
    if (TFormat.isBitfield(dFormat))
    {
      if (dBitField != null) for (TBitfield bf : dBitField) bf.setFieldSeparator(delimiter);
    }
    if (arrayDelimiter.compareTo(delimiter) == 0) return;
    arrayDelimiter = new String(delimiter);
  }
  /**
   * Get the delimiter used in a string representation of an array
   * @return the delimiter used in a string representation of an array
   */
  public String getArrayDelimiter()
  {
    return arrayDelimiter;
  }
  /**
   * NULL element
   */
  public TDataType()
  {
    dFormat = TFormat.CF_NULL;
    dArrayLength = 0;
    dSizeInBytes = 0;
    initCounters();
  }
  private void makeDataObject(int dataSize, short dataFormat)
  {
    dFormat = dataFormat >= 0 ? dataFormat : (short) (256 + dataFormat);
    switch (dFormat % 256)
    {
      case TFormat.CF_DOUBLE:
        hDataObject = new double[dataSize];
        break;
      case TFormat.CF_FLOAT:
        hDataObject = new float[dataSize];
        break;
      case TFormat.CF_BITFIELD16:
      case TFormat.CF_UINT16:
      case TFormat.CF_INT16:
        hDataObject = new short[dataSize];
        break;
      case TFormat.CF_BOOLEAN:
        hDataObject = new boolean[dataSize];
        break;
      case TFormat.CF_BITFIELD32:
      case TFormat.CF_UINT32:
      case TFormat.CF_INT32:
        hDataObject = new int[dataSize];
        break;
      case TFormat.CF_UINT64:
      case TFormat.CF_INT64:
        hDataObject = new long[dataSize];
        break;
      case TFormat.CF_NAME16:
        hDataObject = new NAME16[dataSize];
        for (int i = 0; i < dataSize; i++)
          ((NAME16[]) hDataObject)[i] = new NAME16();
        break;
      case TFormat.CF_NAME32:
        hDataObject = new NAME32[dataSize];
        for (int i = 0; i < dataSize; i++)
          ((NAME32[]) hDataObject)[i] = new NAME32();
        break;
      case TFormat.CF_NAME48:
        hDataObject = new NAME48[dataSize];
        for (int i = 0; i < dataSize; i++)
          ((NAME48[]) hDataObject)[i] = new NAME48();
        break;
      case TFormat.CF_NAME64:
        hDataObject = new NAME64[dataSize];
        for (int i = 0; i < dataSize; i++)
          ((NAME64[]) hDataObject)[i] = new NAME64();
        break;
      case TFormat.CF_NAME8:
        hDataObject = new NAME8[dataSize];
        for (int i = 0; i < dataSize; i++)
          ((NAME8[]) hDataObject)[i] = new NAME8();
        break;
      case TFormat.CF_HISTORY:
        hDataObject = new HISTORY[dataSize];
        break;
      case TFormat.CF_STRUCT:
      case TFormat.CF_BYTE:
      case TFormat.CF_BITFIELD8:
        hDataObject = new byte[dataSize];
        break;
      case TFormat.CF_DBLDBL:
        hDataObject = new DBLDBL[dataSize];
        for (int i = 0; i < dataSize; i++)
          ((DBLDBL[]) hDataObject)[i] = new DBLDBL();
        break;
      case TFormat.CF_DBLTIME:
        hDataObject = new DBLTIME[dataSize];
        for (int i = 0; i < dataSize; i++)
          ((DBLTIME[]) hDataObject)[i] = new DBLTIME();
        break;
      case TFormat.CF_DBLDBLDBL:
        hDataObject = new DBLDBLDBL[dataSize];
        for (int i = 0; i < dataSize; i++)
          ((DBLDBLDBL[]) hDataObject)[i] = new DBLDBLDBL();
        break;
      case TFormat.CF_FLTINT:
        hDataObject = new FLTINT[dataSize];
        for (int i = 0; i < dataSize; i++)
          ((FLTINT[]) hDataObject)[i] = new FLTINT();
        break;
      case TFormat.CF_INTINT:
        hDataObject = new INTINT[dataSize];
        for (int i = 0; i < dataSize; i++)
          ((INTINT[]) hDataObject)[i] = new INTINT();
        break;
      case TFormat.CF_NAME8I:
        hDataObject = new NAME8I[dataSize];
        for (int i = 0; i < dataSize; i++)
          ((NAME8I[]) hDataObject)[i] = new NAME8I();
        break;
      case TFormat.CF_NAME16I:
        hDataObject = new NAME16I[dataSize];
        for (int i = 0; i < dataSize; i++)
          ((NAME16I[]) hDataObject)[i] = new NAME16I();
        break;
      case TFormat.CF_NAME16FI:
        hDataObject = new NAME16FI[dataSize];
        for (int i = 0; i < dataSize; i++)
          ((NAME16FI[]) hDataObject)[i] = new NAME16FI();
        break;
      case TFormat.CF_NAME16II:
        hDataObject = new NAME16II[dataSize];
        for (int i = 0; i < dataSize; i++)
          ((NAME16II[]) hDataObject)[i] = new NAME16II();
        break;
      case TFormat.CF_NAME32I:
        hDataObject = new NAME32I[dataSize];
        for (int i = 0; i < dataSize; i++)
          ((NAME32I[]) hDataObject)[i] = new NAME32I();
        break;
      case TFormat.CF_NAME48I:
        hDataObject = new NAME48I[dataSize];
        for (int i = 0; i < dataSize; i++)
          ((NAME48I[]) hDataObject)[i] = new NAME48I();
        break;
      case TFormat.CF_NAME64I:
        hDataObject = new NAME64I[dataSize];
        for (int i = 0; i < dataSize; i++)
          ((NAME64I[]) hDataObject)[i] = new NAME64I();
        break;
      case TFormat.CF_NAME16DBLDBL:
        hDataObject = new NAME16DBLDBL[dataSize];
        for (int i = 0; i < dataSize; i++)
          ((NAME16DBLDBL[]) hDataObject)[i] = new NAME16DBLDBL();
        break;
      case TFormat.CF_NAME32DBLDBL:
        hDataObject = new NAME32DBLDBL[dataSize];
        for (int i = 0; i < dataSize; i++)
          ((NAME32DBLDBL[]) hDataObject)[i] = new NAME32DBLDBL();
        break;
      case TFormat.CF_NAME64DBL:
        hDataObject = new NAME64DBL[dataSize];
        for (int i = 0; i < dataSize; i++)
          ((NAME64DBL[]) hDataObject)[i] = new NAME64DBL();
        break;
      case TFormat.CF_NAME64TIME:
        hDataObject = new NAME64TIME[dataSize];
        for (int i = 0; i < dataSize; i++)
          ((NAME64TIME[]) hDataObject)[i] = new NAME64TIME();
        break;
      case TFormat.CF_NAME64DBLDBL:
        hDataObject = new NAME64DBLDBL[dataSize];
        for (int i = 0; i < dataSize; i++)
          ((NAME64DBLDBL[]) hDataObject)[i] = new NAME64DBLDBL();
        break;
      case TFormat.CF_NAME64DBLDBLDBL:
        hDataObject = new NAME64DBLDBLDBL[dataSize];
        for (int i = 0; i < dataSize; i++)
          ((NAME64DBLDBLDBL[]) hDataObject)[i] = new NAME64DBLDBLDBL();
        break;
      case TFormat.CF_ASPECTRUM:
        hDataObject = new ASPECTRUM[dataSize];
        for (int i = 0; i < dataSize; i++)
          ((ASPECTRUM[]) hDataObject)[i] = new ASPECTRUM();
        dTag = new String(""+dataSize);
        break;
      case TFormat.CF_SPECTRUM:
        hDataObject = new SPECTRUM(dataSize);
        break;
      case TFormat.CF_GSPECTRUM:
        hDataObject = new GSPECTRUM(dataSize);
        break;
      case TFormat.CF_AIMAGE:
        hDataObject = new IMAGE[dataSize];
        for (int i = 0; i < dataSize; i++)
        ((IMAGE[]) hDataObject)[i] = new IMAGE();
        dTag = new String(""+dataSize);
        break;
      case TFormat.CF_IMAGE:
        hDataObject = new IMAGE(dataSize);
        //by jbobnar: the hDataObject is not an array in this case
//        for (int i = 0; i < dataSize; i++)
//          ((IMAGE[]) hDataObject)[i] = new IMAGE();
//        dTag = new String(""+dataSize);
        break;
      case TFormat.CF_MDA:
        MDA mda = new MDA(dataSize);
        mda.setCarriedFormat(TFormat.getFormatCode(dTag));
        hDataObject = mda;
        break;
      case TFormat.CF_INTFLTINT:
        hDataObject = new INTFLTINT[dataSize];
        for (int i = 0; i < dataSize; i++)
          ((INTFLTINT[]) hDataObject)[i] = new INTFLTINT();
        break;
      case TFormat.CF_XML:
      case TFormat.CF_TEXT:
        hDataObject = new char[dataSize];
        break;
      case TFormat.CF_STRING:
        hDataObject = new String[dataSize];
        for (int i = 0; i < dataSize; i++)
          ((String[])hDataObject)[i] = "";
        break;
      case TFormat.CF_KEYVALUE:
        hDataObject = new KEYVALUE[dataSize];
        for (int i = 0; i < dataSize; i++)
          ((KEYVALUE[])hDataObject)[i] = new KEYVALUE("none","empty");
        break;
      case TFormat.CF_USTRING:
        hDataObject = new USTRING[dataSize];
        for (int i = 0; i < dataSize; i++)
          ((USTRING[]) hDataObject)[i] = new USTRING();
        break;
      case TFormat.CF_UNAME:
        hDataObject = new UNAME[dataSize];
        for (int i = 0; i < dataSize; i++)
          ((UNAME[]) hDataObject)[i] = new UNAME();
        break;
      case TFormat.CF_WINDOW:
        hDataObject = new WINDOW[dataSize];
        for (int i = 0; i < dataSize; i++)
          ((WINDOW[]) hDataObject)[i] = new WINDOW();
        break;
      case TFormat.CF_ADDRESS:
        hDataObject = new INTINTINTINT[dataSize];
        for (int i = 0; i < dataSize; i++)
          ((INTINTINTINT[]) hDataObject)[i] = new INTINTINTINT();
        break;
      case TFormat.CF_INTINTINT:
        hDataObject = new INTINTINT[dataSize];
        for (int i = 0; i < dataSize; i++)
          ((INTINTINT[]) hDataObject)[i] = new INTINTINT();
        break;
      case TFormat.CF_POINT:
        hDataObject = new FLTFLT[dataSize];
        for (int i = 0; i < dataSize; i++)
          ((FLTFLT[]) hDataObject)[i] = new FLTFLT();
        break;
      case TFormat.CF_FILTER:
        hDataObject = new INTFLTFLTFLT[dataSize];
        for (int i = 0; i < dataSize; i++)
          ((INTFLTFLTFLT[]) hDataObject)[i] = new INTFLTFLTFLT();
        break;
      case TFormat.CF_FLTFLTINT:
        hDataObject = new FLTFLTINT[dataSize];
        for (int i = 0; i < dataSize; i++)
          ((FLTFLTINT[]) hDataObject)[i] = new FLTFLTINT();
        break;
      case TFormat.CF_FLTINTINT:
        hDataObject = new FLTINTINT[dataSize];
        for (int i = 0; i < dataSize; i++)
          ((FLTINTINT[]) hDataObject)[i] = new FLTINTINT();
        break;
      case TFormat.CF_FLTINTFLTINT:
        hDataObject = new FLTINTFLTINT[dataSize];
        for (int i = 0; i < dataSize; i++)
          ((FLTINTFLTINT[]) hDataObject)[i] = new FLTINTFLTINT();
        break;
      case TFormat.CF_FWINDOW:
        hDataObject = new INTINTFLTFLT[dataSize];
        for (int i = 0; i < dataSize; i++)
          ((INTINTFLTFLT[]) hDataObject)[i] = new INTINTFLTFLT();
        break;
    }
  }
  /**
   * An unbound DataType object of the specified size and format.
   * @param dataSize
   *          is the array size of the given format.
   * @param dataFormat
   *          is the tine data type of the object to create.
   * @note A call to getData() must be made in order to retrieve the incoming
   *       data into the application array parameters.
   */
  public TDataType(int dataSize, short dataFormat)
  {
    this(dataSize,dataFormat,0); 
  }
  public TDataType(int dataSize, short dataFormat,String tag)
  {
    this(dataSize,dataFormat,tag,0); 
  }
  public TDataType(int dataSize, short dataFormat,int nblks)
  {
    this(dataSize,dataFormat,null,nblks);
  }
  public TDataType(int dataSize, short dataFormat,String tag,int nblks)
  {
    dFormat = dataFormat >= 0 ? dataFormat : (short) (256 + dataFormat);
    int fsiz = TFormat.formatSizeOf(dFormat);
    if (tag != null) dTag = tag;
    switch (dFormat % 256)
    {
      case TFormat.CF_DEFAULT:
        acquireDefaultFormat = true;
        dSizeInBytes = dataSize;
        dArrayLength = dataSize;
        blknum = 1;
        blkcnt = 0;
        return; // no allocation
      case TFormat.CF_MDA:
        fsiz = TFormat.getCarriedFormatSize(dTag);
        if (fsiz < 1) fsiz = 1; // if dTag has not been assigned yet
        makeDataObject(dataSize, dataFormat);
        break;
      case TFormat.CF_STRUCT:
        int ssz = TStructRegistry.getSizeInBytes(tag);
        if (ssz > 0) dataSize *= ssz;
        makeDataObject(dataSize, dataFormat);
        break;
      default:
        makeDataObject(dataSize, dataFormat);
        break;
    }
    dArrayLength = dataSize;
    dSizeInBytes = dataSize * fsiz + TFormat.getFormatHeaderSize(dFormat);
    if (dFormat == TFormat.CF_HISTORY)
    {
      int hsiz = TFormat.getCarriedFormatSize(dTag);
      if (hsiz > 0)
      {
        hsiz += TFormat.getHistoryHeaderSize();
        dArrayLength = dataSize / hsiz;
      }
    }
    initCounters();
    if (numblks < nblks) numblks = nblks;
    dBuffers = new ByteArrayOutputStream[numblks];
    dBuffersReady = new boolean[numblks];
    for (int n = 0; n < numblks; n++)
    {
      dBuffers[n] = new ByteArrayOutputStream(dBufferSegSize);
    }
  }
  /**
   * Returns the associated data timestamp as traditional UNIX timestamp
   *        (seconds since 1.1.1970).
   */
  public int getTimeStamp()
  {
    return timestamp;
  }
  /**
   * Returns the associated data timestamp as JAVA timestamp
   *        (milliseconds since 1.1.1970).
   */
  public long getDataTimeStamp()
  {
    if (dTimestamp > 0) return dTimestamp;
    return (long) timestamp * 1000 + (long) (timestampUSEC /1000);
  }
  /**
   * Sets the associated data timestamp from a JAVA timestamp
   *        (milliseconds since 1.1.1970).
   */
  public void setDataTimeStamp(long TimeStamp)
  {
    dTimestamp = TimeStamp;
    timestamp = (int) (TimeStamp / 1000);
    timestampMSEC = (int) (TimeStamp % 1000);
    timestampUSEC = timestampMSEC * 1000;
  }
  /**
   * Sets the associated data timestamp from a TINE timestamp (seconds
   *        since 1.1.1970 + fractional seconds).
   */
  public void setDataTimeStamp(double TimeStamp)
  {
    long t = ((long) TimeStamp) * 1000 + ((long) (TimeStamp * 1000) % 1000);
    setDataTimeStamp(t);
  }
  /**
   * Returns the associated data completion length
   */
  public int getLastDataSize()
  {
    return dCompletionLength;
  }
  /**
   * A subset of an array of booleans. This corresponds to tine format
   *        CF_BOOLEAN
   * @param data
   *          is a reference to the boolean array used to create and bind to the
   *          TDataType object
   */
  public TDataType(byte[] data)
  {
    initTDataType(data, data.length);
  }
  /**
   * A subset of an array of bytes
   * @param data
   *          is a reference to the byte array used to create and bind to the
   *          TDataType object
   * @param length
   *          is the subset array length to use.
   */
  public TDataType(byte[] data, int length)
  {
    initTDataType(data, length);
  }
  private void initTDataType(byte[] data, int length)
  {
    if (data == null || data.length == 0 || length < 1) return;
    try
    {
      if (length > data.length) length = data.length;
      dFormat = TFormat.CF_BYTE;
      dArrayLength = length;
      dSizeInBytes = length;
      initCounters();
      dBuffers = new ByteArrayOutputStream[numblks];
      dBuffersReady = new boolean[numblks];
      for (int n = 0, i = 0; n < numblks; n++)
      {
        dBuffers[n] = new ByteArrayOutputStream(dBufferSegSize);
        DataOutputStream ds = new DataOutputStream(dBuffers[n]);
        for (; i < dArrayLength && i < (n + 1) * mtu; i++)
          ds.writeByte(data[i]);
        ds.close();
      }
      hDataObject = data;
    }
    catch (IOException e)
    {
      e.printStackTrace();
      MsgLog.log("initTDataType", e.getMessage(), TErrorList.code_failure, e, 0);
    }
  }
  /**
   * An array of 'structs' with the given tag or untagged structs if
   *        tag.length == 0 User-defined structures are allowed in tine and not
   *        too difficult to use. It is strongly encouraged to make use of
   *        'tagged' structures whereby the server and client both register the
   *        structure contents with the tine engine and apply a tag. Only then
   *        can issues of 'byte-swapping' and structure boundaries be handled
   *        correctly. As java does not have structures, the associated class is
   *        responsible for providing a toStruct() method which returns a byte
   *        stream containing the contents of the desired structure.
   * @param data
   *          is a reference to the byte array stream containing the 'struct'
   *          used to create and bind to the TDataType object.
   * @param tag
   *          is the structure tag associated with the 'struct.'
   */
  public TDataType(byte[] data, String tag)
  {
    if (data == null || data.length == 0) return;
    dTag = tag;
    try
    {
      dFormat = TFormat.CF_STRUCT;
      dArrayLength = data.length;
      dSizeInBytes = data.length;
      initCounters();
      dBuffers = new ByteArrayOutputStream[numblks];
      dBuffersReady = new boolean[numblks];
      for (int n = 0, i = 0; n < numblks; n++)
      {
        dBuffers[n] = new ByteArrayOutputStream(dBufferSegSize);
        DataOutputStream ds = new DataOutputStream(dBuffers[n]);
        for (; i < dArrayLength && i < (n + 1) * mtu; i++)
          ds.writeByte(data[i]);
        byte[] b = dBuffers[n].toByteArray();
        dBuffers[n].reset();
        ds.write(b, 0, b.length);
        ds.close();
      }
      hDataObject = data;
    }
    catch (IOException e)
    {
      e.printStackTrace();
      MsgLog.log("TDataType", e.getMessage(), TErrorList.code_failure, e, 0);
    }
  }
  /**
   * A Tagged Structure User-defined structures are allowed in tine and
   *        not too difficult to use. It is strongly encouraged to make use of
   *        'tagged' structures whereby the server and client both register the
   *        structure contents with the tine engine and apply a tag. Only then
   *        can issues of 'byte-swapping' and structure boundaries be handled
   *        correctly. As java does not have structures, the associated class is
   *        responsible for providing a toStruct() method which returns a byte
   *        stream containing the contents of the desired structure.
   * @param data
   *          is a reference to a TTaggedStructure object.
   */
  public TDataType(TTaggedStructure data)
  {
    if (data == null)
    {
      MsgLog.log("TDataType", "Tagged Structure data type unallocated !", TErrorList.invalid_structure_tag,
          null, 0);
      return;
    }
    dTag = data.getName();
    try
    {
      byte[] hb = data.toByteArray();
      dFormat = TFormat.CF_STRUCT;
      int dsiz = data.getSizeInBytes();
      int xsiz = data.getExtendedSizeInBytes();
      dSizeInBytes = dsiz + xsiz;
      dArrayLength = dSizeInBytes; // ??
      initCounters();
      dBuffers = new ByteArrayOutputStream[numblks];
      dBuffersReady = new boolean[numblks];
      for (int n = 0, i = 0; n < numblks; n++)
      {
        dBuffers[n] = new ByteArrayOutputStream(dBufferSegSize);
        DataOutputStream ds = new DataOutputStream(dBuffers[n]);
        for (; i < dSizeInBytes && i < (n + 1) * mtu && i < hb.length; i++)
          ds.writeByte(hb[i]);
        if (xsiz > 0) 
        { // put the extended bytes at the end
          int istart = i;
          hb = data.toExtendedByteArray();
          for (; i < dSizeInBytes && i < (n + 1) * mtu && i < istart + hb.length; i++)
            ds.writeByte(hb[i]);
        }
        ds.close();
      }
      hDataObject = data.getByteArray();
      hStructObject = data;
    }
    catch (IOException e)
    {
      e.printStackTrace();
      MsgLog.log("TDataType", e.getMessage(), TErrorList.code_failure, e, 0);
    }
  }
  /**
   * A array of Tagged Structures User-defined structures are allowed in
   *        tine and not too difficult to use. It is strongly encouraged to make
   *        use of 'tagged' structures whereby the server and client both
   *        register the structure contents with the tine engine and apply a
   *        tag. Only then can issues of 'byte-swapping' and structure
   *        boundaries be handled correctly. As java does not have structures,
   *        the associated class is responsible for providing a toStruct()
   *        method which returns a byte stream containing the contents of the
   *        desired structure.
   * @param data
   *          is a reference to an array of TTaggedStructure object.
   */
  public TDataType(TTaggedStructure[] data)
  {
    if (data == null || data.length == 0)// || data[0] == null)
    {
      MsgLog.log("TDataType", "Tagged Structure data type unallocated !", TErrorList.invalid_structure_tag,null, 0);
      return;
    }
    try
    {
      if (data[0] == null) 
      {
        Class<?> clz = (data).getClass().getComponentType();
        Constructor<?> con = clz.getDeclaredConstructor();
        data[0] = (TTaggedStructure)con.newInstance();
      }
    } 
    catch (Exception x) 
    {
      //x.printStackTrace();  
      MsgLog.log("TDataType", "could not instantiate Tagged Structure data type !", TErrorList.invalid_structure_tag,
          null, 0);
      return;
    };
    for (int k = 1; k < data.length; k++)
      if (data[k] == null) data[k] = (TTaggedStructure) data[0].clone();
    try
    {
      dTag = data[0].getName();
      int dsiz = data[0].getSizeInBytes();
      if (dsiz == 0)
      {
        MsgLog.log("TDataType", "Tagged Structure not registered !", TErrorList.invalid_structure_tag,null, 0);
        return;        
      }
      int xsiz = 0;
      if (data[0].hasExtendedSpace())
      { // suddenly, it's more complicated !
        hasExtendedBytes = true;
        for (int k=0; k<data.length; k++)
        { // set the boundary of the extended space and get the total
          data[k].setExtendedOffset(dsiz * data.length);
          data[k].fixExtendedBytes();
          xsiz += data[k].getExtendedSizeInBytes();
        }
      }
      byte[] hbytes = new byte[dsiz * data.length + xsiz];
      dFormat = TFormat.CF_STRUCT;
      dArrayLength = dsiz * data.length + xsiz;
      dSizeInBytes = dsiz * data.length + xsiz;
      initCounters();
      dBuffers = new ByteArrayOutputStream[numblks];
      dBuffersReady = new boolean[numblks];
      byte[] bsrc;
      for (int k = 0; k < data.length; k++)
      {
        bsrc = data[k].toByteArray();
        if (bsrc == null || bsrc.length == 0) break;
        System.arraycopy(bsrc, 0, hbytes, k * dsiz, dsiz);
      }
      if (data[0].hasExtendedSpace())
      { // now pump the extended bytes onto the stream
        int xoff = data.length * dsiz;
        for (int k = 0; k < data.length; k++)
        {
          xsiz = data[k].getExtendedSizeInBytes();
          System.arraycopy(data[k].toExtendedByteArray(), 0, hbytes, xoff, xsiz);
          xoff += xsiz;
        }
      }
      for (int n = 0, i = 0; n < numblks; n++)
      {
        dBuffers[n] = new ByteArrayOutputStream(dBufferSegSize);
        DataOutputStream ds = new DataOutputStream(dBuffers[n]);
        for (; i < dArrayLength && i < (n + 1) * mtu; i++)
          ds.writeByte(hbytes[i]);
        ds.close();
      }
      hDataObject = hbytes;
      hStructObject = data;
    }
    catch (IOException e)
    {
      e.printStackTrace();
      MsgLog.log("TDataType", e.getMessage(), TErrorList.code_failure, e, 0);
    }
  }
  /**
   * An array of shorts (2-byte integers) This corresponds to tine format
   *        CF_SHORT
   * @param data
   *          is a reference to the short array used to create and bind to the
   *          TDataType object
   */
  public TDataType(short[] data)
  {
    initTDataType(data, data.length);
  }
  /**
   * A subset of an array of shorts (2-byte integers) This corresponds to
   *        tine format CF_SHORT
   * @param data
   *          is a reference to the short array used to create and bind to the
   *          TDataType object
   * @param length
   *          is the subset array length to use.
   */
  public TDataType(short[] data, int length)
  {
    initTDataType(data, length);
  }
  private void initTDataType(short[] data, int length)
  {
    if (data == null || data.length == 0 || length < 1) return;
    try
    {
      if (length > data.length) length = data.length;
      dFormat = TFormat.CF_INT16;
      dArrayLength = length;
      dSizeInBytes = length * 2;
      initCounters();
      dBuffers = new ByteArrayOutputStream[numblks];
      dBuffersReady = new boolean[numblks];
      for (int n = 0, i = 0; n < numblks; n++)
      {
        dBuffers[n] = new ByteArrayOutputStream(dBufferSegSize);
        DataOutputStream ds = new DataOutputStream(dBuffers[n]);
        for (; i < dArrayLength && i < (n + 1) * mtu / 2; i++)
          ds.writeShort(Swap.Short(data[i]));
        ds.close();
      }
      hDataObject = data;
    }
    catch (IOException e)
    {
      e.printStackTrace();
      MsgLog.log("initTDataType", e.getMessage(), TErrorList.code_failure, e, 0);
    }
  }
  /**
   * A single short value This method is only useful for quick creation
   *        of data 'down' objects (i.e. data sets sent to the server). As there
   *        is no reference to the argument passed, it cannot be used to bind
   *        returning data to the object passed as the argument.
   * @param data
   *          is a single short value used to create the TDataType object.
   */
  public TDataType(short data)
  {
    try
    {
      dFormat = TFormat.CF_INT16;
      dArrayLength = 1;
      dSizeInBytes = 2;
      initCounters();
      dBuffers = new ByteArrayOutputStream[numblks];
      dBuffersReady = new boolean[numblks];
      dBuffers[0] = new ByteArrayOutputStream(dBufferSegSize);
      DataOutputStream ds = new DataOutputStream(dBuffers[0]);
      ds.writeShort(data);
      byte[] b = dBuffers[0].toByteArray();
      Swap.Bytes(b, 2);
      dBuffers[0].reset();
      ds.write(b, 0, b.length);
      ds.close();
      hDataObject = null;
    }
    catch (IOException e)
    {
      e.printStackTrace();
      MsgLog.log("TDataType", e.getMessage(), TErrorList.code_failure, e, 0);
    }
  }
  /**
   * An array of ints (4-byte integers) This corresponds to tine format
   *        CF_LONG
   * @param data
   *          is a reference to the int array used to create and bind to the
   *          TDataType object
   */
  public TDataType(int[] data)
  {
    initTDataType(data, data.length);
  }
  /**
   * A subset of an array of ints (4-byte integers) This corresponds to
   *        tine format CF_LONG
   * @param data
   *          is a reference to the int array used to create and bind to the
   *          TDataType object
   * @param length
   *          is the subset array length to use.
   */
  public TDataType(int[] data, int length)
  {
    initTDataType(data, length);
  }
  private void initTDataType(int[] data, int length)
  {
    if (data == null || data.length == 0 || length < 1) return;
    try
    {
      if (length > data.length) length = data.length;
      dFormat = TFormat.CF_INT32;
      dArrayLength = length;
      dSizeInBytes = length * 4;
      initCounters();
      dBuffers = new ByteArrayOutputStream[numblks];
      dBuffersReady = new boolean[numblks];
      for (int n = 0, i = 0; n < numblks; n++)
      {
        dBuffers[n] = new ByteArrayOutputStream(dBufferSegSize);
        DataOutputStream ds = new DataOutputStream(dBuffers[n]);
        for (; i < dArrayLength && i < (n + 1) * mtu / 4; i++)
          ds.writeInt(Swap.Int(data[i]));
        ds.close();
      }
      hDataObject = data;
    }
    catch (IOException e)
    {
      e.printStackTrace();
      MsgLog.log("initTDataType", e.getMessage(), TErrorList.code_failure, e, 0);
    }
  }
  /**
   * A subset of an array of booleans. This corresponds to tine format
   *        CF_BOOLEAN
   * @param data
   *          is a reference to the boolean array used to create and bind to the
   *          TDataType object
   */
  public TDataType(boolean[] data)
  {
    initTDataType(data, data.length);
  }
  /**
   * A subset of an array of booleans. This corresponds to tine format
   *        CF_BOOLEAN
   * @param data
   *          is a reference to the boolean array used to create and bind to the
   *          TDataType object
   * @param length
   *          is the subset array length to use.
   */
  public TDataType(boolean[] data, int length)
  {
    initTDataType(data, length);
  }
  private void initTDataType(boolean[] data, int length)
  {
    if (data == null || data.length == 0 || length < 1) return;
    try
    {
      if (length > data.length) length = data.length;
      dFormat = TFormat.CF_BOOLEAN;
      dArrayLength = length;
      dSizeInBytes = length * 4;
      initCounters();
      int[] idata = new int[data.length];
      for (int i = 0; i < data.length; i++)
        idata[i] = data[i] ? 1 : 0;
      dBuffers = new ByteArrayOutputStream[numblks];
      dBuffersReady = new boolean[numblks];
      for (int n = 0, i = 0; n < numblks; n++)
      {
        dBuffers[n] = new ByteArrayOutputStream(dBufferSegSize);
        DataOutputStream ds = new DataOutputStream(dBuffers[n]);
        for (; i < dArrayLength && i < (n + 1) * mtu / 4; i++)
          ds.writeInt(Swap.Int(idata[i]));
        ds.close();
      }
      hDataObject = data;
    }
    catch (IOException e)
    {
      e.printStackTrace();
      MsgLog.log("initTDataType", e.getMessage(), TErrorList.code_failure, e, 0);
    }
  }
  /**
   * An array of floats This corresponds to tine format CF_FLOAT
   * @param data
   *          is a reference to the float array used to create and bind to the
   *          TDataType object
   */
  public TDataType(float[] data)
  {
    initTDataType(data, data.length);
  }
  /**
   * A subset of an array of floats This corresponds to tine format
   *        CF_FLOAT
   * @param data
   *          is a reference to the float array used to create and bind to the
   *          TDataType object
   * @param length
   *          is the subset array length to use.
   */
  public TDataType(float[] data, int length)
  {
    initTDataType(data, length);
  }
  private void initTDataType(float[] data, int length)
  {
    if (data == null || data.length == 0 || length < 1) return;
    try
    {
      if (length > data.length) length = data.length;
      dFormat = TFormat.CF_FLOAT;
      dArrayLength = length;
      dSizeInBytes = length * 4;
      initCounters();
      dBuffers = new ByteArrayOutputStream[numblks];
      dBuffersReady = new boolean[numblks];
      for (int n = 0, i = 0; n < numblks; n++)
      {
        dBuffers[n] = new ByteArrayOutputStream(dBufferSegSize);
        DataOutputStream ds = new DataOutputStream(dBuffers[n]);
        for (; i < dArrayLength && i < (n + 1) * mtu / 4; i++)
          ds.write(Swap.Float((data[i])));
        ds.close();
      }
      hDataObject = data;
    }
    catch (IOException e)
    {
      e.printStackTrace();
      MsgLog.log("initTDataType", e.getMessage(), TErrorList.code_failure, e, 0);
    }
  }
  /**
   * An array of tine data objects (e.g. FLTINTs, FLTFLTs, NAME16s, etc.)
   *        can be bound to a tine data type object. Tine data type objects are
   *        used in data links between client and server.
   * @param data
   *          is a reference to the data objects used to create and bind to the
   *          TDataType object
   */
  public TDataType(TCompoundDataObject[] data)
  {
    initTDataType(data, data.length);
  }
  /**
   * An array of tine data objects (e.g. FLTINTs, FLTFLTs, NAME16s, etc.)
   *        can be bound to a tine data type object. Tine data type objects are
   *        used in data links between client and server.
   * @param data
   *          is a reference to the data objects used to create and bind to the
   *          TDataType object
   * @param length
   *          is the subset array length to use.
   */
  public TDataType(TCompoundDataObject[] data, int length)
  {
    initTDataType(data, length);
  }
  private void initTDataType(TCompoundDataObject[] data, int length)
  {
    if (data == null || data.length == 0 || length < 1) return;
    try
    {
      Object tdo = data[0] != null ? data[0] : data.getClass().getComponentType().getDeclaredConstructor().newInstance();
      if (length > data.length) length = data.length;
      String tag = ((TCompoundDataObject) tdo).getTag();
      if (tag != null) dTag = tag;        
      dFormat = ((TCompoundDataObject) tdo).getFormat();
      int elementSize = TFormat.formatSizeOf(dFormat);
      if (dFormat == TFormat.CF_HISTORY)
      {
        short cfmt = ((HISTORY[])data)[0].getCarriedFormat();
        elementSize = TFormat.getHistoryHeaderSize() +
                      TFormat.getFormatHeaderSize(cfmt) +
                      TFormat.formatSizeOf(cfmt); 
        switch (cfmt)
        { // let null pointer exceptions fly if they're there ...
          case TFormat.CF_IMAGE:
            elementSize += ((HISTORY[])data)[0].getDataObject().getArrayLength();
            break;
          case TFormat.CF_KEYVALUE:
          case TFormat.CF_STRING:
            elementSize += TFormat.STRING_STD_CAPACITY;
            break;
          default:
            break;
        }
      }
      dArrayLength = length;
      dSizeInBytes = length * elementSize;
      initCounters();
      dBuffers = new ByteArrayOutputStream[numblks];
      dBuffersReady = new boolean[numblks];
      for (int n = 0, i = 0; n < numblks; n++)
      {
        dBuffers[n] = new ByteArrayOutputStream(dBufferSegSize);
        DataOutputStream ds = new DataOutputStream(dBuffers[n]);
        for (; i < length && i < (n + 1) * mtu / elementSize; i++)
          setObjectContents(ds, data, i); //
        ds.close();
      }
      setObjectHandle(data); //
    }
    catch (NullPointerException e)
    {
      e.printStackTrace();
      MsgLog.log("initTDataType", e.getMessage(), TErrorList.invalid_data, e, 0);
    }
    catch (IOException e)
    {
      e.printStackTrace();
      MsgLog.log("initTDataType", e.getMessage(), TErrorList.io_error, e, 0);
    }
    catch (Exception e)
    {
      e.printStackTrace();
      MsgLog.log("initTDataType", e.getMessage(), TErrorList.code_failure, e, 0);
    }
  }
  public TDataType(MDA data)
  {
    try
    {
      dFormat = TFormat.CF_MDA;
      dArrayLength = data.getTotalLength();
      dTag = data.getTag();
      int elementSize = TFormat.getCarriedFormatSize(dTag); 
      dSizeInBytes = TFormat.getFormatHeaderSize(dFormat) + dArrayLength * elementSize;      
      initCounters();
      dBuffers = new ByteArrayOutputStream[numblks];
      dBuffersReady = new boolean[numblks];
      byte[] bdata = data.toByteArray();
      for (int n = 0, i = 0; n < numblks; n++)
      {
        dBuffers[n] = new ByteArrayOutputStream(dBufferSegSize);
        DataOutputStream ds = new DataOutputStream(dBuffers[n]);
        for (; i < dSizeInBytes && i < (n + 1) * mtu; i++)
          ds.writeByte(bdata[i]);
        byte[] b = dBuffers[n].toByteArray();
        dBuffers[n].reset();
        ds.write(b, 0, b.length);
        ds.close();
      }
      hDataObject = data;
    }
    catch (NullPointerException e)
    {
      e.printStackTrace();
      MsgLog.log("initTDataType", e.getMessage(), TErrorList.invalid_data, e, 0);
    }
    catch (IOException e)
    {
      e.printStackTrace();
      MsgLog.log("initTDataType", e.getMessage(), TErrorList.code_failure, e, 0);
    }
  }
  /**
   * A SPECTRUM object. This corresponds to tine format CF_SPECTRUM
   * @param data
   *          is a reference to the SPECTRUM object used to create and bind to
   *          the TDataType object
   */
  public TDataType(SPECTRUM data)
  {
    try
    {
      dFormat = TFormat.CF_SPECTRUM;
      dArrayLength = data.length;
      dSizeInBytes = data.length * 4 + 96;
      initCounters();
      dBuffers = new ByteArrayOutputStream[numblks];
      dBuffersReady = new boolean[numblks];
      byte[] bdata = data.toByteArray();
      for (int n = 0, i = 0; n < numblks; n++)
      {
        dBuffers[n] = new ByteArrayOutputStream(dBufferSegSize);
        DataOutputStream ds = new DataOutputStream(dBuffers[n]);
        for (; i < dSizeInBytes && i < (n + 1) * mtu; i++)
          ds.writeByte(bdata[i]);
        byte[] b = dBuffers[n].toByteArray();
        dBuffers[n].reset();
        ds.write(b, 0, b.length);
        ds.close();
      }
      hDataObject = data;
    }
    catch (NullPointerException e)
    {
      e.printStackTrace();
      MsgLog.log("initTDataType", e.getMessage(), TErrorList.invalid_data, e, 0);
    }
    catch (IOException e)
    {
      e.printStackTrace();
      MsgLog.log("initTDataType", e.getMessage(), TErrorList.code_failure, e, 0);
    }
  }
  public TDataType(GSPECTRUM data)
  {
    try
    {
      dFormat = TFormat.CF_GSPECTRUM;
      dArrayLength = data.length;
      dSizeInBytes = data.length * 4 + GSPECTRUM.hdrSizeInBytes;
      initCounters();
      dBuffers = new ByteArrayOutputStream[numblks];
      dBuffersReady = new boolean[numblks];
      byte[] bdata = data.toByteArray();
      for (int n = 0, i = 0; n < numblks; n++)
      {
        dBuffers[n] = new ByteArrayOutputStream(dBufferSegSize);
        DataOutputStream ds = new DataOutputStream(dBuffers[n]);
        for (; i < dSizeInBytes && i < (n + 1) * mtu; i++)
          ds.writeByte(bdata[i]);
        byte[] b = dBuffers[n].toByteArray();
        dBuffers[n].reset();
        ds.write(b, 0, b.length);
        ds.close();
      }
      hDataObject = data;
    }
    catch (NullPointerException e)
    {
      e.printStackTrace();
      MsgLog.log("initTDataType", e.getMessage(), TErrorList.invalid_data, e, 0);
    }
    catch (IOException e)
    {
      e.printStackTrace();
      MsgLog.log("initTDataType", e.getMessage(), TErrorList.code_failure, e, 0);
    }
  }
  public TDataType(ASPECTRUM[] data)
  {
    initTDataType(data,data.length);
  }
  public TDataType(ASPECTRUM[] data,int length)
  {
    initTDataType(data,length);
  }
  private void initTDataType(ASPECTRUM[] data,int length)
  {
    if (data == null || data.length == 0 || length < 1) return;
    try
    {
      int len = length;
      dFormat = TFormat.CF_ASPECTRUM;
      dTag = new String(""+len);
      int totalHdrBytes = 0, totalSpectraBytes = 0;
      for (int i=0; i<len; i++)
      {
        if (((ASPECTRUM[])data)[i] == null) 
          ((ASPECTRUM[])data)[i] = new ASPECTRUM();
        totalHdrBytes += ASPECTRUM.hdrSizeInBytes + 8;
        totalSpectraBytes += ((ASPECTRUM[])data)[i].length * 4;
      }
      dArrayLength = totalHdrBytes + totalSpectraBytes;
      dSizeInBytes = totalHdrBytes + totalSpectraBytes;
      initCounters();
      dBuffers = new ByteArrayOutputStream[numblks];
      dBuffersReady = new boolean[numblks];
      byte[] b, bdata = new byte[totalHdrBytes+totalSpectraBytes];
      int off = 0, xoff = totalHdrBytes;
      for (int i=0; i<len; i++)
      {
        ((ASPECTRUM[])data)[i].setExtendedOffset(xoff);
        b = ((ASPECTRUM[])data)[i].toByteArray();
        xoff = ((ASPECTRUM[])data)[i].getExtendedOffset();
        System.arraycopy(b, 0, bdata, off, b.length);
        off += b.length;
      }
      for (int i=0; i<len; i++)
      {
        b = ((ASPECTRUM[])data)[i].toSpectrumByteArray();
        if (b == null) continue;
        System.arraycopy(b, 0, bdata, off, b.length);
        off += b.length;
      }
      for (int n = 0, i = 0; n < numblks; n++)
      {
        dBuffers[n] = new ByteArrayOutputStream(dBufferSegSize);
        DataOutputStream ds = new DataOutputStream(dBuffers[n]);
        for (; i < dSizeInBytes && i < (n + 1) * mtu; i++)
          ds.writeByte(bdata[i]);
        ds.close();
      }
      hDataObject = data;
    }
    catch (NullPointerException e)
    {
      e.printStackTrace();
      MsgLog.log("initTDataType", e.getMessage(), TErrorList.invalid_data, e, 0);
    }
    catch (IOException e)
    {
      e.printStackTrace();
      MsgLog.log("initTDataType", e.getMessage(), TErrorList.code_failure, e, 0);
    }
  }
  private void initTDataType(IMAGE[] data,int length)
  {
    if (data == null || data.length == 0 || length < 1) return;
    try
    {
      int len = length;
      dFormat = TFormat.CF_AIMAGE;
      dTag = new String(""+len);
      int totalHdrBytes = 0, totalFrameBytes = 0;
      for (int i=0; i<len; i++)
      {
        if (((IMAGE[])data)[i] == null) 
          ((IMAGE[])data)[i] = new IMAGE();
        totalHdrBytes += IMAGE.HEADER_SIZE + 8;
        totalFrameBytes += ((IMAGE[])data)[i].getImageFrameSize();
        //totalFrameBytes += ((IMAGE[])data)[i].getFrameHeader().appendedFrameSize;
      }
      dArrayLength = totalHdrBytes + totalFrameBytes;
      dSizeInBytes = totalHdrBytes + totalFrameBytes;
      initCounters();
      dBuffers = new ByteArrayOutputStream[numblks];
      dBuffersReady = new boolean[numblks];
      byte[] b, bdata = new byte[totalHdrBytes+totalFrameBytes];
      int off = 0, xoff = totalHdrBytes;
      for (int i=0; i<len; i++)
      {
        ((IMAGE[])data)[i].setExtendedOffset(xoff);
        b = ((IMAGE[])data)[i].toAdjustableHeaderByteArray();
        xoff += ((IMAGE[])data)[i].getExtendedSize();
        System.arraycopy(b, 0, bdata, off, b.length);
        off += b.length;
      }
      for (int i=0; i<len; i++)
      {
        b = ((IMAGE[])data)[i].toFrameByteArray();
        if (b == null) continue;
        System.arraycopy(b, 0, bdata, off, b.length);
        off += b.length;
      }
      for (int n = 0, i = 0; n < numblks; n++)
      {
        dBuffers[n] = new ByteArrayOutputStream(dBufferSegSize);
        DataOutputStream ds = new DataOutputStream(dBuffers[n]);
        for (; i < dSizeInBytes && i < (n + 1) * mtu; i++)
          ds.writeByte(bdata[i]);
        ds.close();
      }
      hDataObject = data;
    }
    catch (NullPointerException e)
    {
      e.printStackTrace();
      MsgLog.log("initTDataType", e.getMessage(), TErrorList.invalid_data, e, 0);
    }
    catch (IOException e)
    {
      e.printStackTrace();
      MsgLog.log("initTDataType", e.getMessage(), TErrorList.code_failure, e, 0);
    }
  }
  public TDataType(IMAGE[] data)
  {
    initTDataType(data, data.length); 
  }
  public TDataType(IMAGE[] data,int length)
  {
    initTDataType(data, length); 
  }
  /**
   * An IMAGE object. This corresponds to tine format CF_IMAGE
   * @param data
   *          is a reference to the IMAGE object used to create and bind to the
   *          TDataType object
   */
  public TDataType(IMAGE data)
  {
    if (data == null) return;
    try
    {
      dFormat = TFormat.CF_IMAGE;
      dArrayLength = data.getImageFrameSize();
      dSizeInBytes = data.getSizeInBytes();
      initCounters();
      dBuffers = new ByteArrayOutputStream[numblks];
      dBuffersReady = new boolean[numblks];
      byte[] bdata = data.toByteArray();
      for (int n = 0, i = 0; n < numblks; n++)
      {
        dBuffers[n] = new ByteArrayOutputStream(dBufferSegSize);
        DataOutputStream ds = new DataOutputStream(dBuffers[n]);
        for (; i < dSizeInBytes && i < (n + 1) * mtu; i++)
          ds.writeByte(bdata[i]);
        byte[] b = dBuffers[n].toByteArray();
        dBuffers[n].reset();
        ds.write(b, 0, b.length);
        ds.close();
      }
      hDataObject = data;
    }
    catch (NullPointerException e)
    {
      e.printStackTrace();
      MsgLog.log("initTDataType", e.getMessage(), TErrorList.invalid_data, e, 0);
    }
    catch (IOException e)
    {
      e.printStackTrace();
      MsgLog.log("initTDataType", e.getMessage(), TErrorList.code_failure, e, 0);
    }
  }
  public void reassignBuffers()
  {
    if (dBuffers == null) return;
    if (numblks <= dBuffers.length) return; // nothing to do
    ByteArrayOutputStream[] newBuffers = new ByteArrayOutputStream[numblks];
    boolean[] newBuffersReady = new boolean[numblks];
    for (int n = 0; n < numblks; n++)
    {
      if (n < dBuffers.length)
      { // pass these along
        newBuffers[n] = dBuffers[n];
        newBuffersReady[n] = dBuffersReady[n];
      }
      else
      { // make these new
        newBuffers[n] = new ByteArrayOutputStream(dBufferSegSize);
        newBuffersReady[n] = false;
      }
    }
    dBuffers = newBuffers;
    dBuffersReady = newBuffersReady;
  }
  private void repackageDataForOutput()
  {
    if (dFormat == TFormat.CF_NULL) 
    {
      initCounters();
      return;
    }
    try
    {
      int fsiz = TFormat.formatSizeOf(dFormat);
      if (TFormat.isAdjustableLength(dFormat)) fsiz = 1;
      initCounters();
      if (dBuffers == null) return;
      if (numblks > dBuffers.length)
      { // possibly due to format type CF_STRING
        reassignBuffers();
      }
      int nElements = dArrayLength;
      for (int i = 0, n = 0, m = 0; n < numblks; n++, m = i)
      {
        dBuffers[n].reset(); 
        DataOutputStream ds = new DataOutputStream(dBuffers[n]);
        for (; i < nElements && i < m + mtu / fsiz; i++)
        {
          setObjectContents(ds, getDataObject(), i);
          switch (dFormat)
          {
            case TFormat.CF_KEYVALUE:
            case TFormat.CF_STRING:
              nElements = dSizeInBytes;
            case TFormat.CF_IMAGE:
            case TFormat.CF_SPECTRUM:
            case TFormat.CF_GSPECTRUM:
            case TFormat.CF_MDA:
            case TFormat.CF_TEXT:
            case TFormat.CF_XML:
              i = m + mtu - 1; // let the i++ do its thing ...
              break;
            case TFormat.CF_AIMAGE:
            case TFormat.CF_ASPECTRUM:
              i = dArrayLength;
              break;
            default:
              break;
          }
        }
        ds.close();
      }
    }
    catch (IOException e)
    {
      e.printStackTrace();
      MsgLog.log("repackageDataForOutput", e.getMessage(), TErrorList.code_failure, e, 0);
    }
  }
  /**
   * An array of doubles. This corresponds to tine format CF_DOUBLE
   * @param data
   *          is a reference to the double array used to create and bind to the
   *          TDataType object
   */
  public TDataType(double[] data)
  {
    initTDataType(data, data.length);
  }
  /**
   * A subset of an array of doubles. This corresponds to tine format
   *        CF_DOUBLE
   * @param data
   *          is a reference to the double array used to create and bind to the
   *          TDataType object
   * @param length
   *          is the subset array length to use.
   */
  public TDataType(double[] data, int length)
  {
    initTDataType(data, length);
  }
  private void initTDataType(double[] data, int length)
  {
    if (data == null || data.length == 0 || length < 1) return;
    try
    {
      if (length > data.length) length = data.length;
      dFormat = TFormat.CF_DOUBLE;
      dArrayLength = length;
      dSizeInBytes = length * 8;
      initCounters();
      dBuffers = new ByteArrayOutputStream[numblks];
      dBuffersReady = new boolean[numblks];
      for (int n = 0, i = 0; n < numblks; n++)
      {
        dBuffers[n] = new ByteArrayOutputStream(dBufferSegSize);
        DataOutputStream ds = new DataOutputStream(dBuffers[n]);
        for (; i < dArrayLength && i < (n + 1) * mtu / 8; i++)
          ds.write(Swap.Double(data[i]));
        ds.close();
      }
      hDataObject = data;
    }
    catch (IOException e)
    {
      e.printStackTrace();
      MsgLog.log("initTDataType", e.getMessage(), TErrorList.code_failure, e, 0);
    }
  }
  /**
   * An array of (Java) longs. This corresponds to tine format CF_DLONG
   *        or CF_INT64
   * @param data
   *          is a reference to the long array used to create and bind to the
   *          TDataType object
   */
  public TDataType(long[] data)
  {
    initTDataType(data, data.length);
  }
  /**
   * A subset of an array of (java) longs. This corresponds to tine
   *        format CF_DLONG, CF_INT64
   * @param data
   *          is a reference to the long array used to create and bind to the
   *          TDataType object
   * @param length
   *          is the subset array length to use.
   */
  public TDataType(long[] data, int length)
  {
    initTDataType(data, length);
  }
  private void initTDataType(long[] data, int length)
  {
    if (data == null || data.length == 0 || length < 1) return;
    try
    {
      if (length > data.length) length = data.length;
      dFormat = TFormat.CF_INT64;
      dArrayLength = length;
      dSizeInBytes = length * 8;
      initCounters();
      dBuffers = new ByteArrayOutputStream[numblks];
      dBuffersReady = new boolean[numblks];
      for (int n = 0, i = 0; n < numblks; n++)
      {
        dBuffers[n] = new ByteArrayOutputStream(dBufferSegSize);
        DataOutputStream ds = new DataOutputStream(dBuffers[n]);
        for (; i < dArrayLength && i < (n + 1) * mtu / 8; i++)
          ds.writeLong(Swap.Long(data[i]));
        ds.close();
      }
      hDataObject = data;
    }
    catch (IOException e)
    {
      e.printStackTrace();
      MsgLog.log("initTDataType", e.getMessage(), TErrorList.code_failure, e, 0);
    }
  }
  public boolean update(byte[] data, int num, int id)
  {
    return update(data, 0, data.length, num, id);
  }
  public boolean update(byte[] data, int off, int len, int num, int id)
  {
    boolean bytesAccepted = false;
    if (TLinkFactory.debugLevel > 1)
      DbgLog.log("TDataType", "update: blk " + num + " id " + id + " (" + len + " bytes)");
    blknum = num;
    // check num vs numblks, and id for integrity
    if (id != blkid)
    {
      if (TLinkFactory.debugLevel > 1)
        DbgLog.log("TDataType", "update: reset block id from "+blkid+" to "+id);
      blkid = id;
      blksin = 0;
      bytesin = 0;
      resetBuffersReady();
    }
    if (data.length == 0)
    {
      blksin = 1;
      if (dBuffersReady == null) dBuffersReady = new boolean[1];
      return false;
    }
    if (dBuffers == null)
    {
      dBuffers = new ByteArrayOutputStream[numblks];
      dBuffersReady = new boolean[numblks];
      for (int n = 0; n < numblks; n++)
      {
        dBuffers[n] = new ByteArrayOutputStream(dBufferSegSize);
      }
    }
    try
    { // possible small mtu server (NIOS) ...
      if (!dBuffersReady[num - 1])
      {
        buffersBusy = true;
        if (dBuffers[num - 1] == null) dBuffers[num - 1] = new ByteArrayOutputStream(dBufferSegSize);
        dBuffers[num - 1].reset();
        DataOutputStream ds = new DataOutputStream(dBuffers[num - 1]);
        ds.write(data, off, len);
        ds.close();
        dBuffersReady[num - 1] = true;
        blksin++;
        bytesAccepted = true;
      }
      else
      {
        if (TLinkFactory.debugLevel > 1) DbgLog.log("TDataType.update","phantom block " + num + " ignored");
      }
      if (TLinkFactory.debugLevel > 1)
      {
        String tag = blksin == numblks ? " transfer complete" : " missing " + (numblks - blksin) + " blocks";
        DbgLog.log("TDataType", "update: blks received " + blksin + tag);
      }
      hasBeenUpdated = true;
    }
    catch (IOException e)
    {
      e.printStackTrace();
      MsgLog.log("TDataType.update", e.toString(), TErrorList.code_failure, e, 0);
    }
    return bytesAccepted;
  }
  /**
   * Gets the canonical data set from the underlying data byte stream
   *        Gets the fully formed (properly assembled and byte-swapped)
   *        canonical data set from the underlying byte stream. This method is
   *        used internally, but can be used to get the underlying data into the
   *        specifically referenced data, for instance when the TDataType object
   *        was instantiated with 'unbound' data.
   * @param data
   *          is the data object into which the underlying data should be
   *          prepared and copied.
   * @return 0 on success or a tine error code
   */
  public int getData(byte[] data)
  {
    return getData(data, data != null ? data.length : 0, 0);
  }
  /**
   * Gets the canonical data set from the underlying data byte stream
   *        Gets the fully formed (properly assembled and byte-swapped)
   *        canonical data set from the underlying byte stream. This method is
   *        used internally, but can be used to get the underlying data into the
   *        specifically referenced data, for instance when the TDataType object
   *        was instantiated with 'unbound' data.
   * @param data
   *          is the data object into which the underlying data should be
   *          prepared and copied.
   * @param length
   *          is the number of elements to get
   * @param offset
   *          is the beginning position of the target data array into which the
   *          data should be copied
   * @return 0 on success or a tine error code
   */
  public int getData(byte[] data, int length, int offset)
  { synchronized (hDataObject) {
    if (data == null || length < 0 || offset < 0) return TErrorList.argument_list_error;
    if (length == 0)
    {
      dCompletionLength = 0;
      isDataObjectInSync = true;
      return 0;
    }
    if (length > data.length - offset) length = data.length - offset;
    if (length <= 0) return TErrorList.argument_list_error;
    if (offset >= data.length) return TErrorList.dimension_error;
    if (buffersBusy) if (waitForBuffers() != 0) return TErrorList.data_stale;
    TStructDescription tsd = null;
    boolean istts = false;
    int slen = 0;
    if (dFormat == TFormat.CF_STRUCT && dTag != null && dTag.length() > 0)
    {
      tsd = TStructRegistry.get(dTag,structKey);
      if (tsd == null) tsd = TStructRegistry.get(dTag);
      if (tsd != null) slen = tsd.getSize();
      istts = slen != 0;
    }
    int len = dArrayLength; // length of the buffered array
//    if (istts && tsd.hasExtendedSpace())
    if (istts )
    {
      len = tsd.hasExtendedSpace() ? length : slen * length;  
    }
    else
    {
      if (len > length) len = length; // can't get more than is there !
      Arrays.fill(data, offset, offset+length, (byte) 0);
    }
    try
    {
      int i, n, m, fmtsiz = TFormat.formatSizeOf(dFormat);
      for (i=0, n=0, m=0; n<numblks && n<dBuffers.length; n++, m=i)
      {
        byte[] b = dBuffers[n].toByteArray();
        Swap.Bytes(b, fmtsiz);
        ByteArrayInputStream bs = new ByteArrayInputStream(b);
        DataInputStream ds = new DataInputStream(bs);
        switch (dFormat)
        {
          case TFormat.CF_DOUBLE:
            for (; i < len && i < m + b.length / fmtsiz; i++)
              data[i + offset] = (byte) ds.readDouble();
            break;
          case TFormat.CF_FLOAT:
            for (; i < len && i < m + b.length / fmtsiz; i++)
              data[i + offset] = (byte) ds.readFloat();
            break;
          case TFormat.CF_BITFIELD32:
            makeBitfield();
            if (dBitField == null) return TErrorList.not_initialized;
            int ival;
            for (; i < len && i < m + b.length / fmtsiz; i++)
            {
              ival = ds.readInt();
              dBitField[i+offset].setValue(ival);
              data[i + offset] = (byte)dBitField[i+offset].getField(dField, ival);
              //data[i + offset] = dBitField.getField(dField, ds.readInt());
            }
            break;
          case TFormat.CF_BOOLEAN:
          case TFormat.CF_INT32:
            for (; i < len && i < m + b.length / fmtsiz; i++)
              data[i + offset] =(byte)ds.readInt();
            break;
          case TFormat.CF_BITFIELD64:
            makeBitfield();
            if (dBitField == null) return TErrorList.not_initialized;
            long lval;
            for (; i < len && i < m + b.length / fmtsiz; i++)
            {
              lval = ds.readLong();
              dBitField[i+offset].setValue(lval);
              data[i + offset] = (byte)dBitField[i+offset].getField(dField, lval);
              //data[i + offset] = dBitField.getField(dField, ds.readLong());
            }
            break;
          case TFormat.CF_INT64:
            for (; i < len && i < m + b.length / fmtsiz; i++)
              data[i + offset] = (byte) ds.readLong();
            break;
          case TFormat.CF_BITFIELD16:
            makeBitfield();
            if (dBitField == null) return TErrorList.not_initialized;
            short sval;
            for (; i < len && i < m + b.length / fmtsiz; i++)
            {
              sval = ds.readShort();
              dBitField[i+offset].setValue(sval);
              data[i + offset] = (byte)dBitField[i+offset].getField(dField, sval);
              //data[i + offset] = dBitField.getField(dField, ds.readShort());
            }
            break;
          case TFormat.CF_INT16:
            for (; i < len && i < m + b.length / fmtsiz; i++)
              data[i + offset] = (byte)ds.readShort();
            break;
          case TFormat.CF_BITFIELD8:
            makeBitfield();
            if (dBitField == null) return TErrorList.not_initialized;
            byte bval;            
            for (; i < len && i < m + b.length / fmtsiz; i++)
            {
              bval = ds.readByte();
              dBitField[i+offset].setValue(bval);
              data[i + offset] = (byte)dBitField[i+offset].getField(dField, bval);
            }
            break;
          default:
          case TFormat.CF_STRUCT:
          case TFormat.CF_BYTE:
            for (; i < len && i < m + b.length / fmtsiz; i++)
            {
              if (i >= data.length-offset) break;
              data[i + offset] = ds.readByte();
            }
            break;
        }
        ds.close();
        bs.close();
      }
      dCompletionLength = i;
      if (istts)
      {
        if (tsd.hasExtendedSpace()) slen = len;
        dCompletionLength /= slen;
      }
      isDataObjectInSync = true;
    }
    catch (IOException e)
    {
      e.printStackTrace();
      MsgLog.log("TDataType.getData", e.getMessage(), TErrorList.code_failure, e, 0);
    }
    return 0;
  } }
  /**
   * Gets the canonical data set from the underlying data byte stream
   *        Gets the fully formed (properly assembled and byte-swapped)
   *        canonical data set from the underlying byte stream. This method is
   *        used internally, but can be used to get the underlying data into the
   *        specifically referenced data, for instance when the TDataType object
   *        was instantiated with 'unbound' data.
   * @param data
   *          is the data object into which the underlying data should be
   *          prepared and copied.
   * @return 0 on success or a tine error code
   */
  public int getData(short[] data)
  {
    return getData(data, data != null ? data.length : 0, 0);
  }
  private void makeBitfield() 
  { 
    makeBitfield(dArrayLength); 
  }
  private void makeBitfield(int len)
  {
    if (dBitField != null) return; // done
    if (dTag == null || dTag.length() == 0) return; // not possible
    dBitField = new TBitfield[len];
    TBitfield bf = TBitfieldRegistry.getBitfield(dTag);
    if (bf == null) bf = new TBitfield(dTag,dFormat); 
    for (int i=0; i<len; i++) dBitField[i] = new TBitfield(bf);
  }
  /**
   * Gets the canonical data set from the underlying data byte stream
   *        Gets the fully formed (properly assembled and byte-swapped)
   *        canonical data set from the underlying byte stream. This method is
   *        used internally, but can be used to get the underlying data into the
   *        specifically referenced data, for instance when the TDataType object
   *        was instantiated with 'unbound' data.
   * @param data
   *          is the data object into which the underlying data should be
   *          prepared and copied.
   * @param length
   *          is the number of elements to get
   * @param offset
   *          is the beginning position of the target data array into which the
   *          data should be copied
   * @return 0 on success or a tine error code
   */
  public int getData(short[] data, int length, int offset)
  { synchronized (hDataObject) {
    if (data == null || length <= 0 || offset < 0) return TErrorList.argument_list_error;
    if (length > data.length - offset) length = data.length - offset;
    if (length <= 0) return TErrorList.argument_list_error;
    if (offset >= data.length) return TErrorList.dimension_error;
    int len = dArrayLength; // length of the buffered array
    if (len > length) len = length; // can't get more than is there !
    if (buffersBusy) if (waitForBuffers() != 0) return TErrorList.data_stale;
    if (isDataObjectInSync && hDataObject != null)
    {
      if (data == hDataObject) return 0;
      Arrays.fill(data, offset, offset+length,(short)0);
      try
      {
        for (int i=0; i < len; i++) data[i+offset] = getShortAtIndex(i);
      }
      catch (TineRuntimeErrorException e)
      {
        return e.getErrorCode();
      }
      dCompletionLength = len;
      return 0;
    }      
    try
    {
      Arrays.fill(data, offset, offset+length, (short) 0);
      int i, n, m, fmtsiz = TFormat.formatSizeOf(dFormat);
      ByteArrayInputStream bs;
      DataInputStream ds;
      for (i=0, n=0, m=0; n<numblks && n<dBuffers.length; n++, m=i)
      {
        if (dBuffers[n] == null) break;
        byte[] b = dBuffers[n].toByteArray();
        Swap.Bytes(b, fmtsiz);
        bs = new ByteArrayInputStream(b);
        ds = new DataInputStream(bs);
        switch (dFormat)
        {
          case TFormat.CF_DOUBLE:
            for (; i < len && i < m + b.length / fmtsiz; i++)
              data[i + offset] = (short) ds.readDouble();
            break;
          case TFormat.CF_FLOAT:
            for (; i < len && i < m + b.length / fmtsiz; i++)
              data[i + offset] = (short) ds.readFloat();
            break;
          case TFormat.CF_BITFIELD64:
            makeBitfield();
            if (dBitField == null) return TErrorList.not_initialized;
            long lval;
            for (; i < len && i < m + b.length / fmtsiz; i++)
            {
              lval = ds.readLong();
              dBitField[i+offset].setValue(lval);
              data[i + offset] = (short)dBitField[i+offset].getField(dField,lval);
            }
            break;
          case TFormat.CF_INT64:
            for (; i < len && i < m + b.length / fmtsiz; i++)
              data[i + offset] = (short) ds.readLong();
            break;
          case TFormat.CF_BITFIELD32:
            int ival;
            makeBitfield();
            if (dBitField == null) return TErrorList.not_initialized;
            for (; i < len && i < m + b.length / fmtsiz; i++)
            {
              ival = ds.readInt();
              dBitField[i+offset].setValue(ival);
              data[i + offset] = (short)dBitField[i+offset].getField(dField,ival);
              //data[i + offset] = (short)dBitField.getField(dField, ds.readInt());
            }
            break;
          case TFormat.CF_BOOLEAN:
          case TFormat.CF_INT32:
            for (; i < len && i < m + b.length / fmtsiz; i++)
              data[i + offset] = (short) ds.readInt();
            break;
          case TFormat.CF_BITFIELD16:
            makeBitfield();
            if (dBitField == null) return TErrorList.not_initialized;
            short sval;
            for (; i < len && i < m + b.length / fmtsiz; i++)
            {
              sval = ds.readShort();
              dBitField[i+offset].setValue(sval);
              data[i + offset] = (short)dBitField[i+offset].getField(dField,sval);
              //data[i + offset] = (short)dBitField.getField(dField, ds.readShort());
            }
            break;
          case TFormat.CF_INT16:
            for (; i < len && i < m + b.length / fmtsiz; i++)
              data[i + offset] = ds.readShort();
            break;
          case TFormat.CF_BITFIELD8:
            makeBitfield();
            if (dBitField == null) return TErrorList.not_initialized;
            byte bval;            
            for (; i < len && i < m + b.length / fmtsiz; i++)
            {
              bval = ds.readByte();
              dBitField[i+offset].setValue(bval);
              data[i + offset] = (short)dBitField[i+offset].getField(dField, bval);
            }
            break;
          case TFormat.CF_BYTE:
            for (; i < len && i < m + b.length / fmtsiz; i++)
              data[i + offset] = ds.readByte();
            break;
          default:
            ds.close();
            bs.close();
            return TErrorList.illegal_format;
        }
        ds.close();
        bs.close();
      }
      dCompletionLength = i;
      isDataObjectInSync = true;
    }
    catch (IOException e)
    {
      e.printStackTrace();
      MsgLog.log("TDataType.getData", e.getMessage(), TErrorList.code_failure, e, 0);
    }
    return 0;
  } }
  /**
   * Puts the current canonical data set into the underlying data byte
   *        stream Puts the input data set into the underlying byte stream,
   *        properly swapped and aligned. It replaces the contents of the
   *        original, referenced data array with that of the data object given.
   * @param data
   *          is data primitive whose value is copied into the underlying byte
   *          stream and referenced data object.
   * @note If this TDataType object is used as input data in persistent links,
   *       then the data object might well be 'locked' meaning that the
   *       underlying data will NOT be changed, as this would constitute a
   *       change in contract.
   * @return 0 upon success or a tine error code
   */
  public int putData(short data)
  {
    short[] d = new short[1];
    d[0] = data;
    return putData(d);
  }
  /**
   * Puts the current canonical data set into the underlying data byte
   *        stream Puts the input data set into the underlying byte stream,
   *        properly swapped and aligned. It replaces the contents of the
   *        original, referenced data array with that of the data object given.
   * @param data
   *          is the data object whose contents should be copied into the
   *          underlying byte stream and referenced data object.
   * @note If this TDataType object is used as input data in persistent links,
   *       then the data object might well be 'locked' meaning that the
   *       underlying data will NOT be changed, as this would constitute a
   *       change in contract.
   * @return 0 upon success or a tine error code
   */
  public int putData(short[] data)
  {
    return putData(data, data != null ? data.length : 0, 0);
  }
  /**
   * Puts the current canonical data set into the underlying data byte
   *        stream Puts the input data set into the underlying byte stream,
   *        properly swapped and aligned. It replaces the contents of the
   *        original, referenced data array with that of the data object given.
   * @param data
   *          is the data object whose contents should be copied into the
   *          underlying byte stream and referenced data object.
   * @param length
   *          is the number of elements from the give data object to copy
   * @param offset
   *          is the starting offset within the array to copy
   * @note If this TDataType object is used as input data in persistent links,
   *       then the data object might well be 'locked' meaning that the
   *       underlying data will NOT be changed, as this would constitute a
   *       change in contract.
   * @return 0 upon success or a tine error code
   */
  public int putData(short[] data, int length, int offset)
  {
    return putData(0,data,length,offset,true);  	
  }
  public int putData(int start, short[] data, int length, int offset, boolean redim)
  { synchronized (hDataObject) {
    if (data == null || length <= 0 || offset < 0) return TErrorList.argument_list_error;
    if (length > data.length - offset) length = data.length - offset;
    if (length <= 0) return TErrorList.argument_list_error;
    if (isLocked) throw new InputDataLockedException(lockedMessage);
    int maxlen = Math.min(dArrayLength, length);
    if (maxlen > dArrayLength - start) maxlen = dArrayLength - start;
    switch (dFormat)
    {
      case TFormat.CF_DOUBLE:
        for (int i = 0; i < maxlen; i++)
          ((double[]) hDataObject)[i+start] = (double) data[i + offset];
        break;
      case TFormat.CF_FLOAT:
        for (int i = 0; i < maxlen; i++)
          ((float[]) hDataObject)[i+start] = (float) data[i + offset];
        break;
      case TFormat.CF_INT16:
        for (int i = 0; i < maxlen; i++)
          ((short[]) hDataObject)[i+start] = (short) data[i + offset];
        break;
      case TFormat.CF_BOOLEAN:
        for (int i = 0; i < maxlen; i++)
          ((boolean[]) hDataObject)[i+start] = (boolean)(data[i + offset] != 0);
        break;
      case TFormat.CF_INT32:
        for (int i = 0; i < maxlen; i++)
          ((int[]) hDataObject)[i+start] = (int) data[i + offset];
        break;
      case TFormat.CF_INT64:
        for (int i = 0; i < maxlen; i++)
          ((long[]) hDataObject)[i+start] = (long) data[i + offset];
        break;
      default:
        return TErrorList.illegal_format;
    }
    if (redim) dArrayLength = Math.min(dArrayLength, length);
    repackageDataForOutput();
    return 0;
  } }
  /**
   * Puts the input data value into the underlying byte stream,
   * properly swapped and aligned. It replaces the contents of the
   * original, referenced data array with that of the data object given.
   * @param value
   *          is the single value which should be copied into the
   *          underlying byte stream and referenced data object at the given offset.
   * @param offset
   *          is the position offset within the targeted array
   * @note If this TDataType object is used as input data in persistent links,
   *       then the data object might well be 'locked' meaning that the
   *       underlying data will NOT be changed, as this would constitute a
   *       change in contract.
   * @return 0 upon success or a tine error code
   * 
   * @throws InputDataLockedException if a link using this TDataType as input 
   * is bound to another link 
   * and an attempt to alter the input data set (which is part of the contract
   * cached at the target server) is made.  
   * 
   */
  public int putData(short value, int offset)
  {
    return putDataValue(Short.valueOf(value),offset);
  }
  /**
   * Puts the input data value array into the underlying byte stream,
   * properly swapped and aligned. It replaces the contents of the
   * original, referenced data array with that of the data object given.
   * @param values
   *          is the values array which should be copied into the
   *          underlying byte stream and referenced data object at the given offset.
   * @param offset
   *          is the position offset within the targeted array
   * @note If this TDataType object is used as input data in persistent links,
   *       then the data object might well be 'locked' meaning that the
   *       underlying data will NOT be changed, as this would constitute a
   *       change in contract.
   * @return 0 upon success or a tine error code
   * 
   * @throws InputDataLockedException if a link using this TDataType as input 
   * is bound to another link 
   * and an attempt to alter the input data set (which is part of the contract
   * cached at the target server) is made.  
   * 
   */
  public int putData(short[] values,int offset)
  {
    return putDataSegment(values, offset);
  }
  public void applyBitField()
  {
    makeBitfield();
    if (dBitField == null) return;
    if (hDataObject instanceof TBitfield[]) return;
    switch (dFormat)
    {
      case TFormat.CF_BITFIELD8:
      case TFormat.CF_BITFIELD16:
      case TFormat.CF_BITFIELD32:
      case TFormat.CF_BITFIELD64:
        if (hDataObject instanceof byte[])
        {
          for (int i = 0; i < dCompletionLength; i++)
          {
            ((byte[]) hDataObject)[i] = (byte) dBitField[i].getField(dField, ((byte[]) hDataObject)[i]);
          }
          break;
        }
        if (hDataObject instanceof short[])
        {
          for (int i = 0; i < dCompletionLength; i++)
          {
            ((short[]) hDataObject)[i] = (short) dBitField[i].getField(dField, ((short[]) hDataObject)[i]);
          }
          break;
        }
        if (hDataObject instanceof int[])
        {
          for (int i = 0; i < dCompletionLength; i++)
          {
            ((int[]) hDataObject)[i] = (int) dBitField[i].getField(dField, ((int[]) hDataObject)[i]);
          }
          break;
        }
        if (hDataObject instanceof long[])
        {
          for (int i = 0; i < dCompletionLength; i++)
          {
            // TODO: allow int64 bitfields !
            ((long[]) hDataObject)[i] = (long) dBitField[i].getField(dField, (int) ((long[]) hDataObject)[i]);
          }
          break;
        }
        break;
    }
    return;
  }
  /**
   * Gets the canonical data set from the underlying data byte stream
   *        Gets the fully formed (properly assembled and byte-swapped)
   *        canonical data set from the underlying byte stream. This method is
   *        mostly used internally, and is of little use to the API programmer.
   * @return 0 on success or a tine error code
   */
  public int getData()
  {
    if (buffersBusy) if (waitForBuffers() != 0) return TErrorList.data_stale;
    switch (dFormat)
    {
      case TFormat.CF_AIMAGE:
        if (hDataObject instanceof IMAGE[])
        {
          IMAGE[] img = (IMAGE[])hDataObject;
          for (int i=0; i<img.length; i++)
            if (img[i] == null) img[i] = new IMAGE(100);
        }
        return getData(hDataObject);
      case TFormat.CF_ASPECTRUM:
        if (hDataObject instanceof ASPECTRUM[])
        {
          ASPECTRUM[] sp = (ASPECTRUM[])hDataObject;
          for (int i=0; i<sp.length; i++)
            if (sp[i] == null) sp[i] = new ASPECTRUM(100);
        }
        return getData(hDataObject);
      case TFormat.CF_STRUCT:
        if (hDataObject != null) getData(hDataObject);
        TTaggedStructure[] tts = null;
        int dsiz = 0;
        if (hStructObject instanceof TTaggedStructure)
        {
          tts = new TTaggedStructure[1];
          tts[0] = (TTaggedStructure) hStructObject;
        }
        else if (hStructObject instanceof TTaggedStructure[])
        {
          tts = (TTaggedStructure[]) hStructObject;
        }
        else
        {
          TStructDescription tsd = TStructRegistry.get(dTag,structKey);
          if (tsd != null && tsd.hasExtendedSpace())
          { // find out how much (i.e. how many) came in:
            tts = new TTaggedStructure[1];
            tts[0] = new TTaggedStructure(dTag);
            if (tts[0].getSizeInBytes() == 0) 
            {
              tts[0] = null;  tts = null;
            }
            else
            {
              TTaggedStructure ref = tsd.getTaggedStructure();
              tts[0].addFieldsFrom(ref);
            }
          }
        }
        if (tts != null)
        {
          dsiz = tts[0].getSizeInBytes();
          int boundary = dsiz * tts.length; 
          for (int i = 0; i < tts.length; i++)
          {
            tts[i].toStruct((byte[]) hDataObject, i * dsiz, dsiz);
            if (i == 0)
            { // find out how much (i.e. how many) came in, in case of extended bytes
              boundary = tts[0].getExtendedOffset();
            }
          }
          dCompletionLength = boundary / dsiz;
          dArrayLength = boundary;
        }
        break;
      default:
        getData(getDataObject());
        break;
    }
    return 0;
  }
  /**
   * Gets the canonical data set from the underlying data byte stream
   *        Gets the fully formed (properly assembled and byte-swapped)
   *        canonical data set from the underlying byte stream. This method is
   *        used internally, but can be used to get the underlying data into the
   *        specifically referenced data, for instance when the TDataType object
   *        was instantiated with 'unbound' data.
   * @param data
   *          is the data object into which the underlying data should be
   *          prepared and copied.
   * @return 0 on success or a tine error code
   */
  public int getData(boolean[] data)
  {
    return getData(data, data != null ? data.length : 0, 0);
  }
  /**
   * Gets the canonical data set from the underlying data byte stream
   *        Gets the fully formed (properly assembled and byte-swapped)
   *        canonical data set from the underlying byte stream. This method is
   *        used internally, but can be used to get the underlying data into the
   *        specifically referenced data, for instance when the TDataType object
   *        was instantiated with 'unbound' data.
   * @param data
   *          is the data object into which the underlying data should be
   *          prepared and copied.
   * @param length
   *          is the number of elements to get
   * @param offset
   *          is the beginning position of the target data array into which the
   *          data should be copied
   * @return 0 on success or a tine error code
   */
  public int getData(boolean[] data, int length, int offset)
  {
    if (data == null || data.length == 0) return TErrorList.argument_list_error;
    int idata[] = new int[data.length];
    int cc = getData(idata, length, offset);
    if (cc == 0)
    {
      for (int i = 0; i < data.length; i++)
        data[i] = idata[i] == 0 ? false : true;
    }
    return cc;
  }
  /**
   * Gets the canonical data set from the underlying data byte stream
   *        Gets the fully formed (properly assembled and byte-swapped)
   *        canonical data set from the underlying byte stream. This method is
   *        used internally, but can be used to get the underlying data into the
   *        specifically referenced data, for instance when the TDataType object
   *        was instantiated with 'unbound' data.
   * @param data
   *          is the data object into which the underlying data should be
   *          prepared and copied.
   * @return 0 on success or a tine error code
   */
  public int getData(int[] data)
  {
    return getData(data, data != null ? data.length : 0, 0);
  }
  /**
   * Gets the canonical data set from the underlying data byte stream
   *        Gets the fully formed (properly assembled and byte-swapped)
   *        canonical data set from the underlying byte stream. This method is
   *        used internally, but can be used to get the underlying data into the
   *        specifically referenced data, for instance when the TDataType object
   *        was instantiated with 'unbound' data.
   * @param data
   *          is the data object into which the underlying data should be
   *          prepared and copied.
   * @param length
   *          is the number of elements to get
   * @param offset
   *          is the beginning position of the target data array into which the
   *          data should be copied
   * @return 0 on success or a tine error code
   */
  public int getData(int[] data, int length, int offset)
  { synchronized (hDataObject) {
    if (data == null || length <= 0 || offset < 0) return TErrorList.argument_list_error;
    if (length > data.length - offset) length = data.length - offset;
    if (length <= 0) return TErrorList.argument_list_error;
    if (offset >= data.length) return TErrorList.dimension_error;
    int len = dArrayLength; // length of the buffered array
    if (len > length) len = length; // can't get more than is there !
    if (buffersBusy) if (waitForBuffers() != 0) return TErrorList.data_stale;
    if (isDataObjectInSync && hDataObject != null)
    {
      if (data == hDataObject) return 0;
      Arrays.fill(data, offset, offset+length, 0);
      try
      {
        for (int i=0; i < len; i++) data[i+offset] = getIntAtIndex(i);
      }
      catch (TineRuntimeErrorException e)
      {
        return e.getErrorCode();
      }
      dCompletionLength = len;
      return 0;
    }      
    try
    {
      Arrays.fill(data, offset, offset+length, 0);
      int i, n, m, fmtsiz = TFormat.formatSizeOf(dFormat);
      ByteArrayInputStream bs;
      DataInputStream ds;
      for (i = 0, n = 0, m = 0; n < numblks; n++, m = i)
      {
        if (dBuffers[n] == null) break;
        byte[] b = dBuffers[n].toByteArray();
        Swap.Bytes(b, fmtsiz);
        bs = new ByteArrayInputStream(b);
        ds = new DataInputStream(bs);
        switch (dFormat)
        {
          case TFormat.CF_DOUBLE:
            for (; i < len && i < m + b.length / fmtsiz; i++)
              data[i + offset] = (int) ds.readDouble();
            break;
          case TFormat.CF_FLOAT:
            for (; i < len && i < m + b.length / fmtsiz; i++)
              data[i + offset] = (int) ds.readFloat();
            break;
          case TFormat.CF_BITFIELD32:
            makeBitfield();
            if (dBitField == null) return TErrorList.not_initialized;
            int ival;
            for (; i < len && i < m + b.length / fmtsiz; i++)
            {
              ival = ds.readInt();
              dBitField[i+offset].setValue(ival);
              data[i + offset] = dBitField[i+offset].getField(dField, ival);
              //data[i + offset] = dBitField.getField(dField, ds.readInt());
            }
            break;
          case TFormat.CF_BOOLEAN:
          case TFormat.CF_UINT32:
          case TFormat.CF_INT32:
            for (; i < len && i < m + b.length / fmtsiz; i++)
              data[i + offset] = ds.readInt();
            break;
          case TFormat.CF_BITFIELD64:
            makeBitfield();
            if (dBitField == null) return TErrorList.not_initialized;
            long lval;
            for (; i < len && i < m + b.length / fmtsiz; i++)
            {
              lval = ds.readLong();
              dBitField[i+offset].setValue(lval);
              data[i + offset] = dBitField[i+offset].getField(dField, lval);
              //data[i + offset] = dBitField.getField(dField, ds.readLong());
            }
            break;
          case TFormat.CF_UINT64:
          case TFormat.CF_INT64:
            for (; i < len && i < m + b.length / fmtsiz; i++)
              data[i + offset] = (int) ds.readLong();
            break;
          case TFormat.CF_BITFIELD16:
            makeBitfield();
            if (dBitField == null) return TErrorList.not_initialized;
            short sval;
            for (; i < len && i < m + b.length / fmtsiz; i++)
            {
              sval = ds.readShort();
              dBitField[i+offset].setValue(sval);
              data[i + offset] = dBitField[i+offset].getField(dField, sval);
              //data[i + offset] = dBitField.getField(dField, ds.readShort());
            }
            break;
          case TFormat.CF_UINT16:
          case TFormat.CF_INT16:
            for (; i < len && i < m + b.length / fmtsiz; i++)
              data[i + offset] = ds.readShort();
            break;
          case TFormat.CF_BITFIELD8:
            makeBitfield();
            if (dBitField == null) return TErrorList.not_initialized;
            byte bval;            
            for (; i < len && i < m + b.length / fmtsiz; i++)
            {
              bval = ds.readByte();
              dBitField[i+offset].setValue(bval);
              data[i + offset] = dBitField[i+offset].getField(dField, bval);
            }
            break;
          case TFormat.CF_BYTE:
            for (; i < len && i < m + b.length / fmtsiz; i++)
              data[i + offset] = ds.readByte();
            break;
          default:
            ds.close();
            bs.close();
            return TErrorList.illegal_format;
        }
        ds.close();
        bs.close();
      }
      dCompletionLength = i;
      isDataObjectInSync = true;
    }
    catch (IOException e)
    {
      e.printStackTrace();
      MsgLog.log("TDataType.getData", e.getMessage(), TErrorList.code_failure, e, 0);
    }
    catch (Exception e)
    {
      e.printStackTrace();
      MsgLog.log("TDataType.getData", e.getMessage(), TErrorList.code_failure, e, 0);      
    }
    return 0;
  } }
  /**
   * Puts the current canonical data set into the underlying data byte
   *        stream Puts the input data set into the underlying byte stream,
   *        properly swapped and aligned. It replaces the contents of the
   *        original, referenced data array with that of the data object given.
   * @param data
   *          is data primitive whose value is copied into the underlying byte
   *          stream and referenced data object.
   * @note If this TDataType object is used as input data in persistent links,
   *       then the data object might well be 'locked' meaning that the
   *       underlying data will NOT be changed, as this would constitute a
   *       change in contract.
   * @return 0 upon success or a tine error code
   */
  public int putData(boolean data)
  {
    boolean[] d = new boolean[1];
    d[0] = data;
    return putData(d);
  }
  /**
   * Puts the current canonical data set into the underlying data byte
   *        stream Puts the input data set into the underlying byte stream,
   *        properly swapped and aligned. It replaces the contents of the
   *        original, referenced data array with that of the data object given.
   * @param data
   *          is the data object whose contents should be copied into the
   *          underlying byte stream and referenced data object.
   * @note If this TDataType object is used as input data in persistent links,
   *       then the data object might well be 'locked' meaning that the
   *       underlying data will NOT be changed, as this would constitute a
   *       change in contract.
   * @return 0 upon success or a tine error code
   */
  public int putData(boolean[] data)
  {
    return putData(data, data != null ? data.length : 0, 0);
  }
  /**
   * Puts the current canonical data set into the underlying data byte
   *        stream Puts the input data set into the underlying byte stream,
   *        properly swapped and aligned. It replaces the contents of the
   *        original, referenced data array with that of the data object given.
   * @param data
   *          is the data object whose contents should be copied into the
   *          underlying byte stream and referenced data object.
   * @param length
   *          is the number of elements from the give data object to copy
   * @param offset
   *          is the starting offset within the array to copy
   * @note If this TDataType object is used as input data in persistent links,
   *       then the data object might well be 'locked' meaning that the
   *       underlying data will NOT be changed, as this would constitute a
   *       change in contract.
   * @return 0 upon success or a tine error code
   */
  public int putData(boolean[] data, int length, int offset)
  {
    return putData(0,data,length,offset,false);	
  }
  public int putData(int start, boolean[] data, int length, int offset, boolean redim)
  {
    if (data == null || length <= 0 || offset < 0) return TErrorList.argument_list_error;
    int idata[] = new int[data.length];
    for (int i = 0; i < data.length; i++)
      idata[i] = data[i] ? 1 : 0;
    return putData(start, idata, length, offset,redim);
  }
  /**
   * Puts the current canonical data set into the underlying data byte
   *        stream Puts the input data value into the underlying byte stream,
   *        properly swapped and aligned. It replaces the contents of the
   *        original, referenced data array with that of the data object given.
   * @param value
   *          is the single value which should be copied into the
   *          underlying byte stream and referenced data object at the given offset.
   * @param offset
   *          is the position offset within the targeted array
   * @note If this TDataType object is used as input data in persistent links,
   *       then the data object might well be 'locked' meaning that the
   *       underlying data will NOT be changed, as this would constitute a
   *       change in contract.
   * @return 0 upon success or a tine error code
   * 
   * @throws InputDataLockedException if a link using this TDataType as input 
   * is bound to another link 
   * and an attempt to alter the input data set (which is part of the contract
   * cached at the target server) is made.  
   * 
   */
  public int putData(boolean value, int offset)
  {
    int idata = value ? 1 : 0;
    return putData(idata, offset);
  }
  /**
   * Puts the current canonical data set into the underlying data byte
   *        stream Puts the input data set into the underlying byte stream,
   *        properly swapped and aligned. It replaces the contents of the
   *        original, referenced data array with that of the data object given.
   * @param data
   *          is data primitive whose value is copied into the underlying byte
   *          stream and referenced data object.
   * @note If this TDataType object is used as input data in persistent links,
   *       then the data object might well be 'locked' meaning that the
   *       underlying data will NOT be changed, as this would constitute a
   *       change in contract.
   * @return 0 upon success or a tine error code
   */
  public int putData(int data)
  {
    int[] d = new int[1];
    d[0] = data;
    return putData(d);
  }
  /**
   * Puts the current canonical data set into the underlying data byte
   *        stream Puts the input data set into the underlying byte stream,
   *        properly swapped and aligned. It replaces the contents of the
   *        original, referenced data array with that of the data object given.
   * @param data
   *          is the data object whose contents should be copied into the
   *          underlying byte stream and referenced data object.
   * @note If this TDataType object is used as input data in persistent links,
   *       then the data object might well be 'locked' meaning that the
   *       underlying data will NOT be changed, as this would constitute a
   *       change in contract.
   * @return 0 upon success or a tine error code
   */
  public int putData(int[] data)
  {
    return putData(data, data != null ? data.length : 0, 0);
  }
  /**
   * Puts the current canonical data set into the underlying data byte
   *        stream Puts the input data set into the underlying byte stream,
   *        properly swapped and aligned. It replaces the contents of the
   *        original, referenced data array with that of the data object given.
   * @param data
   *          is the data object whose contents should be copied into the
   *          underlying byte stream and referenced data object.
   * @param length
   *          is the number of elements from the give data object to copy
   * @param offset
   *          is the starting offset within the array to copy
   * @note If this TDataType object is used as input data in persistent links,
   *       then the data object might well be 'locked' meaning that the
   *       underlying data will NOT be changed, as this would constitute a
   *       change in contract.
   * @return 0 upon success or a tine error code
   */
  public int putData(int[] data, int length, int offset)
  {
    return putData(0,data,length,offset,true); 	
  }
  public int putData(int start, int[] data, int length, int offset, boolean redim)
  { synchronized (hDataObject) {
    if (data == null || length <= 0 || offset < 0) return TErrorList.argument_list_error;
    if (length > data.length - offset) length = data.length - offset;
    if (length <= 0) return TErrorList.argument_list_error;
    if (isLocked) throw new InputDataLockedException(lockedMessage);
    int maxlen = Math.min(dArrayLength, length);
    if (maxlen > dArrayLength - start) maxlen = dArrayLength - start;
    switch (dFormat)
    {
      case TFormat.CF_DOUBLE:
        for (int i = 0; i < maxlen; i++)
          ((double[]) hDataObject)[i+start] = (double) data[i + offset];
        break;
      case TFormat.CF_FLOAT:
        for (int i = 0; i < maxlen; i++)
          ((float[]) hDataObject)[i+start] = (float) data[i + offset];
        break;
      case TFormat.CF_INT16:
        for (int i = 0; i < maxlen; i++)
          ((short[]) hDataObject)[i+start] = (short) data[i + offset];
        break;
      case TFormat.CF_BOOLEAN:
        for (int i = 0; i < maxlen; i++)
          ((boolean[]) hDataObject)[i+start] = (boolean)(data[i + offset] != 0);
        break;
      case TFormat.CF_INT32:
        for (int i = 0; i < maxlen; i++)
          ((int[]) hDataObject)[i+start] = (int) data[i + offset];
        break;
      case TFormat.CF_INT64:
        for (int i = 0; i < maxlen; i++)
          ((long[]) hDataObject)[i+start] = (long) data[i + offset];
        break;
      default:
        return TErrorList.illegal_format;
    }
    if (redim) dArrayLength = Math.min(dArrayLength, length);
    repackageDataForOutput();
    return 0;
  } }
  /**
   * Puts the input data value into the underlying byte stream,
   * properly swapped and aligned. It replaces the contents of the
   * original, referenced data array with that of the data object given.
   * @param value
   *          is the single value which should be copied into the
   *          underlying byte stream and referenced data object at the given offset.
   * @param offset
   *          is the position offset within the targeted array
   * @note If this TDataType object is used as input data in persistent links,
   *       then the data object might well be 'locked' meaning that the
   *       underlying data will NOT be changed, as this would constitute a
   *       change in contract.
   * @return 0 upon success or a tine error code
   * 
   * @throws InputDataLockedException if a link using this TDataType as input 
   * is bound to another link 
   * and an attempt to alter the input data set (which is part of the contract
   * cached at the target server) is made.  
   * 
   */
  public int putData(int value, int offset)
  {
    return putDataValue(Integer.valueOf(value),offset);
  }
  /**
   * Puts the input data value array into the underlying byte stream,
   * properly swapped and aligned. It replaces the contents of the
   * original, referenced data array with that of the data object given.
   * @param values
   *          is the values array which should be copied into the
   *          underlying byte stream and referenced data object at the given offset.
   * @param offset
   *          is the position offset within the targeted array
   * @note If this TDataType object is used as input data in persistent links,
   *       then the data object might well be 'locked' meaning that the
   *       underlying data will NOT be changed, as this would constitute a
   *       change in contract.
   * @return 0 upon success or a tine error code
   * 
   * @throws InputDataLockedException if a link using this TDataType as input 
   * is bound to another link 
   * and an attempt to alter the input data set (which is part of the contract
   * cached at the target server) is made.  
   * 
   */
  public int putData(int[] values, int offset)
  {
    return putDataSegment(values,offset);
  }  
  /**
   * Gets the canonical data set from the underlying data byte stream
   *        Gets the fully formed (properly assembled and byte-swapped)
   *        canonical data set from the underlying byte stream. This method is
   *        used internally, but can be used to get the underlying data into the
   *        specifically referenced data, for instance when the TDataType object
   *        was instantiated with 'unbound' data.
   * @param data
   *          is the data object into which the underlying data should be
   *          prepared and copied.
   * @return 0 on success or a tine error code
   */
  public int getData(float[] data)
  {
    return getData(data, data != null ? data.length : 0, 0);
  }
  /**
   * Gets the canonical data set from the underlying data byte stream
   *        Gets the fully formed (properly assembled and byte-swapped)
   *        canonical data set from the underlying byte stream. This method is
   *        used internally, but can be used to get the underlying data into the
   *        specifically referenced data, for instance when the TDataType object
   *        was instantiated with 'unbound' data.
   * @param data
   *          is the data object into which the underlying data should be
   *          prepared and copied.
   * @param length
   *          is the number of elements to get
   * @param offset
   *          is the beginning position of the target data array into which the
   *          data should be copied
   * @return 0 on success or a tine error code
   */
  public int getData(float[] data, int length, int offset)
  { synchronized (hDataObject) {
    if (data == null || length <= 0 || offset < 0) return TErrorList.argument_list_error;
    if (length > data.length - offset) length = data.length - offset;
    if (length <= 0) return TErrorList.argument_list_error;
    if (offset >= data.length) return TErrorList.dimension_error;
    int len = dArrayLength; // length of the buffered array
    if (len > length) len = length; // can't get more than is there !
    if (buffersBusy) if (waitForBuffers() != 0) return TErrorList.data_stale;
    if (isDataObjectInSync && hDataObject != null)
    {
      if (data == hDataObject) return 0;
      Arrays.fill(data, offset, offset+length, 0);
      try
      {
        for (int i=0; i < len; i++) data[i+offset] = getFloatAtIndex(i);
      }
      catch (TineRuntimeErrorException e)
      {
        return e.getErrorCode();
      }
      dCompletionLength = len;
      return 0;
    }      
    try
    {
      Arrays.fill(data, offset, offset+length, 0);
      int i, n, m, fmtsiz = TFormat.formatSizeOf(dFormat);
      ByteArrayInputStream bs;
      DataInputStream ds;
      for (i=0, n=0, m=0; n<numblks && n<dBuffers.length; n++, m=i)
      {
        if (dBuffers[n] == null) break;
        byte[] b = dBuffers[n].toByteArray();
        Swap.Bytes(b, fmtsiz);
        bs = new ByteArrayInputStream(b);
        ds = new DataInputStream(bs);
        switch (dFormat)
        {
          case TFormat.CF_DOUBLE:
            for (; i < len && i < m + b.length / fmtsiz; i++)
              data[i + offset] = (float) ds.readDouble();
            break;
          case TFormat.CF_FLOAT:
            for (; i < len && i < m + b.length / fmtsiz; i++)
              data[i + offset] = ds.readFloat();
            break;
          case TFormat.CF_BOOLEAN:
          case TFormat.CF_INT32:
            for (; i < len && i < m + b.length / fmtsiz; i++)
              data[i + offset] = ds.readInt();
            break;
          case TFormat.CF_INT64:
            for (; i < len && i < m + b.length / fmtsiz; i++)
              data[i + offset] = (float) ds.readLong();
            break;
          case TFormat.CF_INT16:
            for (; i < len && i < m + b.length / fmtsiz; i++)
              data[i + offset] = ds.readShort();
            break;
          default:
            ds.close();
            bs.close();
            return TErrorList.illegal_format;
        }
        ds.close();
        bs.close();
      }
      dCompletionLength = i;
      isDataObjectInSync = true;
    }
    catch (Exception e)
    {
      e.printStackTrace();
      MsgLog.log("TDataType.getData", e.getMessage(), TErrorList.code_failure, e, 0);
    }
    return 0;
  } }
  /**
   * Gets the canonical data set from the underlying data byte stream
   *        Gets the fully formed (properly assembled and byte-swapped)
   *        canonical data set from the underlying byte stream. This method is
   *        used internally, but can be used to get the underlying data into the
   *        specifically referenced data, for instance when the TDataType object
   *        was instantiated with 'unbound' data.
   * @param data
   *          is the data object into which the underlying data should be
   *          prepared and copied.
   * @return 0 on success or a tine error code
   */
  public int getData(TTaggedStructure[] data)
  {
    return getData(data, data != null ? data.length : 0, 0);
  }
  /**
   * Gets the canonical data set from the underlying data byte stream
   *        Gets the fully formed (properly assembled and byte-swapped)
   *        canonical data set from the underlying byte stream. This method is
   *        used internally, but can be used to get the underlying data into the
   *        specifically referenced data, for instance when the TDataType object
   *        was instantiated with 'unbound' data.
   * @param data
   *          is the data object into which the underlying data should be
   *          prepared and copied.
   * @param length
   *          is the number of elements to get
   * @param offset
   *          is the beginning position of the target data array into which the
   *          data should be copied
   * @return 0 on success or a tine error code
   */
  public int getData(TTaggedStructure[] data, int length, int offset)
  { synchronized (hDataObject) {
    if (data == null || length <= 0 || offset < 0) return TErrorList.argument_list_error;
    if (length > data.length - offset) length = data.length - offset;
    if (length <= 0) return TErrorList.argument_list_error;
    if (dFormat != TFormat.CF_STRUCT) return TErrorList.illegal_format;
    if (data[offset] == null) return TErrorList.invalid_data;
    if (dTag.compareTo(data[offset].getName()) != 0) return TErrorList.invalid_structure_tag;
    if (hDataObject == null) return TErrorList.invalid_data;
    if (buffersBusy) if (waitForBuffers() != 0) return TErrorList.data_stale;
    getData(hDataObject);
    int dsiz = data[offset].getSizeInBytes();
    for (int i = offset; i < length + offset; i++)
    { // copy the byte stream into the referenced data object
      if (data[i] == null) continue;
      data[i].toStruct((byte[])hDataObject, (i-offset) * dsiz, dsiz);
    }
    dCompletionLength /= dsiz;
    isDataObjectInSync = true;
    return 0;
  } }
  /**
   * Gets the canonical data set from the underlying data byte stream
   *        Gets the fully formed (properly assembled and byte-swapped)
   *        canonical data set from the underlying byte stream. This method is
   *        used internally, but can be used to get the underlying data into the
   *        specifically referenced data, for instance when the TDataType object
   *        was instantiated with 'unbound' data.
   * @param data
   *          is the data object into which the underlying data should be
   *          prepared and copied.
   * @return 0 on success or a tine error code
   */
  public int getData(TCompoundDataObject[] data)
  {
    return getData(data, data != null ? data.length : 0, 0);
  }
  /**
   * Gets the canonical data set from the underlying data byte stream
   *        Gets the fully formed (properly assembled and byte-swapped)
   *        canonical data set from the underlying byte stream. This method is
   *        used internally, but can be used to get the underlying data into the
   *        specifically referenced data, for instance when the TDataType object
   *        was instantiated with 'unbound' data.
   * @param data
   *          is the data object into which the underlying data should be
   *          prepared and copied.
   * @param length
   *          is the number of elements to get
   * @param offset
   *          is the beginning position of the target data array into which the
   *          data should be copied
   * @return 0 on success or a tine error code
   */
  public int getData(TCompoundDataObject[] data, int length, int offset)
  { 
    if (data == null || length <= 0 || offset < 0) return TErrorList.argument_list_error;
    if (dFormat == TFormat.CF_STRING) return TErrorList.illegal_format;
    if (data instanceof TBitfield[])
    {
      if (((TBitfield[]) data)[0] == null) return TErrorList.not_initialized;
    }
    if (data instanceof IMAGE[])
    {
      return getData(((IMAGE[]) data)[0]);
    }
    if (data instanceof SPECTRUM[])
    {
      return getData(((SPECTRUM[]) data)[0]);
    }
    if (data instanceof GSPECTRUM[])
    {
      return getData(((GSPECTRUM[]) data)[0]);
    }
    if (length > data.length - offset) length = data.length - offset;
    if (length <= 0) return TErrorList.argument_list_error;
    int len = dArrayLength; // length of this buffered array
    if (len > length) len = length; // can't get more than is there !
    int to_size = 0; // size in bytes of the TCompoundDataObject
    int from_size = TFormat.formatSizeOf(dFormat);
    if (data instanceof HISTORY[])
    {
      return getData(((HISTORY[]) data),length,offset);
    }
    int cpy_size = from_size; // copy size
    if (buffersBusy) if (waitForBuffers() != 0) return TErrorList.data_stale;
    synchronized (hDataObject) {
    try
    {
      int i, n, m, j;
      for (i=0, n=0, m=0, j=0; n<numblks && n<dBuffers.length; n++, m=i)
      {
        if (i + offset >= data.length)
        {
          MsgLog.log("TDataType", "data overflow discarded", TErrorList.buffer_too_small, null, 1);
          break;
        }
        if (data[i + offset] == null) makeNewCompoundDataObject(data, i + offset);
        if (to_size == 0)
        {
          to_size = data[i + offset].getSizeInBytes();
          if (dFormat == TFormat.CF_TEXT)
          { // this is a TEXT object being sent a NAMExx
            to_size = 1;
          }
          if (to_size < cpy_size) cpy_size = to_size;
        }
        if (dBuffers[n] == null) break;
        byte[] b = dBuffers[n].toByteArray(); // this object's byte stream
        ByteArrayInputStream bs = new ByteArrayInputStream(b);
        DataInputStream ds = new DataInputStream(bs);
        for (; i < len && i < m + b.length / from_size; i++)
        {
          if (data[i + offset] == null) makeNewCompoundDataObject(data, i + offset);
          if (data[i + offset].getByteArray() == null) data[i + offset].toByteArray();
          byte[] ba = data[i + offset].getByteArray();
          for (j = 0; j < cpy_size; j++) ba[j] = ds.readByte();
          data[i + offset].toStruct();
        }
        ds.close();
        bs.close();
      }
      dCompletionLength = i;
      isDataObjectInSync = true;
    }
    catch (IOException e)
    {
      e.printStackTrace();
      MsgLog.log("TDataType.getData", e.getMessage(), TErrorList.code_failure, e, 0);
    }
    return 0;
  } }
  public int getData(HISTORY[] data, int length, int offset)
  { // n.b. the initial HISTORY object must not be null !
    synchronized (hDataObject) {
    if (dFormat != TFormat.CF_HISTORY) return TErrorList.illegal_format;
    if (((HISTORY[])data)[0] == null) return TErrorList.un_allocated;
    if (buffersBusy && waitForBuffers() != 0) return TErrorList.data_stale;
    if (length > data.length - offset) length = data.length - offset;
    if (length <= 0) return TErrorList.argument_list_error;
    int len = dArrayLength; // length of this buffered array
    if (len > length) len = length; // can't get more than is there !
    int from_size = ((HISTORY[])data)[0].getSizeInBytes();
    short cfmt = ((HISTORY[])data)[0].getCarriedFormat();
    int capacity = from_size;
    try
    {
      byte[] bset = new byte[numblks*dBuffers[0].size()];
      byte[] b;
      int nbytes = 0;
      for (int n=0; n<numblks; n++)
      { // CF_HISTORY comes in in straight bytes -> have to collect them all first!
        b = dBuffers[n].toByteArray();
        System.arraycopy(b, 0, bset, nbytes, b.length);
        nbytes += b.length;
      }
      ByteArrayInputStream bs = new ByteArrayInputStream(bset);
      DataInputStream ds = new DataInputStream(bs);
      int i, j, nb = 0, fsiz;
      byte[] fb = new byte[4];
      for (i=0; i<len && nb<nbytes-from_size+1; i++)
      {
        if (i + offset >= data.length)
        {
          MsgLog.log("TDataType", "data overflow disgarded", TErrorList.buffer_too_small, null, 1);
          break;
        }
        switch (cfmt)
        {
          case TFormat.CF_IMAGE:
            System.arraycopy(bset, HISTORY.HDR_SIZE+nb+IMAGE.HEADER_SIZE-4, fb, 0, 4);
            fsiz = Swap.Int(fb);
            fsiz += HISTORY.HDR_SIZE + IMAGE.HEADER_SIZE;
            if (fsiz > capacity) return TErrorList.buffer_too_small;
            from_size = fsiz;
            break;
          case TFormat.CF_KEYVALUE:
          case TFormat.CF_STRING:
            for (j=0; j<nbytes-HISTORY.HDR_SIZE-nb; j++) 
            {
              if (bset[HISTORY.HDR_SIZE+nb+j] == 0) break;
            }
            fsiz =  HISTORY.HDR_SIZE + j + 1;
            from_size = fsiz;
            data[i + offset].setByteArray(new byte[from_size]);
            break;
          default:
            break;
        }
        if (data[i + offset] == null) 
        { 
        //makeNewCompoundDataObject(data, i + offset);
          data[i + offset] = data[0].clone();
        }
          
        if (data[i + offset].getByteArray() == null) data[i + offset].toByteArray();
        byte[] ba = data[i + offset].getByteArray();
        for (j = 0; j < from_size; j++) ba[j] = ds.readByte();
        data[i + offset].toStruct();
        nb += from_size;
      }
      ds.close();
      bs.close();
      dCompletionLength = i;
      isDataObjectInSync = true;
    }
    catch (IOException e)
    {
      e.printStackTrace();
      MsgLog.log("TDataType.getData", e.getMessage(), TErrorList.code_failure, e, 0);
    }
    return 0;
  } }
  public int getData(MDA data)
  { synchronized (hDataObject) {
    if (dFormat != TFormat.CF_MDA) return TErrorList.illegal_format;
    int dlen = Math.min(dArrayLength,data.getTotalLength());
    int fsiz = TFormat.getCarriedFormatSize(dTag);
    dlen *= fsiz;
    if (dlen == 0) return TErrorList.dimension_error;
    int dSize = MDA.HDR_SIZE + dlen;
    if (buffersBusy) if (waitForBuffers() != 0) return TErrorList.data_stale;
    try
    {
      int i, n, m;
      byte[] ba = data.getByteArray();
      if (ba == null)
      { // not assigned when constructed
        ba = new byte[dSize];
        data.setByteArray(ba);
      }
      for (i=0, n=0, m=0; n<numblks && n<dBuffers.length; n++, m=i)
      {
        if (dBuffers[n] == null) break;
        byte[] b = dBuffers[n].toByteArray();
        ByteArrayInputStream bs = new ByteArrayInputStream(b);
        DataInputStream ds = new DataInputStream(bs);
        for (; i < dSize && i < m + b.length; i++)
        {
          ba[i] = ds.readByte();
        }
        ds.close();
        bs.close();
      }
      data.toStruct();
      dCompletionLength = (i-MDA.HDR_SIZE)/fsiz;
    }
    catch (IOException e)
    {
      e.printStackTrace();
      MsgLog.log("TDataType.getData", e.getMessage(), TErrorList.code_failure, e, 0);
    }
    return 0;
  } }
  /**
   * Gets the canonical data set from the underlying data byte stream
   *        Gets the fully formed (properly assembled and byte-swapped)
   *        canonical data set from the underlying byte stream. This method is
   *        used internally, but can be used to get the underlying data into the
   *        specifically referenced data, for instance when the TDataType object
   *        was instantiated with 'unbound' data.
   * @param data
   *          is the data object into which the underlying data should be
   *          prepared and copied.
   * @return 0 on success or a tine error code
   */
  public int getData(SPECTRUM data)
  { synchronized (hDataObject) {
    if (dFormat != TFormat.CF_SPECTRUM) return TErrorList.illegal_format;
    int dlen = Math.min(dArrayLength,data.length);
    if (dlen == 0) return TErrorList.dimension_error;
    int dSize = dlen + 24; // effective data size for SPECTRUM
    if (buffersBusy) if (waitForBuffers() != 0) return TErrorList.data_stale;
    try
    {
      int i, n, m;
      for (i=0, n=0, m=0; n<numblks && n<dBuffers.length; n++, m=i)
      {
        if (dBuffers[n] == null) break;
        byte[] b = dBuffers[n].toByteArray();
        ByteArrayInputStream bs = new ByteArrayInputStream(b);
        DataInputStream ds = new DataInputStream(bs);
        for (; i < dSize * 4 && i < m + b.length; i++)
        {
          data.hByteArray[i] = ds.readByte();
        }
        ds.close();
        bs.close();
      }
      dCompletionLength = i/4 - 24;
      data.toStruct();
    }
    catch (IOException e)
    {
      e.printStackTrace();
      MsgLog.log("TDataType.getData", e.getMessage(), TErrorList.code_failure, e, 0);
    }
    return 0;
  } }
  public int getData(GSPECTRUM data)
  { synchronized (hDataObject) {
    if (dFormat != TFormat.CF_GSPECTRUM) return TErrorList.illegal_format;
    int dlen = Math.min(dArrayLength,data.length);
    if (dlen == 0) return TErrorList.dimension_error;
    dlen *= 4; // size of a float
    dlen += GSPECTRUM.hdrSizeInBytes; // header
    if (buffersBusy) if (waitForBuffers() != 0) return TErrorList.data_stale;
    try
    {
      int i, n, m;
      for (i=0, n=0, m=0; n<numblks && n<dBuffers.length; n++, m=i)
      {
        if (dBuffers[n] == null) break;
        byte[] b = dBuffers[n].toByteArray();
        ByteArrayInputStream bs = new ByteArrayInputStream(b);
        DataInputStream ds = new DataInputStream(bs);
        for (; i < dlen && i < m + b.length; i++)
        {
          data.hByteArray[i] = ds.readByte();
        }
        ds.close();
        bs.close();
      }
      dCompletionLength = (i-GSPECTRUM.hdrSizeInBytes)/4;
      data.toStruct();
    }
    catch (IOException e)
    {
      e.printStackTrace();
      MsgLog.log("TDataType.getData", e.getMessage(), TErrorList.code_failure, e, 0);
    }
    return 0;
  } }
  public int getData(ASPECTRUM[] data)
  { synchronized (hDataObject) {
    if (dFormat != TFormat.CF_ASPECTRUM) return TErrorList.illegal_format;
    int len = dArrayLength;
    try { len = Integer.parseInt(dTag); } catch (Exception ignore) {}
    int dlen = Math.min(len,data.length);
    if (dlen == 0) return TErrorList.dimension_error;
    if (buffersBusy) if (waitForBuffers() != 0) return TErrorList.data_stale;
    try
    {
      int i, n, kb=0, k=0;
      int xsiz, xoff;
      boolean scanhdrs = true;
      for (n = 0; n < numblks; n++)
      {
        if (dBuffers[n] == null) break;
        byte[] b = dBuffers[n].toByteArray();
        ByteArrayInputStream bs = new ByteArrayInputStream(b);
        DataInputStream ds = new DataInputStream(bs);
        for (i=0; i < b.length; i++)
        {
          if (scanhdrs)
          { // still scanning thru the fixed header parts
            ((ASPECTRUM[])data)[k].hByteArray[kb++] = ds.readByte();
            if (kb == ASPECTRUM.hdrSizeInBytes+8)
            {
              ((ASPECTRUM[])data)[k].toStruct();
              xsiz = ((ASPECTRUM[])data)[k].getExtendedSize();
              ((ASPECTRUM[])data)[k].setSpectrumByteArray(new byte[xsiz]);
              kb = 0;
              if (k == 0)
              { // re-calculate the number returned ....
                xoff = ((ASPECTRUM[])data)[k].getExtendedOffset();
                len = xoff / (ASPECTRUM.hdrSizeInBytes + 8);
              }
              if (++k == len) 
              {
                scanhdrs = false;
                k = 0 ;
              }
            }
          }
          else
          { // scan thru the extended spectra ...
            byte[] sba = ((ASPECTRUM[])data)[k].hSpectrumByteArray;
            if (sba == null || sba.length == 0)
            {
              i = b.length; continue;
            }
            sba[kb++] = ds.readByte();
            if (kb == sba.length)
            {
              ((ASPECTRUM[])data)[k].toSpectrumStruct();
              kb = 0; k++;
            }
          }
        }
        ds.close();
        bs.close();
      }
      dCompletionLength = len;
      isDataObjectInSync = true;
    }
    catch (IOException e)
    {
      e.printStackTrace();
      MsgLog.log("TDataType.getData", e.getMessage(), TErrorList.code_failure, e, 0);
    }
    return 0;
  } }
  public int getData(IMAGE[] data)
  { synchronized (hDataObject) {
    if (dFormat != TFormat.CF_AIMAGE) return TErrorList.illegal_format;
    int len = dArrayLength; 
    try { len = Integer.parseInt(dTag); } catch (Exception ignore) {}
    int dlen = Math.min(len,data.length);
    if (dlen == 0) return TErrorList.dimension_error;
    if (buffersBusy) if (waitForBuffers() != 0) return TErrorList.data_stale;
    try
    {
      int i, n, kb=0, k=0;
      int xsiz, xoff;
      boolean scanhdrs = true;
      byte[] hdr = new byte[IMAGE.HEADER_SIZE+8];
      for (n = 0; n < numblks; n++)
      {
        if (dBuffers[n] == null) break;
        byte[] b = dBuffers[n].toByteArray();
        ByteArrayInputStream bs = new ByteArrayInputStream(b);
        DataInputStream ds = new DataInputStream(bs);
        for (i=0; i < b.length; i++)
        {
          if (scanhdrs)
          { // still scanning thru the fixed header parts
            hdr[kb++] = ds.readByte();
            if (kb == hdr.length)
            {
              ((IMAGE[])data)[k].readImageArrayHeaders(hdr, 0);
              xsiz = ((IMAGE[])data)[k].getExtendedSize();
              ((IMAGE[])data)[k].ensureFrameBufferCapacity(xsiz);
              kb = 0;
              if (k == 0)
              { // re-calculate the number returned ....
                xoff = ((IMAGE[])data)[k].getExtendedOffset();
                len = xoff / hdr.length;
              }
              if (++k == len) 
              { // done with the headers
                scanhdrs = false;
                k = 0 ;
              }
            }
          }
          else
          { // scan thru the extended spectra ...
            ((IMAGE[])data)[k].fillFrameBuffer(ds.readByte(),kb++);
            if (kb == ((IMAGE[])data)[k].getExtendedSize())
            {
              kb = 0; k++;
            }
          }
        }
        ds.close();
        bs.close();
      }
      dCompletionLength = len;
      isDataObjectInSync = true;
    }
    catch (IOException e)
    {
      e.printStackTrace();
      MsgLog.log("TDataType.getData", e.getMessage(), TErrorList.code_failure, e, 0);
    }
    return 0;
  } }
  /**
   * Gets the canonical data set from the underlying data byte stream
   *        Gets the fully formed (properly assembled and byte-swapped)
   *        canonical data set from the underlying byte stream. This method is
   *        used internally, but can be used to get the underlying data into the
   *        specifically referenced data, for instance when the TDataType object
   *        was instantiated with 'unbound' data.
   * @param data
   *          is the data object into which the underlying data should be
   *          prepared and copied.
   * @return 0 on success or a tine error code
   */
  public int getData(IMAGE image)
  { // mdavid: changed/optimized (~5 times faster!)
    if (dFormat != TFormat.CF_IMAGE) return TErrorList.illegal_format;
    if (dArrayLength > image.getImageFrameSize()) return TErrorList.dimension_error;
    if (buffersBusy && waitForBuffers() != 0) return TErrorList.data_stale;
    try
    {
      int length, srcPos = 0, destPos = 0;
      for (int n = 0; n < numblks && n < dBuffers.length; srcPos = 0, n++)
      {
        if (dBuffers[n] == null) break; // stop loop
        byte[] bytes = dBuffers[n].toByteArray();
        length = bytes.length;
        if (n == 0)
        { // first block contains the image header
          image.readHeaders(bytes);
          length -= (srcPos = IMAGE.HEADER_SIZE);
          if (length < 0)
          {
            DbgLog.log("TDataType.getData", "corrupt image data length");
            length = 0;
          }
        }
        System.arraycopy(bytes, srcPos, image.getImageFrameBuffer(), destPos, length);
        destPos += length;
      }
      dCompletionLength = destPos;
      return TErrorList.success;
    }
    catch (Exception e)
    {
      e.printStackTrace();
      MsgLog.log("TDataType.getData", e.getMessage(), TErrorList.code_failure, e, 0);
      return TErrorList.code_failure;
    }
  }
  /**
   * Puts the current canonical data set into the underlying data byte
   *        stream Puts the input data set into the underlying byte stream,
   *        properly swapped and aligned. It replaces the contents of the
   *        original, referenced data array with that of the data object given.
   * @param data
   *          is data primitive whose value is copied into the underlying byte
   *          stream and referenced data object.
   * @note If this TDataType object is used as input data in persistent links,
   *       then the data object might well be 'locked' meaning that the
   *       underlying data will NOT be changed, as this would constitute a
   *       change in contract.
   * @return 0 upon success or a tine error code
   */
  public int putData(float data)
  {
    float[] d = new float[1];
    d[0] = data;
    return putData(d);
  }
  /**
   * Puts the current canonical data set into the underlying data byte
   *        stream Puts the currently referenced canonical data set into the
   *        underlying byte stream, properly swapped and aligned. This method is
   *        primarily used internally and is of little value to the API
   *        programmer.
   * @return 0 upon success or a tine error code
   */
  public int putData()
  {
    if (hDataObject == null)
    { // doesn't make sense (a primitive was passed as write data ?)
      return 0;
    }
    if (isLocked) throw new InputDataLockedException(lockedMessage);
    repackageDataForOutput();
    return 0;
  }
  /**
   * Puts the current canonical data set into the underlying data byte
   *        stream Puts the input data set into the underlying byte stream,
   *        properly swapped and aligned. It replaces the contents of the
   *        original, referenced data array with that of the data object given.
   * @param data
   *          is the data object whose contents should be copied into the
   *          underlying byte stream and referenced data object.
   * @note If this TDataType object is used as input data in persistent links,
   *       then the data object might well be 'locked' meaning that the
   *       underlying data will NOT be changed, as this would constitute a
   *       change in contract.
   * @return 0 upon success or a tine error code
   */
  public int putData(float[] data)
  {
    return putData(data, data != null ? data.length : 0, 0);
  }
  /**
   * Puts the current canonical data set into the underlying data byte
   *        stream Puts the input data set into the underlying byte stream,
   *        properly swapped and aligned. It replaces the contents of the
   *        original, referenced data array with that of the data object given.
   * @param data
   *          is the data object whose contents should be copied into the
   *          underlying byte stream and referenced data object.
   * @param length
   *          is the number of elements from the give data object to copy
   * @param offset
   *          is the starting offset within the array to copy
   * @note If this TDataType object is used as input data in persistent links,
   *       then the data object might well be 'locked' meaning that the
   *       underlying data will NOT be changed, as this would constitute a
   *       change in contract.
   * @return 0 upon success or a tine error code
   * 
   * @throws InputDataLockedException if a link using this TDataType as input 
   * is bound to another link 
   * and an attempt to alter the input data set (which is part of the contract
   * cached at the target server) is made.  
   * 
   */
  public int putData(float[] data, int length, int offset)
  {
    return putData(0,data,length,offset,true);
  }
  public int putData(int start, float[] data, int length, int offset, boolean redim)
  { synchronized (hDataObject) {
    if (data == null || length <= 0 || offset < 0) return TErrorList.argument_list_error;
    if (length > data.length - offset) length = data.length - offset;
    if (length <= 0) return TErrorList.argument_list_error;
    if (isLocked) throw new InputDataLockedException(lockedMessage);
    int maxlen = Math.min(dArrayLength, length);
    if (maxlen > dArrayLength - start) maxlen = dArrayLength - start;
    switch (dFormat)
    {
      case TFormat.CF_DOUBLE:
        for (int i = 0; i < maxlen; i++)
          ((double[]) hDataObject)[i+start] = (double) data[i + offset];
        break;
      case TFormat.CF_FLOAT:
        for (int i = 0; i < maxlen; i++)
          ((float[]) hDataObject)[i+start] = (float) data[i + offset];
        break;
      case TFormat.CF_INT16:
        for (int i = 0; i < maxlen; i++)
          ((short[]) hDataObject)[i+start] = (short) data[i + offset];
        break;
      case TFormat.CF_INT64:
        for (int i = 0; i < maxlen; i++)
          ((long[]) hDataObject)[i+start] = (long) data[i + offset];
        break;
      case TFormat.CF_BOOLEAN:
        for (int i = 0; i < maxlen; i++)
          ((boolean[]) hDataObject)[i+start] = (boolean)(data[i + offset] != 0);
        break;
      case TFormat.CF_INT32:
        for (int i = 0; i < maxlen; i++)
          ((int[]) hDataObject)[i+start] = (int) data[i + offset];
        break;
      default:
        return TErrorList.illegal_format;
    }
    if (redim) dArrayLength = Math.min(dArrayLength, length);
    repackageDataForOutput();
    return 0;
  } }
  /**
   * Puts the current canonical data set into the underlying data byte
   *        stream Puts the input data value into the underlying byte stream,
   *        properly swapped and aligned. It replaces the contents of the
   *        original, referenced data array with that of the data object given.
   * @param value
   *          is the single value which should be copied into the
   *          underlying byte stream and referenced data object at the given offset.
   * @param offset
   *          is the position offset within the targeted array
   * @note If this TDataType object is used as input data in persistent links,
   *       then the data object might well be 'locked' meaning that the
   *       underlying data will NOT be changed, as this would constitute a
   *       change in contract.
   * @return 0 upon success or a tine error code
   * 
   * @throws InputDataLockedException if a link using this TDataType as input 
   * is bound to another link 
   * and an attempt to alter the input data set (which is part of the contract
   * cached at the target server) is made.  
   * 
   */
  public int putData(float value, int offset)
  {
    return putDataValue(Float.valueOf(value),offset);
  }
  /**
   * Puts the input data value array into the underlying byte stream,
   * properly swapped and aligned. It replaces the contents of the
   * original, referenced data array with that of the data object given.
   * @param values
   *          is the values array which should be copied into the
   *          underlying byte stream and referenced data object at the given offset.
   * @param offset
   *          is the position offset within the targeted array
   * @note If this TDataType object is used as input data in persistent links,
   *       then the data object might well be 'locked' meaning that the
   *       underlying data will NOT be changed, as this would constitute a
   *       change in contract.
   * @return 0 upon success or a tine error code
   * 
   * @throws InputDataLockedException if a link using this TDataType as input 
   * is bound to another link 
   * and an attempt to alter the input data set (which is part of the contract
   * cached at the target server) is made.  
   * 
   */
  public int putData(float[] values,int offset)
  {
    return putDataSegment(values, offset);
  }
  /**
   * Puts the current canonical data set into the underlying data byte
   *        stream Puts the input data set into the underlying byte stream,
   *        properly swapped and aligned. It replaces the contents of the
   *        original, referenced data array with that of the data object given.
   * @param data
   *          is data primitive whose value is copied into the underlying byte
   *          stream and referenced data object.
   * @note If this TDataType object is used as input data in persistent links,
   *       then the data object might well be 'locked' meaning that the
   *       underlying data will NOT be changed, as this would constitute a
   *       change in contract.
   * @return 0 upon success or a tine error code
   */
  public int putData(FLTINT data)
  {
    FLTINT[] d = new FLTINT[1];
    d[0] = data;
    return putData(d);
  }
  /**
   * Puts the current canonical data set into the underlying data byte
   *        stream Puts the input data set into the underlying byte stream,
   *        properly swapped and aligned. It replaces the contents of the
   *        original, referenced data array with that of the data object given.
   * @param data
   *          is the data object whose contents should be copied into the
   *          underlying byte stream and referenced data object.
   * @note If this TDataType object is used as input data in persistent links,
   *       then the data object might well be 'locked' meaning that the
   *       underlying data will NOT be changed, as this would constitute a
   *       change in contract.
   * @return 0 upon success or a tine error code
   */
  public int putData(FLTINT[] data)
  {
    return putData(data, data != null ? data.length : 0, 0);
  }
  /**
   * Puts the current canonical data set into the underlying data byte
   *        stream Puts the input data set into the underlying byte stream,
   *        properly swapped and aligned. It replaces the contents of the
   *        original, referenced data array with that of the data object given.
   * @param data
   *          is the data object whose contents should be copied into the
   *          underlying byte stream and referenced data object.
   * @param length
   *          is the number of elements from the give data object to copy
   * @param offset
   *          is the starting offset within the array to copy
   * @note If this TDataType object is used as input data in persistent links,
   *       then the data object might well be 'locked' meaning that the
   *       underlying data will NOT be changed, as this would constitute a
   *       change in contract.
   * @return 0 upon success or a tine error code
   */
  public int putData(FLTINT[] data, int length, int offset)
  {
    return putData(0,data,length,offset,true);
  }
  public int putData(int start, FLTINT[] data, int length, int offset, boolean redim)
  { synchronized (hDataObject) {
    if (data == null || length <= 0 || offset < 0) return TErrorList.argument_list_error;
    if (length > data.length - offset) length = data.length - offset;
    if (length <= 0) return TErrorList.argument_list_error;
    if (isLocked) throw new InputDataLockedException(lockedMessage);
    int maxlen = Math.min(dArrayLength, length);
    if (maxlen > dArrayLength - start) maxlen = dArrayLength - start;
    switch (dFormat)
    {
      case TFormat.CF_INTINT:
        for (int i = 0; i < maxlen; i++)
        {
          if (data[i + offset] == null) continue;
          if (((INTINT[]) hDataObject)[i+start] != null)
            ((INTINT[]) hDataObject)[i+start].setValues((int) data[i + offset].fval, data[i + offset].ival);
          else
            ((INTINT[]) hDataObject)[i+start] = new INTINT((int) data[i + offset].fval, data[i + offset].ival);
        }
        break;
      case TFormat.CF_FLTINT:
        for (int i = 0; i < maxlen; i++)
        {
          if (data[i + offset] == null) continue;
          if (((FLTINT[]) hDataObject)[i+start] != null)
            ((FLTINT[]) hDataObject)[i+start].setValues(data[i + offset].fval, data[i + offset].ival);
          else
            ((FLTINT[]) hDataObject)[i+start] = data[i + offset];
        }
        break;
      case TFormat.CF_DBLDBL:
        for (int i = 0; i < maxlen; i++)
        {
          if (data[i + offset] == null) continue;
          if (((DBLDBL[]) hDataObject)[i+start] != null)
            ((DBLDBL[]) hDataObject)[i+start].setValues(data[i + offset].fval, data[i + offset].ival);
          else
            ((DBLDBL[]) hDataObject)[i+start] = new DBLDBL(data[i + offset].fval, data[i + offset].ival);
        }
        break;
      case TFormat.CF_DBLDBLDBL:
        for (int i = 0; i < maxlen; i++)
        {
          if (data[i + offset] == null) continue;
          if (((DBLDBLDBL[]) hDataObject)[i+start] != null)
            ((DBLDBLDBL[]) hDataObject)[i+start].setValues(data[i + offset].fval, 0, data[i + offset].ival);
          else
            ((DBLDBLDBL[]) hDataObject)[i+start] = new DBLDBLDBL(data[i + offset].fval, 0, data[i + offset].ival);
        }
        break;
      case TFormat.CF_FLOAT:
        for (int i = 0; i < maxlen; i++)
        {
          if (data[i + offset] == null) continue;
          ((float[]) hDataObject)[i+start] = data[i + offset].fval;
        }
        break;
      case TFormat.CF_DOUBLE:
        for (int i = 0; i < maxlen; i++)
        {
          if (data[i + offset] == null) continue;
          ((double[]) hDataObject)[i+start] = data[i + offset].fval;
        }
        break;
      case TFormat.CF_INT64:
        for (int i = 0; i < maxlen; i++)
        {
          if (data[i + offset] == null) continue;
          ((long[]) hDataObject)[i+start] = data[i + offset].ival;
        }
        break;
      case TFormat.CF_INT32:
        for (int i = 0; i < maxlen; i++)
        {
          if (data[i + offset] == null) continue;
          ((int[]) hDataObject)[i+start] = data[i + offset].ival;
        }
        break;
      default:
        return TErrorList.illegal_format;
    }
    if (redim) dArrayLength = Math.min(dArrayLength, length);
    repackageDataForOutput();
    return 0;
  } }
  /**
   * Puts the current canonical data set into the underlying data byte
   *        stream Puts the input data set into the underlying byte stream,
   *        properly swapped and aligned. It replaces the contents of the
   *        original, referenced data array with that of the data object given.
   * @param data
   *          is the data object whose contents should be copied into the
   *          underlying byte stream and referenced data object.
   * @note If this TDataType object is used as input data in persistent links,
   *       then the data object might well be 'locked' meaning that the
   *       underlying data will NOT be changed, as this would constitute a
   *       change in contract.
   * @return 0 upon success or a tine error code
   */
  public int putData(NAME16FI[] data)
  {
    return putData(data, data != null ? data.length : 0, 0);
  }
  /**
   * Puts the current canonical data set into the underlying data byte
   *        stream Puts the input data set into the underlying byte stream,
   *        properly swapped and aligned. It replaces the contents of the
   *        original, referenced data array with that of the data object given.
   * @param data
   *          is the data object whose contents should be copied into the
   *          underlying byte stream and referenced data object.
   * @param length
   *          is the number of elements from the give data object to copy
   * @param offset
   *          is the starting offset within the array to copy
   * @note If this TDataType object is used as input data in persistent links,
   *       then the data object might well be 'locked' meaning that the
   *       underlying data will NOT be changed, as this would constitute a
   *       change in contract.
   * @return 0 upon success or a tine error code
   */
  public int putData(NAME16FI[] data, int length, int offset)
  {
    return putData(0,data,length,offset,true);  	
  }
  public int putData(int start, NAME16FI[] data, int length, int offset, boolean redim)
  { synchronized (hDataObject) {
    if (data == null || length <= 0 || offset < 0) return TErrorList.argument_list_error;
    if (length > data.length - offset) length = data.length - offset;
    if (length <= 0) return TErrorList.argument_list_error;
    if (isLocked) throw new InputDataLockedException(lockedMessage);
    int maxlen = Math.min(dArrayLength, length);
    if (maxlen > dArrayLength - start) maxlen = dArrayLength - start;
    switch (dFormat)
    {
      case TFormat.CF_NAME16FI:
        for (int i = 0; i < maxlen; i++)
        {
          if (data[i + offset] == null) continue;
          if (((NAME16FI[]) hDataObject)[i+start] != null)
            ((NAME16FI[]) hDataObject)[i+start].setValues(data[i + offset].name, data[i + offset].fval, data[i
                + offset].ival);
          else
            ((NAME16FI[]) hDataObject)[i] = data[i + offset];
        }
        break;
      case TFormat.CF_NAME16II:
        for (int i = 0; i < maxlen; i++)
        {
          if (data[i + offset] == null) continue;
          if (((NAME16II[]) hDataObject)[i+start] != null)
            ((NAME16II[]) hDataObject)[i+start].setValues(data[i + offset].name, (int) data[i + offset].fval,
                data[i + offset].ival);
          else
            ((NAME16II[]) hDataObject)[i+start] = new NAME16II(data[i + offset].name, (int) data[i + offset].fval,
                data[i + offset].ival);
        }
        break;
      case TFormat.CF_NAME16DBLDBL:
        for (int i = 0; i < maxlen; i++)
        {
          if (data[i + offset] == null) continue;
          if (((NAME16DBLDBL[]) hDataObject)[i+start] != null)
            ((NAME16DBLDBL[]) hDataObject)[i+start].setValues(data[i + offset].name, data[i + offset].fval, data[i
                + offset].ival);
          else
            ((NAME16DBLDBL[]) hDataObject)[i+start] = new NAME16DBLDBL(data[i + offset].name,
                data[i + offset].fval, data[i + offset].ival);
        }
        break;
      case TFormat.CF_NAME32DBLDBL:
        for (int i = 0; i < maxlen; i++)
        {
          if (data[i + offset] == null) continue;
          if (((NAME32DBLDBL[]) hDataObject)[i+start] != null)
            ((NAME32DBLDBL[]) hDataObject)[i+start].setValues(data[i + offset].name, data[i + offset].fval, data[i
                + offset].ival);
          else
            ((NAME32DBLDBL[]) hDataObject)[i+start] = new NAME32DBLDBL(data[i + offset].name,
                data[i + offset].fval, data[i + offset].ival);
        }
        break;
      case TFormat.CF_NAME64DBL:
        for (int i = 0; i < maxlen; i++)
        {
          if (data[i + offset] == null) continue;
          if (((NAME64DBL[]) hDataObject)[i+start] != null)
            ((NAME64DBL[]) hDataObject)[i+start].setValues(data[i + offset].name, data[i + offset].fval);
          else
            ((NAME64DBL[]) hDataObject)[i+start] = new NAME64DBL(data[i + offset].name, data[i + offset].fval);
        }
        break;
      case TFormat.CF_NAME64DBLDBL:
        for (int i = 0; i < maxlen; i++)
        {
          if (data[i + offset] == null) continue;
          if (((NAME64DBLDBL[]) hDataObject)[i+start] != null)
            ((NAME64DBLDBL[]) hDataObject)[i+start].setValues(data[i + offset].name, data[i + offset].fval, data[i
                + offset].ival);
          else
            ((NAME64DBLDBL[]) hDataObject)[i+start] = new NAME64DBLDBL(data[i + offset].name,
                data[i + offset].fval, data[i + offset].ival);
        }
        break;
      case TFormat.CF_NAME64DBLDBLDBL:
        for (int i = 0; i < maxlen; i++)
        {
          if (data[i + offset] == null) continue;
          if (((NAME64DBLDBLDBL[]) hDataObject)[i+start] != null)
            ((NAME64DBLDBLDBL[]) hDataObject)[i+start].setValues(data[i + offset].name, data[i + offset].fval,
                data[i + offset].ival, 0);
          else
            ((NAME64DBLDBLDBL[]) hDataObject)[i+start] = new NAME64DBLDBLDBL(data[i + offset].name,
                data[i + offset].fval, data[i + offset].ival);
        }
        break;
      case TFormat.CF_USTRING:
        for (int i = 0; i < maxlen; i++)
        {
          if (data[i] == null) continue;
          if (((USTRING[]) hDataObject)[i+start] != null)
            ((USTRING[]) hDataObject)[i+start].setValues(data[i + offset].ival, data[i + offset].fval, 0, 0, data[i
                + offset].name);
          else
            ((USTRING[]) hDataObject)[i+start] = new USTRING(data[i + offset].ival, data[i + offset].fval, 0, 0,
                data[i + offset].name);
        }
        break;
      case TFormat.CF_UNAME:
        for (int i = 0; i < maxlen; i++)
        {
          if (data[i + offset] == null) continue;
          if (((UNAME[]) hDataObject)[i+start] != null)
            ((UNAME[]) hDataObject)[i+start].setValues(data[i + offset].ival, data[i + offset].fval, 0, 0, data[i
                + offset].name);
          else
            ((UNAME[]) hDataObject)[i+start] = new UNAME(data[i + offset].ival, data[i + offset].fval, 0, 0, data[i
                + offset].name);
        }
        break;
      default:
        return TErrorList.illegal_format;
    }
    if (redim) dArrayLength = Math.min(dArrayLength, length);
    repackageDataForOutput();
    return 0;
  } }
  /**
   * Puts the current canonical data set into the underlying data byte
   *        stream Puts the input data set into the underlying byte stream,
   *        properly swapped and aligned. It replaces the contents of the
   *        original, referenced data array with that of the data object given.
   * @param data
   *          is the data object whose contents should be copied into the
   *          underlying byte stream and referenced data object.
   * @note If this TDataType object is used as input data in persistent links,
   *       then the data object might well be 'locked' meaning that the
   *       underlying data will NOT be changed, as this would constitute a
   *       change in contract.
   * @return 0 upon success or a tine error code
   */
  public int putData(NAME32DBLDBL[] data)
  {
    return putData(data, data != null ? data.length : 0, 0);
  }
  /**
   * Puts the current canonical data set into the underlying data byte
   *        stream Puts the input data set into the underlying byte stream,
   *        properly swapped and aligned. It replaces the contents of the
   *        original, referenced data array with that of the data object given.
   * @param data
   *          is the data object whose contents should be copied into the
   *          underlying byte stream and referenced data object.
   * @param length
   *          is the number of elements from the give data object to copy
   * @param offset
   *          is the starting offset within the array to copy
   * @note If this TDataType object is used as input data in persistent links,
   *       then the data object might well be 'locked' meaning that the
   *       underlying data will NOT be changed, as this would constitute a
   *       change in contract.
   * @return 0 upon success or a tine error code
   */
  public int putData(NAME32DBLDBL[] data, int length, int offset)
  {
    return putData(0,data,length,offset,true);
  }
  public int putData(int start, NAME32DBLDBL[] data, int length, int offset, boolean redim)
  { synchronized (hDataObject) {
    if (data == null || length <= 0 || offset < 0) return TErrorList.argument_list_error;
    if (length > data.length - offset) length = data.length - offset;
    if (length <= 0) return TErrorList.argument_list_error;
    if (isLocked) throw new InputDataLockedException(lockedMessage);
    int maxlen = Math.min(dArrayLength, length);
    if (maxlen > dArrayLength - start) maxlen = dArrayLength - start;
    switch (dFormat)
    {
      case TFormat.CF_NAME16II:
        for (int i = 0; i < maxlen; i++)
        {
          if (data[i + offset] == null) continue;
          if (((NAME16II[]) hDataObject)[i+start] != null)
            ((NAME16II[]) hDataObject)[i+start].setValues(data[i + offset].name, (int) data[i + offset].d1val,
                (int) data[i + offset].d2val);
          else
            ((NAME16II[]) hDataObject)[i+start] = new NAME16II(data[i + offset].name, (int) data[i + offset].d1val,
                (int) data[i + offset].d2val);
        }
        break;
      case TFormat.CF_NAME16FI:
        for (int i = 0; i < maxlen; i++)
        {
          if (data[i + offset] == null) continue;
          if (((NAME16FI[]) hDataObject)[i+start] != null)
            ((NAME16FI[]) hDataObject)[i+start].setValues(data[i + offset].name, (float) data[i + offset].d1val,
                (int) data[i + offset].d2val);
          else
            ((NAME16FI[]) hDataObject)[i+start] = new NAME16FI(data[i].name, (float) data[i + offset].d1val,
                (int) data[i + offset].d2val);
        }
        break;
      case TFormat.CF_NAME16DBLDBL:
        for (int i = 0; i < maxlen; i++)
        {
          if (data[i + offset] == null) continue;
          if (((NAME16DBLDBL[]) hDataObject)[i+start] != null)
            ((NAME16DBLDBL[]) hDataObject)[i+start].setValues(data[i + offset].name, data[i + offset].d1val, data[i
                + offset].d2val);
          else
            ((NAME16DBLDBL[]) hDataObject)[i] = new NAME16DBLDBL(data[i + offset].name,
                data[i + offset].d1val, data[i + offset].d2val);
        }
        break;
      case TFormat.CF_NAME32DBLDBL:
        for (int i = 0; i < maxlen; i++)
        {
          if (data[i + offset] == null) continue;
          if (((NAME32DBLDBL[]) hDataObject)[i+start] != null)
            ((NAME32DBLDBL[]) hDataObject)[i+start].setValues(data[i + offset].name, data[i + offset].d1val, data[i
                + offset].d2val);
          else
            ((NAME32DBLDBL[]) hDataObject)[i] = data[i + offset];
        }
        break;
      case TFormat.CF_NAME64DBL:
        for (int i = 0; i < maxlen; i++)
        {
          if (data[i + offset] == null) continue;
          if (((NAME64DBL[]) hDataObject)[i+start] != null)
            ((NAME64DBL[]) hDataObject)[i+start].setValues(data[i + offset].name, data[i + offset].d1val);
          else
            ((NAME64DBL[]) hDataObject)[i+start] = new NAME64DBL(data[i + offset].name, data[i + offset].d1val);
        }
        break;
      case TFormat.CF_NAME64DBLDBL:
        for (int i = 0; i < maxlen; i++)
        {
          if (data[i + offset] == null) continue;
          if (((NAME64DBLDBL[]) hDataObject)[i+start] != null)
            ((NAME64DBLDBL[]) hDataObject)[i+start].setValues(data[i + offset].name, data[i + offset].d1val, data[i
                + offset].d2val);
          else
            ((NAME64DBLDBL[]) hDataObject)[i+start] = new NAME64DBLDBL(data[i + offset].name,
                data[i + offset].d1val, data[i + offset].d2val);
        }
        break;
      case TFormat.CF_NAME64DBLDBLDBL:
        for (int i = 0; i < maxlen; i++)
        {
          if (data[i + offset] == null) continue;
          if (((NAME64DBLDBLDBL[]) hDataObject)[i+start] != null)
            ((NAME64DBLDBLDBL[]) hDataObject)[i+start].setValues(data[i + offset].name, data[i + offset].d1val,
                data[i + offset].d2val, 0);
          else
            ((NAME64DBLDBLDBL[]) hDataObject)[i+start] = new NAME64DBLDBLDBL(data[i + offset].name,
                data[i + offset].d1val, data[i + offset].d2val);
        }
        break;
      case TFormat.CF_USTRING:
        for (int i = 0; i < maxlen; i++)
        {
          if (data[i + offset] == null) continue;
          if (((USTRING[]) hDataObject)[i+start] != null)
            ((USTRING[]) hDataObject)[i+start].setValues((int) data[i + offset].d2val,
                (float) data[i + offset].d1val, 0, 0, data[i + offset].name);
          else
            ((USTRING[]) hDataObject)[i+start] = new USTRING((int) data[i].d2val, (float) data[i + offset].d1val,
                0, 0, data[i + offset].name);
        }
        break;
      case TFormat.CF_UNAME:
        for (int i = 0; i < maxlen; i++)
        {
          if (data[i + offset] == null) continue;
          if (((UNAME[]) hDataObject)[i+start] != null)
            ((UNAME[]) hDataObject)[i+start].setValues((int) data[i + offset].d2val,
                (float) data[i + offset].d1val, 0, 0, data[i + offset].name);
          else
            ((UNAME[]) hDataObject)[i+start] = new UNAME((int) data[i + offset].d2val,
                (float) data[i + offset].d1val, 0, 0, data[i + offset].name);
        }
        break;
      default:
        return TErrorList.illegal_format;
    }
    if (redim) dArrayLength = Math.min(dArrayLength, length);
    repackageDataForOutput();
    return 0;
  } }
  /**
   * Puts the current canonical data set into the underlying data byte
   *        stream Puts the input data set into the underlying byte stream,
   *        properly swapped and aligned. It replaces the contents of the
   *        original, referenced data array with that of the data object given.
   * @param data
   *          is data primitive whose value is copied into the underlying byte
   *          stream and referenced data object.
   * @note If this TDataType object is used as input data in persistent links,
   *       then the data object might well be 'locked' meaning that the
   *       underlying data will NOT be changed, as this would constitute a
   *       change in contract.
   * @return 0 upon success or a tine error code
   */
  public int putData(INTINT data)
  {
    INTINT[] d = new INTINT[1];
    d[0] = data;
    return putData(d);
  }
  /**
   * Puts the current canonical data set into the underlying data byte
   *        stream Puts the input data set into the underlying byte stream,
   *        properly swapped and aligned. It replaces the contents of the
   *        original, referenced data array with that of the data object given.
   * @param data
   *          is the data object whose contents should be copied into the
   *          underlying byte stream and referenced data object.
   * @note If this TDataType object is used as input data in persistent links,
   *       then the data object might well be 'locked' meaning that the
   *       underlying data will NOT be changed, as this would constitute a
   *       change in contract.
   * @return 0 upon success or a tine error code
   */
  public int putData(INTINT[] data)
  {
    return putData(data, data != null ? data.length : 0, 0);
  }
  /**
   * Puts the current canonical data set into the underlying data byte
   *        stream Puts the input data set into the underlying byte stream,
   *        properly swapped and aligned. It replaces the contents of the
   *        original, referenced data array with that of the data object given.
   * @param data
   *          is the data object whose contents should be copied into the
   *          underlying byte stream and referenced data object.
   * @param length
   *          is the number of elements from the give data object to copy
   * @param offset
   *          is the starting offset within the array to copy
   * @note If this TDataType object is used as input data in persistent links,
   *       then the data object might well be 'locked' meaning that the
   *       underlying data will NOT be changed, as this would constitute a
   *       change in contract.
   * @return 0 upon success or a tine error code
   */
  public int putData(INTINT[] data, int length, int offset)
  {
    return putData(0,data,length,offset,true);
  }
  public int putData(int start, INTINT[] data, int length, int offset, boolean redim)
  { synchronized (hDataObject) {
    if (data == null || length <= 0 || offset < 0) return TErrorList.argument_list_error;
    if (length > data.length - offset) length = data.length - offset;
    if (length <= 0) return TErrorList.argument_list_error;
    if (isLocked) throw new InputDataLockedException(lockedMessage);
    int maxlen = Math.min(dArrayLength, length);
    if (maxlen > dArrayLength - start) maxlen = dArrayLength - start;
    switch (dFormat)
    {
      case TFormat.CF_INTINT:
        for (int i = 0; i < maxlen; i++)
        {
          if (data[i + offset] == null) continue;
          if (((INTINT[]) hDataObject)[i+start] != null)
            ((INTINT[]) hDataObject)[i+start].setValues(data[i + offset].i1val, data[i + offset].i2val);
          else
            ((INTINT[]) hDataObject)[i+start] = data[i + offset];
        }
        break;
      case TFormat.CF_FLTINT:
        for (int i = 0; i < maxlen; i++)
        {
          if (data[i + offset] == null) continue;
          if (((FLTINT[]) hDataObject)[i+start] != null)
            ((FLTINT[]) hDataObject)[i+start].setValues(data[i + offset].i1val, data[i + offset].i2val);
          else
            ((FLTINT[]) hDataObject)[i+start] = new FLTINT(data[i + offset].i1val, data[i + offset].i2val);
        }
        break;
      case TFormat.CF_DBLDBL:
        for (int i = 0; i < maxlen; i++)
        {
          if (data[i + offset] == null) continue;
          if (((DBLDBL[]) hDataObject)[i+start] != null)
            ((DBLDBL[]) hDataObject)[i+start].setValues(data[i + offset].i1val, data[i + offset].i2val);
          else
            ((DBLDBL[]) hDataObject)[i+start] = new DBLDBL(data[i + offset].i1val, data[i + offset].i2val);
        }
        break;
      case TFormat.CF_DBLDBLDBL:
        for (int i = 0; i < maxlen; i++)
        {
          if (data[i + offset] == null) continue;
          if (((DBLDBLDBL[]) hDataObject)[i+start] != null)
            ((DBLDBLDBL[]) hDataObject)[i+start].setValues(data[i + offset].i1val, 0, data[i + offset].i2val);
          else
            ((DBLDBLDBL[]) hDataObject)[i+start] = new DBLDBLDBL(data[i + offset].i1val, 0, data[i + offset].i2val);
        }
        break;
      case TFormat.CF_FLOAT:
        for (int i = 0; i < maxlen; i++)
        {
          if (data[i + offset] == null) continue;
          ((float[]) hDataObject)[i+start] = data[i + offset].i1val;
        }
        break;
      case TFormat.CF_DOUBLE:
        for (int i = 0; i < maxlen; i++)
        {
          if (data[i + offset] == null) continue;
          ((double[]) hDataObject)[i+start] = data[i + offset].i1val;
        }
        break;
      case TFormat.CF_INT64:
        for (int i = 0; i < dArrayLength; i++)
        {
          if (data[i + offset] == null) continue;
          ((long[]) hDataObject)[i+start] = data[i + offset].i1val;
        }
        break;
      case TFormat.CF_INT32:
        for (int i = 0; i < dArrayLength; i++)
        {
          if (data[i + offset] == null) continue;
          ((int[]) hDataObject)[i+start] = data[i + offset].i1val;
        }
        break;
      default:
        return TErrorList.illegal_format;
    }
    if (redim) dArrayLength = Math.min(dArrayLength, length);
    repackageDataForOutput();
    return 0;
  } }
  /**
   * Puts the current canonical data set into the underlying data byte
   *        stream Puts the input data set into the underlying byte stream,
   *        properly swapped and aligned. It replaces the contents of the
   *        original, referenced data array with that of the data object given.
   * @param data
   *          is data primitive whose value is copied into the underlying byte
   *          stream and referenced data object.
   * @note If this TDataType object is used as input data in persistent links,
   *       then the data object might well be 'locked' meaning that the
   *       underlying data will NOT be changed, as this would constitute a
   *       change in contract.
   * @return 0 upon success or a tine error code
   */
  public int putData(DBLDBL data)
  {
    DBLDBL[] d = new DBLDBL[1];
    d[0] = data;
    return putData(d);
  }
  /**
   * Puts the current canonical data set into the underlying data byte
   *        stream Puts the input data set into the underlying byte stream,
   *        properly swapped and aligned. It replaces the contents of the
   *        original, referenced data array with that of the data object given.
   * @param data
   *          is the data object whose contents should be copied into the
   *          underlying byte stream and referenced data object.
   * @note If this TDataType object is used as input data in persistent links,
   *       then the data object might well be 'locked' meaning that the
   *       underlying data will NOT be changed, as this would constitute a
   *       change in contract.
   * @return 0 upon success or a tine error code
   */
  public int putData(DBLDBL[] data)
  {
    return putData(data, data != null ? data.length : 0, 0);
  }
  public int putData(DBLTIME[] data)
  {
    return putData(data, data != null ? data.length : 0, 0);
  }
  /**
   * Puts the current canonical data set into the underlying data byte
   *        stream Puts the input data set into the underlying byte stream,
   *        properly swapped and aligned. It replaces the contents of the
   *        original, referenced data array with that of the data object given.
   * @param data
   *          is the data object whose contents should be copied into the
   *          underlying byte stream and referenced data object.
   * @param length
   *          is the number of elements from the give data object to copy
   * @param offset
   *          is the starting offset within the array to copy
   * @note If this TDataType object is used as input data in persistent links,
   *       then the data object might well be 'locked' meaning that the
   *       underlying data will NOT be changed, as this would constitute a
   *       change in contract.
   * @return 0 upon success or a tine error code
   */
  public int putData(DBLDBL[] data, int length, int offset)
  {
    return putData(0,data,length,offset,true);
  }
  public int putData(int start, DBLDBL[] data, int length, int offset, boolean redim)
  { synchronized (hDataObject) {
    if (data == null || length <= 0 || offset < 0) return TErrorList.argument_list_error;
    if (length > data.length - offset) length = data.length - offset;
    if (length <= 0) return TErrorList.argument_list_error;
    if (isLocked) throw new InputDataLockedException(lockedMessage);
    int maxlen = Math.min(dArrayLength, length);
    if (maxlen > dArrayLength - start) maxlen = dArrayLength - start;
    switch (dFormat)
    {
      case TFormat.CF_INTINT:
        for (int i = 0; i < maxlen; i++)
        {
          if (data[i + offset] == null) continue;
          if (((INTINT[]) hDataObject)[i+start] != null)
            ((INTINT[]) hDataObject)[i+start].setValues((int) data[i + offset].d1val, (int) data[i + offset].d2val);
          else
            ((INTINT[]) hDataObject)[i+start] = new INTINT((int) data[i + offset].d1val,
                (int) data[i + offset].d2val);
        }
        break;
      case TFormat.CF_FLTINT:
        for (int i = 0; i < maxlen; i++)
        {
          if (data[i + offset] == null) continue;
          if (((FLTINT[]) hDataObject)[i+start] != null)
            ((FLTINT[]) hDataObject)[i+start].setValues((float) data[i + offset].d1val,
                (int) data[i + offset].d2val);
          else
            ((FLTINT[]) hDataObject)[i+start] = new FLTINT((float) data[i + offset].d1val,
                (int) data[i + offset].d2val);
        }
        break;
      case TFormat.CF_DBLDBL:
        for (int i = 0; i < maxlen; i++)
        {
          if (data[i + offset] == null) continue;
          if (((DBLDBL[]) hDataObject)[i+start] != null)
            ((DBLDBL[]) hDataObject)[i+start].setValues(data[i + offset].d1val, data[i + offset].d2val);
          else
            ((DBLDBL[]) hDataObject)[i+start] = data[i + offset];
        }
        break;
      case TFormat.CF_DBLTIME:
        for (int i = 0; i < maxlen; i++)
        {
          if (data[i + offset] == null) continue;
          if (((DBLTIME[]) hDataObject)[i+start] != null)
            ((DBLTIME[]) hDataObject)[i+start].setValues(data[i + offset].d1val, data[i + offset].d2val);
          else
            ((DBLTIME[]) hDataObject)[i+start] = new DBLTIME(data[i + offset]);
        }
        break;      case TFormat.CF_FLOAT:
        for (int i = 0; i < maxlen; i++)
        {
          if (data[i + offset] == null) continue;
          ((float[]) hDataObject)[i+start] = (float) data[i + offset].d1val;
        }
        break;
      case TFormat.CF_DOUBLE:
        for (int i = 0; i < maxlen; i++)
        {
          if (data[i + offset] == null) continue;
          ((double[]) hDataObject)[i+start] = data[i + offset].d1val;
        }
        break;
      case TFormat.CF_INT64:
        for (int i = 0; i < maxlen; i++)
        {
          if (data[i + offset] == null) continue;
          ((long[]) hDataObject)[i+start] = (long) data[i + offset].d2val;
        }
        break;
      case TFormat.CF_INT32:
        for (int i = 0; i < maxlen; i++)
        {
          if (data[i + offset] == null) continue;
          ((int[]) hDataObject)[i+start] = (int) data[i + offset].d2val;
        }
        break;
      default:
        return TErrorList.illegal_format;
    }
    if (redim) dArrayLength = Math.min(dArrayLength, length);
    repackageDataForOutput();
    return 0;
  } }
  public int putData(DBLTIME[] data, int length, int offset)
  {
    return putData(0,data,length,offset,true);	
  }
  public int putData(int start, DBLTIME[] data, int length, int offset, boolean redim)
  {    
    return putData(start,(DBLDBL[])data,length,offset,redim);
  }
  /**
   * Puts the current canonical data set into the underlying data byte
   *        stream Puts the input data set into the underlying byte stream,
   *        properly swapped and aligned. It replaces the contents of the
   *        original, referenced data array with that of the data object given.
   * @param data
   *          is data primitive whose value is copied into the underlying byte
   *          stream and referenced data object.
   * @note If this TDataType object is used as input data in persistent links,
   *       then the data object might well be 'locked' meaning that the
   *       underlying data will NOT be changed, as this would constitute a
   *       change in contract.
   * @return 0 upon success or a tine error code
   */
  public int putData(DBLDBLDBL data)
  {
    DBLDBLDBL[] d = new DBLDBLDBL[1];
    d[0] = data;
    return putData(d);
  }
  /**
   * Puts the current canonical data set into the underlying data byte
   *        stream Puts the input data set into the underlying byte stream,
   *        properly swapped and aligned. It replaces the contents of the
   *        original, referenced data array with that of the data object given.
   * @param data
   *          is the data object whose contents should be copied into the
   *          underlying byte stream and referenced data object.
   * @note If this TDataType object is used as input data in persistent links,
   *       then the data object might well be 'locked' meaning that the
   *       underlying data will NOT be changed, as this would constitute a
   *       change in contract.
   * @return 0 upon success or a tine error code
   */
  public int putData(DBLDBLDBL[] data)
  {
    return putData(data, data != null ? data.length : 0, 0);
  }
  /**
   * Puts the current canonical data set into the underlying data byte
   *        stream Puts the input data set into the underlying byte stream,
   *        properly swapped and aligned. It replaces the contents of the
   *        original, referenced data array with that of the data object given.
   * @param data
   *          is the data object whose contents should be copied into the
   *          underlying byte stream and referenced data object.
   * @param length
   *          is the number of elements from the give data object to copy
   * @param offset
   *          is the starting offset within the array to copy
   * @note If this TDataType object is used as input data in persistent links,
   *       then the data object might well be 'locked' meaning that the
   *       underlying data will NOT be changed, as this would constitute a
   *       change in contract.
   * @return 0 upon success or a tine error code
   */
  public int putData(DBLDBLDBL[] data, int length, int offset)
  {
    return putData(0,data,length,offset,true);
  }
  public int putData(int start, DBLDBLDBL[] data, int length, int offset, boolean redim)
  { synchronized (hDataObject) {
    if (data == null || length <= 0 || offset < 0) return TErrorList.argument_list_error;
    if (length > data.length - offset) length = data.length - offset;
    if (length <= 0) return TErrorList.argument_list_error;
    if (isLocked) throw new InputDataLockedException(lockedMessage);
    int maxlen = Math.min(dArrayLength, length);
    if (maxlen > dArrayLength - start) maxlen = dArrayLength - start;
    switch (dFormat)
    {
      case TFormat.CF_FLTINTINT:
        for (int i = 0; i < maxlen; i++)
        {
          if (data[i + offset] == null) continue;
          if (((FLTINTINT[]) hDataObject)[i+start] != null)
            ((FLTINTINT[]) hDataObject)[i+start].setValues((float) data[i + offset].d1val,
                (int) data[i + offset].d2val, (int) data[i + offset].d3val);
          else
            ((FLTINTINT[]) hDataObject)[i+start] = new FLTINTINT((float) data[i + offset].d1val, (int) data[i
                + offset].d2val, (int) data[i + offset].d2val);
        }
        break;
      case TFormat.CF_DBLDBLDBL:
        for (int i = 0; i < maxlen; i++)
        {
          if (data[i + offset] == null) continue;
          if (((DBLDBLDBL[]) hDataObject)[i+start] != null)
            ((DBLDBLDBL[]) hDataObject)[i+start].setValues(data[i + offset].d1val, data[i + offset].d2val, data[i
                + offset].d3val);
          else
            ((DBLDBLDBL[]) hDataObject)[i+start] = data[i + offset];
        }
        break;
      case TFormat.CF_FLOAT:
        for (int i = 0; i < maxlen; i++)
        {
          if (data[i + offset] == null) continue;
          ((float[]) hDataObject)[i+start] = (float) data[i + offset].d1val;
        }
        break;
      case TFormat.CF_DOUBLE:
        for (int i = 0; i < maxlen; i++)
        {
          if (data[i + offset] == null) continue;
          ((double[]) hDataObject)[i+start] = data[i + offset].d1val;
        }
        break;
      case TFormat.CF_INT64:
        for (int i = 0; i < maxlen; i++)
        {
          if (data[i + offset] == null) continue;
          ((long[]) hDataObject)[i+start] = (long) data[i + offset].d3val;
        }
        break;
      case TFormat.CF_INT32:
        for (int i = 0; i < maxlen; i++)
        {
          if (data[i + offset] == null) continue;
          ((int[]) hDataObject)[i+start] = (int) data[i + offset].d3val;
        }
        break;
      default:
        return TErrorList.illegal_format;
    }
    if (redim) dArrayLength = Math.min(dArrayLength, length);
    repackageDataForOutput();
    return 0;
  } }
  /**
   * Puts the current canonical data set into the underlying data byte
   *        stream Puts the input data set into the underlying byte stream,
   *        properly swapped and aligned. It replaces the contents of the
   *        original, referenced data array with that of the data object given.
   * @param data
   *          is the data object whose contents should be copied into the
   *          underlying byte stream and referenced data object.
   * @note If this TDataType object is used as input data in persistent links,
   *       then the data object might well be 'locked' meaning that the
   *       underlying data will NOT be changed, as this would constitute a
   *       change in contract.
   * @return 0 upon success or a tine error code
   */
  public int putData(String[] data)
  {
    return putData(data, data != null ? data.length : 0, 0);
  }
  /**
   * Puts the current canonical data set into the underlying data byte
   *        stream Puts the input data set into the underlying byte stream,
   *        properly swapped and aligned. It replaces the contents of the
   *        original, referenced data array with that of the data object given.
   * @param data
   *          is the data object whose contents should be copied into the
   *          underlying byte stream and referenced data object.
   * @param length
   *          is the number of elements from the give data object to copy
   * @param offset
   *          is the starting offset within the array to copy
   * @note If this TDataType object is used as input data in persistent links,
   *       then the data object might well be 'locked' meaning that the
   *       underlying data will NOT be changed, as this would constitute a
   *       change in contract.
   * @return 0 upon success or a tine error code
   */
  public int putData(String[] data, int length, int offset)
  {
    return putData(0,data,length,offset,true);
  }
  public int putData(int start, String[] data, int length, int offset, boolean redim)
  { synchronized (hDataObject) {
    if (data == null || length <= 0 || offset < 0) return TErrorList.argument_list_error;
    if (length > data.length - offset) length = data.length - offset;
    int maxlen = dArrayLength;
    if (maxlen > length) maxlen = length;
    if (maxlen > dArrayLength - start) maxlen = dArrayLength - start;
    if (length <= 0) return TErrorList.argument_list_error;
    if (isLocked) throw new InputDataLockedException(lockedMessage);
    if (TFormat.isVariableLengthString(dFormat))
    { // String array to CF_STRING, etc.
      if (hDataObject instanceof String[])
      {
        for (int i = 0; i < maxlen; i++)
          ((String[]) hDataObject)[i+start] = data[i + offset];
      }
      else if (hDataObject instanceof KEYVALUE[])
      {
        for (int i = 0; i < maxlen; i++)
          ((KEYVALUE[]) hDataObject)[i+start] = new KEYVALUE(data[i + offset]);      
      }
      //if (dArrayLength > maxlen) dArrayLength = maxlen;
      dSizeInBytes = 0;
      for (int i = 0; i < length; i++)
      { // first see how much string space we need (going out)
        dSizeInBytes += (data[i] != null) ? data[i].length() : 0;
        dSizeInBytes++; // add terminating 0
      }
    }
    else if (TFormat.isString(dFormat))
    { // String array to non-CF_STRING type
      try 
      {
        if (hDataObject instanceof NAME[])
        {
          for (int i = 0; i < maxlen; i++)
            ((NAME[]) hDataObject)[i+start].setName(data[i+offset]);
        }
        else if (hDataObject instanceof char[])
        { // TEXT object being passed a NAMExx
          hDataObject = data[offset].toCharArray();
          maxlen *= data[offset].length();
        }
        else if (dFormat == TFormat.CF_USTRING)
        {
          for (int i = 0; i < maxlen; i++)
            ((USTRING[]) hDataObject)[i+start].setString(data[i+offset]);          
        }
      }
      catch (Exception e)
      {
        e.printStackTrace();
        MsgLog.log("TDataType", "cannot convert " + TFormat.toString(getObjectFormat(data)) + " to "
            + TFormat.toString(dFormat), TErrorList.illegal_format, e, 0);
      }
      if (redim) dArrayLength = Math.min(dArrayLength, length);
    }
    repackageDataForOutput();
    return 0;
  } }
  /**
   * Puts the current canonical data set into the underlying data byte
   *        stream Puts the input data set into the underlying byte stream,
   *        properly swapped and aligned. It replaces the contents of the
   *        original, referenced data array with that of the data object given.
   * @param data
   *          is the data object whose contents should be copied into the
   *          underlying byte stream and referenced data object.
   * @note If this TDataType object is used as input data in persistent links,
   *       then the data object might well be 'locked' meaning that the
   *       underlying data will NOT be changed, as this would constitute a
   *       change in contract.
   * @return 0 upon success or a tine error code
   */
  public int putData(KEYVALUE[] data)
  {
    return putData(data, data != null ? data.length : 0, 0);
  }
  /**
   * Puts the current canonical data set into the underlying data byte
   *        stream Puts the input data set into the underlying byte stream,
   *        properly swapped and aligned. It replaces the contents of the
   *        original, referenced data array with that of the data object given.
   * @param data
   *          is the data object whose contents should be copied into the
   *          underlying byte stream and referenced data object.
   * @param length
   *          is the number of elements from the give data object to copy
   * @param offset
   *          is the starting offset within the array to copy
   * @note If this TDataType object is used as input data in persistent links,
   *       then the data object might well be 'locked' meaning that the
   *       underlying data will NOT be changed, as this would constitute a
   *       change in contract.
   * @return 0 upon success or a tine error code
   */
  public int putData(KEYVALUE[] data, int length, int offset)
  {
    return putData(0,data,length,offset,true);
  }
  public int putData(int start, KEYVALUE[] data, int length, int offset, boolean redim)
  { synchronized (hDataObject) {
    if (data == null || length <= 0 || offset < 0) return TErrorList.argument_list_error;
    if (length > data.length - offset) length = data.length - offset;
    int maxlen = dArrayLength;
    if (maxlen > length) maxlen = length;
    if (maxlen > dArrayLength - start) maxlen = dArrayLength - start;
    if (length <= 0) return TErrorList.argument_list_error;
    if (!TFormat.isVariableLengthString(dFormat)) return TErrorList.illegal_format;
    if (isLocked) throw new InputDataLockedException(lockedMessage);
    if (hDataObject instanceof KEYVALUE[])
    {
      for (int i = 0; i < maxlen; i++)
        ((KEYVALUE[]) hDataObject)[i+start] = data[i + offset];
    }
    else if (hDataObject instanceof String[])
    {
      for (int i = 0; i < maxlen; i++)
        ((String[]) hDataObject)[i+start] = data[i + offset].getString();      
    }
    if (dArrayLength > maxlen) dArrayLength = maxlen;
    dSizeInBytes = 0;
    for (int i = 0; i < length; i++)
    { // first see how much string space we need (going out)
      dSizeInBytes += (data[i] != null) ? data[i].getString().length() : 0;
      dSizeInBytes++; // add terminating 0
    }
    if (redim) dArrayLength = Math.min(dArrayLength, length);
    repackageDataForOutput();
    return 0;
  } }
  public int putData(String value, int offset)
  {
    if (offset < 0) return TErrorList.argument_list_error;
    if (dFormat != TFormat.CF_STRING) return TErrorList.illegal_format;
    if (dArrayLength < offset) return TErrorList.dimension_error;
    if (isLocked) throw new InputDataLockedException(lockedMessage);
    ((String[]) hDataObject)[offset] = value;
    dSizeInBytes = 0;
    String s = null;
    for (int i = 0; i < dArrayLength; i++)
    { // first see how much string space we need (going out)
      s = ((String[]) hDataObject)[i];
      dSizeInBytes += (s != null) ? s.length() : 0;
      dSizeInBytes++; // add terminating 0
    }
    repackageDataForOutput();
    return 0;
  }
  public int putData(KEYVALUE value, int offset)
  {
    if (offset < 0) return TErrorList.argument_list_error;
    if (dFormat != TFormat.CF_KEYVALUE) return TErrorList.illegal_format;
    if (dArrayLength < offset) return TErrorList.dimension_error;
    if (isLocked) throw new InputDataLockedException(lockedMessage);
    ((KEYVALUE[]) hDataObject)[offset] = value;
    dSizeInBytes = 0;
    KEYVALUE s = null;
    for (int i = 0; i < dArrayLength; i++)
    { // first see how much string space we need (going out)
      s = ((KEYVALUE[]) hDataObject)[i];
      dSizeInBytes += (s != null) ? s.getString().length() : 0;
      dSizeInBytes++; // add terminating 0
    }
    repackageDataForOutput();
    return 0;
  }
  public int putData(String value, short format, int offset)
  {
    switch (format)
    {
      case TFormat.CF_INT16:
        return putData(Short.parseShort(value),offset);
      case TFormat.CF_BOOLEAN:
        boolean bv = false;
        if (value != null) bv = (value.compareToIgnoreCase("true") == 0)
                             || (value.compareToIgnoreCase("yes") == 0);
        return putData(bv,offset);
      case TFormat.CF_INT32:
        return putData(Integer.parseInt(value),offset);
      case TFormat.CF_INT64:
        return putData(Long.parseLong(value),offset);
      case TFormat.CF_FLOAT:
        return putData(Float.parseFloat(value),offset);
      case TFormat.CF_DOUBLE:
        return putData(Double.parseDouble(value),offset);
      case TFormat.CF_NAME8: 
      case TFormat.CF_NAME16: 
      case TFormat.CF_NAME32: 
      case TFormat.CF_NAME48: 
      case TFormat.CF_NAME64: 
        return putData(new NAME64(value),offset);
      case TFormat.CF_STRING:
        return putData(value,offset);
      case TFormat.CF_KEYVALUE:
        return putData(value,offset);
      default:
        break;
    }
    return TErrorList.illegal_format;    
  }
  public int putData(ASPECTRUM[] data)
  {
    return putData(data,data.length,0);
  }
  public int putData(ASPECTRUM[] data, int length, int offset)
  {
    return putData(0,data,length,offset,true);
  }
  public int putData(int start, ASPECTRUM[] data, int length, int offset, boolean redim)
  { synchronized (hDataObject) {
    if (dFormat != TFormat.CF_ASPECTRUM) return TErrorList.illegal_format;
    if (data == null || length <= 0 || offset < 0) return TErrorList.argument_list_error;
    if (length > data.length - offset) length = data.length - offset;
    int len = 0;
    try { len = Integer.parseInt(dTag); } catch (Exception ignore) {};
    if (len == 0) len = dArrayLength;
    len -= start;
    if (len > length) len = length;
    if (length <= 0) return TErrorList.argument_list_error;
    if (isLocked) throw new InputDataLockedException(lockedMessage);
    dSizeInBytes = 0;
    for (int i = 0; i < len; i++)
    {
      dSizeInBytes += ASPECTRUM.hdrSizeInBytes + 8 + data[i+offset].length * 4;
      ((ASPECTRUM[]) hDataObject)[i+start] = data[i + offset];
    }
    dArrayLength = dSizeInBytes;
    dTag = new String(""+len);
    if (redim) dArrayLength = Math.min(dArrayLength, length);
    repackageDataForOutput();
    return 0;
  } }
  /**
   * Puts the current canonical data set into the underlying data byte
   *        stream Puts the input data set into the underlying byte stream,
   *        properly swapped and aligned. It replaces the contents of the
   *        original, referenced data array with that of the data object given.
   * @param data
   *          is data primitive whose value is copied into the underlying byte
   *          stream and referenced data object.
   * @note If this TDataType object is used as input data in persistent links,
   *       then the data object might well be 'locked' meaning that the
   *       underlying data will NOT be changed, as this would constitute a
   *       change in contract.
   * @return 0 upon success or a tine error code
   */
  public int putData(String data)
  {
    if (data == null) return 0;
    if (hDataObject == null || dFormat == TFormat.CF_NULL || dArrayLength == 0)
    {
      if (TLinkFactory.debugLevel > 0) DbgLog.log("TDataType.putData","attempt to assign string to null data object");
      return 0;
    }
    int len = data.length();
    if (isLocked) throw new InputDataLockedException(lockedMessage);
    switch (dFormat)
    {
      case TFormat.CF_NAME8:
        ((NAME8[]) hDataObject)[0] = new NAME8(data);
        break;
      case TFormat.CF_NAME16:
        ((NAME16[]) hDataObject)[0] = new NAME16(data);
        break;
      case TFormat.CF_NAME32:
        ((NAME32[]) hDataObject)[0] = new NAME32(data);
        break;
      case TFormat.CF_NAME48:
        ((NAME48[]) hDataObject)[0] = new NAME48(data);
        break;
      case TFormat.CF_NAME64:
        ((NAME64[]) hDataObject)[0] = new NAME64(data);
        break;
      case TFormat.CF_USTRING:
        ((USTRING[]) hDataObject)[0] = new USTRING(data);
        break;
      case TFormat.CF_XML:
      case TFormat.CF_TEXT:
        if (len < dArrayLength) dArrayLength = len;
        hDataObject = data.toCharArray();
        break;
      case TFormat.CF_STRING:
        ((String[]) hDataObject)[0] = new String(data);
        break;
      case TFormat.CF_BYTE:
        if (len < dArrayLength) dArrayLength = len;
        hDataObject = data.getBytes();
        break;
      case TFormat.CF_BOOLEAN:
        if (len < dArrayLength) dArrayLength = len;
        ((boolean[]) hDataObject)[0] = (data.compareToIgnoreCase("true") == 0)
                                    || (data.compareToIgnoreCase("yes") == 0);
      default:
        return TErrorList.illegal_format;
    }
    repackageDataForOutput();
    return 0;
  }
  public int putData(char[] data)
  {
    String s;
    if (data == null || data.length == 0) s = "";
    s = new String(data);
    return putData(s);
  }
  public int putData(char[] data,int length,int offset)
  {
    String s;
    if (data == null || data.length == 0) s = "";
    if (length <= 0 || offset < 0) return TErrorList.argument_list_error;
    if (length > data.length - offset) length = data.length - offset;
    if (length <= 0) return TErrorList.argument_list_error;
    s = new String(data,offset,length);
    return putData(s);
  }
  /**
   * Puts the current canonical data set into the underlying data byte
   *        stream Puts the input data set into the underlying byte stream,
   *        properly swapped and aligned. It replaces the contents of the
   *        original, referenced data array with that of the data object given.
   * @param data
   *          is Tagged Structure whose contents are to be copied into the
   *          underlying byte stream and referenced data object.
   * @note If this TDataType object is used as input data in persistent links,
   *       then the data object might well be 'locked' meaning that the
   *       underlying data will NOT be changed, as this would constitute a
   *       change in contract.
   * @return 0 upon success or a tine error code
   */
  public int putData(TTaggedStructure data)
  {
    if (data == null) return TErrorList.argument_list_error;
    if (dFormat != TFormat.CF_STRUCT) return TErrorList.illegal_format;
    if (dTag.compareTo(data.getName()) != 0) return TErrorList.invalid_structure_tag;
    if (hDataObject == null) return TErrorList.invalid_data;
    int dsiz = data.getSizeInBytes();
    int xsiz = 0; //data.getExtendedSizeInBytes();
    if (data.hasExtendedSpace())
    {
      data.setExtendedOffset(dsiz + xsiz); // <- watch here '+ xsiz
      data.fixExtendedBytes();
      xsiz += data.getExtendedSizeInBytes();
    }
    byte[] dbytes = data.toByteArray();
    if (dbytes == null) return TErrorList.code_failure;
    if (isLocked) throw new InputDataLockedException(lockedMessage);
    if (((byte[])hDataObject).length < dsiz + xsiz)
      hDataObject = new byte[dsiz + xsiz];
    System.arraycopy(dbytes, 0, (byte[]) hDataObject, 0, dsiz);
    if (xsiz > 0)
    {
      dbytes = data.toExtendedByteArray();
      System.arraycopy(dbytes, 0, (byte[]) hDataObject, dsiz, xsiz);
    }
    dbytes = (byte[]) hDataObject;
    if (hStructObject != null && hStructObject instanceof TTaggedStructure)
    { // this TDataType references another tagged structure !
      TTaggedStructure tts = (TTaggedStructure) hStructObject;
      tts.toStruct(dbytes);
    }
    else if (hStructObject instanceof TTaggedStructure[])
    { // this TDataType references an array of tagged structures !
      TTaggedStructure tts[] = (TTaggedStructure[]) hStructObject;
      tts[0].toStruct(dbytes);
    }
    repackageDataForOutput();
    return 0;
  }
  /**
   * Puts the current canonical data set into the underlying data byte
   *        stream Puts the input data set into the underlying byte stream,
   *        properly swapped and aligned. It replaces the contents of the
   *        original, referenced data array with that of the data object given.
   * @param data
   *          is the data object whose contents should be copied into the
   *          underlying byte stream and referenced data object.
   * @note If this TDataType object is used as input data in persistent links,
   *       then the data object might well be 'locked' meaning that the
   *       underlying data will NOT be changed, as this would constitute a
   *       change in contract.
   * @return 0 upon success or a tine error code
   */
  public int putData(TTaggedStructure[] data)
  {
    return putData(data, data != null ? data.length : 0, 0);
  }
  /**
   * Puts the current canonical data set into the underlying data byte
   *        stream Puts the input data set into the underlying byte stream,
   *        properly swapped and aligned. It replaces the contents of the
   *        original, referenced data array with that of the data object given.
   * @param data
   *          is the data object whose contents should be copied into the
   *          underlying byte stream and referenced data object.
   * @param length
   *          is the number of elements from the given data object to copy
   * @param offset
   *          is the starting offset within the array to copy
   * @note If this TDataType object is used as input data in persistent links,
   *       then the data object might well be 'locked' meaning that the
   *       underlying data will NOT be changed, as this would constitute a
   *       change in contract.
   * @return 0 upon success or a tine error code
   */
  public int putData(TTaggedStructure[] data, int length, int offset)
  {
    return putData(0,data,length,offset,true);
  }
  public int putData(int start, TTaggedStructure[] data, int length, int offset, boolean redim)
  { synchronized (hDataObject) {
    //if (data == null || length <= 0 || offset < 0) return TErrorList.argument_list_error;
    if (data == null || length < 0 || offset < 0) return TErrorList.argument_list_error;
    if (length > data.length - offset) length = data.length - offset;
    if (length == 0) 
    {
      setArrayLength(0,0);
      repackageDataForOutput();
      return 0;
    }
    if (dFormat != TFormat.CF_STRUCT) return TErrorList.illegal_format;
    if (dTag.compareTo(data[0].getName()) != 0) return TErrorList.invalid_structure_tag;
    if (hDataObject == null) return TErrorList.invalid_data;
    if (isLocked) throw new InputDataLockedException(lockedMessage);
    int xsiz = 0;
    
    int dsiz = data[0].getSizeInBytes();
    int asiz = getArrayLength();
    if (asiz > length) asiz = length;
    if (start > 0) asiz -= start;
    int extsiz = 0;
    if (data[0].hasExtendedSpace())
    { // suddenly, it's more complicated !
      for (int k=0; k<asiz; k++)
      { // set the boundary of the extended space and get the total
        data[k].setExtendedOffset(dsiz * asiz + xsiz); // <- watch here '+ xsiz
        data[k].fixExtendedBytes();
        xsiz += data[k].getExtendedSizeInBytes();
      }
      extsiz = xsiz;
    }
    if (((byte[])hDataObject).length < data.length*dsiz + xsiz)
      hDataObject = new byte[data.length*dsiz + xsiz];
    for (int i = 0; i < asiz; i++)
    {
      data[i + offset].toByteArray();
      System.arraycopy(data[i + offset].getByteArray(), 0, (byte[]) hDataObject, i * dsiz, dsiz);
    }
    if (data[0].hasExtendedSpace())
    { // now pump the extended bytes onto the stream
      int xoff = asiz * dsiz;
      for (int k = 0; k < asiz; k++)
      {
        xsiz = data[k].getExtendedSizeInBytes();
        System.arraycopy(data[k].toExtendedByteArray(), 0, (byte[]) hDataObject, xoff, xsiz);
        xoff += xsiz;
      }
    }
    if (hStructObject != null && hStructObject instanceof TTaggedStructure)
    {
      TTaggedStructure tts = (TTaggedStructure) hStructObject;
      tts.toStruct(data[offset].getByteArray());
    }
    else if (hStructObject instanceof TTaggedStructure[])
    {
      TTaggedStructure tts[] = (TTaggedStructure[]) hStructObject;
      for (int i = 0; i < asiz; i++)
      {
        tts[i+start].toStruct(data[i + offset].getByteArray());
      }
    }
    if (redim) setArrayLength(asiz,extsiz);
    repackageDataForOutput();
    return 0;
  } }
  public int putData(MDA data)
  { synchronized (hDataObject) {
    if (data == null) return TErrorList.argument_list_error;
    if (dFormat != TFormat.CF_MDA) return TErrorList.illegal_format;
    if (dTag.compareTo(data.getTag()) != 0) return TErrorList.illegal_format;
    if (hDataObject == null) return TErrorList.invalid_data;
    if (!(hDataObject instanceof MDA)) return TErrorList.invalid_data;
    MDA tgt = (MDA)hDataObject;
    int totlen = dArrayLength;
    int ndim = data.getNumberDimension();
    if (tgt.getNumberDimension() != 0)
    {
      if (ndim != tgt.getNumberDimension()) return TErrorList.dimension_error;
      int dlen = 1;
      for (int i=0; i<ndim; i++)
      {
        dlen = data.getAxis(i).getLength();
        if (i < ndim-1)
        {
          if (dlen != tgt.getAxis(i).getLength()) return TErrorList.dimension_error;
        }
        else
        {
          if (dlen > tgt.getAxis(i).getLength()) dlen = tgt.getAxis(i).getLength();
        }
        totlen *= dlen;
      }
    }
    hDataObject = data;
    dArrayLength = totlen;
    byte[] dbytes = data.toByteArray();
    if (dbytes == null) return TErrorList.code_failure;
    if (isLocked) throw new InputDataLockedException(lockedMessage);
    
    repackageDataForOutput();
    return 0;
  } }
  /**
   * Puts the current canonical data set into the underlying data byte
   *        stream Puts the input data set into the underlying byte stream,
   *        properly swapped and aligned. It replaces the contents of the
   *        original, referenced data array with that of the data object given.
   * @param data
   *          is the data object whose contents should be copied into the
   *          underlying byte stream and referenced data object.
   * @note If this TDataType object is used as input data in persistent links,
   *       then the data object might well be 'locked' meaning that the
   *       underlying data will NOT be changed, as this would constitute a
   *       change in contract.
   * @return 0 upon success or a tine error code
   */
  public int putData(byte[] data)
  {
    return putData(data, data != null ? data.length : 0, 0);
  }
  /**
   * Puts the current canonical data set into the underlying data byte
   *        stream Puts the input data set into the underlying byte stream,
   *        properly swapped and aligned. It replaces the contents of the
   *        original, referenced data array with that of the data object given.
   * @param data
   *          is the data object whose contents should be copied into the
   *          underlying byte stream and referenced data object.
   * @param length
   *          is the number of elements from the give data object to copy
   * @param offset
   *          is the starting offset within the array to copy
   * @note If this TDataType object is used as input data in persistent links,
   *       then the data object might well be 'locked' meaning that the
   *       underlying data will NOT be changed, as this would constitute a
   *       change in contract.
   * @return 0 upon success or a tine error code
   */
  public int putData(byte[] data, int length, int offset)
  {
    return putData(0,data,length,offset,true);
  }
  public int putData(int start, byte[] data, int length, int offset, boolean redim)
  { synchronized (hDataObject) {
    if (data == null || length <= 0 || offset < 0) return TErrorList.argument_list_error;
    if (length > data.length - offset) length = data.length - offset;
    if (length <= 0) return TErrorList.argument_list_error;
    if (isLocked) throw new InputDataLockedException(lockedMessage);
    int maxlen = Math.min(dArrayLength, length);
    if (maxlen > dArrayLength - start) maxlen = dArrayLength - start;
    switch (dFormat)
    {
      case TFormat.CF_TEXT:
        if (hDataObject instanceof char[])
        {
          for (int i = 0; i < maxlen; i++)
            ((char[]) hDataObject)[i+start] = (char) data[i + offset];
          break;
        }
        if (hDataObject instanceof StringBuffer)
        {
          String s = new String(data);
          ((StringBuffer) hDataObject).append(s, offset, dArrayLength);
          break;
        }
        if (hDataObject instanceof byte[])
        {
          for (int i = 0; i < maxlen; i++)
            ((byte[]) hDataObject)[i+start] = data[i + offset];
          break;
        }
        return TErrorList.illegal_format;
      case TFormat.CF_BITFIELD8:
      case TFormat.CF_BYTE:
        System.arraycopy(data, offset, (byte[]) hDataObject, start, maxlen);
        break;
      case TFormat.CF_BITFIELD16:
      case TFormat.CF_UINT16:
      case TFormat.CF_INT16:
        for (int i = 0; i < maxlen; i++)
          ((short[]) hDataObject)[i+start] = (short) data[i + offset];
        break;
      case TFormat.CF_BOOLEAN:
        if (hDataObject instanceof boolean[])
        {
          for (int i = 0; i < maxlen; i++)
            ((boolean[]) hDataObject)[i+start] = (boolean)(data[i + offset] != 0);
          break;
        }
      case TFormat.CF_BITFIELD32:
      case TFormat.CF_UINT32:
      case TFormat.CF_INT32:
        for (int i = 0; i < maxlen; i++)
          ((int[]) hDataObject)[i+start] = (int) data[i + offset];
        break;
      default:
        return TErrorList.illegal_format;
    }
    if (redim) dArrayLength = Math.min(dArrayLength, length);
    repackageDataForOutput();
    return 0;
  } }
  /**
   * Puts the input data value into the underlying byte stream,
   * properly swapped and aligned. It replaces the contents of the
   * original, referenced data array with that of the data object given.
   * @param value
   *          is the single value which should be copied into the
   *          underlying byte stream and referenced data object at the given offset.
   * @param offset
   *          is the position offset within the targeted array
   * @note If this TDataType object is used as input data in persistent links,
   *       then the data object might well be 'locked' meaning that the
   *       underlying data will NOT be changed, as this would constitute a
   *       change in contract.
   * @return 0 upon success or a tine error code
   * 
   * @throws InputDataLockedException if a link using this TDataType as input 
   * is bound to another link 
   * and an attempt to alter the input data set (which is part of the contract
   * cached at the target server) is made.  
   * 
   */
  public int putData(byte value,int offset)
  {
    return putDataValue(Byte.valueOf(value),offset);
  }
  /**
   * Puts the input data value array into the underlying byte stream,
   * properly swapped and aligned. It replaces the contents of the
   * original, referenced data array with that of the data object given.
   * @param values
   *          is the values array which should be copied into the
   *          underlying byte stream and referenced data object at the given offset.
   * @param offset
   *          is the position offset within the targeted array
   * @note If this TDataType object is used as input data in persistent links,
   *       then the data object might well be 'locked' meaning that the
   *       underlying data will NOT be changed, as this would constitute a
   *       change in contract.
   * @return 0 upon success or a tine error code
   * 
   * @throws InputDataLockedException if a link using this TDataType as input 
   * is bound to another link 
   * and an attempt to alter the input data set (which is part of the contract
   * cached at the target server) is made.  
   * 
   */
  public int putData(byte[] values,int offset)
  {
    return putDataSegment(values, offset);
  }
  public int putData(byte[] data, String tag)
  { synchronized (hDataObject) {
    if (data != null && isLocked) throw new InputDataLockedException(lockedMessage);
    if (hDataObject != null)
    {
      if (tag != null && TStructRegistry.getSizeInBytes(tag,dTagKey) < 1) return TErrorList.invalid_structure_tag;
      if (data.length < dArrayLength) dArrayLength = data.length;
      System.arraycopy(data, 0, (byte[]) hDataObject, 0, dArrayLength);
    }
    else if (hStructObject != null && hStructObject instanceof TTaggedStructure)
    {
      TTaggedStructure tts = (TTaggedStructure) hStructObject;
      if (data.length != tts.getSizeInBytes()) return TErrorList.invalid_structure_tag;
      System.arraycopy(data, 0, tts.getByteArray(), 0, data.length);
    }
    else if (hStructObject instanceof TTaggedStructure[])
    {
      TTaggedStructure tts[] = (TTaggedStructure[]) hStructObject;
      int dsiz = tts[0].getSizeInBytes();
      for (int i = 0; i < tts.length; i++)
      {
        System.arraycopy(data, i * dsiz, tts[i].getByteArray(), 0, dsiz);
      }
    }
    else
    {
      return TErrorList.illegal_format;
    }
    repackageDataForOutput();
    return 0;
  } }
  public int replaceData(Object data,int sourceIndex,int targetIndex)
  {
    int len = 0;
    try { len = Array.getLength(data); } catch (Exception any) {};
    if (len == 0) return TErrorList.argument_list_error;
    if (sourceIndex < 0 || sourceIndex >= len) return TErrorList.dimension_error;
    if (data instanceof byte[])
      return putData(((byte[])data)[sourceIndex],targetIndex);
    else if (data instanceof short[])
      return putData(((short[])data)[sourceIndex],targetIndex);
    else if (data instanceof int[])
      return putData(((int[])data)[sourceIndex],targetIndex);
    else if (data instanceof boolean[])
      return putData(((boolean[])data)[sourceIndex],targetIndex);
    else if (data instanceof float[])
      return putData(((float[])data)[sourceIndex],targetIndex);
    else if (data instanceof double[])
      return putData(((double[])data)[sourceIndex],targetIndex);
    else if (data instanceof long[])
      return putData(((long[])data)[sourceIndex],targetIndex);
    else if (data instanceof String[])
      return putData(((String[])data)[sourceIndex],targetIndex);
    else if (data instanceof TCompoundDataObject[])
      return putData(((TCompoundDataObject[])data)[sourceIndex],targetIndex);
    return TErrorList.illegal_format;
  }
  public int replaceData(Object data,int sourceIndex,int targetIndex,int length)
  {
    if (length < 2) return replaceData(data,sourceIndex,targetIndex);
    int len = 0;
    try { len = Array.getLength(data); } catch (Exception any) {};
    if (len > 0)
    {
      if (sourceIndex < 0 || sourceIndex >= len) return TErrorList.dimension_error;  
    }
    if (data instanceof byte[])
    { // CF_STRUCT will land here too
      byte[] a = Arrays.copyOfRange((byte[])data, sourceIndex, sourceIndex+length);
      return putData(a,targetIndex);
    }
    else if (data instanceof short[])
    {
      short[] a = Arrays.copyOfRange((short[])data, sourceIndex, sourceIndex+length);
      return putData(a,targetIndex);
    }
    else if (data instanceof int[])
    {
      int[] a = Arrays.copyOfRange((int[])data, sourceIndex, sourceIndex+length);
      return putData(a,targetIndex);
    }
    else if (data instanceof float[])
    {
      float[] a = Arrays.copyOfRange((float[])data, sourceIndex, sourceIndex+length);
      return putData(a,targetIndex);
    }
    else if (data instanceof double[])
    {
      double[] a = Arrays.copyOfRange((double[])data, sourceIndex, sourceIndex+length);
      return putData(a,targetIndex);
    }
    else if (data instanceof long[])
    {
      long[] a = Arrays.copyOfRange((long[])data, sourceIndex, sourceIndex+length);
      return putData(a,targetIndex);
    }
    else if (data instanceof TCompoundDataObject[])
    {
      TCompoundDataObject[] tdo = (TCompoundDataObject[])data;
      TCompoundDataObject[] a = Arrays.copyOfRange(tdo,sourceIndex,sourceIndex+length);
      return putData(a,length,targetIndex);
    }
    else if (data instanceof TCompoundDataObject)
    {
      TCompoundDataObject tdo = (TCompoundDataObject)data;
      return putData(tdo);
    }
    return TErrorList.illegal_format;
  }
  public int putData(int start,Object data,int length,int offset, boolean redim)
  {
    if (data instanceof byte[])
      return putData(start,(byte[]) data,length,offset,redim);
    else if (data instanceof short[])
      return putData(start,(short[]) data,length,offset,redim);
    else if (data instanceof int[])
      return putData(start,(int[]) data,length,offset,redim);
    else if (data instanceof boolean[])
      return putData(start,(boolean[]) data,length,offset,redim);
    else if (data instanceof float[])
      return putData(start,(float[]) data,length,offset,redim);
    else if (data instanceof double[])
      return putData(start,(double[]) data,length,offset,redim);
    else if (data instanceof long[])
      return putData(start,(long[]) data,length,offset,redim);
    else if (data instanceof String[])
      return putData(start,(String[]) data,length,offset,redim);
    else if (data instanceof KEYVALUE[])
      return putData(start,(KEYVALUE[]) data,length,offset,redim);
    else if (data instanceof char[])
      return putData((char[]) data,length,offset);
    else if (data instanceof TCompoundDataObject[])
    {
      int cc = putData(start,(TCompoundDataObject[]) data,length,offset,redim);
      if (dFormat == TFormat.CF_HISTORY)
      { // reset the 'arraylength' to the length in bytes (that's how it's done!)
        dArrayLength *= (TFormat.getHistoryHeaderSize() + TFormat.getCarriedFormatSize(dTag));
      }
      return cc;
    } 
    return TErrorList.not_implemented;
  }
  public int putData(Object data)
  {
    if (data instanceof byte[])
      return putData((byte[]) data);
    else if (data instanceof short[])
      return putData((short[]) data);
    else if (data instanceof int[])
      return putData((int[]) data);
    else if (data instanceof boolean[])
      return putData((boolean[]) data);
    else if (data instanceof float[])
      return putData((float[]) data);
    else if (data instanceof double[])
      return putData((double[]) data);
    else if (data instanceof long[])
      return putData((long[]) data);
    else if (data instanceof String[])
      return putData((String[]) data);
    else if (data instanceof KEYVALUE[])
      return putData((KEYVALUE[]) data);
    else if (data instanceof char[])
      return putData((char[]) data);
    else if (data instanceof TCompoundDataObject[])
    {
      int cc = putData((TCompoundDataObject[]) data);
      if (dFormat == TFormat.CF_HISTORY)
      { // reset the 'arraylength' to the length in bytes (that's how it's done!)
        dArrayLength *= (TFormat.getHistoryHeaderSize() + TFormat.getCarriedFormatSize(dTag));
      }
      return cc;
    }
    if (dFormat != getObjectFormat(data)) return TErrorList.illegal_format;
    int len = getObjectSize(data);
    if (dFormat != TFormat.CF_TEXT && dFormat != TFormat.CF_XML && len < dArrayLength) dArrayLength = len;
    setObjectHandle(data);
    repackageDataForOutput();
    return 0;
  }
  /**
   * Puts the current canonical data set into the underlying data byte
   *        stream Puts the input data set into the underlying byte stream,
   *        properly swapped and aligned. It replaces the contents of the
   *        original, referenced data array with that of the data object given.
   * @param data
   *          is IMAGE type whose contents are to copied into the underlying
   *          byte stream and referenced data object.
   * @note If this TDataType object is used as input data in persistent links,
   *       then the data object might well be 'locked' meaning that the
   *       underlying data will NOT be changed, as this would constitute a
   *       change in contract.
   * @return 0 upon success or a tine error code
   */
  public int putData(IMAGE data)
  {
    if (data == null) return TErrorList.argument_list_error;
    if (dFormat != TFormat.CF_IMAGE) return TErrorList.illegal_format;
    if (isLocked) throw new InputDataLockedException(lockedMessage);
    setObjectHandle(data);
    int cc = 0;
    int dsize = data.getHeaderSize() + data.getFrameHeader().appendedFrameSize;
    if (dsize < 0 || dsize > data.getHeaderSize() + data.getImageFrameSize())
    {
      cc = TErrorList.dimension_error;
    }
    else
    {
      dArrayLength = dsize;
      dSizeInBytes = dsize; // data.getSizeInBytes();
      repackageDataForOutput();
    }
    return cc;
  }
  public int putData(IMAGE[] data)
  {
    return putData(data,data.length,0);
  }
  public int putData(IMAGE[] data, int length, int offset)
  { synchronized (hDataObject) {
    if (data == null || length <= 0 || offset < 0) return TErrorList.argument_list_error;
    if (length > data.length - offset) length = data.length - offset;
    int len = 0;
    try { len = Integer.parseInt(dTag); } catch (Exception ignore) {};
    if (len == 0) len = dArrayLength;
    if (len > length) len = length;
    if (length <= 0) return TErrorList.argument_list_error;
    if (dFormat != TFormat.CF_AIMAGE) return TErrorList.illegal_format;
    if (isLocked) throw new InputDataLockedException(lockedMessage);
    dSizeInBytes = 0;
    int totalHdrBytes = 0, totalFrameBytes = 0;
    for (int i=0; i<len; i++)
    {
      if (((IMAGE[])data)[i+offset] == null) ((IMAGE[])data)[i+offset] = new IMAGE(100);
      totalHdrBytes += IMAGE.HEADER_SIZE + 8;
      totalFrameBytes += ((IMAGE[])data)[i+offset].getFrameHeader().appendedFrameSize;
    }
    dSizeInBytes = totalHdrBytes + totalFrameBytes;
    int xoff = totalHdrBytes; 
    for (int i=0; i<len; i++)
    {
      ((IMAGE[])data)[i+offset].setExtendedOffset(xoff);
      ((IMAGE[])data)[i+offset].toAdjustableHeaderByteArray();
      xoff += ((IMAGE[])data)[i+offset].getExtendedSize();
      ((IMAGE[]) hDataObject)[i] = data[i+offset];
    }    
    dArrayLength = dSizeInBytes;
    dTag = new String(""+len);
    repackageDataForOutput();
    return 0;
  } }
  /**
   * Puts the current canonical data set into the underlying data byte
   *        stream Puts the input data set into the underlying byte stream,
   *        properly swapped and aligned. It replaces the contents of the
   *        original, referenced data array with that of the data object given.
   * @param data
   *          is data primitive whose value is copied into the underlying byte
   *          stream and referenced data object.
   * @note If this TDataType object is used as input data in persistent links,
   *       then the data object might well be 'locked' meaning that the
   *       underlying data will NOT be changed, as this would constitute a
   *       change in contract.
   * @return 0 upon success or a tine error code
   */
  public int putData(TCompoundDataObject data)
  {
    if (data == null) return TErrorList.argument_list_error;
    if (hDataObject instanceof TCompoundDataObject[])
    {
      TCompoundDataObject[] d = new TCompoundDataObject[]{data};
      return putData(d);
    }
    if (!(hDataObject instanceof TCompoundDataObject)) return TErrorList.illegal_format;
    if (isLocked) throw new InputDataLockedException(lockedMessage);
    if (dFormat != getObjectFormat(data)) return TErrorList.illegal_format;
    switch (dFormat)
    {
      case TFormat.CF_GSPECTRUM:
      case TFormat.CF_SPECTRUM:
      case TFormat.CF_IMAGE:
        setObjectHandle(data);
        dArrayLength = data.getSizeInBytes() - TFormat.getHeaderSize(dFormat);
        dArrayLength /= TFormat.formatSizeOf(dFormat);
        break;
      default:
        return TErrorList.illegal_format;
    }
    repackageDataForOutput();
    return 0;
  }
  /**
   * Puts the current canonical data set into the underlying data byte
   *        stream Puts the input data set into the underlying byte stream,
   *        properly swapped and aligned. It replaces the contents of the
   *        original, referenced data array with that of the data object given.
   * @param data
   *          is the data object whose contents should be copied into the
   *          underlying byte stream and referenced data object.
   * @note If this TDataType object is used as input data in persistent links,
   *       then the data object might well be 'locked' meaning that the
   *       underlying data will NOT be changed, as this would constitute a
   *       change in contract.
   * @return 0 upon success or a tine error code
   */
  public int putData(TCompoundDataObject[] data)
  {
    return putData(data, data != null ? data.length : 0, 0);
  }
  /**
   * Puts the current canonical data set into the underlying data byte
   *        stream Puts the input data set into the underlying byte stream,
   *        properly swapped and aligned. It replaces the contents of the
   *        original, referenced data array with that of the data object given.
   * @param data
   *          is the data object whose contents should be copied into the
   *          underlying byte stream and referenced data object.
   * @param length
   *          is the number of elements from the give data object to copy
   * @param offset
   *          is the starting offset within the array to copy
   * @note If this TDataType object is used as input data in persistent links,
   *       then the data object might well be 'locked' meaning that the
   *       underlying data will NOT be changed, as this would constitute a
   *       change in contract.
   * @return 0 upon success or a tine error code
   */
  public int putData(TCompoundDataObject[] data, int length, int offset)
  {
  	return putData(0,data,length,offset,true);
  }
  public int putData(int start, TCompoundDataObject[] data, int length, int offset, boolean redim)
  { synchronized (hDataObject) {
    if (data == null || length < 0 || offset < 0) return TErrorList.argument_list_error;
    if (length == 0) 
    {
      setArrayLength(0,0);
      repackageDataForOutput();
      return 0;
    }
    if (length > data.length - offset) length = data.length - offset;
    int maxlen = Math.min(dArrayLength, length);
    if (maxlen > dArrayLength - start) maxlen = dArrayLength - start;
    if (length <= 0) return TErrorList.argument_list_error;
    if (isLocked) throw new InputDataLockedException(lockedMessage);
    boolean formatsEqual = dFormat == getObjectFormat(data);
    if ((data instanceof NAME[] && hDataObject instanceof NAME[])
        || (data instanceof NAMEINT[] && hDataObject instanceof NAMEINT[]))
    {
      int objlen = ((TCompoundDataObject[]) hDataObject).length;
      if (objlen < maxlen)
      { // how did this happen?  Is this Josef's problem ?
        maxlen = objlen;
      }
      if (formatsEqual)
      {
        for (int i = 0; i < maxlen; i++)
          ((TCompoundDataObject[]) hDataObject)[i+start] = data[i + offset];
      }
      else
      {
        try
        {
          @SuppressWarnings("rawtypes")
          Class[] params = new Class[1];
          params[0] = String.class;
          Object[] objs = new Object[1];
          for (int i = 0; i < maxlen; i++)
          {
            objs[0] = (data[i + offset].toString());
            ((TCompoundDataObject[]) hDataObject)[i+start] = (TCompoundDataObject) hDataObject.getClass()
                .getComponentType().getConstructor(params).newInstance(objs);
          }
        }
        catch (Exception e)
        {
          e.printStackTrace();
          MsgLog.log("TDataType", "cannot convert " + TFormat.toString(getObjectFormat(data)) + " to "
              + TFormat.toString(dFormat), TErrorList.illegal_format, e, 0);
        }
      }
    }
    else if (data instanceof NAME[] && hDataObject instanceof char[])
    { // TEXT object being passed a NAMExx
      hDataObject = data[offset].toString().toCharArray();
      maxlen *= data[offset].getSizeInBytes();
    }
    else if (data instanceof NAME[] && hDataObject instanceof String[])
    { // TEXT object being passed a NAMExx
      int objlen = ((String[]) hDataObject).length;
      if (objlen < maxlen) maxlen = objlen;
      for (int i = 0; i < maxlen; i++)
      {
        ((String[]) hDataObject)[i+start] = data[i + offset].toString();
      }
    }
    else if (!formatsEqual && hDataObject instanceof USTRING[])
    {
    	if (data instanceof NAME[])
    	{
        int objlen = ((USTRING[]) hDataObject).length;
        if (objlen < maxlen) maxlen = objlen;
        for (int i = 0; i < maxlen; i++)
        {
          ((USTRING[]) hDataObject)[i+start].setString(data[i + offset].toString());
        }    		
    	}
    	else if (data instanceof NAME64DBLDBL[])
    	{
        int objlen = ((USTRING[]) hDataObject).length;
        if (objlen < maxlen) maxlen = objlen;
        NAME64DBLDBL[] d = (NAME64DBLDBL[])data; 
        int ts = (int)(System.currentTimeMillis()/1000);
        for (int i = 0; i < maxlen; i++)
        {
          ((USTRING[]) hDataObject)[i+start].setValues((int)d[i+offset].d2val, (float)d[i+offset].d1val, 0, ts, d[i+offset].getName());;
        }    		
    	}
    	else if (data instanceof NAME32DBLDBL[])
    	{
        int objlen = ((USTRING[]) hDataObject).length;
        if (objlen < maxlen) maxlen = objlen;
        NAME32DBLDBL[] d = (NAME32DBLDBL[])data; 
        int ts = (int)(System.currentTimeMillis()/1000);
        for (int i = 0; i < maxlen; i++)
        {
          ((USTRING[]) hDataObject)[i+start].setValues((int)d[i+offset].d2val, (float)d[i+offset].d1val, 0, ts, d[i+offset].getName());;
        }    		
    	}
    	else
    	{
    		return TErrorList.illegal_format;
    	}
    }
    else if (!formatsEqual && hDataObject instanceof NAME64DBLDBL[])
    {
    	if (data instanceof NAME[])
    	{
        int objlen = ((NAME64DBLDBL[]) hDataObject).length;
        if (objlen < maxlen) maxlen = objlen;
        for (int i = 0; i < maxlen; i++)
        {
          ((NAME64DBLDBL[]) hDataObject)[i+start].setName(data[i+offset].toString());
        }    		
    	}
    	else if (data instanceof USTRING[])
    	{
        int objlen = ((USTRING[]) hDataObject).length;
        if (objlen < maxlen) maxlen = objlen;
        USTRING[] d = (USTRING[])data; 
        for (int i = 0; i < maxlen; i++)
        {
          ((NAME64DBLDBL[]) hDataObject)[i+start].setValues(d[i+offset].getString(),d[i+offset].f1val,d[i+offset].ival);
        }    		
    	}
    	else if (data instanceof NAME32DBLDBL[])
    	{
        int objlen = ((USTRING[]) hDataObject).length;
        if (objlen < maxlen) maxlen = objlen;
        NAME32DBLDBL[] d = (NAME32DBLDBL[])data; 
        for (int i = 0; i < maxlen; i++)
        {
          ((NAME64DBLDBL[]) hDataObject)[i+start].setValues(d[i+offset].getName(),d[i+offset].d1val,d[i+offset].d2val);
        }    		
    	}
    	else
    	{
    		return TErrorList.illegal_format;
    	}
    }
    else
    {
      if (!formatsEqual) return TErrorList.illegal_format;
      if (offset == 0 && start == 0)
      {
        setObjectHandle(data);
      }
      else
      {
        for (int i = 0; i < maxlen; i++)
          ((TCompoundDataObject[]) hDataObject)[i+start] = data[i + offset];        
      }
    }
    if (redim) dArrayLength = Math.min(dArrayLength, length);
    repackageDataForOutput();
    return 0;
  } }
  /**
   * Puts the current canonical data set into the underlying data byte
   *        stream Puts the input data value into the underlying byte stream,
   *        properly swapped and aligned. It replaces the contents of the
   *        original, referenced data array with that of the data object given.
   * @param value
   *          is the single value which should be copied into the
   *          underlying byte stream and referenced data object at the given offset.
   * @param offset
   *          is the position offset within the targeted array
   * @note If this TDataType object is used as input data in persistent links,
   *       then the data object might well be 'locked' meaning that the
   *       underlying data will NOT be changed, as this would constitute a
   *       change in contract.
   * @return 0 upon success or a tine error code
   * 
   * @throws InputDataLockedException if a link using this TDataType as input 
   * is bound to another link 
   * and an attempt to alter the input data set (which is part of the contract
   * cached at the target server) is made.  
   * 
   */
  public int putData(TCompoundDataObject value, int offset)
  {
    if (offset < 0) return TErrorList.argument_list_error;
    if (isLocked) throw new InputDataLockedException(lockedMessage);
    if (dArrayLength < offset) return TErrorList.dimension_error;
    TCompoundDataObject[] d = new TCompoundDataObject[1];
    d[0] = value;
    short sFormat = getObjectFormat(d);
    boolean formatsEqual = dFormat == sFormat;
    if (formatsEqual)
    {
      ((TCompoundDataObject[]) hDataObject)[offset] = value;
    }
    else if ((TFormat.isName(dFormat) && TFormat.isName(sFormat))
        || (TFormat.isNameInt(dFormat) && TFormat.isNameInt(sFormat)))
    {
      try
      {
        @SuppressWarnings("rawtypes")
        Class[] params = new Class[1];
        params[0] = String.class;
        Object[] objs = new Object[1];
        objs[0] = (value.toString());
        ((TCompoundDataObject[]) hDataObject)[offset] = (TCompoundDataObject) hDataObject.getClass()
            .getComponentType().getConstructor(params).newInstance(objs);
      }
      catch (Exception e)
      {
        e.printStackTrace();
        MsgLog.log("TDataType", "cannot convert " + TFormat.toString(getObjectFormat(d)) + " to "
            + TFormat.toString(dFormat), TErrorList.illegal_format, e, 0);
        return TErrorList.illegal_format;
      }      
    }
    else if (TFormat.isName(dFormat) && hDataObject instanceof USTRING[])
    {
      for (int i = 0; i < dArrayLength; i++)
      {
        ((USTRING[]) hDataObject)[offset].setString(value.toString());
      }
    }
    else
    {
      return TErrorList.illegal_format;
    }
    repackageDataForOutput();
    return 0;
  }
  /**
   * Gets the canonical data set from the underlying data byte stream
   *        Gets the fully formed (properly assembled and byte-swapped)
   *        canonical data set from the underlying byte stream. This method is
   *        used internally, but can be used to get the underlying data into the
   *        specifically referenced data, for instance when the TDataType object
   *        was instantiated with 'unbound' data.
   * @param data
   *          is the data object into which the underlying data should be
   *          prepared and copied.
   * @return 0 on success or a tine error code
   */
  public int getData(double[] data)
  {
    return getData(data, data != null ? data.length : 0, 0);
  }
  /**
   * Gets the canonical data set from the underlying data byte stream
   *        Gets the fully formed (properly assembled and byte-swapped)
   *        canonical data set from the underlying byte stream. This method is
   *        used internally, but can be used to get the underlying data into the
   *        specifically referenced data, for instance when the TDataType object
   *        was instantiated with 'unbound' data.
   * @param data
   *          is the data object into which the underlying data should be
   *          prepared and copied.
   * @param length
   *          is the number of elements to get
   * @param offset
   *          is the beginning position of the target data array into which the
   *          data should be copied
   * @return 0 on success or a tine error code
   */
  public int getData(double[] data, int length, int offset)
  { synchronized (hDataObject) {
    if (data == null || length <= 0 || offset < 0) return TErrorList.argument_list_error;
    if (length > data.length - offset) length = data.length - offset;
    if (length <= 0) return TErrorList.argument_list_error;
    if (offset >= data.length) return TErrorList.dimension_error;
    int len = dArrayLength; // length of the buffered array
    if (len > length) len = length; // can't get more than is there !
    if (buffersBusy) if (waitForBuffers() != 0) return TErrorList.data_stale;
    if (isDataObjectInSync && hDataObject != null)
    {
      if (data == hDataObject) return 0;
      Arrays.fill(data, offset, offset+length, 0);
      try
      {
        for (int i=0; i < len; i++) data[i+offset] = getDoubleAtIndex(i);
      }
      catch (TineRuntimeErrorException e)
      {
        return e.getErrorCode();
      }
      dCompletionLength = len;
      return 0;
    }      
    try
    {
      Arrays.fill(data, offset, offset+length, 0);
      int i, n, m, fmtsiz = TFormat.formatSizeOf(dFormat);
      ByteArrayInputStream bs;
      DataInputStream ds;
      for (i=0, n=0, m=0; n<numblks && n<dBuffers.length; n++, m=i)
      {
        if (dBuffers[n] == null) break;
        byte[] b = dBuffers[n].toByteArray();
        Swap.Bytes(b, fmtsiz);
        bs = new ByteArrayInputStream(b);
        ds = new DataInputStream(bs);
        switch (dFormat)
        {
          case TFormat.CF_DOUBLE:
            for (; i < len && i < m + b.length / fmtsiz; i++)
              data[i + offset] = ds.readDouble();
            break;
          case TFormat.CF_FLOAT:
            for (; i < len && i < m + b.length / fmtsiz; i++)
              data[i + offset] = ds.readFloat();
            break;
          case TFormat.CF_INT64:
            for (; i < len && i < m + b.length / fmtsiz; i++)
              data[i + offset] = ds.readLong();
            break;
          case TFormat.CF_BOOLEAN:
          case TFormat.CF_INT32:
            for (; i < len && i < m + b.length / fmtsiz; i++)
              data[i + offset] = ds.readInt();
            break;
          case TFormat.CF_INT16:
            for (; i < len && i < m + b.length / fmtsiz; i++)
              data[i + offset] = ds.readShort();
            break;
          default:
            ds.close();
            bs.close();
            return TErrorList.illegal_format;
        }
        ds.close();
        bs.close();
      }
      dCompletionLength = i;
      isDataObjectInSync = true;
    }
    catch (IOException e)
    {
      e.printStackTrace();
      MsgLog.log("TDataType.getData", e.getMessage(), TErrorList.code_failure, e, 0);
    }
    return 0;
  } }
  /**
   * Puts the current canonical data set into the underlying data byte
   *        stream Puts the input data set into the underlying byte stream,
   *        properly swapped and aligned. It replaces the contents of the
   *        original, referenced data array with that of the data object given.
   * @param data
   *          is data primitive whose value is copied into the underlying byte
   *          stream and referenced data object.
   * @note If this TDataType object is used as input data in persistent links,
   *       then the data object might well be 'locked' meaning that the
   *       underlying data will NOT be changed, as this would constitute a
   *       change in contract.
   * @return 0 upon success or a tine error code
   */
  public int putData(double data)
  {
    return putData(new double[]{data});
  }
  /**
   * Puts the current canonical data set into the underlying data byte
   *        stream Puts the input data set into the underlying byte stream,
   *        properly swapped and aligned. It replaces the contents of the
   *        original, referenced data array with that of the data object given.
   * @param data
   *          is the data object whose contents should be copied into the
   *          underlying byte stream and referenced data object.
   * @note If this TDataType object is used as input data in persistent links,
   *       then the data object might well be 'locked' meaning that the
   *       underlying data will NOT be changed, as this would constitute a
   *       change in contract.
   * @return 0 upon success or a tine error code
   */
  public int putData(double[] data)
  {
    return putData(data, data != null ? data.length : 0, 0);
  }
  /**
   * Puts the current canonical data set into the underlying data byte
   *        stream Puts the input data set into the underlying byte stream,
   *        properly swapped and aligned. It replaces the contents of the
   *        original, referenced data array with that of the data object given.
   * @param data
   *          is the data object whose contents should be copied into the
   *          underlying byte stream and referenced data object.
   * @param length
   *          is the number of elements from the give data object to copy
   * @param offset
   *          is the starting offset within the array to copy
   * @note If this TDataType object is used as input data in persistent links,
   *       then the data object might well be 'locked' meaning that the
   *       underlying data will NOT be changed, as this would constitute a
   *       change in contract.
   * @return 0 upon success or a tine error code
   */
  public int putData(double[] data, int length, int offset)
  {
    return putData(0,data,length,offset,true);
  }
  public int putData(int start, double[] data, int length, int offset, boolean redim)
  { synchronized (hDataObject) {
    if (data == null || length <= 0 || offset < 0) return TErrorList.argument_list_error;
    if (length > data.length - offset) length = data.length - offset;
    if (length <= 0) return TErrorList.argument_list_error;
    if (isLocked) throw new InputDataLockedException(lockedMessage);
    int maxlen = Math.min(dArrayLength, length);
    if (maxlen > dArrayLength - start) maxlen = dArrayLength - start;
    switch (dFormat)
    {
      case TFormat.CF_DOUBLE:
        for (int i = 0; i < maxlen; i++)
          ((double[]) hDataObject)[i+start] = (double) data[i + offset];
        break;
      case TFormat.CF_FLOAT:
        for (int i = 0; i < maxlen; i++)
          ((float[]) hDataObject)[i+start] = (float) data[i + offset];
        break;
      case TFormat.CF_INT16:
        for (int i = 0; i < maxlen; i++)
          ((short[]) hDataObject)[i+start] = (short) data[i + offset];
        break;
      case TFormat.CF_INT64:
        for (int i = 0; i < maxlen; i++)
          ((long[]) hDataObject)[i+start] = (long) data[i + offset];
        break;
      case TFormat.CF_BOOLEAN:
        for (int i = 0; i < maxlen; i++)
          ((boolean[]) hDataObject)[i+start] = (boolean)(data[i + offset] != 0);
        break;
      case TFormat.CF_INT32:
        for (int i = 0; i < maxlen; i++)
          ((int[]) hDataObject)[i+start] = (int) data[i + offset];
        break;
      default:
        return TErrorList.illegal_format;
    }
    if (redim) dArrayLength = Math.min(dArrayLength, length);
    repackageDataForOutput();
    return 0;
  } }
  /**
   * Puts the input data value into the underlying byte stream,
   * properly swapped and aligned. It replaces the contents of the
   * original, referenced data array with that of the data object given.
   * @param value
   *          is the single value which should be copied into the
   *          underlying byte stream and referenced data object at the given offset.
   * @param offset
   *          is the position offset within the targeted array
   * @note If this TDataType object is used as input data in persistent links,
   *       then the data object might well be 'locked' meaning that the
   *       underlying data will NOT be changed, as this would constitute a
   *       change in contract.
   * @return 0 upon success or a tine error code
   * 
   * @throws InputDataLockedException if a link using this TDataType as input 
   * is bound to another link 
   * and an attempt to alter the input data set (which is part of the contract
   * cached at the target server) is made.  
   * 
   */
  public int putData(double value, int offset)
  {
    return putDataValue(Double.valueOf(value),offset);
  }
  /**
   * Puts the input data value array into the underlying byte stream,
   * properly swapped and aligned. It replaces the contents of the
   * original, referenced data array with that of the data object given.
   * @param values
   *          is the values array which should be copied into the
   *          underlying byte stream and referenced data object at the given offset.
   * @param offset
   *          is the position offset within the targeted array
   * @note If this TDataType object is used as input data in persistent links,
   *       then the data object might well be 'locked' meaning that the
   *       underlying data will NOT be changed, as this would constitute a
   *       change in contract.
   * @return 0 upon success or a tine error code
   * 
   * @throws InputDataLockedException if a link using this TDataType as input 
   * is bound to another link 
   * and an attempt to alter the input data set (which is part of the contract
   * cached at the target server) is made.  
   * 
   */
  public int putData(double[] values, int offset)
  {
    return putDataSegment(values,offset);
  }
  private double getDoubleAtIndex(int index)
  {
    return getDoubleAtIndex(hDataObject,index);
  }
  private double getDoubleAtIndex(Object d,int index)
  {
    try
    {
      if (d instanceof double[]) return ((double[])d)[index];
      if (d instanceof float[]) return (double)((float[])d)[index];
      if (d instanceof short[]) return (double)((short[])d)[index];
      if (d instanceof int[]) return (double)((int[])d)[index];
      if (d instanceof long[]) return (double)((long[])d)[index];
      if (d instanceof byte[]) return (double)((byte[])d)[index];
    }
    catch (Exception e) 
    {
      MsgLog.log("getDoubleAtIndex",e.getMessage(),TErrorList.code_failure,e,0);      
    }
    throw new TineRuntimeErrorException(TErrorList.illegal_format);
  }
  private float getFloatAtIndex(int index)
  {
    return getFloatAtIndex(hDataObject,index);
  }
  private float getFloatAtIndex(Object d,int index)
  {
    try
    {
      if (d instanceof double[]) return (float)((double[])d)[index];
      if (d instanceof float[]) return ((float[])d)[index];
      if (d instanceof short[]) return (float)((short[])d)[index];
      if (d instanceof int[]) return (float)((int[])d)[index];
      if (d instanceof long[]) return (float)((long[])d)[index];
      if (d instanceof byte[]) return (float)((byte[])d)[index];
    }
    catch (Exception e) 
    {
      MsgLog.log("getFloatAtIndex",e.getMessage(),TErrorList.code_failure,e,0);      
    }
    throw new TineRuntimeErrorException(TErrorList.illegal_format);
  }
  private int getIntAtIndex(int index)
  {
    return getIntAtIndex(hDataObject,index);
  }
  private int getIntAtIndex(Object d,int index)
  {
    try
    {
      if (d instanceof double[]) return (int)((double[])d)[index];
      if (d instanceof float[]) return (int)((float[])d)[index];
      if (d instanceof short[]) return (int)((short[])d)[index];
      if (d instanceof int[]) return ((int[])d)[index];
      if (d instanceof long[]) return (int)((long[])d)[index];
      if (d instanceof byte[]) return (int)((byte[])d)[index];
      if (d instanceof boolean[]) return (int)(((boolean[])d)[index] ? 1 : 0);
    }
    catch (Exception e) 
    {
      MsgLog.log("getIntAtIndex",e.getMessage(),TErrorList.code_failure,e,0);      
    }
    throw new TineRuntimeErrorException(TErrorList.illegal_format);
  }
  private short getShortAtIndex(int index)
  {
    return getShortAtIndex(hDataObject,index);
  }
  private short getShortAtIndex(Object d,int index)
  {
    try
    {
      if (d instanceof double[]) return (short)((double[])d)[index];
      if (d instanceof float[]) return (short)((float[])d)[index];
      if (d instanceof short[]) return ((short[])d)[index];
      if (d instanceof int[]) return (short)((int[])d)[index];
      if (d instanceof long[]) return (short)((long[])d)[index];
      if (d instanceof byte[]) return (short)((byte[])d)[index];
    }
    catch (Exception e) 
    {
      MsgLog.log("getShortAtIndex",e.getMessage(),TErrorList.code_failure,e,0);      
    }
    throw new TineRuntimeErrorException(TErrorList.illegal_format);
  }
  private long getLongAtIndex(int index)
  {
    return getLongAtIndex(hDataObject,index);
  }
  private long getLongAtIndex(Object d,int index)
  {
    try
    {
      if (d instanceof double[]) return (long)((double[])d)[index];
      if (d instanceof float[]) return (long)((float[])d)[index];
      if (d instanceof short[]) return (long)((short[])d)[index];
      if (d instanceof int[]) return (long)((int[])d)[index];
      if (d instanceof long[]) return ((long[])d)[index];
      if (d instanceof byte[]) return (long)((byte[])d)[index];
    }
    catch (Exception e) 
    {
      MsgLog.log("getLongAtIndex",e.getMessage(),TErrorList.code_failure,e,0);      
    }
    throw new TineRuntimeErrorException(TErrorList.illegal_format);
  }
  private int putDataSegment(Object values, int offset)
  {
    if (values == null || offset < 0) return TErrorList.argument_list_error;
    if (isLocked) throw new InputDataLockedException(lockedMessage);
    if (dArrayLength < offset) return TErrorList.dimension_error;
    if (!values.getClass().isArray()) return TErrorList.not_supported;
    int vlen = Array.getLength(values);
    if (vlen > dArrayLength - offset) vlen = dArrayLength - offset;
    switch (dFormat)
    {
      case TFormat.CF_DOUBLE:
        for (int i=0; i<vlen; i++)
          ((double[]) hDataObject)[offset+i] = getDoubleAtIndex(values,i);
        break;
      case TFormat.CF_FLOAT:
        for (int i=0; i<vlen; i++)
          ((float[]) hDataObject)[offset+i] = (float)getDoubleAtIndex(values,i);
        break;
      case TFormat.CF_UINT16:
      case TFormat.CF_INT16:
        for (int i=0; i<vlen; i++)
          ((short[]) hDataObject)[offset+i] = (short)getIntAtIndex(values,i);
        break;
      case TFormat.CF_INT64:
        for (int i=0; i<vlen; i++)
          ((long[]) hDataObject)[offset+i] = (long)getIntAtIndex(values,i);
        break;
      case TFormat.CF_BOOLEAN:
        if (hDataObject instanceof boolean[])
        {
          for (int i=0; i<vlen; i++)
            ((boolean[]) hDataObject)[offset+i] = (boolean)(getIntAtIndex(values,i) != 0);
          break;
        }
      case TFormat.CF_UINT32:
      case TFormat.CF_INT32:
        for (int i=0; i<vlen; i++)
          ((int[]) hDataObject)[offset+i] = (int)getIntAtIndex(values,i);
        break;
      case TFormat.CF_STRUCT:
      case TFormat.CF_BYTE:
        if (values instanceof byte[])
        {
          System.arraycopy(values, 0, hDataObject, offset, vlen);      
          break;
        }
      default:
        return TErrorList.illegal_format;
    }
    repackageDataForOutput();
    return 0;
  }
  private double getDoubleValue(Object d)
  {
    try
    {
      if (d instanceof Double) return ((Double)d).doubleValue();
      if (d instanceof Float) return (((Float)d)).doubleValue();
      if (d instanceof Short) return (((Short)d)).doubleValue();
      if (d instanceof Integer) return (((Integer)d)).doubleValue();
      if (d instanceof Long) return (((Long)d)).doubleValue();
      if (d instanceof Byte) return (((Byte)d)).doubleValue();
    }
    catch (Exception e) { }
    MsgLog.log("getDoubleValue", "could not get primitive value",TErrorList.code_failure,null,0);
    return -1;
  }
  private double getIntValue(Object d)
  {
    try
    {
      if (d instanceof Double) return ((Double)d).intValue();
      if (d instanceof Float) return (((Float)d)).intValue();
      if (d instanceof Short) return (((Short)d)).intValue();
      if (d instanceof Integer) return (((Integer)d)).intValue();
      if (d instanceof Long) return (((Long)d)).intValue();
      if (d instanceof Byte) return (((Byte)d)).intValue();
    }
    catch (Exception e) { }
    MsgLog.log("getIntValue", "could not get primitive value",TErrorList.code_failure,null,0);
    return -1;
  }
  private int putDataValue(Object d, int offset)
  {
    if (offset < 0) return TErrorList.argument_list_error;
    if (isLocked) throw new InputDataLockedException(lockedMessage);
    if (dArrayLength < offset) return TErrorList.dimension_error;
    switch (dFormat)
    {
      case TFormat.CF_DOUBLE:
        ((double[]) hDataObject)[offset] = getDoubleValue(d);
        break;
      case TFormat.CF_FLOAT:
          ((float[]) hDataObject)[offset] = (float)getDoubleValue(d);
        break;
      case TFormat.CF_INT16:
          ((short[]) hDataObject)[offset] = (short)getIntValue(d);
        break;
      case TFormat.CF_INT64:
          ((long[]) hDataObject)[offset] = (long)getIntValue(d);
        break;
      case TFormat.CF_BOOLEAN:
        ((boolean[]) hDataObject)[offset] = (boolean)(getIntValue(d) != 0);
        break;
      case TFormat.CF_INT32:
          ((int[]) hDataObject)[offset] = (int)getIntValue(d);
        break;
      default:
        return TErrorList.illegal_format;
    }
    repackageDataForOutput();
    return 0;  
  }
  /**
   * Gets the canonical data set from the underlying data byte stream
   *        Gets the fully formed (properly assembled and byte-swapped)
   *        canonical data set from the underlying byte stream. This method is
   *        used internally, but can be used to get the underlying data into the
   *        specifically referenced data, for instance when the TDataType object
   *        was instantiated with 'unbound' data.
   * @param data
   *          is the data object into which the underlying data should be
   *          prepared and copied.
   * @return 0 on success or a tine error code
   */
  public int getData(long[] data)
  {
    return getData(data, data != null ? data.length : 0, 0);
  }
  /**
   * Gets the canonical data set from the underlying data byte stream
   *        Gets the fully formed (properly assembled and byte-swapped)
   *        canonical data set from the underlying byte stream. This method is
   *        used internally, but can be used to get the underlying data into the
   *        specifically referenced data, for instance when the TDataType object
   *        was instantiated with 'unbound' data.
   * @param data
   *          is the data object into which the underlying data should be
   *          prepared and copied.
   * @param length
   *          is the number of elements to get
   * @param offset
   *          is the beginning position of the target data array into which the
   *          data should be copied
   * @return 0 on success or a tine error code
   */
  public int getData(long[] data, int length, int offset)
  { synchronized (hDataObject) {
    boolean isUnsigned = false;
    if (data == null || length <= 0 || offset < 0) return TErrorList.argument_list_error;
    if (length > data.length - offset) length = data.length - offset;
    if (length <= 0) return TErrorList.argument_list_error;
    if (offset >= data.length) return TErrorList.dimension_error;
    int len = dArrayLength; // length of the buffered array
    if (len > length) len = length; // can't get more than is there !
    if (buffersBusy) if (waitForBuffers() != 0) return TErrorList.data_stale;
    if (isDataObjectInSync && hDataObject != null)
    {
      if (data == hDataObject) return 0;
      Arrays.fill(data, offset, offset+length, 0);
      try
      {
        for (int i=0; i < len; i++) data[i+offset] = getLongAtIndex(i);
      }
      catch (TineRuntimeErrorException e)
      {
        return e.getErrorCode();
      }
      dCompletionLength = len;
      return 0;
    }      
    try
    {
      Arrays.fill(data, offset, offset+length, 0);
      int i, n, m, fmtsiz = TFormat.formatSizeOf(dFormat);
      ByteArrayInputStream bs;
      DataInputStream ds;
      for (i=0, n=0, m=0; n<numblks && n<dBuffers.length; n++, m=i)
      {
        if (dBuffers[n] == null) break;
        byte[] b = dBuffers[n].toByteArray();
        Swap.Bytes(b, fmtsiz);
        bs = new ByteArrayInputStream(b);
        ds = new DataInputStream(bs);
        switch (dFormat)
        {
          case TFormat.CF_DOUBLE:
            for (; i < len && i < m + b.length / fmtsiz; i++)
              data[i + offset] = (long) ds.readDouble();
            break;
          case TFormat.CF_FLOAT:
            for (; i < len && i < m + b.length / fmtsiz; i++)
              data[i + offset] = (long) ds.readFloat();
            break;
          case TFormat.CF_BITFIELD64:
            makeBitfield();
            if (dBitField == null) return TErrorList.not_initialized;
            long lval;
            for (; i < len && i < m + b.length / fmtsiz; i++)
            {
              lval = ds.readLong();
              dBitField[i+offset].setValue(lval);
              data[i + offset] = dBitField[i+offset].getField(dField, lval);
              //data[i + offset] = dBitField.getField(dField, ds.readLong());
            }
            break;
          case TFormat.CF_UINT64:
          case TFormat.CF_INT64:
            for (; i < len && i < m + b.length / fmtsiz; i++)
              data[i + offset] = ds.readLong();
            break;
          case TFormat.CF_BITFIELD32:
            makeBitfield();
            if (dBitField == null) return TErrorList.not_initialized;
            int ival;
            for (; i < len && i < m + b.length / fmtsiz; i++)
            {
              ival = ds.readInt();
              dBitField[i+offset].setValue(ival);
              data[i + offset] = dBitField[i+offset].getField(dField, ival);
             // data[i + offset] = dBitField.getField(dField, ds.readInt());
            }
            break;
          case TFormat.CF_UINT32:
            isUnsigned = true;
          case TFormat.CF_BOOLEAN:
          case TFormat.CF_INT32:
            for (; i < len && i < m + b.length / fmtsiz; i++)
            {
              ival = ds.readInt();
              data[i + offset] =  isUnsigned ? UnsignedInt(ival) : ival;
            }
            break;
          case TFormat.CF_BITFIELD16:
            makeBitfield();
            if (dBitField == null) return TErrorList.not_initialized;
            short sval;
            for (; i < len && i < m + b.length / fmtsiz; i++)
            {
              sval = ds.readShort();
              dBitField[i+offset].setValue(sval);
              data[i + offset] = dBitField[i+offset].getField(dField, sval);
             // data[i + offset] = dBitField.getField(dField, ds.readShort());
            }
            break;
          case TFormat.CF_INT16:
            for (; i < len && i < m + b.length / fmtsiz; i++)
              data[i + offset] = ds.readShort();
            break;
          case TFormat.CF_BITFIELD8:
            makeBitfield();
            if (dBitField == null) return TErrorList.not_initialized;
            byte bval;            
            for (; i < len && i < m + b.length / fmtsiz; i++)
            {
              bval = ds.readByte();
              dBitField[i+offset].setValue(bval);
              data[i + offset] = dBitField[i+offset].getField(dField, bval);
            }
            break;
          case TFormat.CF_BYTE:
            for (; i < len && i < m + b.length / fmtsiz; i++)
              data[i + offset] = UnsignedShort(ds.readByte());
            break;
          default:
            ds.close();
            bs.close();
            return TErrorList.illegal_format;
        }
        ds.close();
        bs.close();
      }
      dCompletionLength = i;
      isDataObjectInSync = true;
    }
    catch (IOException e)
    {
      e.printStackTrace();
      MsgLog.log("TDataType.getData", e.getMessage(), TErrorList.code_failure, e, 0);
    }
    return 0;
  } }
  /**
   * Puts the current canonical data set into the underlying data byte
   *        stream Puts the input data set into the underlying byte stream,
   *        properly swapped and aligned. It replaces the contents of the
   *        original, referenced data array with that of the data object given.
   * @param data
   *          is data primitive whose value is copied into the underlying byte
   *          stream and referenced data object.
   * @note If this TDataType object is used as input data in persistent links,
   *       then the data object might well be 'locked' meaning that the
   *       underlying data will NOT be changed, as this would constitute a
   *       change in contract.
   * @return 0 upon success or a tine error code
   */
  public int putData(long data)
  {
    long[] d = new long[1];
    d[0] = data;
    return putData(d);
  }
  /**
   * Puts the current canonical data set into the underlying data byte
   *        stream Puts the input data set into the underlying byte stream,
   *        properly swapped and aligned. It replaces the contents of the
   *        original, referenced data array with that of the data object given.
   * @param data
   *          is the data object whose contents should be copied into the
   *          underlying byte stream and referenced data object.
   * @note If this TDataType object is used as input data in persistent links,
   *       then the data object might well be 'locked' meaning that the
   *       underlying data will NOT be changed, as this would constitute a
   *       change in contract.
   * @return 0 upon success or a tine error code
   */
  public int putData(long[] data)
  {
    return putData(data, data != null ? data.length : 0, 0);
  }
  /**
   * Puts the current canonical data set into the underlying data byte
   *        stream Puts the input data set into the underlying byte stream,
   *        properly swapped and aligned. It replaces the contents of the
   *        original, referenced data array with that of the data object given.
   * @param data
   *          is the data object whose contents should be copied into the
   *          underlying byte stream and referenced data object.
   * @param length
   *          is the number of elements from the give data object to copy
   * @param offset
   *          is the starting offset within the array to copy
   * @note If this TDataType object is used as input data in persistent links,
   *       then the data object might well be 'locked' meaning that the
   *       underlying data will NOT be changed, as this would constitute a
   *       change in contract.
   * @return 0 upon success or a tine error code
   */
  public int putData(long[] data, int length, int offset)
  {
    return putData(0,data,length,offset,true);  	
  }
  public int putData(int start, long[] data, int length, int offset, boolean redim)
  { synchronized (hDataObject) {
    if (data == null || length <= 0 || offset < 0) return TErrorList.argument_list_error;
    if (length > data.length - offset) length = data.length - offset;
    if (length <= 0) return TErrorList.argument_list_error;
    if (isLocked) throw new InputDataLockedException(lockedMessage);
    int maxlen = Math.min(dArrayLength, length);
    if (maxlen > dArrayLength - start) maxlen = dArrayLength - start;
    switch (dFormat)
    {
      case TFormat.CF_DOUBLE:
        for (int i = 0; i < maxlen; i++)
          ((double[]) hDataObject)[i+start] = (double) data[i + offset];
        break;
      case TFormat.CF_FLOAT:
        for (int i = 0; i < maxlen; i++)
          ((float[]) hDataObject)[i+start] = (float) data[i + offset];
        break;
      case TFormat.CF_INT16:
        for (int i = 0; i < maxlen; i++)
          ((short[]) hDataObject)[i+start] = (short) data[i + offset];
        break;
      case TFormat.CF_UINT16:
        for (int i = 0; i < maxlen; i++)
          ((short[]) hDataObject)[i+start] = UnsignedShort((int)data[i + offset]);
        break;
      case TFormat.CF_INT64:
        for (int i = 0; i < maxlen; i++)
          ((long[]) hDataObject)[i+start] = (long) data[i + offset];
        break;
      case TFormat.CF_BOOLEAN:
        for (int i = 0; i < maxlen; i++)
          ((boolean[]) hDataObject)[i+start] = (boolean)(data[i + offset] != 0);
        break;
      case TFormat.CF_INT32:
        for (int i = 0; i < maxlen; i++)
          ((int[]) hDataObject)[i+start] = (int) data[i + offset];
        break;
      case TFormat.CF_UINT32:
        for (int i = 0; i < maxlen; i++)
          ((int[]) hDataObject)[i+start] = UnsignedInt(data[i + offset]);        
        break;
      default:
        return TErrorList.illegal_format;
    }
    if (redim) dArrayLength = Math.min(dArrayLength, length);
    repackageDataForOutput();
    return 0;
  } }
  /**
   * Puts the input data value into the underlying byte stream,
   * properly swapped and aligned. It replaces the contents of the
   * original, referenced data array with that of the data object given.
   * @param value
   *          is the single value which should be copied into the
   *          underlying byte stream and referenced data object at the given offset.
   * @param offset
   *          is the position offset within the targeted array
   * @note If this TDataType object is used as input data in persistent links,
   *       then the data object might well be 'locked' meaning that the
   *       underlying data will NOT be changed, as this would constitute a
   *       change in contract.
   * @return 0 upon success or a tine error code
   * 
   * @throws InputDataLockedException if a link using this TDataType as input 
   * is bound to another link 
   * and an attempt to alter the input data set (which is part of the contract
   * cached at the target server) is made.  
   * 
   */
  public int putData(long value, int offset)
  {
    return putDataValue(Long.valueOf(value),offset);
  }
  /**
   * Puts the input data value array into the underlying byte stream,
   * properly swapped and aligned. It replaces the contents of the
   * original, referenced data array with that of the data object given.
   * @param values
   *          is the values array which should be copied into the
   *          underlying byte stream and referenced data object at the given offset.
   * @param offset
   *          is the position offset within the targeted array
   * @note If this TDataType object is used as input data in persistent links,
   *       then the data object might well be 'locked' meaning that the
   *       underlying data will NOT be changed, as this would constitute a
   *       change in contract.
   * @return 0 upon success or a tine error code
   * 
   * @throws InputDataLockedException if a link using this TDataType as input 
   * is bound to another link 
   * and an attempt to alter the input data set (which is part of the contract
   * cached at the target server) is made.  
   * 
   */
  public int putData(long[] values,int offset)
  {
    return putDataSegment(values, offset);
  }
  public byte[] getDataBuffer()
  {
    try
    {
      if (dBuffers == null) return null;
      if (dBuffer == null)
        dBuffer = new ByteArrayOutputStream(dSizeInBytes);
      else
        dBuffer.reset();
      DataOutputStream dos = new DataOutputStream(dBuffer);
      for (int n = 0; n < numblks; n++)
      {
        if (dBuffers[n] == null) break;
        dos.write(dBuffers[n].toByteArray());
      }
      dos.close();
    }
    catch (IOException e)
    {
      e.printStackTrace();
      MsgLog.log("TDataType.getDataBuffer", e.getMessage(), TErrorList.code_failure, e, 0);
    }
    return dBuffer.toByteArray();
  }
  public byte[] getDataBuffer(int n)
  {
    try
    {
      if (dBuffer == null)
        dBuffer = new ByteArrayOutputStream();
      else
        dBuffer.reset();
      DataOutputStream dos = new DataOutputStream(dBuffer);
      if (n < numblks && dBuffers[n] != null)
      {
        dos.write(dBuffers[n].toByteArray());
      }
      dos.close();
    }
    catch (IOException e)
    {
      e.printStackTrace();
      MsgLog.log("TDataType.getDataBuffer", e.getMessage(), TErrorList.code_failure, e, 0);
    }
    return dBuffer.toByteArray();
  }
  /**
   * Returns the data array length of the associated tine Data Type
   *        object
   * @return Returns the data array length of the associated tine Data Type
   *         object
   */
  public int getArrayLength()
  {
    int dlen = dArrayLength;
    if (dFormat == TFormat.CF_STRUCT)
    { // if tagged structure, then this call returns the number of structures, not bytes !
      int ssiz = TStructRegistry.getSizeInBytes(dTag,dTagKey);
      if (ssiz < 1) ssiz = 1;
      dlen /= ssiz;
    }
    if (TFormat.isAdjustableLength(dFormat))
    {
      try
      {
        if (dTag.length() > 0) dlen = Integer.parseInt(dTag);
      }
      catch (NumberFormatException e) {};
    }
    return dlen;
  }
  /**
   * Returns the data type of the associated tine Data Type object
   * @return Returns the data type of the associated tine Data Type object
   */
  public short getFormat()
  {
    return dFormat >= 0 ? dFormat : (short) (256 + dFormat);
  }
  public void removeTagFromLink()
  {
    removeTagAtLinkTime = true;
  }
  public String getTag()
  {
    return removeTagAtLinkTime ? "" : dTag;
  }
  public int getSizeInBytes()
  {
    return dSizeInBytes;
  }
  public int getDataSize()
  {
    int fmtsiz = 0;
    if ((fmtsiz = TFormat.formatSizeOf(dFormat)) > 0)
    {
      int datasize = fmtsiz * dArrayLength;
      if (dFormat == TFormat.CF_SPECTRUM) datasize += SPECTRUM.hdrSizeInBytes;
      if (dFormat == TFormat.CF_GSPECTRUM) datasize += GSPECTRUM.hdrSizeInBytes;
      return datasize;
    }
    return 0;
  }
  private void makeNewCompoundDataObject(TCompoundDataObject[] Data, int index)
  {
    try
    {
      Data[index] = (TCompoundDataObject) Data.getClass().getComponentType().getDeclaredConstructor().newInstance();
    }
    catch (Exception e)
    {
      e.printStackTrace();
      MsgLog.log("makeNewCompoundDataObject", e.toString(), TErrorList.code_failure, e, 0);
    }
  }
  private byte[] getStringArrayBytes(Object[] data)
  {
    int sptr = 0, slen = 0;
    byte[] bdata = new byte[dSizeInBytes];
    String s;
    for (int i = 0; i < data.length; i++)
    { // now fill in the string space
      if (data[i] != null)
      {
        if (data[i] instanceof String) 
          s = (String)data[i];
        else if (data[i] instanceof KEYVALUE)
          s = ((KEYVALUE)data[i]).getString();
        else continue; 
        slen = s.length();
        if (slen + sptr >= dSizeInBytes) break;
        System.arraycopy(s.getBytes(), 0, bdata, sptr, slen);
        sptr += slen;
      }
      if (sptr < dSizeInBytes) bdata[sptr++] = 0;
    }
    return bdata;
  }
  private byte[] inptBytes = null;
  private void writeInptBytesToStream(DataOutputStream ds,int index) throws IOException
  {
    if (inptBytes == null) return;
    byte[] b = inptBytes;
    int len = Math.min(mtu, b.length-index);
    ds.write(b,index,len);
    if (b.length == len + index) inptBytes = null;    
  }
  private void setObjectContents(DataOutputStream ds, Object Data, int index)
  {
    try
    {
      if (Data instanceof byte[])
      {
        ds.writeByte(((byte[]) Data)[index]);
      }
      else if (Data instanceof short[])
      {
        ds.writeShort(Swap.Short(((short[]) Data)[index]));
      }
      else if (Data instanceof int[])
      {
        ds.writeInt(Swap.Int(((int[]) Data)[index]));
      }
      else if (Data instanceof boolean[])
      {
        ds.writeInt(Swap.Int(((boolean[]) Data)[index] ? 1 : 0));
      }
      else if (Data instanceof long[])
      {
        ds.writeLong(Swap.Long(((long[]) Data)[index]));
      }
      else if (Data instanceof float[])
      {
        ds.write(Swap.Float(((float[]) Data)[index]));
      }
      else if (Data instanceof double[])
      {
        ds.write(Swap.Double(((double[]) Data)[index]));
      }
      else if (Data instanceof char[])
      {
        if (index == 0 || inptBytes == null) inptBytes = (new String((char[]) Data)).getBytes();
        writeInptBytesToStream(ds,index);
      }
      else if (Data instanceof StringBuffer)
      {
        if (index == 0 || inptBytes == null) inptBytes = ((StringBuffer) Data).toString().getBytes();
        writeInptBytesToStream(ds,index);
      }
      else if (Data instanceof String)
      {
        if (index == 0 || inptBytes == null) inptBytes = ((String) Data).getBytes();
        writeInptBytesToStream(ds,index);
      }
      else if (Data instanceof ASPECTRUM[])
      { // this must be rescanned and re-serialized (currently will NOT work as input data)
        int totalHdrBytes = 0, totalSpectraBytes = 0;
        int len = ((ASPECTRUM[]) Data).length;
        for (int i=0; i<len; i++)
        {
          if (((ASPECTRUM[])Data)[i] == null) ((ASPECTRUM[])Data)[i] = new ASPECTRUM();
          totalHdrBytes += ASPECTRUM.hdrSizeInBytes + 8;
          totalSpectraBytes += ((ASPECTRUM[])Data)[i].length * 4;
        }
        dArrayLength = totalHdrBytes + totalSpectraBytes;
        dSizeInBytes = totalHdrBytes + totalSpectraBytes;
        byte[] b, bdata = new byte[totalHdrBytes+totalSpectraBytes];
        int off = 0, xoff = totalHdrBytes;
        for (int i=0; i<len; i++)
        {
          ((ASPECTRUM[])Data)[i].setExtendedOffset(xoff);
          b = ((ASPECTRUM[])Data)[i].toByteArray();
          xoff = ((ASPECTRUM[])Data)[i].getExtendedOffset();
          System.arraycopy(b, 0, bdata, off, b.length);
          off += b.length;
        }
        for (int i=0; i<len; i++)
        {
          b = ((ASPECTRUM[])Data)[i].toSpectrumByteArray();
          if (b == null) continue;
          System.arraycopy(b, 0, bdata, off, b.length);
          off += b.length;
        }
        ds.write(bdata);
      }
      else if (Data instanceof IMAGE[])
      { // CF_AIMAGE: this must be rescanned and re-serialized (currently will NOT work as input data)
        int len = ((IMAGE[]) Data).length;
        int[] xsiz = new int[len];
        for (int i=0; i<len; i++)
        {
          if (((IMAGE[])Data)[i] == null) ((IMAGE[])Data)[i] = new IMAGE(100);
          ((IMAGE[])Data)[i].writeHeaders(ds);
          ds.writeInt(Swap.Int(((IMAGE[])Data)[i].getExtendedOffset()));
          xsiz[i] = ((IMAGE[])Data)[i].getExtendedSize();
          ds.writeInt(Swap.Int(xsiz[i]));
        }
        for (int i=0; i<len; i++)
        {
          ds.write(((IMAGE[])Data)[i].getImageFrameBuffer(),0,xsiz[i]);
        }
      }
      else if (Data instanceof TCompoundDataObject[])
      {
        if (((TCompoundDataObject[]) Data)[index] == null)
          makeNewCompoundDataObject((TCompoundDataObject[]) Data, index);
        ds.write(((TCompoundDataObject[]) Data)[index].toByteArray());
      }
      else if (Data instanceof String[])
      {
        if (index == 0 || inptBytes == null) inptBytes = getStringArrayBytes((String[]) Data);
        writeInptBytesToStream(ds,index);
      }
      else if (Data instanceof KEYVALUE[])
      {
        if (index == 0 || inptBytes == null) inptBytes = getStringArrayBytes((KEYVALUE[]) Data);
        writeInptBytesToStream(ds,index);
      }
      else if (Data instanceof SPECTRUM)
      {
        if (index == 0 || inptBytes == null) inptBytes = ((SPECTRUM) Data).toByteArray();
        writeInptBytesToStream(ds,index);
      }
      else if (Data instanceof GSPECTRUM)
      {
        if (index == 0 || inptBytes == null) inptBytes = ((GSPECTRUM) Data).toByteArray();
        writeInptBytesToStream(ds,index);
      }
      else if (Data instanceof IMAGE)
      {
        if (index == 0 || inptBytes == null) inptBytes = ((IMAGE) Data).toByteArray();
        writeInptBytesToStream(ds,index);
        //((IMAGE) Data).write(ds); // mdavid: optimized
      }
      else if (Data instanceof MDA)
      {
        if (index == 0 || inptBytes == null) inptBytes = ((MDA) Data).toByteArray();
        writeInptBytesToStream(ds,index);
      }
    }
    catch (IOException e)
    {
      e.printStackTrace();
      MsgLog.log("TDataType.setObjectContents", e.getMessage(), TErrorList.code_failure, e, 0);
    }
  }
  public void setDataObject(int dataSize, short dataFormat)
  { // only allow this if the data object is null (used for CF_DEFAULT)
    if (hDataObject == null) 
    {
      makeDataObject(dataSize, dataFormat);
      dSizeInBytes = dataSize * TFormat.formatSizeOf(dataFormat);
      initCounters();
    }
    if (numblks < 1) numblks = 1;
    boolean allocBuffers = dBuffers == null;
    if (dBuffers != null)
    {
      if (dBuffers.length > 0 && dBuffers.length < numblks) numblks = dBuffers.length;
      if (dBuffers.length > numblks) allocBuffers = true;
    }
    if (allocBuffers) 
    {
      dBuffers = new ByteArrayOutputStream[numblks];
      for (int i=0; i<numblks; i++)
      {
        dBuffers[i] = new ByteArrayOutputStream(dBufferSegSize);
      }
    }
    if (dBuffersReady == null) dBuffersReady = new boolean[numblks];
  }
  private void setObjectHandle(TCompoundDataObject[] Data)
  {
    hDataObject = Data;
  }
  private void setObjectHandle(Object Data)
  {
    hDataObject = Data;
  }
  public static int getObjectSize(Object Data)
  {
    if (Data instanceof TCompoundDataObject[])
      return ((TCompoundDataObject[]) Data).length;
    if (Data instanceof TTaggedStructure[])
      return ((TTaggedStructure[]) Data).length;
    else if (Data instanceof byte[])
      return ((byte[]) Data).length;
    else if (Data instanceof short[])
      return ((short[]) Data).length;
    else if (Data instanceof int[])
      return ((int[]) Data).length;
    else if (Data instanceof long[])
      return ((long[]) Data).length;
    else if (Data instanceof boolean[])
      return ((boolean[]) Data).length;
    else if (Data instanceof float[])
      return ((float[]) Data).length;
    else if (Data instanceof double[])
      return ((double[]) Data).length;
    else if (Data instanceof SPECTRUM)
      return ((SPECTRUM) Data).hByteArray.length;
    else if (Data instanceof GSPECTRUM)
      return ((GSPECTRUM) Data).hByteArray.length;
    else if (Data instanceof IMAGE)
      return ((IMAGE) Data).getImageFrameSize();
    else if (Data instanceof char[])
      return ((char[]) Data).length;
    else if (Data instanceof String)
      return ((String) Data).length();
    else if (Data instanceof StringBuffer) return ((StringBuffer) Data).toString().length();
    return 0;
  }
  public static short getObjectFormat(Object Data)
  {
    if (Data instanceof TCompoundDataObject[])
    {
      try
      {
        Object tdo = ((TCompoundDataObject[]) Data)[0] != null ? ((TCompoundDataObject[]) Data)[0] : Data
            .getClass().getComponentType().getDeclaredConstructor().newInstance();
        return ((TCompoundDataObject) tdo).getFormat();
      }
      catch (Exception e)
      {
        e.printStackTrace();
        MsgLog.log("TDataType.getObjectFormat", e.toString(), TErrorList.code_failure, e, 0);
      }
    }
    if (Data instanceof TCompoundDataObject)
      return ((TCompoundDataObject) Data).getFormat();
    if (Data instanceof byte[])
      return TFormat.CF_BYTE;
    else if (Data instanceof short[])
      return TFormat.CF_INT16;
    else if (Data instanceof int[])
      return TFormat.CF_INT32;
    else if (Data instanceof float[])
      return TFormat.CF_FLOAT;
    else if (Data instanceof long[])
      return TFormat.CF_INT64;
    else if (Data instanceof double[])
      return TFormat.CF_DOUBLE;
    else if (Data instanceof boolean[])
      return TFormat.CF_BOOLEAN;
    else if (Data instanceof String)
      return TFormat.CF_TEXT;
    else if (Data instanceof StringBuffer)
      return TFormat.CF_TEXT;
    else if (Data instanceof char[])
      return TFormat.CF_TEXT;
    else if (Data instanceof String[]) 
      return TFormat.CF_STRING;
    else if (Data instanceof KEYVALUE[]) 
      return TFormat.CF_KEYVALUE;
    else if (Data instanceof TTaggedStructure[])
      return TFormat.CF_STRUCT;
    return TFormat.CF_NULL;
  }
  /**
   * Gets the canonical data set from the underlying data byte stream
   *        Gets the fully formed (properly assembled and byte-swapped)
   *        canonical data set from the underlying byte stream. This method is
   *        used internally, but can be used to get the underlying data into the
   *        specifically referenced data, for instance when the TDataType object
   *        was instantiated with 'unbound' data.
   * @param data
   *          is the data object into which the underlying data should be
   *          prepared and copied.
   * @return 0 on success or a tine error code
   */
  public int getData(Object Data)
  {
    int len = 1;
    try { len = Array.getLength(Data); } catch (Exception ignore) {}
    return getData(Data,len,0);
  }
  /**
   * Gets the canonical data set from the underlying data byte stream
   *        Gets the fully formed (properly assembled and byte-swapped)
   *        canonical data set from the underlying byte stream. This method is
   *        used internally, but can be used to get the underlying data into the
   *        specifically referenced data, for instance when the TDataType object
   *        was instantiated with 'unbound' data.
   * @param data
   *          is the data object into which the underlying data should be
   *          prepared and copied.
   * @param length
   *          is the number of elements to get
   * @param offset
   *          is the beginning position of the target data array into which the
   *          data should be copied
   * @return 0 on success or a tine error code
   */
  public int getData(Object Data, int length, int offset)
  {
    if (Data instanceof byte[])
      getData((byte[]) Data, length, offset);
    else if (Data instanceof short[])
      getData((short[]) Data, length, offset);
    else if (Data instanceof int[])
      getData((int[]) Data, length, offset);
    else if (Data instanceof boolean[])
      getData((boolean[]) Data, length, offset);
    else if (Data instanceof float[])
      getData((float[]) Data, length, offset);
    else if (Data instanceof double[])
      getData((double[]) Data, length, offset);
    else if (Data instanceof long[])
      getData((long[]) Data, length, offset);
    else if (Data instanceof char[])
      getData((char[]) Data);
    else if (Data instanceof StringBuffer)
      getData((StringBuffer) Data);
    else if (Data instanceof SPECTRUM)
      getData((SPECTRUM) Data);
    else if (Data instanceof GSPECTRUM)
      getData((GSPECTRUM) Data);
    else if (Data instanceof MDA)
      getData((MDA) Data);
    else if (Data instanceof ASPECTRUM[])
      getData((ASPECTRUM[]) Data);
    else if (Data instanceof IMAGE)
      getData((IMAGE) Data);
    else if (Data instanceof IMAGE[])
      getData((IMAGE[]) Data);
    else if (Data instanceof TCompoundDataObject[])
      getData((TCompoundDataObject[]) Data, length, offset);
    else if (Data instanceof KEYVALUE[])
      getData((KEYVALUE[]) Data, length, offset);
    else if (Data instanceof String[]) 
      getData((String[]) Data, length, offset);
    else if (Data instanceof TTaggedStructure[])
      getData((TTaggedStructure[]) Data, length, offset);
    else
      return TErrorList.illegal_format;
    return 0;
  }
  public int dataFill(String fromStr)
  {
    // TODO: add the new formats
    int n = 0;
    switch (dFormat)
    {
      case TFormat.CF_NULL:
        break;
      case TFormat.CF_NAME16FI:
        for (n = 0; n < dArrayLength && ((NAME16FI[]) hDataObject)[n] != null; n++)
          ((NAME16FI[]) hDataObject)[n].setName(fromStr);
        break;
      case TFormat.CF_NAME16II:
        for (n = 0; n < dArrayLength && ((NAME16II[]) hDataObject)[n] != null; n++)
          ((NAME16II[]) hDataObject)[n].setName(fromStr);
        break;
      case TFormat.CF_NAME16DBLDBL:
        for (n = 0; n < dArrayLength && ((NAME16DBLDBL[]) hDataObject)[n] != null; n++)
          ((NAME16DBLDBL[]) hDataObject)[n].setName(fromStr);
        break;
      case TFormat.CF_NAME32DBLDBL:
        for (n = 0; n < dArrayLength && ((NAME32DBLDBL[]) hDataObject)[n] != null; n++)
          ((NAME32DBLDBL[]) hDataObject)[n].setName(fromStr);
        break;
      case TFormat.CF_NAME64DBL:
        for (n = 0; n < dArrayLength && ((NAME64DBL[]) hDataObject)[n] != null; n++)
          ((NAME64DBL[]) hDataObject)[n].setName(fromStr);
        break;
      case TFormat.CF_NAME64TIME:
        for (n = 0; n < dArrayLength && ((NAME64TIME[]) hDataObject)[n] != null; n++)
          ((NAME64TIME[]) hDataObject)[n].setName(fromStr);
        break;
      case TFormat.CF_NAME64DBLDBL:
        for (n = 0; n < dArrayLength && ((NAME64DBLDBL[]) hDataObject)[n] != null; n++)
          ((NAME64DBLDBL[]) hDataObject)[n].setName(fromStr);
        break;
      case TFormat.CF_NAME64DBLDBLDBL:
        for (n = 0; n < dArrayLength && ((NAME64DBLDBLDBL[]) hDataObject)[n] != null; n++)
          ((NAME64DBLDBLDBL[]) hDataObject)[n].setName(fromStr);
        break;
      case TFormat.CF_NAME8:
      case TFormat.CF_NAME16:
      case TFormat.CF_NAME32:
      case TFormat.CF_NAME48:
      case TFormat.CF_NAME64:
        for (n = 0; n < dArrayLength && ((TCompoundDataObject[]) hDataObject)[n] != null; n++)
          ((NAME[]) hDataObject)[n].setName(fromStr);
        break;
      case TFormat.CF_NAME8I:
      case TFormat.CF_NAME16I:
      case TFormat.CF_NAME32I:
      case TFormat.CF_NAME48I:
      case TFormat.CF_NAME64I:
        for (n = 0; n < dArrayLength && ((TCompoundDataObject[]) hDataObject)[n] != null; n++)
          ((NAMEINT[]) hDataObject)[n].setName(fromStr);
        break;
      case TFormat.CF_USTRING:
        for (n = 0; n < dArrayLength && ((USTRING[]) hDataObject)[n] != null; n++)
          ((USTRING[]) hDataObject)[n].setString(fromStr);
        break;
      case TFormat.CF_UNAME:
        for (n = 0; n < dArrayLength && ((UNAME[]) hDataObject)[n] != null; n++)
          ((UNAME[]) hDataObject)[n].setString(fromStr);
        break;
      case TFormat.CF_XML:
      case TFormat.CF_TEXT:
        int size = dArrayLength < fromStr.length() ? dArrayLength : fromStr.length();
        char[] s = fromStr.toCharArray();
        for (n = 0; n < size && hDataObject != null; n++)
          ((char[]) hDataObject)[n] = s[n];
        break;
    }
    return n;
  }
  public void dataFill()
  {
    switch (dFormat)
    {
      case TFormat.CF_NULL:
        break;
      case TFormat.CF_NAME16FI:
        Arrays.fill((NAME16FI[]) hDataObject, new NAME16FI());
        break;
      case TFormat.CF_NAME16II:
        Arrays.fill((NAME16II[]) hDataObject, new NAME16II());
        break;
      case TFormat.CF_NAME16DBLDBL:
        Arrays.fill((NAME16DBLDBL[]) hDataObject, new NAME16DBLDBL());
        break;
      case TFormat.CF_NAME32DBLDBL:
        Arrays.fill((NAME32DBLDBL[]) hDataObject, new NAME32DBLDBL());
        break;
      case TFormat.CF_NAME64DBL:
        Arrays.fill((NAME64DBL[]) hDataObject, new NAME64DBL());
        break;
      case TFormat.CF_NAME64DBLDBL:
        Arrays.fill((NAME64DBLDBL[]) hDataObject, new NAME64DBLDBL());
        break;
      case TFormat.CF_NAME64DBLDBLDBL:
        Arrays.fill((NAME64DBLDBLDBL[]) hDataObject, new NAME64DBLDBLDBL());
        break;
      case TFormat.CF_STRUCT:
      case TFormat.CF_BYTE:
        Arrays.fill((byte[]) hDataObject, (byte)0);
        break;
      case TFormat.CF_INT16:
        Arrays.fill((short[]) hDataObject, (short)0);
        break;
      case TFormat.CF_INT32:
        Arrays.fill((int[]) hDataObject, (int)0);
        break;
      case TFormat.CF_INT64:
        Arrays.fill((long[]) hDataObject, (long)0);
        break;
      case TFormat.CF_FLOAT:
        Arrays.fill((float[]) hDataObject, (float)0);
        break;
      case TFormat.CF_DOUBLE:
        Arrays.fill((double[]) hDataObject, (double)0);
        break;
      case TFormat.CF_INTINT:
        Arrays.fill((INTINT[]) hDataObject, new INTINT());
        break;
      case TFormat.CF_FLTINT:
        Arrays.fill((FLTINT[]) hDataObject, new FLTINT());
        break;
      case TFormat.CF_FLTFLT:
        Arrays.fill((FLTFLT[]) hDataObject, new FLTFLT());
        break;
      case TFormat.CF_DBLTIME:
        Arrays.fill((DBLTIME[]) hDataObject, new DBLTIME());
        break;
      case TFormat.CF_DBLDBL:
        Arrays.fill((DBLDBL[]) hDataObject, new DBLDBL());
        break;
      case TFormat.CF_DBLDBLDBL:
        Arrays.fill((DBLDBLDBL[]) hDataObject, new DBLDBLDBL());
        break;
      case TFormat.CF_INTFLTINT:
        Arrays.fill((INTFLTINT[]) hDataObject, new INTFLTINT());
        break;
      case TFormat.CF_FLTFLTINT:
        Arrays.fill((FLTFLTINT[]) hDataObject, new FLTFLTINT());
        break;
      case TFormat.CF_FLTINTINT:
        Arrays.fill((FLTINTINT[]) hDataObject, new FLTINTINT());
        break;
      case TFormat.CF_INTINTINT:
        Arrays.fill((INTINTINT[]) hDataObject, new INTINTINT());
        break;
      case TFormat.CF_FLTINTFLTINT:
        Arrays.fill((FLTINTFLTINT[]) hDataObject, new FLTINTFLTINT());
        break;
      case TFormat.CF_ADDRESS:
        Arrays.fill((INTINTINTINT[]) hDataObject, new INTINTINTINT());
        break;
      case TFormat.CF_FILTER:
        Arrays.fill((INTFLTFLTFLT[]) hDataObject, new INTFLTFLTFLT());
        break;
      case TFormat.CF_FWINDOW:
        Arrays.fill((INTINTFLTFLT[]) hDataObject, new INTINTFLTFLT());
        break;
      case TFormat.CF_USTRING:
        Arrays.fill((USTRING[]) hDataObject, new USTRING());
        break;
      case TFormat.CF_UNAME:
        Arrays.fill((UNAME[]) hDataObject, new UNAME());
        break;
      case TFormat.CF_SPECTRUM:
        Arrays.fill((SPECTRUM[]) hDataObject, new SPECTRUM());
        break;
      case TFormat.CF_GSPECTRUM:
        Arrays.fill((GSPECTRUM[]) hDataObject, new GSPECTRUM());
        break;
    }
    return;
  }
  public int dataFill(double fromVal)
  {
    return dataFill(fromVal, 0);
  }
  public int dataFill(double fromVal, int offset)
  {
    // TODO: add the new formats
    int n = 0, len;
    if (offset >= dArrayLength) return 0;
    switch (dFormat)
    {
      case TFormat.CF_NULL:
        break;
      case TFormat.CF_NAME16FI:
        if ((len = ((NAME16FI[]) hDataObject).length) > dArrayLength) len = dArrayLength;
        for (n = offset; n < len && ((NAME16FI[]) hDataObject)[n] != null; n++)
          ((NAME16FI[]) hDataObject)[n].fval = (float) fromVal;
        break;
      case TFormat.CF_NAME16II:
        if ((len = ((NAME16II[]) hDataObject).length) > dArrayLength) len = dArrayLength;
        for (n = offset; n < len && ((NAME16II[]) hDataObject)[n] != null; n++)
          ((NAME16II[]) hDataObject)[n].i1val = (int) fromVal;
        break;
      case TFormat.CF_NAME16DBLDBL:
        if ((len = ((NAME16DBLDBL[]) hDataObject).length) > dArrayLength) len = dArrayLength;
        for (n = offset; n < len && ((NAME16DBLDBL[]) hDataObject)[n] != null; n++)
          ((NAME16DBLDBL[]) hDataObject)[n].d1val = fromVal;
        break;
      case TFormat.CF_NAME32DBLDBL:
        if ((len = ((NAME32DBLDBL[]) hDataObject).length) > dArrayLength) len = dArrayLength;
        for (n = offset; n < len && ((NAME32DBLDBL[]) hDataObject)[n] != null; n++)
          ((NAME32DBLDBL[]) hDataObject)[n].d1val = fromVal;
        break;
      case TFormat.CF_NAME64DBL:
        if ((len = ((NAME64DBL[]) hDataObject).length) > dArrayLength) len = dArrayLength;
        for (n = offset; n < len && ((NAME64DBL[]) hDataObject)[n] != null; n++)
          ((NAME64DBL[]) hDataObject)[n].dval = fromVal;
        break;
      case TFormat.CF_NAME64DBLDBL:
        if ((len = ((NAME64DBLDBL[]) hDataObject).length) > dArrayLength) len = dArrayLength;
        for (n = offset; n < len && ((NAME64DBLDBL[]) hDataObject)[n] != null; n++)
          ((NAME64DBLDBL[]) hDataObject)[n].d1val = fromVal;
        break;
      case TFormat.CF_NAME64DBLDBLDBL:
        if ((len = ((NAME64DBLDBLDBL[]) hDataObject).length) > dArrayLength) len = dArrayLength;
        for (n = offset; n < len && ((NAME64DBLDBLDBL[]) hDataObject)[n] != null; n++)
          ((NAME64DBLDBLDBL[]) hDataObject)[n].d1val = fromVal;
        break;
      case TFormat.CF_STRUCT:
      case TFormat.CF_BYTE:
        if ((len = ((byte[]) hDataObject).length) > dArrayLength) len = dArrayLength;
        for (n = offset; n < len && ((byte[]) hDataObject) != null; n++)
          ((byte[]) hDataObject)[n] = (byte) fromVal;
        break;
      case TFormat.CF_INT16:
        if ((len = ((short[]) hDataObject).length) > dArrayLength) len = dArrayLength;
        for (n = offset; n < len && ((short[]) hDataObject) != null; n++)
          ((short[]) hDataObject)[n] = (short) fromVal;
        break;
      case TFormat.CF_INT32:
        if ((len = ((int[]) hDataObject).length) > dArrayLength) len = dArrayLength;
        for (n = offset; n < len && ((int[]) hDataObject) != null; n++)
          ((int[]) hDataObject)[n] = (int) fromVal;
        break;
      case TFormat.CF_INT64:
        if ((len = ((long[]) hDataObject).length) > dArrayLength) len = dArrayLength;
        for (n = offset; n < len && ((long[]) hDataObject) != null; n++)
          ((long[]) hDataObject)[n] = (long) fromVal;
        break;
      case TFormat.CF_FLOAT:
        if ((len = ((float[]) hDataObject).length) > dArrayLength) len = dArrayLength;
        for (n = offset; n < len && ((float[]) hDataObject) != null; n++)
          ((float[]) hDataObject)[n] = (float) fromVal;
        break;
      case TFormat.CF_DOUBLE:
        if ((len = ((double[]) hDataObject).length) > dArrayLength) len = dArrayLength;
        for (n = offset; n < len && ((double[]) hDataObject) != null; n++)
          ((double[]) hDataObject)[n] = fromVal;
        break;
      case TFormat.CF_INTINT:
        if ((len = ((INTINT[]) hDataObject).length) > dArrayLength) len = dArrayLength;
        for (n = offset; n < len && ((INTINT[]) hDataObject)[n] != null; n++)
          ((INTINT[]) hDataObject)[n].i1val = (int) fromVal;
        break;
      case TFormat.CF_FLTINT:
        if ((len = ((FLTINT[]) hDataObject).length) > dArrayLength) len = dArrayLength;
        for (n = offset; n < len && ((FLTINT[]) hDataObject)[n] != null; n++)
          ((FLTINT[]) hDataObject)[n].fval = (float) fromVal;
        break;
      case TFormat.CF_FLTFLT:
        if ((len = ((FLTFLT[]) hDataObject).length) > dArrayLength) len = dArrayLength;
        for (n = offset; n < len && ((FLTFLT[]) hDataObject)[n] != null; n++)
          ((FLTFLT[]) hDataObject)[n].f1val = (float) fromVal;
        break;
      case TFormat.CF_DBLTIME:
        if ((len = ((DBLTIME[]) hDataObject).length) > dArrayLength) len = dArrayLength;
        for (n = offset; n < len && ((DBLTIME[]) hDataObject)[n] != null; n++)
          ((DBLTIME[]) hDataObject)[n].d1val = fromVal;
        break;
      case TFormat.CF_DBLDBL:
        if ((len = ((DBLDBL[]) hDataObject).length) > dArrayLength) len = dArrayLength;
        for (n = offset; n < len && ((DBLDBL[]) hDataObject)[n] != null; n++)
          ((DBLDBL[]) hDataObject)[n].d1val = fromVal;
        break;
      case TFormat.CF_DBLDBLDBL:
        if ((len = ((DBLDBLDBL[]) hDataObject).length) > dArrayLength) len = dArrayLength;
        for (n = offset; n < len && ((DBLDBLDBL[]) hDataObject)[n] != null; n++)
          ((DBLDBLDBL[]) hDataObject)[n].d1val = fromVal;
        break;
      case TFormat.CF_INTFLTINT:
        if ((len = ((INTFLTINT[]) hDataObject).length) > dArrayLength) len = dArrayLength;
        for (n = offset; n < len && ((INTFLTINT[]) hDataObject)[n] != null; n++)
          ((INTFLTINT[]) hDataObject)[n].fval = (float) fromVal;
        break;
      case TFormat.CF_FLTFLTINT:
        if ((len = ((FLTFLTINT[]) hDataObject).length) > dArrayLength) len = dArrayLength;
        for (n = offset; n < len && ((FLTFLTINT[]) hDataObject)[n] != null; n++)
          ((FLTFLTINT[]) hDataObject)[n].f1val = (float) fromVal;
        break;
      case TFormat.CF_FLTINTINT:
        if ((len = ((FLTINTINT[]) hDataObject).length) > dArrayLength) len = dArrayLength;
        for (n = offset; n < len && ((FLTINTINT[]) hDataObject)[n] != null; n++)
          ((FLTINTINT[]) hDataObject)[n].f1val = (float) fromVal;
        break;
      case TFormat.CF_INTINTINT:
        if ((len = ((INTINTINT[]) hDataObject).length) > dArrayLength) len = dArrayLength;
        for (n = offset; n < len && ((INTINTINT[]) hDataObject)[n] != null; n++)
          ((INTINTINT[]) hDataObject)[n].i1val = (int) fromVal;
        break;
      case TFormat.CF_FLTINTFLTINT:
        if ((len = ((FLTINTFLTINT[]) hDataObject).length) > dArrayLength) len = dArrayLength;
        for (n = offset; n < len && ((FLTINTFLTINT[]) hDataObject)[n] != null; n++)
          ((FLTINTFLTINT[]) hDataObject)[n].f1val = (float) fromVal;
        break;
      case TFormat.CF_ADDRESS:
        if ((len = ((INTINTINTINT[]) hDataObject).length) > dArrayLength) len = dArrayLength;
        for (n = offset; n < len && ((INTINTINTINT[]) hDataObject)[n] != null; n++)
          ((INTINTINTINT[]) hDataObject)[n].i1val = (int) fromVal;
        break;
      case TFormat.CF_FILTER:
        if ((len = ((INTFLTFLTFLT[]) hDataObject).length) > dArrayLength) len = dArrayLength;
        for (n = offset; n < len && ((INTFLTFLTFLT[]) hDataObject)[n] != null; n++)
          ((INTFLTFLTFLT[]) hDataObject)[n].f1val = (float) fromVal;
        break;
      case TFormat.CF_FWINDOW:
        if ((len = ((INTINTFLTFLT[]) hDataObject).length) > dArrayLength) len = dArrayLength;
        for (n = offset; n < len && ((INTINTFLTFLT[]) hDataObject)[n] != null; n++)
          ((INTINTFLTFLT[]) hDataObject)[n].f1val = (float) fromVal;
        break;
      case TFormat.CF_USTRING:
        if ((len = ((USTRING[]) hDataObject).length) > dArrayLength) len = dArrayLength;
        for (n = offset; n < len && ((USTRING[]) hDataObject)[n] != null; n++)
          ((USTRING[]) hDataObject)[n].f1val = (float) fromVal;
        break;
      case TFormat.CF_UNAME:
        if ((len = ((UNAME[]) hDataObject).length) > dArrayLength) len = dArrayLength;
        for (n = offset; n < len && ((UNAME[]) hDataObject)[n] != null; n++)
          ((UNAME[]) hDataObject)[n].f1val = (float) fromVal;
        break;
      case TFormat.CF_SPECTRUM:
        for (n = offset; n < dArrayLength && ((SPECTRUM) hDataObject) != null; n++)
          ((SPECTRUM) hDataObject).d_spect_array[n] = (float) fromVal;
        break;
      case TFormat.CF_GSPECTRUM:
        for (n = offset; n < dArrayLength && ((GSPECTRUM) hDataObject) != null; n++)
          ((GSPECTRUM) hDataObject).d_spect_array[n] = (float) fromVal;
        break;
    }
    return n - offset;
  }
  private int copyContents(TDataType fromData, int size, int offset)
  {
    if (fromData == null) return TErrorList.invalid_parameter;
    if (dFormat != fromData.dFormat) return TErrorList.illegal_format;
    int cc = 0, n = 0, len = Math.min(size,fromData.dArrayLength-offset);
    try 
    {
      if (hDataObject instanceof TCompoundDataObject[])
      {
        for (n=0; n<len; n++) 
          ((TCompoundDataObject[])hDataObject)[n].copy(((TCompoundDataObject[])fromData.hDataObject)[n+offset]);
        dCompletionLength = Math.min(n,fromData.dCompletionLength);
      }
    } 
    catch (Exception x) 
    { 
      MsgLog.log("TDataType.copyContents", x.getMessage(), TErrorList.code_failure, x, 1);
      cc = TErrorList.code_failure;
    }
    return cc;
  }
  public int dataCopy(TDataType fromData)
  {
    if (fromData == null) return 0;
    return dataCopy(fromData, 0);
  }
  public int dataCopy(TDataType fromData, int offset)
  {
    return dataCopy(fromData,offset,false);
  }
  private boolean isAllowedBooleanCopy(int fromFormat, int toFormat)
  {
    switch (fromFormat)
    {
      case TFormat.CF_BOOLEAN:
        if (toFormat == fromFormat) return true;
        if (toFormat == TFormat.CF_INT32 || toFormat == TFormat.CF_UINT32) return true;
        break;
      case TFormat.CF_INT32:
      case TFormat.CF_UINT32:
        if (toFormat == fromFormat) return true;
        if (toFormat == TFormat.CF_BOOLEAN) return true;
        break;
    }
    return false;
  }
  public int dataCopy(TDataType fromData, int offset,boolean isMca)
  { 
    boolean assignBuffers = false;
    if (dFormat != fromData.dFormat)
    {
      if (dFormat == TFormat.CF_DEFAULT) synchronized (fromData)
      { // just for jDDD ...
        dFormat = fromData.dFormat;
        if (isMca)
        {
          dArrayLength = 1;
          dCompletionLength = 1;            
        }
        else
        {
          dArrayLength = fromData.dArrayLength;
          dCompletionLength = fromData.dCompletionLength;
        }
        if (hDataObject == null && hStructObject == null) 
          setDataObject(dArrayLength, dFormat);
        assignBuffers = true;
      }
      else if (!isAllowedBooleanCopy(fromData.dFormat,dFormat))
      { // no can do -> jump out
         return 0;
      }
    }
    if (fromData.hasBeenUpdated && !hasBeenUpdated)
    {
      assignBuffers = true;
    }
    if (assignBuffers)
    {
      if (dBuffers == null) 
      {
        if (!isMca) 
        {
          dBuffers = fromData.dBuffers;
        }
        else 
        {
          dBuffers = new ByteArrayOutputStream[1];
          dBuffers[0] = new ByteArrayOutputStream(dBufferSegSize);
        }
      }
      if (dBuffersReady == null) dBuffersReady = fromData.dBuffersReady;
      hasBeenUpdated = true;
    }
    numblks = fromData.numblks;
    int n = 0;
    int size = dArrayLength < fromData.dArrayLength ? dArrayLength : fromData.dArrayLength;
    if (size > 1)
    {
      try
      { // there are one or two formats where hDataObject isn't an array
        int mx = Array.getLength(hDataObject);
        if (size > mx) size = mx;
        mx = Array.getLength(fromData.hDataObject) - offset;
        if (size > mx) size = mx;
      }
      catch (Exception ignore) {} 
    }
    // copy the innards ...
    dTimestamp = fromData.dTimestamp;
    timestamp = fromData.timestamp;
    timestampMSEC = fromData.timestampMSEC;
    timestampUSEC = fromData.timestampUSEC;
    sysDataStamp = fromData.sysDataStamp;
    usrDataStamp = fromData.usrDataStamp;
    copy(fromData, offset);
    if (fromData.hDataObject == null || hDataObject == null) return 0;
    // copy into the referenced object ...
//    if (hDataObject instanceof TCompoundDataObject[])
//    {
//      System.arraycopy((TCompoundDataObject[]) fromData.hDataObject, offset,
//          (TCompoundDataObject[]) hDataObject, 0, size);
//      n = size;
//    }
    // TODO: catch SPECTRUM and any other special cases ?
    // TODO: LNGINT and INTINT both have the same format (also TDS and
    // INTFLTINT)
    switch (dFormat)
    {
      default:
        if (hDataObject instanceof TCompoundDataObject[])
        {
          n = size;
          copyContents(fromData,size,offset);
        }
        break;        
      case TFormat.CF_DEFAULT:
      case TFormat.CF_NULL:
        break;
      case TFormat.CF_TEXT:
        if (hDataObject instanceof char[])
        {
          System.arraycopy((char[]) fromData.hDataObject, offset, (char[]) hDataObject, 0, size);
          n = size;
          break;
        }
        else if (hDataObject instanceof StringBuffer)
        {
          String s = ((StringBuffer) fromData.hDataObject).substring(0);
          n = s.length();
          ((StringBuffer) hDataObject).ensureCapacity(n);
          ((StringBuffer) hDataObject).insert(0, s);
          break;
        }
        break; // Karol 3.12.09  
      case TFormat.CF_STRING: // this just copies the references ...
        for (int i = 0; i < size; i++)
          ((String[]) hDataObject)[i] = ((String[]) fromData.hDataObject)[i];
        break;
      case TFormat.CF_KEYVALUE: // this just copies the references ...
        for (int i = 0; i < size; i++)
          ((KEYVALUE[]) hDataObject)[i] = ((KEYVALUE[]) fromData.hDataObject)[i];
        break;
      case TFormat.CF_BITFIELD8:
        if (this.dField == null || this.dField.length() == 0) this.dField = fromData.dField;
        this.dBitField = fromData.dBitField;
        if (fromData.hDataObject instanceof int[])
        {
          int[] ivals = (int[])fromData.hDataObject;
          byte[] bvals = (byte[]) hDataObject;
          for (int i=0; i<ivals.length; i++) bvals[i] = (byte)ivals[i];
        }
        break;
      case TFormat.CF_STRUCT:
      case TFormat.CF_BYTE:
        System.arraycopy((byte[]) fromData.hDataObject, offset, (byte[]) hDataObject, 0, size);
        n = size;
        break;
      case TFormat.CF_BITFIELD16: // can be coerced from an int !
        if (this.dField == null || this.dField.length() == 0) this.dField = fromData.dField;
        this.dBitField = fromData.dBitField;
        if (fromData.hDataObject instanceof int[])
        {
          if (hDataObject instanceof short[])
          {
            short[] vals = (short[]) hDataObject;
            for (int i=0; i<vals.length; i++) 
              vals[i] = (short)dBitField[i].getField(this.dField);
          }
          else if (hDataObject instanceof int[])
          {
            int[] vals = (int[]) hDataObject;
            for (int i=0; i<vals.length; i++) 
              vals[i] = dBitField[i].getField(this.dField);
          }
        }
        break;
      case TFormat.CF_UINT16:
      case TFormat.CF_INT16:
        System.arraycopy((short[]) fromData.hDataObject, offset, (short[]) hDataObject, 0, size);
        n = size;
        break;
      case TFormat.CF_BITFIELD32:
        if (this.dField == null || this.dField.length() == 0) this.dField = fromData.dField;
        this.dBitField = fromData.dBitField;
        if (fromData.hDataObject instanceof int[])
        {
          if (hDataObject instanceof short[])
          {
            short[] vals = (short[]) hDataObject;
            for (int i=0; i<vals.length; i++) 
              vals[i] = (short)dBitField[i].getField(this.dField);
          }
          else if (hDataObject instanceof int[])
          {
            int[] vals = (int[]) hDataObject;
            for (int i=0; i<vals.length; i++) 
              vals[i] = dBitField[i].getField(this.dField);
          }
        }
        break;
      case TFormat.CF_BOOLEAN:
        if (hDataObject instanceof boolean[])
        {
          boolean[] vals = null; 
          if (fromData.hDataObject instanceof boolean[])
          {
            vals = (boolean[]) fromData.hDataObject;
          }
          else
          {
            if (!(hDataObject instanceof int[])) break;
            vals = new boolean[fromData.dArrayLength];
            for (int i= 0; i<vals.length; i++) vals[i] = ((int[])fromData.hDataObject)[i] != 0;
          }
          System.arraycopy(vals, offset, (boolean[]) hDataObject, 0, size);
          n = size;
          break;          
        }
      case TFormat.CF_UINT32:
      case TFormat.CF_INT32:
        int[] ivals = null;
        if (fromData.hDataObject instanceof boolean[])
        {
          ivals = new int[fromData.dArrayLength];
          for (int i=0; i<ivals.length; i++) ivals[i] = ((boolean[])fromData.hDataObject)[i] ? 1 : 0;
        }
        else
        {
          ivals = (int[])fromData.hDataObject;
        }
        System.arraycopy(ivals, offset, (int[]) hDataObject, 0, size);
        n = size;
        break;
      case TFormat.CF_BITFIELD64:
        if (this.dField == null || this.dField.length() == 0) this.dField = fromData.dField;
        this.dBitField = fromData.dBitField;
        if (fromData.hDataObject instanceof long[])
        {
          if (hDataObject instanceof short[])
          {
            short[] vals = (short[]) hDataObject;
            for (int i=0; i<vals.length; i++) 
              vals[i] = (short)dBitField[i].getField(this.dField);
          }
          else if (hDataObject instanceof int[])
          {
            int[] vals = (int[]) hDataObject;
            for (int i=0; i<vals.length; i++) 
              vals[i] = dBitField[i].getField(this.dField);
          }
          else if (hDataObject instanceof long[])
          {
            long[] vals = (long[]) hDataObject;
            for (int i=0; i<vals.length; i++) 
              vals[i] = dBitField[i].getField(this.dField);
          }
        }
        break;
      case TFormat.CF_UINT64:
      case TFormat.CF_INT64:
        System.arraycopy((long[]) fromData.hDataObject, offset, (long[]) hDataObject, 0, size);
        n = size;
        break;
      case TFormat.CF_FLOAT:
        System.arraycopy((float[]) fromData.hDataObject, offset, (float[]) hDataObject, 0, size);
        n = size;
        break;
      case TFormat.CF_DOUBLE:
        System.arraycopy((double[]) fromData.hDataObject, offset, (double[]) hDataObject, 0, size);
        n = size;
        break;
      case TFormat.CF_SPECTRUM:
        ((SPECTRUM) hDataObject).s_inc = ((SPECTRUM) fromData.hDataObject).s_inc;
        ((SPECTRUM) hDataObject).s_start = ((SPECTRUM) fromData.hDataObject).s_start;
        ((SPECTRUM) hDataObject).tm = ((SPECTRUM) fromData.hDataObject).tm;
        ((SPECTRUM) hDataObject).status = ((SPECTRUM) fromData.hDataObject).status;
        ((SPECTRUM) hDataObject).comment = ((SPECTRUM) fromData.hDataObject).comment;
        System.arraycopy(((SPECTRUM) fromData.hDataObject).d_spect_array, offset,
            ((SPECTRUM) hDataObject).d_spect_array, 0, size);
        n = size;
        break;
      case TFormat.CF_GSPECTRUM:
        ((GSPECTRUM) hDataObject).id = ((GSPECTRUM) fromData.hDataObject).id;
        ((GSPECTRUM) hDataObject).s_inc = ((GSPECTRUM) fromData.hDataObject).s_inc;
        ((GSPECTRUM) hDataObject).s_start = ((GSPECTRUM) fromData.hDataObject).s_start;
        ((GSPECTRUM) hDataObject).tm = ((GSPECTRUM) fromData.hDataObject).tm;
        ((GSPECTRUM) hDataObject).status = ((GSPECTRUM) fromData.hDataObject).status;
        ((GSPECTRUM) hDataObject).groups = ((GSPECTRUM) fromData.hDataObject).groups;
        ((GSPECTRUM) hDataObject).group_size = ((GSPECTRUM) fromData.hDataObject).group_size;
        ((GSPECTRUM) hDataObject).group_inc = ((GSPECTRUM) fromData.hDataObject).group_inc;
        ((GSPECTRUM) hDataObject).stat_mask = ((GSPECTRUM) fromData.hDataObject).stat_mask;
        ((GSPECTRUM) hDataObject).comment = ((GSPECTRUM) fromData.hDataObject).comment;
        System.arraycopy(((GSPECTRUM) fromData.hDataObject).statistics, offset,
            ((GSPECTRUM) hDataObject).statistics, 0, 24);
        System.arraycopy(((GSPECTRUM) fromData.hDataObject).timing_param, offset,
            ((GSPECTRUM) hDataObject).timing_param, 0, 8);
        System.arraycopy(((GSPECTRUM) fromData.hDataObject).d_spect_array, offset,
            ((GSPECTRUM) hDataObject).d_spect_array, 0, size);
        n = size;
        break;
    }
    return n;
  }
  public Object getStructObject()
  {
    return hStructObject;
  }
  /**
   * Retrieves the returned data as an Object container. This method is
   *        useful for obtaining the unbound data when the data type is unknown
   *        at link time. A call to getFormat() can be made to ascertain the
   *        tine data type and hence the array nature of the object returned.
   */
  public Object getDataObject()
  {
    return hDataObject;
  }
  /**
   * Retrieves the returned data as a Vector. This method is useful for
   *        obtaining the unbound data when the data type is unknown at link
   *        time. A call to getFormat() can be made to ascertain the tine data
   *        type and hence the contents of the Vector.
   */
  public Vector<Object> getVector()
  {
    // TODO: clean up the individual type allocations and follow the NAME8
    // example
    // TODO: add the new formats
    // We still have to have the long switch on the format so that we know
    // what we have to create!
    int i = 0;
    if (dVector == null) dVector = new Vector<Object>();
    Vector<Object> v = dVector;
    v.clear();
    switch (dFormat)
    {
      case TFormat.CF_NULL:
        break;
      case TFormat.CF_NAME8:
      {
        NAME8 d[] = (NAME8[]) hDataObject;
        if (d == null)
        {
          d = new NAME8[dArrayLength];
          getData(d);
        }
        for (i = 0; i < d.length; i++)
          v.addElement(d[i]);
        break;
      }
      case TFormat.CF_NAME16:
      {
        NAME16 d[] = (NAME16[]) hDataObject;
        if (d == null)
        {
          d = new NAME16[dArrayLength];
          getData(d);
        }
        for (i = 0; i < d.length; i++)
          v.addElement(d[i]);
        break;
      }
      case TFormat.CF_NAME32:
      {
        NAME32 d[] = (NAME32[]) hDataObject;
        if (d == null)
        {
          d = new NAME32[dArrayLength];
          getData(d);
        }
        for (i = 0; i < d.length; i++)
          v.addElement(d[i]);
        break;
      }
      case TFormat.CF_NAME48:
      {
        NAME48 d[] = (NAME48[]) hDataObject;
        if (d == null)
        {
          d = new NAME48[dArrayLength];
          getData(d);
        }
        for (i = 0; i < d.length; i++)
          v.addElement(d[i]);
        break;
      }
      case TFormat.CF_NAME64:
      {
        NAME64 d[] = (NAME64[]) hDataObject;
        if (d == null)
        {
          d = new NAME64[dArrayLength];
          getData(d);
        }
        for (i = 0; i < d.length; i++)
          v.addElement(d[i]);
        break;
      }
      case TFormat.CF_NAME8I:
      {
        NAME8I d[] = (NAME8I[]) hDataObject;
        if (d == null)
        {
          d = new NAME8I[dArrayLength];
          getData(d);
        }
        for (i = 0; i < d.length; i++)
          v.addElement(d[i]);
        break;
      }
      case TFormat.CF_NAME16I:
      {
        NAME16I d[] = (NAME16I[]) hDataObject;
        if (d == null)
        {
          d = new NAME16I[dArrayLength];
          getData(d);
        }
        for (i = 0; i < d.length; i++)
          v.addElement(d[i]);
        break;
      }
      case TFormat.CF_NAME16FI:
      {
        NAME16FI d[] = (NAME16FI[]) hDataObject;
        if (d == null)
        {
          d = new NAME16FI[dArrayLength];
          getData(d);
        }
        for (i = 0; i < d.length; i++)
          v.addElement(d[i]);
        break;
      }
      case TFormat.CF_NAME16II:
      {
        NAME16II d[] = (NAME16II[]) hDataObject;
        if (d == null)
        {
          d = new NAME16II[dArrayLength];
          getData(d);
        }
        for (i = 0; i < d.length; i++)
          v.addElement(d[i]);
        break;
      }
      case TFormat.CF_NAME16DBLDBL:
      {
        NAME16DBLDBL d[] = (NAME16DBLDBL[]) hDataObject;
        if (d == null)
        {
          d = new NAME16DBLDBL[dArrayLength];
          getData(d);
        }
        for (i = 0; i < d.length; i++)
          v.addElement(d[i]);
        break;
      }
      case TFormat.CF_NAME32DBLDBL:
      {
        NAME32DBLDBL d[] = (NAME32DBLDBL[]) hDataObject;
        if (d == null)
        {
          d = new NAME32DBLDBL[dArrayLength];
          getData(d);
        }
        for (i = 0; i < d.length; i++)
          v.addElement(d[i]);
        break;
      }
      case TFormat.CF_NAME64DBL:
      {
        NAME64DBL d[] = (NAME64DBL[]) hDataObject;
        if (d == null)
        {
          d = new NAME64DBL[dArrayLength];
          getData(d);
        }
        for (i = 0; i < d.length; i++)
          v.addElement(d[i]);
        break;
      }
      case TFormat.CF_NAME64DBLDBL:
      {
        NAME64DBLDBL d[] = (NAME64DBLDBL[]) hDataObject;
        if (d == null)
        {
          d = new NAME64DBLDBL[dArrayLength];
          getData(d);
        }
        for (i = 0; i < d.length; i++)
          v.addElement(d[i]);
        break;
      }
      case TFormat.CF_NAME64DBLDBLDBL:
      {
        NAME64DBLDBLDBL d[] = (NAME64DBLDBLDBL[]) hDataObject;
        if (d == null)
        {
          d = new NAME64DBLDBLDBL[dArrayLength];
          getData(d);
        }
        for (i = 0; i < d.length; i++)
          v.addElement(d[i]);
        break;
      }
      case TFormat.CF_NAME32I:
      {
        NAME32I d[] = (NAME32I[]) hDataObject;
        if (d == null)
        {
          d = new NAME32I[dArrayLength];
          getData(d);
        }
        for (i = 0; i < d.length; i++)
          v.addElement(d[i]);
        break;
      }
      case TFormat.CF_NAME48I:
      {
        NAME48I d[] = (NAME48I[]) hDataObject;
        if (d == null)
        {
          d = new NAME48I[dArrayLength];
          getData(d);
        }
        for (i = 0; i < d.length; i++)
          v.addElement(d[i]);
        break;
      }
      case TFormat.CF_NAME64I:
      {
        NAME64I d[] = (NAME64I[]) hDataObject;
        if (d == null)
        {
          d = new NAME64I[dArrayLength];
          getData(d);
        }
        for (i = 0; i < d.length; i++)
          v.addElement(d[i]);
        break;
      }
      case TFormat.CF_STRING:
      {
        String d[] = (String[]) hDataObject;
        if (d == null)
        {
          d = new String[dArrayLength];
          getData(d);
        }
        for (i = 0; i < d.length; i++)
          v.addElement(new String(d[i]));
        break;
      }
      case TFormat.CF_KEYVALUE:
      {
        KEYVALUE d[] = (KEYVALUE[]) hDataObject;
        if (d == null)
        {
          d = new KEYVALUE[dArrayLength];
          getData(d);
        }
        for (i = 0; i < d.length; i++)
          v.addElement(d[i]);
        break;
      }
      case TFormat.CF_CHAR:
      {
        char d[] = (char[]) hDataObject;
        if (d == null)
        {
          d = new char[dArrayLength];
          getData(d);
        }
        v.addElement(new String(d));
        break;
      }        
      case TFormat.CF_BYTE:
      {
        byte d[] = (byte[]) hDataObject;
        if (d == null)
        {
          d = new byte[dArrayLength];
          getData(d);
        }
        for (i = 0; i < d.length; i++)
          v.addElement(d[i]);
        break;
      }
      case TFormat.CF_HISTORY:
      case TFormat.CF_STRUCT:
        // TODO: could check the tag and make a vector of the struct object
        break;
      case TFormat.CF_INT16:
      {
        short d[] = (short[]) hDataObject;
        if (d == null)
        {
          d = new short[dArrayLength];
          getData(d);
        }
        for (i = 0; i < d.length; i++)
          v.addElement(d[i]);
        break;
      }
      case TFormat.CF_INT32:
      {
        int d[] = (int[]) hDataObject;
        if (d == null)
        {
          d = new int[dArrayLength];
          getData(d);
        }
        for (i = 0; i < d.length; i++)
          v.addElement(d[i]);
        break;
      }
      case TFormat.CF_INT64:
      {
        long d[] = (long[]) hDataObject;
        if (d == null)
        {
          d = new long[dArrayLength];
          getData(d);
        }
        for (i = 0; i < d.length; i++)
          v.addElement(d[i]);
        break;
      }
      case TFormat.CF_FLOAT:
      {
        float d[] = (float[]) hDataObject;
        if (d == null)
        {
          d = new float[dArrayLength];
          getData(d);
        }
        for (i = 0; i < d.length; i++)
          v.addElement(d[i]);
        break;
      }
      case TFormat.CF_DOUBLE:
      {
        double d[] = (double[]) hDataObject;
        if (d == null)
        {
          d = new double[dArrayLength];
          getData(d);
        }
        for (i = 0; i < d.length; i++)
          v.addElement(d[i]);
        break;
      }
      case TFormat.CF_INTINT:
      {
        INTINT d[] = (INTINT[]) hDataObject;
        if (d == null)
        {
          d = new INTINT[dArrayLength];
          getData(d);
        }
        for (i = 0; i < d.length; i++)
          v.addElement(d[i]);
        break;
      }
      case TFormat.CF_FLTINT:
      {
        FLTINT d[] = (FLTINT[]) hDataObject;
        if (d == null)
        {
          d = new FLTINT[dArrayLength];
          getData(d);
        }
        for (i = 0; i < d.length; i++)
          v.addElement(d[i]);
        break;
      }
      case TFormat.CF_FLTFLT:
      {
        FLTFLT d[] = (FLTFLT[]) hDataObject;
        if (d == null)
        {
          d = new FLTFLT[dArrayLength];
          getData(d);
        }
        for (i = 0; i < d.length; i++)
          v.addElement(d[i]);
        break;
      }
      case TFormat.CF_DBLDBL:
      {
        DBLDBL d[] = (DBLDBL[]) hDataObject;
        if (d == null)
        {
          d = new DBLDBL[dArrayLength];
          getData(d);
        }
        for (i = 0; i < d.length; i++)
          v.addElement(d[i]);
        break;
      }
      case TFormat.CF_DBLTIME:
      {
        DBLTIME d[] = (DBLTIME[]) hDataObject;
        if (d == null)
        {
          d = new DBLTIME[dArrayLength];
          getData(d);
        }
        for (i = 0; i < d.length; i++)
          v.addElement(d[i]);
        break;
      }
      case TFormat.CF_DBLDBLDBL:
      {
        DBLDBLDBL d[] = (DBLDBLDBL[]) hDataObject;
        if (d == null)
        {
          d = new DBLDBLDBL[dArrayLength];
          getData(d);
        }
        for (i = 0; i < d.length; i++)
          v.addElement(d[i]);
        break;
      }
      case TFormat.CF_INTFLTINT:
      {
        INTFLTINT d[] = (INTFLTINT[]) hDataObject;
        if (d == null)
        {
          d = new INTFLTINT[dArrayLength];
          getData(d);
        }
        for (i = 0; i < d.length; i++)
          v.addElement(d[i]);
        break;
      }
      case TFormat.CF_FLTFLTINT:
      {
        FLTFLTINT d[] = (FLTFLTINT[]) hDataObject;
        if (d == null)
        {
          d = new FLTFLTINT[dArrayLength];
          getData(d);
        }
        for (i = 0; i < d.length; i++)
          v.addElement(d[i]);
        break;
      }
      case TFormat.CF_FLTINTINT:
      {
        FLTINTINT d[] = (FLTINTINT[]) hDataObject;
        if (d == null)
        {
          d = new FLTINTINT[dArrayLength];
          getData(d);
        }
        for (i = 0; i < d.length; i++)
          v.addElement(d[i]);
        break;
      }
      case TFormat.CF_INTINTINT:
      {
        INTINTINT d[] = (INTINTINT[]) hDataObject;
        if (d == null)
        {
          d = new INTINTINT[dArrayLength];
          getData(d);
        }
        for (i = 0; i < d.length; i++)
          v.addElement(d[i]);
        break;
      }
      case TFormat.CF_FLTINTFLTINT:
      {
        FLTINTFLTINT d[] = (FLTINTFLTINT[]) hDataObject;
        if (d == null)
        {
          d = new FLTINTFLTINT[dArrayLength];
          getData(d);
        }
        for (i = 0; i < d.length; i++)
          v.addElement(d[i]);
        break;
      }
      case TFormat.CF_ADDRESS:
      {
        INTINTINTINT d[] = (INTINTINTINT[]) hDataObject;
        if (d == null)
        {
          d = new INTINTINTINT[dArrayLength];
          getData(d);
        }
        for (i = 0; i < d.length; i++)
          v.addElement(d[i]);
        break;
      }
      case TFormat.CF_FILTER:
      {
        INTFLTFLTFLT d[] = (INTFLTFLTFLT[]) hDataObject;
        if (d == null)
        {
          d = new INTFLTFLTFLT[dArrayLength];
          getData(d);
        }
        for (i = 0; i < d.length; i++)
          v.addElement(d[i]);
        break;
      }
      case TFormat.CF_FWINDOW:
      {
        INTINTFLTFLT d[] = (INTINTFLTFLT[]) hDataObject;
        if (d == null)
        {
          d = new INTINTFLTFLT[dArrayLength];
          getData(d);
        }
        for (i = 0; i < d.length; i++)
          v.addElement(d[i]);
        break;
      }
      case TFormat.CF_USTRING:
      {
        USTRING d[] = (USTRING[]) hDataObject;
        ;
        if (d == null)
        {
          d = new USTRING[dArrayLength];
          getData(d);
        }
        for (i = 0; i < d.length; i++)
          v.addElement(d[i]);
        break;
      }
      case TFormat.CF_UNAME:
      {
        UNAME undata[] = (UNAME[]) hDataObject;
        ;
        if (undata == null)
        {
          undata = new UNAME[dArrayLength];
          getData(undata);
        }
        for (i = 0; i < undata.length; i++)
          v.addElement(undata[i]);
        break;
      }
    }
    dCompletionLength = i;
    return v;
  }
  private void appendElementName(StringBuffer sb, int index)
  {
    if (arrayElementNames != null && index < arrayElementNames.length)
    {
      if (arrayElementNames[index] != null)
        sb.append(arrayElementNames[index] + ": ");
      else
        sb.append("#" + index + ": ");
    }
  }
  public static String pushToCsvString(TCompoundDataObject[] cdo,int index)
  {
    if (cdo == null) return "(null)";
    int len = cdo.length;
    if (len == 0) return "(empty)";
    if (index < 0 || index >= len) return "(out of range)";
    StringBuffer sb = new StringBuffer(cdo[0].getSizeInBytes() + 64);
    sb.append(cdo[index].getCsvString());
    return sb.toString();
  }
  public static String pushToCsvString(String[] sa,int index)
  {
    if (sa == null) return "(null)";
    int len = sa.length;
    if (len == 0) return "(empty)";
    if (index < 0 || index >= len) return "(out of range)";
    String tname = sa[index];
    if (tname == null) return "";
    tname = tname.trim();
    if (tname.contains(",") || tname.contains(" ") ||
        tname.contains("\t") || tname.contains("\n"))
    {
      return "\""+tname+"\"";
    }
    return tname;
  }
  public static String pushToCsvString(Object obj,int index)
  {
    if (obj == null) return "(null)";
    int len = 0;
    try 
    { 
      len = Array.getLength(obj); 
      if (len == 0) return "(empty)";
      if (obj instanceof byte[]) return ""+Array.getByte(obj, index);
      if (obj instanceof short[]) return ""+Array.getShort(obj, index);
      if (obj instanceof int[]) return ""+Array.getInt(obj, index);
      if (obj instanceof long[]) return ""+Array.getLong(obj, index);
      if (obj instanceof boolean[]) return ""+Array.getBoolean(obj, index);
      if (obj instanceof double[]) return ""+Array.getDouble(obj, index);
      if (obj instanceof float[]) return ""+Array.getFloat(obj, index);
      if (obj instanceof TCompoundDataObject[])
      {
        return pushToCsvString((TCompoundDataObject[])obj,index);
      }
      if (obj instanceof String[])
      {
        return pushToCsvString((String[])obj,index);
      }
      if (obj instanceof char[]) 
      {
        String s = new String((char [])obj).trim();
        if (s.contains(",") || s.contains(" ") ||
            s.contains("\t") || s.contains("\n"))
        {
          return "\""+s+"\"";
        }
        return s;        
      }
    } 
    catch (ArrayIndexOutOfBoundsException e)
    {
      return "(out of range)";
    }
    catch (IllegalArgumentException e)
    {
      return "(datatype not an array?)";
    }
    catch (Exception e) 
    { 
      MsgLog.log("TDataType.pushToCsvString", "cannot parse to string",TErrorList.code_failure,e,0);
      return "(parse error)";
    }
    return "(data type not parse-able)";
  }
  private String pushToString(TCompoundDataObject[] cdo)
  {
  	return pushToString(cdo, 0);
  }
  private String pushToString(TCompoundDataObject[] cdo,int maximumLength)
  {
    if (dCompletionLength == 0) return "";
    int len = dCompletionLength;
    if (maximumLength > 0) len = Math.min(maximumLength, dCompletionLength);
    StringBuffer sb = new StringBuffer(cdo[0].getSizeInBytes() * len);
    len = Math.min(len,cdo.length);
    for (int i = 0; i < len; i++)
    {
      if (cdo[i] == null) break;
      appendElementName(sb, i);
      cdo[i].setFieldSeparator(fieldDelimiter);
      sb.append(cdo[i].toString());
      if (i < len -1) sb.append(arrayDelimiter);
    }
    return sb.toString();
  }
  public String getObjectInfo(boolean isInput)
  {
    if (dArrayLength == 0 || dFormat == TFormat.CF_NULL) return "(null)";
    String tag = " element";
    if (dArrayLength > 1) tag += "s";
    String info = ""+dArrayLength+" "+TFormat.toString(dFormat)+tag;
    if (dTag.length() > 0) info += " (tag = "+dTag+")";
    if (isInput)
    {
      info += "\n"+"input data value(s):";
      info += "\n\t"+toString();      
    }
    else
    {
      info += "\n"+"last update:";
      info += "\n\t"+dCompletionLength+tag+" returned";
      info += "\n\t"+"time stamp: "+TDataTime.toString(dTimestamp)+"; utc: "+dTimestamp;
      info += "\n\t"+"system stamp: "+sysDataStamp;
      info += "\n\t"+"user stamp: "+usrDataStamp;
      info += "\n\t"+"has been updated: "+hasBeenUpdated;
      info += "\n\t"+"update reason: "+TTransferCodes.toString(xferReason);
    }
    return info;
  }
  private String getRealTrimmedString(char[] cb)
  {
    if (cb == null || cb.length == 0) return "";
    int idx0, idx1=cb.length-1;
    for (idx0=0; idx0<cb.length && cb[idx0] == 0; idx0++);
    for (idx1=idx0; idx1<cb.length && cb[idx1] != 0; idx1++);
    return new String(cb,idx0,idx1-idx0);
  }
  private String fieldDelimiter = ", ";
  public void setFieldDelimiter(String value)
  {
    if (value != null) fieldDelimiter = value;    
  }
  /**
   * Outputs the returned data as a String. This method is useful for
   *        displaying the unbound data when the data type is unknown at link
   *        time.
   */
  public String toString()
  {
  	return toString(0);
  }
  public String toString(int maximumLength)
  {
    try
    {
      StringBuffer sb = new StringBuffer();
      int dlen = dCompletionLength;
      if (dlen == 0) dlen = dArrayLength;
      if (maximumLength > 0) dlen = maximumLength;
      // TODO: clean up, add the new formats
      boolean isUnsigned = false;
      switch (dFormat)
      {
        case TFormat.CF_NULL:
          return new String("");
        case TFormat.CF_XML:
        case TFormat.CF_TEXT:
        {
          if (hDataObject instanceof char[]) return getRealTrimmedString((char[]) hDataObject);
          if (hDataObject instanceof StringBuffer) return ((StringBuffer) hDataObject).toString().trim();
          char[] t = new char[dArrayLength];
          getData(t);
          return getRealTrimmedString(t);
        }
        case TFormat.CF_KEYVALUE:
        {
          KEYVALUE[] d = (KEYVALUE[]) hDataObject;
          if (d == null)
          {
            d = new KEYVALUE[dlen];
            getData(d);
          }
          String ds = new String();
          for (int i = 0; i < dlen; i++)
            ds += d[i].getString() + arrayDelimiter;
          return ds;
        }
        case TFormat.CF_STRING:
        {
          String[] d = (String[]) hDataObject;
          if (d == null)
          {
            d = new String[dlen];
            getData(d);
          }
          String ds = new String();
          for (int i = 0; i < dlen; i++)
            ds += d[i] + arrayDelimiter;
          return ds;
        }
        case TFormat.CF_USTRING:
        {
          USTRING d[] = (USTRING[]) hDataObject;
          if (d == null)
          {
            d = new USTRING[dArrayLength];
            getData(d);
          }
          return pushToString(d, maximumLength);
        }
        case TFormat.CF_UNAME:
        {
          UNAME d[] = (UNAME[]) hDataObject;
          if (d == null)
          {
            d = new UNAME[dArrayLength];
            getData(d);
          }
          return pushToString(d, maximumLength);
        }
        case TFormat.CF_INTINTINT:
        {
          INTINTINT d[] = (INTINTINT[]) hDataObject;
          if (d == null)
          {
            d = new INTINTINT[dArrayLength];
            getData(d);
          }
          return pushToString(d, maximumLength);
        }
        case TFormat.CF_ADDRESS:
        {
          INTINTINTINT d[] = (INTINTINTINT[]) hDataObject;
          if (d == null)
          {
            d = new INTINTINTINT[dArrayLength];
            getData(d);
          }
          return pushToString(d, maximumLength);
        }
        case TFormat.CF_WINDOW:
        {
          WINDOW d[] = (WINDOW[]) hDataObject;
          if (d == null)
          {
            d = new WINDOW[dArrayLength];
            getData(d);
          }
          return pushToString(d, maximumLength);
        }          
        case TFormat.CF_FILTER:
        {
          INTFLTFLTFLT d[] = (INTFLTFLTFLT[]) hDataObject;
          if (d == null)
          {
            d = new INTFLTFLTFLT[dArrayLength];
            getData(d);
          }
          return pushToString(d, maximumLength);
        }
        case TFormat.CF_FWINDOW:
        {
          INTINTFLTFLT d[] = (INTINTFLTFLT[]) hDataObject;
          if (d == null)
          {
            d = new INTINTFLTFLT[dArrayLength];
            getData(d);
          }
          return pushToString(d, maximumLength);
        }
        case TFormat.CF_FLTINTFLTINT:
        {
          FLTINTFLTINT d[] = (FLTINTFLTINT[]) hDataObject;
          if (d == null)
          {
            d = new FLTINTFLTINT[dArrayLength];
            getData(d);
          }
          return pushToString(d, maximumLength);
        }
        case TFormat.CF_INTFLTINT:
        {
          INTFLTINT d[] = (INTFLTINT[]) hDataObject;
          if (d == null)
          {
            d = new INTFLTINT[dArrayLength];
            getData(d);
          }
          return pushToString(d, maximumLength);
        }
        case TFormat.CF_FLTFLTINT:
        {
          FLTFLTINT d[] = (FLTFLTINT[]) hDataObject;
          if (d == null)
          {
            d = new FLTFLTINT[dArrayLength];
            getData(d);
          }
          return pushToString(d, maximumLength);
        }
        case TFormat.CF_FLTINTINT:
        {
          FLTINTINT d[] = (FLTINTINT[]) hDataObject;
          if (d == null)
          {
            d = new FLTINTINT[dArrayLength];
            getData(d);
          }
          return pushToString(d, maximumLength);
        }
        case TFormat.CF_NAME8:
        {
          NAME8 d[] = (NAME8[]) hDataObject;
          if (d == null)
          {
            d = new NAME8[dArrayLength];
            getData(d);
          }
          return pushToString(d, maximumLength);
        }
        case TFormat.CF_NAME16:
        {
          NAME16 d[] = (NAME16[]) hDataObject;
          if (d == null)
          {
            d = new NAME16[dArrayLength];
            getData(d);
          }
          return pushToString(d, maximumLength);
        }
        case TFormat.CF_NAME32:
        {
          NAME32 d[] = (NAME32[]) hDataObject;
          if (d == null)
          {
            d = new NAME32[dArrayLength];
            getData(d);
          }
          return pushToString(d, maximumLength);
        }
        case TFormat.CF_NAME48:
        {
          NAME48 d[] = (NAME48[]) hDataObject;
          if (d == null)
          {
            d = new NAME48[dArrayLength];
            getData(d);
          }
          getData(d);
          return pushToString(d, maximumLength);
        }
        case TFormat.CF_NAME64:
        {
          NAME64 d[] = (NAME64[]) hDataObject;
          if (d == null)
          {
            d = new NAME64[dArrayLength];
            getData(d);
          }
          return pushToString(d, maximumLength);
        }
        case TFormat.CF_NAME8I:
        {
          NAME8I d[] = (NAME8I[]) hDataObject;
          if (d == null)
          {
            d = new NAME8I[dArrayLength];
            getData(d);
          }
          return pushToString(d, maximumLength);
        }
        case TFormat.CF_NAME16I:
        {
          NAME16I d[] = (NAME16I[]) hDataObject;
          if (d == null)
          {
            d = new NAME16I[dArrayLength];
            getData(d);
          }
          return pushToString(d, maximumLength);
        }
        case TFormat.CF_NAME16FI:
        {
          NAME16FI d[] = (NAME16FI[]) hDataObject;
          if (d == null)
          {
            d = new NAME16FI[dArrayLength];
            getData(d);
          }
          return pushToString(d, maximumLength);
        }
        case TFormat.CF_NAME16II:
        {
          NAME16II d[] = (NAME16II[]) hDataObject;
          if (d == null)
          {
            d = new NAME16II[dArrayLength];
            getData(d);
          }
          return pushToString(d, maximumLength);
        }
        case TFormat.CF_NAME16DBLDBL:
        {
          NAME16DBLDBL d[] = (NAME16DBLDBL[]) hDataObject;
          if (d == null)
          {
            d = new NAME16DBLDBL[dArrayLength];
            getData(d);
          }
          return pushToString(d, maximumLength);
        }
        case TFormat.CF_NAME32DBLDBL:
        {
          NAME32DBLDBL d[] = (NAME32DBLDBL[]) hDataObject;
          if (d == null)
          {
            d = new NAME32DBLDBL[dArrayLength];
            getData(d);
          }
          return pushToString(d, maximumLength);
        }
        case TFormat.CF_NAME64DBL:
        {
          NAME64DBL d[] = (NAME64DBL[]) hDataObject;
          if (d == null)
          {
            d = new NAME64DBL[dArrayLength];
            getData(d);
          }
          return pushToString(d, maximumLength);
        }
        case TFormat.CF_NAME64TIME:
        {
          NAME64TIME d[] = (NAME64TIME[]) hDataObject;
          if (d == null)
          {
            d = new NAME64TIME[dArrayLength];
            getData(d);
          }
          return pushToString(d, maximumLength);
        }
        case TFormat.CF_NAME64DBLDBL:
        {
          NAME64DBLDBL d[] = (NAME64DBLDBL[]) hDataObject;
          if (d == null)
          {
            d = new NAME64DBLDBL[dArrayLength];
            getData(d);
          }
          return pushToString(d, maximumLength);
        }
        case TFormat.CF_NAME64DBLDBLDBL:
        {
          NAME64DBLDBLDBL d[] = (NAME64DBLDBLDBL[]) hDataObject;
          if (d == null)
          {
            d = new NAME64DBLDBLDBL[dArrayLength];
            getData(d);
          }
          return pushToString(d, maximumLength);
        }
        case TFormat.CF_NAME32I:
        {
          NAME32I d[] = (NAME32I[]) hDataObject;
          if (d == null)
          {
            d = new NAME32I[dArrayLength];
            getData(d);
          }
          return pushToString(d, maximumLength);
        }
        case TFormat.CF_NAME48I:
        {
          NAME48I d[] = (NAME48I[]) hDataObject;
          if (d == null)
          {
            d = new NAME48I[dArrayLength];
            getData(d);
          }
          return pushToString(d, maximumLength);
        }
        case TFormat.CF_NAME64I:
        {
          NAME64I d[] = (NAME64I[]) hDataObject;
          if (d == null)
          {
            d = new NAME64I[dArrayLength];
            getData(d);
          }
          return pushToString(d, maximumLength);
        }
        case TFormat.CF_BYTE:
        {
          String ds;        
          byte d[] = (byte[]) hDataObject;
          if (d == null)
          {
            d = new byte[dArrayLength];
            getData(d);
          }
          int len = Math.min(dlen, d.length);
          for (int i = 0; i < len; i++)
          {
            ds = new String("" + d[i]);
            appendElementName(sb, i);
            sb.append(ds);
            if (i < len-1) sb.append(arrayDelimiter);
          }
          return sb.toString();
        }
        case TFormat.CF_HISTORY:
        {
          HISTORY d[] = (HISTORY[]) hDataObject;
          if (d == null)
          {
            d = new HISTORY[dArrayLength];
            getData(d);
          }
          return pushToString(d, maximumLength);
        }
        case TFormat.CF_STRUCT:
        {
          if (hStructObject != null && hStructObject instanceof TTaggedStructure)
          {
            TTaggedStructure tts = (TTaggedStructure) hStructObject;
            tts.setFieldSeparator(fieldDelimiter);
            tts.setArraySeparator(arrayDelimiter);
            return tts.toString(true);
          }
          if (hStructObject instanceof TTaggedStructure[])
          {
            int len = ((TTaggedStructure[]) hStructObject).length;
            for (int i = 0; i < len && i<dCompletionLength; i++)
            {
              TTaggedStructure tts = ((TTaggedStructure[]) hStructObject)[i];
              tts.setFieldSeparator(fieldDelimiter);
              //tts.setArraySeparator(arrayDelimiter); <- this is already an array of structs: leave this be
              sb.append(tts.toString(true));
              if (i < len-1) sb.append(arrayDelimiter);
            }
            return sb.toString();
          }
          if (hDataObject != null)
          {
            int len = TStructRegistry.getSizeInBytes(dTag,dTagKey) * dCompletionLength;
            TStructDescription tds = dTag != null ? TStructRegistry.get(dTag,structKey) : null;
            if (tds != null)
            {
              if (tds.hasExtendedSpace()) len = ((byte[])hDataObject).length; 
              if (dField != null && dField.length() > 0)
              {
              }
              return TStructRegistry.toString(dTag, dTagKey, (byte[])hDataObject, len);
            }
          }
          return new String("user defined structure");
        }
        case TFormat.CF_UINT16:
          isUnsigned = true;
        case TFormat.CF_INT16:
        {
          short d[] = (short[]) hDataObject;
          if (d == null)
          {
            d = new short[dArrayLength];
            getData(d);
          }
          String ds;
          int len = Math.min(dlen, d.length);
          for (int i = 0; i < len; i++)
          {
            ds = isUnsigned ? new String(""+UnsignedShort(d[i])) :  new String("" + d[i]);
            appendElementName(sb, i);
            sb.append(ds);
            if (i < len -1) sb.append(arrayDelimiter);
          }
          return sb.toString();
        }
        case TFormat.CF_BITFIELD8:
        case TFormat.CF_BITFIELD16:
        case TFormat.CF_BITFIELD32:
        case TFormat.CF_BITFIELD64:
        {
          TBitfield d[] = null;
          if (hDataObject instanceof TBitfield[])
          {
            d = (TBitfield[])hDataObject;
          }
          else if (dBitField != null)
          {
            d = dBitField;
          }
          else
          {
            d = new TBitfield[dArrayLength];
            TBitfield bf = TBitfieldRegistry.getBitfield(dTag);
            if (bf != null)
            {
              for (int i=0; i<d.length; i++) 
              { 
                d[i] = new TBitfield(bf);
              }
            }
            getData(d);          
          }
          String ds;
          int len = Math.min(dlen, d.length);
          for (int i = 0; i < len; i++)
          {
            ds = new String("" + d[i].toString());
            appendElementName(sb, i);
            sb.append(ds);
            if (i < len -1) sb.append(arrayDelimiter);
          }
          return sb.toString();
        }
        case TFormat.CF_UINT32:
          isUnsigned = true;
        case TFormat.CF_INT32:
        {
          int d[] = (int[]) hDataObject;
          if (d == null)
          {
            d = new int[dArrayLength];
            getData(d);
          }
          String ds;
          int len = Math.min(dlen, d.length);
          for (int i = 0; i < len; i++)
          {
            ds = isUnsigned ? new String(""+UnsignedInt(d[i])) :  new String("" + d[i]);
            appendElementName(sb, i);
            sb.append(ds);
            if (i < len -1) sb.append(arrayDelimiter);
          }
          return sb.toString();
        }
        case TFormat.CF_UINT64:
        case TFormat.CF_INT64:
        {
          long d[] = (long[]) hDataObject;
          if (d == null)
          {
            d = new long[dArrayLength];
            getData(d);
          }
          String ds;
          int len = Math.min(dlen, d.length);
          for (int i = 0; i < len; i++)
          {
            ds = new String("" + d[i]);
            appendElementName(sb, i);
            sb.append(ds);
            if (i < len -1) sb.append(arrayDelimiter);
          }
          return sb.toString();
        }
        case TFormat.CF_BOOLEAN:
        {
          if (hDataObject instanceof boolean[])
          {
            int len = Math.min(dlen, ((boolean[])hDataObject).length);
            for (int i = 0; i < len; i++)
            {
              appendElementName(sb, i);
              sb.append(((boolean[])hDataObject)[i]);
              if (i < len -1) sb.append(arrayDelimiter);
            }          
          }
          else if (hDataObject instanceof int[])
          {
            int d[] = (int[]) hDataObject;
            if (d == null)
            {
              d = new int[dArrayLength];
              getData(d);
            }
            int len = Math.min(dlen, d.length);
            for (int i = 0; i < len; i++)
            {
              appendElementName(sb, i);
              sb.append(d[i] == 0 ? "false" : "true");
              if (i < len -1) sb.append(arrayDelimiter);
            }
          }
          return sb.toString();
        }
        case TFormat.CF_INTINT:
        {
          INTINT d[] = (INTINT[]) hDataObject;
          if (d == null)
          {
            d = new INTINT[dArrayLength];
            getData(d);
          }
          return pushToString(d);
        }
        case TFormat.CF_FLOAT:
        {
          float d[] = (float[]) hDataObject;
          if (d == null)
          {
            d = new float[dArrayLength];
            getData(d);
          }
          int len = Math.min(dlen, d.length);
          for (int i = 0; i < len; i++)
          {
            appendElementName(sb, i);
            sb.append(d[i]);
            if (i < len -1) sb.append(arrayDelimiter);
          }
          return sb.toString();
        }
        case TFormat.CF_FLTINT:
        {
          FLTINT d[] = (FLTINT[]) hDataObject;
          if (d == null)
          {
            d = new FLTINT[dArrayLength];
            getData(d);
          }
          return pushToString(d);
        }
        case TFormat.CF_FLTFLT:
        {
          FLTFLT d[] = (FLTFLT[]) hDataObject;
          if (d == null)
          {
            d = new FLTFLT[dArrayLength];
            getData(d);
          }
          return pushToString(d);
        }
        case TFormat.CF_SPECTRUM:
        {
          SPECTRUM d = (SPECTRUM) hDataObject;
          if (d == null)
          {
            d = new SPECTRUM();
            getData(d);
          }
          return d.toString();
        }
        case TFormat.CF_GSPECTRUM:
        {
          GSPECTRUM d = (GSPECTRUM) hDataObject;
          if (d == null)
          {
            d = new GSPECTRUM();
            getData(d);
          }
          return d.toString();
        }
        case TFormat.CF_MDA:
        {
          MDA d = (MDA) hDataObject;
          if (d == null)
          {
            return "not instantiated!";
          }
          return d.toString();
        }
        case TFormat.CF_ASPECTRUM:
        {
          ASPECTRUM[] d = (ASPECTRUM[]) hDataObject;
          if (d == null)
          {
            d = new ASPECTRUM[dArrayLength];
            getData(d);
          }
          return pushToString(d);
        }
        case TFormat.CF_DOUBLE:
        {
          double d[] = (double[]) hDataObject;
          if (d == null)
          {
            d = new double[dArrayLength];
            getData(d);
          }
          int len = Math.min(dlen, d.length);
          for (int i = 0; i < len; i++)
          {
            appendElementName(sb, i);
            sb.append(d[i]);
            if (i < len -1) sb.append(arrayDelimiter);
          }
          return sb.toString();
        }
        case TFormat.CF_DBLDBL:
        {
          DBLDBL d[] = (DBLDBL[]) hDataObject;
          if (d == null)
          {
            d = new DBLDBL[dArrayLength];
            getData(d);
          }
          return pushToString(d);
        }
        case TFormat.CF_DBLTIME:
        {
          DBLTIME d[] = (DBLTIME[]) hDataObject;
          if (d == null)
          {
            d = new DBLTIME[dArrayLength];
            getData(d);
          }
          return pushToString(d);
        }
        case TFormat.CF_DBLDBLDBL:
        {
          DBLDBLDBL d[] = (DBLDBLDBL[]) hDataObject;
          if (d == null)
          {
            d = new DBLDBLDBL[dArrayLength];
            getData(d);
          }
          return pushToString(d);
        }
      }
    }
    catch (Exception e)
    { // an exception (if any) is most liable to be an illegal cast, but whatever it is, catch it 
      MsgLog.log("TDataType.toString", TFormat.toString(dFormat)+" "+e.getMessage(),TErrorList.code_failure,e,0);      
    }
    return new String("");
  }
  /**
   * An array of chars (to be interpretted as a text string)
   * @param data
   *          is a reference to the char[] array used to create and bind to the
   *          TDataType object
   */
  public TDataType(char[] data)
  {
    initTDataType(data, data.length);
  }
  /**
   * A subset of an array of chars (to be interpretted as a text string)
   * @param data
   *          is a reference to the char[] array used to create and bind to the
   *          TDataType object
   * @param length
   *          is the subset array length to use.
   */
  public TDataType(char[] data, int length)
  {
    initTDataType(data, length);
  }
  private void initTDataType(char[] data, int length)
  {
    if (data == null || data.length == 0 || length < 1) return;
    try
    {
      if (length > data.length) length = data.length;
      dFormat = TFormat.CF_TEXT;
      dArrayLength = length;
      dSizeInBytes = length;
      initCounters();
      dBuffers = new ByteArrayOutputStream[numblks];
      dBuffersReady = new boolean[numblks];
      for (int n = 0, i = 0; n < numblks; n++)
      {
        dBuffers[n] = new ByteArrayOutputStream(dBufferSegSize);
        DataOutputStream ds = new DataOutputStream(dBuffers[n]);
        byte bdata;
        for (; i < dArrayLength && i < (n + 1) * mtu; i++)
        {
          bdata = (byte) data[i];
          ds.writeByte(bdata);
        }
        ds.close();
      }
      hDataObject = data;
    }
    catch (IOException e)
    {
      e.printStackTrace();
      MsgLog.log("initTDataType", e.getMessage(), TErrorList.code_failure, e, 0);
    }
  }
  /**
   * A text string buffer
   * @param data
   *          is a reference to the StringBuffer used to create and bind to the
   *          TDataType object
   */
  public TDataType(StringBuffer data)
  {
    initTDataType(data, data.capacity());
  }
  /**
   * A subset of a text string buffer
   * @param data
   *          is a reference to the StringBuffer used to create and bind to the
   *          TDataType object
   * @param length
   *          is the subset length to use.
   */
  public TDataType(StringBuffer data, int length)
  {
    initTDataType(data, length);
  }
  private void initTDataType(StringBuffer data, int length)
  {
    if (data == null || data.capacity() == 0 || length < 1) return;
    try
    {
      if (length > data.capacity()) length = data.capacity();
      dFormat = TFormat.CF_TEXT;
      dArrayLength = length;
      dSizeInBytes = dArrayLength;
      byte[] bdata = data.toString().getBytes();
      initCounters();
      dBuffers = new ByteArrayOutputStream[numblks];
      dBuffersReady = new boolean[numblks];
      for (int n = 0, i = 0; n < numblks; n++)
      {
        dBuffers[n] = new ByteArrayOutputStream(dBufferSegSize);
        DataOutputStream ds = new DataOutputStream(dBuffers[n]);
        for (; i < bdata.length && i < (n + 1) * mtu; i++)
          ds.writeByte(bdata[i]);
        ds.close();
      }
      hDataObject = data;
    }
    catch (IOException e)
    {
      e.printStackTrace();
      MsgLog.log("initTDataType", e.getMessage(), TErrorList.code_failure, e, 0);
    }
  }
  /**
   * A subset of an array of Strings. This corresponds to tine format
   *        CF_STRING
   * @param data
   *          is a reference to the String array used to create and bind to the
   *          TDataType object
   */
  public TDataType(String[] data)
  {
    initTDataType(data, data.length);
  }
  /**
   * A subset of an array of Strings. This corresponds to tine format
   *        CF_STRING
   * @param data
   *          is a reference to the String array used to create and bind to the
   *          TDataType object
   * @param length
   *          is the subset array length to use.
   */
  public TDataType(String[] data, int length)
  {
    initTDataType(data, length);
  }
  private void initTDataType(String[] data, int length)
  {
    if (data == null || data.length == 0 || length < 1) return;
    try
    {
      if (length > data.length) length = data.length;
      dFormat = TFormat.CF_STRING;
      dArrayLength = length;
      dSizeInBytes = 0;
      for (int i = 0; i < length; i++)
      { // first see how much string space we need (going out)
        dSizeInBytes += (data[i] != null) ? data[i].length() : 0;
        dSizeInBytes++; // add terminating 0
      }
      byte[] bdata = getStringArrayBytes(data);
      initCounters();
      dBuffers = new ByteArrayOutputStream[numblks];
      dBuffersReady = new boolean[numblks];
      for (int n = 0, i = 0; n < numblks; n++)
      {
        dBuffers[n] = new ByteArrayOutputStream(dBufferSegSize);
        DataOutputStream ds = new DataOutputStream(dBuffers[n]);
        for (; i < bdata.length && i < (n + 1) * mtu; i++)
          ds.writeByte(bdata[i]);
        ds.close();
      }
      hDataObject = data;
    }
    catch (IOException e)
    {
      e.printStackTrace();
      MsgLog.log("initTDataType", e.getMessage(), TErrorList.code_failure, e, 0);
    }
  }
  /**
   * A subset of an array of KEYVALUE Strings. This corresponds to tine format
   *        CF_KEYVALUE 
   * @param data
   *          is a reference to the String array used to create and bind to the
   *          TDataType object
   */
  public TDataType(KEYVALUE[] data)
  {
    initTDataType(data, data.length);
  }
  /**
   * A subset of an array of KEYVALUE Strings. This corresponds to tine format
   *        CF_KEYVALUE
   * @param data
   *          is a reference to the KEYVALUE array used to create and bind to the
   *          TDataType object
   * @param length
   *          is the subset array length to use.
   */
  public TDataType(KEYVALUE[] data, int length)
  {
    initTDataType(data, length);
  }
  private void initTDataType(KEYVALUE[] data, int length)
  {
    if (data == null || data.length == 0 || length < 1) return;
    try
    {
      if (length > data.length) length = data.length;
      dFormat = TFormat.CF_KEYVALUE;
      dArrayLength = length;
      dSizeInBytes = 0;
      for (int i = 0; i < length; i++)
      { // first see how much string space we need (going out)
        dSizeInBytes += (data[i] != null) ? data[i].getString().length() : 0;
        dSizeInBytes++; // add terminating 0
      }
      byte[] bdata = getStringArrayBytes(data);
      initCounters();
      dBuffers = new ByteArrayOutputStream[numblks];
      dBuffersReady = new boolean[numblks];
      for (int n = 0, i = 0; n < numblks; n++)
      {
        dBuffers[n] = new ByteArrayOutputStream(dBufferSegSize);
        DataOutputStream ds = new DataOutputStream(dBuffers[n]);
        for (; i < bdata.length && i < (n + 1) * mtu; i++)
          ds.writeByte(bdata[i]);
        ds.close();
      }
      hDataObject = data;
    }
    catch (IOException e)
    {
      e.printStackTrace();
      MsgLog.log("initTDataType", e.getMessage(), TErrorList.code_failure, e, 0);
    }
  }
  /**
   * A text string
   * @param data
   *          is a reference to the String used to create and bind to the
   *          TDataType object
   */
  public TDataType(String data)
  {
    initTDataType(data, data.length());
  }
  /**
   * A text string
   * @param data
   *          is a reference to the String used to create and bind to the
   *          TDataType object
   * @param length
   *          is the subset length to use.
   */
  public TDataType(String data, int length)
  {
    initTDataType(data, length);
  }
  private void initTDataType(String data, int length)
  {
    if (data == null || length < 1) return;
    try
    {
      if (length > data.length()) length = data.length();
      dFormat = TFormat.CF_TEXT;
      dArrayLength = length;
      dSizeInBytes = dArrayLength;
      byte[] bdata = data.getBytes();
      initCounters();
      dBuffers = new ByteArrayOutputStream[numblks];
      dBuffersReady = new boolean[numblks];
      for (int n = 0, i = 0; n < numblks; n++)
      {
        dBuffers[n] = new ByteArrayOutputStream(dBufferSegSize);
        DataOutputStream ds = new DataOutputStream(dBuffers[n]);
        for (; i < bdata.length && i < (n + 1) * mtu; i++)
          ds.writeByte(bdata[i]);
        ds.close();
      }
      hDataObject = new char[dArrayLength];
      if (dArrayLength > 0) data.getChars(0, dArrayLength, (char[]) hDataObject, 0);
    }
    catch (IOException e)
    {
      e.printStackTrace();
      MsgLog.log("initTDataType", e.getMessage(), TErrorList.code_failure, e, 0);
    }
  }
  /**
   * Gets the canonical data set from the underlying data byte stream
   *        Gets the fully formed (properly assembled and byte-swapped)
   *        canonical data set from the underlying byte stream. This method is
   *        used internally, but can be used to get the underlying data into the
   *        specifically referenced data, for instance when the TDataType object
   *        was instantiated with 'unbound' data.
   * @param data
   *          is the data object into which the underlying data should be
   *          prepared and copied.
   * @return 0 on success or a tine error code
   */
  public int getData(char[] data)
  { synchronized (hDataObject) {
    int slen;
    switch (dFormat)
    {
      case TFormat.CF_XML:
      case TFormat.CF_TEXT:  
        slen = dArrayLength; 
        break;
      case TFormat.CF_NAME8:
      case TFormat.CF_NAME16:
      case TFormat.CF_NAME32:
      case TFormat.CF_NAME48:
      case TFormat.CF_NAME64:
        if (dArrayLength > 1) return TErrorList.dimension_error;
        slen = TFormat.formatSizeOf(dFormat);
        break;
      case TFormat.CF_STRING:
        if (dArrayLength > 1) return TErrorList.dimension_error;
        slen = ((String[])hDataObject)[0].length();
        break;
      default:
        return TErrorList.illegal_format;    
    }
    if (slen > data.length) return TErrorList.string_too_long;
    int i, n, m, cplen;
    if (buffersBusy) if (waitForBuffers() != 0) return TErrorList.data_stale;
    Arrays.fill(data, (char) 0);
    byte bdata[] = new byte[data.length];
    for (i=0, n=0, m=0; n<numblks && n<dBuffers.length; n++, m=i)
    {
      if (dBuffers[n] == null) break;
      byte[] b = dBuffers[n].toByteArray();
      //for (cplen=0; cplen<b.length; cplen++) if (b[cplen] == 0) break;
      cplen = b.length;
      if (cplen > data.length - m) cplen = data.length - m;
      if (cplen > 0)
      {
        System.arraycopy(b, 0, bdata, m, cplen);
        i += cplen;
      }
    }
    if (i > slen) i = slen;
    String sdata = new String(bdata, 0, i);
    sdata.getChars(0, sdata.length(), data, 0);
    dCompletionLength = i;
    isDataObjectInSync = true;
    return 0;
  } }
  /**
   * Gets the canonical data set from the underlying data byte stream
   *        Gets the fully formed (properly assembled and byte-swapped)
   *        canonical data set from the underlying byte stream. This method is
   *        used internally, but can be used to get the underlying data into the
   *        specifically referenced data, for instance when the TDataType object
   *        was instantiated with 'unbound' data.
   * @param data
   *          is the data object into which the underlying data should be
   *          prepared and copied.
   * @return 0 on success or a tine error code
   */
  public int getData(StringBuffer data)
  { synchronized (hDataObject) {
    int slen;
    switch (dFormat)
    {
      case TFormat.CF_TEXT:  
        slen = dArrayLength; 
        break;
      case TFormat.CF_NAME8:
      case TFormat.CF_NAME16:
      case TFormat.CF_NAME32:
      case TFormat.CF_NAME48:
      case TFormat.CF_NAME64:
        if (dArrayLength > 1) return TErrorList.dimension_error;
        slen = TFormat.formatSizeOf(dFormat);
        break;
      case TFormat.CF_STRING:
        if (dArrayLength > 1) return TErrorList.dimension_error;
        slen = ((String[])hDataObject)[0].length();
        break;
      default:
        return TErrorList.illegal_format;    
    }
    if (slen > data.capacity()) return TErrorList.string_too_long;
    int i, n, m;
    if (buffersBusy) if (waitForBuffers() != 0) return TErrorList.data_stale;
    byte bdata[] = new byte[data.capacity()];
    for (i = 0, n = 0, m = 0; n < numblks; n++, m = i)
    {
      if (dBuffers[n] == null) break;
      byte[] b = dBuffers[n].toByteArray();
      int cplen = b.length;
      if (cplen > data.capacity() - m) cplen = data.capacity() - m;
      if (cplen > 0)
      {
        System.arraycopy(b, 0, bdata, m, b.length);
        i += cplen;
      }
    }
    if (i > slen) i = slen;
    String sdata = new String(bdata, 0, i);
    data.replace(0, sdata.length(), sdata);
    if (hDataObject instanceof StringBuffer)
      ((StringBuffer) hDataObject).replace(0, sdata.length(), sdata);
    else if (hDataObject instanceof char[])
      System.arraycopy(data.toString().toCharArray(), 0, (char[]) hDataObject, 0, sdata.length());
    dCompletionLength = i;
    isDataObjectInSync = true;
    return 0;
  } }
  public int getData(KEYVALUE[] data)
  {
    if (data == null) return TErrorList.invalid_parameter;
    String[] sdata = new String[data.length];
    int cc = getData(sdata);
    if (cc != 0) return cc;
    for (int i=0; i<data.length; i++)
    {
      if (data[i] == null) 
        data[i] = new KEYVALUE(sdata[i]);
      else
        data[i].putString(sdata[i]); 
    }
    return 0;
  }
  /**
   * Gets the canonical data set from the underlying data byte stream
   *        Gets the fully formed (properly assembled and byte-swapped)
   *        canonical data set from the underlying byte stream. This method is
   *        used internally, but can be used to get the underlying data into the
   *        specifically referenced data, for instance when the TDataType object
   *        was instantiated with 'unbound' data.
   * @param data
   *          is the data object into which the underlying data should be
   *          prepared and copied.
   * @return 0 on success or a tine error code
   */
  public int getData(String[] data)
  {
    return getData(data, data != null ? data.length : 0, 0);
  }
  /**
   * Gets the canonical data set from the underlying data byte stream
   *        Gets the fully formed (properly assembled and byte-swapped)
   *        canonical data set from the underlying byte stream. This method is
   *        used internally, but can be used to get the underlying data into the
   *        specifically referenced data, for instance when the TDataType object
   *        was instantiated with 'unbound' data.
   * @param data
   *          is the data object into which the underlying data should be
   *          prepared and copied.
   * @param length
   *          is the number of elements to get
   * @param offset
   *          is the beginning position of the target data array into which the
   *          data should be copied
   * @return 0 on success or a tine error code
   */
  public int getData(String[] data,int length,int offset)
  { synchronized (hDataObject) {
    if (data == null) return TErrorList.argument_list_error;
    int dlen = dArrayLength;
    if (dArrayLength > data.length) dlen = data.length;
    int i, n, m, cc;
    if (length > data.length-offset) length = data.length - offset;
    switch (dFormat)
    {
      case TFormat.CF_TEXT:
        char[] s =  new char[dArrayLength];
        if ((cc=getData(s)) != 0) return cc;
        data[offset] = new String(s);
        dCompletionLength = 1;
        return 0;
      case TFormat.CF_NAME8:
      case TFormat.CF_NAME16: 
      case TFormat.CF_NAME32: 
      case TFormat.CF_NAME48: 
      case TFormat.CF_NAME64:
        NAME64[] n64 = new NAME64[dlen];
        if ((cc=getData(n64)) != 0) return cc;
        for (i = 0; i < dCompletionLength && i<length; i++) 
          data[i+offset] = n64[i].name;
        return 0;
      case TFormat.CF_USTRING:
        USTRING[] ustr = new USTRING[dlen];
        if ((cc=getData(ustr)) != 0) return cc;
        for (i = 0; i < dCompletionLength && i<length; i++) 
          data[i+offset] = ustr[i].str;
        return 0;
      case TFormat.CF_UNAME:
        UNAME[] unam = new UNAME[dlen];
        if ((cc=getData(unam)) != 0) return cc;
        for (i = 0; i < dCompletionLength && i<length; i++) 
          data[i+offset] = unam[i].str;
        return 0;
      case TFormat.CF_KEYVALUE:
      case TFormat.CF_STRING:
        if (bytesin == 0 && dSizeInBytes == 0) 
          bytesin = data.length * TFormat.STRING_STD_CAPACITY;
        break;
      default:
        return TErrorList.illegal_format;
    }
    if (bytesin > 0) dSizeInBytes = bytesin;
    if (buffersBusy) if (waitForBuffers() != 0) return TErrorList.data_stale;
    byte[] bdata = new byte[dSizeInBytes];
    for (i = 0, n = 0, m = 0; n < numblks; n++, m = i)
    {
      if (dBuffers[n] == null) break;
      byte[] b = dBuffers[n].toByteArray();
      System.arraycopy(b, 0, bdata, m, b.length);
      i += b.length;
    }
    int off = 0, len = 0;
    for (i = 0; i < dlen && i < length; i++)
    {
      for (len = 0; len + off < dSizeInBytes; len++)
      { // look for the terminating '0's ...
        if (bdata[off + len] == 0) break;
      }
      if (len >= dSizeInBytes - off) break;
      // make the strings from the underlying byte stream :
      data[i+offset] = new String(bdata, off, len);
      // double assignment in case byte stream has not yet been filled:
      if (hDataObject instanceof StringBuffer)
      {
        data[offset] = ((StringBuffer) hDataObject).toString().trim();
        break;
      }
      else if (hDataObject instanceof char[])
      {
        data[offset] = new String((char[]) hDataObject).trim();
        break;
      }
      else if (hDataObject instanceof String[])
      {
        data[i+offset] = ((String[]) hDataObject)[i];
      }
      off += len + 1;
    }
    dCompletionLength = i;
    isDataObjectInSync = true;
    return 0;
  } }
  public int getData(KEYVALUE[] data,int length,int offset)
  { synchronized (hDataObject) {
    if (data == null) return TErrorList.argument_list_error;
    int dlen = dArrayLength;
    if (dArrayLength > data.length) dlen = data.length;
    int i, n, m, cc;
    if (length > data.length-offset) length = data.length - offset;
    switch (dFormat)
    {
      case TFormat.CF_TEXT:
        char[] s =  new char[dArrayLength];
        if ((cc=getData(s)) != 0) return cc;
        data[offset] = new KEYVALUE(new String(s));
        dCompletionLength = 1;
        return 0;
      case TFormat.CF_NAME8:
      case TFormat.CF_NAME16: 
      case TFormat.CF_NAME32: 
      case TFormat.CF_NAME48: 
      case TFormat.CF_NAME64:
        NAME64[] n64 = new NAME64[dlen];
        if ((cc=getData(n64)) != 0) return cc;
        for (i = 0; i < dCompletionLength && i<length; i++) 
          data[i+offset] = new KEYVALUE(n64[i].name);
        return 0;
      case TFormat.CF_USTRING:
        USTRING[] ustr = new USTRING[dlen];
        if ((cc=getData(ustr)) != 0) return cc;
        for (i = 0; i < dCompletionLength && i<length; i++) 
          data[i+offset] = new KEYVALUE(ustr[i].str);
        return 0;
      case TFormat.CF_UNAME:
        UNAME[] unam = new UNAME[dlen];
        if ((cc=getData(unam)) != 0) return cc;
        for (i = 0; i < dCompletionLength && i<length; i++) 
          data[i+offset] = new KEYVALUE(unam[i].str);
        return 0;
      case TFormat.CF_KEYVALUE:
      case TFormat.CF_STRING:
        if (bytesin == 0 && dSizeInBytes == 0) 
          bytesin = data.length * TFormat.STRING_STD_CAPACITY;
        break;
      default:
        return TErrorList.illegal_format;
    }
    if (bytesin > 0) dSizeInBytes = bytesin;
    if (buffersBusy) if (waitForBuffers() != 0) return TErrorList.data_stale;
    byte[] bdata = new byte[dSizeInBytes];
    for (i = 0, n = 0, m = 0; n < numblks; n++, m = i)
    {
      if (dBuffers[n] == null) break;
      byte[] b = dBuffers[n].toByteArray();
      System.arraycopy(b, 0, bdata, m, b.length);
      i += b.length;
    }
    int off = 0, len = 0;
    for (i = 0; i < dlen && i < length; i++)
    {
      for (len = 0; len + off < dSizeInBytes; len++)
      { // look for the terminating '0's ...
        if (bdata[off + len] == 0) break;
      }
      if (len >= dSizeInBytes - off) break;
      // make the strings from the underlying byte stream :
      data[i+offset] = new KEYVALUE(new String(bdata, off, len));
      // double assignment in case byte stream has not yet been filled:
      if (hDataObject instanceof StringBuffer)
      {
        data[offset] = new KEYVALUE(((StringBuffer) hDataObject).toString().trim());
        break;
      }
      else if (hDataObject instanceof char[])
      {
        data[offset] = new KEYVALUE(new String((char[]) hDataObject).trim());
        break;
      }
      else if (hDataObject instanceof String[])
      {
        data[i+offset] = new KEYVALUE(((String[]) hDataObject)[i]);
      }
      else if (hDataObject instanceof KEYVALUE[])
      {
        data[i+offset] = ((KEYVALUE[])hDataObject)[i];
      }
      off += len + 1;
    }
    dCompletionLength = i;
    isDataObjectInSync = true;
    return 0;
  } }
  public void setTag(String tag)
  {
    if (dFormat == TFormat.CF_MDA)
    {
      MDA mda = (MDA)hDataObject;
      if (mda != null && mda.getCarriedFormat() == TFormat.CF_NULL)
        mda.setCarriedFormat(TFormat.getFormatCode(tag));
      int fsiz = TFormat.getCarriedFormatSize(tag);
      dSizeInBytes = dArrayLength * fsiz + TFormat.getFormatHeaderSize(dFormat);
      initCounters();
      reassignBuffers();
    }
    else
    {
      TStructDescription tsd = TStructRegistry.get(tag,structKey);
      if (tsd != null && tsd.hasExtendedSpace()) hasExtendedBytes = true;
    }
    dTag = tag;
  }
  /**
   * Returns the last data completion length.
   * @return the last data completion length. This can be different (smaller)
   *         than the data length used to construct the object.
   */
  public int getCompletionLength()
  {
    return dCompletionLength;
  }
  public void setArrayLength(int arrayLength)
  {
    setArrayLength(arrayLength, 0);
  }
  public void setArrayLength(int arrayLength,int extLength)
  {
    int dlen = arrayLength;
    if (dFormat == TFormat.CF_STRUCT)
    { // if tagged structure, then this call takes the number of structures, not
      // bytes !
      int ssiz = TStructRegistry.getSizeInBytes(dTag,dTagKey);
      if (ssiz < 1) ssiz = 1;
      dlen *= ssiz;
      dlen += extLength;
//      if (hasExtendedBytes) dSizeInBytes = dlen; //<< superseded below !!
    }
    if (dFormat == TFormat.CF_STRING)
    {
      dlen = 0;
      if (hDataObject instanceof String[])
      {
        String[] sa = (String[])hDataObject;
        for (int i=0; i<sa.length; i++)
        {
          if (sa[i] == null) break;
          dlen += sa[i].length()+1;
        }
      }
      if (dlen == 0) dlen = arrayLength * TFormat.STRING_STD_CAPACITY; 
    }
    dArrayLength = dlen;
    dSizeInBytes = dlen * TFormat.formatSizeOf(dFormat) + TFormat.getFormatHeaderSize(dFormat);
  }
  public String getField()
  {
    return dField;
  }
  public void setField(String field)
  {
    dField = field;
  }
  public TBitfield[] getBitField()
  {
    return dBitField;
  }
  public void setBitField(TBitfield[] bitfield)
  {
    dBitField = bitfield;
  }
}